# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  dfbinop.x
# 	Author	:  akeith
# 	Date	:  03/30/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) dfbinop.x -r1.3 03/30/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	04/24/95   akeith	updated to better handle INDEF's, better io docs, and now df's can be defined as delta df's
# 	04/19/94   akeith	changed binop() to zbinop() to avoid conflict
# 	03/30/94   akeith	date and time created 94/03/30 12:36:56 by akeith
# H-
# 
#


#       Version: @(#)dfbinop.x	1.3 4/24/95 UCB CEA EGO

include	"df.h"
include	<dfset.h>

#
# df_binop() computes the binary operation between the
#	two df's, a and b, and puts the result in c.
#

procedure	df_binop(a, b, c, opr, k) #I

pointer	a			#I discrete function
pointer	b			#I discrete function
pointer	c			#I discrete function
int	opr			#I operation
double	k			#I operational value

int	df_type()
pointer	tmp_c, df_tmp()

begin

	if(a == c || b == c) {
		if(df_type(a) == DF_PD) {

			if(df_type(c) == DF_PF)
				call eu_error(1, "df_binop(): ", PDPF_ERR)

			tmp_c = df_tmp(c, df_type(c))

			if(df_type(b) == DF_PD) call pdpd_binop(a, b, tmp_c, opr, k)
			else			call pdpf_binop(a, b, tmp_c, opr, k)

			call df_mov(tmp_c, c)
			call df_tmpclose(tmp_c)

		} else {
			if(df_type(b) == DF_PD) {
				if(df_type(c) == DF_PD) {
					tmp_c = df_tmp(c, df_type(c))
					call pdpf_binop(b, a, tmp_c, opr+DF_OPSWCH, k)
					call df_mov(tmp_c, c)
					call df_tmpclose(tmp_c)
				} else call eu_error(1, "df_binop(): ", PDPF_ERR)

			} else {
				tmp_c = df_tmp(c, df_type(c))
				call pfpf_binop(a, b, tmp_c, opr, k)
				call df_mov(tmp_c, c)
				call df_tmpclose(tmp_c)
			}
		}
	} else {
		if(df_type(a) == DF_PD) {
			if(df_type(c) == DF_PF)
				call eu_error(1, "df_binop(): ", PDPF_ERR)
			if(df_type(b) == DF_PD)	call pdpd_binop(a, b, c, opr, k)
			else			call pdpf_binop(a, b, c, opr, k)

		} else {
			if(df_type(b) == DF_PD) {
				if(df_type(c) == DF_PD)
					call pdpf_binop(b, a, c, opr+DF_OPSWCH, k)
				else	call eu_error(1, "df_binop(): ", PDPF_ERR)

			} else {
				if(df_type(c) == DF_PD) {
					tmp_c = df_tmp(c, df_type(a))
					call pfpf_binop(a, b, tmp_c, opr, k)
					call df_mov(tmp_c, c)
					call df_tmpclose(tmp_c)
				} else
					call pfpf_binop(a, b, c, opr, k)
			}
		}
	}

end


#
# pdpd_binop() computes the binary operation, opr, between
#	two pixel data discrete functions, a and b, and sets
#	the output discrete function, c. The resulting df
#	will be a pixel data df who's domain will
#	be the union of the initial df's.
#

procedure	pdpd_binop(a, b, c, opr, k) #I

pointer	a			#I discrete function
pointer	b			#I discrete function
pointer	c			#I discrete function
int	opr			#I operation
double	k			#I operational value

double	a_ledge, a_redge	# edge values
double	b_ledge, b_redge	# edge values
double	c_ledge, c_redge	# edge values
double	ax, ay			# discrete values
double	bx, by			# discrete values
double	cy			# discrete values

int	a_state, b_state	# get states
int	bo_dgetd()
pointer	a_bo, b_bo, bo_open()

begin

	call df_reset(c)

	call df_gedge(a, a_ledge, a_redge)
	call df_gedge(b, b_ledge, b_redge)
	call zbinop(a_ledge, b_ledge, opr, c_ledge, k)
	call zbinop(a_redge, b_redge, opr, c_redge, k)
	call df_pedge(c, c_ledge, c_redge)

	call df_gx0(a, ax)
	if(IS_INDEFD(ax)) call df_mkx0(a)
	call df_gx0(b, bx)
	if(IS_INDEFD(bx)) call df_mkx0(b)

	a_state = 1; b_state = 1
        a_bo = bo_open(a)
        b_bo = bo_open(b)

	call df_gx0(a, ax); call df_gx0(b, bx)
	if(DOUBLE_LT(ax, bx)) {
		call df_px0(c, ax)
		a_state = bo_dgetd(a_bo, ax, ay)
		by = b_ledge
	} else if(DOUBLE_EQ(ax, bx)) {
		call df_px0(c, bx)
		a_state = bo_dgetd(a_bo, ax, ay)
		b_state = bo_dgetd(b_bo, bx, by)
	} else {
		call df_px0(c, bx)
		b_state = bo_dgetd(b_bo, bx, by)
		ay = a_ledge
	}

		# Run through each and every point.

	while(a_state != 0 && b_state != 0) {

		while(DOUBLE_LT(ax, bx) && a_state != 0) {
			call zbinop(ay, by, opr, cy, k)
			call df_dputd(c, ax, cy, 1)
		        a_state = bo_dgetd(a_bo, ax, ay)
		}

		while(DOUBLE_LT(bx, ax) && b_state != 0) {
			call zbinop(ay, by, opr, cy, k)
			call df_dputd(c, bx, cy, 1)
		        b_state = bo_dgetd(b_bo, bx, by)
		}

		 while(DOUBLE_EQ(ax, bx) && (a_state != 0 || b_state != 0)) {
			call zbinop(ay, by, opr, cy, k)
			call df_dputd(c, ax, cy, 1)
		        a_state = bo_dgetd(a_bo, ax, ay)
		        b_state = bo_dgetd(b_bo, bx, by)
		}
	}

	while(a_state != 0){
		call zbinop(ay, b_redge, opr, cy, k)
		call df_dputd(c, ax, cy, 1)
		a_state = bo_dgetd(a_bo, ax, ay)
	}

	while(b_state != 0) {
		call zbinop(a_redge, by, opr, cy, k)
		call df_dputd(c, bx, cy, 1)
		b_state = bo_dgetd(b_bo, bx, by)
	}

        call bo_close(a_bo)
        call bo_close(b_bo)

	call df_rewind(a)
	call df_rewind(b)

end


#
# pfpf_binop() computes the binary operation, opr, between
#	two piecewise function discrete functions, a and b,
#	and sets the output discrete function, c. The resulting
#	df will be a piecewise function df who's domain will
#	be the union of the initial df's.
#

procedure	pfpf_binop(a, b, c, opr, k) #I

pointer	a			#I discrete function
pointer	b			#I discrete function
pointer	c			#I discrete function
int	opr			#I operation
double	k			#I operational value

double	a_ledge, a_redge	# edge values
double	b_ledge, b_redge	# edge values
double	c_ledge, c_redge	# edge values
double	a_ox, a_oy		# discrete values
double	a_nx, a_ny		# discrete values
double	b_ox, b_oy		# discrete values
double	b_nx, b_ny		# discrete values
double	a_x0, b_x0		# discrete values
double	a_gap, b_gap		# gap values
double	a_delta, b_delta	# delta values
double	a_base, b_base		# base values
double	cy
double	y, bo_interp(), pf_interp()

int	a_state, b_state	# get states
int	df_dgetd()
bool	a_is_delta, b_is_delta, df_gdelta()

begin

	call df_reset(c)

        a_is_delta = df_gdelta(a, a_gap, a_delta, a_base)
        b_is_delta = df_gdelta(b, b_gap, b_delta, b_base)

	call df_gedge(a, a_ledge, a_redge)
	call df_gedge(b, b_ledge, b_redge)
	call zbinop(a_ledge, b_ledge, opr, c_ledge, k)
	call zbinop(a_redge, b_redge, opr, c_redge, k)
	call df_pedge(c, c_ledge, c_redge)

	call df_gx0(a, a_x0)
	if(IS_INDEFD(a_x0)) call df_mkx0(a)
	call df_gx0(b, b_x0)
	if(IS_INDEFD(b_x0)) call df_mkx0(b)

	call df_gx0(a, a_x0); call df_gx0(b, b_x0);
	if(DOUBLE_LT(a_x0, b_x0)) call df_px0(c, a_x0)
	else		call df_px0(c, b_x0)

	a_state = df_dgetd(a, a_nx, a_ny, 1)
	b_state = df_dgetd(b, b_nx, b_ny, 1)
	a_x0 = a_nx; b_x0 = b_nx

	while(DOUBLE_LT(a_nx, b_x0) && a_state != 0) {
		call zbinop(a_ny, b_ledge, opr, cy, k)
		call df_dputd(c, a_nx, cy, 1)
		a_ox = a_nx; a_oy = a_ny
		a_state = df_dgetd(a, a_nx, a_ny, 1)
	}

	while(DOUBLE_LT(b_nx, a_x0) && b_state != 0) {
		call zbinop(a_ledge, b_ny, opr, cy, k)
		call df_dputd(c, b_nx, cy, 1)
		b_ox = b_nx; b_oy = b_ny
		b_state = df_dgetd(b, b_nx, b_ny, 1)
	}

		# Run through each and every point.

	while(a_state != 0 && b_state != 0) {

            if(b_is_delta)
		while(DOUBLE_LT(a_nx, b_nx) && a_state != 0) {
			y = bo_interp(b_nx, b_ny, b_ox, b_oy,
                                         b_gap, b_base, a_nx)
			call zbinop(a_ny, y, opr, cy, k)
			call df_dputd(c, a_nx, cy, 1)
			a_ox = a_nx; a_oy = a_ny
			a_state = df_dgetd(a, a_nx, a_ny, 1)
		}
            else
		while(DOUBLE_LT(a_nx, b_nx) && a_state != 0) {
			y = pf_interp(b_nx, b_ny, b_ox, b_oy, a_nx)
			call zbinop(a_ny, y, opr, cy, k)
			call df_dputd(c, a_nx, cy, 1)
			a_ox = a_nx; a_oy = a_ny
			a_state = df_dgetd(a, a_nx, a_ny, 1)
		}

            if(a_is_delta)
		while(DOUBLE_LT(b_nx, a_nx) && b_state != 0) {
			y = bo_interp(a_nx, a_ny, a_ox, a_oy,
                                         a_gap, a_base, b_nx)
			call zbinop(y, b_ny, opr, cy, k)
			call df_dputd(c, b_nx, cy, 1)
			b_ox = b_nx; b_oy = b_ny
			b_state = df_dgetd(b, b_nx, b_ny, 1)
		}
            else
		while(DOUBLE_LT(b_nx, a_nx) && b_state != 0) {
			y = pf_interp(a_nx, a_ny, a_ox, a_oy, b_nx)
			call zbinop(y, b_ny, opr, cy, k)
			call df_dputd(c, b_nx, cy, 1)
			b_ox = b_nx; b_oy = b_ny
			b_state = df_dgetd(b, b_nx, b_ny, 1)
		}

		while(DOUBLE_EQ(a_nx, b_nx) && (a_state != 0 || b_state != 0)) {
			call zbinop(a_ny, b_ny, opr, cy, k)
			call df_dputd(c, a_nx, cy, 1)
			a_ox = a_nx; a_oy = a_ny
			b_ox = b_nx; b_oy = b_ny
			a_state = df_dgetd(a, a_nx, a_ny, 1)
			b_state = df_dgetd(b, b_nx, b_ny, 1)
		}
	}

	while(a_state != 0) {
		call zbinop(a_ny, b_redge, opr, cy, k)
		call df_dputd(c, a_nx, cy, 1)
		a_state = df_dgetd(a, a_nx, a_ny, 1)
	}

	while(b_state != 0) {
		call zbinop(a_redge, b_ny, opr, cy, k)
		call df_dputd(c, b_nx, cy, 1)
		b_state = df_dgetd(b, b_nx, b_ny, 1)
	}

	call df_rewind(a)
	call df_rewind(b)

end


#
# pdpf_binop() computes the binary operation, opr, between
#	a pixel data discrete function, a, and a piecewise
#	discrete functions, b, and sets the output discrete
#	function, c. The resulting df will be a piecewise
#	function df who's domain will be the union of the
#	initial df's.
#

procedure	pdpf_binop(a, b, c, opr, k) #I

pointer	a			#I discrete function
pointer	b			#I discrete function
pointer	c			#I discrete function
int	opr			#I operation
double	k			#I operational value

double	a_ledge, a_redge	# edge values
double	b_ledge, b_redge	# edge values
double	c_ledge, c_redge	# edge values
double	ax, ay			# discrete values
double	b_ox, b_oy		# discrete values
double	b_nx, b_ny		# discrete values
double	t_oy			# discrete values
double	t_nx, t_ny		# discrete values
double	b_gap, b_delta, b_base	# gap, delta and base values

double	cy
double	y, bo_interp(), pf_interp()

int	a_state, b_state	# get states
int	df_dgetd()
int	bo_dgetd()
pointer	a_bo, bo_open()
bool	b_is_delta, df_gdelta()

begin

	call df_reset(c)

        a_bo = bo_open(a)
        b_is_delta = df_gdelta(b, b_gap, b_delta, b_base)

	call df_gedge(a, a_ledge, a_redge)
	call df_gedge(b, b_ledge, b_redge)
	call zbinop(a_ledge, b_ledge, opr, c_ledge, k)
	call zbinop(a_redge, b_redge, opr, c_redge, k)
	call df_pedge(c, c_ledge, c_redge)

	call df_gx0(a, ax)
	if(IS_INDEFD(ax)) call df_mkx0(a)
	call df_gx0(b, b_nx)
	if(IS_INDEFD(b_nx)) call df_mkx0(b)

	call df_gx0(a, ax); call df_gx0(b, b_nx)
	ay = a_ledge; b_ny = b_ledge
	a_state = 1
	b_state = df_dgetd(b, b_ox, t_oy, 1)
	b_nx = b_ox; b_oy = t_oy

	if(DOUBLE_LT(ax, b_nx)) {

		call df_px0(c, ax)
	        a_state = bo_dgetd(a_bo, ax, ay)

		while(DOUBLE_LT(ax, b_nx) && a_state != 0) {
			call zbinop(ay, b_ny, opr, cy, k)
			call df_dputd(c, ax, cy, 1)
	                a_state = bo_dgetd(a_bo, ax, ay)
		}

		call zbinop(ay, b_ny, opr, cy, k)
		call df_dputd(c, b_nx, cy, 1)
		if(DOUBLE_EQ(ax, b_ox)) a_state = bo_dgetd(a_bo, ax, ay)
		b_state = df_dgetd(b, b_nx, b_ny, 1)

	} else {

		call df_px0(c, b_nx)
		b_state = df_dgetd(b, b_nx, b_ny, 1)
	}


		# Run through each and every point.

	while(a_state != 0 && b_state != 0) {

            if(b_is_delta)
		while(DOUBLE_LT(ax, b_nx) && a_state != 0) {
			t_nx = ax
			t_ny = bo_interp(b_nx, b_ny, b_ox, b_oy,
                                         b_gap, b_base, t_nx)
			if(IS_INDEFD(t_ny) || IS_INDEFD(t_oy)) y = INDEFD
			else				       y = 0.5*(t_ny + t_oy)
			call zbinop(ay, y, opr, cy, k)
			call df_dputd(c, t_nx, cy, 1)
			t_oy = t_ny
	                a_state = bo_dgetd(a_bo, ax, ay)
		}
            else
		while(DOUBLE_LT(ax, b_nx) && a_state != 0) {
			t_nx = ax
			t_ny = pf_interp(b_nx, b_ny, b_ox, b_oy, t_nx)
			if(IS_INDEFD(t_ny) || IS_INDEFD(t_oy)) y = INDEFD
			else				       y = 0.5*(t_ny + t_oy)
			call zbinop(ay, y, opr, cy, k)
			call df_dputd(c, t_nx, cy, 1)
			t_oy = t_ny
	                a_state = bo_dgetd(a_bo, ax, ay)
		}

		while(DOUBLE_LT(b_nx, ax) && b_state != 0) {
			t_nx = b_nx; t_ny = b_ny
			if(IS_INDEFD(t_ny) || IS_INDEFD(t_oy)) y = INDEFD
			else				       y = 0.5*(t_ny + t_oy)
			call zbinop(ay, y, opr, cy, k)
			call df_dputd(c, t_nx, cy, 1)
			t_oy = t_ny
			b_ox = b_nx; b_oy = b_ny
			b_state = df_dgetd(b, b_nx, b_ny, 1)
		}

		while(DOUBLE_EQ(ax, b_nx) && (a_state != 0 || b_state != 0)) {
			t_nx = b_nx; t_ny = b_ny
			if(IS_INDEFD(t_ny) || IS_INDEFD(t_oy)) y = INDEFD
			else				       y = 0.5*(t_ny + t_oy)
			call zbinop(ay, y, opr, cy, k)
			call df_dputd(c, t_nx, cy, 1)
			t_oy = t_ny
			b_ox = b_nx; b_oy = b_ny
	                a_state = bo_dgetd(a_bo, ax, ay)
			b_state = df_dgetd(b, b_nx, b_ny, 1)
		}
	}

	while(a_state != 0) {
		call zbinop(ay, b_redge, opr, cy, k)
		call df_dputd(c, ax, cy, 1)
	        a_state = bo_dgetd(a_bo, ax, ay)
	}

	while(b_state != 0) {
		t_nx = b_nx; t_ny = b_ny
		if(IS_INDEFD(t_ny) || IS_INDEFD(t_oy)) y = INDEFD
		else				       y = 0.5*(t_ny + t_oy)
		call zbinop(a_redge, y, opr, cy, k)
		call df_dputd(c, t_nx, cy, 1)
		t_oy = t_ny
		b_state = df_dgetd(b, b_nx, b_ny, 1)
	}

        call bo_close(a_bo)

	call df_rewind(a)
	call df_rewind(b)

end


#
# zbinop() computes the binary operation between the
#	y values, a and b, and puts the result in c.
#

procedure	zbinop(a, b, opr, c, k) #I

double	a			#I a value
double	b			#I b value
double	c			#O c value
int	opr			#I operation
double	k			#I operational value

begin

	if(a == INDEFD || b == INDEFD) {
		c = INDEFD
		return
	}

	switch(opr) {
	  case DF_ADD:
		c = a + b
	  case DF_SWCHADD:
		c = a + b
	  case DF_SUB:
		c = a - b
	  case DF_SWCHSUB:
		c = b - a
	  case DF_MULT:
		c = a * b
	  case DF_SWCHMULT:
		c = a * b
	  case DF_INTERP:
		c = (b - a) * k + a
	  case DF_SWCHINTERP:
		c = (a - b) * k + b
	  case DF_DIV:
		if(DOUBLE_EQ(b, 0.0)) c = INDEFD
		else c = a / b
	  case DF_SWCHDIV:
		if(DOUBLE_EQ(a, 0.0)) c = INDEFD
		else c = b / a
	}

end

