# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  dfbinwidth.x
# 	Author	:  akeith
# 	Date	:  06/26/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) dfbinwidth.x -r1.2 06/26/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	04/24/95   akeith	updated to better handle INDEF's, better io docs, and now df's can be defined as delta df's
# 	06/26/94   akeith	date and time created 94/06/26 21:34:11 by akeith
# H-
# 
#


#	Version: @(#)dfbinwidth.x	1.2 4/24/95 UCB CEA EGO

include	"df.h"
include	<dfset.h>

#
# df_dbw() divides the range of a by its bin width and
#	places the resulting df in b. a can equal b
#	and if a is a piecewise function and b is
#	pixel data, the finial result is translated
#	using df_pfpd(). The resulting df must exist.
#

procedure	df_dbw(a, b) #E

pointer a			#I input df
pointer b			#I existing output df

pointer op, op_open()

begin
	call df_zdbw(a, op_open(op, a, b))
	call op_close(op, a, b)
end


#
# df_mbw() multiplies the range of a by its bin width and
#	places the resulting df in b. a can equal b
#	and if a is a piecewise function and b is
#	pixel data, the finial result is translated
#	using df_pfpd(). The resulting df must exist.
#

procedure	df_mbw(a, b) #E

pointer a			#I input df
pointer b			#I existing output df

pointer op, op_open()

begin
	call df_zmbw(a, op_open(op, a, b))
	call op_close(op, a, b)
end


#
# df_zdbw() performs the calculation for two distinct
#       df's of the same type.
#

procedure	df_zdbw(a, b) #I

pointer a			#I input df
pointer b			#I existing output df

double	diff
double	x0, x, y
double	gap, delta, base
double	ledge, redge

int	df_dgetd()
bool	df_gdelta()

begin

	call df_gx0(a, x0)
	if(IS_INDEFD(x0)) {
		call df_mkx0(a)
		call df_gx0(a, x0)
	}
	call df_px0(b, x0)

	call df_gedge(a, ledge, redge)
	call df_pedge(b, ledge, redge)

	if(df_gdelta(a, gap, delta, base))
            if(!IS_INDEFD(base))
		call df_pdelta(b, gap, delta, base)
        else call df_undelta(b)

	if(df_gdelta(a, gap, delta, base))
	    while(df_dgetd(a, x, y, 1) != 0) {
                    diff = x - x0
                    if(diff > delta) y = y/delta
                    else y = y/diff
		    call df_dputd(b, x, y, 1)
		    x0 = x
	    }
        else
	    while(df_dgetd(a, x, y, 1) != 0) {
		    y = y/(x - x0)
		    call df_dputd(b, x, y, 1)
		    x0 = x
	    }
end


#
# df_zmbw() performs the calculation for two distinct
#       df's of the same type.
#

procedure	df_zmbw(a, b) #I

pointer a			#I input df
pointer b			#I existing output df

double	diff
double	x0, x, y
double	ledge, redge
double	gap, delta, base

int	df_dgetd()
bool	df_gdelta()

begin

	call df_gx0(a, x0)
	if(IS_INDEFD(x0)) {
		call df_mkx0(a)
		call df_gx0(a, x0)
	}
	call df_px0(b, x0)

	call df_gedge(a, ledge, redge)
	call df_pedge(b, ledge, redge)

	if(df_gdelta(a, gap, delta, base))
            if(!IS_INDEFD(base))
		call df_pdelta(b, gap, delta, base)
        else call df_undelta(b)

	if(df_gdelta(a, gap, delta, base))
	    while(df_dgetd(a, x, y, 1) != 0) {
                    diff = x - x0
                    if(diff > delta) y = y*delta
                    else y = y*diff
		    call df_dputd(b, x, y, 1)
		    x0 = x
	    }
        else
	    while(df_dgetd(a, x, y, 1) != 0) {
		    y = y*(x - x0)
		    call df_dputd(b, x, y, 1)
		    x0 = x
	    }
end

