# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  dfconv.x
# 	Author	:  akeith
# 	Date	:  03/30/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) dfconv.x -r1.3 03/30/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	04/24/95   akeith	updated to better handle INDEF's, better io docs, and now df's can be defined as delta df's
# 	09/14/94   akeith	fixed bool comparison
# 	03/30/94   akeith	date and time created 94/03/30 12:36:57 by akeith
# H-
# 
#


#       Version: @(#)dfconv.x	1.3 4/24/95 UCB CEA EGO

include	"df.h"
include	<dfset.h>

#
# pdpd_1conv()
#

double	procedure	pdpd_1conv(a, a_shift, b, b_shift) #I

pointer	a			#I discrete function
double	a_shift			#I domain shift
pointer	b			#I discrete function
double	b_shift			#I domain shift

double	area			# area
double	c_ox			# discrete values
double	ax, ay			# discrete values
double	bx, by			# discrete values

pointer	a_bo, b_bo
pointer	bo_open()

int	a_state, b_state	# get states
int	bo_dgetd()

begin

	call df_gx0(a, ax)
	if(IS_INDEFD(ax)) call df_mkx0(a)
	call df_gx0(b, bx)
	if(IS_INDEFD(bx)) call df_mkx0(b)

	a_bo = bo_open(a)
	b_bo = bo_open(b)

	area = 0.0
	a_state = 1; b_state = 1

	call df_gx0(a, ax); call df_gx0(b, bx)
	ax = ax + a_shift; bx = bx + b_shift

	if(DOUBLE_LT(ax, bx)) {

		c_ox = bx
		while(DOUBLE_LT(ax, c_ox) && a_state != 0) {
			a_state = bo_dgetd(a_bo, ax, ay)
			ax = ax + a_shift
		}
		b_state = bo_dgetd(b_bo, bx, by)
		bx = bx + b_shift

	} else {

		c_ox = ax
		while(DOUBLE_LT(bx, c_ox) && b_state != 0) {
			b_state = bo_dgetd(b_bo, bx, by)
			bx = bx + b_shift
		}
		a_state = bo_dgetd(a_bo, ax, ay)
		ax = ax + a_shift
	}

		# Run through each and every point.

	while(a_state != 0 && b_state != 0) {

		while(DOUBLE_LT(ax, bx) && a_state != 0) {
			area = area + ay*by*(ax - c_ox)
			c_ox = ax
			a_state = bo_dgetd(a_bo, ax, ay)
			ax = ax + a_shift
		}

		while(DOUBLE_LT(bx, ax) && b_state != 0) {
			area = area + ay*by*(bx - c_ox)
			c_ox = bx
			b_state = bo_dgetd(b_bo, bx, by)
			bx = bx + b_shift
		}

		 while(DOUBLE_EQ(ax, bx) && (a_state != 0 || b_state != 0)) {
			area = area + ay*by*(ax - c_ox)
			c_ox = ax
			a_state = bo_dgetd(a_bo, ax, ay)
			ax = ax + a_shift
			b_state = bo_dgetd(b_bo, bx, by)
			bx = bx + b_shift
		}
	}

	call bo_close(a_bo)
	call bo_close(b_bo)

	call df_rewind(a)
	call df_rewind(b)

	return(area)

end


#
# pfpf_1conv()
#

double	procedure	pfpf_1conv(a, a_shift, b, b_shift) #I

pointer	a			#I discrete function
double	a_shift			#I domain shift
pointer	b			#I discrete function
double	b_shift			#I domain shift

double	area			# area
double	a_ox, a_oy		# discrete values
double	a_nx, a_ny		# discrete values
double	b_ox, b_oy		# discrete values
double	b_nx, b_ny		# discrete values
double	c_ox, c_oy		# discrete values
double	c_nx, c_ny		# discrete values
double	a_x0, b_x0		# discrete values
double	y, pf_interp(), bo_interp()
double	a_gap, b_gap
double	a_delta, b_delta
double	a_base, b_base

int	a_state, b_state	# get states
int	df_dgetd()

bool	a_is_delta, b_is_delta 
bool	df_gdelta()

begin

	area = 0.0

	call df_gx0(a, a_x0)
	if(IS_INDEFD(a_x0)) call df_mkx0(a)
	call df_gx0(b, b_x0)
	if(IS_INDEFD(b_x0)) call df_mkx0(b)

	a_is_delta = df_gdelta(a, a_gap, a_delta, a_base)
	b_is_delta = df_gdelta(b, b_gap, b_delta, b_base)

	a_state = df_dgetd(a, a_nx, a_ny, 1)
	a_nx = a_nx + a_shift
	b_state = df_dgetd(b, b_nx, b_ny, 1)
	b_nx = b_nx + b_shift
	a_x0 = a_nx; b_x0 = b_nx

	while(DOUBLE_LT(a_nx, b_x0) && a_state != 0) {
		c_ox = b_x0; c_oy = 0.0
		a_ox = a_nx; a_oy = a_ny
		a_state = df_dgetd(a, a_nx, a_ny, 1)
		a_nx = a_nx + a_shift
	}

	while(DOUBLE_LT(b_nx, a_x0) && b_state != 0) {
		c_ox = a_x0; c_oy = 0.0
		b_ox = b_nx; b_oy = b_ny
		b_state = df_dgetd(b, b_nx, b_ny, 1)
		b_nx = b_nx + b_shift
	}

		# Run through each and every point.

	while(a_state != 0 && b_state != 0) {

	    if(b_is_delta)
		while(DOUBLE_LT(a_nx, b_nx) && a_state != 0) {

			y = bo_interp(b_nx, b_ny, b_ox, b_oy, b_gap, b_base, a_nx)
			c_ny = y*a_ny; c_nx = a_nx
			area = area + (c_ny + c_oy)*(c_nx - c_ox)*0.5
			c_ox = c_nx; c_oy = c_ny

			a_ox = a_nx; a_oy = a_ny
			a_state = df_dgetd(a, a_nx, a_ny, 1)
			a_nx = a_nx + a_shift
		}
	    else
		while(DOUBLE_LT(a_nx, b_nx) && a_state != 0) {

			y = pf_interp(b_nx, b_ny, b_ox, b_oy, a_nx)
			c_ny = y*a_ny; c_nx = a_nx
			area = area + (c_ny + c_oy)*(c_nx - c_ox)*0.5
			c_ox = c_nx; c_oy = c_ny

			a_ox = a_nx; a_oy = a_ny
			a_state = df_dgetd(a, a_nx, a_ny, 1)
			a_nx = a_nx + a_shift
		}

	    if(a_is_delta)
		while(DOUBLE_LT(b_nx, a_nx) && b_state != 0) {

			y = bo_interp(a_nx, a_ny, a_ox, a_oy, a_gap, a_base, b_nx)
			c_ny = y*b_ny; c_nx = b_nx
			area = area + (c_ny + c_oy)*(c_nx - c_ox)*0.5
			c_ox = c_nx; c_oy = c_ny

			b_ox = b_nx; b_oy = b_ny
			b_state = df_dgetd(b, b_nx, b_ny, 1)
			b_nx = b_nx + b_shift
		}
	    else
		while(DOUBLE_LT(b_nx, a_nx) && b_state != 0) {

			y = pf_interp(a_nx, a_ny, a_ox, a_oy, b_nx)
			c_ny = y*b_ny; c_nx = b_nx
			area = area + (c_ny + c_oy)*(c_nx - c_ox)*0.5
			c_ox = c_nx; c_oy = c_ny

			b_ox = b_nx; b_oy = b_ny
			b_state = df_dgetd(b, b_nx, b_ny, 1)
			b_nx = b_nx + b_shift
		}

		while(DOUBLE_EQ(a_nx, b_nx) && (a_state != 0 || b_state != 0)) {

			c_ny = a_ny*b_ny; c_nx = a_nx
			area = area + (c_ny + c_oy)*(c_nx - c_ox)*0.5
			c_ox = c_nx; c_oy = c_ny

			a_ox = a_nx; a_oy = a_ny
			a_state = df_dgetd(a, a_nx, a_ny, 1)
			a_nx = a_nx + a_shift

			b_ox = b_nx; b_oy = b_ny
			b_state = df_dgetd(b, b_nx, b_ny, 1)
			b_nx = b_nx + b_shift
		}
	}

	call df_rewind(a)
	call df_rewind(b)

	return(area)

end


#
# pdpf_1conv()
#

double	procedure	pdpf_1conv(a, a_shift, b, b_shift) #I

pointer	a			#I discrete function
double	a_shift			#I domain shift
pointer	b			#I discrete function
double	b_shift			#I domain shift

double	area			# area
double	ax, ay			# discrete values
double	b_ox, b_oy		# discrete values
double	b_nx, b_ny		# discrete values
double	t_ox, t_oy		# discrete values
double	t_nx, t_ny		# discrete values
double	b_gap
double	b_delta
double	b_base

double	y, pf_interp(), bo_interp()

pointer a_bo, bo_open()
bool	b_is_delta, df_gdelta()
int	a_state, b_state	# get states
int	df_dgetd()
int	bo_dgetd()

begin

	area = 0.0

	call df_gx0(a, ax)
	if(IS_INDEFD(ax)) call df_mkx0(a)
	call df_gx0(b, b_nx)
	if(IS_INDEFD(b_nx)) call df_mkx0(b)

	a_bo = bo_open(a)
	b_is_delta = df_gdelta(b, b_gap, b_delta, b_base)

	call df_gx0(a, ax)
	ax = ax + a_shift
	b_state = df_dgetd(b, b_ox, b_oy, 1)
	b_ox = b_ox + b_shift

	a_state = 1
	if(DOUBLE_LT(ax, b_ox)) {

		t_ox = t_ox; t_ny = b_oy
		while(DOUBLE_LT(ax, b_ox) && a_state != 0) {
			a_state = bo_dgetd(a_bo, ax, ay)
			ax = ax + a_shift
		}
		b_state = df_dgetd(b, b_nx, b_ny, 1)
		b_nx = b_nx + b_shift

	} else {

		t_ox = ax
		if(b_is_delta)
                        t_oy = bo_interp(b_nx, b_ny, b_ox, b_oy, b_gap, b_base, t_nx)
		else    t_oy = pf_interp(b_nx, b_ny, b_ox, b_oy, t_nx)
		b_nx = b_ox; b_ny = b_oy
		while(DOUBLE_LT(b_nx, ax) && b_state != 0) {
			b_ox = b_nx; b_oy = b_ny
			b_state = df_dgetd(b, b_nx, b_ny, 1)
			b_nx = b_nx + b_shift
		}
		a_state = bo_dgetd(a_bo, ax, ay)
		ax = ax + a_shift
	}


		# Run through each and every point.

	while(a_state != 0 && b_state != 0) {

	    if(b_is_delta)
		while(DOUBLE_LT(ax, b_nx) && a_state != 0) {
			t_nx = ax
			t_ny = bo_interp(b_nx, b_ny, b_ox, b_oy, b_gap, b_base, t_nx)
			y = 0.5*(t_ny + t_oy)
			area = area + ay*y*(t_nx - t_ox)
			t_ox = t_nx; t_oy = t_ny
			a_state = bo_dgetd(a_bo, ax, ay)
			ax = ax + a_shift
		}
	    else
		while(DOUBLE_LT(ax, b_nx) && a_state != 0) {
			t_nx = ax
			t_ny = pf_interp(b_nx, b_ny, b_ox, b_oy, t_nx)
			y = 0.5*(t_ny + t_oy)
			area = area + ay*y*(t_nx - t_ox)
			t_ox = t_nx; t_oy = t_ny
			a_state = bo_dgetd(a_bo, ax, ay)
			ax = ax + a_shift
		}

		while(DOUBLE_LT(b_nx, ax) && b_state != 0) {
			t_nx = b_nx; t_ny = b_ny
			y = 0.5*(t_ny + t_oy)
			area = area + ay*y*(t_nx - t_ox)
			t_ox = t_nx; t_oy = t_ny
			b_ox = b_nx; b_oy = b_ny
			b_state = df_dgetd(b, b_nx, b_ny, 1)
			b_nx = b_nx + b_shift
		}

		while(DOUBLE_EQ(ax, b_nx) && (a_state != 0 || b_state != 0)) {
			t_nx = b_nx; t_ny = b_ny
			y = 0.5*(t_ny + t_oy)
			area = area + ay*y*(t_nx - t_ox)
			t_ox = t_nx; t_oy = t_ny
			b_ox = b_nx; b_oy = b_ny
			a_state = bo_dgetd(a_bo, ax, ay)
			ax = ax + a_shift
			b_state = df_dgetd(b, b_nx, b_ny, 1)
			b_nx = b_nx + b_shift
		}
	}

	call bo_close(a_bo)

	call df_rewind(a)
	call df_rewind(b)

	return(area)

end


