# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  tbldf.gx
# 	Author	:  akeith
# 	Date	:  03/30/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) tbldf.gx -r1.3 03/30/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/22/96   tnk	search for TIMDEL, then DELTIME headers (primary & alternate)
# 	05/01/95   akeith	added row number domain
# 	03/30/94   akeith	date and time created 94/03/30 12:45:38 by akeith
# H-
# 
#


#       Version: @(#)tbldf.gx	1.3 10/22/96 UCB CEA EGO

include "iotb.h"
include <dfset.h>
include <tbset.h>

#
# pf_tbl() creates a piecewise discrete function directly
#	from a ST table. The domain and range are specified
#	by the column names of the table, tb.
#

pointer	procedure	pf_tbl(tb, domain, range) #E

pointer	tb			#I table pointer
char	domain[SZ_COLNAME]	#I domain name
char	range[SZ_COLNAME]	#I range name

pointer	tbl_df()
begin
	return(tbl_df(DF_PF, tb, domain, range))
end


#
# pd_tbl() creates a pixel data discrete function directly
#	from a ST table. The domain and range are specified
#	by the column names of the table, tb.
#

pointer	procedure	pd_tbl(tb, domain, range) #E

pointer	tb			#I table pointer
char	domain[SZ_COLNAME]	#I domain name
char	range[SZ_COLNAME]	#I range name

pointer	tbl_df()
begin
	return(tbl_df(DF_PD, tb, domain, range))
end


#
# tbl_df() creates a discrete function directly from a
#	ST table. The domain and range are specified by
#	the column names of the table, tb. The type of 
#	discrete function is indicated by type.
#

pointer	procedure	tbl_df(type, tb, domain, range) #E

int	type			#I df type
pointer	tb			#I table pointer
char	domain[SZ_COLNAME]	#I domain name
char	range[SZ_COLNAME]	#I range name


pointer	rcp, dcp		# range/domain column pointers
pointer df, df_open()		# discrete function structure

int	dmn_type		# data type
int	rng_type		# data type
int	tbcigi()		# data type

double	t0, t1, del_t
double	tbhgtd()

int	tbpsta(), strlen()
bool	streq()
pointer	tbcnum()
pointer	io, iotb_open()
extern	iotb_free()


extern	iotb_gets(), iotb_puts()

extern	iotb_geti(), iotb_puti()

extern	iotb_getr(), iotb_putr()

extern	iotb_getd(), iotb_putd()


begin

	if(streq(domain, ROW_NUMBER_DOMAIN)) dcp = NULL
	else if(strlen(domain) == 0) dcp = tbcnum(tb, 1)
	else call tbcfnd(tb, domain, dcp, 1)

	if(strlen(range) != 0) call tbcfnd(tb, range, rcp, 1)
	else rcp = tbcnum(tb, 2)
	if(rcp == NULL) return(NULL)

	dmn_type = tbcigi(dcp, TBL_COL_DATATYPE)
	rng_type = tbcigi(rcp, TBL_COL_DATATYPE)

#  Open the discrete function.

	df = df_open(type, dmn_type, rng_type)

#  Set the symbolic domain.

	switch(dmn_type) {

	  case TY_SHORT:
		io = iotb_open(tb, dcp)
		call df_symdmn(df, io, iotb_free, iotb_gets, iotb_puts)

	  case TY_INT:
		io = iotb_open(tb, dcp)
		call df_symdmn(df, io, iotb_free, iotb_geti, iotb_puti)

	  case TY_REAL:
		io = iotb_open(tb, dcp)
		call df_symdmn(df, io, iotb_free, iotb_getr, iotb_putr)

	  case TY_DOUBLE:
		io = iotb_open(tb, dcp)
		call df_symdmn(df, io, iotb_free, iotb_getd, iotb_putd)

	}

# Set the symbolic range.

	switch(rng_type) {

	  case TY_SHORT:
		io = iotb_open(tb, rcp)
		call df_symrng(df, io, iotb_free, iotb_gets, iotb_puts)

	  case TY_INT:
		io = iotb_open(tb, rcp)
		call df_symrng(df, io, iotb_free, iotb_geti, iotb_puti)

	  case TY_REAL:
		io = iotb_open(tb, rcp)
		call df_symrng(df, io, iotb_free, iotb_getr, iotb_putr)

	  case TY_DOUBLE:
		io = iotb_open(tb, rcp)
		call df_symrng(df, io, iotb_free, iotb_getd, iotb_putd)

	}

# Set the x0.

	if(tbpsta(tb, TBL_NROWS) > 0) {
		if(type == DF_PD) {
			# look for the primary, then alternate deltime header
			ifnoerr( del_t = tbhgtd(tb, PDELT_HEADER_NAME) ) {
				call tbegtd(tb, dcp, 1, t0)
				call df_px0(df, (t0 - del_t))
			} else {
			   ifnoerr( del_t = tbhgtd(tb, ADELT_HEADER_NAME) ) {
				call tbegtd(tb, dcp, 1, t0)
				call df_px0(df, (t0 - del_t))
			   } else {
				# no header, use the first two rows
				call tbegtd(tb, dcp, 1, t0)
				call tbegtd(tb, dcp, 2, t1)
				call df_px0(df, (t0 - (t1 - t0)))
			   }
			}
		}
	}

	return(df)
end


#
# iotb_open()
#

pointer	procedure	iotb_open(tp, cp) #I

pointer tp			#I table pointer
pointer cp			#I column pointer

pointer	io			# input/output pointer

begin

	call iotb_malloc(io)
	IOTB_TP(io) = tp
	IOTB_CP(io) = cp

	return(io)
end




#
# iotb_get[sird]()
#

procedure	iotb_gets(io, buf, strt, nelms) #I

pointer	io			#O input/output pointer
short	buf[nelms]		#O buffer
int	strt			#I starting position
int	nelms			#O number of elms to get

int	i
int	last
pointer tp, cp			# table and column pointer
pointer sp			# stack pointer
pointer nflg			# null data buffer

int	tnelms, tbpsta()

begin

	tp = IOTB_TP(io)
	cp = IOTB_CP(io)
	tnelms = tbpsta(tp, TBL_NROWS)

	if(strt == EOF || strt > tnelms) {
		nelms = 0
		return
	}

	last = strt + nelms - 1
	if(last > tnelms) last = tnelms
	nelms = last - strt + 1

	if(cp != NULL) {
		call smark(sp)
		call salloc(nflg, nelms, TY_BOOL)

		call tbcgts(tp, cp, buf, Memb[nflg], strt, last)

		call sfree(sp)

	} else {
		do i = strt, last
		     buf[i-strt-1] = i
	}

end


#
# iotb_get[sird]()
#

procedure	iotb_geti(io, buf, strt, nelms) #I

pointer	io			#O input/output pointer
int	buf[nelms]		#O buffer
int	strt			#I starting position
int	nelms			#O number of elms to get

int	i
int	last
pointer tp, cp			# table and column pointer
pointer sp			# stack pointer
pointer nflg			# null data buffer

int	tnelms, tbpsta()

begin

	tp = IOTB_TP(io)
	cp = IOTB_CP(io)
	tnelms = tbpsta(tp, TBL_NROWS)

	if(strt == EOF || strt > tnelms) {
		nelms = 0
		return
	}

	last = strt + nelms - 1
	if(last > tnelms) last = tnelms
	nelms = last - strt + 1

	if(cp != NULL) {
		call smark(sp)
		call salloc(nflg, nelms, TY_BOOL)

		call tbcgti(tp, cp, buf, Memb[nflg], strt, last)

		call sfree(sp)

	} else {
		do i = strt, last
		     buf[i-strt-1] = i
	}

end


#
# iotb_get[sird]()
#

procedure	iotb_getr(io, buf, strt, nelms) #I

pointer	io			#O input/output pointer
real	buf[nelms]		#O buffer
int	strt			#I starting position
int	nelms			#O number of elms to get

int	i
int	last
pointer tp, cp			# table and column pointer
pointer sp			# stack pointer
pointer nflg			# null data buffer

int	tnelms, tbpsta()

begin

	tp = IOTB_TP(io)
	cp = IOTB_CP(io)
	tnelms = tbpsta(tp, TBL_NROWS)

	if(strt == EOF || strt > tnelms) {
		nelms = 0
		return
	}

	last = strt + nelms - 1
	if(last > tnelms) last = tnelms
	nelms = last - strt + 1

	if(cp != NULL) {
		call smark(sp)
		call salloc(nflg, nelms, TY_BOOL)

		call tbcgtr(tp, cp, buf, Memb[nflg], strt, last)

		call sfree(sp)

	} else {
		do i = strt, last
		     buf[i-strt-1] = i
	}

end


#
# iotb_get[sird]()
#

procedure	iotb_getd(io, buf, strt, nelms) #I

pointer	io			#O input/output pointer
double	buf[nelms]		#O buffer
int	strt			#I starting position
int	nelms			#O number of elms to get

int	i
int	last
pointer tp, cp			# table and column pointer
pointer sp			# stack pointer
pointer nflg			# null data buffer

int	tnelms, tbpsta()

begin

	tp = IOTB_TP(io)
	cp = IOTB_CP(io)
	tnelms = tbpsta(tp, TBL_NROWS)

	if(strt == EOF || strt > tnelms) {
		nelms = 0
		return
	}

	last = strt + nelms - 1
	if(last > tnelms) last = tnelms
	nelms = last - strt + 1

	if(cp != NULL) {
		call smark(sp)
		call salloc(nflg, nelms, TY_BOOL)

		call tbcgtd(tp, cp, buf, Memb[nflg], strt, last)

		call sfree(sp)

	} else {
		do i = strt, last
		     buf[i-strt-1] = i
	}

end





#
# iotb_put[sird]()
#

procedure	iotb_puts(io, buf, strt, nelms) #I

pointer	io			#O input/output pointer
short	buf[nelms]		#O buffer
int	strt			#I starting position
int	nelms			#O number of elms to get

int	tbpsta()

begin

	if(strt == EOF) strt = tbpsta(IOTB_TP(io), TBL_NROWS) + 1

	if(IOTB_CP(io) != NULL)
	      call tbcpts(IOTB_TP(io), IOTB_CP(io), buf, strt, strt+nelms-1)

end


#
# iotb_put[sird]()
#

procedure	iotb_puti(io, buf, strt, nelms) #I

pointer	io			#O input/output pointer
int	buf[nelms]		#O buffer
int	strt			#I starting position
int	nelms			#O number of elms to get

int	tbpsta()

begin

	if(strt == EOF) strt = tbpsta(IOTB_TP(io), TBL_NROWS) + 1

	if(IOTB_CP(io) != NULL)
	      call tbcpti(IOTB_TP(io), IOTB_CP(io), buf, strt, strt+nelms-1)

end


#
# iotb_put[sird]()
#

procedure	iotb_putr(io, buf, strt, nelms) #I

pointer	io			#O input/output pointer
real	buf[nelms]		#O buffer
int	strt			#I starting position
int	nelms			#O number of elms to get

int	tbpsta()

begin

	if(strt == EOF) strt = tbpsta(IOTB_TP(io), TBL_NROWS) + 1

	if(IOTB_CP(io) != NULL)
	      call tbcptr(IOTB_TP(io), IOTB_CP(io), buf, strt, strt+nelms-1)

end


#
# iotb_put[sird]()
#

procedure	iotb_putd(io, buf, strt, nelms) #I

pointer	io			#O input/output pointer
double	buf[nelms]		#O buffer
int	strt			#I starting position
int	nelms			#O number of elms to get

int	tbpsta()

begin

	if(strt == EOF) strt = tbpsta(IOTB_TP(io), TBL_NROWS) + 1

	if(IOTB_CP(io) != NULL)
	      call tbcptd(IOTB_TP(io), IOTB_CP(io), buf, strt, strt+nelms-1)

end


