# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  wimgdf.gx
# 	Author	:  akeith
# 	Date	:  09/14/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) wimgdf.gx -r1.5 09/14/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	11/02/94   akeith	added error rebinning
# 	11/02/94   akeith	rearanged
# 	11/01/94   akeith	fixed generic error on img buf
# 	10/25/94   akeith	documentation
# 	09/14/94   akeith	date and time created 94/09/14 16:25:45 by akeith
# H-
# 
#


#	Version: @(#)wimgdf.gx	1.5 11/2/94 UCB CEA EGO

include	<dfset.h>
include	<imhdr.h>

define	PF_NORMAL	-1
define	PD_NORMAL	1
define	PD_RAW		2
define	PD_ERRDATA	3

#
# pd_wimg() writes a pd type df to an image line and band.
#	The df is domain mapped to the mwcs of the image.
#

procedure	pd_wimg(im, line, band, df) #E

pointer	im			#I image
int	line			#I line
int	band			#I band
pointer	df			#I discrete function
extern	null_func

begin
	call df_wimg(PD_NORMAL, im, line, band, df, null_func)
end


#
# pd_wrim() writes a pd type df to an image line and band.
#	The df is domain mapped to the mwcs of the image.
#

procedure	pd_wrim(im, line, band, df) #E

pointer	im			#I image
int	line			#I line
int	band			#I band
pointer	df			#I discrete function
extern	null_func()

begin
	call df_wimg(PD_RAW, im, line, band, df, null_func)
end


#
# pd_weim() writes a pd type df to an image line and band.
#	The df is domain mapped to the mwcs of the image.
#

procedure	pd_weim(im, line, band, df, sqrt_func) #E

pointer	im			#I image
int	line			#I line
int	band			#I band
pointer	df			#I discrete function
extern	sqrt_func()		#I 

begin
	call df_wimg(PD_ERRDATA, im, line, band, df, sqrt_func)
end


#
# pf_wimg() writes a pf type df to an image line and band.
#	The df is domain mapped to the mwcs of the image.
#

procedure	pf_wimg(im, line, band, df) #E

pointer	im			#I image
int	line			#I line
int	band			#I band
pointer	df			#I discrete image
extern	null_func()
begin
	call df_wimg(PF_NORMAL, im, line, band, df, null_func)
end


#
# df_wimg() writes a type of df to an image line and band.
#	The df is domain mapped to the mwcs of the image.
#

procedure	df_wimg(type, im, line, band, df, func) #E

int	type			#I write type
pointer	im			#I image
int	line			#I line
int	band			#I band
pointer	df			#I discrete function
extern	func			#I generic user function

int	df_type			# df type
long	lp[3]			# line position
pointer	buf			#
pointer	mw, mw_openim()		#


pointer	impl1s()		#
pointer	impl2s()		#
pointer	impl3s()		#

pointer	impl1i()		#
pointer	impl2i()		#
pointer	impl3i()		#

pointer	impl1l()		#
pointer	impl2l()		#
pointer	impl3l()		#

pointer	impl1r()		#
pointer	impl2r()		#
pointer	impl3r()		#

pointer	impl1d()		#
pointer	impl2d()		#
pointer	impl3d()		#


begin

	if(type > 0) df_type = DF_PD
	else	df_type = DF_PF

	lp[2] = line
	lp[3] = band
	mw = mw_openim(im)
	switch(IM_PIXTYPE(im)) {

	  case TY_SHORT:
		switch(IM_NDIM(im)) {
		  case 1: buf = impl1s(im)
		  case 2: buf = impl2s(im, line)
		  case 3: buf = impl3s(im, line, band)
		}
		call df_2imgs(df_type, type, buf, mw, lp,
			IM_NDIM(im), IM_LEN(im, 1), df, func)

	  case TY_INT:
		switch(IM_NDIM(im)) {
		  case 1: buf = impl1i(im)
		  case 2: buf = impl2i(im, line)
		  case 3: buf = impl3i(im, line, band)
		}
		call df_2imgi(df_type, type, buf, mw, lp,
			IM_NDIM(im), IM_LEN(im, 1), df, func)

	  case TY_LONG:
		switch(IM_NDIM(im)) {
		  case 1: buf = impl1l(im)
		  case 2: buf = impl2l(im, line)
		  case 3: buf = impl3l(im, line, band)
		}
		call df_2imgl(df_type, type, buf, mw, lp,
			IM_NDIM(im), IM_LEN(im, 1), df, func)

	  case TY_REAL:
		switch(IM_NDIM(im)) {
		  case 1: buf = impl1r(im)
		  case 2: buf = impl2r(im, line)
		  case 3: buf = impl3r(im, line, band)
		}
		call df_2imgr(df_type, type, buf, mw, lp,
			IM_NDIM(im), IM_LEN(im, 1), df, func)

	  case TY_DOUBLE:
		switch(IM_NDIM(im)) {
		  case 1: buf = impl1d(im)
		  case 2: buf = impl2d(im, line)
		  case 3: buf = impl3d(im, line, band)
		}
		call df_2imgd(df_type, type, buf, mw, lp,
			IM_NDIM(im), IM_LEN(im, 1), df, func)

	}
	call mw_close(mw)

end



#
# df_2img$t()
#

procedure	df_2imgs(df_type, type, buf, mw, lp, ndim, npix, df, func) #I

int	df_type			#I df type
int	type			#I type of write
pointer	buf			#I image buffer
pointer	mw			#I mwcs of image
long	lp[ndim]		#I line position
int	ndim			#I number of dimensions
int	npix			#I number of pixels
pointer	df			#I discrete function
extern	func			#I generic user function

pointer	sp, x			#

pointer df_b, df_open()		#
int	npts, df_dgets()	#

begin

	call smark(sp)
	call salloc(x, npix, TY_DOUBLE)

	df_b = df_open(df_type, TY_DOUBLE, TY_SHORT)

	if(type == PD_ERRDATA)
		call df_ctdmnerr(df, mw, "world", 1, lp, ndim, npix, df_b)
	else	call df_ctdomain(df, mw, "world", 1, lp, ndim, npix, df_b)

	if(type == PD_NORMAL)
		call df_mbw(df_b, df_b)

	if(type == PD_ERRDATA)
		call df_sqrt(df_b, df_b, func)

	npts = df_dgets(df_b, Memd[x], Mems[buf], npix)
	call df_close(df_b)

	call sfree(sp)

end


#
# df_2img$t()
#

procedure	df_2imgi(df_type, type, buf, mw, lp, ndim, npix, df, func) #I

int	df_type			#I df type
int	type			#I type of write
pointer	buf			#I image buffer
pointer	mw			#I mwcs of image
long	lp[ndim]		#I line position
int	ndim			#I number of dimensions
int	npix			#I number of pixels
pointer	df			#I discrete function
extern	func			#I generic user function

pointer	sp, x			#

pointer df_b, df_open()		#
int	npts, df_dgeti()	#

begin

	call smark(sp)
	call salloc(x, npix, TY_DOUBLE)

	df_b = df_open(df_type, TY_DOUBLE, TY_INT)

	if(type == PD_ERRDATA)
		call df_ctdmnerr(df, mw, "world", 1, lp, ndim, npix, df_b)
	else	call df_ctdomain(df, mw, "world", 1, lp, ndim, npix, df_b)

	if(type == PD_NORMAL)
		call df_mbw(df_b, df_b)

	if(type == PD_ERRDATA)
		call df_sqrt(df_b, df_b, func)

	npts = df_dgeti(df_b, Memd[x], Memi[buf], npix)
	call df_close(df_b)

	call sfree(sp)

end


#
# df_2img$t()
#

procedure	df_2imgl(df_type, type, buf, mw, lp, ndim, npix, df, func) #I

int	df_type			#I df type
int	type			#I type of write
pointer	buf			#I image buffer
pointer	mw			#I mwcs of image
long	lp[ndim]		#I line position
int	ndim			#I number of dimensions
int	npix			#I number of pixels
pointer	df			#I discrete function
extern	func			#I generic user function

pointer	sp, x			#

pointer df_b, df_open()		#
int	npts, df_dgetl()	#

begin

	call smark(sp)
	call salloc(x, npix, TY_DOUBLE)

	df_b = df_open(df_type, TY_DOUBLE, TY_LONG)

	if(type == PD_ERRDATA)
		call df_ctdmnerr(df, mw, "world", 1, lp, ndim, npix, df_b)
	else	call df_ctdomain(df, mw, "world", 1, lp, ndim, npix, df_b)

	if(type == PD_NORMAL)
		call df_mbw(df_b, df_b)

	if(type == PD_ERRDATA)
		call df_sqrt(df_b, df_b, func)

	npts = df_dgetl(df_b, Memd[x], Meml[buf], npix)
	call df_close(df_b)

	call sfree(sp)

end


#
# df_2img$t()
#

procedure	df_2imgr(df_type, type, buf, mw, lp, ndim, npix, df, func) #I

int	df_type			#I df type
int	type			#I type of write
pointer	buf			#I image buffer
pointer	mw			#I mwcs of image
long	lp[ndim]		#I line position
int	ndim			#I number of dimensions
int	npix			#I number of pixels
pointer	df			#I discrete function
extern	func			#I generic user function

pointer	sp, x			#

pointer df_b, df_open()		#
int	npts, df_dgetr()	#

begin

	call smark(sp)
	call salloc(x, npix, TY_DOUBLE)

	df_b = df_open(df_type, TY_DOUBLE, TY_REAL)

	if(type == PD_ERRDATA)
		call df_ctdmnerr(df, mw, "world", 1, lp, ndim, npix, df_b)
	else	call df_ctdomain(df, mw, "world", 1, lp, ndim, npix, df_b)

	if(type == PD_NORMAL)
		call df_mbw(df_b, df_b)

	if(type == PD_ERRDATA)
		call df_sqrt(df_b, df_b, func)

	npts = df_dgetr(df_b, Memd[x], Memr[buf], npix)
	call df_close(df_b)

	call sfree(sp)

end


#
# df_2img$t()
#

procedure	df_2imgd(df_type, type, buf, mw, lp, ndim, npix, df, func) #I

int	df_type			#I df type
int	type			#I type of write
pointer	buf			#I image buffer
pointer	mw			#I mwcs of image
long	lp[ndim]		#I line position
int	ndim			#I number of dimensions
int	npix			#I number of pixels
pointer	df			#I discrete function
extern	func			#I generic user function

pointer	sp, x			#

pointer df_b, df_open()		#
int	npts, df_dgetd()	#

begin

	call smark(sp)
	call salloc(x, npix, TY_DOUBLE)

	df_b = df_open(df_type, TY_DOUBLE, TY_DOUBLE)

	if(type == PD_ERRDATA)
		call df_ctdmnerr(df, mw, "world", 1, lp, ndim, npix, df_b)
	else	call df_ctdomain(df, mw, "world", 1, lp, ndim, npix, df_b)

	if(type == PD_NORMAL)
		call df_mbw(df_b, df_b)

	if(type == PD_ERRDATA)
		call df_sqrt(df_b, df_b, func)

	npts = df_dgetd(df_b, Memd[x], Memd[buf], npix)
	call df_close(df_b)

	call sfree(sp)

end



#
# null_func
#

procedure	null_func()
begin
end

