# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  dftrans.x
# 	Author	:  akeith
# 	Date	:  03/30/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) dftrans.x -r1.2 03/30/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	04/24/95   akeith	updated to better handle INDEF's, better io docs, and now df's can be defined as delta df's
# 	03/30/94   akeith	date and time created 94/03/30 12:37:26 by akeith
# H-
# 
#


#	Version: @(#)dftrans.x	1.2 4/24/95 UCB CEA EGO

include	"df.h"
include	<dfset.h>
include	<mach.h>

define	DF_RSD		1
define	DF_CTD		2
define	DF_DFD		3

#
# df_rsdomain() changes the domain of a to a regularly
#	spaced domain specified by x0, delta x and the
#	number of elements. The results are placed in
#	b. a and b may be the same df. If a is a pf type
#	df and b is a pd type df, the domain of a is first
#	combined with the specified domain. Df a with its
#	combined domain is translated into a pd and then
#	areas are combined to the  specified domain to
#	conserve information. The resulting df must exist.
#

procedure	df_rsdomain(a, x0, delx, nelms, b) #E

pointer	a			#I input df
double	x0			#I left edge
double	delx			#I step size
int	nelms			#I number of elements
pointer	b			#I output df

int	df_type()
pointer op, tmp, op_open()	#

begin
	tmp = op_open(op, a, b)
	if(df_type(a) == DF_PF)
		call pf_zrsd(a, x0, delx, nelms, tmp)
	else	call pd_zrsd(a, x0, delx, nelms, tmp)
	call op_close(op, a, b)
end


#
# df_xdomain() changes the domain of a to a domain specified
#	by a coordinate transformation from a buffer and the
#	number of elements. The results are placed in
#	b. a and b may be the same df. If a is a pf type
#	df and b is a pd type df, the domain of a is first
#	combined with the specified domain. Df a with its
#	combined domain is translated into a pd and then
#	areas are combined to the  specified domain to
#	conserve information. The resulting df must exist.
#

procedure	df_xdomain(a, x0, xvals, nelms, b) #E

pointer	a			#I input df
double	x0			#I x0
double	xvals[nelms]		#I coordinate transformation
int	nelms			#I number of elements
pointer	b			#I output df

int	df_type()
pointer op, tmp, op_open()	#

begin
	tmp = op_open(op, a, b)
	if(df_type(a) == DF_PF)
		call pf_zxd(a, x0, xvals, nelms, tmp)
	else	call pd_zxd(a, x0, xvals, nelms, tmp)
	call op_close(op, a, b)
end


#
# df_dfdomain() changes the domain of a to a domain specified
#	by the domain of c and its number of elements. The results
#	are placed in b. a and b may be the same df. If a is a pf type
#	df and b is a pd type df, the domain of a is first
#	combined with the specified domain. Df a with its
#	combined domain is translated into a pd and then
#	areas are combined to the  specified domain to
#	conserve information. The resulting df must exist.
#

procedure	df_dfdomain(a, c, b) #E

pointer	a			#I input df
pointer	c			#I match df
pointer	b			#I output df

int	df_type()
pointer op, tmp, op_open()	#

begin
	tmp = op_open(op, a, b)
	if(df_type(a) == DF_PF)
		call pf_zdfd(a, c, tmp)
	else	call pd_zdfd(a, c, tmp)
	call op_close(op, a, b)
end


#
# df_ctdomain() changes the domain of a to a domain specified
#	by the domain of a wcs and its number of elements. The results
#	are placed in b. a and b may be the same df. If a is a pf type
#	df and b is a pd type df, the domain of a is first
#	combined with the specified domain. Df a with its
#	combined domain is translated into a pd and then
#	areas are combined to the  specified domain to
#	conserve information. The resulting df must exist.
#

procedure	df_ctdomain(a, mw, sys, axis, lp, ndim, npts, b) #E

pointer	a			#I input df
pointer mw			#I mwcs
char	sys[ARB]		#I coord. sys.
int	axis			#I axis (1 - 7)
long	lp[ndim]		#I line position
int	ndim			#I number of dimensions
int	npts			#I number of points
pointer	b			#I output df

int	i, p, l			#
int	pdim			#
int	paxis			#
double	x0			#
pointer	ct			#
pointer	sp, x			#
pointer	lx, wx			#
pointer	axno, axval		#
int	df_type()		#

begin

	call smark(sp)
	call salloc(x, npts, TY_DOUBLE)

	call wcs_gct(mw, sys, FALSE, ct, axno, axval, pdim)

	call salloc(lx, pdim, TY_DOUBLE)
	call salloc(wx, pdim, TY_DOUBLE)

	do p = 1, pdim {
		l = Memi[axno+p-1]
		if(l == 0) Memd[lx+p-1] = Memi[axval+p-1] + 1
		else if(l == axis) paxis = p
		else Memd[lx+p-1] = lp[l]
	}

	if(df_type(a) == DF_PD) {

		Memd[lx+paxis-1] = 0.5
		call mw_ctrand(ct, Memd[lx], Memd[wx], pdim)
		x0 = Memd[wx+paxis-1]

		do i = 1, npts {
			Memd[lx+paxis-1] = i + 0.5
			call mw_ctrand(ct, Memd[lx], Memd[wx], pdim)
			Memd[x+i-1] = Memd[wx+paxis-1]
		}

		call df_xdomain(a, x0, Memd[x], npts, b) #E

	} else {

		do i = 1, npts {
			Memd[lx+paxis-1] = i
			call mw_ctrand(mw, Memd[lx], Memd[wx], pdim)
			Memd[x+i-1] = Memd[wx+paxis-1]
		}
		x0 = Memd[x]
		call df_xdomain(a, x0, Memd[x], npts, b) #E
	}

	call sfree(sp)
	call mw_ctfree(ct)

end



#
# pd_zrsd()
#

procedure	pd_zrsd(a, x0, delx, nelms, b) #I

pointer	a			#I input df
double	x0			#I left edge
double	delx			#I step size
int	nelms			#I number of elements
pointer	b			#I output df

int	i			# counter
int	state			# get state

double	a_x0			# x0 of a
double	a_ledge, a_redge	# left and right edge
double	a_nx, a_ox, a_ny	# discrete values
double	xn, xo, tx		# discrete values
double	area, y			# area and value
double	gap, delta, base

pointer	a_bo, bo_open()
bool	df_gdelta()
int	bo_dgetd()

begin


	call df_gx0(a, a_x0)
	if(IS_INDEFD(a_x0)) call df_mkx0(a)

	a_bo = bo_open(a)

#  Get and Set x0

	call df_gx0(a, a_x0)
	call df_px0(b, x0)

#  Get and Set the edges

	call df_gedge(a, a_ledge, a_redge)
	call df_pedge(b, a_ledge, a_redge)

#  Get and Set delta values

	if(df_gdelta(a, gap, delta, base))
	      call df_pdelta(b, gap, delta, base)
	else  call df_undelta(b)

#  Do the Calculation

	a_ox = a_x0
	state = bo_dgetd(a_bo, a_nx, a_ny)

	i = 1
	xo = x0
	xn = delx + x0

	while(DOUBLE_LE(xn,a_x0) && i <= nelms) {
		call df_dputd(b, xn, a_ledge, 1)
		xo = xn
		i = i + 1
		xn = i*delx + x0
	}

	while(DOUBLE_LE(a_nx,x0) && state != 0) {
		a_ox = a_nx
		state = bo_dgetd(a_bo, a_nx, a_ny)
	}

	while(state != 0 && i <= nelms) {

		area = 0.0
		tx = xo

		while(DOUBLE_LE(a_nx,xn)) {
			area = area + (a_nx - tx)*a_ny
			a_ox = a_nx
			tx = a_nx
			state = bo_dgetd(a_bo, a_nx, a_ny)
			if(state == 0) {
				a_nx = MAX_DOUBLE
				a_ny = a_redge
			}
		}

		area = area + (xn - tx)*a_ny
		y = area/(xn - xo)
		call df_dputd(b, xn, y, 1)

		xo = xn
		i = i + 1
		xn = i*delx + x0

		while(DOUBLE_LE(xn,a_nx) && i <= nelms) {
			call df_dputd(b, xn, a_ny, 1)
			xo = xn
			i = i + 1
			xn = i*delx + x0
		}
	}

	call bo_close(a_bo)

end


#
# pf_zrsd()
#

procedure	pf_zrsd(a, x0, delx, nelms, b) #I

pointer	a			#I input df
double	x0			#I left edge
double	delx			#I step size
int	nelms			#I number of elements
pointer	b			#I output df

int	i			#
int	state			#
double	x, y			#
double	xo, yo			#
double	xn, yn			#
double	a_ledge, a_redge	#
double	a_gap, a_delta, a_base

int	df_dgetd()
bool	a_is_delta, df_gdelta()
double	pf_interp()
double	bo_interp()

begin

#  Get and Set x0

	i = 1
	x = x0
	call df_px0(b, x0)

#  Get and Set the edges

	call df_gedge(a, a_ledge, a_redge)
	call df_pedge(b, a_ledge, a_redge)

#  Get and Set delta values

	a_is_delta = df_gdelta(a, a_gap, a_delta, a_base)
	if(a_is_delta)
	      call df_pdelta(b, a_gap, a_delta, a_base)
	else  call df_undelta(b)

#  Do the Calculation

	state = df_dgetd(a, xo, yo, 1)

	while(DOUBLE_LE(x,xo) && i <= nelms) {
		call df_dputd(b, x, a_ledge, 1)
		i = i + 1
		x = (i-1)*delx + x0
	}

	while(df_dgetd(a, xn, yn, 1) != 0 && i <= nelms) {
	     if(a_is_delta)
		while(DOUBLE_LE(x,xn) && i <= nelms) {
			y = bo_interp(xn, yn, xo, yo, a_gap, a_base, x)
			call df_dputd(b, x, y, 1)
			i = i + 1
			x = (i-1)*delx + x0
		}
	     else
		while(DOUBLE_LE(x,xn) && i <= nelms) {
			y = pf_interp(xn, yn, xo, yo, x)
			call df_dputd(b, x, y, 1)
			i = i + 1
			x = (i-1)*delx + x0
		}
		xo = xn; yo = yn
	}

	while(i <= nelms) {
		call df_dputd(b, x, a_redge, 1)
		i = i + 1
		x = (i-1)*delx + x0
	}

end


#
# pd_zxd()
#

procedure	pd_zxd(a, x0, xvals, nelms, b) #I

pointer	a			#I input df
double	x0			#I x0
double	xvals[nelms]		#I coordinate transformation
int	nelms			#I number of elements
pointer	b			#I output df

int	i			# counter
int	state			# get state

double	a_x0			# x0 of a
double	a_ledge, a_redge	# left and right edge
double	a_nx, a_ox, a_ny	# discrete values
double	xn, xo, tx		# discrete values
double	area, y			# area and value
double	gap, delta, base

int	bo_dgetd()
bool	df_gdelta()
pointer	a_bo, bo_open()

begin

	call df_gx0(a, a_x0)
	if(IS_INDEFD(a_x0)) call df_mkx0(a)

	a_bo = bo_open(a)

#  Get and Set x0

	call df_gx0(a, a_x0)
	call df_px0(b, x0)

#  Get and Set the edges

	call df_gedge(a, a_ledge, a_redge)
	call df_pedge(b, a_ledge, a_redge)

#  Get and Set delta values

	if(df_gdelta(a, gap, delta, base))
	      call df_pdelta(b, gap, delta, base)
	else  call df_undelta(b)

#  Do the Calculation

	a_ox = a_x0
	state = bo_dgetd(a_bo, a_nx, a_ny)

	i = 1
	xo = x0
	xn = xvals[i]

	while(DOUBLE_LE(xn,a_x0) && i <= nelms) {
		call df_dputd(b, xn, a_ledge, 1)
		xo = xn
		i = i + 1
		xn = xvals[i]
	}

	while(DOUBLE_LE(a_nx,x0) && state != 0) {
		a_ox = a_nx
		state = bo_dgetd(a_bo, a_nx, a_ny)
	}

	while(state != 0 && i <= nelms) {

		area = 0.0
		tx = xo

		while(DOUBLE_LE(a_nx,xn)) {
			area = area + (a_nx - tx)*a_ny
			a_ox = a_nx
			tx = a_nx
			state = bo_dgetd(a_bo, a_nx, a_ny)
			if(state == 0) {
				a_nx = MAX_DOUBLE
				a_ny = a_redge
			}
		}

		area = area + (xn - tx)*a_ny
		y = area/(xn - xo)
		call df_dputd(b, xn, y, 1)

		xo = xn
		i = i + 1
		xn = xvals[i]

		while(DOUBLE_LE(xn,a_nx) && i <= nelms) {
			call df_dputd(b, xn, a_ny, 1)
			xo = xn
			i = i + 1
			xn = xvals[i]
		}
	}

	call bo_close(a_bo)

end


#
# pf_zxd()
#

procedure	pf_zxd(a, x0, xvals, nelms, b) #I

pointer	a			#I input df
double	x0			#I x0
double	xvals[nelms]		#I coord. trans.
int	nelms			#I number of elements
pointer	b			#I output df

int	i			#
int	state			#
double	y, x 			#
double	xo, yo			#
double	xn, yn			#
double	a_ledge, a_redge	#
double	a_gap, a_delta, a_base	#

int	df_dgetd()
bool	a_is_delta, df_gdelta()
double	pf_interp()
double	bo_interp()

begin

#  Get and Set x0

	i = 1
	x = xvals[i]
	call df_px0(b, x)

#  Get and Set the edges

	call df_gedge(a, a_ledge, a_redge)
	call df_pedge(b, a_ledge, a_redge)

#  Get and Set delta values

	a_is_delta = df_gdelta(a, a_gap, a_delta, a_base)
	if(a_is_delta)
	      call df_pdelta(b, a_gap, a_delta, a_base)
	else  call df_undelta(b)

#  Do the Calculation

	state = df_dgetd(a, xo, yo, 1)

	while(DOUBLE_LE(x,xo) && i <= nelms) {
		call df_dputd(b, x, a_ledge, 1)
		i = i + 1
		x = xvals[i]
	}

	while(df_dgetd(a, xn, yn, 1) != 0 && i <= nelms) {
	     if(a_is_delta)
		while(DOUBLE_LE(x,xn) && i <= nelms) {
			y = bo_interp(xn, yn, xo, yo, a_gap, a_base, x)
			call df_dputd(b, x, y, 1)
			i = i + 1
			x = xvals[i]
		}
	     else
		while(DOUBLE_LE(x,xn) && i <= nelms) {
			y = pf_interp(xn, yn, xo, yo, x)
			call df_dputd(b, x, y, 1)
			i = i + 1
			x = xvals[i]
		}
		xo = xn; yo = yn
	}

	while(i <= nelms) {
		call df_dputd(b, x, a_redge, 1)
		i = i + 1
		x = xvals[i]
	}

end


#
# pd_zdfd()
#

procedure	pd_zdfd(a, c, b) #I

pointer	a			#I input df
pointer	c			#I input df
pointer	b			#I output df

int	a_state, c_state	# get state

double	x0, a_x0		# x0 of a
double	a_ledge, a_redge	# left and right edge
double	a_nx, a_ox, a_ny	# discrete values
double	xn, xo, tx		# discrete values
double	area, y			# area and value
double	gap, delta, base	#

pointer	a_bo, bo_open()
bool	df_gdelta()
int	df_type()
int	df_dgetd()
int	bo_dgetd()

begin

	call df_gx0(a, a_x0)
	if(IS_INDEFD(a_x0)) call df_mkx0(a)

	a_bo = bo_open(a)

	call df_gx0(a, a_x0)
	call df_gedge(a, a_ledge, a_redge)

#  Get and Set delta values

	if(df_gdelta(a, gap, delta, base))
	      call df_pdelta(b, gap, delta, base)
	else  call df_undelta(b)

#  Do the Calculation

	a_ox = a_x0
	a_state = bo_dgetd(a_bo, a_nx, a_ny)

	if(df_type(c) == DF_PD) {
		call df_gx0(c, x0)
		c_state = df_dgetd(c, xn, y, 1)
		xo = x0
	} else {
		c_state = df_dgetd(c, x0, y, 1)
		c_state = df_dgetd(c, xn, y, 1)
		xo = x0
	}
	
	call df_px0(b, x0)
	call df_pedge(b, a_ledge, a_redge)
	xo = x0

	while(DOUBLE_LE(xn,a_x0) && c_state != 0) {
		call df_dputd(b, xn, a_ledge, 1)
		xo = xn
		c_state = df_dgetd(c, xn, y, 1)
	}

	while(DOUBLE_LE(a_nx,x0) && a_state != 0) {
		a_ox = a_nx
		a_state = bo_dgetd(a_bo, a_nx, a_ny)
	}

	while(a_state != 0 && c_state != 0) {

		area = 0.0
		tx = xo

		while(DOUBLE_LE(a_nx,xn)) {
			area = area + (a_nx - tx)*a_ny
			a_ox = a_nx
			tx = a_nx
			a_state = bo_dgetd(a_bo, a_nx, a_ny)
			if(a_state == 0) {
				a_nx = MAX_DOUBLE
				a_ny = a_redge
			}
		}

		area = area + (xn - tx)*a_ny
		y = area/(xn - xo)
		call df_dputd(b, xn, y, 1)

		xo = xn
		c_state = df_dgetd(c, xn, y, 1)

		while(DOUBLE_LE(xn,a_nx) && c_state != 0) {
			call df_dputd(b, xn, a_ny, 1)
			xo = xn
			c_state = df_dgetd(c, xn, y, 1)
		}
	}

	call bo_close(a_bo)

	call df_rewind(c)
end


#
# pf_zdfd()
#

procedure	pf_zdfd(a, c, b) #I

pointer	a			#I input df
pointer	c			#I input df
pointer	b			#I output df

int	a_state, c_state	# df get state
double	y, x 			# x and y values
double	xo, yo			#
double	xn, yn			#
double	a_ledge, a_redge	#
double	a_gap, a_delta, a_base	#

int	df_dgetd()
bool	a_is_delta, df_gdelta()
double	pf_interp()
double	bo_interp()

begin

#  Get and Set x0

	c_state = df_dgetd(c, x, y, 1)
	call df_px0(b, x)

#  Get and Set the edges

	call df_gedge(a, a_ledge, a_redge)
	call df_pedge(b, a_ledge, a_redge)

#  Get and Set delta values

	a_is_delta = df_gdelta(a, a_gap, a_delta, a_base)
	if(a_is_delta)
	      call df_pdelta(b, a_gap, a_delta, a_base)
	else  call df_undelta(b)

#  Do the Calculation

	a_state = df_dgetd(a, xo, yo, 1)

	while(DOUBLE_LE(x,xo) && c_state != 0) {
		call df_dputd(b, x, a_ledge, 1)
		c_state = df_dgetd(c, x, y, 1)
	}

	while(df_dgetd(a, xn, yn, 1) != 0 && c_state != 0) {
	     if(a_is_delta)
		while(DOUBLE_LE(x,xn) && c_state != 0) {
			y = bo_interp(xn, yn, xo, yo, a_gap, a_base, x)
			call df_dputd(b, x, y, 1)
			c_state = df_dgetd(c, x, y, 1)
		}
	     else
		while(DOUBLE_LE(x,xn) && c_state != 0) {
			y = pf_interp(xn, yn, xo, yo, x)
			call df_dputd(b, x, y, 1)
			c_state = df_dgetd(c, x, y, 1)
		}
		xo = xn; yo = yn
	}

	while(c_state != 0) {
		call df_dputd(b, x, a_redge, 1)
		c_state = df_dgetd(c, x, y, 1)
	}

	call df_rewind(c)
end

