# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  dftrim.x
# 	Author	:  akeith
# 	Date	:  03/30/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) dftrim.x -r1.3 03/30/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	04/24/95   akeith	updated to better handle INDEF's, better io docs, and now df's can be defined as delta df's
# 	04/07/94   akeith	the values of x_low and x_high are no longer effected.
# 	03/30/94   akeith	date and time created 94/03/30 12:37:27 by akeith
# H-
# 
#


#	Version: @(#)dftrim.x	1.3 4/24/95 UCB CEA EGO

include	"df.h"
include	<dfset.h>

#
# df_trim() trims the domain of a to the range of x_low
#	and x_high and places the resulting df in b. a
#	can equal b and if a is a piecewise function and
#	b is pixel data, the finial result is translated
#       using df_pfpd(). The resulting df must exist.
#

procedure	df_trim(a, x_low, x_high, b) #E

pointer	a			#I input DF
double  x_low			#I lowest acceptable domain value
double  x_high			#I highest acceptable domain value
pointer b			#I output DF (must already exist)

pointer op, op_open()

begin
	call df_ztrim(a, x_low, x_high, op_open(op, a, b))
	call op_close(op, a, b)
end


#
# df_ztrim() performs the calculation for two distinct
#       df's of the same type.
#

procedure	df_ztrim(a, xl, xh, b) #I

pointer	a			#I input DF
double  xl			#I lowest acceptable domain value
double  xh			#I highest acceptable domain value
pointer b			#I output DF (must already exist)

double	x0
double	x, y 
double	x_low, x_high 
double	ledge, redge
double	gap, delta, base
double	ox, oy
double	pf_interp()
double	bo_interp()

int	state, df_dgetd()
int	df_type()
bool	is_delta, df_gdelta()

begin

	x_low = xl; x_high = xh

	call df_gedge(a, ledge, redge)
	call df_pedge(b, ledge, redge)

   # Get each element, check if x value is in range.

	call df_gx0(a, x0)
	if(IS_INDEFD(x0)) {
		call df_mkx0(a)
		call df_gx0(a, x0)
	}

	if(x_low < x0) {
		call df_px0(b, x0)
		x_low = x0
	} else	call df_px0(b, x_low)

	if(x_high < x0) call error(1, "df_trim(): x_high < x0")

        is_delta = df_gdelta(a, gap, delta, base)
	if(is_delta) call df_pdelta(b, gap, delta, base)
        else call df_undelta(b)

	if(df_type(a) == DF_PD) {

		repeat {
			state = df_dgetd(a, x, y, 1)
		} until(DOUBLE_GT(x, x_low) || state == 0)

		if(state == 0) call error(1, "df_trim(): x_low > df domain")

		while(DOUBLE_LT(x, x_high) && state != 0) {
			call df_dputd(b, x, y, 1)
			state = df_dgetd(a, x, y, 1)
		}

		if(state != 0)
	            if(is_delta)
                        if(x_high > delta) call df_dputd(b, x_high, base, 1)
                    else call df_dputd(b, x_high, y, 1)

	} else {

		state = df_dgetd(a, x, y, 1)
		repeat {
			ox = x; oy = y
			state = df_dgetd(a, x, y, 1)
		} until(DOUBLE_GT(x, x_low) || state == 0)

		if(state == 0) call error(1, "df_trim(): x_low > df domain")

                if(is_delta) oy = bo_interp(x, y, ox, oy, gap, base, x_low)
                else oy = pf_interp(x, y, ox, oy, x_low)
		call df_dputd(b, x_low, oy, 1)

		while(DOUBLE_LT(x, x_high) && state != 0) {
			call df_dputd(b, x, y, 1)
			ox = x; oy = y
			state = df_dgetd(a, x, y, 1)
		}

		if(state != 0) {
                        if(is_delta)
                             y = bo_interp(x, y, ox, oy, gap, base, x_high)
                        else y = pf_interp(x, y, ox, oy, x_high)
			call df_dputd(b, x_high, y, 1)
		}

	}

end

