# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  getfiledf.x
# 	Author	:  tnk
# 	Date	:  04/28/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) getfiledf.x -r1.3 04/28/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	09/14/94   tnk	adjusted df height/area code
# 	09/13/94   tnk	switched to p[fd]_rimg routines
# 	04/28/94   tnk	date and time created 94/04/28 12:02:32 by tnk
# H-
# 
#


include <dfset.h>
include <tbset.h>
include <imhdr.h>
include <imio.h>

# @(#)getfiledf.x	1.3 9/14/94 UCB CEA EGO

# Given the name of a file, read all its contents & return as a df
# of same type as "file_df".
# domain_col and range_col are column names for ST table or
# column numbers (eg, "c1") for an ASCII file.
# File can be ASCII, ST table, or IRAF image format.
# Returns "true" on success

bool procedure get_file_df(file_df,filename, domain_col, range_col, ftype)

pointer file_df         #I/O existing DF, which we read into
char    filename[ARB]   #I name of file to read 
char    domain_col[ARB] #I name or number of domain column
char    range_col[ARB]  #I name or number of range column
char    ftype           #O returned file type ("I"mage "T"able "E"rror)

pointer read_df         #temporary df used to force a read
pointer tbl_ptr
pointer im_ptr
bool    status
int     getdftype
int     line
int     band

# FUNCTIONS
pointer tbtopn()            # open ST table
pointer tbl_df()            # create DF from ST table
int	tbtacc()
pointer immap()
pointer pf_rimg()
pointer pd_rimg()
int     df_type()
int     df_len()

begin

   status = true

   tbl_ptr = NULL
   im_ptr = NULL
   ftype = 'E'
   getdftype = df_type(file_df)

   # figure out input file type:  image or Table (binary or ascii)
   # try image first  

   iferr ( im_ptr = immap(filename, READ_ONLY, 0) ) 
      im_ptr = NULL

   if (im_ptr != NULL) {
     ftype = 'I'
     line = 1
     band = 1
     if (getdftype == DF_PF)
         read_df  = pf_rimg(im_ptr, line, band)
     else
         read_df  = pd_rimg(im_ptr, line, band)

     if (read_df == NULL) {
         call eprintf("Could not read data from image file %s\n")
         call pargstr(filename)
         status = false
     }
     else {
        call df_dbw(read_df, read_df)  #convert from areas to heights
	call df_mov(read_df, file_df)
     }
   }
   else if ( tbtacc(filename) == YES) { # ST table or ASCII

      tbl_ptr = tbtopn(filename,READ_ONLY,0)    #open the table
      if (tbl_ptr == NULL) {
            call eprintf("Cannot open table %s\n")
            call pargstr(filename)
            status = false
      } else {

         ftype='T'
         # open file DF
         read_df  = tbl_df(getdftype, tbl_ptr, domain_col, range_col)
         if (read_df == NULL) {
             call eprintf("could not read columns from table file %s\n")
             call pargstr(filename)
             status = false
         } else {
           
            # convert from areas to heights if this is a PD
            if (getdftype == DF_PD)
	       call df_dbw(read_df,file_df)
            else
               call df_mov(read_df,file_df)

         }
      }
   }
   else {  #not table (binary or ascii) or image 
      call eprintf("Cannot read input file %s.\n")
      call pargstr(filename)
      call eprintf("It must be a valid IRAF image, ST binary table or ASCII table.\n")
      status = false
   }

   if (status == true) {

      call df_close(read_df)

      # verify that df is not empty
      if (  df_len(file_df) == 0 ) {
         call eprintf("input file is empty\n")
         status = false
      } 

      call df_rewind(file_df)
   }

   # close input file
   if (tbl_ptr != NULL) call tbtclo(tbl_ptr)
   if (im_ptr != NULL) call imunmap(im_ptr)

   return (status)

end



