# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  ebaspect.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) ebaspect.x -r1.7 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	09/30/96   tnk	adjusted ra range to be 0 to 360
# 	06/29/94   akeith	Now does ds and scanner remap
# 	05/04/94   akeith	Got rid of DEBUG statements
# 	09/28/92   akeith	put in debug statements
# 	10/28/91   ericco	get source position in detector coordinates instead of
# computing the "distance" between an event and the source.  These
# positions are the x and y angles from boresight to the source.
# 	05/07/91   ericco	date and time created 91/05/07 14:05:24 by ericco
# H-
# 
#

#	Revised: 9/30/96

include <edata.h>
include <math.h>
include "ebuffer.h"
include	<dbg.h>

#	This Z value is the focal length of the telescope
#	optics.  Currently, the range of x is 204.8 to 1843.2.
#	The field of view is 2.5 degrees. so, the focal length
#	must be: -(1843.2-204.8)/atan(2.5)*180/pi/2
#	which is:
define		DEF_Z	-688.236
#	DEF_X, DEF_Y is the position that corresponds to the
#	z-axis on the detector.
define		DEF_X	1024.0
define		DEF_Y	1024.0

#	This should be changed in the future.  right now, we
#	add the source position offset (from boresight) to a
#	photon position.  In the future, source position should
#	be added to new fields in the photon record.  Then,
#	later in the processing, photon position can be modified
#	or flat fielding, or whatever else is source position
#	dependent -- like colimeter response.
#	Also, the boresight position and aperature high should
#	be input parameters for each detector.  Perhaps it is
#	even necessary to have different factors for x and y
#	positions! Talk to Vince about this.
#	Currently the output buffer is always set to the input
#	buffer.
procedure	eb_aspect(asp, ib, xi, yi, ti, src, aspect, ob)	#correct aspect
pointer		asp		#I aspect buffer descriptor
pointer		ib		#I input buffer
int		xi, yi, ti	#I indexes to src dx, dy, time fields
double		src[4]		#I assumed source position quaternion
pointer		aspect		#I aspect data pointer
pointer		ob		#I output buffer, if any

long		index
pointer		ptr
#double		z
#double		cx, cy
#data		z/DEF_Z/
#data		cx/DEF_X/
#data		cy/DEF_Y/
double		sx, sy

begin
  EB_FROMFIRST(index, ib) {
    ptr = EB_POINTER(ib,index)
    call er_aspeval(asp, EB_VALUE(ptr, ti))
    call er_aspang(asp, src, Memd[aspect], sx, sy)
    EB_VALUE(ptr,xi) = sx
    EB_VALUE(ptr,yi) = sy
#    call er_asprot(asp, src, Memd[aspect], z, sx, sy)
#    EB_VALUE(ptr,xi) = EB_VALUE(ptr,xi) - (sx + cx)
#    EB_VALUE(ptr,yi) = EB_VALUE(ptr,yi) - (sy + cy)
  }
  ob = ib
end



#				Z
#
#			yref	|
#			   |	|
#			   |	|
#			   v	|
#		       -  ------|-----------
#		     /  /	|	   /
#	     xref-->/  /	|	  /
#		    - /		+----------------------- Y
#		     /	       /      | /
#		    /	      /       |<-- zref
#		   /	     /        /
#		   ---------/--------
#		           /
#
#			 X



#
# eb_imaspect()
#

procedure	eb_imaspect(asp, ib, xi, yi, ti, acs2ds, ob)

pointer	asp			#I aspect buffer descriptor
pointer	ib			#I input buffer
int	xi, yi, ti		#I indexes to src dx, dy, time fields
double	acs2ds[4]		#I DS aspect offset: acs to ds
pointer	ob			#I output buffer, if any

long	index
pointer	ptr
double	j2ds[4]
double	ds2j[4]
double	j2acs[4]

begin

	EB_FROMFIRST(index, ib) {

		ptr = EB_POINTER(ib,index)

		call er_aspeval(asp, EB_VALUE(ptr, ti))
		call er_aspcur(asp, j2acs)
		call er_qmult(j2acs, acs2ds, j2ds)
		call er_qinvert(j2ds, ds2j)
		call er_qrotate(EB_VALUE(ptr,xi), EB_VALUE(ptr,yi), ds2j)
		if (EB_VALUE(ptr,xi) < 0.0)
			EB_VALUE(ptr,xi) = EB_VALUE(ptr,xi) + TWOPI

	}

	ob = ib
end
