# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  ebpht.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) ebpht.x -r1.8 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	01/13/97   mabbott	added check for GT_BAD table causing null column ptrs
# 	05/12/94   akeith	took out debug statements
# 	05/04/94   akeith	Got rid of DEBUG statements
# 	09/28/92   akeith	put in debug statements
# 	10/28/91   ericco	added output percent completion message
# 	06/19/91   ericco	add er_getd to fix pointer (type) evaluation bug
# 	05/07/91   ericco	date and time created 91/05/07 14:06:00 by ericco
# H-
# 
#

#	Revised: 1/13/97

#	EUV buffer photon operations

include <edata.h>
include <eerror.h>
include <dbg.h>
include	"../er/ercol.h"
include	"ebuffer.h"

define	EBP_GROUP	VEC_ELEMENT($1,1)
define	EBP_TP		VEC_ELEMENT($1,2)
define	EBP_BUFFER	VEC_ELEMENT($1,3)
define	EBP_TPDONE	VEC_ELEMENT($1,4)

define	EBP_LENGTH	long(4)

define	X_INDEX		1
define	Y_INDEX		2
define	W_INDEX		3
define	S_INDEX		4
define	Z_INDEX		5
define	T_INDEX		6
define	D_INDEX		7

define	NUM_INDEX	7

#	Release event buffer resources
procedure	ebp_close (pht)	# close
pointer		pht		#I photon data

begin
  call er_gclose(EBP_GROUP(pht))
  call eb_free(EBP_BUFFER(pht))
  call ev_free(pht)
end

#
#	Field names should not be hard coded here.
#	This routine gets the very first event as
#	well as initializing information.  This way
#	"get next" event doesn't have to check for
#	special conditions.
#

procedure	ebp_open (ebp, tp, names, length) # initialize event buffer

pointer		ebp			#O photon data
pointer		tp			#I table pointer
pointer		names			#I list of column names
long		length			#I buffer length

pointer		buf

bool	eu_nullinvec()

begin

#  This sets up the EUV buffer photon operations structure, EBP_.
#  Two additional structures are stored within the EBP_ structure.
#  The first is created in er_gopen() and stored in EBP_GROUP, the
#	second is allocated by eb_alloc() and stored in EBP_BUFFER.

	call ev_new(ebp, EBP_LENGTH)
	EBP_TPDONE(ebp) = 1

#  This opens a group of columns whose names are stored within names.
#  The pointer to the group list is stored in EBP_GROUP.
#  Here the names are the columns w/in table0.

	call er_gopen(tp, VEC_ELEMENT(names,1), VEC_SIZE(names),
			long(0), EBP_GROUP(ebp))

    # Die if we couldn't find all of the columns we need
    if (eu_nullinvec(EBP_GROUP(ebp))) {
	  call eprintf("cannot find all event list columns\n")
	  call eerror(EES_NOTAB)
    }

	EBP_TP(ebp) = tp		# Table pointer of table containing
					#     the photons.

#  Allocate space for the EB_ structure and store the pointer in EBP_BUFFER.

	call eb_alloc(EBP_BUFFER(ebp))
	buf = EBP_BUFFER(ebp)

	call eb_addfield(buf, "X")	# This incr. EBG_WIDTH and places
	call eb_addfield(buf, "Y")	# the value of width in one of the
	call eb_addfield(buf, "W")	# EBE_ fields, eg. EBE_TIME.
	call eb_addfield(buf, "S")
	call eb_addfield(buf, "Z")
	call eb_addfield(buf, "TIME")
	call eb_addfield(buf, "DETECTOR")

	call eb_addbuffer(buf, int(length))	# allocates space to EB_DATA_PTR
						# of type double.
	call er_phtfirst(EBP_GROUP(ebp))
	call ebp_pushpht(ebp)

end


#	Get the next additional events until the event
#	buffer is filled.
procedure	ebp_morepht(pht, flag) # fill event buffer
pointer		pht				#I photon data
bool		flag				#O EOD flag

pointer		buf
int		new
long		did
long		total

begin


	total = max(GT_NROWS(EBP_TP(pht)),1)
	new = EBP_TPDONE(pht)
	did = new * 10 / total

	buf = EBP_BUFFER(pht)
	flag = FALSE
	if (EBG_USED(buf) > 0) return

	while (EB_NUMSLOT(buf) > 0) {

		call er_phtnext(EBP_GROUP(pht), flag)
		if (flag == TRUE) {
			if (EBG_USED(buf) > 0) flag = FALSE
			break
		}
		call ebp_pushpht(pht)
		new = new + 1
	}

	EBP_TPDONE(pht) = new
	new = new * 10 / total
	if (did < new) {
		call eprintf("%d%% read\n")
		call pargi(new*10)
	}

end

#	return the buffer associated with an event
#	buffer
procedure	ebp_gbuf(pht, buf) # get event buffer
pointer		pht				#I photon data
pointer		buf				#O current buffer
begin
  buf = EBP_BUFFER(pht)
end

#	Copy an event from the group into a event buffer.
#	This routine assumes a lot of type information
#	about events.  This information is correct for EUV
#	but is not necessarily true in general.
procedure	ebp_pushpht(pht) # store event in a event buffer
pointer		pht				#I photon data

pointer		buf
pointer		grp

pointer		ptr

begin
  buf = EBP_BUFFER(pht)
  grp = EBP_GROUP(pht)
  EBG_USED(buf) = EBG_USED(buf) + 1
  ptr = EB_POINTER(buf, EBG_USED(buf))

  # need to modify to run with various data types
  call er_getd(grp, X_INDEX, EC_HI_OFFSET, EB_VALUE(ptr,X_INDEX))
  call er_getd(grp, Y_INDEX, EC_HI_OFFSET, EB_VALUE(ptr,Y_INDEX))
  call er_getd(grp, W_INDEX, EC_HI_OFFSET, EB_VALUE(ptr,W_INDEX))
  call er_getd(grp, S_INDEX, EC_HI_OFFSET, EB_VALUE(ptr,S_INDEX))
  call er_getd(grp, Z_INDEX, EC_HI_OFFSET, EB_VALUE(ptr,Z_INDEX))
  call er_getd(grp, D_INDEX, EC_HI_OFFSET, EB_VALUE(ptr,D_INDEX))
  call er_getd(grp, T_INDEX, EC_HI_OFFSET, EB_VALUE(ptr,T_INDEX))
end
