# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  ebuffer.h
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) ebuffer.h -r1.1 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	05/07/91   ericco	date and time created 91/05/07 14:06:10 by ericco
# H-
# 
#

#	Revised: 5/7/91

#	EUV buffer operations

#	Currently there is a fixed number of known field names.
#	In the future this will be generalized to handle any
#	number of names.

define	EB_FIELD_DICT	",X,Y,W,S,Z,TIME,TYPE,DETECTOR,FLAT,"

#	Event Buffer data structure.
#	EBE entries are "event" information.  These are
#	special fields that the software needs to know about
#	inorder to process EUV data correctly.  In the future,
#	other field may be included in a event buffer -- which
#	are not of particular interest to the EUV data processing.
#	EBG entries are "global" information about the event
#	buffer.  For example, EBG_DETECTOR is the detector number
#	that the buffer contains.  Typically this is information
#	that is not necessary to carry along with individual events.

define	EBE_X			VEC_ELEMENT($1,1)	# photon description
define	EBE_Y			VEC_ELEMENT($1,2)
define	EBE_W			VEC_ELEMENT($1,3)
define	EBE_S			VEC_ELEMENT($1,4)
define	EBE_Z			VEC_ELEMENT($1,5)
define	EBE_TIME		VEC_ELEMENT($1,6)
define	EBE_TYPE		VEC_ELEMENT($1,7)
define	EBE_DETECTOR		VEC_ELEMENT($1,8)
define	EBE_FLAT		VEC_ELEMENT($1,9)
define	EB_DATA_PTR		VEC_ELEMENT($1,10)
define	EBG_DETECTOR		VEC_ELEMENT($1,11)	# global description
define	EBG_TYPE		VEC_ELEMENT($1,12)
define	EBG_FORMAT		VEC_ELEMENT($1,13)
define	EBG_WIDTH		VEC_ELEMENT($1,14)
define	EBG_LENGTH		VEC_ELEMENT($1,15)
define	EBG_INDIRECT		VEC_ELEMENT($1,16)
define	EBG_USED		VEC_ELEMENT($1,17)
define	EBG_INT_Y		VEC_ELEMENT($1,18)

define	EB_LENGTH		long(20)

#	Initial development focused on two types of event buffers:
#	direct and indirect.  Direct buffers contained actual events,
#	whereas indirect buffers contained pointers to events.  During
#	development it became clear that the use of indirect buffers
#	was not very robust.  At the current time, 4/91, no indirect
#	buffers are being used -- although there is still some code
#	that handles indirect buffers.  Indirect buffers could still
#	be used on a scratch buffer basis.  That is, they could be
#	used to store temporary results.
#	The problem with indirect buffers in general, is in garbage
#	collection.  That is, when are all references to a event
#	buffer extinct?
#	In any case, indirect buffer contain pointers to the direct
#	buffers that they reference.
#	EB prefix are used with direct buffers.
#	EBI prefix are used with indirect buffers.

define	EB_VALID_INDEX		(($1) != 0)
define	EB_DIRECT		(EBG_INDIRECT($1) == ($1))
define	EB_VALUE		Memd[$1+($2)-1]

define	EB_POINTER		(EB_DATA_PTR($1)+(($2)-1)*EBG_WIDTH($1))
define	EB_DOUBLE		EB_VALUE(EB_POINTER($1,$2),$3)
define	EB_X			EB_DOUBLE($1,EBE_X($1))
define	EB_Y			EB_DOUBLE($1,EBE_Y($1))
define	EB_W			EB_DOUBLE($1,EBE_W($1))
define	EB_S			EB_DOUBLE($1,EBE_S($1))
define	EB_Z			EB_DOUBLE($1,EBE_Z($1))
define	EB_TIME			EB_DOUBLE($1,EBE_TIME($1))
define	EB_TYPE			EB_DOUBLE($1,EBE_TYPE($1))
define	EB_DETECTOR		EB_DOUBLE($1,EBE_DETECTOR($1))
define	EB_FLAT			EB_DOUBLE($1,EBE_FLAT($1))

define	EBI_POINTER		Meml[EB_DATA_PTR($1)+($2)-1]
define	EBI_DOUBLE		EB_VALUE(EBI_POINTER($1,$2),$3)
define	EBI_X			EBI_DOUBLE($1,EBE_X($1))
define	EBI_Y			EBI_DOUBLE($1,EBE_Y($1))
define	EBI_W			EBI_DOUBLE($1,EBE_W($1))
define	EBI_S			EBI_DOUBLE($1,EBE_S($1))
define	EBI_Z			EBI_DOUBLE($1,EBE_Z($1))
define	EBI_TIME		EBI_DOUBLE($1,EBE_TIME($1))
define	EBI_TYPE		EBI_DOUBLE($1,EBE_TYPE($1))
define	EBI_DETECTOR		EBI_DOUBLE($1,EBE_DETECTOR($1))
define	EBI_FLAT		EBI_DOUBLE($1,EBE_FLAT($1))

#	Convience macros for iterating over a event buffer.
define	EB_FROMFIRST	for ($1=1; $1 <= EBG_USED($2); $1 = $1 + 1)
define	EB_FROMLAST	for ($1=EBG_USED($2); $1 >= 1; $1 = $1 - 1)

#	Indirect macros (not tested)
define	EB_NEXTSLOT	EBI_POINTER($1,EBG_USED($1)+1)
define	EB_LASTSLOT	EBI_POINTER($1,EBG_USED($1))
define	EB_PUSHSLOT	EB_NEXTSLOT($1) = $2; EBG_USED($1) = EBG_USED($1) + 1
define	EB_POPSLOT	$2 = EB_LASTSLOT($1); EBG_USED($1) = EBG_USED($1) - 1

#	Convience macros for buffer information
define	EB_NUMSLOT	(EBG_LENGTH($1) - EBG_USED($1))
define	EB_HAVESLOT	(EB_NUMSLOT($1) > 0)
define	EB_NOSLOT	(EB_NUMSLOT($1) == 0)
