# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  ebuffer.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) ebuffer.x -r1.6 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	05/04/94   akeith	Got rid of DEBUG statements
# 	12/06/92   ericco	added clear buffer function
# 	09/28/92   akeith	put in debug statements
# 	07/17/91   ericco	modified malloc/mfree calls
# 	05/07/91   ericco	date and time created 91/05/07 14:06:15 by ericco
# H-
# 
#

#	Revised: 5/4/94

#	general	buffer operations

include	<dbg.h>
include	<eerror.h>
include <edata.h>
include "ebuffer.h"

#	Get the field index for a given column name.  Note
#	that columns are not in any particular order.`
#	Currently, there are a fixed number of column names.
procedure	eb_fieldid (name, id)	# get field index
char		name[ARB]			#I field name
int		id				#O field id

int		strdic()
char		junk

begin
  id = strdic(name, junk, 0, EB_FIELD_DICT)
  CHK_VALFIELD(id != 0)
end

#	Allocate a buffer of photon data
procedure	eb_alloc (ptr) # get a processing buffer
pointer		ptr				#O buffer

begin
  call ev_new_clear(ptr, EB_LENGTH)
  EBG_INDIRECT(ptr) = ptr
end

#	Add a field (not necessarily a new field) to a processing`
#	buffer.
procedure	eb_addfield (ptr, fname)	# add field
pointer		ptr				#I buffer
char		fname[ARB]			#I field name

int		id

begin
  CHK_INDBUF(EBG_INDIRECT(ptr) == ptr)
  CHK_EXISTPTR(EB_DATA_PTR(ptr) == 0)

  call eb_fieldid(fname, id)
  if (VEC_ELEMENT(ptr, id) == 0) {
    EBG_WIDTH(ptr) = EBG_WIDTH(ptr) + 1
    VEC_ELEMENT(ptr, id) = EBG_WIDTH(ptr)
  }
end

#	Get the length of the processing buffer's event buffer.
procedure	eb_glength(ptr, length) # get length
pointer		ptr				#I buffer
int		length				#O buffer length

begin
  length = EBG_LENGTH(ptr)
end

#	Get the column associated with a field index.
#	In order to handle names in general, this function
#	needs to change drastically.  Currently, the "index"
#	is a number from 1 to 9 that access one of the
#	first 9 elements in the processing buffer data
#	structure.  The buffer should really keep an a-list
#	of field indexes and column pointers and column
#	names, and provide access routines.
procedure	eb_gcolumn(ptr, index, column)
pointer		ptr				#I buffer
int		index				#I element index
long		column				#O column

begin
  column = VEC_ELEMENT(ptr, index)
end

#	This is only for direct buffers.
#	add event buffer to processing buffer

procedure	eb_addbuffer (ptr, length) # add event buffer

pointer		ptr				#I buffer
int		length				#I buffer length

begin
	CHK_EXISTPTR(EB_DATA_PTR(ptr) == 0)
	call eu_malloc (EB_DATA_PTR(ptr), length * EBG_WIDTH(ptr), TY_DOUBLE)
	CHK_NEXISTPTR(EB_DATA_PTR(ptr) != 0)
	EBG_USED(ptr) = 0
	EBG_LENGTH(ptr) = length

end

#	Extent an event buffer to a (presummable) new width.
#	Buffer length is not changed.
procedure	eb_extbuffer (ptr, length) # extent buffer
pointer		ptr				#I buffer
int		length				#I buffer length

begin
  CHK_NEXISTPTR(EB_DATA_PTR(ptr) != 0)
  call realloc (EB_DATA_PTR(ptr), length * EBG_WIDTH(ptr), TY_DOUBLE)
  CHK_NEXISTPTR(EB_DATA_PTR(ptr) != 0)
  EBG_LENGTH(ptr) = length
end

#	Release event buffer in a processing buffer
procedure	eb_freebuffer (ptr) # release event buffer
pointer		ptr				#I buffer

begin
  CHK_NEXISTPTR(EB_DATA_PTR(ptr) != 0)
  if (EBG_INDIRECT(ptr) != ptr) {
    call eu_mfree (EB_DATA_PTR(ptr), TY_LONG)
  } else {
    call eu_mfree (EB_DATA_PTR(ptr), TY_DOUBLE)
  }
  EB_DATA_PTR(ptr) = 0
  EBG_LENGTH(ptr) = 0
  EBG_USED(ptr) = 0
end

#	Create an indirect buffer from a direct buffer
procedure	eb_indbuffer (ptr, inp) # create indirect buffer
pointer		ptr				#O indirect buffer
pointer		inp				#I direct buffer

begin
  call eb_alloc(ptr)
  EBG_INDIRECT(ptr) = inp
end

#	Create a copy of a direct buffer
procedure	eb_copybuffer (ptr, inp, length) # copy direct buffer
pointer		ptr				#O direct buffer
pointer		inp				#I direct buffer
long		length				#I buffer length

begin
  call eb_alloc(ptr)

  call eb_headernew(ptr, inp)
  if (length == 0) {
    call eb_addbuffer (ptr, EBG_LENGTH(inp))
  } else {
    call eb_addbuffer (ptr, length)
  }
end

#	Create a copy of a direct buffer header
procedure	eb_headernew (ptr, inp) # copy processing buffer
pointer		ptr				#O direct buffer
pointer		inp				#I direct buffer

int		i

begin
  call eb_alloc(ptr)
  
  do i = 1, 9 {
    if (VEC_ELEMENT(inp, i) == 0) next
    EBG_WIDTH(ptr) = EBG_WIDTH(ptr) + 1
    VEC_ELEMENT(ptr, i) = EBG_WIDTH(ptr)
  }
end

#	Copy an event between direct buffers
#	To handle general event buffers the 1 to 9 limit
#	needs to change to lookup values in the field
#	a-list structure.
procedure	eb_copye (inp, index, out, pos) # general event copy
pointer		inp				#I direct buffer
long		index				#I location
pointer		out				#I direct buffer
long		pos				#I location

int		i
int		inpf
int		outf
pointer		iptr, optr

begin
  iptr = EB_POINTER(inp, index)
  optr = EB_POINTER(out, pos)
  inpf = 1
  outf = 1
  do i = 1, 9 {
    if (VEC_ELEMENT(out, i) == 0) {
      if (VEC_ELEMENT(inp, i) != 0) inpf = inpf + 1
      next
    }
    if (VEC_ELEMENT(inp, i) == 0) {
      EB_VALUE(optr, outf) = INDEFD
    } else {
      EB_VALUE(optr, outf) = EB_VALUE(iptr, inpf)
      inpf = inpf + 1
    }
    outf = outf + 1
  }
end

#	Bind indirect list to direct buffer
procedure	eb_bindbuffer (ptr) # initialize indirect buffer
pointer		ptr				#I indirect buffer

pointer		ind
int		length, index

begin
  CHK_DIRBUF(EBG_INDIRECT(ptr) != ptr)
  ind = EBG_INDIRECT(ptr)

  CHK_INDLIST(EB_DATA_PTR(ptr) == 0)
  length = EBG_LENGTH(ind)
  call eu_malloc(EB_DATA_PTR(ptr), length, TY_LONG)
  CHK_NINDLIST(EB_DATA_PTR(ptr) != 0)
  do index = 1, EBG_USED(ind) {
    EBI_POINTER(ptr, index) = EB_POINTER(ind, index)
  }
  EBG_LENGTH(ptr) = length
  EBG_USED(ptr) = EBG_USED(ind)
end

#	Release processing buffer
procedure	eb_free(ptr) # release buffer
pointer		ptr				#I buffer

begin
  if (EB_DATA_PTR(ptr) != 0) call eb_freebuffer(ptr)
  call ev_free(ptr)
  ptr = 0
end

#	Clear the contents of a buffer

procedure	eb_clear_buffer(eb) #E
pointer		eb		#B data pointer

begin
  if (eb != NULL) {
    EBG_USED(eb) = 0
  }
end
