# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  ecorrect.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) ecorrect.x -r1.7 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	05/04/94   akeith	Got rid of DEBUG statements
# 	09/21/93   akeith	fixed the yes/no true false problem
# 	09/28/92   akeith	put in debug statements
# 	10/28/91   ericco	added another input buffer to keep rescaled input values
# in.  I alson increased the size of each raster buffer.  Now there
# should be one event per pixel -- minimizing image io.  Also
# noted that only one buffer is open at a time on an image.  so
# i'm copying the x buffer into a tmp location before getting
# the y buffer.  Finally, I let the data scale back into real
# coordiates instead of always applying the inverse scaling
# 	07/17/91   ericco	modified malloc/mfree calls.
# 	05/07/91   ericco	date and time created 91/05/07 14:06:19 by ericco
# H-
# 
#

#	Revised: 5/4/94

define	MIN_BUFLEN	20

include	<imhdr.h>

include	<dbg.h>
include <edata.h>
include	"ebuffer.h"

#	EC is the EUV Correction buffer data structure.
#	2D distortion correction is a multi-step process.
#	Input events are first convert from their "world"
#	coordinate system to the correction coordinate
#	system.  This system is the image's physical coordinates
#	Next, the events are split into buffers cooresponding
#	to rows in the correction image.
#	When enough photons are collected in a given row, that
#	part of the image is read into memory, and a bi-linear
#	interpolation of the correction value is returned.
#	This value is added to the original position.

define	EC_IM		VEC_ELEMENT($1,1)
define	EC_IMNAME	VEC_ELEMENT($1,2)
define	EC_IMFULLNAME	VEC_ELEMENT($1,3)
define	EC_DT		VEC_ELEMENT($1,4)
define	EC_RECORD	VEC_ELEMENT($1,5)
define	EC_FCONST	VEC_ELEMENT($1,6)
define	EC_DATA		VEC_ELEMENT($1,7)
define	EC_XPLANE	VEC_ELEMENT($1,8)
define	EC_YPLANE	VEC_ELEMENT($1,9)
define	EC_FPLANE	VEC_ELEMENT($1,10)
define	EC_FROM		VEC_ELEMENT($1,11)
define	EC_TO		VEC_ELEMENT($1,12)
define	EC_SPLIT	VEC_ELEMENT($1,13)
define	EC_MW		VEC_ELEMENT($1,14)
define	EC_DETECTOR	VEC_ELEMENT($1,15)
define	EC_BUF		VEC_ELEMENT($1,15)

define	EC_LENGTH	long(20)
define	EC_FVALUE	Memd[EC_DATA($1)+0]
define	EC_DLENGTH	int(1)

#	Get an image header value based on a tdb field name
procedure	ec_hdrl(ec, ref, val) # get long header value
pointer		ec				#I data
char		ref[ARB]			#I reference
long		val				#O value

pointer		sp, buf
long		imgetl()

begin
  call smark(sp)
  call dtgptr(EC_DT(ec), int(EC_RECORD(ec)), ref, buf)
  val = imgetl(EC_IM(ec), Memc[buf])
  call sfree(sp)
end

#	Get optional flat field constant from image header.
procedure	ec_hdrflat(ec) # get flat constant value
pointer		ec				#I data

pointer		sp, buf
real		imgetr()
bool		dtgetb()

begin
  call smark(sp)
  if (dtgetb(EC_DT(ec), int(EC_RECORD(ec)), "fconstant") == TRUE) {
    call dtgptr(EC_DT(ec), int(EC_RECORD(ec)), "fvalref", buf)
    EC_FCONST(ec) = 1
    EC_FVALUE(ec) = imgetr(EC_IM(ec), Memc[buf])
  } else {
    EC_FCONST(ec) = 0
    call ec_hdrl(ec, "fplaneref", EC_FPLANE(ec))
  }
  call sfree(sp)
end

#	Get the transformation to and from the
#	world coordinate system and the physical
#	coordinate system.
procedure	ec_hdrwcs(ec) # get mwcs from image
pointer		ec				#I data

pointer		sp, buf, mw
pointer		mw_sctran()
pointer		mw_openim()

string		image_system "physical"

begin
  call smark(sp)
  mw = mw_openim(EC_IM(ec))
  EC_MW(ec) = mw
  call dtgptr(EC_DT(ec), int(EC_RECORD(ec)), "wcsref", buf)
  EC_TO(ec) = mw_sctran(mw, Memc[buf], image_system, 3B)
  EC_FROM(ec) = mw_sctran(mw, image_system, Memc[buf], 3B)
  call sfree(sp)
end


#	Initialize a correction buffer from the tdb

procedure	ec_read (dt, rec, path, ec, eb, det) # read correction data
pointer		dt				#I text database
int		rec				#I record name
char		path[ARB]			#I directory
pointer		eb				#I input buffer
pointer		ec				#O data pointer
int		det				#I detector id

pointer		buf, ref
pointer		sp
pointer		immap()
long		buflen

begin

  call ev_new(ec, EC_LENGTH)
  call eu_malloc(EC_DATA(ec), EC_DLENGTH, TY_DOUBLE)
  call smark(sp)
  EC_DT(ec) = dt
  EC_RECORD(ec) = rec
  call dtgptr(dt, rec, "Name", buf)
  call eu_file(path, Memc[buf], ref)
  call eu_newstr(Memc[buf], EC_IMNAME(ec))
  call eu_newstr(Memc[ref], EC_IMFULLNAME(ec))

  EC_IM(ec) = immap(Memc[ref], READ_ONLY, 0)
  EC_DETECTOR(ec) = det
  call ec_hdrl(ec, "XPlaneref", EC_XPLANE(ec))
  call ec_hdrl(ec, "YPlaneref", EC_YPLANE(ec))
  call ec_hdrflat(ec)
  call ec_hdrwcs(ec)

  call eb_copybuffer (EC_BUF(ec), eb, long(0))

#  buflen = EBG_LENGTH(eb) / IM_LEN(EC_IM(ec), 2)
  buflen = IM_LEN(EC_IM(ec), 1)
  if (buflen < MIN_BUFLEN) buflen = MIN_BUFLEN
  call es_open(eb, long(1),
	       long(IM_LEN(EC_IM(ec), 2)),
	       buflen,
	       EC_SPLIT(ec))
  call sfree(sp)
end

#	Release correction buffer resources
procedure	ec_free(ec)	# release correction resources
pointer		ec		#I correction data

begin
  call eb_free(EC_BUF(ec))
  call es_close(EC_SPLIT(ec))
  call mw_ctfree(EC_TO(ec))
  call mw_ctfree(EC_FROM(ec))
  call mw_close(EC_MW(ec))
  call imunmap(EC_IM(ec))
  call eu_mfree(EC_IMNAME(ec), TY_CHAR)
  call eu_mfree(EC_IMFULLNAME(ec), TY_CHAR)
  call eu_mfree(EC_DATA(ec), TY_DOUBLE)
  call ev_free(ec)
end

#	Apply a correction to a event buffer.
#	Events are transformed to image coordinates, and
#	then split up by row number.  If any of the rows
#	are ready, then those events are distortion corrected
#	and transformed back into world coordinates.

procedure	eb_correct(ec, eb, xi, yi, fi, eo, force)

pointer		ec				#I correction data
pointer		eb				#I input events
long		xi, yi, fi			#I x and y offsets
pointer		eo				#O results
int		force				#I force output

pointer		es
long		index
pointer		ib
long		op

begin
  ib = EC_BUF(ec)
  if (eb != 0 && EBG_USED(eb) != 0 && EB_NUMSLOT(ib) >= EBG_USED(eb)) {
    call eb_ctran2(EC_TO(ec), eb, xi, yi, xi, yi)
    op = EBG_USED(ib)
    EB_FROMFIRST(index,eb) {
      op = op + 1
      call eb_copye(eb, index, ib, op)
    }
    EBG_USED(ib) = op
    EBG_USED(eb) = 0
#  } else { # stop further processing
#    eb = 0
  }
  es = EC_SPLIT(ec)
  if (EBG_USED(ib) != 0) {
    call es_split(es, ib, yi)
  }
  index = 0
  if (force == 0) {
    call es_popfull(es, index, eo)
  } else {
    call es_popany(es, index, eo)
  }
  if (eo != 0) {
    # do distortion step in place
    call eb_distortion(ec, eo, xi, yi, fi) # includes ctrans
  }
  eb = eo
end


#	corrects position in place.
#	Includes ctrans from image space to instrument space.
#	only one image section can be in the cache at any given
#	time, so the xplane data are copied out of the cache before
#	the yplane data are read in.
procedure	eb_distortion(ec, eb, xi, yi, fi) # apply correction
pointer		ec				#I correction data
pointer		eb				#I/O event buffer
long		xi, yi, fi			#I field offset

long		y
pointer		ptr
double		dx, dy, df, ddf
pointer		im
int		width, len, xp, yp, fp
pointer		px, py, pf
int		fc
pointer		imgs3d()
int		index
#double		in[2],out[2]
pointer		sp, tp

begin

	im = EC_IM(ec)
	width = IM_LEN(im,1)
	len = IM_LEN(im,2)
	xp = EC_XPLANE(ec)
	yp = EC_YPLANE(ec)
	y = EBG_INT_Y(eb)
	fp = EC_FPLANE(eb)
	fc = EC_FCONST(ec)
	ddf = EC_FVALUE(ec)
	pf = NULL

	call smark(sp)
	call salloc(px, 3 * width, TY_DOUBLE)

	if (y == 0) {			# could be any line in image

		EB_FROMFIRST(index, eb) {

			ptr = EB_POINTER(eb, index)
			dx = EB_VALUE(ptr, xi)
			dy = EB_VALUE(ptr, yi)
			y = dy

			if (y + 1 > len) y = len - 1
			else if (y - 1 < 1) y = 2

			tp = imgs3d(im, 1, width, y-1, y+1, xp, xp)
			call amovd(Memd[tp], Memd[px], 3 * width)
			py = imgs3d(im, 1, width, y-1, y+1, yp, yp)

			if (fi != 0 && fc == 0 && fp != 0) {
				pf = imgs3d(im, 1, width, y-1, y+1, fp, fp)
				df = ddf
			}

			call eb_corr(Memd[px], Memd[py], pf, y, dx, dy, df, width)

			if (fi != 0)
				EB_VALUE(ptr, fi) = EB_VALUE(ptr, fi) * df

#in[1] = EB_VALUE(ptr, xi)
#in[2] = EB_VALUE(ptr, yi)
#call mw_ctrand(EC_FROM(ec), in, out, 2)
#EB_VALUE(ptr, xi) = out[1] + dx
#EB_VALUE(ptr, yi) = out[2] + dy

			EB_VALUE(ptr, xi) = dx
			EB_VALUE(ptr, yi) = dy
		}

	} else {			# must be line y in image

		if (y + 1 > len) y = len - 1
		else if (y - 1 < 1) y = 2

		tp = imgs3d(im, 1, width, y-1, y+1, xp, xp)
		call amovd(Memd[tp], Memd[px], 3 * width)
		py = imgs3d(im, 1, width, y-1, y+1, yp, yp)

		if (fi != 0 && fc == 0 && fp != 0) {
			pf = imgs3d(im, 1, width, y-1, y+1, fp, fp)
			df = ddf
		}

		EB_FROMFIRST(index, eb) {

			ptr = EB_POINTER(eb, index)
			dx = EB_VALUE(ptr, xi)
			dy = EB_VALUE(ptr, yi)
			if (y + 1 > len) y = len - 1
			call eb_corr(Memd[px], Memd[py], pf, y, dx, dy, df, width)
			if (fi != 0) EB_VALUE(ptr, fi) = EB_VALUE(ptr, fi) * df

#in[1] = EB_VALUE(ptr, xi)
#in[2] = EB_VALUE(ptr, yi)
#call mw_ctrand(EC_FROM(ec), in, out, 2)
#EB_VALUE(ptr, xi) = out[1] + dx
#EB_VALUE(ptr, yi) = out[2] + dy

			EB_VALUE(ptr, xi) = dx
			EB_VALUE(ptr, yi) = dy
		}
	}

#  call eb_ctran2(EC_FROM(ec), eb, xi, yi, xi, yi)

call sfree(sp)

end

