# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  esplit.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) esplit.x -r1.9 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	05/12/94   akeith	took out debug statements
# 	05/04/94   akeith	Got rid of DEBUG statements
# 	09/28/92   akeith	put in debug statements
# 	10/30/91   ericco	removed unused locals
# 	10/28/91   ericco	added time stamp keyword
# 	10/28/91   ericco	added protection against null pointers and weird data
# 	07/17/91   ericco	fixed memory leak by closing all buffers
# 	05/07/91   ericco	date and time created 91/05/07 14:06:28 by ericco
# H-
# 
#

#	Revised: 5/12/94

include	<dbg.h>
include	<edata.h>
include	"ebuffer.h"

define		ES_BUFFER	VEC_ELEMENT($1,($2)+1)
define		ES_MINIMUM	VEC_ELEMENT($1,1)
define		ES_LENGTH	(VEC_SIZE($1)-1)

#	Get first buffer
procedure	es_gfbuf (es, buf) # get first buffer
pointer		es			#I split buffer data
pointer		buf			#O first output buffer

begin
  buf = ES_BUFFER(es, 1)
end

#	Get number of buffers
procedure	es_glength(es, length) # get number of buffers
pointer		es		#I split processing buffer
int		length		#O number of buffers

begin
  length = ES_LENGTH(es)
end

#	Release split resources
procedure	es_close (es) # release resources
pointer		es			#I split buffer data

long		index

begin
  do index = 1, ES_LENGTH(es) + 1 {
    call eb_free(ES_BUFFER(es,index))
  }
  call ev_free(es)
end

#	Open a split buffer based on a vector string.

procedure	es_open1 (vstr, min, max, len, es) # open

pointer		vstr			#I string vector
long		min, max		#I index range
long		len			#I len of buffer
pointer		es			#O split buffer data

long		length
long		index
pointer		buf

int		i

begin

	length = max - min + 1 + 1	# length is the number of detectors.
	call ev_new(es, length)		# es = EPP_BUF(pointer)

	ES_MINIMUM(es) = min
	do index = 1, length {

		call eb_alloc(buf)
		ES_BUFFER(es,index) = buf
		EBG_INT_Y(buf) = min + index - 1

		do i = 1, VEC_SIZE(vstr) {
			call eb_addfield (buf, VEC_STRING(vstr,i))
		}
		call eb_addbuffer(buf, len)
	}

end

#	Open a split buffer based on a previous buffer type

procedure	es_open (eb, min, max, len, es)

pointer		eb			#I input buffer
long		min, max		#I index range
long		len			#I len of buffer
pointer		es			#O split buffer data

long		length
long		index
pointer		buf

begin

	length = max - min + 1 + 1
	call ev_new(es, length)
	ES_MINIMUM(es) = min

	do index = 1, length {
		call eb_copybuffer(ES_BUFFER(es, index), eb, len)
		buf = ES_BUFFER(es, index)
		EBG_INT_Y(buf) = min + index - 1
	}

end

#	Apply a split process to a event buffer
procedure	es_split(es, eb, findex) # split buffer
pointer		es			#I split buffer
pointer		eb			#I input buffer
long		findex			#I field index

int		split
pointer		ptr
long		minimum
pointer		sb
long		length
long		index
double		dtmp

begin

	minimum = ES_MINIMUM(es) - 1
	length = ES_LENGTH(es)

	if (EB_DIRECT(eb)) {

		EB_FROMLAST(index, eb) {

			ptr = EB_POINTER(eb, index)
			dtmp = EB_VALUE(ptr, findex) - minimum
			if (dtmp < 1 || dtmp > length) next
			split = dtmp
			sb = ES_BUFFER(es, split)

			if (EB_HAVESLOT(sb)) {

				EBG_USED(sb) = EBG_USED(sb) + 1
				call eb_copye (eb, index, sb, EBG_USED(sb))

			} else {
				EBG_USED(eb) = index
				return
			}
		}

	} else {

		EB_FROMLAST(index, eb) {

			ptr = EBI_POINTER(eb, index)
			split = EB_VALUE(ptr, index) - minimum

			if (split > 0 && split <= length) {

				sb = ES_BUFFER(es, split)

				if (EB_HAVESLOT(sb)) {
					EB_PUSHSLOT(sb,ptr)
				} else {
					EBG_USED(eb) = index
					return
				}
			}
		}
	}

	EBG_USED(eb) = 0

end

#	Emit a split buffer that is full.
procedure	es_popfull(es, index, buf) # get next available ptr
pointer		es			#I split buffer
long		index			#I/O index information
pointer		buf			#O available buffer

int		flag

begin

	if (es == 0) {
		buf = 0
		return
	}
	if (index == 0) {
		flag = 1
		index = ES_LENGTH(es)
	} else {
		buf = ES_BUFFER(es, index)
		if (EBG_USED(buf) == 0) flag = 1; else flag = 0
	}
	if (flag == 1) {

		for( buf = ES_BUFFER(es, index);
		     index > 0 && EBG_USED(buf) != EBG_LENGTH(buf);
		     buf = ES_BUFFER(es, index) )
						index = index - 1
	}

	if (index == 0) buf = 0

end

#	Emit any split buffer.
procedure	es_popany(es, index, buf) # get next available ptr
pointer		es			#I split buffer
long		index			#I/O index information
pointer		buf			#O next buffer

begin
  if (es == 0) {
    buf = 0
    return
  }
  if (index == 0) index = ES_LENGTH(es)
  for (buf = ES_BUFFER(es, index);
       index > 0			&&
       EBG_USED(buf) == 0;
       buf = ES_BUFFER(es, index)) {
    index = index - 1
  }
  if (index == 0) buf = 0
end

