# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  edread.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) edread.x -r1.7 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	12/27/96   mabbott	changes to support parameter overrides
# 	05/12/94   akeith	took out debug statements
# 	09/28/92   akeith	put in debug statements
# 	07/01/92   ericco	removed out of date records
# 	07/17/91   ericco	fixed memory on source part of data structure
# 	05/07/91   ericco	date and time created 91/05/07 14:07:39 by ericco
# H-
# 
#

#	Revised: 12/27/96

include <edata.h>
include <override.h>
include <dbg.h>

#	Read table information from an Observation
#	Store the table names in the ED_ structure.

procedure ed_rdtable (dt, rec, override, ed_tbls)	# Read Table list

pointer	dt				#I Text database
char	rec[ARB]			#I Record Name
pointer	override		#I Parameter override structure
pointer	ed_tbls				#O loaded information

long	i
long	len

pointer	sp
pointer	buf
pointer tmp

int	record				# record index

bool	streq()
int	dtgeti()
int	dtlocate()			# gets the record index

begin

	call smark(sp)

	record = dtlocate(dt, rec)		# locate the record
	len = dtgeti(dt, record, "Length")	# get number of fields
	call ev_new(ed_tbls, len)		# allocate mem for each tbl name

	do i = 1, len {				# store ptr's to table name strings
		call dtscan(dt)
		call dtgptr(dt, record, "", buf)

		# If filename starts with _%_, it is really a parameter name
		# to read the filename from.  Only two names are supported:
		# events and aspect. 
		if (streq(Memc[buf], "_%_events")) {
          if (EVENTS(override) != '/' && EVENTS(override) != '.' &&
              EVENTS(override) != EOS ) {
            call eu_file(".", EVENTS(override), tmp)
          }
          else call eu_snewstr(EVENTS(override), tmp)

		  call eu_newstr(Memc[tmp], VEC_ELEMENT(ed_tbls, i))
		} else if (streq(Memc[buf], "_%_aspect")) {
          if (ASPECT(override) != '/' && ASPECT(override) != '.' &&
              ASPECT(override) != EOS ) {
            call eu_file(".", ASPECT(override), tmp)
          }
          else call eu_snewstr(ASPECT(override), tmp)

		  call eu_newstr(Memc[tmp], VEC_ELEMENT(ed_tbls, i))
		} else {
          # no indirection, just store the name
		  call eu_newstr(Memc[buf], VEC_ELEMENT(ed_tbls, i))
		}
        call eu_free(buf, TY_CHAR)
	}

	call sfree(sp)

end


#
#	Read EUV data source information
#	Only stores the path and the name which isn't much.
#

procedure ed_rdsource (dt, rec, ed)	# Read Source record

pointer	dt				#I Text database
char	rec[ARB]			#I Record Name
pointer	ed				#I/O Modified source structure

pointer	ptr				# loaded information
int	record, dtlocate()
#int	dtgeti()

begin

  record = dtlocate(dt, rec)
  call ev_new_clear(ptr, ED_SRC_LEN)
  ED_SOURCE(ed) = ptr

#  ED_START(ed) =  dtgeti(dt, record, "start")
#  ED_FINISH(ed) = dtgeti(dt, record, "finish")
#  call dtgptr(dt, record, "Root", VEC_ELEMENT(ptr, ED_ROOT_ID))

  call dtgptr(dt, record, "path", VEC_ELEMENT(ptr, ED_DIR_ID))

#  call dtgptr(dt, record, "class", VEC_ELEMENT(ptr, ED_SUBCL_ID))
#  call dtgptr(dt, record, "specification", VEC_ELEMENT(ptr, ED_CL_ID))
#  call dtgptr(dt, record, "paramlabel", VEC_ELEMENT(ptr, ED_LABEL_ID))

  call dtgptr(dt, record, "name", VEC_ELEMENT(ptr, ED_NAME_ID))

end

#	Read column location information
#	Store the list of column names and the corresponding
#	table from which the columns come from.

procedure ed_rdcolumns(dt, rec, ed)	# Read Source record

pointer	dt				#I Text database
char	rec[ARB]			#I Record Name
pointer	ed				#O loaded information

int	record, dtlocate()
int	dtgeti()
long	len
long	i
pointer	cptr, tptr

begin

	record = dtlocate(dt, rec)
	len = dtgeti(dt, record, "Length")

	call ev_new(ED_COLUMN(ed), len)
	cptr = ED_COLUMN(ed)

	call ev_new(ED_TBL_REF(ed), len)
	tptr = ED_TBL_REF(ed)

	do i = 1, len {
		call dtscan(dt)
		call gargl(VEC_ELEMENT(tptr,i))
		call dtgptr(dt, record, "", VEC_ELEMENT(cptr, i))
	}

end


#	Read special column information

procedure ed_rdspecial(dt, rec, ed)	# Read Special records

pointer	dt				#I Text database
char	rec[ARB]			#I Record Name
pointer	ed				#I/O data pointer

int	record, dtlocate()
long	len
int	dtgeti()
pointer	tptr
pointer	cptr
long	tbl, col
int	spc_indx, strdic()     # special index as defined by the SPC_DICT.
pointer	sp, buf
long	i

begin

	call smark(sp)
	call salloc(buf, 1024, TY_CHAR)

	record = dtlocate(dt, rec)
	len = dtgeti(dt, record, "Length")

	call ev_new(tptr, ED_SPC_LEN)		# Allocate and clear the special
	ED_SPC_TBL(ed) = tptr			# column and table list.

	call ev_new(cptr, ED_SPC_LEN)
	ED_SPC_COL(ed) = cptr

	do i = 1, ED_SPC_LEN {
		VEC_ELEMENT(tptr, i) = 0
		VEC_ELEMENT(cptr, i) = 0
	}

	do i = 1, len {
		call dtscan(dt)
		call gargwrd(Memc[buf], 1024)
		spc_indx = strdic(Memc[buf], Memc[buf], 0, SPC_DICT)
		call gargwrd(Memc[buf], 1024)
		call edgtc (ed, Memc[buf], tbl, col)
		VEC_ELEMENT(tptr, spc_indx) = tbl
		VEC_ELEMENT(cptr, spc_indx) = col
	}

	call sfree(sp)

end

#	Read table time column information
#	This creates a list of the time columns indexed by
#	the table number. The pointer to the list is stored
#	in ED_TIMES.

procedure	ed_rdtimes(dt, record, ed) # Read time column names

pointer	dt				#I Text database
char	record[ARB]			#I record name
pointer	ed				#I loaded information

int	rec, dtlocate()
int	tbl, col
pointer	sp, buf
pointer	tptr
int	len
int	dtgeti()
int	i

begin

	call smark (sp)
	call salloc (buf, 1024, TY_CHAR)

	call ev_new_clear(ED_TIMES(ed), VEC_SIZE(ED_TABLES(ed)))
	tptr = ED_TIMES(ed)

	rec = dtlocate(dt, record)
	len = dtgeti(dt, rec, "Length")

	do i = 1, len {
		call dtscan(dt)
		call gargi(tbl)
		call gargwrd(Memc[buf], 1024)
		call edgtc (ed, Memc[buf], tbl, col)
		VEC_ELEMENT(tptr, tbl) = col
	}

end


#	Read valid detector list
#	A pointer, stored in ED_DETECTORS, to a list of valid
#	detectors as indicated by a value of one, otherwise a
#	zero. This list is indexed by detector number.

procedure	ed_rddets(dt, record, ed) # Read time column names

pointer	dt				#I Text database
char	record[ARB]			#I record name
pointer	ed				#I loaded information

int	rec, dtlocate()
int	tbl, tptr
pointer	sp, buf

int	len, dtgeti()
int	i

begin

	call smark (sp)
	call salloc (buf, 1024, TY_CHAR)

	call ev_new_clear(ED_DETECTORS(ed), MAX_DETECTORS)
	tptr = ED_DETECTORS(ed)

	rec = dtlocate(dt, record)
	len = dtgeti(dt, rec, "Length")

	do i = 1, len {
		call dtscan(dt)
		call gargi(tbl)
		VEC_ELEMENT(tptr, tbl) = 1
	}

end

#
#	Read information related to an EUV Observation.
#
#  The ED_ structure is malloced and stored in EP_OBS_PTR.
#  The ED_structure is then filled with info from the obs.
#  text database by calling ed_rd* functions.
#

procedure ed_read (dt, override, ed)	# Read Observation Processing data

pointer	dt				#I Text database
pointer	override		#I Parameter override structure
pointer	ed				#O loaded information

begin

#    The ED_ (EUV Data) structure definition can be found in edata.h.

	call eu_malloc(ed, ED_LENGTH, TY_LONG)

	ED_ASPECT(ed) = 0
	ED_PHTBUF(ed) = 0
    ED_SRCPOS(ed) = SRCPOS(override)

#    store the table names here.
	call ed_rdtable(dt, "Table List", override, ED_TABLES(ed))

#    soon the GT_ structure will stored here.
	call ev_new_clear(ED_TABLE_PTR(ed), VEC_SIZE(ED_TABLES(ed)))

#    there is little use of this.
	call ed_rdsource(dt, "EES Source", ed)

#    store column names and corres. tables of residence.
	call ed_rdcolumns(dt, "Column List", ed)

#    store the table and column indicies for the special monitors.
	call ed_rdspecial(dt, "Special List", ed)

#    store the list of column indicies for the time columns indx by table number.
	call ed_rdtimes(dt, "TimeStamps", ed)

#    store the list of valid detectors.
	call ed_rddets(dt, "Detector List", ed)

end


#	Release EUV observation information
procedure	ed_free (state) # release resources
pointer		state		#I/O EUV data

begin
  call ev_free_string(ED_TABLES(state))
  call ev_free_string(ED_SOURCE(state))
  call ev_free_string(ED_COLUMN(state))
  call ev_free(ED_SPC_TBL(state))
  call ev_free(ED_SPC_COL(state))
  call ev_free(ED_TIMES(state))
  call ev_free(ED_DETECTORS(state))
  call ev_free(ED_TBL_REF(state))
  call ev_free(ED_TABLE_PTR(state))
  call eu_mfree(state, TY_LONG)
end
