# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  edd.x
# 	Author	:  ericco
# 	Date	:  04/30/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) edd.x -r1.1 04/30/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	04/30/92   ericco	date and time created 92/04/30 00:26:49 by ericco
# H-
# 
#

#	External interface to the EUV detector data structures
#	See also eddops.x

include	"dd.h"
include "det.h"
include "ea.h"
include "eao.h"
include "hs.h"
include "pd.h"
include "psf.h"

#	Returns the number of hot spots on the given detector.
#	A return status of YES indicates success.
int procedure	edd_hotspots(dd, number)
pointer		dd			#I detector structure
int		number			#O number of hotspots

begin
	number = 0
	if (dd == 0) return NO
	number = DD_HOTSPOT_LEN(dd)
	return YES
end

#	Return the characteristics of a hot spot.
#	A return status of YES indicates success.
int procedure	edd_hotspot(dd, i, x, y, xfwhm, yfwhm, rate)
pointer		dd			#I detector structure
int		i			#I hot spot index
real		x, y			#O hot spot position
real		xfwhm, yfwhm		#O hot spot fwhm
real		rate			#O counts per second

pointer		hs

begin
	if (dd == 0) return NO
	if (i < 1 || i > DD_HOTSPOT_LEN(dd)) return NO

	hs = DD_HOTSPOT(dd, i)
	if (hs == 0) return NO

	x = HS_X(hs)
	y = HS_Y(hs)

	xfwhm = HS_XFWHM(hs)
	yfwhm = HS_YFWHM(hs)

	rate = HS_RATE(hs)

	return YES
end

#	Returns pointers to an array of psfs.  psfs are
#	wavelength dependent, so each psf is characterized
#	by a wavelength and the weighted of two gaussians.
#	A return status of YES indicates success.
int procedure	edd_psfs (dd, length, wavelength, fwhm1, fwhm2, ratio)
pointer		dd			#I detector structure
int		length			#O length of data arrays
pointer		wavelength		#O pointer to real wavelength values
pointer		fwhm1			#O pointer to real fwhm values
pointer		fwhm2			#O pointer to real fwhm values
pointer		ratio			#O pointer to real fwhm ratio values

pointer		pd

begin
	if (dd == 0) return NO

	pd = DD_PSF(dd)
	if (pd == 0) return NO

	length = PD_LENGTH(pd)
	wavelength = PD_WAVELENGTHS(pd)
	fwhm1 = PD_FWHM1(pd)
	fwhm2 = PD_FWHM2(pd)
	ratio = PD_RATIO(pd)

	return YES
end

#	A return status of YES indicates success.
int procedure	edd_boresight(dd, w, x, y, z)
pointer		dd
double		w, x, y, z

begin
	if (dd == 0) return NO

	w = DD_BORESIGHT_W(dd)
	x = DD_BORESIGHT_X(dd)
	y = DD_BORESIGHT_Y(dd)
	z = DD_BORESIGHT_Z(dd)

	return YES
end

#	Returns the number of orders for the given detector
#	A return status of YES indicates success.
int procedure	edd_ea_orders(dd, orders)
pointer		dd			#I detector structure
int		orders			#O number of orders

begin
	orders = 0
	if (dd == 0) return NO

	orders = DD_EA_LEN(dd)

	return YES
end

#	Returns the characteristics of the detectors nth order.
#	A return status of YES indicates success.
int procedure	edd_ea_order(dd, order, length, wavelength, effarea)
pointer		dd			#I detector structure
int		order			#I order
int		length			#O length of data arrays
pointer		wavelength		#O pointer to real wavelengths
pointer		effarea			#O pointer to real effective areas

pointer		eao

begin
	length = 0
	if (dd == 0) return NO
	if (order < 1 || order > DD_EA_LEN(dd)) return NO

	eao = DD_EA(dd, order)
	if (eao == 0) return NO

	length = EAO_LENGTH(eao)
	wavelength = EAO_WAVELENGTH(eao)
	effarea = EAO_EFFECTIVE_AREA(eao)

	return YES
end
