# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  eddops.x
# 	Author	:  ericco
# 	Date	:  04/30/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) eddops.x -r1.1 04/30/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	04/30/92   ericco	date and time created 92/04/30 00:26:50 by ericco
# H-
# 
#

#	External interface to the EUV detector data structures
#	See also edd.x

include	"dd.h"
include	"det.h"
include	"qd.h"

#	Read in detector corresponding to a given label.
procedure	edd_new(dd, et, label)
pointer		dd			#O new detector structure
pointer		et			#I EUV table pointer
char		label[ARB]		#I detector label

int		i
int		len
pointer		detectors
pointer		det
pointer		eths
pointer		qd, qds
int		qd_len

begin
  len = ARB
  detectors = 0

  call et_add_string(et, "name", label)
  call et_det_read(et, detectors, len)
  call et_pop_restriction(et)

  if (len < 1) {
    dd = 0
    return
  }
  do i = 2, len {
    det = Meml[detectors+i-1]
    call det_close(det)
  }

  call dd_malloc(dd)
  DD_DETECTOR(dd) = Meml[detectors]
  det = DD_DETECTOR(dd)

  call mfree(detectors, TY_LONG)

  # find associated detector information
  call et_open_new(eths, DET_HOTSPOTS(det), et)
  DD_HOTSPOT_PTR(dd) = 0
  if (eths == 0) {
    DD_HOTSPOT_LEN(dd) = 0
  } else {
    call et_add_string(eths, "detector", DET_NAME(det))
    DD_HOTSPOT_LEN(dd) = ARB
    call et_hs_read(eths, DD_HOTSPOT_PTR(dd), DD_HOTSPOT_LEN(dd))
    call et_close(eths)
  }

  call et_open_new(eths, DET_PSF(det), et)
  if (eths == 0) {
    DD_PSF(dd) = 0
  } else {
    call et_pd_get(eths, DD_PSF(dd))
    call et_close(eths)
  }

  call et_open_new(eths, DET_EFFAREA(det), et)
  DD_EA_PTR(dd) = 0
  if (eths == 0) {
    DD_EA_LEN(dd) = 0
  } else {
    DD_EA_LEN(dd) = ARB
    call et_eao_read(eths, DD_EA_PTR(dd), DD_EA_LEN(dd))
    call et_close(eths)
  }

  call et_open_new(eths, "quaternion", et)
  DD_BORESIGHT_W(dd) = 1.0
  DD_BORESIGHT_X(dd) = 0.0
  DD_BORESIGHT_Y(dd) = 0.0
  DD_BORESIGHT_Z(dd) = 0.0
  if (eths != 0) {
    call et_add_string(eths, "detector", label)
    call et_add_string(eths, "name", "boresight")
    qd_len = 1
    call malloc(qds, 1, TY_LONG)
    call et_qd_read(eths, qds, qd_len)
    call et_pop_restriction(et)
    call et_pop_restriction(et)
    if (qd_len > 0) {
      qd = Meml[qds]
      DD_BORESIGHT_W(dd) = QD_W(qd)
      DD_BORESIGHT_X(dd) = QD_X(qd)
      DD_BORESIGHT_Y(dd) = QD_Y(qd)
      DD_BORESIGHT_Z(dd) = QD_Z(qd)
      call qd_free(qd)
    }
    call mfree(qds, TY_LONG)
    call et_close(eths)
  }
end

#	Release detector resources
procedure	edd_close(dd)
pointer		dd			#I/O detector structure

int		i

begin
  if (dd == 0) return

  do i = 1, DD_HOTSPOT_LEN(dd) {
    call hs_free(DD_HOTSPOT(dd,i))
  }
  if (DD_HOTSPOT_LEN(dd) != 0)	call mfree(DD_HOTSPOT_PTR(dd), TY_LONG)

  call det_close(DD_DETECTOR(dd))
  call pd_close(DD_PSF(dd))

  do i = 1, DD_EA_LEN(dd) {
    call eao_close(DD_EA(dd,i))
  }
  if (DD_EA_LEN(dd) != 0)	call mfree(DD_EA_PTR(dd), TY_LONG)

  call dd_free(dd)
  dd = 0
end
