# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  emask.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) emask.x -r1.4 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	09/28/92   akeith	put in debug statements
# 	10/29/91   ericco	changed threshold check for symmetry
# 	05/07/91   ericco	date and time created 91/05/07 14:08:41 by ericco
# H-
# 
#

#	Revised: 9/28/92

include	<imhdr.h>
include	<edata.h>
include	<dbg.h>

#	Get mask buffer

procedure	em_gbuf (ptr, buf)
pointer		ptr				#I mask data
pointer		buf				#O output buffer

begin
DEBUG(
	call dbg_printf("em_gbuf,fun", "em_gbuf()\n")
)
  buf = VEC_ELEMENT(ptr, 4)
end

#	Read mask from text database.

procedure	em_read (dt, rec, ptr, lp)	# read a mask
pointer		dt				#I text database
int		rec				#I record id
pointer		ptr				#O mask data
pointer		lp				#I/O output buffer

int		dtgeti()
int		len
pointer		x1, x2
int		index

begin

DEBUG(
	call dbg_printf("em_read,fun", "em_read()\n")
)
	call ev_new(ptr, long(4))
	VEC_ELEMENT(ptr, 3) = dtgeti(dt, rec, "Offset")

	len = dtgeti(dt, rec, "Length")
	call ev_new(VEC_ELEMENT(ptr, 1), long(len)); x1 = VEC_ELEMENT(ptr, 1)
	call ev_new(VEC_ELEMENT(ptr, 2), long(len)); x2 = VEC_ELEMENT(ptr, 2)

	do index = 1, len {
		call dtscan(dt)
		call gargl (VEC_ELEMENT(x1, index))
		call gargl (VEC_ELEMENT(x2, index))
	}
	call eb_copybuffer(VEC_ELEMENT(ptr,4), lp, 0)
	lp = VEC_ELEMENT(ptr, 4)

end

#	Write a mask to a text database

procedure	em_write (dt, record, ptr)	# read a mask

pointer		dt				#I text database
char		record[ARB]			#I record name
pointer		ptr				#I mask data

int		len
pointer		x1, x2
long		started
long		i, j
long		index

begin

DEBUG(
	call dbg_printf("em_write,func", "em_write() record = %s\n")
	call pargstr(record)
)
	call dtptime(dt)
	call dtput(dt, "begin\t%s\n")
	call pargstr(record)

	started = VEC_ELEMENT(ptr,3)
	x1 = VEC_ELEMENT(ptr, 1)
	len = VEC_SIZE(x1)
	x2 = VEC_ELEMENT(ptr, 2)

	for (i = 1; i <= len && VEC_ELEMENT(x1,i) == VEC_ELEMENT(x2,i); i = i + 1);
	for (j = len; j >= 1 && VEC_ELEMENT(x1,j) == VEC_ELEMENT(x2,j); j = j - 1);

	call dtput(dt, "\tOffset\t%d\n"); call pargl(started + i - 1)
	call dtput(dt, "\tLength\t%d\n"); call pargl(j - i + 1)
	do index = i, j {
		call dtput(dt,"\t\t%d\t%d\n")
		call pargl (VEC_ELEMENT(x1, index))
		call pargl (VEC_ELEMENT(x2, index))
	}

end

#	Release mask resources
procedure	em_free (ptr)		# free a mask
pointer		ptr			#I mask data

begin
DEBUG(
	call dbg_printf("em_free,fun", "em_free()\n")
)
  call ev_free(VEC_ELEMENT(ptr, 1))
  call ev_free(VEC_ELEMENT(ptr, 2))
  call ev_free(ptr)
end

#	Compare a set of positions with a mask
procedure	em_check (ptr, x, y, t, len)
pointer		ptr
long		x[len], y[len], t[len]
int		len

int		index
long		start
long		finish
int		i
pointer		x1, x2

begin

DEBUG(
	call dbg_printf("em_check,func", "em_check()\n")
)
	start = VEC_ELEMENT(ptr, 3) - 1
	x1 = VEC_ELEMENT(ptr, 1)
	x2 = VEC_ELEMENT(ptr, 2)
	finish = VEC_SIZE(x1)

	do index = 1, len {
		i = y[index] - start
		if (i < 1 || i > finish) {
			t[index] = 0
		} else
		if (x[index] < VEC_ELEMENT(x1,i) || x[index] > VEC_ELEMENT(x2,i)) {
			t[index] = 0
		} else {
			t[index] = 1
		}
	}

end

#	Generate a simple mask from an image using a threshold
procedure	em_threshold (im, thr, ptr)	# mask from threshold image
pointer		im				#I image
short		thr				#I threshold
pointer		ptr				#O mask data

int		len
pointer		x1, x2
int		width
pointer		buf, imgl2s()
int		sp, ep
int		out
int		index
pointer		x

begin
DEBUG(
	call dbg_printf("em_threshold,func", "em_threshold()\n")
)
  call ev_new(ptr, long(3))
  len = IM_LEN(im,2)
  width = IM_LEN(im,1)
  call ev_new(VEC_ELEMENT(ptr, 1), long(len)); x1 = VEC_ELEMENT(ptr, 1)
  call ev_new(VEC_ELEMENT(ptr, 2), long(len)); x2 = VEC_ELEMENT(ptr, 2)
  VEC_ELEMENT(ptr, 3) = 0

  do index = 1, len {
    buf = imgl2s (im, index)
    sp = 0
    ep = 0
    out = 1
    do x = buf, buf+width-1 {
      if (out == 1 && Mems[x] > thr) {
	sp = x - buf + 1
	out = 0
      }
      if (out == 0 && Mems[x] <= thr) {
	ep = x - buf + 1
	break
      }
    }
    if (out == 0 && ep == 0) ep = width
    VEC_ELEMENT(x1, index) = sp
    VEC_ELEMENT(x2, index) = ep
  }
end
