# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  epproc.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) epproc.x -r1.9 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	12/27/96   mabbott	changes to support parameter overrides
# 	06/02/94   akeith	Put in a large number of spaces to make the code visible
# 	05/12/94   akeith	took out debug statements
# 	09/28/92   akeith	put in debug statements
# 	07/07/92   ericco	adjusted processing algorithm to be more efficient and corrected hanging buffers by not forcing them to be skipped
# 	10/29/91   ericco	replaced eu_malloc calls.
# Modified pipeline to push events as far as possible
# through the processing steps.  This happens when
# flushing and when processing intermediate results.
# 	05/07/91   ericco	date and time created 91/05/07 14:09:55 by ericco
# H-
# 
#

#	Revised: 12/27/96

include	<eerror.h>
include	<edata.h>
include <override.h>
include	<dbg.h>
include	"ep.h"

#	Execute a single run through a list of processes.
procedure	ep_loop (prc, flag) # run a processs loop
pointer		prc		#I processing list
bool		flag		#I end of data flag

pointer		pl, ed
int		force
pointer		rp
int		num
int		index
pointer		buf
int		i
int		det
int		len
int		j
pointer		sp
pointer		obuf
int		li

begin

    if (NUM_PROC(prc) < 2) {          # must read and write at least
        flag = TRUE
        return
    }

    pl = EP_PRC_PTR(prc)
    ed = EP_OBS_PTR(prc)
    rp = VEC_ELEMENT(pl,1)
    num = NUM_PROC(prc)

    call smark(sp)
    call salloc(obuf, num, TY_LONG)

    call ebp_morepht(ED_PHTBUF(ed), flag)

    if (flag == TRUE) {

        force = 1
        index = 0
        call es_popany(EPP_BUF(rp), index, buf)
        flag = FALSE
        if (buf == 0) {

            flag = TRUE
            # really force everything out
            call es_glength(EPP_BUF(rp), len)
            do index = 1, len {

                det = index + EPP_FIELD2(rp) - 1
                buf = 0
                Meml[obuf+1-1] = buf
                for (i = 2; i <= num; i = i + 1) {

                    call ep_run(VEC_ELEMENT(pl,i), det, buf, force)
                    Meml[obuf+i-1] = buf
                    if (buf != 0) {

                        force = 0            # process reminder without forcing it
                        li = num
                        for (j = i+1; j <= num; j = j + 1) {

                            call ep_run(VEC_ELEMENT(pl,j), det, buf, force)
                            Meml[obuf+j-1] = buf
                            while (buf == 0 && j > i) {

                                j = j - 1
                                if (li > j) {
                                    buf = Meml[obuf+j-1]
                                    li = j
                                }
                            }

                            if (buf == 0) break
                        }

                        i = i - 1 # rerun task
                        force = 1
                        buf = 0
                    }

                }
            }
            call sfree(sp)
            return
        }

    } else {

        force = 0

#  Otherwise there are some photons to process
#  First process "reads" photons

        call ebp_gbuf(ED_PHTBUF(ed), buf)
        call es_split(EPP_BUF(rp), buf, EPP_FIELD(rp))

        index = 0
        call es_popfull(EPP_BUF(rp), index, buf)
        if (buf == 0) return
    }

# One of the pipe streams is ready

    det = index
    det = det + EPP_FIELD2(rp) - 1
    Meml[obuf+1-1] = buf
#   li = num

    for (i = 2; i <= num; i = i + 1) {

        call ep_run(VEC_ELEMENT(pl,i), det, buf, force)
        Meml[obuf+i-1] = buf
        if (buf == 0) Meml[obuf+i-2] = 0
        if (buf == Meml[obuf+i-2]) Meml[obuf+i-2] = 0

        while (buf == 0 && i > 1) {
            i = i - 1
#            if (li > i) {
                buf = Meml[obuf+i-1]
#                li = i
#            }
        }
        if (buf == 0) break
    }

    call sfree(sp)

end


#	Initialize processing pipeline from text database.
#	It also allocates the EUV Pipeline structure EP_.

procedure	ep_proc (dt, record, override, prc) # open a processing environment

pointer		dt			#I text database
char		record[ARB]		#I record name
pointer     override    #I parameter override structure
pointer		prc			#O process data

int	rec, dtlocate()
pointer	ptr
int	type
bool	streq()
pointer	dtmap()
pointer	lp
int	i

pointer	obs, ref, sp

begin

  rec = dtlocate(dt, record)
  call dtgtype(dt, rec, type)
  if (type != EP_PIPE) call eerror(EES_REQPIPE)

# Allocate space for the EP_ structure, store the pointer
#	and the path to the text database.

  call smark(sp)
  call ev_new_clear(prc, EP_LENGTH)
  EP_PRC_DT(prc) = dt
  call eu_newstr(".", EP_PRC_PATH_PTR(prc))

# Get and store the path to the referance text database.

  if (streq(REFERENCE(override), "")) {
	call dtgptr(dt, rec, "RefPath", ptr)
	call eu_newstr(Memc[ptr], EP_REF_PATH_PTR(prc))
	call eu_mfree(ptr, TY_CHAR)
  } else {
	call malloc(ptr, 1024, TY_CHAR)
	call fnldir(REFERENCE(override), Memc[ptr], SZ_LINE)
	if (streq(Memc[ptr], "")) call strcpy(".", Memc[ptr], SZ_LINE)
	call eu_newstr(Memc[ptr], EP_REF_PATH_PTR(prc))
	call mfree(ptr, TY_CHAR)
  }

# Get and store the path to the observation text database.

  if (streq(TABLEINDEX(override), "")) {
	call dtgptr(dt, rec, "ObsPath", ptr)
	call eu_newstr(Memc[ptr], EP_OBS_PATH_PTR(prc))
	call eu_mfree(ptr, TY_CHAR)
  } else {
	call malloc(ptr, 1024, TY_CHAR)
	call fnldir(TABLEINDEX(override), Memc[ptr], SZ_LINE)
	if (streq(Memc[ptr], "")) call strcpy(".", Memc[ptr], SZ_LINE)
	call eu_newstr(Memc[ptr], EP_OBS_PATH_PTR(prc))
	call mfree(ptr, TY_CHAR)
  }

# Get and store the Observation's text database pointer.

  if (streq(TABLEINDEX(override), "")) {
	call dtgptr(dt, rec, "Observation", obs)
	call eu_file(EP_OBS_PATH(prc), Memc[obs], ptr)   
	call eu_mfree(obs, TY_CHAR)
	EP_OBS_DT(prc) = dtmap(Memc[ptr], READ_ONLY)
  } else {
	call eu_newstr(TABLEINDEX(override), ptr)   
	EP_OBS_DT(prc) = dtmap(Memc[ptr], READ_ONLY)
    call eu_free(ptr, TY_CHAR)
  }

# Get and store the Reference's text database pointer.

  if (streq(REFERENCE(override), "")) {
	call dtgptr(dt, rec, "Reference", ref)
	call eu_file(EP_REF_PATH(prc), Memc[ref], ptr)
	call eu_mfree(ref, TY_CHAR)
	EP_REF_DT(prc) = dtmap(Memc[ptr], READ_ONLY)
  } else {
	call eu_newstr(REFERENCE(override), ptr)
	EP_REF_DT(prc) = dtmap(Memc[ptr], READ_ONLY)
    call eu_free(ptr, TY_CHAR)
  }
  call sfree(sp)

# Read the Observation's text database and store the pointer
#	to the ED_ structure in EP_OBS_PTR.

  call ed_read(EP_OBS_DT(prc), override, EP_OBS_PTR(prc))

  call dtgvstr(dt, rec, "ProcessList", EP_PRC_LST(prc))
  call ev_new_clear(EP_PRC_PTR(prc), NUM_PROC(prc))

# Open all the st table for access and for each table store
#	a pointer to the GT_ structure in ED_TABLE_PTR.

  call ed_otab(EP_OBS_PTR(prc), EP_OBS_PATH(prc))

# This still needs to be explaned. It's a bit complicated to explain,
#	but this basically sets up the EUV buffer photon operations
#	structure, EBP_.

  call edopht(EP_OBS_PTR(prc))

  lp = NULL
  do i = 1, NUM_PROC(prc) {
    call ep_start(prc, lp,
		  VEC_STRING(EP_PRC_LST(prc),i),
		  VEC_ELEMENT(EP_PRC_PTR(prc),i))
  }

end

#	Release processing pipeline resources
procedure	ep_free(state)	# relase resources
pointer		state		#I process list

int		i
pointer		ed

begin
  ed = EP_OBS_PTR(state)
  call edctab(ed)
  call edcpht(ed)
  call ed_casp(ed)
  call ed_free(ed)

  do i = 1, NUM_PROC(state) {
    call ep_finish(VEC_ELEMENT(EP_PRC_PTR(state),i))
  }

  call ev_free(EP_PRC_PTR(state))
  call ev_free_string(EP_PRC_LST(state))

  call dtunmap(EP_OBS_DT(state))
  call dtunmap(EP_REF_DT(state))

  call eu_mfree(EP_PRC_PATH_PTR(state), TY_CHAR)
  call eu_mfree(EP_REF_PATH_PTR(state), TY_CHAR)
  call eu_mfree(EP_OBS_PATH_PTR(state), TY_CHAR)

  call ev_free(state)
end

