# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  epstart.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) epstart.x -r1.13 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	01/13/97   mabbott	added coord override for imaging scripts
# 	12/27/96   mabbott	changes to support parameter override of source position
# 	06/29/94   akeith	Now remaps deepserv and scanners
# 	05/12/94   akeith	took out debug statements
# 	12/11/92   ericco	removed info pointer
# 	12/11/92   ericco	added info parameter
# 	12/06/92   ericco	added looklist
# 	12/04/92   ericco	added print modulke
# 	09/28/92   akeith	put in debug statements
# 	10/29/91   ericco	fixed eu_malloc.  Changed field names of aspect record
# 	05/07/91   ericco	date and time created 91/05/07 14:10:13 by ericco
# H-
# 
#

#	Revised: 1/13/97

include	<eerror.h>
include <edata.h>
include <override.h>
include <dbg.h>
include	"ep.h"

#	Release processing resources
procedure	ep_finish(prc) #E
pointer		prc			#I/O process record

int		i
pointer		pd
int		type

begin
  pd = EPP_PERDET(prc)
  type = EPP_TYPE(prc)
  switch (type) {
    case EP_READ:
      call es_close (EPP_BUF(prc))
    case EP_TIMELIST:
      do i = 1, VEC_SIZE(pd)
        if (VEC_ELEMENT(pd, i) != 0) call et_old(VEC_ELEMENT(pd, i))
    case EP_CONVLIST:
      do i = 1, VEC_SIZE(pd)
        if (VEC_ELEMENT(pd, i) != 0) call ex_close(VEC_ELEMENT(pd, i))
    case EP_CALCLIST:
      do i = 1, VEC_SIZE(pd)
        if (VEC_ELEMENT(pd, i) != 0) call ea_close(VEC_ELEMENT(pd, i))
    case EP_PRINTLIST:
      do i = 1, VEC_SIZE(pd)
        if (VEC_ELEMENT(pd, i) != 0) call cp_close(VEC_ELEMENT(pd, i))
    case EP_LOOKLIST:
      do i = 1, VEC_SIZE(pd)
        if (VEC_ELEMENT(pd, i) != 0) call clu_close(VEC_ELEMENT(pd, i))
    case EP_MASKLIST:
      do i = 1, VEC_SIZE(pd)
        if (VEC_ELEMENT(pd, i) != 0) call em_free(VEC_ELEMENT(pd, i))
    case EP_CORRLIST:
      do i = 1, VEC_SIZE(pd)
        if (VEC_ELEMENT(pd, i) != 0) call ec_free(VEC_ELEMENT(pd, i))
    case EP_ASPLIST:
      EPP_ASP(prc) = 0
      call eu_mfree(EPP_DATA(prc), TY_DOUBLE)
      do i = 1, VEC_SIZE(pd)
        if (VEC_ELEMENT(pd, i) != 0) call eu_mfree(VEC_ELEMENT(pd, i), TY_DOUBLE)
    case EP_IMASPLIST:
      EPP_ASP(prc) = 0
      do i = 1, VEC_SIZE(pd)
        if (VEC_ELEMENT(pd, i) != 0) call eu_mfree(VEC_ELEMENT(pd, i), TY_DOUBLE)
    case EP_WRITELIST:
      do i = 1, VEC_SIZE(pd)
        if (VEC_ELEMENT(pd, i) != 0) call ew_close(VEC_ELEMENT(pd, i))
    default: call eerror(EES_REQLIST)
  }

  call ev_free(pd)
  call ev_free(prc)
end

#	Initialize a processing list element

procedure	ep_start (ep, lp, name, prc) # initialize a process

pointer		ep			#I processing data
pointer		lp			#I last process buffer
char		name[ARB]		#I record name
pointer		prc			#O process record

int	record
pointer	dets
int	i
int	dmin, dmax
int	type
pointer		tmp
pointer		id
bool		strne()
pointer		buf
pointer		dt
pointer		path
pointer		tdt, trec, tpath, ptr
long		sz, dtgeti()
pointer		ed
pointer	    spor

begin

	ed = EP_OBS_PTR(ep)
	call ev_new_clear(prc, EPP_LENGTH)
	call ev_new_clear(EPP_PERDET(prc), MAX_DETECTORS)

	call ep_dtfind(ep, name, dt, record, path)

	dets = ED_DETECTORS(ed)
	call dtgtype(dt, record, type)
	EPP_TYPE(prc) = type

	if (lp == NULL) { # must be a read

		if (type != EP_READ) call eerror(EES_REQREAD)	# Get the list of
		call dtgvstr(dt, record, "DetectorList", tmp)   # detectors, store
								# names in a vector
								# list, tmp.
		do i = 1, VEC_SIZE(tmp) {

		  #   Convert the name into an id number.
		  call eu_detect_id(VEC_STRING(tmp,i), id)

		  #   Check that this is a valid detector list as recorded
		  #   by ed_rddets() and stored in ED_DETECTORS(ed).
		  if (VEC_ELEMENT(dets, id) == 0) call eerror(EES_DETOFF)

		  #   Find the min and max detector id number.
		  if (i == 1) {dmin = id; dmax = id}
		  else if (dmin > id) dmin = id;
		  else if (dmax < id) dmax = id;
		}
		call ev_free_string(tmp)

		call ebp_gbuf(ED_PHTBUF(ed), buf)      # get EBP_BUFFER of ED_PHTBUF.

		sz = dtgeti(dt, record, "EventsPerBuffer")

		call dtgvstr(dt, record, "OutputFields", tmp)
		call es_open1(tmp, dmin, dmax, sz, EPP_BUF(prc))
		call ev_free_string(tmp)

		call dtgfield(dt, record, "DetectorField", EPP_FIELD(prc))

		call eb_gcolumn(buf, EPP_FIELD(prc), EPP_FIELD(prc))

		EPP_FIELD2(prc) = dmin

		call es_gfbuf(EPP_BUF(prc), lp)	# set up last buffer pointer

	} else {

		if (type == EP_READ) call eerror(EES_NOREAD)

		switch (type) {

		  case EP_CALCLIST:
			call ep_rdrec(ep, dt, record, "ProcessList",
					EP_CALC, EPP_PERDET(prc), lp)

		  case EP_PRINTLIST:
			call ep_rdrec(ep, dt, record, "ProcessList",
					EP_PRINTBUFFER, EPP_PERDET(prc), lp)

		  case EP_LOOKLIST:
			call ep_rdrec(ep, dt, record, "ProcessList",
					EP_LOOKUP, EPP_PERDET(prc), lp)

		  case EP_CONVLIST:
			call ep_rdrec(ep, dt, record, "ProcessList",
						EP_CONV, EPP_PERDET(prc), lp)

		  case EP_TIMELIST:
			call dtgptr(dt, record, "AcceptMask", tmp)
			if (strne(Memc[tmp],"Out")) EPP_MASKIN(prc) = 1;
			else EPP_MASKIN(prc) = 0
			call eu_mfree(tmp, TY_CHAR)
			call dtgfield(dt, record, "TimeField", EPP_FIELD(prc))
			call eb_gcolumn(lp, EPP_FIELD(prc), EPP_FIELD(prc))
			call ep_rdrec(ep, dt, record, "ProcessList",
						EP_TIME, EPP_PERDET(prc), lp)

		  case EP_MASKLIST:
			call dtgptr(dt, record, "AcceptMask", tmp)
			if (strne(Memc[tmp],"Out")) EPP_MASKIN(prc) = 1;
			else EPP_MASKIN(prc) = 0
			call eu_mfree(tmp, TY_CHAR)
			call dtgfield(dt, record, "XField", EPP_FIELD(prc))
			call eb_gcolumn(lp, EPP_FIELD(prc), EPP_FIELD(prc))
			call dtgfield(dt, record, "YField", EPP_FIELD2(prc))
			call eb_gcolumn(lp, EPP_FIELD2(prc), EPP_FIELD2(prc))
			call ep_rdrec(ep, dt, record, "ProcessList",
						EP_MASK, EPP_PERDET(prc), lp)

		  case EP_CORRLIST:
			call dtgfield(dt, record, "XInputField", EPP_FIELD(prc))
			call eb_gcolumn(lp, EPP_FIELD(prc), EPP_FIELD(prc))
			call dtgfield(dt, record, "YInputField", EPP_FIELD2(prc))
			call eb_gcolumn(lp, EPP_FIELD2(prc), EPP_FIELD2(prc))
			call ep_rdrec(ep, dt, record, "ProcessList",
						EP_CORR, EPP_PERDET(prc), lp)

		  case EP_ASPLIST:
#			call dtgfield(dt, record, "XInputField", EPP_FIELD(prc))
			call dtgfield(dt, record, "SourceAngleXField", EPP_FIELD(prc))
			call eb_gcolumn(lp, EPP_FIELD(prc), EPP_FIELD(prc))
#			call dtgfield(dt, record, "YInputField", EPP_FIELD2(prc))
			call dtgfield(dt, record, "SourceAngleYField", EPP_FIELD2(prc))
			call eb_gcolumn(lp, EPP_FIELD2(prc), EPP_FIELD2(prc))
			call dtgfield(dt, record, "TimeField", EPP_FIELD3(prc))
			call eb_gcolumn(lp, EPP_FIELD3(prc), EPP_FIELD3(prc))

			# fetch the source position
			# look for parameter overrides
			spor = ED_SRCPOS(ed)
			if (IS_INDEFD(SRC_RA(spor)) || IS_INDEFD(SRC_DEC(spor))) {
			  # read from ST table specified by Source record 
			  call dtgptr(dt, record, "Source", ptr)
			  call ep_dtfind(ep, Memc[ptr], tdt, trec, tpath)
			  call eu_mfree(ptr, TY_CHAR)
			  call eq_read(tdt, trec, EPP_DATA(prc))
			} else {
			  # use ra & dec from parameters
			  call eq_scalloc(EPP_DATA(prc))
			  call eq_scstore(SRC_RA(spor), SRC_DEC(spor), EPP_DATA(prc))
			}

			if (ED_ASPECT(ed) == 0) {
				call edoasp(ed)
				if (ED_ASPECT(ed) == 0) call eerror(EES_NOASP)
			}
			EPP_ASP(prc) = ED_ASPECT(ed)
			call ep_rdrec(ep, dt, record, "ProcessList",
			EP_QUATERNION, EPP_PERDET(prc), lp)



		  case EP_IMASPLIST:

			call dtgfield(dt, record, "XInputField", EPP_FIELD(prc))
			call eb_gcolumn(lp, EPP_FIELD(prc), EPP_FIELD(prc))
			call dtgfield(dt, record, "YInputField", EPP_FIELD2(prc))
			call eb_gcolumn(lp, EPP_FIELD2(prc), EPP_FIELD2(prc))
			call dtgfield(dt, record, "TimeField", EPP_FIELD3(prc))
			call eb_gcolumn(lp, EPP_FIELD3(prc), EPP_FIELD3(prc))

			if (ED_ASPECT(ed) == 0) {
				call edoasp(ed)
				if (ED_ASPECT(ed) == 0) call eerror(EES_NOASP)
			}
			EPP_ASP(prc) = ED_ASPECT(ed)

			call ep_rdrec(ep, dt, record, "ProcessList",
			              EP_QUATERNION, EPP_PERDET(prc), lp)



		  case EP_WRITELIST:
			call ep_rdrec(ep, dt, record, "ProcessList", EP_WRITE,
						EPP_PERDET(prc), lp)
		  default: call eerror(EES_REQLIST)
		}
	}

end
