# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  eqread.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) eqread.x -r1.10 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	01/13/97   mabbott	added coord override for imaging scripts
# 	12/27/96   mabbott	changes to support parameter overrides
# 	05/12/94   akeith	took out debug statements
# 	07/12/93   tnk	changed tbcgtt to tbcgttn
# 	11/16/92   ericco	added check to look for hour units in star catalog
# 	09/28/92   akeith	put in debug statements
# 	10/29/91   ericco	fixed eu_malloc.  added comments
# 	06/19/91   ericco	added eq_scread to layer star positions over quaternions
# 	05/07/91   ericco	date and time created 91/05/07 14:10:18 by ericco
# H-
# 
#

#	Revised: 1/13/97

include	<eerror.h>
include <edata.h>
include	<math.h>
include <dbg.h>
include <tbset.h>

#	Read a quaternion from a text database
procedure	eq_read(dt, rec, elem)
pointer		dt
int		rec
pointer		elem

int	junk
errchk	dtgad

begin

	call eq_scalloc (elem)
	iferr (call dtgad(dt, rec, "Quaternion", Memd[elem], 4, junk)) {
		iferr(call eq_scread (dt, rec, elem)) {
			call eu_mfree(elem, TY_DOUBLE)
			call eerror(EES_NOQUAT)
		}
	}

end

#	Convert a star catalog position into a quaternion
procedure	eq_scread(dt, rec, elem) # read star catalog
pointer		dt		#I text database
int		rec		#I star catalog record
pointer		elem		#O quaternion converted position

pointer	catp, rap, decp, colp, valuep
bool	streq()
pointer	tp, tbtopn()
int	rows
int	i, row
int	inra, indec, incol
int	len, strlen()
bool	nullflag
double	ra, dec
pointer	buf
errchk	tbtopn()
#int	tbtacc()

char	units[2]

begin

  call dtgptr(dt, rec, "Catalog", catp)
  call dtgptr(dt, rec, "Column", colp)
  call dtgptr(dt, rec, "Value", valuep)
  call dtgptr(dt, rec, "ColumnRA", rap)
  call dtgptr(dt, rec, "ColumnDEC", decp)
  len = strlen(Memc[valuep])
  call eu_malloc (buf, len+1, TY_CHAR)

#  if (tbtacc(Memc[catp]) != YES) {
#    call eerror(EES_NOTAB)
#  }
  tp = tbtopn(Memc[catp], READ_ONLY, 0)
  call tbtrows(tp, rows)
  row = 1
  if (rows > 1) {
    if (streq("ROW", Memc[colp])) {
      i = 1
      call ctoi(Memc[valuep], i, row)
      if (row > rows || row < 1) {
	call eerror(EES_NOQUAT)
      }
    } else {
      call tbcfnd (tp, Memc[colp], incol, 1)
      if (incol == NULL) call eerror(EES_NOQUAT)
      for (row = 1; row <= rows; row = row + 1) {
	call tbcgttn(tp, incol, Memc[buf], nullflag, len, row, row)
	if (nullflag) next
	if (streq(Memc[buf], Memc[valuep])) break
      }
      if (row > rows) call eerror(EES_NOQUAT)
    }
  }
  call tbcfnd (tp, Memc[rap], inra, 1)
  if (inra == NULL) call eerror(EES_NOQUAT)
  call tbcfnd (tp, Memc[decp], indec, 1)
  if (indec == NULL) call eerror(EES_NOQUAT)

  call tbcgtd(tp, inra, ra, nullflag, row, row)
  if (nullflag) call eerror(EES_NOQUAT)

  call tbcigt(inra, TBL_COL_UNITS, units, 2)

  if (units[1] == 'h' || units[1] == 'H') {
    ra = 15.0 * ra
  } else if (units[1] != 'd' && units[1] != 'D') {
	call eerror(EES_NOQUAT)
  }

  call tbcgtd(tp, indec, dec, nullflag, row, row)
  if (nullflag) call eerror(EES_NOQUAT)

  call tbtclo(tp)

  call eu_mfree(buf, TY_CHAR)
  call eu_mfree(catp, TY_CHAR)
  call eu_mfree(colp, TY_CHAR)
  call eu_mfree(valuep, TY_CHAR)
  call eu_mfree(rap, TY_CHAR)
  call eu_mfree(decp, TY_CHAR)

  call eq_scstore(ra, dec, elem)
end

# convert an RA,DEC pair to a quaternion and store
#    RA and DEC must be in degrees
procedure	eq_scstore(ra, dec, elem) # read star catalog

double ra, dec
pointer elem

double rra, rdec

begin
  rra = TWOPI * ra / 360.0
  rdec = TWOPI * dec / 360.0

  call er_qfrompos(rra, rdec, Memd[elem])
end

# allocate space for a quaternion
procedure   eq_scalloc(elem)

pointer elem

begin
	call eu_malloc (elem, 4, TY_DOUBLE)
end
