# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  ewread.gx
# 	Author	:  ericco
# 	Date	:  10/29/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) ewread.gx -r1.45 10/29/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	08/24/98   mabbott	added comments about batch queues
# 	08/21/98   bboyd	commented out section where DISPAXIS gets added
# 	05/26/98   mabbott	added declarations for qp_geti hack
# 	05/26/98   mabbott	added qp_geti hack to fix qp_puti error on dc-flag
# 	02/26/97   mabbott	fixed bug which omitted position overrides for imaging instr.
# 	01/30/97   tnk	took INHERIT from kill list
# 	01/13/97   tnk	moved kill list initialization into subroutine
# 	10/24/96   tnk	added TIMEDEL to kill list
# 	06/06/96   gareth	changed RATE[0-9]* to RATE? RATE?? RATE???
# 	04/24/96   gareth	changed RATE card agin to RATE[0-9]* to be safer
# 	04/24/96   gareth	changed RATE_??? to RATE?* for removal of rate header cards
# 	10/03/95   ericco	added indexing to qpsrt style sort option
# 	01/27/95   ericco	fixed macro name prefix for long to be 'i' instead of 'l'
# 	09/30/94   akeith	add new qpsrt fields to write record
# 	07/25/94   akeith	Fixed mwcs.
# 	06/29/94   akeith	fixed imdir problem for QPSRT only
# 	06/13/94   tnk	added override kill list
# 	06/13/94   akeith	Now writes Deep Survey remap qppoes
# 	03/10/94   tnk	added hcpkill call (and associated table pointers)
# 	03/09/94   tnk	changed EV_NUMOR, EV_NUMIR to EVENTSOR, EVENTSIR
# 	02/03/94   ericco	added in/out of range header cards
# 	05/03/93   ericco	fixed up sorting temp files
# 	03/24/93   akeith	put event buffer var
# 	03/09/93   ericco	added pros headers and HeaderList field
# 	03/05/93   ericco	added WCS options and PROS option
# 	02/12/93   ericco	added optional ADVICE ability
# 	02/01/93   ericco	possible sort optimization
# 	01/26/93   ericco	added optional PageSize and BucketLength fields
# 	01/26/93   ericco	apparently ADVICE doesn't speed up photon sorting
# 	12/23/92   ericco	fixed output field offsets
# 	12/18/92   akeith	put in qpsrt_ functions and the new sort options
# 	11/17/92   ericco	remove qp_add_macro
# 	09/28/92   akeith	put in debug statements
# 	08/12/92   ericco	added wcs attributes
# 	08/12/92   ericco	added dispaxis and dc-flag header parameters
# 	08/11/92   ericco	added exptime defattr
# 	08/11/92   ericco	removed w0 and wpc header values
# 	06/23/92   ericco	added output alias option
# 	06/17/92   ericco	add check for INDEF value on type conversion
# 	04/30/92   ericco	corrected byte offset computation
# 	10/30/91   ericco	added fix for large event lists
# 	10/29/91   ericco	date and time created 91/10/29 15:46:41 by ericco
# H-
# 
#

# @(#)ewread.gx	1.45	Revised: 8/24/98

#include <config.h>
include <mach.h>
include <iraf.h>
include <qpset.h>
include <imhdr.h>
include <fset.h>
include <math.h>

include <eerror.h>
include	<edata.h>
include	<override.h>
include	<qpsrt.h>
include	<dbg.h>
include "ep.h"

#	This numbers should be computed from the total number
#	of events expected.  Or they should be selected by the
#	user.  These values make it work on long lists.
define	EUV_PAGESIZE	(4 * 1024)
define	EUV_BUCKETLEN	(1 * EUV_PAGESIZE)

#	Name of the event out-of-range header card
define	EV_OOR		"EVENTSOR"
define	EV_IOR		"EVENTSIR"

define	STR_LEN		1024

#	EQP write data structure
define		EQP_QP	VEC_ELEMENT($1,1)	# QPOE file
define		EQP_MW	VEC_ELEMENT($1,2)	# MWCS
define		EQP_CT	VEC_ELEMENT($1,3)	# world to phys transform
define		EQP_IO	VEC_ELEMENT($1,4)	# QPIO
define		EQP_TFD	VEC_ELEMENT($1,5)	# temp file
define		EQP_IM	VEC_ELEMENT($1,6)	# raw count image
define		EQP_RUN	VEC_ELEMENT($1,7)	# mode flag
define		EQP_XO	VEC_ELEMENT($1,8)	# x offset
define		EQP_YO	VEC_ELEMENT($1,9)	# y offset
define		EQP_COL	VEC_ELEMENT($1,10)	# column pointers
define		EQP_TYP	VEC_ELEMENT($1,11)	# column type
define		EQP_BUF	VEC_ELEMENT($1,12)	# output buffer
define		EQP_PTR	VEC_ELEMENT($1,13)	# short output pointer
define		EQP_XPTR VEC_ELEMENT($1,14)	# short x pointer
define		EQP_YPTR VEC_ELEMENT($1,15)	# short y pointer
define		EQP_IMNM VEC_ELEMENT($1,16)	# image name if delete-able
define		EQP_ARRAY VEC_ELEMENT($1,17)	# use array not image
define		EQP_XLEN VEC_ELEMENT($1,18)	# x axis length
define		EQP_YLEN VEC_ELEMENT($1,19)	# y axis length
define		EQP_2PASS VEC_ELEMENT($1,20)	# sort algorithm
define		EQP_TFD2 VEC_ELEMENT($1,21)	# temp file 2
define		EQP_ORDERED VEC_ELEMENT($1,22)	# in order flag
define		EQP_SORT VEC_ELEMENT($1,23)	# do we sort?
define		EQP_QPSRT VEC_ELEMENT($1,24)	# do we use qpsrt?
define		EQP_TP VEC_ELEMENT($1,25)	# timing profile
define		EQP_BUFSIZE VEC_ELEMENT($1,26)	# tmp file buf size
define		EQP_BUFSIZE2 VEC_ELEMENT($1,27)	# tmp2 file buf size
define		EQP_EVBUFSZ VEC_ELEMENT($1,28)	# event buf size
define		EQP_OOR	VEC_ELEMENT($1,29)	# out of range events
define		EQP_IOR	VEC_ELEMENT($1,30)	# in range events

define		EQP_LENGTH	long(30)


# Utility function to strip the white space from a string



#	Process a event buffer
procedure	ew_proc (ew, eb, force)	# process buffer
pointer		ew		#I write state
pointer		eb		#I input buffer
int		force		#I force output

begin
  call ewb_proc(ew, EQP_CT(ew), EQP_XO(ew), EQP_YO(ew), # EQP_SPLIT(ew),
		eb, force)
end

#	Remove a file associated with a "temp" file descriptor
procedure	rm_tfd(fd)	# delete a temp file
int	fd			#I file descriptor

pointer	sp, str
errchk	delete

begin
  if (fd == 0) return
  call smark(sp)
  call salloc(str, STR_LEN, TY_CHAR)
  call fstats(fd, F_FILENAME, Memc[str], STR_LEN)
  call close(fd)
  iferr(call delete(Memc[str]));
  call sfree(sp)
end

#	Release write resources
procedure	ew_close (ew) # release write resources
pointer		ew		#I write state

begin

  if (EQP_SORT(ew) == NO) {
    call rm_tfd(EQP_TFD(ew))
    call rm_tfd(EQP_TFD2(ew))
  }

  if( EQP_SORT(ew) == YES ) {

    if (EQP_QPSRT(ew) == YES) {
      call eprintf("Sorting ...\n")
      call qpsrt_close(EQP_IM(ew))
      EQP_ORDERED(ew) = YES
    } else {
      if (EQP_ARRAY(ew) == YES) {
        call eu_mfree(EQP_IM(ew), TY_LONG)
      } else {
        call imunmap(EQP_IM(ew))
      }
    }

  }

  call mw_ctfree (EQP_CT(ew))
  call mw_close(EQP_MW(ew))

  call qp_addl(EQP_QP(ew), EV_OOR, EQP_OOR(ew), "Out of range events")
  call qp_addl(EQP_QP(ew), EV_IOR, EQP_IOR(ew), "In range events")
  if ( EQP_ORDERED(ew) == YES) call qpio_mkindex(EQP_IO(ew),"")

  call qpio_close (EQP_IO(ew))
  call qp_close(EQP_QP(ew))

  if (EQP_IMNM(ew) != NULL) {
    call imdelete(Memc[EQP_IMNM(ew)])
    call eu_mfree(EQP_IMNM(ew), TY_CHAR)
  }

#  call tp_print(EQP_TP(ew))
#  call tp_close(EQP_TP(ew))

  call ev_free(EQP_TYP(ew))
  call ev_free(EQP_COL(ew))
  call eu_mfree(EQP_BUF(ew), TY_REAL)
  call ev_free(EQP_PTR(ew))
  call ev_free(ew)

end


procedure	ewr_new_qpoe(qp)
pointer		qp		#I/O new qpoe descriptor

long	now, clktime()
int	datamin[8], datamax[8]
data	datamin/0,0,0,0,0,0,0,0/
data	datamax/0,0,0,0,0,0,0,0/

begin
  now = clktime(0)
  call qp_addf(qp, "cretime", "i", 1, "image creation time", 0)
  call qp_puti(qp, "cretime", now)
  call qp_addf(qp, "modtime", "i", 1, "image modify time", 0)
  call qp_puti(qp, "modtime", now)
  call qp_addf(qp, "limtime", "i", 1, "data min/max update time", 0)
  call qp_puti(qp, "limtime", now)
  call qp_addf(qp, "datamin", "i", 8, "minimum pixel value", 0)
  call qp_write(qp, "datamin", datamin, 8, 1, "i")
  call qp_addf(qp, "datamax", "i", 8, "maximum pixel value", 0)
  call qp_write(qp, "datamax", datamax, 8, 1, "i")
end


#	Read write state
procedure	ew_read(dt, rec, eqp, lp, ed, ep) # read write state
pointer		dt		#I text database
char		rec[ARB]	#I record name
pointer		eqp		#O write structure
pointer		lp		#I input buffer-like
pointer		ed		#I EUV data pointer 
pointer		ep		#I Event process

int	re, dtlocate()
int	i
pointer	ptr

pointer	im, immap(), qpsrt_open()
pointer	impl2l()
pointer	impl1l()

extern	mw_ctfree()
pointer	qp, qp_open ()
extern	qp_close()
int	qpio_open()
extern	qpio_close()

int	axlen[2]
extern	mw_close()
pointer	mw_sctran()
pointer	tmp
errchk	qp_delete, imdelete
int	dtgeti()
int	xo, yo
pointer	outs, col, typ
int	bytes
int	size, si
pointer	name, kind, buf, sp, ptrs, sys
bool	streq()
short	ls, lr
data	ls/'s'/
data	lr/'r'/
pointer	edef, op
bool	bval, dtgetb(), pros
int	lenlp
int	pagesize
int	bucketlen
int	tmpbufsz
int	evbufsz
int	qspagesize
int	qsmaxlfiles
int	qsfcachesize
int	qsoptfbsize
pointer ed_tptr, tp      # table pointers

string	string_c	"char"
string	string_s	"short"
string	string_i	"integer"
string	string_l	"long"
string	string_r	"real"
string	string_d	"double"
string	string_x	"complex"

define	qpoemac_c	'c'
define	qpoemac_s	's'
define	qpoemac_i	'i'
define	qpoemac_l	'l'
define	qpoemac_r	'r'
define	qpoemac_d	'd'
define	qpoemac_x	'x'

pointer kill

pointer	alias
#pointer tp_open()
pointer	xsevent
char	ft[3]

int xxx
int qp_geti()

begin

	ft[2] = ':'
	ft[3] = EOS

	call smark(sp)
	call salloc (xsevent, STR_LEN, TY_CHAR)
	call salloc (name, STR_LEN, TY_CHAR)
	call salloc (kind, STR_LEN, TY_CHAR)
	call salloc (alias, STR_LEN, TY_CHAR)
	call salloc (buf, STR_LEN, TY_CHAR)
	call salloc (sys, STR_LEN, TY_CHAR)
	call salloc (edef, STR_LEN, TY_CHAR)
	op = edef
	Memc[op] = '{'; op = op + 1

	call strcpy("{", Memc[xsevent], ARB)

	call ev_new_clear(eqp, EQP_LENGTH)
	re = dtlocate(dt, rec)
	axlen[1] = dtgeti(dt, re, "Xsize")
	axlen[2] = dtgeti(dt, re, "Ysize")
	call dtgptr(dt, re, "File", tmp)
	iferr(call qp_delete(Memc[tmp]));

	iferr(pagesize = dtgeti(dt, re, "PageSize")) {
		pagesize = EUV_PAGESIZE
	}

	iferr(bucketlen = dtgeti(dt, re, "BucketLength")) {
		bucketlen = EUV_BUCKETLEN
	}

	iferr(tmpbufsz = dtgeti(dt, re, "TmpBufSize")) {
		tmpbufsz = 0
	}

	if (tmpbufsz < 0) tmpbufsz = 0

	qp = qp_open (Memc[tmp], NEW_FILE, NULL)
	call qp_seti(qp, QPOE_PAGESIZE, pagesize)
	call qp_seti(qp, QPOE_BUCKETLEN, bucketlen)

	EQP_QP(eqp) = qp
	call eu_mfree(tmp, TY_CHAR)

#  EQP_TP(eqp) = tp_open(10)
#  call qp_seti(qp, QPOE_DEBUGLEVEL, 5)


	call ew_gmwcs(dt, re, ep, Memc[sys], STR_LEN, EQP_MW(eqp))
	EQP_CT(eqp) = mw_sctran (EQP_MW(eqp), Memc[sys], "physical", 3B)

	call qp_savewcs (qp, EQP_MW(eqp))

	call dtgfield(dt, re, "Xfield", xo)
	call eb_gcolumn(lp, xo, EQP_XO(eqp))
	xo = EQP_XO(eqp)
	call dtgfield(dt, re, "Yfield", yo)
	call eb_gcolumn(lp, yo, EQP_YO(eqp))
	yo = EQP_YO(eqp)


	# copy header cards from the first ST table
	# Note that we override the default hcpkill kill string 

	ed_tptr = ED_TABLE_PTR(ed)
	tp = GT_PTR(VEC_ELEMENT(ed_tptr, 1))

define SZ_KILL 1024
        call salloc(kill, SZ_KILL, TY_CHAR)
        call set_kill_list(Memc[kill], SZ_KILL)
        call hcpkill(tp, "T", qp, "Q", Memc[kill])

  # Add QPOE header cards

	iferr (call dtgvstr(dt, re, "HeaderList", outs)) ; else {
		do i = 1, VEC_SIZE(outs) {
			call sscan(VEC_STRING(outs, i))
			call gargwrd(Memc[name], STR_LEN)
			call gargstr(Memc[alias], STR_LEN)
			call trimwhite( Memc[alias] )
			call qp_add_string(qp, Memc[name], Memc[alias])
		}
		call ev_free_string(outs)
	}


	call dtgvstr(dt, re, "FieldList", outs)
	call ev_new (col, VEC_SIZE(outs))
	call ev_new (typ, VEC_SIZE(outs))
	EQP_COL(eqp) = col
	EQP_TYP(eqp) = typ
	bytes = 0
	size = 0

	do i = 1, VEC_SIZE(outs) {

		call sscan(VEC_STRING(outs,i))
		call gargwrd(Memc[name], STR_LEN)
		call eb_fieldid(Memc[name], VEC_ELEMENT(col,i))
		call eb_gcolumn(lp, VEC_ELEMENT(col,i), VEC_ELEMENT(col,i))
		call gargwrd(Memc[kind], STR_LEN)
		Memc[alias] = EOS
		call gargwrd(Memc[alias], STR_LEN)

		if (streq("TIME", Memc[name])) {
			call sprintf(Memc[buf], STR_LEN, "exptime = integral %s:d")

			if (Memc[alias] != EOS) call pargstr(Memc[alias])
			else			call pargstr(Memc[name])

			call qp_add_string(qp, "defattr1", Memc[buf])
		}

		call sprintf(Memc[buf], STR_LEN, "%c%d")

		if (VEC_ELEMENT(col,i) == xo || VEC_ELEMENT(col,i) == yo) {

			VEC_ELEMENT(typ,i) = TY_SHORT
			call pargs(ls)
			bytes = bytes + SZ_SHORT
			call pargi(size * SZB_CHAR)
			size = size + SZ_SHORT
			Memc[op] = ls; op = op + 1

			if (VEC_ELEMENT(col,i) == xo) {
				Memc[op] = ':'; op = op + 1
				Memc[op] = 'x'; op = op + 1
				call strcat("s:x", Memc[xsevent], ARB)
			} else if (VEC_ELEMENT(col,i) == yo) {
				Memc[op] = ':'; op = op + 1
				Memc[op] = 'y'; op = op + 1
				call strcat("s:y", Memc[xsevent], ARB)
			}
			Memc[op] = ','; op = op + 1


		} else if (streq(Memc[kind], string_c)) {
			VEC_ELEMENT(typ,i) = TY_CHAR
			call pargc(Memc[kind])
			bytes = bytes + SZ_CHAR
			size = (size + SZ_CHAR - 1) / SZ_CHAR
			call pargi(size * SZ_CHAR * SZB_CHAR)
			size = size * SZ_CHAR + SZ_CHAR
			Memc[op] = qpoemac_c; op = op + 1
			Memc[op] = ','; op = op + 1
			ft[1] = qpoemac_c
			call strcat(ft, Memc[xsevent], ARB)
			if (Memc[alias] != EOS) {
				call strcat(Memc[alias], Memc[xsevent], ARB)
			} else {
				call strcat(Memc[name], Memc[xsevent], ARB)
			}

		} else if (streq(Memc[kind], string_s)) {
			VEC_ELEMENT(typ,i) = TY_SHORT
			call pargc(Memc[kind])
			bytes = bytes + SZ_SHORT
			size = (size + SZ_SHORT - 1) / SZ_SHORT
			call pargi(size * SZ_SHORT * SZB_CHAR)
			size = size * SZ_SHORT + SZ_SHORT
			Memc[op] = qpoemac_s; op = op + 1
			Memc[op] = ','; op = op + 1
			ft[1] = qpoemac_s
			call strcat(ft, Memc[xsevent], ARB)
			if (Memc[alias] != EOS) {
				call strcat(Memc[alias], Memc[xsevent], ARB)
			} else {
				call strcat(Memc[name], Memc[xsevent], ARB)
			}

		} else if (streq(Memc[kind], string_i)) {
			VEC_ELEMENT(typ,i) = TY_INT
			call pargc(Memc[kind])
			bytes = bytes + SZ_INT
			size = (size + SZ_INT - 1) / SZ_INT
			call pargi(size * SZ_INT * SZB_CHAR)
			size = size * SZ_INT + SZ_INT
			Memc[op] = qpoemac_i; op = op + 1
			Memc[op] = ','; op = op + 1
			ft[1] = qpoemac_i
			call strcat(ft, Memc[xsevent], ARB)
			if (Memc[alias] != EOS) {
				call strcat(Memc[alias], Memc[xsevent], ARB)
			} else {
				call strcat(Memc[name], Memc[xsevent], ARB)
			}

		} else if (streq(Memc[kind], string_l)) {
			VEC_ELEMENT(typ,i) = TY_LONG
			call pargc(Memc[kind])
			bytes = bytes + SZ_LONG
			size = (size + SZ_LONG - 1) / SZ_LONG
			call pargi(size * SZ_LONG * SZB_CHAR)
			size = size * SZ_LONG + SZ_LONG
			Memc[op] = qpoemac_l; op = op + 1
			Memc[op] = ','; op = op + 1
			ft[1] = qpoemac_l
			call strcat(ft, Memc[xsevent], ARB)
			if (Memc[alias] != EOS) {
				call strcat(Memc[alias], Memc[xsevent], ARB)
			} else {
				call strcat(Memc[name], Memc[xsevent], ARB)
			}

		} else if (streq(Memc[kind], string_r)) {
			VEC_ELEMENT(typ,i) = TY_REAL
			call pargc(Memc[kind])
			bytes = bytes + SZ_REAL
			size = (size + SZ_REAL - 1) / SZ_REAL
			call pargi(size * SZ_REAL * SZB_CHAR)
			size = size * SZ_REAL + SZ_REAL
			Memc[op] = qpoemac_r; op = op + 1
			Memc[op] = ','; op = op + 1
			ft[1] = qpoemac_r
			call strcat(ft, Memc[xsevent], ARB)
			if (Memc[alias] != EOS) {
				call strcat(Memc[alias], Memc[xsevent], ARB)
			} else {
				call strcat(Memc[name], Memc[xsevent], ARB)
			}

		} else if (streq(Memc[kind], string_d)) {
			VEC_ELEMENT(typ,i) = TY_DOUBLE
			call pargc(Memc[kind])
			bytes = bytes + SZ_DOUBLE
			size = (size + SZ_DOUBLE - 1) / SZ_DOUBLE
			call pargi(size * SZ_DOUBLE * SZB_CHAR)
			size = size * SZ_DOUBLE + SZ_DOUBLE
			Memc[op] = qpoemac_d; op = op + 1
			Memc[op] = ','; op = op + 1
			ft[1] = qpoemac_d
			call strcat(ft, Memc[xsevent], ARB)
			if (Memc[alias] != EOS) {
				call strcat(Memc[alias], Memc[xsevent], ARB)
			} else {
				call strcat(Memc[name], Memc[xsevent], ARB)
			}

		} else call eerror(EES_BAD_QPOE_TYPE)

		if (i != VEC_SIZE(outs)) call strcat(",", Memc[xsevent], ARB)
		else			 call strcat("}", Memc[xsevent], ARB)

		if (Memc[alias] != EOS) {
			call qp_add_macro (qp, Memc[alias], Memc[buf])
		} else {
			call qp_add_macro (qp, Memc[name], Memc[buf])
		}

	}
	Memc[op-1]='}'; Memc[op] = EOS

	iferr(pros = dtgetb(dt, re, "Pros")) {
		pros = FALSE
	} else {
		if(pros) call qp_add_string(qp, "XS-EVENT", Memc[xsevent])
	}


	call ev_new (ptrs, VEC_SIZE(outs))
	EQP_PTR(eqp) = ptrs
	size = (size + SZ_REAL - 1) / SZ_REAL
	call eu_malloc (EQP_BUF(eqp), size, TY_REAL)
	si = P2S(EQP_BUF(eqp))

	do i = 1, VEC_SIZE(outs) {

		switch (VEC_ELEMENT(typ,i)) {	# Align to NEXT TY_PIXEL slot
						#      convert to short again

		  case TY_CHAR :
			si = ((si-1) + SZ_CHAR - 1) * SZ_SHORT / SZ_CHAR + 1
			VEC_ELEMENT(ptrs,i) = si
			si = (si - 1 + 1) * SZ_CHAR / SZ_SHORT + 1

		  case TY_SHORT :
			si = ((si-1) + SZ_SHORT - 1) * SZ_SHORT / SZ_SHORT + 1
			VEC_ELEMENT(ptrs,i) = si
			si = (si - 1 + 1) * SZ_SHORT / SZ_SHORT + 1

		  case TY_INT :
			si = ((si-1) + SZ_INT - 1) * SZ_SHORT / SZ_INT + 1
			VEC_ELEMENT(ptrs,i) = si
			si = (si - 1 + 1) * SZ_INT / SZ_SHORT + 1

		  case TY_LONG :
			si = ((si-1) + SZ_LONG - 1) * SZ_SHORT / SZ_LONG + 1
			VEC_ELEMENT(ptrs,i) = si
			si = (si - 1 + 1) * SZ_LONG / SZ_SHORT + 1

		  case TY_REAL :
			si = ((si-1) + SZ_REAL - 1) * SZ_SHORT / SZ_REAL + 1
			VEC_ELEMENT(ptrs,i) = si
			si = (si - 1 + 1) * SZ_REAL / SZ_SHORT + 1

		  case TY_DOUBLE :
			si = ((si-1) + SZ_DOUBLE - 1) * SZ_SHORT / SZ_DOUBLE + 1
			VEC_ELEMENT(ptrs,i) = si
			si = (si - 1 + 1) * SZ_DOUBLE / SZ_SHORT + 1

		}

		if (VEC_ELEMENT(col,i) == xo) {
			EQP_XPTR(eqp) = VEC_ELEMENT(ptrs, i)
		} else if (VEC_ELEMENT(col,i) == yo) {
			EQP_YPTR(eqp) = VEC_ELEMENT(ptrs, i)
		}
	}


	call qp_addf (qp, "naxes", "i", 1, "number of image axis", 0)
	call qp_puti(qp, "naxes", 2)
 
	call qp_addf(qp, "axlen", "i", 2, "length of each axis", 0)
	call qp_write(qp, "axlen", axlen, 2, 1, "i")

	call qp_addf(qp, "event", Memc[edef], 1, "event type", 0)
	call qp_addf(qp, "events", "event", 0, "event list", 0)
	EQP_IO(eqp) = qpio_open(qp, "events", NEW_FILE)


	call ewr_new_qpoe(qp)

# Commenting out entire section to avoid runtime error in cep.  The 
# bug fix mentioned in the removed section did not work at some sites
# outside of CEA, as we feared would be the case.
#
#	iferr (i = dtgeti(dt, re, "DispAxis")) ; else {
#		call qp_addf(qp, "dispaxis", "i", 1, "dispersion axis", 0)
#		call qp_puti(qp, "dispaxis", i)
#		# This qp_geti call does nothing and is here only because it prevents
#		# an error from occuring on the following qp_puti call.  The reason
#		# for this bug is unknown and this hack is all we can find so far to
#		# make it go away. Sorry...
#		xxx = qp_geti(qp, "dispaxis")
#	}

#
#  Removal of the section above has the following implications:
#
#    In mkdataproducts.cl, line 854 (in egodata), the task euvextract
#    is run without explicitly setting the dispaxis parameter to 1.
#    However, the parameter defaults to 1.  So this will probably be ok
#    even with the DISPAXIS header entry not present in the image. It
#    should be checked and perhaps fixed if we are ever switching the 
#    queues to euv1.8/IRAF2.11.
#
#    In mkarchivelib.sh, line 3814, the DISPAXIS header keyword is
#    deleted from the FITS archive products.  This may cause an error
#    if the keyword is not present.  Therefore, this line should be removed
#    before releasing the batch queues to work with euv1.8/IRAF2.11
#

	iferr (i = dtgeti(dt, re, "DC-Flag")) ; else {
		call qp_addf(qp, "dc-flag", "i", 1, "distortion correction flag", 0)
		call qp_puti(qp, "dc-flag", i)
	}

	call eb_glength(lp, lenlp)
	lenlp = lenlp / axlen[2]
	if (lenlp < 2) lenlp = 2

	EQP_XLEN(eqp) = axlen[1]
	EQP_YLEN(eqp) = axlen[2]

	iferr(bval = dtgetb(dt, re, "Sort")) {
		EQP_SORT(eqp) = YES
		EQP_ORDERED(eqp) = YES
		if (pros) call qp_add_string(qp, "XS-SORT", "y x")
	} else {
		if (bval) {
			if (pros) call qp_add_string(qp, "XS-SORT", "y x")
			EQP_SORT(eqp) = YES
			EQP_ORDERED(eqp) = YES
		} else {
			EQP_SORT(eqp) = NO
			EQP_ORDERED(eqp) = NO
		}
	}

	iferr(bval = dtgetb(dt, re, "QPSort")) {
		EQP_QPSRT(eqp) = NO
	} else {
		if (bval) {

			EQP_QPSRT(eqp) = YES

			iferr(evbufsz = dtgeti(dt, re, "EventBufSize")) {
				EQP_EVBUFSZ(eqp) = DEF_BUFSIZE
			} else {
				EQP_EVBUFSZ(eqp) = evbufsz
			}

		} else EQP_QPSRT(eqp) = NO
	}

	iferr(bval = dtgetb(dt, re, "Array")) {
		EQP_ARRAY(eqp) = YES
	} else {
		if (bval) EQP_ARRAY(eqp) = YES
		else	  EQP_ARRAY(eqp) = NO
	}

	iferr(bval = dtgetb(dt, re, "TwoPass")) {
		EQP_2PASS(eqp) = YES
	} else {
		if (bval) EQP_2PASS(eqp) = YES
		else	  EQP_2PASS(eqp) = NO
	}


	if (EQP_QPSRT(eqp) == NO) {
		call mk_tfd (EQP_TFD(eqp))
		call mk_tfd (EQP_TFD2(eqp))
	}

# Get the block sizes for the 2 types of temp files
# Each sort record consists of the data in bytes,
# less the known x and/or y fields, plus the pointer
# to the next event.  Also, round off to the next
# LEN_RANDBUF just so the buffer is really small.

	EQP_BUFSIZE(eqp)  = 0
	EQP_BUFSIZE2(eqp) = 0

	if (EQP_2PASS(eqp) == YES) {
		EQP_BUFSIZE(eqp)  = bytes + SZ_LONG - SZ_SHORT
		EQP_BUFSIZE2(eqp) = bytes + SZ_LONG - SZ_SHORT - SZ_SHORT
	} else {
		EQP_BUFSIZE(eqp)  = bytes + SZ_LONG - SZ_SHORT - SZ_SHORT
	}

	EQP_BUFSIZE(eqp) = EQP_BUFSIZE(eqp) * tmpbufsz
	EQP_BUFSIZE2(eqp) = EQP_BUFSIZE2(eqp) * tmpbufsz

# initially set the advice to sequential, since
# writing only occurs that the end of the file.

	if (EQP_QPSRT(eqp) == NO) {
		if (EQP_BUFSIZE(eqp) != 0)
		call fseti(EQP_TFD(eqp), F_ADVICE, SEQUENTIAL)
		if (EQP_BUFSIZE2(eqp) != 0)
		call fseti(EQP_TFD2(eqp), F_ADVICE, SEQUENTIAL)
	}
    
	if (EQP_QPSRT(eqp) == YES) {

		im = qpsrt_open(EQP_IO(eqp))

		ifnoerr(call dtgptr(dt, re, "TempDir", tmp)) {
			call qpsrt_dir(im, Memc[tmp])
		}

		call qpsrt_seti(im, QS_EVBUFSZ, EQP_EVBUFSZ(eqp))

		ifnoerr(qspagesize = dtgeti(dt, re, "QSPageSize")) {
			call qpsrt_seti(im, QS_PAGESIZE, qspagesize)
		} else	call qpsrt_seti(im, QS_PAGESIZE, EUV_PAGESIZE)

		ifnoerr(qsmaxlfiles = dtgeti(dt, re, "QSMaxLFiles")) {
			call qpsrt_seti(im, QS_MAXLFILES, qsmaxlfiles)
		}

		ifnoerr(qsfcachesize = dtgeti(dt, re, "QSFCacheSize")) {
			call qpsrt_seti(im, QS_FCACHESIZE, qsfcachesize)
		}

		ifnoerr(qsoptfbsize = dtgeti(dt, re, "QSOptFBSize")) {
			call qpsrt_seti(im, QS_OPTFBSIZE, qsoptfbsize)
		}

	} else {

		if (EQP_ARRAY(eqp) == YES) {

			EQP_IMNM(eqp) = NULL
			if (EQP_2PASS(eqp) == YES) {
				call eu_malloc(im, axlen[2], TY_LONG)
				call aclrl (Meml[im], axlen[2])
			} else {
				call eu_malloc(im, axlen[1]*axlen[2], TY_LONG)
				call aclrl (Meml[im], axlen[1]*axlen[2])
			}

		} else {

			call dtgptr(dt, re, "Image", tmp)
			iferr (call imdelete(Memc[tmp]));
			im = immap(Memc[tmp], NEW_IMAGE, 0)

			if (dtgetb(dt, re, "Delete")) {
				call eu_newstr(Memc[tmp], EQP_IMNM(eqp))
			} else {
				EQP_IMNM(eqp) = NULL
			}
			call eu_mfree(tmp, TY_CHAR)

			if ( EQP_2PASS(eqp) == YES ) {

				IM_NDIM(im) = 1
				IM_PIXTYPE(im) = TY_LONG
				IM_LEN(im,1) = axlen[2]
				ptr = impl1l (im)
				call aclrl (Meml[ptr], axlen[2])

			} else {

				IM_NDIM(im) = 2
				IM_PIXTYPE(im) = TY_LONG
				IM_LEN(im,1) = axlen[1]
				IM_LEN(im,2) = axlen[2]
				call mw_saveim(EQP_MW(eqp), im)
				do i = 1, axlen[2] {
					ptr = impl2l (im, i)
					call aclrl (Meml[ptr], axlen[1])
				}

			}

		}
	}

	EQP_IM(eqp) = im
	EQP_RUN(eqp) = 1
	EQP_ORDERED(eqp) = NO
	EQP_OOR(eqp) = 0
	EQP_IOR(eqp) = 0

	call ev_free_string(outs)
	call sfree(sp)

end

#	Get a temporary file descriptor
procedure	mk_tfd (fd) # get temp fd
long		fd		#O file descriptor

pointer		sp, buf
long		open()

begin
  call smark(sp)
  call salloc (buf, STR_LEN, TY_CHAR)
  call mktemp ("imdir$tempsort", Memc[buf], STR_LEN)
  fd = open(Memc[buf], NEW_FILE, BINARY_FILE)
  call sfree(sp)
end

#	Write an event
procedure	ewt_out(ew, buf) # write event
pointer		ew		#I write state
double		buf[ARB]	#I event

begin

#	call tp_start(EQP_TP(ew), 1)
	if( EQP_SORT(ew) == YES ) {
		if( EQP_QPSRT(ew) == YES ) call ewt_newsrt(ew, buf)
		else call ewt_oldsrt(ew, buf)
	} else call ewt_newsrt(ew, buf)
#	call tp_stop(EQP_TP(ew), 1)

end

#	Write an event for new sorting
procedure	ewt_newsrt(ew, buf) # write event
pointer		ew		#I write state
double		buf[ARB]	#I event

int	i, offset
int	xo, yo
int	ix, iy
int	lx, ly
double	dx, dy

pointer	attributes
pointer types
pointer buf_ptr
pointer event_buf

begin

	event_buf = P2S(EQP_BUF(ew))

	xo = EQP_XO(ew)
	yo = EQP_YO(ew)
	lx = EQP_XLEN(ew)
	ly = EQP_YLEN(ew)

	dx = buf[xo]
	dy = buf[yo]

	if (dx < -MAX_INT || dx > MAX_INT	||
	    dy < -MAX_INT || dy > MAX_INT) {
		EQP_OOR(ew) = EQP_OOR(ew) + 1
		return
	}

	ix = dx
	iy = dy
	if ((ix < 1)  || (iy < 1) ||
	    (ix > lx) || (iy > ly)) {
		EQP_OOR(ew) = EQP_OOR(ew) + 1
		return
	}

	EQP_IOR(ew) = EQP_IOR(ew) + 1

	attributes = EQP_COL(ew)
	types = EQP_TYP(ew)
	do i = 1, VEC_SIZE(attributes) {

		offset  = VEC_ELEMENT( attributes,  i)
		buf_ptr = VEC_ELEMENT( EQP_PTR(ew), i)

		switch( VEC_ELEMENT(types, i) ) {

		  case TY_SHORT: 
			if(buf[offset] == INDEFD) Mems[buf_ptr] = INDEFS
			else Mems[buf_ptr] = buf[offset]

		  case TY_INT: 
			if(buf[offset] == INDEFD) Memi[buf_ptr] = INDEFI
			else Memi[buf_ptr] = buf[offset]

		  case TY_LONG: 
			if(buf[offset] == INDEFD) Meml[buf_ptr] = INDEFL
			else Meml[buf_ptr] = buf[offset]

		  case TY_REAL: 
			if(buf[offset] == INDEFD) Memr[buf_ptr] = INDEFR
			else Memr[buf_ptr] = buf[offset]

		  case TY_DOUBLE: 
			if(buf[offset] == INDEFD) Memd[buf_ptr] = INDEFD
			else Memd[buf_ptr] = buf[offset]

		  case TY_CHAR: 
			if(buf[offset] == INDEFD) Memc[buf_ptr] = -128
			else Memc[buf_ptr] = buf[offset]
		}
	}

	if(EQP_SORT(ew) == YES)
		call qpsrt_putevents(EQP_IM(ew), event_buf, 1)
	else
		call qpio_putevents(EQP_IO(ew), event_buf, 1)

end



#	Write an event for old sorting
procedure	ewt_oldsrt(ew, buf) # write event
pointer		ew		#I write state
double		buf[ARB]	#I event

int		fd
pointer		im

int		i
long		last
pointer		imgs2l()
pointer		imps2l()
pointer		imgs1l()
pointer		imps1l()
long		note()

int		ix, iy, o
pointer		out, typ
int		xo, yo
int		lx, ly
double		dx, dy
int		ap
int		tp


char		datum_c

short		datum_s

int		datum_i

long		datum_l

real		datum_r

double		datum_d


begin
  ap = EQP_ARRAY(ew)
  fd = EQP_TFD(ew)
  im = EQP_IM(ew)
  xo = EQP_XO(ew)
  yo = EQP_YO(ew)
  lx = EQP_XLEN(ew)
  ly = EQP_YLEN(ew)
  tp = EQP_2PASS(ew)

  dx = buf[xo]
  dy = buf[yo]
  if (dx < -MAX_INT || dx > MAX_INT	||
      dy < -MAX_INT || dy > MAX_INT) {
	EQP_OOR(ew) = EQP_OOR(ew) + 1
	return
  }

  ix = dx
  iy = dy
  if ((ix < 1) || (iy < 1) || (ix > lx) || (iy > ly)) {
	EQP_OOR(ew) = EQP_OOR(ew) + 1
	return
  }

  EQP_IOR(ew) = EQP_IOR(ew) + 1

  # push last pointer, then push non-position fields
  if (ap == YES) {
    if (tp == YES) {
      last = Meml[im+iy-1]
      Meml[im+iy-1] = note(fd)
    } else {
      last = Meml[im+ix-1+(iy-1)*lx]
      Meml[im+ix-1+(iy-1)*lx] = note(fd)
    }
  } else {
    if (tp == YES) {
      last = Meml[imgs1l(im, iy, iy)]
      Meml[imps1l(im, iy, iy)] = note(fd)
    } else {
      last = Meml[imgs2l(im, ix, ix, iy, iy)]
      Meml[imps2l(im, ix, ix, iy, iy)] = note(fd)
    }
      call imflsl(im)
  }
  call write (fd, last, SZ_LONG)
  out = EQP_COL(ew)
  typ = EQP_TYP(ew)
  do i = 1, VEC_SIZE(out) {
    o = VEC_ELEMENT(out,i)
    if (o == yo		||
	(tp == NO && o == xo)) next
    switch(VEC_ELEMENT(typ,i)) {

    case TY_CHAR :
      if (buf[VEC_ELEMENT(out,i)] == INDEFD) {
	datum_c = -128
      } else {
	datum_c = buf[VEC_ELEMENT(out,i)]
      }
      call write (fd, datum_c, SZ_CHAR)
#  call eprintf("field %d: %s\n"); call pargi(o);call parg$t(datum_$t)

    case TY_SHORT :
      if (buf[VEC_ELEMENT(out,i)] == INDEFD) {
	datum_s = INDEFS
      } else {
	datum_s = buf[VEC_ELEMENT(out,i)]
      }
      call write (fd, datum_s, SZ_SHORT)
#  call eprintf("field %d: %s\n"); call pargi(o);call parg$t(datum_$t)

    case TY_INT :
      if (buf[VEC_ELEMENT(out,i)] == INDEFD) {
	datum_i = INDEFI
      } else {
	datum_i = buf[VEC_ELEMENT(out,i)]
      }
      call write (fd, datum_i, SZ_INT)
#  call eprintf("field %d: %s\n"); call pargi(o);call parg$t(datum_$t)

    case TY_LONG :
      if (buf[VEC_ELEMENT(out,i)] == INDEFD) {
	datum_l = INDEFL
      } else {
	datum_l = buf[VEC_ELEMENT(out,i)]
      }
      call write (fd, datum_l, SZ_LONG)
#  call eprintf("field %d: %s\n"); call pargi(o);call parg$t(datum_$t)

    case TY_REAL :
      if (buf[VEC_ELEMENT(out,i)] == INDEFD) {
	datum_r = INDEFR
      } else {
	datum_r = buf[VEC_ELEMENT(out,i)]
      }
      call write (fd, datum_r, SZ_REAL)
#  call eprintf("field %d: %s\n"); call pargi(o);call parg$t(datum_$t)

    case TY_DOUBLE :
      if (buf[VEC_ELEMENT(out,i)] == INDEFD) {
	datum_d = INDEFD
      } else {
	datum_d = buf[VEC_ELEMENT(out,i)]
      }
      call write (fd, datum_d, SZ_DOUBLE)
#  call eprintf("field %d: %s\n"); call pargi(o);call parg$t(datum_$t)

    }
  }
end

#	Flush write state
#
#	This sort algorithm writes out events once to a temp
#	file.  An array of heads of linked lists is kept in
#	memory.  Each event is put on the appropriate list and
#	pushed into the temp file.  At the end, the heads are
#	processed in sorting order, and each list is written
#	out (in sort order).
#
#	Fil Machi had an idea of sorting many photon events.
#	We have 22 bits of sorting information.  Look at, say,
#	the 3 bits at a time, starting with the LSB.  Sort the
#	photons into the 8 (2^number_bits) files.  Cat the files
#	back together and repeat the process with the next 3
#	LSBs until all the bits are exhausted.
#
#	Another possibility is to write events to a temp file
#	and simply count the frequency distribution of the events.
#	Then write another temp file by seeking to the right
#	file position (using the frequency information) and writting
#	the event.  An optimization is to cache to must likely
#	bins in memory.
procedure	ewt_flush(ew) # flush write state
pointer		ew		#I write state

int		x, y
pointer		xi, yi

pointer		buf, sp
pointer		hds, cts
long		len, pos
pointer		imgl2l(), impl2l()
pointer		imgl1l(), impl1l()
int		xo, yo, o
int		fd
pointer		im
pointer		qpio, ptrs, typ, out
int		i
int		lx,ly
int		ap
int		tp

pointer	yhds, ycts, yfd, xhds, xcts
int	xfd
long	last
long	note()
short	px, py

begin
  if (EQP_SORT(ew) == NO) return
  if (EQP_QPSRT(ew) == YES) return

  if (EQP_RUN(ew) == 0) return
  call eprintf("Sorting ...\n")
  EQP_RUN(ew) = 0
  EQP_ORDERED(ew) = YES
  xo = EQP_XO(ew)
  yo = EQP_YO(ew)
  lx = EQP_XLEN(ew)
  ly = EQP_YLEN(ew)
  ap = EQP_ARRAY(ew)
  tp = EQP_2PASS(ew)

  ptrs = EQP_PTR(ew)		# general information
  xi = EQP_XPTR(ew)
  yi = EQP_YPTR(ew)
  buf = P2S(EQP_BUF(ew))
  qpio = EQP_IO(ew)
  fd = EQP_TFD(ew)
  im = EQP_IM(ew)
  out = EQP_COL(ew)
  typ = EQP_TYP(ew)

  call smark(sp)
  call salloc(xhds, lx, TY_LONG)
  call salloc(xcts, lx, TY_LONG)
  xfd = EQP_TFD2(ew)

  # We have finished writing on the y file.  make sure
  # that all the data are written out, then change to
  # random access mode.  The block sizes have already
  # been set up, so we just get rid of the old buffer,
  # and advise on the new one.

  call flush(fd)
  if (EQP_BUFSIZE(ew) != 0) {
    call fseti(fd, F_BUFSIZE, EQP_BUFSIZE(ew))
    call fseti(fd, F_ADVICE, RANDOM)
  }

  # go through image in position order.  pop off last pointer.
  # reconstruct position from pixel position.
  # pop previous pointer from file.
  # reconstruct non-position fields from file.
  # emit one event at a time to QPIO.
  py = 1
  do y = 1, ly {
    Mems[yi] = y
    if (ap == YES) {
      if (tp == YES) {
	yhds = im
	ycts = im
      } else {
	hds = im + (y - 1) * lx
	cts = im + (y - 1) * lx
      }
    } else {
      if (tp == YES) {
	yhds = imgl1l (im)
	ycts = impl1l (im)
      } else {
	hds = imgl2l (im, y)
	cts = impl2l (im, y)
      }
    }
    if (tp == YES) {	# sort on x here
      yfd = fd
      fd = xfd
      call aclrl (Meml[xhds], lx)
      call aclrl (Meml[xcts], lx)
      pos = Meml[yhds+y-1]
      Meml[ycts+y-1] = 0
      while (pos != 0) {
	call seek(yfd, pos)
	call read(yfd, pos, SZ_LONG)
	do i = 1, VEC_SIZE(out) {
	  o = VEC_ELEMENT(ptrs, i)
	  if (o == yi) next
	  switch (VEC_ELEMENT(typ, i)) {

	  case TY_CHAR :
	    call read(yfd, Memc[o], SZ_CHAR)

	  case TY_SHORT :
	    call read(yfd, Mems[o], SZ_SHORT)

	  case TY_INT :
	    call read(yfd, Memi[o], SZ_INT)

	  case TY_LONG :
	    call read(yfd, Meml[o], SZ_LONG)

	  case TY_REAL :
	    call read(yfd, Memr[o], SZ_REAL)

	  case TY_DOUBLE :
	    call read(yfd, Memd[o], SZ_DOUBLE)

          }
	}
	last = Meml[xhds+Mems[xi]-1]
	Meml[xhds+Mems[xi]-1] = note(fd)
	call write (fd, last, SZ_LONG)
	do i = 1, VEC_SIZE(ptrs) {
	  o = VEC_ELEMENT(ptrs, i)
	  if (o == yi || o == xi) next
	  switch(VEC_ELEMENT(typ,i)) {

          case TY_CHAR :
	    call write(fd, Memc[o], SZ_CHAR)

          case TY_SHORT :
	    call write(fd, Mems[o], SZ_SHORT)

          case TY_INT :
	    call write(fd, Memi[o], SZ_INT)

          case TY_LONG :
	    call write(fd, Meml[o], SZ_LONG)

          case TY_REAL :
	    call write(fd, Memr[o], SZ_REAL)

          case TY_DOUBLE :
	    call write(fd, Memd[o], SZ_DOUBLE)

          }
	}
      }
      hds = xhds
      cts = xcts
    }

    # we have finished writing to the x file now.
    # make sure everything is written, then switch
    # to random acces mode just like the y file.
    call flush(fd)
    if (EQP_BUFSIZE2(ew) != 0) {
      call fseti(fd, F_BUFSIZE, EQP_BUFSIZE2(ew))
      call fseti(fd, F_ADVICE, RANDOM)
    }

    px = 1
    do x = 1, lx {
      Mems[xi] = x
      len = 0
      pos = Meml[hds+x-1]
      while (pos != 0) {
	call seek (fd, pos)
	call read(fd, pos, SZ_LONG)
	do i = 1, VEC_SIZE(out) {
	  o = VEC_ELEMENT(ptrs, i)
	  if (o == yi || o == xi) next
	  switch (VEC_ELEMENT(typ, i)) {

	  case TY_CHAR :
	    call read(fd, Memc[o], SZ_CHAR)

	  case TY_SHORT :
	    call read(fd, Mems[o], SZ_SHORT)

	  case TY_INT :
	    call read(fd, Memi[o], SZ_INT)

	  case TY_LONG :
	    call read(fd, Meml[o], SZ_LONG)

	  case TY_REAL :
	    call read(fd, Memr[o], SZ_REAL)

	  case TY_DOUBLE :
	    call read(fd, Memd[o], SZ_DOUBLE)

          }
	}
	len = len + 1
	if (Mems[xi] < px) {
	  EQP_ORDERED(ew) = NO
#	  call eprintf("Event is out of x order: (%d,%d)\n")
#	  call pargs(Mems[xi])
#	  call pargs(Mems[yi])
	} else if (Mems[yi] < py) {
	  EQP_ORDERED(ew) = NO
#	  call eprintf("Event is out of y order: (%d,%d)\n")
#	  call pargs(Mems[xi])
#	  call pargs(Mems[yi])
	}
	call qpio_putevents (qpio, buf, 1)
	px = Mems[xi]
	py = Mems[yi]
      }
      if (tp == YES) {
	Meml[ycts+y-1] = Meml[ycts+y-1] + len
      } else {
	Meml[cts+x-1] = len
      }
      if (ap != YES) {
	call imflsl(im)
      }
    }
    if (tp == YES) {
      call seek(fd, long(1))

      # now all the data has been read, the x file will
      # (possibly) be written to again.  Drop the random
      # buffer, and get the sequential one.
      if (EQP_BUFSIZE2(ew) != 0) {
	call fseti(fd, F_BUFSIZE, NULL)
	call fseti(fd, F_ADVICE, SEQUENTIAL)
      }
      fd = yfd
    }
  }

  # actually unnecessary, but it restores the y file
  # to the state that it came into the routine.
  if (EQP_BUFSIZE(ew) != 0) {
    call fseti(fd, F_BUFSIZE, NULL)
    call fseti(fd, F_ADVICE, SEQUENTIAL)
  }

  call sfree(sp)
end


#
# ew_gmwcs()
#

procedure	ew_gmwcs(dt, re, ep, sys, maxchar, mw)

pointer	dt			#I text database
int	re			#I text database position of record
pointer	ep			#I euv process
char	sys[ARB]		#O world coordinate system name
int	maxchar			#I max length of sys string
pointer	mw			#O mwcs

int	naxes
int	axes[2]			#
int	x_sz, y_sz		#
real	xlo, xhi		#
real	ylo, yhi		#

double	r[2], w[2], cd[2,2]	# world coord spec's
double	ra, dec

pointer	sys_name		# world coordinate system name
pointer	tmp			# tmp string
pointer	ndt, rec, path		# text database
pointer	quat			#

pointer ed
pointer spor

int	dtgeti()
real	dtgetr()
pointer	mw_open()

begin

	ed = EP_OBS_PTR(ep)

# Creating the mwcs.

	mw = mw_open(NULL, 2)

	iferr(call dtgptr(dt, re, "WCSName", sys_name)) {
		call mw_newsystem (mw, "world", 2)
		call strcpy("world", sys, maxchar)
	} else {
		call mw_newsystem (mw, Memc[sys_name], 2)
		call strcpy(Memc[sys_name], sys, maxchar)
	}
	call eu_mfree(sys_name, TY_CHAR)


	x_sz = dtgeti(dt, re, "Xsize")
	y_sz = dtgeti(dt, re, "Ysize")

	xlo = dtgetr(dt, re, "Xmin")
	xhi = dtgetr(dt, re, "Xmax")
	ylo = dtgetr(dt, re, "Ymin")
	yhi = dtgetr(dt, re, "Ymax")

# Set the transformation from physical to world coordinates

	cd[1,1] = double(xhi-xlo)/x_sz;	cd[2,1] = 0.0D0
	cd[1,2] = 0.0D0;		cd[2,2] = double(yhi-ylo)/y_sz

# Set the reference position in physical coordinates

	r[1] = double(x_sz+1)/2.0
	r[2] = double(y_sz+1)/2.0


# Set the reference position in world coordinates

  iferr(call dtgptr(dt, re, "RefPosition", tmp)) {

    # If we get here we have a Spectrometer write record,
    # for which there is never a user-specified reference
    # position.  So just use the center of the field of view
	w[1] = (xhi+xlo)/2.0
	w[2] = (yhi+ylo)/2.0

  } else {

    # If we reach this section, we have a tangent plane projection.
    # We need a position to use as the tangent point, so we try
	# to fetch the source position

	# first look for parameter overrides
	spor = ED_SRCPOS(ed)
	if (IS_INDEFD(SRC_RA(spor)) || IS_INDEFD(SRC_DEC(spor))) {

      # no override - get position from location specified in
      # the Write record "RefPosition" card
	  call ep_dtfind(ep, Memc[tmp], ndt, rec, path)
	  call eu_mfree(tmp, TY_CHAR)
	  call eq_read(ndt, rec, quat)

	} else {

	  # override - use ra & dec from the parameters
	  call eq_scalloc(quat)
	  call eq_scstore(SRC_RA(spor), SRC_DEC(spor), quat)

	}

    call er_qtopos(Memd[quat], ra, dec)
	call eu_mfree(quat, TY_DOUBLE)

	w[1] = ra * RADIAN
	w[2] = dec * RADIAN

	axes[1] = 1; axes[2] = 2; naxes = 2
	call mw_swtype(mw, axes, naxes, "tan", 
           "axis 1: axtype=ra  axis 2: axtype=dec")

  }

# Save it in mw
	call mw_swtermd (mw, r, w, cd, 2)

	ifnoerr(call dtgptr(dt, re, "Xlabel", tmp)) {
		call mw_swattrs(mw, 1, "label", Memc[tmp])
		call eu_mfree(tmp, TY_CHAR)
	}

	ifnoerr(call dtgptr(dt, re, "Xunits", tmp)) {
		call mw_swattrs(mw, 1, "units", Memc[tmp])
		call eu_mfree(tmp, TY_CHAR)
	}

	ifnoerr(call dtgptr(dt, re, "Ylabel", tmp)) {
		call mw_swattrs(mw, 2, "label", Memc[tmp])
		call eu_mfree(tmp, TY_CHAR)
	}

	ifnoerr(call dtgptr(dt, re, "Yunits", tmp)) {
		call mw_swattrs(mw, 2, "units", Memc[tmp])
		call eu_mfree(tmp, TY_CHAR)
	}

end

# set the kill string used in eliminating header cards 
procedure set_kill_list(kill, sz_kill)
char kill[ARB]      #O kill list buffer
int sz_kill         #I length of kill list buffer 

begin

    call strcpy("RATE[0-9][0-9]* DELTIME TIMEDEL TIME TTYPE[0-9][0-9]* ", kill, sz_kill)
    call strcat("TFORM[0-9][0-9]* TUNIT[0-9][0-9]* TDISP[0-9][0-9]* ", kill, sz_kill)
    call strcat("TNULL[0-9][0-9]* TDMIN[0-9][0-9]* TDMAX[0-9][0-9]* ", kill, sz_kill)
    call strcat("TLMIN[0-9][0-9]* TLMAX[0-9][0-9]* EXTNAME TFIELDS ", kill, sz_kill)
    call strcat("GCOUNT PCOUNT NAXIS[0-9]* BITPIX XTENSION ", kill, sz_kill)
    call strcat("COMMENT", kill, sz_kill)
   
end
