# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  eraspect.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) eraspect.x -r1.9 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	01/13/97   mabbott	added check for GT_BAD table causing null column ptrs
# 	05/27/94   akeith	documentation
# 	05/12/94   akeith	took out debug statements
# 	09/28/92   akeith	put in debug statements
# 	07/20/92   ericco	convert off-axis angles to degrees
# 	10/29/91   ericco	added support of source off-boresight angles
# 	05/07/91   ericco	date and time created 91/05/07 14:10:44 by ericco
# H-
# 
#

#	Revised: 1/13/97

#	EUV run time aspect operations

include	<mach.h>
include	<math.h>
include	<edata.h>
include	<eerror.h>
include	<dbg.h>
include	"ercol.h"

define	EA_GRP		VEC_ELEMENT($1,1)	# column descriptors
define	EA_W_IND	VEC_ELEMENT($1,2)	# W index
define	EA_X_IND	VEC_ELEMENT($1,3)	# X index
define	EA_Y_IND	VEC_ELEMENT($1,4)	# Y index
define	EA_Z_IND	VEC_ELEMENT($1,5)	# Z index
define	EA_T_IND	VEC_ELEMENT($1,6)	# T index

define	EA_LENGTH	long(10)

define	EA_W_COL	VEC_ELEMENT(EA_GRP($1), EA_W_IND($1))
define	EA_X_COL	VEC_ELEMENT(EA_GRP($1), EA_X_IND($1))
define	EA_Y_COL	VEC_ELEMENT(EA_GRP($1), EA_Y_IND($1))
define	EA_Z_COL	VEC_ELEMENT(EA_GRP($1), EA_Z_IND($1))
define	EA_T_COL	VEC_ELEMENT(EA_GRP($1), EA_T_IND($1))

define	EA_W_DATA	Memd[EC_DATA(EA_W_COL($1))+$2]
define	EA_X_DATA	Memd[EC_DATA(EA_X_COL($1))+$2]
define	EA_Y_DATA	Memd[EC_DATA(EA_Y_COL($1))+$2]
define	EA_Z_DATA	Memd[EC_DATA(EA_Z_COL($1))+$2]
define	EA_T_DATA	Memd[EC_DATA(EA_T_COL($1))+$2]

#	Release aspect columns
procedure	er_aspclose (asp)	# close aspect columns
pointer		asp			#I aspect column list

begin
  call er_gclose(EA_GRP(asp))
  call ev_free(asp)
end

#	Set aspect W column
procedure	er_aspsetw (asp, index) # set W column
pointer		asp			#I aspect data
int		index			#I group column index

begin
  EA_W_IND(asp) = index
end

#	Set aspect X column
procedure	er_aspsetx (asp, index) # set X column
pointer		asp			#I aspect data
int		index			#I group column index

begin
  EA_X_IND(asp) = index
end

#	Set aspect Y column
procedure	er_aspsety (asp, index) # set Y column
pointer		asp			#I aspect data
int		index			#I group column index

begin
  EA_Y_IND(asp) = index
end

#	Set aspect Z column
procedure	er_aspsetz (asp, index) # set Z column
pointer		asp			#I aspect data
int		index			#I group column index

begin
  EA_Z_IND(asp) = index
end

#	Set aspect T column
procedure	er_aspsett (asp, index) # set T column
pointer		asp			#I aspect data
int		index			#I group column index

begin
  EA_T_IND(asp) = index
end


#	Open aspect columns
procedure	er_aspopen (tp, names, len, ti, asp) # open aspect columns
pointer		tp			#I table pointer
pointer		names[len]		#I column name pointers
long		len			#I number of names
long		ti			#I time column index
pointer		asp			#O column descriptors

bool	eu_nullinvec()

begin
  call ev_new_clear(asp, EA_LENGTH)
  call er_gopen (tp, names, len, ti, EA_GRP(asp))

  # Die if we couldn't find any of the columns we need
  if (eu_nullinvec(EA_GRP(asp))) {
	call eprintf("cannot find all aspect table columns\n")
	call eerror(EES_NOTAB)
  }
end

#	Get first aspect value
procedure	er_aspfirst (asp)	# get first aspect value
pointer		asp			#I aspect columns

pointer		grp

begin
  grp = EA_GRP(asp)
  call er_gfirst(grp)
  call er_gcopy(grp, EC_HI_OFFSET, EC_LO_OFFSET)
  call er_gcopy(grp, EC_HI_OFFSET, EC_TMP_OFFSET)
end

#	Get next aspect value
procedure	er_aspnext (asp, flag)	# get next aspect value
pointer		asp			#I aspect columns
bool		flag			#O eof flag

pointer		grp

begin
  grp = EA_GRP(asp)
  call er_gnext(grp, flag)
  if (flag == FALSE) {
    call er_gcopy(grp, EC_HI_OFFSET, EC_LO_OFFSET)
    call er_gcopy(grp, EC_HI_OFFSET, EC_TMP_OFFSET)
  }
end

#	Get current rotation
procedure	er_aspcur(asp, r)
pointer		asp			#I aspect columns
double		r[4]			#O rotation quaternion

begin
  r[W] = EA_W_DATA(asp, EC_TMP_OFFSET)
  r[X] = EA_X_DATA(asp, EC_TMP_OFFSET)
  r[Y] = EA_Y_DATA(asp, EC_TMP_OFFSET)
  r[Z] = EA_Z_DATA(asp, EC_TMP_OFFSET)
end

#	Eric says "Do not use this routine!"
#	Rotate a position into detector coordinates
procedure	er_asprot (asp, in, off, z, x, y) # rotate a position
pointer		asp			#I aspect columns
double		in[4]			#I evaluation time
double		off[4]			#I offset quaternion
double		z			#I focal length
double		x, y			#O detector position

double		out[4]
pointer		grp

double		r[4], x1, y1, z1, f

begin
  grp = EA_GRP(asp)
  call er_aspcur(asp, r)
  out[W] = in[W]
  out[X] = in[X]
  out[Y] = in[Y]
  out[Z] = in[Z]
#  call er_qinvert(r, r)
  call er_qrotateq(out, r)
  call er_qrotateq(out, off)
  call er_qinvert(out, out)
  call er_qtovec(out, x1, y1, z1)
  if (z1 > -.001) {
    x = MAX_DOUBLE
    y = MAX_DOUBLE
  } else {
    f = z / z1
    x = x1 * f
    y = y1 * f
  }
end

#	Rotate a position into boresight coordinates
procedure	er_aspang (asp, in, off, angx, angy) # rotate a position
pointer		asp			#I aspect columns
double		in[4]			#I evaluation time
double		off[4]			#I offset quaternion
double		angx, angy		#O boresight position

double		out[4]
pointer		grp

double		r[4], x1, y1, z1
double		atan()

begin
  grp = EA_GRP(asp)
  call er_aspcur(asp, r)
  out[W] = in[W]
  out[X] = in[X]
  out[Y] = in[Y]
  out[Z] = in[Z]
#  call er_qinvert(r, r)
  call er_qrotateq(out, r)
  call er_qrotateq(out, off)
#  call er_qinvert(out, out)
  call er_qtovec(out, x1, y1, z1)
  if (z1 < .001) {	# too far away
    angx = HALFPI
    angy = HALFPI
  } else {
    angx = atan(x1/z1)
    angy = atan(y1/z1)
    angx = angx * 180 / PI
    angy = angy * 180 / PI
  }
end

#	Get the aspect at a particular time
#	aspect is store in groups temporary fields
procedure	er_aspeval (asp, time)	# eval aspect at time
pointer		asp			#I aspect columns
double		time			#I evaluation time

pointer		grp, ec
double		lo[4], hi[4], tmp[4], tlo, thi
int		status

begin
  grp = EA_GRP(asp)
  ec = VEC_ELEMENT(grp, 1)
  call er_gdelims(grp, time, status)
  if (EC_LO_ROW(ec) == EC_HI_ROW(ec)) {
    call er_gcopy(grp, EC_LO_OFFSET, EC_TMP_OFFSET)
  } else {
    lo[W] = EA_W_DATA(asp, EC_LO_OFFSET)
    lo[X] = EA_X_DATA(asp, EC_LO_OFFSET)
    lo[Y] = EA_Y_DATA(asp, EC_LO_OFFSET)
    lo[Z] = EA_Z_DATA(asp, EC_LO_OFFSET)
    tlo = EA_T_DATA(asp, EC_LO_OFFSET)
    
    hi[W] = EA_W_DATA(asp, EC_HI_OFFSET)
    hi[X] = EA_X_DATA(asp, EC_HI_OFFSET)
    hi[Y] = EA_Y_DATA(asp, EC_HI_OFFSET)
    hi[Z] = EA_Z_DATA(asp, EC_HI_OFFSET)
    thi = EA_T_DATA(asp, EC_HI_OFFSET)

    if (status != 0) {
      call er_qnormalize (lo, lo)
      EA_W_DATA(asp, EC_LO_OFFSET) = lo[W]
      EA_X_DATA(asp, EC_LO_OFFSET) = lo[X]
      EA_Y_DATA(asp, EC_LO_OFFSET) = lo[Y]
      EA_Z_DATA(asp, EC_LO_OFFSET) = lo[Z]
      call er_qnormalize (hi, hi)
      EA_W_DATA(asp, EC_HI_OFFSET) = hi[W]
      EA_X_DATA(asp, EC_HI_OFFSET) = hi[X]
      EA_Y_DATA(asp, EC_HI_OFFSET) = hi[Y]
      EA_Z_DATA(asp, EC_HI_OFFSET) = hi[Z]
    }
    call er_qinterp(lo, tlo, hi, thi, tmp, time)

    EA_W_DATA(asp, EC_TMP_OFFSET) = tmp[W]
    EA_X_DATA(asp, EC_TMP_OFFSET) = tmp[X]
    EA_Y_DATA(asp, EC_TMP_OFFSET) = tmp[Y]
    EA_Z_DATA(asp, EC_TMP_OFFSET) = tmp[Z]
    EA_T_DATA(asp, EC_TMP_OFFSET) = time
  }
end

#	Interpolate between two quaternion times
procedure	er_qinterp(qlo, tlo, qhi, thi, qt, time)
double		qlo[4], tlo
double		qhi[4], thi
double		qt[4], time

double		part

begin
  part = (time - tlo) / (thi - tlo)
  call er_qextrapolate (qlo, qhi, part, qt)
end
