# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  ergroup.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) ergroup.x -r1.10 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	07/14/94   akeith	fixed requiste for non-existent table columns
# 	05/12/94   akeith	took out debug statements
# 	02/02/94   ericco	remove the dangerous er_gptr routine
# 	02/03/93   ericco	fixed eof INDEF value error
# 	09/28/92   akeith	put in debug statements
# 	06/17/92   ericco	added INDEF checks to type conversions
# 	10/29/91   ericco	added utility routines to access double value data
# also added a status returning group delimiter
# 	06/19/91   ericco	created er_getd to force a group value into a double
# 	05/07/91   ericco	date and time created 91/05/07 14:10:53 by ericco
# H-
# 
#

#	Revised: 7/14/94

#	EUV run time group operations

include	<edata.h>
include	<dbg.h>
include	"ercol.h"

#	Release group resources
procedure	er_gclose (grp)		# close column group
pointer		grp			#I group column list

int		index

begin
  do index = 1, VEC_SIZE(grp) {
    call er_cclose (VEC_ELEMENT(grp,index))
  }
  call ev_free(grp)
  grp = 0
end

#	Given an array of string pointers (of column names in
#	a single table), open up the columns as a group.

procedure	er_gopen (tp, names, len, ti, grp) # open column group

pointer		tp			#I table pointer
pointer		names[len]		#I column name pointers
long		len			#I number of names
long		ti			#I time column index
pointer		grp			#O column descriptors

int		index
pointer		tcp

begin

#  grp = EBP_GROUP()

	call ev_new(grp, long(len))

	if (ti != 0) {
		tcp = -1
		if(names[ti] != NULL)
			call er_copen(tp, tcp, Memc[names[ti]], VEC_ELEMENT(grp, ti))
		else	call er_copen(tp, tcp, EOS, VEC_ELEMENT(grp, ti))
	} else tcp = 0

#  For each name in the list of names, er_copen() gathers the info
#    from the st tables and stores it within the EC_ structure whose
#    pointer is stored in the EBP_GROUP list, [VEC_ELEMENT(grp, index)]. 

	do index = 1, len {
		if (index != ti) {

			if(names[index] != NULL)
				call er_copen(tp, tcp, Memc[names[index]],
							VEC_ELEMENT(grp, index))
			else	call er_copen(tp, tcp, EOS,
							VEC_ELEMENT(grp, index))

		}
	}

end

#	Get first valid group element
procedure	er_gfirst (grp)		# get first group value
pointer		grp			#I group data

begin
  call er_cinit1(VEC_ELEMENT(grp, 1), int(VEC_SIZE(grp)))
end

#	Copy between group registers
procedure	er_gcopy (grp, from, to) # copy group value
pointer		grp			#I group data
int		from			#I from buffer
int		to			#I to buffer

int		index

begin
  do index = 1, VEC_SIZE(grp) {
    call er_ccopy (VEC_ELEMENT(grp,index), from, to)
  }
end

#	Get High and Low register limits
procedure	er_ghilo (grp, tlo, thi, rlo, rhi) # delimit group
pointer		grp			#I group data
double		tlo, thi		#I time limit
long		rlo, rhi		#I row limit

pointer		ec, tc

begin
  ec = VEC_ELEMENT(grp, 1)
  tc = EC_TIME_CP(ec)
  tlo = Memd[EC_LO_DATA(tc)]
  thi = Memd[EC_HI_DATA(tc)]
  rlo = EC_LO_ROW(ec)
  rhi = EC_HI_ROW(ec)
end

#	Get previous valid group value
procedure	er_gprev (grp, flag) # previous value
pointer		grp			#I group data
bool		flag			#I eof flag

begin
  call er_cprev(VEC_ELEMENT(grp, 1), int(VEC_SIZE(grp)), flag)
end

#	Get next valid group value
procedure	er_gnext (grp, flag) # next value
pointer		grp			#I group data
bool		flag			#I eof flag

begin
  call er_cnext1(VEC_ELEMENT(grp, 1), int(VEC_SIZE(grp)), flag)
end

#	Delimit a time without returning a status
procedure	er_gdelim(grp, time) # delimit time
pointer		grp			#I group data
double		time			#I time to find

int		status

begin
  call er_gdelims(grp, time, status)
end

#	this could be modified to only copy from the tmp register
#	to the lo/hi register when there is a hit.  expectation is
#	that most of the time, we will only have to move a single
#	slot -- so this may not be important.
procedure	er_gdelims (grp, time, status) # set lo < time < hi 
pointer		grp			#I group data
double		time			#I time to find
int		status			#O indicate updated values

pointer		ec, tc
bool		flag

begin
  status = 0
  ec = VEC_ELEMENT(grp, 1)
  tc = EC_TIME_CP(ec)

  if (Memd[EC_LO_DATA(tc)] > time) {
    status = 1
    repeat {
      call er_gcopy (grp, EC_LO_OFFSET, EC_HI_OFFSET)
      call er_gprev(grp, flag)
    } until ((flag == TRUE) || (Memd[EC_LO_DATA(tc)] <= time))

    # restore last lo value if the beginning of the table is
    # encountered.
    if (flag == TRUE) {
      call er_gcopy (grp, EC_HI_OFFSET, EC_LO_OFFSET)
    }
  } else if (Memd[EC_HI_DATA(tc)] < time) {
    status = 1
    repeat {
      call er_gcopy (grp, EC_HI_OFFSET, EC_LO_OFFSET)
      call er_gnext(grp, flag)
    } until ((flag == TRUE) || (Memd[EC_HI_DATA(tc)] >= time))

    # restore last hi value if the end of the table is
    # encountered.
    if (flag == TRUE) {
      call er_gcopy (grp, EC_LO_OFFSET, EC_HI_OFFSET)
    }
  }
end

#	Get a pointer to a group value.
#	This is dangerous since the type of the
#	pointer is not necessarily known.
#
#	This looks like it is too dangerous since it
#	caused a release to fail.
#procedure	er_gptr (grp, index, offset, ptr) # get data pointer
#pointer		grp			#I group data
#long		index			#I column index
#int		offset			#I line offset
#pointer		ptr			#O data pointer
#
#pointer		ec
#
#begin
#  ec = VEC_ELEMENT(grp, index)
#  ptr = EC_DATA(ec) + offset
#end


#	Get a value as a type double

procedure	er_getd(grp, index, offset, value)

pointer		grp			#I group data
long		index			#I column index
int		offset			#I line offset
double		value			#O data value

pointer		ec
pointer		ptr

begin

	ec = VEC_ELEMENT(grp, index)
	ptr = EC_DATA(ec) + offset
	value = INDEFD

	switch (EC_DATA_TYPE(ec)) {
	  case TY_DOUBLE : if (Memd[ptr] != INDEFD) value = Memd[ptr]
	  case TY_LONG   : if (Meml[ptr] != INDEFL) value = Meml[ptr]
	  case TY_SHORT  : if (Mems[ptr] != INDEFS) value = Mems[ptr]
	  case TY_INT    : if (Memi[ptr] != INDEFI) value = Memi[ptr]
	  case TY_REAL   : if (Memr[ptr] != INDEFR) value = Memr[ptr]
	}

end

#	Get the group's high delimiter value as a double
procedure	er_ghid(grp, index, value)
pointer		grp			#I group data
long		index			#I column index
double		value			#O data value

begin
	call er_getd(grp, index, EC_HI_OFFSET, value)
end

#	Get the group's low delimiter value as a double
procedure	er_glod(grp, index, value)
pointer		grp			#I group data
long		index			#I column index
double		value			#O data value

begin
  call er_getd(grp, index, EC_LO_OFFSET, value)
end

#	Get the group's tmp value as a double
procedure	er_gtmpd(grp, index, value)
pointer		grp			#I group data
long		index			#I column index
double		value			#O data value

begin
  call er_getd(grp, index, EC_TMP_OFFSET, value)
end
