# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  ertables.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) ertables.x -r1.14 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	12/27/96   mabbott	did some debugging work
# 	11/08/96   ericco	added column alias in header support
# 	10/22/96   ericco	added stronger checking for invalid values in constant columns
# 	10/22/96   ericco	fixed flag indicating valid value for constant columns
# 	07/11/96   ericco	Added constant column support with ST Tables.
# 	07/14/94   akeith	fixed requiste for non-existent table columns
# 	07/12/94   ericco	silently accept unknown columns are INDEFD values
# 	05/12/94   akeith	took out debug statements
# 	11/20/93   ericco	added direct access to short table columns
# 	09/28/92   akeith	put in debug statements
# 	10/29/91   ericco	added eu_malloc calls
# 	06/19/91   ericco	added general table handling and er_cqp for QPOE tables
# 	05/07/91   ericco	date and time created 91/05/07 14:11:12 by ericco
# H-
# 
#

#	Revised: 12/27/96

#	EUV run time table operations

#	this code assumes that there is at least one row of good
#	data for every "group" of columns


include	<mach.h>
include	<eerror.h>
include	<edata.h>
include	<dbg.h>
include	"ercol.h"

#	read a qpoe field
procedure	er_cqp(tp, index, dtype, dp, rawtype, cp, flag)
pointer		tp			#I table pointer
long		index			#I event "index"
long		dtype			#I output data type
pointer		dp			#I/O data pointer
long		rawtype			#I raw data type
long		cp			#I byte offset
bool		flag			#O null value

int		i
pointer		io
int		mask
int		ev
double		value
pointer		buf, qpbuf

begin
  io = GT_QPIO(tp)
  buf = GT_QPBUF(tp)
  flag = FALSE
  if (index < GT_INDEX(tp)) {
    flag = TRUE
  } else if (GT_INDEX(tp) < index) {
    do i = GT_INDEX(tp) + 1, index {
      call qpio_getevents(io, qpbuf, mask, 1, ev)
      if (ev != 1) {
	flag = TRUE
	GT_EOF(tp) = 1
	break
      }
      GT_INDEX(tp) = i
    }
    if (flag == FALSE) call amovs (Mems[qpbuf], Mems[buf], GT_QPLEN(tp))
  }
  if (flag == FALSE) {
    buf = buf + cp / (SZ_SHORT *  SZB_CHAR)
    switch (rawtype) {
    case TY_DOUBLE : value = Memd[((buf-1)*SZ_SHORT)/SZ_DOUBLE+1]
    case TY_LONG   : value = Meml[((buf-1)*SZ_SHORT)/SZ_LONG+1]
    case TY_SHORT  : value = Mems[((buf-1)*SZ_SHORT)/SZ_SHORT+1]
    case TY_INT    : value = Memi[((buf-1)*SZ_SHORT)/SZ_INT+1]
    case TY_REAL   : value = Memr[((buf-1)*SZ_SHORT)/SZ_REAL+1]
    default        : flag = TRUE
    }
  }
  switch (dtype) {
  case TY_DOUBLE : if (flag == TRUE) Memd[dp] = INDEFD; else Memd[dp] = value
  case TY_LONG   : if (flag == TRUE) Meml[dp] = INDEFL; else Meml[dp] = value
  case TY_SHORT  : if (flag == TRUE) Mems[dp] = INDEFS; else Mems[dp] = value
  case TY_INT    : if (flag == TRUE) Memi[dp] = INDEFI; else Memi[dp] = value
  case TY_REAL   : if (flag == TRUE) Memr[dp] = INDEFR; else Memr[dp] = value
  default        : flag = TRUE
  }
end

#	read a column value into a column register

procedure	er_cget (ec, offset, index, flag) # get data point

pointer		ec			#I column data
int		offset			#I insert spot
int		index			#I row number
bool		flag			#O null value

int		spaceint
pointer		tp			# ST table pointer
pointer		cp			# ST column pointer
pointer		dp			# I don't know
pointer		gtp			# ptr to general table structure, GT_

begin

	gtp = EC_TP(ec)
	cp = EC_CP(ec)
	dp = EC_DATA(ec) + offset

	# if the column doesn't exist, the type must be double, the
	# value must be constant and it must already be initialized.
	if (cp == NULL) {
		flag = FALSE
		switch(EC_DATA_TYPE(ec)) {
		  case TY_DOUBLE :
			if (Memd[dp] == INDEFD) flag = TRUE
		  case TY_LONG   :
			if (Meml[dp] == INDEFL) flag = TRUE
		  case TY_SHORT  :
			if (Mems[dp] == INDEFS) flag = TRUE
		  case TY_INT    :
			if (Memi[dp] == INDEFI) flag = TRUE
		  case TY_REAL   :
			if (Memr[dp] == INDEFR) flag = TRUE
		  default        :
			flag = TRUE
		}
		return
	}

	tp = GT_PTR(gtp)
	flag = TRUE

	switch (GT_TYPE(gtp)) {

	  case GT_QPOE :	# for QPOE we can only return the next value
		call er_cqp(gtp, index, EC_DATA_TYPE(ec), dp,
				EC_RDATA_TYPE(ec), cp, flag)

	  case GT_STBT :
		flag = FALSE

		switch(EC_DATA_TYPE(ec)) {

		  case TY_DOUBLE :
			call tbcgtd (tp, cp, Memd[dp], flag, index, index)
			if (flag == TRUE) Memd[dp] = INDEFD

		  case TY_LONG   :
			call tbcgti (tp, cp, spaceint, flag, index, index)
			if (flag == TRUE) Meml[dp] = INDEFL; else Meml[dp] = spaceint

		  case TY_SHORT  :
			call tbcgts (tp, cp, Mems[dp], flag, index, index)
			if (flag == TRUE) Mems[dp] = INDEFS

		  case TY_INT    :
			call tbcgti (tp, cp, Memi[dp], flag, index, index)
			if (flag == TRUE) Memi[dp] = INDEFI

		  case TY_REAL   :
			call tbcgtr (tp, cp, Memr[dp], flag, index, index)
			if (flag == TRUE) Memr[dp] = INDEFR

		  default        :
			flag = TRUE

		}
	}

end

#	copy between column registers
procedure	er_ccopy (ec, from, to) # copy cache data
pointer		ec			#I column data
int		from			#I from position
int		to			#I to position

pointer		dp

begin
  EC_ROW(ec, to) = EC_ROW(ec, from)
  dp = EC_DATA(ec)
  switch (EC_DATA_TYPE(ec)) {
  case TY_DOUBLE	: Memd[dp+to] = Memd[dp+from]
  case TY_LONG		: Meml[dp+to] = Meml[dp+from]
  case TY_SHORT		: Mems[dp+to] = Mems[dp+from]
  case TY_INT		: Memi[dp+to] = Memi[dp+from]
  case TY_REAL		: Memr[dp+to] = Memr[dp+from]
  }
end

#	Get next set of column values whether or
#	not they are valid.

procedure	er_cnext0 (ecs, len, flag)	# load very next column values

pointer		ecs[len]			#I column data
int		len				#I number columns
bool		flag				#O null value

int		row
int		nrow
int		i
bool		cflag
pointer		tp

begin

#  ecs[] is a pointer to the EC_ structure.

	tp = EC_TP(ecs[1])
	row = EC_HI_ROW(ecs[1])

	do i = 2, len {
		if (row < EC_HI_ROW(ecs[i])) row = EC_HI_ROW(ecs[i])
	}

	nrow = EC_ROWS(ecs[1])
	flag = TRUE

	if (nrow == 0) {
		if (GT_EOF(tp) == 1) return
	} else if (row == nrow) return
	flag = FALSE

	row = row + 1
	do i = 1, len {
		call er_cget (ecs[i], EC_HI_OFFSET, row, cflag)
		EC_HI_ROW(ecs[i]) = row
	}

end


#	Get next set of valid (non-INDEF) column values.

procedure	er_cnext1 (ecs, len, flag)	# load next column values
pointer		ecs[len]			#I column data
int		len				#I number columns
bool		flag				#O null value

int		index
int		row
int		nrow
int		i
pointer		tp

begin
  tp = EC_TP(ecs[1])
  row = EC_HI_ROW(ecs[1])
  do index = 2, len {
    if (row < EC_HI_ROW(ecs[index])) row = EC_HI_ROW(ecs[index])
  }
  nrow = EC_ROWS(ecs[1])
  flag = TRUE
  if (nrow == 0) {
    index = row + 1
    repeat {
      do i = 1, len {
	call er_cget (ecs[i], EC_HI_OFFSET, index, flag)
        EC_HI_ROW(ecs[i]) = index
        if (flag == TRUE) break
      }
      if (flag == FALSE) break
      index = index + 1
    } until (GT_EOF(tp) == 1)
  } else if (row < nrow) {
    do index = row + 1, nrow {
      do i = 1, len {
	call er_cget (ecs[i], EC_HI_OFFSET, index, flag)
	EC_HI_ROW(ecs[i]) = index
	if (flag == TRUE) break
      }
      if (flag == FALSE) break
    }
  }
end

#	Get next set of valid columns, while
#	pushing current values into the low register.
procedure	er_cnext (ecs, len, flag)	# find next column values
pointer		ecs[len]			#I column data
int		len				#I number columns
bool		flag				#O null value

int		index

begin
  call er_cnext1(ecs, len, flag)
  if (flag == FALSE) {
    do index = 1, len {
      call er_ccopy (ecs[index], EC_HI_OFFSET, EC_LO_OFFSET)
      call er_ccopy (ecs[index], EC_TMP_OFFSET, EC_HI_OFFSET)
    }
  }
end

#	Get previous set of valid column values.
procedure	er_cprev1 (ecs, len, flag)	# load prev column values
pointer		ecs[len]			#I column data
int		len				#I number columns
bool		flag				#O null value

int		index
int		row
int		i

begin
  row = EC_LO_ROW(ecs[1])
  do index = 2, len {
    if (row > EC_LO_ROW(ecs[index])) row = EC_LO_ROW(ecs[index])
  }
  flag = TRUE
  if (row == 1) return
  do index = row - 1, 1, -1 {
    do i = 1, len {
      call er_cget (ecs[i], EC_TMP_OFFSET, index, flag)
      EC_TMP_ROW(ecs[i]) = index
      if (flag == TRUE) break
    }
    if (flag == FALSE) break
  }
end

#	Get previous set of value columns, while
#	pushing current values into high register.
procedure	er_cprev (ecs, len, flag)	# find prev column values
pointer		ecs[len]			#I column data
int		len				#I number columns
bool		flag				#O null value

int		index

begin
  call er_cprev1(ecs, len, flag)
  if (flag == FALSE) {
    do index = 1, len {
      call er_ccopy (ecs[index], EC_LO_OFFSET, EC_HI_OFFSET)
      call er_ccopy (ecs[index], EC_TMP_OFFSET, EC_LO_OFFSET)
    }
  }
end

#	Initialize a set of column values.

procedure	er_cinit0 (ecs, len)	# initialize columns

pointer		ecs[len]		#I column data
int		len			#I number columns

int		index
bool		flag

begin

	do index = 1, len {
		EC_HI_ROW(ecs[index]) = 0
	}

	call er_cnext0(ecs, len, flag)

end

#	Initialize a set of value column values.
procedure	er_cinit1 (ecs, len)	# initialize columns
pointer		ecs[len]		#I column data
int		len			#I number columns

int		index
bool		flag

begin
  do index = 1, len {
    EC_HI_ROW(ecs[index]) = 0
  }
  call er_cnext1(ecs, len, flag)
  if (flag) call eerror(EES_NULCOL)
end

#	Initialize and set low and high registers
#	to valid column values.
procedure	er_cinit (ecs, len)	# initialize columns
pointer		ecs[len]		#I column data
int		len			#I number columns

int		index

begin
  call er_cinit1(ecs, len)
  do index = 1, len {
    call er_ccopy (ecs[index], EC_TMP_OFFSET, EC_LO_OFFSET)
    call er_ccopy (ecs[index], EC_TMP_OFFSET, EC_HI_OFFSET)
  }
end


#	Open a column optionally with an associated
#	time column.

procedure	er_copen (gtp, tcp, name, ec)	# open a column

pointer		gtp			#I table pointer
pointer		tcp			#I table time column pointer
char		name[ARB]		#I column name
pointer		ec			#O column data

pointer		sp, buf
int		number
int		lendata, lenfmt
double		v

pointer		tp
int		ip
int		qp_read()

double		tbhgtd()
errchk		tbhgtd
errchk		tbhgtt

begin

	if (GT_PTR(gtp) == GT_BAD) {ec = 0; return}
	lendata = 0
	lenfmt = 0

	call smark(sp)
	call salloc(buf, 1024, TY_CHAR)

#  Allocation of the EC_ structure is stored in the EBP_GROUP list.

	call ev_new_clear(ec, EC_LENGTH)
	EC_TP(ec) = gtp
	tp = GT_PTR(gtp)
	EC_CP(ec) = NULL

	switch (GT_TYPE(gtp)) {

	  case GT_STBT :
		if (name[1] != EOS) {
			call tbcfnd(tp, name, EC_CP(ec), 1)

			# if the column wasn't found, see if there is
			# a header card that has an alias.  If so, then
			# try to open that column instead.
			if (EC_CP(ec) == NULL) {
				ifnoerr (call tbhgtt(tp, name, Memc[buf], 1024)) {
					call tbcfnd(tp, Memc[buf], EC_CP(ec), 1)
				}
			}
		}

		call tbtrows(tp, EC_ROWS(ec))

		# skip non-existent columns
		if (EC_CP(ec) != NULL)
			call tbcinf (EC_CP(ec), number, Memc[buf], Memc[buf], Memc[buf],
						EC_DATA_TYPE(ec), lendata, lenfmt)

	  case GT_QPOE :
		EC_ROWS(ec) = 0 # indicates unknown length, sequential access only
		ip = qp_read(tp, name, Memc[buf], 1024, 1, "macro")
		if (ip > 1) {
		  ip = 2
		  call ctoi(Memc[buf], ip, EC_CP(ec))
		  switch (Memc[buf]) {
		  case 'r' :      EC_DATA_TYPE(ec) = TY_REAL
		  case 's' :      EC_DATA_TYPE(ec) = TY_SHORT
		  case 'l' :      EC_DATA_TYPE(ec) = TY_LONG
		  case 'i' :      EC_DATA_TYPE(ec) = TY_INT
		  case 'd' :      EC_DATA_TYPE(ec) = TY_DOUBLE
		  }
		}
	}

	if (EC_CP(ec) == NULL)
		EC_DATA_TYPE(ec) = TY_DOUBLE

	EC_RDATA_TYPE(ec) = EC_DATA_TYPE(ec)
	if (tcp == -1) {
		tcp = ec
		EC_DATA_TYPE(ec) = TY_DOUBLE	# force to double data type
						#    for time columns
	}

	EC_TIME_CP(ec) = tcp

	# Allocate data buffer and initialize constant values
	# header initialization is only for st tables.
	call eu_malloc (EC_DATA(ec), 3, EC_DATA_TYPE(ec))
	if (EC_CP(ec) == NULL) {
	  v = INDEFD
	  if (GT_TYPE(gtp) == GT_STBT && name[1] != EOS) {
		iferr (v = tbhgtd(tp, name))
			v = INDEFD
	  }
	  Memd[EC_LO_DATA(ec)] = v
	  Memd[EC_HI_DATA(ec)] = v
	  Memd[EC_TMP_DATA(ec)] = v
	}

call sfree(sp)

end

#	Release column resources
procedure	er_cclose (ec)			# close column
pointer		ec				#I column data

begin
  call eu_mfree(EC_DATA(ec), EC_DATA_TYPE(ec))
  call ev_free(ec)
end

#	Evaluate a column at a particular time
#	using a particular method.
procedure	er_ceval(ec, time, how, ptr)	# get column value at time
pointer		ec				#I column data
double		time				#I time to evaluate
int		how				#I interpolation method
pointer		ptr				#O pointer to value or NULL

pointer		sp, buf
double		dt, t1
pointer		lo, hi
pointer		tc

begin
  tc = EC_TIME_CP(ec)

  call smark (sp)
  call salloc (buf, 3, TY_LONG)	# maybe i should new_vector this
  Meml[buf] = 3
  Meml[buf+2] = ec
  Meml[buf+1] = tc
  if (Meml[buf+1] == 0) {ptr = 0; return}

  call er_gdelim(buf, time)
  switch (how) {
  case EVAL_PREV	: ptr = EC_DATA(ec) + EC_LO_OFFSET
  case EVAL_NEXT	: ptr = EC_DATA(ec) + EC_HI_OFFSET
  case EVAL_LINEAR	:
    dt = Memd[EC_HI_DATA(tc)] - Memd[EC_LO_DATA(tc)]
    if (dt == 0.0) {
      ptr = EC_DATA(ec) + EC_LO_OFFSET
    } else {
      ptr = EC_DATA(ec) + EC_TMP_OFFSET
      lo = EC_DATA(ec) + EC_LO_OFFSET
      hi = EC_DATA(ec) + EC_HI_OFFSET
      t1 = (time - Memd[EC_LO_ROW(ec)]) / dt
      switch (EC_DATA_TYPE(ec)) {
      case TY_DOUBLE	: Memd[ptr] = Memd[lo] + (Memd[hi] - Memd[lo]) * t1
      case TY_LONG	: Meml[ptr] = Meml[lo] + (Meml[hi] - Meml[lo]) * t1
      case TY_SHORT	: Mems[ptr] = Mems[lo] + (Mems[hi] - Mems[lo]) * t1
      case TY_INT	: Memi[ptr] = Memi[lo] + (Memi[hi] - Memi[lo]) * t1
      case TY_REAL	: Memr[ptr] = Memr[lo] + (Memr[hi] - Memr[lo]) * t1
      }
    }
  default		: ptr = NULL
  }
end
