# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  etops.x
# 	Author	:  ericco
# 	Date	:  04/30/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) etops.x -r1.4 04/30/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	07/12/93   tnk	changed tbrgtt to tbrgttn
# 	04/30/92   ericco	date and time created 92/04/30 00:50:58 by ericco
# H-
# 
#

include <tbset.h>

include	"et.h"
include	"etr.h"

#	Add a string restriction to a column
procedure	etr_new_string(etr, tb, column, str, )
pointer		etr		#I/O restrictions
pointer		tb		#I table
char		column[ARB]	#I name
char		str[ARB]	#I restriction

pointer		loc

begin
  etr = 0
  call tbcfnd (tb, column, loc, 1)
  if (loc == NULL) return

  call etr_malloc(etr)
  call ds_strcpy(column, ETR_COLUMN_PTR(etr))
  call ds_strcpy(str, ETR_MATCH_PTR(etr))
  ETR_COLUMN_ID(etr) = loc
  ETR_STRINGP(etr) = 1
  ETR_EXACTP(etr) = 1
end

#	Add a real number range to a table restriction
procedure	etr_new_range(etr, tb, column, minv, maxv)
pointer		etr			#I/O restrictions
pointer		tb			#I table
char		column[ARB]		#I column nme
real		minv			#I low value
real		maxv			#I high value

pointer		loc

begin
  etr = 0
  call tbcfnd (tb, column, loc, 1)
  if (loc == NULL) return

  call etr_malloc(etr)
  call ds_strcpy(column, ETR_COLUMN_PTR(etr))
  ETR_COLUMN_ID(etr) = loc
  ETR_MATCH_PTR(etr) = 0
  ETR_STRINGP(etr) = 0
  ETR_EXACTP(etr) = 0
  ETR_MINIMUM(etr) = minv
  ETR_MAXIMUM(etr) = maxv
end

#	Release table restriction resource
procedure	etr_close(etr)
pointer		etr		#I restrictions

begin
  if (ETR_STRINGP(etr) != 0) {
    call mfree(ETR_MATCH_PTR(etr), TY_CHAR)
  }
  call mfree(ETR_COLUMN_PTR(etr), TY_CHAR)
  call etr_free(etr)
  etr = 0
end

#	Open a euv style table
procedure	et_open(et, table)
pointer		et			#O table pointer
char		table[ARB]		#I table name

pointer		tbtopn()
int		tbtacc()

begin
  et = 0
  if (tbtacc(table) == NO) return

  call et_malloc(et)
  call ds_strcpy(table, ET_NAME_PTR(et))
  ET_TB(et) = tbtopn(table, READ_ONLY, 0)
  ET_RESTRICTION_LEN(et) = 0
  ET_RESTRICTION_PTR(et) = 0
end


#	Add a set of restrictions to a table
procedure	et_add_restriction(et, etr)
pointer		et			#I table
pointer		etr			#I restrictions

int		i

begin
  if (ET_RESTRICTION_LEN(et) == 0) {
    i = 1
    call malloc (ET_RESTRICTION_PTR(et), i, TY_LONG)
  } else {
    i = ET_RESTRICTION_LEN(et) + 1
    call realloc (ET_RESTRICTION_PTR(et), i, TY_LONG)
  }
  ET_RESTRICTION_LEN(et) = i
  ET_RESTRICTION(et,i) = etr
end

#	Add a string restriction to a table
procedure	et_add_string(et, column, str)
pointer		et			#I table
char		column[ARB]		#I column
char		str[ARB]		#I restriction

pointer		etr

begin
  call etr_new_string(etr, ET_TB(et), column, str)
  call et_add_restriction(et, etr)
end

#	Add range restriction to table
procedure	et_add_range(et, column, minv, maxv)
pointer		et			#I table
char		column[ARB]		#I column
real		minv			#I low value
real		maxv			#I high value

pointer		etr

begin
  call etr_new_range(etr, ET_TB(et), column, minv, maxv)
  call et_add_restriction(et, etr)
end

#	Remove last restriction from table
procedure	et_pop_restriction(et)
pointer		et			#I table

int		i

begin
  if (et == 0) return
  if (ET_RESTRICTION_LEN(et) == 0) return
  if (ET_RESTRICTION_LEN(et) == 1) {
    ET_RESTRICTION_LEN(et) = 0
    call mfree (ET_RESTRICTION_PTR(et), TY_LONG)
    ET_RESTRICTION_PTR(et) = 0
  } else {
    i = ET_RESTRICTION_LEN(et) - 1
    call realloc (ET_RESTRICTION_PTR(et), i, TY_LONG)
    ET_RESTRICTION_LEN(et) = i
  }
end

#	Get st table from euv style table
procedure	et_table(et, tb)
pointer		et		#I table
pointer		tb		#O st table

begin
  tb = 0
  if (et == 0) return
  tb = ET_TB(et)
end

#	Release table resources
procedure	et_close(et)
pointer		et			#I table

int		i

begin
  if (et == 0) return

  do i = 1, ET_RESTRICTION_LEN(et) {
    call etr_close(ET_RESTRICTION(et,i))
  }

  if (ET_RESTRICTION_PTR(et) != 0)
    call mfree(ET_RESTRICTION_PTR(et), TY_LONG)

  if (ET_NAME_PTR(et) != 0)
    call mfree(ET_NAME_PTR(et), TY_CHAR)

  call tbtclo(ET_TB(et))
  call et_free(et)
  et = 0
end

#	Get the preceding or succive row value.
procedure	et_move(et, row, dir)
pointer		et			#I table
int		row			#I/O row
int		dir			#I directions

int		newrow
int		i
pointer		tb
real		value
bool		nullp
pointer		sp
pointer		buf
int		goodflag
int		nrows
pointer		etr

int		tbpsta()
bool		streq()

begin
  newrow = row
  row = EOF
  if (et == 0) return
  tb = ET_TB(et)
  if (tb == 0) return

  nrows = tbpsta(tb, TBL_NROWS)

  call smark(sp)
  call salloc(buf, SZ_LINE, TY_CHAR)

  goodflag = NO
  repeat {
    newrow = newrow + dir
    if (newrow < 1 || newrow > nrows) break

    goodflag = YES

    do i = 1, ET_RESTRICTION_LEN(et) {
      etr = ET_RESTRICTION(et, i)
      if (ETR_STRINGP(etr) == 0) {
	call tbrgtr(tb, ETR_COLUMN_ID(etr), value, nullp, 1, newrow)
	if (! nullp) {
	  if (value < ETR_MINIMUM(etr)	||
	      value > ETR_MAXIMUM(etr)) {
	    goodflag = NO
	    break
	  }
	}
      } else {
	call tbrgttn(tb, ETR_COLUMN_ID(etr), Memc[buf], nullp, SZ_LINE,
		    1, newrow)
	if (! nullp) {
	  if (! streq(Memc[buf], ETR_MATCH(etr))) {
	    goodflag = NO
	    break
	  }
	}
      }
    }
  } until (goodflag == YES)

  if (goodflag == YES) {
    row = newrow
  }
  call sfree(sp)
end

#	Get the next row
procedure	et_next(et, row)
pointer		et		#I table
int		row		#I/O new row

begin
  call et_move(et, row, 1)
end

#	Get the previous row
procedure	et_prev(et, row)
pointer		et			#I table
int		row			#I/O new row

begin
  call et_move(et, row, -1)
end

#	Open another table where a previous table
#	existed.
procedure	et_open_new(et, name, et_old)
pointer		et			#O table
char		name[ARB]		#I new table name
pointer		et_old			#I old table pointer

int		len
pointer		sp, buf
int		nlen, dlen, fnldir()
int		strlen()

begin
  call et_open(et, name)
  if (et == 0 && et_old != 0) {
    len = strlen(ET_NAME(et_old))
    nlen = strlen(name)
    call smark(sp)
    call salloc(buf, len+nlen, TY_CHAR)

    dlen = fnldir (ET_NAME(et_old), Memc[buf], len)
    call strcpy(name, Memc[buf+dlen], nlen)
    call et_open(et, Memc[buf])

    call sfree(sp)
  }
end
