# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  mlsops.x
# 	Author	:  ericco
# 	Date	:  04/30/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) mlsops.x -r1.2 04/30/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	05/14/92   ericco	added documentation
# 	04/30/92   ericco	date and time created 92/04/30 00:51:04 by ericco
# H-
# 
#

#	5/14/92

#	Operations on mls structure.  The ml structure is
#	only a component of the mls structure, and should
#	not be accessed directly.

#	An mls is a list of limits.  Limits have a high and
#	low value.  Values of INDEF are allowed.  Limits can
#	be inclusive or exclusive.  Exclusive limits are
#	denoted in the ST Table with a '!' before the monitor
#	name.

include <tbset.h>
include "ml.h"
include "mls.h"

#	Create a dictionary for an array of ml
procedure	mls_make_dictionary (dictionary, mls, length)
pointer		dictionary			#O pointer to string
pointer		mls[length]			#I arrays of limits
int		length				#I array length

int		dictlen
int		i
pointer		ml
pointer		nameptr
int		strlen()

begin
  dictlen = 1
  do i = 1, length {
    ml = mls[i]
    nameptr = ML_NAME_PTR(ml)
    if (ML_NAME(ml) == '!') nameptr = nameptr + 1
    dictlen = dictlen + strlen(Memc[nameptr]) + 1
  }
  call malloc(dictionary, dictlen, TY_CHAR)
  Memc[dictionary] = ','
  Memc[dictionary+1] = EOS
  do i = 1, length {
    ml = mls[i]
    nameptr = ML_NAME_PTR(ml)
    if (ML_NAME(ml) == '!') nameptr = nameptr + 1
    call strcat(Memc[nameptr], Memc[dictionary], dictlen)
    call strcat(",", Memc[dictionary], dictlen)
  }
end

#	Initialize an mls structure
procedure	mls_open(mls)
pointer		mls		#O created mls structure

begin
  call mls_malloc(mls)
  MLS_LIMITS_LEN(mls) = 0
  MLS_MONITORS(mls) = 0
  MLS_LIMITS_PTR(mls) = NULL
  MLS_DICTIONARY_PTR(mls) = NULL
end

#	Create a new mls structure from an ST Table.
procedure	mls_load (mls, table, et_old) # create mls
pointer		mls			#O new mls structure pointer
char		table[ARB]		#I new of ST Table
pointer		et_old			#I old et table or NULL

pointer		et

int		nrows
pointer		ml

begin
  mls = 0

  call et_open_new(et, table, et_old)
  if (et == 0) return

  nrows = ARB
  call et_ml_read(et, ml, nrows)
  call et_close(et)
  if (nrows == 0) return

  call mls_malloc(mls)
  MLS_LIMITS_LEN(mls) = nrows
  MLS_MONITORS(mls) = nrows
  MLS_LIMITS_PTR(mls) = ml

  call mls_make_dictionary (MLS_DICTIONARY_PTR(mls), MLS_LIMITS(mls,1), nrows)
end

#	Release mls structure
procedure	mls_close(mls) # release mls
pointer		mls			#I/O structure pointer

int		i, nrows
pointer		ml

begin
  if (mls == 0) return

  nrows = MLS_LIMITS_LEN(mls)
  do i = 1, nrows {
    ml = MLS_LIMITS(mls,i)
    call mfree(ML_NAME_PTR(ml), TY_CHAR)
    call ml_free(ml)
  }
  call mfree(MLS_DICTIONARY_PTR(mls), TY_CHAR)
  call mfree(MLS_LIMITS_PTR(mls), TY_LONG)
  call mls_free(mls)
end

#	Given a monitor name, extract the associated limits.
#	Currently, each monitor name can have at most a single
#	limit range.  Each range has a low and a high value
#	with a flag indicating inclusive/exclusive.  Length
#	is the maximum number of values to return.  Length
#	returns the actual number of values returned.
procedure	mls_limits (mls, monitor, lows, highs, includeps, length)
pointer		mls			#I structure
char		monitor[ARB]		#I monitor name
real		lows[length]		#O lower limits
real		highs[length]		#O upper limits
int		includeps[length]	#O inclusive flags
int		length			#I/O array lengths

pointer		ml
int		strdic()
int		index

char		junk[1]

begin
  if (length < 1) return

  if (mls == 0) {
    length = 0
    return
  }
  if (MLS_DICTIONARY_PTR(mls) == NULL) {
    length = 0
    return
  }

  index = strdic(monitor, junk, 1, MLS_DICTIONARY(mls))
  if (index < 1  || index > MLS_LIMITS_LEN(mls)) {
    length = 0
    return
  }

  ml = MLS_LIMITS(mls, index)
  if (ml == 0) {
    length = 0
    return
  }
  
  # later we might return more than a single interval
  length = 1
  lows[1] = ML_LOW(ml)
  highs[1] = ML_HIGH(ml)
  if (ML_NAME(ml) == '!') {
    includeps[1] = NO
  } else {
    includeps[1] = YES
  }
end

#	Returns the number of monitors that have limits.
#	Currently this is the same as the number of limits.
#	In the future, the number of monitors could be less
#	than the number of limits.
procedure	mls_number_monitors(mls, length) # number of monitors
pointer		mls			#I structure
int		length			#O number of monitors

begin
  length = 0
  if (mls != 0)  length = MLS_MONITORS(mls)
end

#	Returns the name of the indexth monitor
procedure	mls_monitor(mls, index, name) # get monitor name
pointer		mls			#I structure
int		index			#I which monitor to get
pointer		name			#O pointer to monitor name

pointer		ml
pointer		nameptr

begin
  name = 0
  if (mls == 0) return
  if (index < 1  || index > MLS_LIMITS_LEN(mls)) return

  ml = MLS_LIMITS(mls, index)
  if (ml == 0) return

  nameptr = ML_NAME_PTR(ml)
  if (nameptr == 0) return

  if (Memc[nameptr] == '!') nameptr = nameptr + 1
  name = nameptr
end

#	Set the limits for an existing or new monitor.
procedure	mls_set (mls, monitor, lows, highs, includeps, length)
pointer		mls			#I/O structure
char		monitor[ARB]		#I monitor name
real		lows[length]		#I lower limits
real		highs[length]		#I upper limits
int		includeps[length]	#I inclusive flags
int		length			#I array lengths

pointer		ml
int		strdic()
int		index

char		junk[1]
int		strlen()
int		slen

begin
  if (length < 1) return

  if (mls == 0) {
    length = 0
    return
  }

  if (MLS_DICTIONARY_PTR(mls) != NULL) {
    index = strdic(monitor, junk, 1, MLS_DICTIONARY(mls))
  } else {
	index = 0
	call ds_strcpy(",",MLS_DICTIONARY_PTR(mls))
  }

  if (index < 1  || index > MLS_LIMITS_LEN(mls)) {
    call ml_malloc(ml)
    call ds_strcpy(monitor, ML_NAME_PTR(ml))
    ML_LOW(ml) = INDEFR
    ML_HIGH(ml) = INDEFR
    MLS_LIMITS_LEN(mls) = MLS_LIMITS_LEN(mls) + 1
    MLS_MONITORS(mls) = MLS_LIMITS_LEN(mls)
    call realloc(MLS_LIMITS_PTR(mls), MLS_LIMITS_LEN(mls), TY_LONG)
    MLS_LIMITS(mls, MLS_LIMITS_LEN(mls)) = ml

    slen = strlen(monitor) + strlen(MLS_DICTIONARY(mls)) + 1
    call realloc(MLS_DICTIONARY_PTR(mls), slen, TY_CHAR)
    call strcat(monitor, MLS_DICTIONARY(mls), slen)
    call strcat(",", MLS_DICTIONARY(mls), slen)
    
  } else {
    ml = MLS_LIMITS(mls, index)
  }

  ML_LOW(ml) = lows[1]
  ML_HIGH(ml) = highs[1]
  if (includeps[1] == NO && ML_NAME(ml) != '!') {
    slen = strlen(monitor)
    call realloc(ML_NAME_PTR(ml), slen+1, TY_CHAR)
    call amovc(Memc[ML_NAME_PTR(ml)+1], ML_NAME(ml), slen+1)
    ML_NAME(ml) = '!'
  } else if (includeps[1] == YES && ML_NAME(ml) == '!') {
    slen = strlen(monitor)
    call amovc(ML_NAME(ml), Memc[ML_NAME_PTR(ml)+1], slen)
    call realloc(ML_NAME_PTR(ml), slen-1, TY_CHAR)
  }
end

#	Delete an existing monitor.
procedure	mls_delete (mls, monitor)
pointer		mls			#I/O structure
char		monitor[ARB]		#I monitor name

pointer		ml
int		strdic()
int		index

char		junk[1]
int		i

begin
  if (mls == 0) return

  if (MLS_DICTIONARY_PTR(mls) == NULL)  return

  index = strdic(monitor, junk, 1, MLS_DICTIONARY(mls))
  if (index < 1  || index > MLS_LIMITS_LEN(mls)) return

  # delete element from list
  ml = MLS_LIMITS(mls,index)
  call mfree(ML_NAME_PTR(ml), TY_CHAR)
  call ml_free(ml)

  do i = index+1, MLS_LIMITS_LEN(mls) {
    MLS_LIMITS(mls, i-1) = MLS_LIMITS(mls, i)
  }
  MLS_LIMITS_LEN(mls) = MLS_LIMITS_LEN(mls) - 1
  MLS_MONITORS(mls) = MLS_LIMITS_LEN(mls)
  call realloc(MLS_LIMITS_PTR(mls), MLS_LIMITS_LEN(mls), TY_LONG)

  if (MLS_DICTIONARY_PTR(mls) != 0) {
    call mfree (MLS_DICTIONARY_PTR(mls), TY_CHAR)
  }
  call mls_make_dictionary (MLS_DICTIONARY_PTR(mls),
			    MLS_LIMITS(mls,1), MLS_LIMITS_LEN(mls))
end

#	Write an mls to an ST table
procedure	mls_write(mls, table, status)
pointer		mls				#I structure
char		table[ARB]			#I table name
int		status				#O YES or NO

pointer		tbtopn()
pointer		tb
pointer		ml
int		len
int		nlen
pointer		name
pointer		low
pointer		high

int		i
int		strlen()

begin
  status = NO
  if (mls == 0) return

  if (MLS_LIMITS_LEN(mls) ==  0) return

  len = 0
  do i = 1, MLS_LIMITS_LEN(mls) {
    ml = MLS_LIMITS(mls, i)
    nlen = strlen(ML_NAME(ml))
    if (nlen > len) len = nlen
  }

  iferr { tb = tbtopn(table, NEW_FILE, 0) } then { return }

  call tbcdef (tb, name, "name", "", "%s", -len, 1, 1)
  call tbcdef (tb, low, "low", "", "%f", TY_REAL, 1, 1)
  call tbcdef (tb, high, "high", "", "%f", TY_REAL, 1, 1)

  if (name != 0 && low != 0 && high != 0) {
    iferr {call tbtcre(tb)} then { return }

    do i = 1, MLS_LIMITS_LEN(mls) {
      ml = MLS_LIMITS(mls, i)
      nlen = strlen(ML_NAME(ml))
      call tbrptt (tb, name, ML_NAME(ml), nlen, 1, i)
      call tbrptr (tb, low, ML_LOW(ml), 1, i)
      call tbrptr (tb, high, ML_HIGH(ml), 1, i)
    }
  }

  call tbtclo(tb)
  status = YES
end
