/*	Revised: 4/8/98 					*/
/*      @(#)c_contact.c	1.4 UCB CEA EGO */
/*
/*	8 April 1998: Modified by Bill Boyd
/*	I removed the "ifdef BSDUNIX" and "ifdef SUNOS4" conditions in
/*	order to compile with IRAF 2.11, where these are not defined
/*	in the iraf.h file.  This code should only be compiled under
/*	SunOS and never under Solaris.
/*
/*
**	Make a connection to the telemetry server
*/

#define import_libc
#define NOLIBCNAMES
#define import_spp
#define import_xnames  
#include	<iraf.h>
#include	<stdio.h>

#define	EGOServerOpen	eusopn_
#define	EGOServerMode	eusmod_
#define	EGOServerClose	euscls_
#define	EGOServerCloseWrite	eusclw_
#define	EGOServerSend	eussnd_
#define	EGOServerRecv	eusrcv_
#define	EGOServerError	euserr_

#define	BAD_CONTACT	-1
#define MAX_CONNECTIONS 10

typedef struct {
  FILE*	soc_read;
  FILE*	soc_write;
  int	descriptor;
  int	mode;
} soc_connection;

static soc_connection soc_conn[MAX_CONNECTIONS+1]; /* An array of socket connection 
						 ** structures. This allows for more 
						 ** than one connection to the server 
						 */





/*
**	Get the length of an SPP string.
*/
static	Xstrlen (str)
XCHAR	*str;
{
  int	length = 0;
  if (str) {
    while (*str++)
      length++;
  }
  return length;
}

/*
**	Get a new copy of a C string from an SPP string
*/
static	char	*newcstr (str)
XCHAR	*str;
{
  int	length = Xstrlen(str)+1;
  char  *malloc();
  char	*ret = malloc(length);
  char	*ptr = ret;

  while (length--)
    *ptr++ = *str++;
  return ret;
}


#define	MADE
/*#include <stdio.h>
#include <fcntl.h>*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <netinet/in.h>
#include <time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <syslog.h>
#include <errno.h>

static	int	save_errno = 0;

/*
**
*/
XINT	EGOServerError()
{
  return (XINT)save_errno;
}

/*
**	Open a connection to a service on a host
*/
XINT	EGOServerOpen (Xservice, Xhost)
XCHAR	*Xservice;
XCHAR	*Xhost;
{
  char	*ptr;
  struct sockaddr_in	sin;
  struct servent	*sp;
  struct hostent	*hp;
  int	descriptor, client;
  char	*protocol = "tcp";
  int	conn;
  int 	i;
  FILE	*read_stream, *write_stream;

  ptr = newcstr(Xservice);
  sp = getservbyname (ptr, protocol);
  free(ptr);
  if (sp == NULL)
    return BAD_CONTACT;

  ptr = newcstr(Xhost);
  hp = gethostbyname(ptr);
  free(ptr);
  if (hp == NULL)
    return BAD_CONTACT;

  memset ((char *)&sin, 0, sizeof(sin));
  memcpy((char *)&sin.sin_addr, hp->h_addr, hp->h_length);
  sin.sin_family= hp->h_addrtype;
  sin.sin_port= htons(sp->s_port);

  if (descriptor = socket(hp->h_addrtype, SOCK_STREAM, 0),
      descriptor < 0)
    return BAD_CONTACT;


  for ( i=1; MAX_CONNECTIONS; i++ ){
    if ( soc_conn[i].soc_read == NULL ){
      if ( (read_stream = fdopen( descriptor, "r")) 		&&
	   (write_stream = fdopen( descriptor, "w")) ){
	soc_conn[i].soc_read = read_stream;
	soc_conn[i].soc_write = write_stream;
	soc_conn[i].descriptor = descriptor;
	break;
      }
      else{
	return BAD_CONTACT;	  
      }
    }
  }

  conn = connect (descriptor, (struct sockaddr *)&sin, sizeof(sin));
  if (conn != 0) {
    close(descriptor);
    return BAD_CONTACT;
  }

  return (XINT)( i );
}

/*
**	Close a connection to a service
*/
	EGOServerClose (descriptor_index)
XINT	*descriptor_index;
{
  fclose(soc_conn[(int)*descriptor_index].soc_read);
  fclose(soc_conn[(int)*descriptor_index].soc_write);
  soc_conn[(int)*descriptor_index].descriptor = 0;
}

/*
**	Toggle a connection mode between ASCII and binary
*/
	EGOServerMode (descriptor_index)
XINT	*descriptor_index;
{
  if ( soc_conn[(int)*descriptor_index].mode )
    soc_conn[(int)*descriptor_index].mode = 0;
  else
    soc_conn[(int)*descriptor_index].mode = 1;

}

/*
**	Close a write connection to a service
*/
	EGOServerCloseWrite (descriptor_index)
XINT	*descriptor_index;
{
  
  shutdown((int)(soc_conn[(int)*descriptor_index].descriptor), 1);
}

/*
**	Send a string to a service
*/
XINT	EGOServerSend (descriptor_index, msg)
XINT	*descriptor_index;
XCHAR	*msg;
{
  char	*ptr = newcstr(msg);
  int	ret = BAD_CONTACT;

  if ( fprintf(soc_conn[(int)*descriptor_index].soc_write,"%s",ptr) > 0 ){
    fflush(soc_conn[(int)*descriptor_index].soc_write);
    ret = 0;
  }

  free(ptr);
  return ret;
}

/*
**	Recieve a string from a service
*/
XINT	EGOServerRecv (descriptor_index, msg, msglen)
XINT	*descriptor_index;
XCHAR	*msg;
XINT    *msglen;
{
  XCHAR	*spp_ptr = msg;
  char	*bptr = (char *)msg;
  int   len = (int)*msglen;
  int   ret;
  int	red;
  char	*ptr, *tmp;
  int	i = 0;


#ifdef	FORCE_ERROR
  static int	testerror = 1000;
  if (testerror-- < 0) {testerror = 1000; save_errno = 31415;
			return BAD_CONTACT; }
#endif

  if (len == 0) return 0;
  save_errno = 0;

  if (soc_conn[(int)*descriptor_index].mode != 0){
    if ( fread( bptr, len * sizeof(XCHAR), 1,
	 soc_conn[(int)*descriptor_index].soc_read ) != 1 ){
      save_errno = errno;
      return BAD_CONTACT;
    }
  }
  else
  {
      ret = fgetc(soc_conn[(int)*descriptor_index].soc_read);
      if ( ret == EOF )
      {
	  save_errno = errno;
	  return BAD_CONTACT;
      }else
	  *spp_ptr = (char)ret;
      while ( *spp_ptr != '\n' && i < (len-1) )
      {
	  i = i + sizeof(XCHAR)/sizeof(char);
	  ret = fgetc(soc_conn[(int)*descriptor_index].soc_read);
	  if ( ret == EOF )
	  {
	      save_errno = errno;
	      return BAD_CONTACT;
	  }else
	      *(++spp_ptr) = (char)ret;
      }
      *(++spp_ptr) = 0;

    }

  return 0;
}

