# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  hcpkill.x
# 	Author	:  tnk
# 	Date	:  04/08/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) hcpkill.x -r1.19 04/08/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/13/97   bboyd	changed SF_INHERIT to public QPF_INHERIT and added qpset.h include
# 	10/13/97   bboyd	removed obsolete dsym
# 	10/13/97   bboyd	replaced qp_dtype with qp_dtypeeuv.  Removed private QPOE 
# includes.  Added qpeuv.h include
# 	01/13/97   tnk	changed patterns in kill lists
# 	01/10/97   mabbott	new pattern matching for regex
# 	12/20/96   tnk	added TDMIN/MAX TLMIN/MAX COMMENT cards to TB_KILL list
# 	10/24/96   tnk	added TIMEDEL to kill list
# 	10/10/96   tnk	allow type bool for qpoe input
# 	06/06/96   gareth	changed RATE[0-9]* to RATE? RATE?? RATE??? to fix bug ( apparently in
# iraf code )
# 	04/24/96   gareth	changed RATE_??? to RATE[0-9]*
# 	10/18/95   tnk	closed list
# 	09/12/95   tnk	added INHERIT flag check to qpoe headers
# 	02/08/95   tnk	fixed NULL pointer in qp_dtype call
# 	08/30/94   tnk	switched from imgftype to imhdtype
# 	08/19/94   tnk	added call to imcpcmnt
# 	08/18/94   tnk	added COMMENT and HISTORY to imkill list;  increased kill bufferlen
# 	08/03/94   tnk	convert reals to doubles on image headers
# 	06/13/94   tnk	added BANDID to IM_KILL list
# 	04/08/94   tnk	date and time created 94/04/08 15:41:35 by tnk
# H-
# 
#

include <tbset.h>
include <imhdr.h>
include <imio.h>
include <qpeuv.h>
include <qpset.h>

# @(#)hcpkill.x	1.19 10/13/97 UCB CEA EGO

# define the internal cards to omit
define  IM_OMIT "i_?* WCSDIM WAT?_??? CTYPE? CRVAL? CRPIX? CDELT? CD?_? LTM?_? IRAF-??? LTV?_? XPLANE YPLANE FVALUE"  
define  QP_OMIT "naxes axlen datamin datamax cretime modtime limtime"
define  TB_OMIT " "

# define default cards to kill, based on input file type
define IM_KILL "DEFATTR? DEFFILT? QPFILT?? XS-EVENT XS-SORT BANDID? COMMENT HISTORY"
define QP_KILL "DEFATTR? DEFFILT? QPFILT?? XS-EVENT XS-SORT"
define TB_KILL "RATE[0-9][0-9]* DELTIME TIMEDEL TLMIN[0-9][0-9]* TLMAX[0-9][0-9]* TDMIN[0-9][0-9]* TDMAX[0-9][0-9]*"

define SZ_KILL 1024
#-------------------------------------------------------------
# Copy all header cards from in_fp to out_fp, 
# except those that match entries in the user_kill list.
# The user_kill list entries can have single char '?' wildcards in them.
# Sample call:   hcpkill(in,"QPOE",out,"IMAGE","FOO BAR??")
# If user_kill is null, then a predefined kill list (defined by input
# file type) is used.
# The kill list is case insensitive.
#
# Note that all header cards are also filtered by a separate 'OMIT'
# list, that eliminates housekeeping type cards.
#
procedure hcpkill(in_fp, in_ftype, out_fp, out_ftype, user_kill)

pointer in_fp           #I   opened input file
char    in_ftype[ARB]   #I   file type ("TABLE" | "QPOE" | "IMAGE")  
pointer out_fp          #I/O opened output file
char    out_ftype[ARB]  #I   file type ("TABLE" | "QPOE" | "IMAGE")
char    user_kill[ARB]  #I   user's list of cards to kill

pointer kill_list
pointer sp

begin

   call smark(sp)
   call salloc(kill_list, SZ_KILL,  TY_CHAR)

   switch (in_ftype[1]) {
   case 'Q':   # QPOE
      call cat_strings(QP_OMIT, user_kill, QP_KILL, Memc[kill_list], SZ_KILL)
      call hqp_cpy(in_fp, out_fp, out_ftype, Memc[kill_list])

   case 'I':   # IMAGE
      call cat_strings(IM_OMIT, user_kill, IM_KILL, Memc[kill_list], SZ_KILL)
      call him_cpy(in_fp, out_fp, out_ftype, Memc[kill_list])

   case 'T':   # TABLE 
      call cat_strings(TB_OMIT, user_kill, TB_KILL, Memc[kill_list], SZ_KILL)
      call htb_cpy(in_fp, out_fp, out_ftype, Memc[kill_list])

   default:
      call error(100,"hcpkill: invalid input file type")
   }    

   call sfree(sp)

end

#----------------------------------------------------------------------
# copy header cards from a QPOE file to the output file,
# except for those on the omit or kill lists
#
procedure hqp_cpy(in_fp, out_fp, out_ftype, kill_list)
pointer in_fp           #I   opened input file
pointer out_fp          #I/O opened output file
char    out_ftype[ARB]  #I   file type ("TABLE" | "QPOE" | "IMAGE" min match)
char    kill_list[ARB]  #I   string of kill patterns

pointer sp
pointer parmname   # parameter name
pointer list       # list of symbols in qpoe table
pointer qpdatatype # qp datatype string (i, d, macro, events, etc.)
pointer comment    # qpoe symbol comment string
int     maxelem    # max elements allowed in a qpoe entry (determines lfile)
int     flags      # qpoe flags
pointer text
int     int_dtype
int     nelem


# FUNCTIONS
pointer qp_ofnlu()
int     qp_gnfn()
int     qp_queryf()   # Get information describing the named parameter
int     qp_dtypeeuv()
bool    match_list()

begin 

   call smark(sp)
   call salloc(parmname,   SZ_FNAME, TY_CHAR)
   call salloc(qpdatatype, SZ_FNAME, TY_CHAR)
   call salloc(comment,    SZ_LINE,  TY_CHAR)
   call salloc(text,       SZ_LINE,  TY_CHAR)

   list = qp_ofnlu (in_fp, "*")
   while (qp_gnfn (list, Memc[parmname], SZ_FNAME) != EOF) {

      nelem = qp_queryf (in_fp, Memc[parmname], Memc[qpdatatype], maxelem,
         Memc[comment], flags)

      # if it's of suitable data type and the INHERIT flag is set, copy it
      int_dtype = qp_dtypeeuv(in_fp, Memc[qpdatatype])
      if (int_dtype >= TY_BOOL && int_dtype < TY_COMPLEX &&
	   (and (flags, QPF_INHERIT) != 0))
      {
         # pull each pattern out of kill list & compare against parm_name
         if (! match_list(kill_list,Memc[parmname]) )
         {
             call hdrcpy(in_fp,"QPOE",out_fp,out_ftype, Memc[parmname],
                int_dtype)
         }
      }
   }

   call sfree(sp)

end

#----------------------------------------------------------------------
# copy header cards from an IRAF image file to the output file
# except for those on the omit or kill lists
#
procedure him_cpy(in_fp, out_fp, out_ftype, kill_list)
pointer in_fp           #I   opened input file
pointer out_fp          #I/O opened output file
char    out_ftype[ARB]  #I   file type ("TABLE" | "QPOE" | "IMAGE" min match)
char    kill_list[ARB]  #I   string of kill patterns

pointer sp
pointer parmname   # parameter name
pointer list       # list of symbols in image
int     int_dtype

# FUNCTIONS
pointer im_ofnlu()
int     im_gnfn()
bool    match_list()
int     imhdtype()

begin

   call smark(sp)
   call salloc(parmname,   SZ_FNAME, TY_CHAR)

   list = im_ofnlu (in_fp, "*")
   while (im_gnfn (list, Memc[parmname], SZ_FNAME) != EOF) {

      # pull each pattern out of kill list & compare against parm_name
      if (! match_list(kill_list,Memc[parmname]) )
      {
         int_dtype = imhdtype(in_fp, Memc[parmname])
   
         # convert all real values to doubles to prevent loss of accuracy
         if (int_dtype == TY_REAL)
            int_dtype = TY_DOUBLE

         call hdrcpy(in_fp,"IMAGE",out_fp,out_ftype,Memc[parmname],
            int_dtype)
      }   
   }
   call imcfnl(list)

   # if this is an image to image transfer, copy over COMMENT/HISTORY
   if (out_ftype[1]=='I')
      call imcpcmnt(in_fp,out_fp)

   call sfree(sp)

end

#----------------------------------------------------------------------
# copy header cards from an ST table file to the output file,
# except for those on the omit or kill lists
#
procedure htb_cpy(in_fp, out_fp, out_ftype, kill_list)
pointer in_fp           #I   opened input file
pointer out_fp          #I/O opened output file
char    out_ftype[ARB]  #I   file type ("TABLE" | "QPOE" | "IMAGE" min match)
char    kill_list[ARB]  #I   string of kill patterns

pointer sp
pointer parmname   # parameter name
int     n_cards
int     i
pointer parmval    # parameter value
int     int_dtype

# FUNCTIONS
int     tbpsta()      # table status information
bool    match_list()
bool	streq()

begin

   call smark(sp)
   call salloc(parmname,   SZ_FNAME, TY_CHAR)
   call salloc(parmval,    SZ_LINE,  TY_CHAR)

   n_cards = tbpsta(in_fp, TBL_NPAR)
   for (i=1; i<= n_cards; i=i+1)
   {
      call tbhgnp (in_fp, i, Memc[parmname], int_dtype, Memc[parmval])

      # skip blank cards
      if (streq(Memc[parmname], "")) next

      # skip the known unwanted cards
      if (! match_list(kill_list,Memc[parmname]) )
      {
         call hdrcpy(in_fp,"TABLE",out_fp,out_ftype,Memc[parmname],
            int_dtype)
      }
   }
   call sfree(sp)

end

#-----------------------------------------
# cpy str1 onto outstr,
# then cat a space, then
# if str2 is defined, cat it to outstr, otherwise cat alt_str2

procedure cat_strings(str1, str2, alt_str2, outstr, maxout)
char  str1[ARB]
char  str2[ARB]
char  alt_str2[ARB]
char  outstr[ARB]
int   maxout

#FUNCTIONS
int   strlen()

begin

   call strcpy(str1, outstr, maxout)
   call strcat(" ", outstr, maxout)
   if (strlen(str2) != 0)
      call strcat(str2, outstr, maxout)
   else 
      call strcat(alt_str2, outstr, maxout)

end
