# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  qpparseeuv.x
# 	Author	:  bboyd
# 	Date	:  09/12/97
# 	Synopsis:  
# 	Keywords:
# 		@(#) qpparseeuv.x -r1.7 09/12/97 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/02/97   bboyd	FORTRAN arrays begin with 1, not 0!!! 
# 	09/25/97   bboyd	Fixed some logical bugs in the processing flow
# 	09/18/97   bboyd	rewritten for euv package
# 	09/15/97   mabbott	temp fixes to compile cleanly
# 	09/12/97   mabbott	replaced qpeuv.h reference
# 	09/12/97   mabbott	temp remove include qpeuv.h
# 	09/12/97   bboyd	date and time created 97/09/12 11:57:29 by bboyd
# H-
# 
#


include <iraf.h>
include <mach.h>
include	<qpeuv.h>

define MAX_FIELDS 50

# @(#)qpparseeuv.x	1.7 10/2/97 UCB CEA EGO
#
#
# QP_PARSEEUV -- Parse the field list, or declarations string for a user
# defined datatype (structure or domain).
#
# Syntax:	{ type1, type2, ..., typeN }
#
# e.g.,		{d,s:x,s:y,s,s,s,s}	(Rosat/PROS event structure)
#
# where the TYPEi are primitive types, e.g., "r" or "real", "i" or "int",
# etc.  Selected fields may have ":x" or ":y" appended to indicate that these
# are the default coordinate fields to be used for position based extraction.
# Fields will be automatically aligned as necessary, and the computed structure
# size will be forced to be an integral multiple of the largest datatype
# within the structure, to ensure proper alignment in arrays of the structures.
#
# Based upon the IRAF QPOE system routine QP_PARSEFL.  Rewritten for 
# use in the euv package.

int procedure qp_parseeuv (qp, fieldlist, dd)

pointer	qp			#I QPOE descriptor
char	fieldlist[ARB]		#I field list defining new datatype (domain)
pointer	dd			#U pointer to domain descriptor

int	nfields, offset, maxsize, xfield, yfield, token, dtype, fsize, i
char	tokchar


int	sizeof()
string	qperr "qp_parseeuv: QPOE structdef"
define	nextfield_ 91

begin

	# Advance to structure terms list.
	i = 1
	while (fieldlist[i] != EOS) {
	    token = fieldlist[i]
	    if (token == '{') {
		i = i+1
		break
	    } 
	    else
		i = i+1
	}

	nfields = 0
	offset  = 0
	maxsize = 0
	xfield  = 0
	yfield  = 0
	
	# Process the structure terms list.
	repeat {
	    token = fieldlist[i]

	    switch (token) {

	    case EOS, '}':
		break
	    case ',':
		i = i+1
		next
	    case 'b','c','s','i','l','r','d','x':
		i = i+1
		nfields = nfields + 1

	  	switch (token) {
		case 'b':
		    dtype = TY_BOOL
	    	case 'c':
		    dtype = TY_CHAR
	    	case 's':
		    dtype = TY_SHORT
	    	case 'i':
		    dtype = TY_INT
	    	case 'l':
		    dtype = TY_LONG
	    	case 'r':
		    dtype = TY_REAL
	    	case 'd':
		    dtype = TY_DOUBLE
	    	case 'x':
		    dtype = TY_COMPLEX
		}
		if ( dtype < TY_BOOL || dtype > TY_COMPLEX) {
		    tokchar = token
		    call eprintf ("%s: bad field type `%s'\n")
		    call pargstr (qperr)
		    call pargstr (tokchar)
		    goto nextfield_
		} else 
		    fsize = sizeof (dtype) * SZB_CHAR
		    
		
		# Output field descriptor.
		if (nfields > MAX_FIELDS) {
		    tokchar = token
		    call eprintf ("%s: too many fields `%s'\n")
			call pargstr (qperr)
			call pargstr (tokchar)
		    break
		}

		EUVDD_FOFFSET(dd,nfields) = (offset + fsize-1) / fsize

		EUVDD_FTYPE(dd,nfields)   = dtype

		# Update structure size parameters.
		offset = (EUVDD_FOFFSET(dd,nfields) * fsize) + fsize
		maxsize = max (maxsize, fsize)

		if (fieldlist[i] == ':') {
		    i = i+1
		    repeat {
		    	token = fieldlist[i]

		    	switch (token) {
		    	case 'x':
		 	    if (xfield != 0) {
			    	tokchar = token
			    	call eprintf ("%s: duplicate X field `%s'\n")
			    	call pargstr (qperr)
			    	call pargstr (tokchar)
			    }
			    i = i+1
			    xfield = nfields
			    break
		    	case 'y':
		 	    if (yfield != 0) {
			    	tokchar = token
			    	call eprintf ("%s: duplicate Y field `%s'\n")
			    	call pargstr (qperr)
			    	call pargstr (tokchar)
			    }
			    i = i+1
			    yfield = nfields
			    break
		    	default:
			    tokchar = token
			    call eprintf ("%s: unknown : field modifier `%s'\n")
			    call pargstr (qperr)
			    call pargstr (tokchar)

		    	}
		    }
		}
		goto nextfield_
	    default:
		tokchar = token
		call eprintf ("%s: unexpected datatype `%s'\n")
		    call pargstr (qperr)
		    call pargstr (tokchar)
	    }

nextfield_
	    # Read and discard tokens until we get to the next field.
	    while (fieldlist[i] != EOS && fieldlist[i] != '}') {
		if (fieldlist[i] == ',')
		    break
	    	i = i+1
	    }

	}

	# Complete the domain descriptor initialization.
	EUVDD_NFIELDS(dd) = nfields
	EUVDD_XFIELD(dd)  = xfield
	EUVDD_YFIELD(dd)  = yfield

	# Pad the struct size to an integral multiple of the max field size.
	if (nfields > 0) {
	    maxsize = max (SZ_STRUCT, maxsize)
	    EUVDD_STRUCTLEN(dd) = (offset+maxsize-1)/maxsize*maxsize / SZ_STRUCT
	} else
	    EUVDD_STRUCTLEN(dd) = 0


	return (nfields)
end
