# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  qprlmergeeuv.gx
# 	Author	:  bboyd
# 	Date	:  10/13/97
# 	Synopsis:  
# 	Keywords:
# 		@(#) qprlmergeeuv.gx -r1.1 10/13/97 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/13/97   bboyd	date and time created 97/10/13 14:07:18 by bboyd
# H-
# 
#


include <mach.h>
include <qpeuv.h>

# QP_RLMERGEEUV -- Merge (AND) two range lists.  Only ranges which are
# common to both range lists are output.  The number of ranges in the
# output range list is returned as the function value.
#
#  Based on qp_rlmerge and modified for use in the euv package.
#
#  @(#)qprlmergeeuv.gx	1.1 10/13/97 UCB CEA EGO

int procedure qp_rlmergeeuvr (os,oe,olen, xs,xe,nx, ys,ye,ny)

pointer	os, oe			#U output range list
int	olen			#U allocated length of OS, OE arrays

real	xs[ARB], xe[ARB]	#I range list to be merged with
int	nx			#I number of ranges in X list
real	ys[ARB], ye[ARB]	#I range list to be merged with X
int	ny			#I number of ranges in Y list

real	o1, o2
int	nx_out, xi, yi, i
real	qp_minvaleuvr(), qp_maxvaleuvr()
bool	qp_lessthaneuvr()
errchk	realloc

begin
	nx_out = 0
	if (nx <= 0 || ny <= 0)
	    return (0)

	xi = 1
	yi = 1

	do i = 1, ARB {
	    # Find a pair of ranges which intersect.
	    repeat {
		if (qp_lessthaneuvr (xe[xi], ys[yi])) {
		    if (xi >= nx)
			return (nx_out)
		    else
			xi = xi + 1
		} else if (qp_lessthaneuvr (ye[yi], xs[xi])) {
		    if (yi >= ny)
			return (nx_out)
		    else
			yi = yi + 1
		} else
		    break
	    }

	    # Compute the intersection.
	    o1 = qp_maxvaleuvr (xs[xi], ys[yi])
	    o2 = qp_minvaleuvr (xe[xi], ye[yi])

	    # Output the range.
	    if (nx_out + 1 > olen) {
		olen = max (DEF_XLEN, olen * 2)
		call realloc (os, olen, TY_REAL)
		call realloc (oe, olen, TY_REAL)
	    }

	    Memr[os+nx_out] = o1
	    Memr[oe+nx_out] = o2
	    nx_out = nx_out + 1

	    # Advance to the next range.
	    if (xi < nx && qp_lessthaneuvr (xe[xi], ye[yi]))
		xi = xi + 1
	    else if (yi < ny)
		yi = yi + 1
	    else
		break
	}

	return (nx_out)
end


# QP_MINVALEUV -- Return the lesser of two values, where either value can
# be an open range.

real procedure qp_minvaleuvr (x, y)

real	x			#I first value
real	y			#I second value

bool	qp_lessthaneuvr()

begin
	if (qp_lessthaneuvr (x, y))
	    return (x)
	else
	    return (y)
end


# QP_MAXVALEUV -- Return the greater of two values, where either value can
# be an open range.

real procedure qp_maxvaleuvr (x, y)

real	x			#I first value
real	y			#I second value

bool	qp_lessthaneuvr()

begin
	if (qp_lessthaneuvr (x, y))
	    return (y)
	else
	    return (x)
end


# QP_LESSTHANEUV -- Test if X is less than Y, where X and Y can be open
# range values.

bool procedure qp_lessthaneuvr (x, y)

real	x			#I first value
real	y			#I second value

begin
	if (IS_EUVLEFTR(x))
	    return (!IS_EUVLEFTR(y))
	else if (IS_EUVRIGHTR(x))
	    return (false)
	else if (IS_EUVLEFTR(y))
	    return (false)
	else if (IS_EUVRIGHTR(y))
	    return (true)
	else
	    return (x < y)
end
