# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  hashops.x
# 	Author	:  ericco
# 	Date	:  04/30/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) hashops.x -r1.4 04/30/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	11/24/92   ericco	use IRAF system hash
# 	11/17/92   ericco	added map function
# 	04/30/92   ericco	date and time created 92/04/30 01:45:48 by ericco
# H-
# 
#

include	<ctype.h>

include "il.h"
include "hash.h"

define	DEFAULT_HASH_SIZE	26

#	Release hash resources
procedure	hash_close(hash) #E 
pointer		hash			#I hash table

int		i

begin
  do i = 1, HASH_LIST_LEN(hash) {
    call il_close(HASH_LIST(hash,i), HASH_RELEASE(hash))
  }
  call hash_free(hash)
end

#	Open a hash table
procedure	hash_open(hash, release) #E
pointer		hash			#O hash table
pointer		release			#I release function

int		hlen

begin
  call hash_malloc(hash)
  HASH_RELEASE(hash) = release
  hlen = DEFAULT_HASH_SIZE
  HASH_LIST_LEN(hash) = hlen
  call calloc(HASH_LIST_PTR(hash), hlen, TY_LONG)
  HASH_NEXT(hash) = NULL
end

#	Get a copy of the label value
procedure	hash_get(hash, str, item) #E 
pointer		hash			#I table
char		str[ARB]		#I label
pointer		item			#O value

pointer		il

begin
  item = 0
  call hash_get_il(hash, str, il)
  if (il != 0) call item_deep_copy(IL_VALUE(il), item)
end

#	Lookup hash value
procedure	hash_get_il(hash, str, il) #E 
pointer		hash			#I table
char		str[ARB]		#I label
pointer		il			#O pointer

string		default_label "default"

begin
  il = 0
  if (hash == 0) return
  if (str[1] == EOS) {
    call hash_get_location(hash, default_label, il, 0)
  } else {
    call hash_get_location(hash, str, il, 0)
  }
end

#	Check if label exists in hash table
procedure	hash_exists(hash, str, item) #E 
pointer		hash			#I table
char		str[ARB]		#I label
pointer		item			#O pointer

pointer		il

begin
  call hash_get_location(hash, str, il, 1)
  if (il != 0)	item = IL_VALUE(il)
  else		item = 0
end

#	Hash a string
procedure	hash_hash(hash, str, hvalue)
pointer		hash			#I table
char		str[ARB]		#I label
int		hvalue			#O hash

#int		i
#int		hlen
int		sthash()	# iraf system hash function

begin
  hvalue = sthash(str, HASH_LIST_LEN(hash))

#  hlen = HASH_LIST_LEN(hash)
#  hvalue = 0
#  for (i = 1; str[i] != EOS; i=i+1) hvalue = hvalue + str[i]
#  hvalue = hvalue - (hvalue / hlen) * hlen
#  hvalue = hvalue + 1
end

#	Get a pointer to the itemlist node
procedure	hash_get_location(hash, str, il, exists_only)
pointer		hash			#I table
char		str[ARB]		#I label
pointer		il			#O pointer
int		exists_only		#I checking only flag

int		hvalue
char		fch
pointer		lptr
pointer		top
pointer		sp, name
int		i

bool		streq()
int		strlen()

begin
  call smark(sp)
  call salloc(name, strlen(str), TY_CHAR)
  for (i = 1; str[i] != EOS; i = i + 1) {
    if (IS_LOWER(str[i])) {
      Memc[name+i-1] = TO_UPPER(str[i])
    } else {
      Memc[name+i-1] = str[i]
    }
  }
  Memc[name+i-1] = EOS

  call hash_hash(hash, Memc[name], hvalue)
  top = HASH_LIST(hash, hvalue)
  fch = Memc[name]
  lptr = 0

  for (il = top; il != 0; il = IL_NEXT(il)) {
    if (IL_NAME(il) == fch && streq(IL_NAME(il), Memc[name])) break
    lptr = il
  }

  # make item, if not found
  if (exists_only == 0 && il == 0) call il_new(il, Memc[name])

  if (il != 0) {
    # unlink item, if necessary
    if (lptr != 0) IL_NEXT(lptr) = IL_NEXT(il)

    # move item to the front of the existing list
    if (top != il) IL_NEXT(il) = top

    # set the hash list to the item
    HASH_LIST(hash, hvalue) = il
  }

  call sfree(sp)
end

#	Copy item into label value
procedure	hash_put(hash, str, item) #E 
pointer		hash			#I table
char		str[ARB]		#I label
pointer		item			#I value

pointer		il

begin
  if (item == 0) return
  call hash_get_il(hash, str, il)
  if (il != 0) {
    call item_release(IL_VALUE(il), HASH_RELEASE(hash))    
    call item_deep_copy(item, IL_VALUE(il))
  }
end


#
# hash_next()
#

int	procedure	hash_next(hash, nxt, str, item)

pointer	hash			#I hash table
int	nxt			#O next hash item
char	str[ARB]		#O string value
pointer	item			#O item value

int	il_next()

begin

	if(nxt > HASH_LIST_LEN(hash)) return(0)
	if(nxt <= 0) {
		nxt = 1
		HASH_NEXT(hash) = HASH_LIST(hash, nxt)
	}

	while (il_next(HASH_NEXT(hash), str, item) == 0) {
		nxt = nxt + 1
		if(nxt > HASH_LIST_LEN(hash)) return(0)
		HASH_NEXT(hash) = HASH_LIST(hash, nxt)
	}

	return(1)
end

