# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  hdops.x
# 	Author	:  ericco
# 	Date	:  04/30/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) hdops.x -r1.3 04/30/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	08/23/94   tnk	added data type to mfree call
# 	04/30/92   ericco	date and time created 92/04/30 01:45:50 by ericco
# H-
# 
#

include "hd.h"
# @(#)hdops.x	1.3 8/23/94

define DEFAULT_HISTORY_SIZE	100

#	Release history resources
procedure	hd_close(hd)
pointer		hd			#I history

begin
  call hd_purge(hd)
  call mfree(HD_HISTORY_PTR(hd), TY_LONG)
  call hd_free(hd)
  hd = 0
end

#	Remove data from history
procedure	hd_purge(hd)
pointer		hd			#I history

int		hlen
int		i

begin
  hlen = HD_HISTORY_LEN(hd)
  do i = 1, hlen {
    call item_release(HD_HISTORY(hd,i),HD_RELEASE(hd))
  }
end

#	Create a history
procedure	hd_open(hd, ds_release)
pointer		hd			#O history
pointer		ds_release		#I structure release function

int		hlen

begin
  call hd_malloc(hd)
  hlen = DEFAULT_HISTORY_SIZE
  HD_HISTORY_LEN(hd) = hlen
  call calloc(HD_HISTORY_PTR(hd), hlen, TY_LONG)
  HD_CURRENT(hd) = 1
  HD_RELEASE(hd) = ds_release
end

#	Resize length of history
procedure	hd_resize(hd, length)
pointer		hd			#I history
int		length			#I new length

int		hlen
int		current
pointer		hptr
int		i
int		hitem

begin
  if (hd == 0) return
  hlen = HD_HISTORY_LEN(hd)
  if (length < 1) return
  if (length == hlen) return

  current = HD_CURRENT(hd)
  hptr = HD_HISTORY_PTR(hd)
  call malloc(HD_HISTORY_PTR(hd), length, TY_LONG)

  if (length < hlen) {
    do i = length+1, hlen {
      hitem = current - i
      if (hitem < 1) hitem = hitem + hlen
      call item_release(HD_HISTORY(hd,hitem),HD_RELEASE(hd))
    }
    do i = 1, length {
      hitem = current - i
      if (hitem < 1) hitem = hitem + hlen
      HD_HISTORY(hd,i) = Meml[hptr+hitem-1]
    }
    current = hlen
  } else {
    call malloc(HD_HISTORY_PTR(hd), length, TY_LONG)
    do i = 1, hlen {
      hitem = current - i
      if (hitem < 1) hitem = hitem + hlen
      HD_HISTORY(hd,i) = Meml[hptr+hitem-1]
    }
    current = hlen
    do i = hlen+1, length {
      call item_new_empty(HD_HISTORY(hd,i))
    }
  }

  call mfree(hptr,TY_LONG)
  HD_HISTORY_LEN(hd) = length
  HD_CURRENT(hd) = current
end

#	Returns a copy of the indexth item in the history
procedure	hd_get(hd, index, item)
pointer		hd			#I history
int		index			#I history index
pointer		item			#O data pointer

int		hitem
int		current

begin
  item = 0
  if (hd == 0) return
  if (index > HD_HISTORY_LEN(hd)) return
  if (index < 1) return
  current = HD_CURRENT(hd)

  hitem = current - index
  if (hitem < 1) hitem = hitem + HD_HISTORY_LEN(hd)
  call item_deep_copy(HD_HISTORY(hd,hitem), item)
end

#	Sets the indexth item in the history
procedure	hd_put(hd, index, item)
pointer		hd			#I history
int		index			#I history index
pointer		item			#I data pointer

int		hitem
int		current

begin
  if (hd == 0) return
  if (index > HD_HISTORY_LEN(hd)) return
  if (index < 1) return
  current = HD_CURRENT(hd)

  hitem = current - index
  if (hitem < 1) hitem = hitem + HD_HISTORY_LEN(hd)
  call item_release(HD_HISTORY(hd,hitem),HD_RELEASE(hd))
  HD_HISTORY(hd,hitem) = item
end

#	Sets the current item in the history
procedure	hd_push(hd, item)
pointer		hd			#I history
pointer		item			#I item to add

int		hlen
int		current

begin
  if (hd == 0) return
  current = HD_CURRENT(hd)
  hlen = HD_HISTORY_LEN(hd)

  call item_release(HD_HISTORY(hd,current),HD_RELEASE(hd))
  HD_HISTORY(hd, current) = item
  if (current == hlen)	HD_CURRENT(hd) = 1
  else			HD_CURRENT(hd) = current + 1
end

#	Pops the current item in the history
procedure	hd_pop(hd, item)
pointer		hd			#I history
pointer		item			#O last item pointer

int		hlen
int		current

begin
  if (hd == 0) return
  current = HD_CURRENT(hd)
  hlen = HD_HISTORY_LEN(hd)

  item = HD_HISTORY(hd, current)
  HD_HISTORY(hd, current) = 0
  if (current == 1)	HD_CURRENT(hd) = hlen
  else			HD_CURRENT(hd) = current - 1
end
