# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  iiops.gx
# 	Author	:  ericco
# 	Date	:  04/30/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) iiops.gx -r1.6 04/30/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	03/30/94   akeith	fixed small generic type error.
# 	02/28/94   akeith	got rid of a useless pointer
# 	05/14/92   ericco	added documentation
# 	04/30/92   ericco	date and time created 92/04/30 01:45:56 by ericco
# H-
# 
#

include <ctype.h>

include	"ii.h"
include	<iierr.h>


#
#	Create a new interpreter
#
procedure	ii_open(ii, eval, release) #E
pointer		ii			#O interp
extern		eval			#I evaluating function
extern		release			#I releaseing function

int		locpr()

begin
  call ii_malloc(ii)
  II_INCLUDE(ii) = 0
  II_ERROR(ii) = NO
  II_ARGV_LEN(ii) = 0
  II_ARGV_PTR(ii) = 0
  II_COMMAND(ii) = 0
  II_HELP(ii) = 0
  II_EVAL(ii) = locpr(eval)
  II_RELEASE(ii) = locpr(release)
  II_PREPROC(ii) = 0
  II_POSTPROC(ii) = 0
  II_PRIVATE(ii) = 0
  II_PRIVATE_RELEASE(ii) = 0
  call hd_open(II_HISTORY(ii), II_RELEASE(ii))
  call hash_open(II_HASH(ii), II_RELEASE(ii))
end


#
#	Register pre processing function
#
procedure	ii_pre(ii, pre, state) #E
pointer		ii			#I interp
extern		pre			#I function
pointer		state			#I function's state

int		locpr()
begin
  II_PREPROC(ii) = locpr(pre)
  II_PRESTATE(ii) = state
end


#
#	Register the post processing function
#
procedure	ii_post(ii, post, state) #E
pointer		ii			#I interp
extern		post			#I function
pointer		state			#I function's state

int		locpr()
begin
  II_POSTPROC(ii) = locpr(post)
  II_POSTSTATE(ii) = state
end


#
#	Post a pointer to the private data.
#
procedure	ii_post_private(ii, private, destroy) #E
pointer		ii			#I interp
pointer		private			#I pointer
extern		destroy			#I releasing function

int		locpr()

begin
  II_PRIVATE(ii) = private
  II_PRIVATE_RELEASE(ii) = locpr(destroy)
end


#
#	Process a command
#
procedure	ii_evaluate (ii, cmd) #E
pointer		ii			#I interp
char		cmd[ARB]		#I command string

begin
  call zcall2(II_EVAL(ii), ii, cmd)
end


#
#	Retrive any errors after ii_evaluate()
#
procedure	ii_posteval(ii, stat, err, maxch) #E
pointer		ii			#I interp
int		stat			#O error status
char		err[ARB]		#O error string
int		maxch			#I max chars

pointer	sp, list

begin
	call ii_status(ii, stat)
	err[1] = EOS

	if(stat == 0) return

	call smark(sp)
	call salloc(list, 1, TY_LONG)

	call ii_list_history(ii, Meml[list], 1)
	call strcpy(Memc[Meml[list]], err, SZ_LINE)

	call sfree(sp)
	call ii_pop(ii)
end


#
#	Pop the history
#
procedure	ii_pop(ii) #I
pointer		ii			#I interp

begin
  call ii_pop_1(ii)
end


#
#	Returns an element in the history as a string.
#
procedure	ii_list_history(ii, list, len)
pointer		ii			#I interp
long		list[ARB]		#O list of history
int		len			#I length of list

int		i
pointer		item

begin

	do i = 1, len {
		call hd_get(II_HISTORY(ii), i, item)
		call item_to_string(item, list[i])
		call ii_item_release(ii, item)
	}
end


#
#	Release interp resources
#
procedure	ii_close(ii) #E
pointer		ii			#I interp

begin
  call hash_close(II_HASH(ii))
  call hd_close(II_HISTORY(ii))
  call help_release(II_HELP(ii))
  if (II_COMMAND(ii) != 0) call mfree(II_COMMAND(ii), TY_CHAR)
  if (II_ARGV_LEN(ii) != 0) call mfree(II_ARGV_PTR(ii), TY_LONG)
  if (II_PRIVATE(ii) != 0 && II_PRIVATE_RELEASE(ii) != 0) {
    call zcall1(II_PRIVATE_RELEASE(ii), II_PRIVATE(ii))
  }
  call ii_free(ii)
  ii = 0
end



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#
#  These are internal functions which are not guaranteed.	      #
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#
#
#	Need to add ';' delimited commands for many commands
#	on a single line.
#	Need to add "define" command to define a new command.
#	defining could include "compiling" -- converting symbols
#	into pointers.
#	To implement above, i need a "runtime stack" to replace
#	redo.  This "todo" stack would contain a series of
#	frames: an include file frame, an evaluate frame, and
#	a defined frame.  Parsing adds to the todo stack.
#	Evaluating pops an element for the todo stack and perhaps
#	adds new items onto the todo stack.


#	Purge interp history
procedure	ii_purge(ii) #I
pointer		ii			#I interpreter

begin
  call hd_purge(II_HISTORY(ii))
end

#	Release interp item
procedure	ii_item_release(ii, item) #I
pointer		ii			#I interpreter
pointer		item			#I value

begin
  call item_release(item, II_RELEASE(ii))
end

#	Add an include file to the input stream
procedure	ii_include(ii, file) #I
pointer		ii			#I interp
char		file[ARB]		#I include file

int		fd, open()
pointer		inc

begin
  fd = open(file, READ_ONLY, TEXT_FILE)
  if (fd == 0) return
  call malloc(inc, 2, TY_LONG)
  Meml[inc+1] = II_INCLUDE(ii)
  Meml[inc] = fd
  II_INCLUDE(ii) = inc
end

#	Pop n items off the history
procedure	ii_pop_n(ii, num) #I
pointer		ii			#I interp
int		num			#I number of items

int		i

begin
  do i = 1, num {
    call ii_pop_1(ii)
  }
end

#	Pop an item off of the history
procedure	ii_pop_1(ii) #I
pointer		ii			#I interp

pointer		item

begin
  call hd_pop(II_HISTORY(ii), item)
  call ii_item_release(ii, item)
end

#	Flag an error condition in the interp
procedure	ii_error(ptr, cond) #I
pointer		ptr			#I interp
int		cond			#I condition

begin
    II_ERROR(ptr) = cond
end

#	Return an error condition in the interp
procedure	ii_status(ptr, cond) #I
pointer		ptr			#I interp
int		cond			#I condition

begin
  cond = II_ERROR(ptr)
end

#	Release include file links
procedure	ii_clean_include(ii) #I
pointer		ii			#I interp

pointer		inc

begin
  while (II_INCLUDE(ii) != 0) {
    call close(Meml[II_INCLUDE(ii)])
    inc = Meml[II_INCLUDE(ii)+1]
    call mfree(II_INCLUDE(ii), TY_LONG)
  }
end

#	Command not found function
procedure	ii_not_found(ii) #I
pointer		ii			#I interp
begin
  call ii_err_push(ii, ": Command not found.")
end

#	No arguments given.
procedure	ii_no_args(ii) #I
pointer		ii			#I interp
begin
  call ii_err_push(ii, "No arguments given.")
end

#	Bad number of arguments.
procedure	ii_bad_nargs(ii) #I
pointer		ii			#I interp
begin
  call ii_err_push(ii, ": Bad number of arguments.")
end

#	Bad argument value.
procedure	ii_bad_val(ii) #I
pointer		ii			#I interp
begin
  call ii_err_push(ii, ": Bad argument value.")
end

#	Bad argument value.
procedure	ii_usr_err(ii) #I
pointer		ii			#I interp
int	code, errget()
pointer	str
begin
  call salloc(str, SZ_LINE, TY_CHAR)
  Memc[str] = ':'
  Memc[str+1] = ' '
  code = errget(Memc[str+2], SZ_LINE)
  call ii_err_push(ii, Memc[str])
end

define	II_ARGV_VALUE	Memc[II_ARGV($1,$2)]

#	Simply command not found function
procedure	ii_err_push(ii, str) #I
pointer		ii			#I interp
char		str[ARB]		#I error string

int	len, strlen()
pointer	sp, buf

begin
  if(II_ARGV_LEN(ii) > 0) len = strlen(II_ARGV_VALUE(ii, 1)) 
  else len = 0
  len = len + strlen(str)
  call smark(sp)
  call salloc(buf, len+1, TY_CHAR)
  Memc[buf] = EOS
  if(II_ARGV_LEN(ii) > 0) call strcat(II_ARGV_VALUE(ii,1), Memc[buf], len)
  call strcat(str, Memc[buf], len)
  call ii_push_string(ii, Memc[buf])
  call sfree(sp)
end

#	Redo a command
procedure	ii_redo(ii, redo) #I
pointer		ii			#I interp
pointer		redo			#I redo structure

pointer		inc
pointer		cmd
data		cmd / 0 /

int		getline()

begin
  if (cmd == 0) call malloc(cmd, SZ_LINE, TY_CHAR)

  if (II_REDO(ii) != 0)	{
    redo = II_REDO(ii)
  } else {
    redo = 0
    inc = II_INCLUDE(ii)
    while (inc != 0 && getline(Meml[inc], Memc[cmd]) == EOF) {
      inc= Meml[II_INCLUDE(ii)+1]
      call close(Meml[II_INCLUDE(ii)])
      call mfree(II_INCLUDE(ii), TY_LONG)
      II_INCLUDE(ii) = inc
    }
    if (inc != 0) redo = cmd
  }
end

define	IS_OCTAL	($1>='0'&&$1<='7')
define	IS_SPACE	($1==' '||$1=='\t'||$1=='\n')

#	Convert a string into internal format
procedure	ii_parse_string(str, ac, av, delim) #I
char		str[ARB]			#I command
int		ac				#O arguments
pointer		av				#O pointers to args
int		delim				#I delimiter to use

pointer		cp, op
int		dquote, chr
int		ip
int		len
int		strlen()

begin
  ac = 0
  av = 0
  cp = 0

  for (ip = 1; IS_SPACE(str[ip]); ip = ip + 1) ;
  if (str[ip] != EOS && str[ip] != '#') {
    len = strlen(str[ip])
    call malloc (cp, len, TY_CHAR)

    ac = 1
    call malloc (av, ac, TY_LONG)
    Meml[av+ac-1] = cp
    op = cp

    dquote = 0
    while (str[ip] != EOS) {
      for (; str[ip] != EOS; ip = ip + 1) {
	if (delim == ' ') {
	  if (IS_SPACE(str[ip]) && dquote == 0) break
	} else {
	  if (str[ip] == delim && dquote == 0) break
	}
	switch (str[ip]) {
	case '\\':
	  if (IS_OCTAL(str[ip+1])) {
	    ip = ip + 1; chr = str[ip] - '0'
	    if (IS_OCTAL(str[ip+1])) {
	      ip = ip + 1; chr = chr * 8 + str[ip] - '0'
	    }
	    if (IS_OCTAL(str[ip+1])) {
	      ip = ip + 1; chr = chr * 8 + str[ip] - '0'
	    }
	  } else if (str[ip+1] == EOS) {
	    break
	  } else {
	    chr = str[ip+1]
	  }
	case '"':
	  if (dquote == 1) {
	    dquote = 0
	  } else {
	    dquote = 1
	  }
	  next
	default:
	  chr = str[ip]
	}
	Memc[op] = chr; op = op + 1
      }
      Memc[op] = EOS; op = op + 1
      if (str[ip] == EOS) break;
      for (ip = ip + 1; IS_SPACE(str[ip]); ip = ip + 1) ;
      if (str[ip] == EOS) break;
      ac = ac + 1
      call realloc(av, ac, TY_LONG)
      Meml[av+ac-1] = op
    }
  }
end

#	Parse a command
procedure	ii_parse(ii, str) #I
pointer		ii			#I interp
char		str[ARB]		#I command line

int		ac
pointer		av
pointer		cp

begin
  II_REDO(ii) = 0
  call ii_parse_string(str, ac, av, ' ')
  if (ac > 0)	cp = Meml[av]
  else		cp = 0

  if (II_COMMAND(ii) != 0) call mfree(II_COMMAND(ii), TY_CHAR)
  if (II_ARGV_LEN(ii) != 0) call mfree(II_ARGV_PTR(ii), TY_LONG)

  II_COMMAND(ii) = cp
  II_ARGV_LEN(ii) = ac
  II_ARGV_PTR(ii) = av
end

#	Push a string into the input command stream
procedure	ii_evaluate_string (ii, cmd) #I
pointer		ii			#I interp
char		cmd[ARB]		#I new command

int		strlen()
pointer		buf

begin
  call salloc(buf, strlen(cmd), TY_CHAR)
  call strcpy(cmd, Memc[buf], ARB)
  II_REDO(ii) = buf
end
#	Execute any pre processing function
procedure	ii_pre_process(ii) #I
pointer		ii		#I interp

begin
  if (II_PREPROC(ii) == 0) return
  call zcall3(II_PREPROC(ii), II_POSTSTATE(ii),
	      II_ARGV_LEN(ii), II_ARGV(ii,1))
end
#	Execute any post processing function
procedure	ii_post_process(ii, resc) #I
pointer		ii		#I interp
int		resc		#I

begin
  if (II_POSTPROC(ii) == 0) return
  call zcall4(II_POSTPROC(ii), II_POSTSTATE(ii),
	      II_ARGV_LEN(ii), II_ARGV(ii,1), resc)
end

#	Get the private pointer
procedure	ii_private(ii, private) #I
pointer		ii			#I interp
pointer		private			#O pointer

begin
  private = II_PRIVATE(ii)
end

#	Initialize the help info
procedure	ii_help_open (ii, name) #I
pointer		ii			#I interp
char		name[ARB]		#I help file

begin
  if (II_HELP(ii) != 0) call help_release(II_HELP(ii))
  call help_open(II_HELP(ii), name)
end

#	Simply help function
procedure	ii_help(ii, topic, ptr) #I
pointer		ii			#I interp
char		topic[ARB]		#I help topic
pointer		ptr			#O help message

begin
  call help_info(II_HELP(ii), topic, ptr)
end

#	Push a string pointer into the history
procedure	ii_push_t (ii, dat) #I
pointer		ii		#I interpreter state
pointer		dat		#I data to push onto history

begin
  call ii_push_string (ii, Memc[dat])
end
#	Push a string into the history
procedure	ii_push_string (ii, dat) #I
pointer		ii		#I interpreter state
char		dat[ARB]	#I data to push onto history

pointer		item

begin
  call item_new_string(item, dat)
  call hd_push(II_HISTORY(ii), item)
end

#	Push a pointer into the history
procedure	ii_push_p (ii, dat) #I
pointer		ii		#I interpreter state
pointer		dat		#I data to push onto history

pointer		item

begin
  call item_new_p(item, dat)
  call hd_push(II_HISTORY(ii), item)
end

#	Push a structure into the history
procedure	ii_push_z(ii, dat, ds_type) #I
pointer		ii		#I interpreter state
pointer		dat		#I data to push onto history
int		ds_type		#I ds type

begin
  call ii_push_structure(ii, dat, ds_type)
end

#	Push a typed sturcture into the history
procedure	ii_push_structure(ii, dat, ds_type) #I
pointer		ii		#I interpreter state
pointer		dat		#I data to push onto history
int		ds_type		#I ds type

pointer		item

begin
  call item_new_structure(item, ds_type, dat)
  call hd_push(II_HISTORY(ii), item)
end


#	Push a data values into the history
procedure	ii_push_d (ii, dat) #I
pointer		ii		#I interpreter state
double		dat		#I data to push onto history

pointer		item

begin
  call item_new_d(item, dat)
  call hd_push(II_HISTORY(ii), item)
end

#	Push a data values into the history
procedure	ii_push_r (ii, dat) #I
pointer		ii		#I interpreter state
real		dat		#I data to push onto history

pointer		item

begin
  call item_new_r(item, dat)
  call hd_push(II_HISTORY(ii), item)
end

#	Push a data values into the history
procedure	ii_push_s (ii, dat) #I
pointer		ii		#I interpreter state
short		dat		#I data to push onto history

pointer		item

begin
  call item_new_s(item, dat)
  call hd_push(II_HISTORY(ii), item)
end

#	Push a data values into the history
procedure	ii_push_l (ii, dat) #I
pointer		ii		#I interpreter state
long		dat		#I data to push onto history

pointer		item

begin
  call item_new_l(item, dat)
  call hd_push(II_HISTORY(ii), item)
end

#	Push a data values into the history
procedure	ii_push_i (ii, dat) #I
pointer		ii		#I interpreter state
int		dat		#I data to push onto history

pointer		item

begin
  call item_new_i(item, dat)
  call hd_push(II_HISTORY(ii), item)
end

#	Push a data values into the history
procedure	ii_push_c (ii, dat) #I
pointer		ii		#I interpreter state
char		dat		#I data to push onto history

pointer		item

begin
  call item_new_c(item, dat)
  call hd_push(II_HISTORY(ii), item)
end


#	Get data from the history
procedure	ii_history(ii, history, item) #I
pointer		ii			#I interp
int		history			#I history index
pointer		item			#O value

begin
  call hd_get(II_HISTORY(ii), history, item)
end


#	Get a command line argument
procedure	ii_argument(ii, argc, item) #I
pointer		ii		#I interp
int		argc		#I argumnet index
pointer		item		#O value

int		ip
int		history
pointer		arg

begin
  arg = II_ARGV(ii,argc)
  if (Memc[arg] == '$') {
    if (IS_DIGIT(Memc[arg+1])) {
      ip = 2
      call ctol(Memc[arg], ip, history)
      call ii_history(ii, int(history), item)
    } else if (IS_ALPHA(Memc[arg+1])) {
      call hash_get(II_HASH(ii), Memc[arg+1], item)
    } else {
      call item_new_string(item, Memc[arg])
    }
  } else {
      call item_new_string(item, Memc[arg])
  }
end

#	Set a variable
procedure	ii_set(ii, str, value) #I
pointer		ii		#I interp
char		str[ARB]	#I label
char		value[ARB]	#I value

pointer		item

begin
  # the type of the item is needed, not just its string
  # representation
  call ii_argument(ii, 3, item)
  call hash_put(II_HASH(ii), str, item)
  call ii_item_release(ii, item)
end


#	Get a data string pointer
procedure	ii_get_t(ii, argc, dat) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
pointer		dat		#O converted data

begin
  call ii_get_string(ii, argc, dat)
  if (dat == 0)
    call error(1, "arg. is not a string.")
end

#	Get string data
procedure	ii_get_string(ii, argc, dat) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
pointer		dat		#O converted data

pointer		item
bool		item_is_string()

begin
  call ii_argument(ii, argc, item)
  if(item_is_string(item)) call item_to_string(item, dat)
  call ii_item_release(ii, item)
end

#	Get a structure
procedure	ii_get_z(ii, argc, dat, ds_type) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
pointer		dat		#O converted data
int		ds_type		#I ds type

begin
  call ii_get_structure(ii, argc, dat, ds_type)
  if (dat == NULL)
    call error(1, "arg. is not a data structure.")
end

#	Push a typed sturcture into the history
procedure	ii_get_structure(ii, argc, dat, ds_type) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
pointer		dat		#O converted data
int		ds_type		#I ds type

pointer		item

begin
  call ii_argument(ii, argc, item)
  call item_to_structure(item, ds_type, dat)
  call ii_item_release(ii, item)
end

#	Get pointer data
procedure	ii_get_p(ii, argc, dat) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
pointer		dat		#O converted data

pointer		item		#

begin

  call ii_argument(ii, argc, item)
  call item_to_p(item, dat)
  call ii_item_release(ii, item)
  if(dat == NULL)
    call error(1, "arg. is not a pointer")
end


#	Get a primitive data types

procedure	ii_get_d(ii, argc, dat) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
double		dat		#O converted data

pointer		item

begin
  call ii_argument(ii, argc, item)
  call item_to_d(item, dat)
  call ii_item_release(ii, item)
  if(IS_INDEFD(dat))
    call error(1, "Bad data type for argument.")
end

procedure	ii_get_r(ii, argc, dat) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
real		dat		#O converted data

pointer		item

begin
  call ii_argument(ii, argc, item)
  call item_to_r(item, dat)
  call ii_item_release(ii, item)
  if(IS_INDEFR(dat))
    call error(1, "Bad data type for argument.")
end

procedure	ii_get_s(ii, argc, dat) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
short		dat		#O converted data

pointer		item

begin
  call ii_argument(ii, argc, item)
  call item_to_s(item, dat)
  call ii_item_release(ii, item)
  if(IS_INDEFS(dat))
    call error(1, "Bad data type for argument.")
end

procedure	ii_get_l(ii, argc, dat) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
long		dat		#O converted data

pointer		item

begin
  call ii_argument(ii, argc, item)
  call item_to_l(item, dat)
  call ii_item_release(ii, item)
  if(IS_INDEFL(dat))
    call error(1, "Bad data type for argument.")
end

procedure	ii_get_i(ii, argc, dat) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
int		dat		#O converted data

pointer		item

begin
  call ii_argument(ii, argc, item)
  call item_to_i(item, dat)
  call ii_item_release(ii, item)
  if(IS_INDEFI(dat))
    call error(1, "Bad data type for argument.")
end

procedure	ii_get_c(ii, argc, dat) #I
pointer		ii		#I interpreter state
int		argc		#I argument index
char		dat		#O converted data

pointer		item

begin
  call ii_argument(ii, argc, item)
  call item_to_c(item, dat)
  call ii_item_release(ii, item)
  if(IS_INDEF(dat))
    call error(1, "Bad data type for argument.")
end



#	Get the number of arguments
procedure	ii_args(ii,args) #I
pointer		ii		#I interp
int		args		#I number of arguments

begin
  args = II_ARGV_LEN(ii)
end

#	Explicitly call a macro
procedure	ii_call (ii,label) #I
pointer		ii		#I interp
char		label[ARB]	#I macro name

int		ip
pointer		item
int		history

begin
  if (IS_DIGIT(label[1])) {
    ip = 2
    call ctol(label, ip, history)
    call ii_history(ii, int(history), item)
  } else if (IS_ALPHA(label[1])) {
    call hash_get(II_HASH(ii), label, item)
  } else {
    call item_new_string(item, label)
  }
  call item_to_string(item, II_REDO(ii))
  call ii_item_release(ii, item)
end

#	Lookup a convert value
procedure	ii_lookup(ii,dict, key) #I
pointer		ii		#I interp
char		dict[ARB]	#I dictionary to use
int		key		#O entry in dictionary

pointer		dat
char		junk[1]
int		strdic()

begin
  key = 0
  if (II_ARGV_LEN(ii) < 1) return
  call ii_get_string(ii, 1, dat)
  if (dat == 0) return

  key = strdic(Memc[dat], junk, 1, dict)
end
