# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  itemops.gx
# 	Author	:  ericco
# 	Date	:  04/30/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) itemops.gx -r1.7 04/30/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	06/29/94   akeith	added bool functionality
# 	02/28/94   akeith	updated documentation
# 	07/16/93   akeith	fixed the appearent recursion in ds_release
# 	07/01/92   ericco	added argument error
# 	05/14/92   ericco	added documentation
# 	04/30/92   ericco	date and time created 92/04/30 01:46:02 by ericco
# H-
# 
#

include "item.h"

define TY_NONE		100
define TY_DS		101
define TY_STRING	102

#	Check if item is empty
procedure	item_empty_p (item, pred) #E
pointer		item		#I item
int		pred		#O flag

begin
  if (ITEM_TYPE(item) == TY_NONE) {
    pred = YES
  } else {
    pred = NO
  }
end


#
#  Routines to create a new item.
#


#	Make a new primitive type item
procedure	item_new_b (item, dat) #E
pointer		item		#O new item
bool		dat		#I value

pointer		buf

begin
  call item_malloc(item)
  call malloc(buf, 1, TY_BOOL)
  Memb[buf] = dat
  ITEM_TYPE(item) = TY_BOOL
  ITEM_VALUE(item) = buf
end

#	Make a new primitive type item
procedure	item_new_d (item, dat) #E
pointer		item		#O new item
double		dat		#I value

pointer		buf

begin
  call item_malloc(item)
  call malloc(buf, 1, TY_DOUBLE)
  Memd[buf] = dat
  ITEM_TYPE(item) = TY_DOUBLE
  ITEM_VALUE(item) = buf
end

#	Make a new primitive type item
procedure	item_new_r (item, dat) #E
pointer		item		#O new item
real		dat		#I value

pointer		buf

begin
  call item_malloc(item)
  call malloc(buf, 1, TY_REAL)
  Memr[buf] = dat
  ITEM_TYPE(item) = TY_REAL
  ITEM_VALUE(item) = buf
end

#	Make a new primitive type item
procedure	item_new_s (item, dat) #E
pointer		item		#O new item
short		dat		#I value

pointer		buf

begin
  call item_malloc(item)
  call malloc(buf, 1, TY_SHORT)
  Mems[buf] = dat
  ITEM_TYPE(item) = TY_SHORT
  ITEM_VALUE(item) = buf
end

#	Make a new primitive type item
procedure	item_new_l (item, dat) #E
pointer		item		#O new item
long		dat		#I value

pointer		buf

begin
  call item_malloc(item)
  call malloc(buf, 1, TY_LONG)
  Meml[buf] = dat
  ITEM_TYPE(item) = TY_LONG
  ITEM_VALUE(item) = buf
end

#	Make a new primitive type item
procedure	item_new_i (item, dat) #E
pointer		item		#O new item
int		dat		#I value

pointer		buf

begin
  call item_malloc(item)
  call malloc(buf, 1, TY_INT)
  Memi[buf] = dat
  ITEM_TYPE(item) = TY_INT
  ITEM_VALUE(item) = buf
end

#	Make a new primitive type item
procedure	item_new_c (item, dat) #E
pointer		item		#O new item
char		dat		#I value

pointer		buf

begin
  call item_malloc(item)
  call malloc(buf, 1, TY_CHAR)
  Memc[buf] = dat
  ITEM_TYPE(item) = TY_CHAR
  ITEM_VALUE(item) = buf
end


#	Make an empty item
procedure	item_new_empty(item) #E
pointer		item		#O new item

begin
  call item_malloc(item)
  ITEM_TYPE(item) = TY_NONE
end

#	Make a pointer item
procedure	item_new_p (item, dat) #E
pointer		item		#O new item
pointer		dat		#I value

pointer		buf

begin
  call item_malloc(item)
  call malloc(buf, 1, TY_LONG)
  Meml[buf] = dat
  ITEM_TYPE(item) = TY_POINTER
  ITEM_VALUE(item) = buf
end

#	Make a string item
procedure	item_new_string(item, str) #E
pointer		item			#O new item
char		str[ARB]		#I value

int		len, strlen()
pointer		buf

begin
  len = strlen(str)
  call malloc(buf, len, TY_CHAR)
  call strcpy(str, Memc[buf], len)
  call item_malloc(item)
  ITEM_TYPE(item) = TY_STRING
  ITEM_VALUE(item) = buf
end

#	Make a new structure item
procedure	item_new_structure(item, ds_type, ds_ptr) #E
pointer		item			#O new item
int		ds_type			#I structure type
pointer		ds_ptr			#I structure pointer

begin
  call item_malloc(item)
  ITEM_TYPE(item) = TY_DS
  call ds_new(ITEM_VALUE(item), ds_type, ds_ptr)
end

#	Make a new item from an old item
procedure	item_deep_copy(old, new) #E
pointer		old		#I old item
pointer		new		#O new item

pointer		dptr

begin
  new = 0
  if (old == 0) return

  dptr = ITEM_VALUE(old)

  switch(ITEM_TYPE(old)) {
  case TY_DS:
    call ds_indirect_copy(dptr)
    call item_malloc(new)
    ITEM_TYPE(new) = TY_DS
    ITEM_VALUE(new) = dptr
  case TY_STRING:
    call item_new_string(new, Memc[dptr])

  case TY_BOOL:
    call item_new_b(new, Memb[dptr])

  case TY_DOUBLE:
    call item_new_d(new, Memd[dptr])

  case TY_REAL:
    call item_new_r(new, Memr[dptr])

  case TY_SHORT:
    call item_new_s(new, Mems[dptr])

  case TY_LONG:
    call item_new_l(new, Meml[dptr])

  case TY_INT:
    call item_new_i(new, Memi[dptr])

  case TY_CHAR:
    call item_new_c(new, Memc[dptr])

  }
end


#
#  Routine to release an item.
#

#	Release item's resouces
procedure	item_release(item, ds_rel) #E
pointer		item			#I item
pointer		ds_rel			#I release function

begin
  if (item == 0) return
  switch(ITEM_TYPE(item)) {
  case TY_NONE:
    # don't release any subpart
  case TY_STRING:
    call mfree(ITEM_VALUE(item), TY_CHAR)
  case TY_POINTER:
    call mfree(ITEM_VALUE(item), TY_LONG)
  case TY_DS:
    call ds_release(ITEM_VALUE(item), ds_rel)
  default:
    call mfree(ITEM_VALUE(item), ITEM_TYPE(item))
  }
  call item_free(item)
  item = 0
end


#
#  Routines to check an item's type.
#

procedure	item_gtype(item, type) #E
pointer		item		#I value
int		type		#I type
begin
	type = 0
	if(item == NULL) return
	if(ITEM_TYPE(item) == TY_STRING) type = -TY_CHAR
	else type = ITEM_TYPE(item)
end



bool	procedure	item_is_b(item) #E
pointer		item		#I value

begin
  if (item == 0) return(FALSE)
  if(ITEM_TYPE(item) == TY_BOOL) return(TRUE)
  else return(FALSE)
end

bool	procedure	item_is_d(item) #E
pointer		item		#I value

begin
  if (item == 0) return(FALSE)
  if(ITEM_TYPE(item) == TY_DOUBLE) return(TRUE)
  else return(FALSE)
end

bool	procedure	item_is_r(item) #E
pointer		item		#I value

begin
  if (item == 0) return(FALSE)
  if(ITEM_TYPE(item) == TY_REAL) return(TRUE)
  else return(FALSE)
end

bool	procedure	item_is_s(item) #E
pointer		item		#I value

begin
  if (item == 0) return(FALSE)
  if(ITEM_TYPE(item) == TY_SHORT) return(TRUE)
  else return(FALSE)
end

bool	procedure	item_is_l(item) #E
pointer		item		#I value

begin
  if (item == 0) return(FALSE)
  if(ITEM_TYPE(item) == TY_LONG) return(TRUE)
  else return(FALSE)
end

bool	procedure	item_is_i(item) #E
pointer		item		#I value

begin
  if (item == 0) return(FALSE)
  if(ITEM_TYPE(item) == TY_INT) return(TRUE)
  else return(FALSE)
end

bool	procedure	item_is_c(item) #E
pointer		item		#I value

begin
  if (item == 0) return(FALSE)
  if(ITEM_TYPE(item) == TY_CHAR) return(TRUE)
  else return(FALSE)
end


#	Check that an item is a string
bool	procedure	item_is_string(item) #E
pointer		item		#I value

begin
  if (item == 0) return(FALSE)
  if(ITEM_TYPE(item) == TY_STRING) return(TRUE)
  else return(FALSE)
end

#	Check that an item is a pointer
bool	procedure	item_is_p(item) #E
pointer		item		#I value

begin
  if (item == 0) return(FALSE)
  if(ITEM_TYPE(item) == TY_POINTER) return(TRUE)
  else return(FALSE)
end

#	Check that an item is a ds
bool	procedure	item_is_structure(item) #E
pointer		item		#I value

begin
  if (item == 0) return(FALSE)
  if(ITEM_TYPE(item) == TY_DS) return(TRUE)
  else return(FALSE)
end


#
#  Routines to convert an item's type.
#

#	Convert item into string
procedure	item_to_string(item, ptr) #E
pointer		item		#I value
pointer		ptr		#O string pointer

int		len, strlen()
pointer		p

begin
  ptr = 0
  if (item == 0) return

  switch (ITEM_TYPE(item)) {

  case TY_BOOL:
    call salloc(ptr, 20, TY_CHAR)
    call sprintf(Memc[ptr], 20, "%s")
    call pargb(Memb[ITEM_VALUE(item)])

  case TY_DOUBLE:
    call salloc(ptr, 20, TY_CHAR)
    call sprintf(Memc[ptr], 20, "%s")
    call pargd(Memd[ITEM_VALUE(item)])

  case TY_REAL:
    call salloc(ptr, 20, TY_CHAR)
    call sprintf(Memc[ptr], 20, "%s")
    call pargr(Memr[ITEM_VALUE(item)])

  case TY_SHORT:
    call salloc(ptr, 20, TY_CHAR)
    call sprintf(Memc[ptr], 20, "%s")
    call pargs(Mems[ITEM_VALUE(item)])

  case TY_LONG:
    call salloc(ptr, 20, TY_CHAR)
    call sprintf(Memc[ptr], 20, "%s")
    call pargl(Meml[ITEM_VALUE(item)])

  case TY_INT:
    call salloc(ptr, 20, TY_CHAR)
    call sprintf(Memc[ptr], 20, "%s")
    call pargi(Memi[ITEM_VALUE(item)])

  case TY_CHAR:
    call salloc(ptr, 20, TY_CHAR)
    call sprintf(Memc[ptr], 20, "%s")
    call pargc(Memc[ITEM_VALUE(item)])

  case TY_DS:
    call salloc(ptr, 20, TY_CHAR)
    call sprintf(Memc[ptr], 20, "%s")
    call ds_value(ITEM_VALUE(item), p)
    call pargl(p)
  case TY_POINTER:
    call salloc(ptr, 20, TY_CHAR)
    call sprintf(Memc[ptr], 20, "%s")
    call pargl(Meml[ITEM_VALUE(item)])
  case TY_STRING:
    len = strlen(Memc[ITEM_VALUE(item)])
    call salloc(ptr, len, TY_CHAR)
    call strcpy(Memc[ITEM_VALUE(item)], Memc[ptr], ARB)
  }
end

#	Convert item to double
procedure	item_to_d(item, dbl) #E
pointer		item		#I item
double		dbl		#O value

begin
  dbl = INDEFD
  if (item == 0) return
  switch (ITEM_TYPE(item)) {

  case TY_DOUBLE: dbl = Memd[ITEM_VALUE(item)]

  case TY_REAL: dbl = Memr[ITEM_VALUE(item)]

  case TY_LONG: dbl = Meml[ITEM_VALUE(item)]

  case TY_INT: dbl = Memi[ITEM_VALUE(item)]

  case TY_SHORT: dbl = Mems[ITEM_VALUE(item)]

  case TY_STRING:
    call string_to_double(Memc[ITEM_VALUE(item)], dbl)
  }

end

#	Convert item to real
procedure	item_to_r(item, flt) #E
pointer		item		#I item
real		flt		#O value

double	dbl

begin
  flt = INDEFR
  if (item == 0) return
  switch (ITEM_TYPE(item)) {

  case TY_REAL: flt = Memr[ITEM_VALUE(item)]

  case TY_LONG: flt = Meml[ITEM_VALUE(item)]

  case TY_INT: flt = Memi[ITEM_VALUE(item)]

  case TY_SHORT: flt = Mems[ITEM_VALUE(item)]

  case TY_STRING:
    call string_to_double(Memc[ITEM_VALUE(item)], dbl)
    if(!IS_INDEFD(dbl)) flt = dbl
  }
end

#	Convert item to long
procedure	item_to_l(item, lng) #E
pointer		item			#I item
long		lng			#O value

begin
  lng = INDEFL
  if (item == 0) return
  switch (ITEM_TYPE(item)) {

  case TY_LONG: lng = Meml[ITEM_VALUE(item)]

  case TY_INT: lng = Memi[ITEM_VALUE(item)]

  case TY_SHORT: lng = Mems[ITEM_VALUE(item)]

  case TY_STRING:
    call string_to_long(Memc[ITEM_VALUE(item)], lng)
  }
end

#	Convert item to int
procedure	item_to_i(item, i) #E
pointer		item			#I item
int		i			#O value

long	lng

begin
  i = INDEFI
  if (item == 0) return
  switch (ITEM_TYPE(item)) {

  case TY_LONG: i = Meml[ITEM_VALUE(item)]

  case TY_INT: i = Memi[ITEM_VALUE(item)]

  case TY_SHORT: i = Mems[ITEM_VALUE(item)]

  case TY_STRING:
    call string_to_long(Memc[ITEM_VALUE(item)], lng)
    if(!IS_INDEFL(lng)) i = lng
  }
end

#	Convert item to short
procedure	item_to_s(item, s) #E
pointer		item			#I item
short		s			#O value

long	lng

begin
  s = INDEFS
  if (item == 0) return
  if(ITEM_TYPE(item) == TY_SHORT)
  	s = Mems[ITEM_VALUE(item)]

  if(ITEM_TYPE(item) == TY_STRING) {
    call string_to_long(Memc[ITEM_VALUE(item)], lng)
    if(!IS_INDEFL(lng)) s = lng
  }
end

#	Convert item to char
procedure	item_to_c(item, c) #E
pointer		item			#I item
char		c			#O value

int	strlen()
begin
  c = INDEFS
  if (item == 0) return
  if(ITEM_TYPE(item) == TY_CHAR)
	c = Memc[ITEM_VALUE(item)]
  else if(ITEM_TYPE(item) == TY_STRING) {
    if(strlen(Memc[ITEM_VALUE(item)]) == 1)
	c = Memc[ITEM_VALUE(item)]
  }
end

#	Convert item to pointer
procedure	item_to_p(item, p) #E
pointer		item			#I item
pointer		p			#O value

begin
  p = NULL
  if (item == 0) return
  if(ITEM_TYPE(item) == TY_POINTER)
	p = Meml[ITEM_VALUE(item)]
end

#	Convert item to bool
procedure	item_to_b(item, b) #E
pointer		item			#I item
bool		b			#O value

begin
  b = FALSE
  if (item == 0) return
  if(ITEM_TYPE(item) == TY_BOOL)
	b = Memb[ITEM_VALUE(item)]
end

#	Convert item to data structure
procedure	item_to_structure(item, ds_type, ds_ptr) #E
pointer		item			#I item
int		ds_type			#I ds type
pointer		ds_ptr			#O value

int	type
begin
  ds_ptr = NULL
  if (item == 0) return
  if(ITEM_TYPE(item) == TY_DS) {
	call ds_get(ITEM_VALUE(item), type, ds_ptr)
	if(type != ds_type) ds_ptr = NULL
  }
end



#	Get item's data pointer
procedure	item_data_pointer(item, ptr) #I
pointer		item		#I item
pointer		ptr		#O data pointer

begin
  if (item == 0)	ptr = 0
  else			ptr = ITEM_VALUE(item)
end


define	IS_WHITE	($1==' '||$1=='\t')

#	Convert a string to double
procedure	string_to_double(str, dbl) #I
char		str[ARB]	#I
double		dbl		#O value

int	ip, nused, ctod()
int	len, strlen()

begin
    len = strlen(str)
    while(IS_WHITE(str[len]) && len>0)
      len = len - 1
    ip = 1
    nused = ctod(str, ip, dbl)
    if(nused < len) dbl = INDEFD
end

#	Convert a string to long
procedure	string_to_long(str, val) #I
char		str[ARB]	#I
long		val		#O value

int	ip, nused, ctol()
int	len, strlen()

begin
    len = strlen(str)
    while(IS_WHITE(str[len]) && len>0)
      len = len - 1
    ip = 1
    nused = ctol(str, ip, val)
    if(nused < len) val = INDEFL
end

