# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  iliops.x
# 	Author	:  akeith
# 	Date	:  02/28/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) iliops.x -r1.3 02/28/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	06/28/94   akeith	fixed a small memory leak
# 	02/28/94   akeith	date and time created 94/02/28 16:35:14 by akeith
# H-
# 
#


#	Version: @(#)iliops.x	1.3 6/28/94 UCB CEA EGO

include	"ili.h"
include	<imhdr.h>

#
# ili_open() return a pointer to the image line interpolation
#	structure and opens the named image, image.
#

pointer	procedure	ili_open(image) #E

char	image[ARB]		#I image name

pointer	ili			# image line interpolation
pointer	immap()
pointer mw_openim()

int	errcode()

begin

	call ili_malloc(ili)
	call ili_init(ili)

	iferr(ILI_IM(ili) = immap(image, READ_ONLY, NULL)) {
		call ili_close(ili)
		ili = NULL
		call erract(errcode())
	}

	ILI_MW(ili) = mw_openim(ILI_IM(ili))

	return(ili)
end


#
# ili_close() closes the named image and frees the ili structure.
#

procedure	ili_close(ili) #E

pointer	ili			#I image line interpolation

begin

	if(ILI_MW(ili) != NULL) call mw_close(ILI_MW(ili))

	if(ILI_IM(ili) != NULL) call imunmap(ILI_IM(ili))

	if(ILI_DF1(ili) != NULL) call df_close(ILI_DF1(ili))

	if(ILI_DF2(ili) != NULL) call df_close(ILI_DF2(ili))

	call ili_free(ili)
end


#
# ili_retrieve() retrieves a discrete function at the reference
#	coordinate.
#

procedure	ili_retrieve(ili, rc, df) #E

pointer	ili			#I image line interpolation
double	rc			#I reference coordinate
pointer df			#I discrete function

int	ndim
int	low, high
real	k
real	logical_val
pointer	im
pointer sp, w, l

begin

	im = ILI_IM(ili)
	ndim = IM_NDIM(im)

	call smark(sp)
	call salloc(w, ndim, TY_REAL)
	call salloc(l, ndim, TY_REAL)
	call aclrr(Memr[w], ndim)
	call aclrr(Memr[l], ndim)
	
	Memr[w+1] = rc

	call wl_ctranr(ILI_MW(ili), "world", Memr[w], Memr[l], ndim)

	logical_val = Memr[l+1]
	low = int(logical_val)
	high = low + 1
	k = logical_val - low

	if(low < 1) {
		low = 1
		high = 2
		k = 0.0

	} else if(high > IM_LEN(im, 2)) {
		low = IM_LEN(im, 2) - 1
		high = IM_LEN(im, 2)
		k = 1.0
	}

	call ili_setup(ili, low, high)
	call df_interp(ILI_DF1(ili), double(k), ILI_DF2(ili), df)

	call sfree(sp)
end


#
# ili_seti()
#

procedure	ili_seti(ili, param, value) #E

pointer	ili			#I image line interpolation
int	param			#I ili set parameter
int	value			#I ili set value

begin

	switch(param) {

	  case ILI_BAND_PARM:
		ILI_BAND(ili) = value

	  default:
		call error(1, "ili_seti(): unknown parameter.")
	}

end


#
# ili_stati()
#

int	procedure	ili_stati(ili, param) #E

pointer	ili			#I image line interpolation
int	param			#I ili set parameter

int	value			# parameter value

begin

	switch(param) {

	  case ILI_BAND_PARM:
		value = ILI_BAND(ili)

	  default:
		call error(1, "ili_seti(): unknown parameter.")
	}

	return(value)

end


#
#
#  ******     INTERNAL FUNCTIONS    ******     
#
#


#
# ili_init()
#

procedure	ili_init(ili) #I

pointer	ili			#I image line interpolation

begin

	ILI_IM(ili) = NULL
	ILI_MW(ili) = NULL
	ILI_BAND(ili) = 1
	ILI_ROW1(ili) = 0
	ILI_ROW2(ili) = 0
	ILI_DF1(ili) = NULL
	ILI_DF2(ili) = NULL
end


#
# ili_setup()
#

procedure	ili_setup(ili, low, high)

pointer	ili			#I image line interpolation
int	low			#I low row number
int	high			#I high row number

pointer	pf_rimg()

begin

	if(low == ILI_ROW1(ili)) return

	if(low == ILI_ROW2(ili)) {
		ILI_ROW1(ili) = low
		ILI_ROW2(ili) = high
		call df_close(ILI_DF1(ili))
		ILI_DF1(ili) = ILI_DF2(ili)
		ILI_DF2(ili) = pf_rimg(ILI_IM(ili), ILI_ROW2(ili), ILI_BAND(ili))
		call ili_sedge(ILI_DF2(ili), NULL)
		return
	}

	if(high == ILI_ROW1(ili)) {
		ILI_ROW1(ili) = low
		ILI_ROW2(ili) = high
		call df_close(ILI_DF2(ili))
		ILI_DF2(ili) = ILI_DF1(ili)
		ILI_DF1(ili) = pf_rimg(ILI_IM(ili), ILI_ROW1(ili), ILI_BAND(ili))
		call ili_sedge(ILI_DF1(ili), NULL)
		return

	} else {

		ILI_ROW1(ili) = low
		ILI_ROW2(ili) = high
		call df_close(ILI_DF1(ili))
		call df_close(ILI_DF2(ili))
		ILI_DF1(ili) = pf_rimg(ILI_IM(ili), ILI_ROW1(ili), ILI_BAND(ili))
		ILI_DF2(ili) = pf_rimg(ILI_IM(ili), ILI_ROW2(ili), ILI_BAND(ili))
		call ili_sedge(ILI_DF1(ili), ILI_DF2(ili))
		return
	}

end


#
# ili_sedge()
#

procedure	ili_sedge(df1, df2)

pointer	df1			#I
pointer	df2			#I

int	stat, df_dgetd()	#
int	buf_size		#
pointer	sp, x, y		#
double	redge

begin

	buf_size = 100

	call smark(sp)
	call salloc(x, buf_size, TY_DOUBLE)
	call salloc(y, buf_size, TY_DOUBLE)

	call df_rewind(df1)
	stat = df_dgetd(df1, Memd[x], redge, 1)

	stat = df_dgetd(df1, Memd[x], Memd[y], buf_size)

	while(stat == buf_size)
		stat = df_dgetd(df1, Memd[x], Memd[y], buf_size)

	if(stat == 0) stat = buf_size

	call df_pedge(df1, redge, Memd[y+stat-1])
	call df_rewind(df1)


	if(df2 != NULL) {
		call df_rewind(df2)
		stat = df_dgetd(df2, Memd[x], redge, 1)

		stat = df_dgetd(df2, Memd[x], Memd[y], buf_size)

		while(stat == buf_size)
			stat = df_dgetd(df2, Memd[x], Memd[y], buf_size)

		if(stat == 0) stat = buf_size

		call df_pedge(df2, redge, Memd[y+stat-1])
		call df_rewind(df2)
	}

	call sfree(sp)
end


