# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  nr_poidev.x
# 	Author	:  tnk
# 	Date	:  08/01/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) nr_poidev.x -r1.2 08/01/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	08/10/95   tnk	added copyright notice
# 	08/01/94   tnk	date and time created 94/08/01 17:48:39 by tnk
# H-
# 
#

# Copyright(c) 1986 Association of Universities for Research in Astronomy Inc.
# Copyright (C) 1986 Numerical Recipes Software

include	<math.h>

# @(#)nr_poidev.x	1.2 8/10/95 UCB CEA EGO

# GAMMLN -- Return natural log of gamma function.
# POIDEV -- Returns Poisson deviates for a given mean.

# GAMMLN -- Return natural log of gamma function.
# Argument must greater than 0.  Full accuracy is obtained for values
# greater than 1.  For 0<xx<1, the reflection formula can be used first.
#
# Based on Numerical Recipes by Press, Flannery, Teukolsky, and Vetterling.
# Used by permission of the authors.
# Copyright(c) 1986 Numerical Recipes Software.

real procedure nr_gammln2 (xx)

real	xx		# Value to be evaluated

int	j
double	cof[6], stp, x, tmp, ser
data	cof, stp / 76.18009173D0, -86.50532033D0, 24.01409822D0,
		-1.231739516D0,.120858003D-2,-.536382D-5,2.50662827465D0/

begin
	x = xx - 1.0D0
	tmp = x + 5.5D0
	tmp = (x + 0.5D0) * log (tmp) - tmp
	ser = 1.0D0
	do j = 1, 6 {
	    x = x + 1.0D0
	    ser = ser + cof[j] / x
	}
	return (tmp + log (stp * ser))
end


# POIDEV -- Returns Poisson deviates for a given mean.
# The real value returned is an integer.
#
# Based on Numerical Recipes by Press, Flannery, Teukolsky, and Vetterling.
# Used by permission of the authors.
# Copyright(c) 1986 Numerical Recipes Software.

real procedure nr_poidev (xm, seed)

real	xm		# Poisson mean
long	seed		# Random number seed

real	oldm, g, em, t, y, sq, alxm, nr_gammln2(), urand()
data	oldm /-1./

begin
	if (xm < 12) {
	    if (xm != oldm) {
		oldm = xm
		g = exp (-xm)
	    }
	    em = 0
	    for (t = urand (seed); t > g; t = t * urand (seed))
		em = em + 1
	} else {
	    if (xm != oldm) {
		oldm = xm
		sq = sqrt (2. * xm)
		alxm = log (xm)
		g = xm * alxm - nr_gammln2 (xm+1.)
	    }
	    repeat {
		repeat {
	            y = tan (PI * urand(seed))
		    em = sq * y + xm
		} until (em >= 0.)
	        em = int (em)
	        t = 0.9 * (1 + y**2) * exp (em * alxm - nr_gammln2 (em+1) - g)
	    } until (urand(seed) <= t)
	}
	return (em)
end

