# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  plfbinops.x
# 	Author	:  ericco
# 	Date	:  04/13/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) plfbinops.x -r1.7 04/13/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	07/02/93   akeith	documented
# 	06/17/92   gareth	edge effects corrected
# 	04/13/92   ericco	removed imhdr include file
# 	04/13/92   ericco	gareth's fix for spikes in output image of euvcombine
# 	04/13/92   ericco	date and time created 92/04/13 14:08:34 by ericco
# H-
# 
#

# DATE 7/2/93
# PLF BINOPS
#
# INCLUDE FILES
#

#
# AUTHOR : GARETH BEVAN
# HISTORY : CREATED 03 APRIL 1992

#
# plf_binops performs mathamatical operations between two piece-wise
# linear functions.
# Allowable operations
# +  add plfs,
# -  subtract plfs,
# %  mod plfs,
# /  devide plfs,
# *  multiply plfs,
# ^  plf 1 to the power of plf 2,
# &  bitwise and plfs resulting plf is 0 or 1,
# =  compare plfs resulting plf is 0 or 1,
# |  bitwise or plfs resulting plf is 0 or 1,
# <  compare plfs resulting plf is 0 or 1,
# >  compare plfs resulting plf is 0 or 1,
# t  atan2(plf1, plf2)
#

procedure plf_binops( x1, y1, len1, x2, y2, len2, op, outx, outy, outlen ) #E

#
# IMPORT VARIABLES :
real x1[len1], y1[len1]             #I INPUT PLF FUNCTIONS
real x2[len2], y2[len2]             #I SECOND INPUT PLF

int  len2			    #I LENGTH OF PLF
int  op                             #I THE OPERATION TO BE PERFORMED
int  len1                           #I LENGTH OF PLF FUCTION

# IMPORT/EXPORT VARIABLES :
# EXPORT VARIABLES :
pointer outx, outy                  #O OUTPUT PLF
int  outlen                         #O SIZE OF PLF

# FUNCTION
real plf_left_get()

# LOCAL VARIABLES :
int i,j,k                           # ARRAY INDECIES
real newpty, y_val                  # Y VALUE OF PLF
bool one_plf                        # TRUE IF WE ARE ONLY DEALING WITH ONE PLF
bool right1, right2                 # True if there is a right most value
bool lend1, lend2                   # True if the zero pts HAVE been inseted
                                    # into the PLF
bool rend1, rend2
bool done, connected                # are the plfs connected ie overlap

#############################################################################

begin
  # SET INITIAL VALUES
  outlen = len1 + len2 + 4
  i = 1
  j = 1
  k = 0

  if ( (len1 == 0 ) || ( len2 == 0 ))
  {
    call printf("error")
    outlen = 0
    outx = 0
    outy = 0
  }

  done = false
  connected = true  
  one_plf = false
  lend1 = false
  lend2 = false
  rend1 = false
  rend2 = false

  if ( y1[1] == 0.0 ) lend1 = true
  if ( y2[1] == 0.0 ) lend2 = true
  if ( y1[len1] == 0.0 ) rend1 = true
  if ( y2[len2] == 0.0 ) rend2 = true
  
  
  call malloc( outx, outlen, TY_REAL )
  call malloc( outy, outlen, TY_REAL )
  
  # FIND OUT THE STARTING CONDITIONS
  if ( x1[1] < x2[1] )
  {
    if ( !lend1 ) # add implied zero
      {
        Memr[outx+k] = x1[1]
        Memr[outy+k] = 0.0
        k = k + 1
      }

    repeat
    {
      Memr[outx + k] = x1[i]
      call plf_op( y1[i], 0.0, op, Memr[outy+k] )
      i = i + 1
      k = k + 1
      if ( i > len1 )
      {
        done = true
      }
      else 
      {
	if ( x1[i] > x2[1] )
	{
	  done = true
	}
      }

    } until ( done )

    # allow for plfs that do not overlap
    if ( i > len1 )
    {
      
      one_plf = true
      connected = false
      i = i - 1
    }
    else  # they do overlap
    {           
      if ( ! lend2 )
      {
        # INCLUDE IMPLIED ZERO POINT
        y_val = plf_left_get( x1, y1, len1, i, x2[1] )
        Memr[outx + k] = x2[1] 
        newpty = 0.0
        call plf_op( y_val, newpty, op, Memr[outy + k ] )
        k = k + 1
      }
    }
      
  }
  else if ( x2[1] < x1[1] )
  {

    if ( !lend2 ) # add implied zero point
    {
      Memr[outx+k] = x2[1]
      Memr[outy+k] = 0.0
      k = k + 1
    }

    repeat
    {
      Memr[outx+k] = x2[j]
      call plf_op( 0.0, y2[j], op, Memr[outy+k] )
      j = j + 1
      k = k + 1
      if ( j > len2 )
      {
        done = true
      }
      else
      {
        if ( x2[j] > x1[1] )
        {
          done = true
        }               
      }
    } until ( done )

    # Allow for non over lapping plfs

    if ( j > len2 )
    {
      
      one_plf = true
      connected = false
      j = j - 1
    }
    else  # they do overlap
    {
      
      if ( ! lend1 )
      {
        # INCLUDE IMPLIED ZERO POINT
        y_val =  plf_left_get( x2, y2, len2, j, x1[1] )
        Memr[outx+k] = x1[1]
        newpty = 0.0
        call plf_op( newpty, y_val, op, Memr[outy+k] )
        k = k + 1
      }
    }  
  }

  else # Start points are equal
  {
    
    # Include initial zero pt
    Memr[outx+k] = x1[1]
    Memr[outy+k] = 0.0
    k = k + 1
    
    call plf_right_get( x1, y1, len1, i, right1 )
    call plf_right_get( x2, y2, len2, j, right2 )


    if ( right1 && right2 )
    {
      
      call plf_op( y1[i], y2[j], op, Memr[outy+k] )
      Memr[outx+k] = x1[1]
      k = k + 1
      i = i + 1
      j = j + 1
    }	
    else if ( right1 )
    {
      
      call plf_op( y1[i], y2[j], op, Memr[outy+k] )
      Memr[outx+k] = x1[1]
      k = k + 1
      i = i + 1
      j = j + 1
    }
    else if ( right2  )
    {
      
      call plf_op( y1[i], y2[j], op, Memr[outy+k] )
      Memr[outx+k] = x1[1]
      k = k + 1
      i = i + 1
      j = j + 1
    }
    else # Niether plf starts with a discontinuity
    {
      
      Memr[outx+k] = x1[1]
      call plf_op( y1[1], y2[1], op, Memr[outy+k] )
      i = i + 1
      j = j + 1
      k = k + 1
    }      
  } # ENDIF
  
  
  # NOW WE ARE CONSIDERING BOTH PLFS AT THE SAME TIME
  while ( ! one_plf ) 
  {

    if ( x1[i] < x2[j] )
    {
      y_val = plf_left_get( x2, y2, len2, j, x1[i] )
      call plf_op( y1[i], y_val, op, Memr[outy+k] )
      Memr[outx+k] = x1[i]

      call plf_right_get( x1, y1, len1, i, right1 )

      if ( right1 )
      {
	
	k = k + 1
	call plf_op( y1[i], y_val, op, Memr[outy+k] )
	Memr[outx+k] = x1[i]
      }
      k = k + 1
      i = i + 1
      if ( i > len1 )
	{ 
          one_plf = true
          i = len1       # allow for end condition
	}

    }
    else if ( x2[j] < x1[i] )
    {
      
      y_val =  plf_left_get( x1, y1, len1, i, x2[j] )

      call plf_op( y_val, y2[j], op, Memr[outy+k] )
      Memr[outx+k] = x2[j]
      
      call plf_right_get( x2, y2, len2, j, right2 )
      if ( right2 )
      {
	k = k + 1
	call plf_op( y_val, y2[j], op, Memr[outy+k] )
	Memr[outx+k] = x2[j]
      }
      k = k + 1
      j = j + 1
      if ( j > len2 ) 
	{
          one_plf = true
          j = len2       # allow for end condition
	}
    }
    else # THE DISCONTINUITIES ARE IN THE SAME PLACE
    {

      Memr[outx+k] = x1[i]
      call plf_op( y1[i], y2[j], op, Memr[outy+k] )
      
      call plf_right_get( x1, y1, len1, i, right1 )
      call plf_right_get( x2, y2, len2, j, right2 )
      
      if ( right1 && right2  )
      {
	
	k = k + 1
	call plf_op( y1[i], y2[j], op, Memr[outy+k] )
	Memr[outx+k] = x1[i]
      }
      else if ( right1  )
      {
	
	k = k + 1
	Memr[outx+k] = x1[i]
	call plf_op( y1[i], y2[j], op, Memr[outy+k] )
      }
      else if ( right2  )
      {
      
	k = k + 1
	Memr[outx+k] = x1[i]
	call plf_op( y1[i], y2[j], op, Memr[outy+k] )
      }

      if ( ( i >= len1 ) || ( j >= len2 ) ) 
      {
	one_plf = true
      }
      else
      {
	i = i + 1
        j = j + 1
      }

      k = k + 1

    }    

  } # END WHILE

  # Deal with the last part of the last plf
  if ( (i >= len1) && ( j < len2 ) )
  {

    if ( !connected ) 
    {
      if ( ! rend1 )
      {
        
        # Insert implied zero pt at end of 1st plf
        Memr[outx+k] = x1[len1]
        Memr[outy+k] = 0.0
        k = k + 1
      }
      
      if ( !lend2 )# insert zero at start of 2nd plf
      { 
        Memr[outx+k] = x2[1]
        Memr[outy+k] = 0.0
        k = k + 1
      }
    }
    else
    {
 
      if ( ! rend1 )
      {
        
        # Insert implied zero from end of 1st plf
        Memr[outx+k] = x1[len1]
        newpty = 0.0
        y_val = plf_left_get( x2, y2, len2, j, x1[len1] )
        call plf_op( newpty, y_val , op, Memr[outy+k] )
        k = k + 1
      }
    }

    
    for ( ; j<=len2; j=j+1 )
    {
      
      Memr[outx+k] = x2[j]
      call plf_op( 0.0, y2[j], op, Memr[outy+k] )
      k = k + 1
    }

    if ( !rend2 ) # add zero end point
    {
      Memr[outx+k] = x2[len2]
      Memr[outy+k] = 0.0
      k = k + 1
    }

  }
  else if ( ( j >= len2 ) && ( i < len1 ) ) 
  {

    if ( !connected ) # add first zero point
    {
      if ( ! rend2 )
      {
        
        # Insert zero at end of 1st plf
        Memr[outx+k] = x2[len2]
        Memr[outy+k] = 0.0
        k = k + 1
      }
      if ( !lend1 )# insert zero at start of 2nd plf
      { 
        Memr[outx+k] = x1[1]
        Memr[outy+k] = 0.0
        k = k + 1
      }
    }
    else
    {
   
      if ( ! rend2 )
      {
      
        # Insert implied zero at end of first plf
        Memr[outx+k] = x2[len2]
        newpty = 0.0
        y_val = plf_left_get( x1, y1, len1, i, x2[len2] )
        call plf_op( y_val, newpty, op, Memr[outy+k] )
        k = k + 1
      }
    }  
  
    for ( ; i<=len1; i=i+1 )
    {
      
      Memr[outx+k] = x1[i]
      call plf_op( y1[i], 0.0, op, Memr[outy+k] )
      k = k + 1
    }
    if ( !rend1 ) # add zero end point
    {
      Memr[outx+k] = x1[len1]
      Memr[outy+k] = 0.0
      k = k + 1
    }

  }
  else # Both plfs run out together
  {
    if ( x1[len1] > x2[len2] )
    {
      if ( !rend2 ) # implied zero at end of short plf
      {
	y_val = plf_left_get( x1, y1, len1, len1, x2[len2] )
	call plfop( y_val, 0.0, op, Memr[outy+k] )              
        Memr[outx+k] = x2[len2]
        k = k + 1
      }
      if ( !rend1 ) # undefined region and end of long plf
      {
	call plfop( y1[len1], 0.0, op, Memr[outy+k] )              
        Memr[outx+k] = x1[len1]
        k = k + 1
       }
    }
    else if ( x2[len2] > x1[len1] )
    {
      if ( !rend1 ) # implied zero at end of short plf
      {
	y_val = plf_left_get( x2, y2, len2, len2, x1[len1] )
	call plfop( 0.0, y_val, op, Memr[outy+k] )              
        Memr[outx+k] = x1[len1]
        k = k + 1
      }
      if ( !rend2 ) # undefined region and end of long plf
      {
	call plfop( 0.0, y2[len2], op, Memr[outy+k] )              
        Memr[outx+k] = x2[len2]
        k = k + 1
      }
    }
      
    if ( Memr[outy+k-1] != 0.0 )
    {
      
      Memr[outx+k] = max( x1[len1], x2[len2] )
      Memr[outy+k] = 0.0
      k = k + 1
    }
    
     
  } # END OF COPY SECTION
  


  outlen = k 
  
end  # OF PROCEDURE PLF_BIN_OP
