# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  qpsrtclose.x
# 	Author	:  akeith
# 	Date	:  01/27/93
# 	Synopsis:  
# 	Keywords:
# 		@(#) qpsrtclose.x -r1.5 01/27/93 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	09/28/94   akeith	documentation
# 	09/27/94   akeith	now utilizes the FMIO library
# 	06/29/94   akeith	fixed imdir for tmp files
# 	03/24/93   akeith	everything has changed
# 	01/27/93   akeith	date and time created 93/01/27 17:25:47 by akeith
# H-
# 
#


#       Version: @(#)qpsrtclose.x	1.5 9/28/94 UCB CEA EGO

include "qs.h"

define	LFILE		Meml[($1)+($2)-1]
define	EVENT		Mems[($1)+(($2)-1)*($3)]

#
# qpsrt_close() proforms the merge sort across the N temporary files,
#	writing the events out to qpio_putevents() as it goes, then
#	the qpsrt structure is closed.
#

procedure	qpsrt_close(qs)  #E

pointer	qs				#I/O qpoe sort structure pointer

int	nev_left			# num events left
pointer	fm_open()			#

begin

	nev_left = QS_NEVBUF(qs) - QS_NEV_AVAIL(qs)
	call qs_sortout(qs, nev_left)

	if(QS_FM(qs) != NULL) {
		call fm_close(QS_FM(qs))
		QS_FM(qs) = fm_open(QS_FNAME(qs), READ_ONLY)
	}

	call qs_merge(qs)


	if(QS_BUF_PTR(qs) != NULL)
		call mfree(QS_BUF_PTR(qs), TY_SHORT)
	QS_BUF_PTR(qs) = NULL
	QS_BUF_LEN(qs) = 0

	if(QS_SBUF_PTR(qs) != NULL)
		call mfree(QS_SBUF_PTR(qs), TY_LONG)
	QS_SBUF_PTR(qs) = NULL
	QS_SBUF_LEN(qs) = 0

	if(QS_DIR_PTR(qs) != NULL)
		call mfree(QS_DIR_PTR(qs), TY_CHAR)
	QS_DIR_PTR(qs) = NULL

	if(QS_FM(qs) != NULL) {
		call fm_close(QS_FM(qs))
		call fm_delete(QS_FNAME(qs))
	}
	QS_FM(qs) = NULL

	if(QS_FNAME_PTR(qs) != NULL)
		call mfree(QS_FNAME_PTR(qs), TY_CHAR)
	QS_FNAME_PTR(qs) = NULL
	
	call qs_free(qs)

end


#
# qs_merge() merges separately sorted batches of event records
#	      and write them to the qpoe file
#

procedure qs_merge(qs)

pointer	qs				#I/O qpoe sort structure pointer

int	lf				# lfile index
int	op				# # events in obuf
int	done				# number of temp files completed
int	nchars				# number of SPP chars just read
int	ev_sz				# event size
int	nlfs				# number of lightweight file
pointer ebuf				# qpc recs from the temp files
pointer obuf				# buffer to hold output events
pointer	optr				# output pointer
pointer ev				# current event pointer
pointer	fd				# list of lightweight files
pointer	sp				# stack pointer

int	qs_read()			# read data

begin

	ev_sz = QS_EVSZ(qs)
	nlfs = QS_NLF(qs)

	# mark the stack and
	# allocate space for event recs (1 per file) for merge

	call smark(sp)
	call salloc (fd, nlfs, TY_LONG)
	call salloc (ebuf, nlfs*ev_sz, TY_SHORT)
	call salloc (obuf, QS_NEVBUF(qs)*ev_sz, TY_SHORT)
	call salloc (optr, QS_NEVBUF(qs), TY_POINTER)

	call amovkl(1, Meml[fd], nlfs)

	# read the first record of each file

	do lf = 1, nlfs
	    nchars = qs_read(QS_FM(qs), fd, lf, ebuf, ev_sz)

	# process all data records in the temp files

	done = 0
	op = 0

	while( done != nlfs ) {

	    # find next qpc to write
	    call qs_min(qs, fd, ebuf, lf)

	    # move record photon to a safe place
	    ev = obuf + (op * ev_sz)
	    Memi[optr+op] = ev
	    call amovs(EVENT(ebuf, lf, ev_sz), Mems[ev], ev_sz)

	    # bump pointer and flush output buffer when it fills
	    op = op + 1
	    if( op >= QS_NEVBUF(qs) ) {
		call qpio_putevents (QS_QPIO(qs), Memi[optr], op)
		op = 0
	    }

	    # read the next record from this temp file

	    if(qs_read(QS_FM(qs), fd, lf, ebuf, ev_sz) == EOF)
	        done = done + 1

	}

	# flush any unbuffered events
	if( op > 0 )
	      call qpio_putevents(QS_QPIO(qs), Memi[optr], op)

	# free stack space
	call sfree(sp)
end


#
# qs_read() reads a single event out of a lfile.
#

int	procedure	qs_read(fm, fd, lf, ebuf, ev_sz)

pointer	fm			#I fmio struct
pointer fd			#I pointer to array lightweight files
int	lf			#I lfile index
pointer	ebuf			#I pointer to array of qpc recs
int	ev_sz			#I event size

int	fm_getfd()		# fmio get file desc
int	nchars, read()
long	fpos, note()

begin

	fpos = LFILE(fd,lf)
	LFILE(fd,lf) = fm_getfd(fm, lf, READ_ONLY, BINARY_FILE)
	call seek(LFILE(fd,lf), fpos)
	nchars = read(LFILE(fd,lf), EVENT(ebuf, lf, ev_sz), ev_sz)
	fpos = note(LFILE(fd,lf))
	call fm_retfd(fm, lf)

	if( nchars == EOF ) LFILE(fd,lf) = -1
	else		    LFILE(fd,lf) = fpos

	return(nchars)

end


#
# qs_min() - find minimum of a group of qpc records
#

procedure qs_min(qs, fd, ebuf, min)

pointer	qs			#I/O qpoe sort structure pointer
pointer fd			#I pointer to array lightweight files
pointer	ebuf			#I pointer to array of qpc recs
int	min			#O min qpc record

int	i			# loop counter
int	size			# event size
int	nlfs			# num of file descr

int	qs_cmp()

begin

	min = 0
	size = QS_EVSZ(qs)
	nlfs = QS_NLF(qs)

	# now look for the smallest qpc record
	# find the minimum, based on the sort type

	do i = 1, nlfs
	    if(LFILE(fd,i) != -1 ) {
		min = i
		break
	    }

	if(min == 0)
		call error(1, "qs_min(): min = 0")

	do i = min+1, nlfs
	    if(LFILE(fd,i) != -1 ){
		if(qs_cmp(ebuf+((i-1)*size), ebuf+((min-1)*size)) < 0)
		    min = i
	    }

end


