# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_tbverfsort.x
# 	Author	:  tnk
# 	Date	:  06/24/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_tbverfsort.x -r1.6 06/24/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	01/17/97   tnk	changed message format, added diffs
# 	04/17/96   tnk	added tolcount, maxreports
# 	04/16/96   tnk	added tolerance; now continues after 1st out-of-order
# 	03/16/95   tnk	changed percent printout on gaps
# 	03/16/95   tnk	added gap checking
# 	06/24/94   tnk	date and time created 94/06/24 13:46:36 by tnk
# H-
# 
#

# @(#)t_tbverfsort.x	1.6 1/17/97 UCB CEA EGO

include <tbset.h>
include <mach.h>

# determine if a table is sorted by two columns

procedure	t_tbverfsort()

# input params
pointer	tbname		 # QPIO filename param. ptr
pointer ycol             # primary key
pointer xcol             # secondary key

#local vars
pointer 	tb
pointer		sp
int             i
double          x
double          y
int             num_rows
int             pct
int             oldpct
double          old_x
double          old_y
bool		is_sorted
pointer         ycolcp
pointer         xcolcp
bool            read_err
int             startrow
double		gap
bool		onecol
double		tolerance
int		tolcount
int 		maxreports
int		nreports
double		diff
double		maxdiff

# FUNCTIONS 
pointer tbtopn()                # table initialization
int     tbpsta()
int     clgeti()
double  clgetd()
int     strcmp()

begin


  call smark(sp)

  call eu_clstr(tbname,   "table")   # get qpoe filename 
  call eu_clstr(ycol,     "ycol")    # primary sort col
  call eu_clstr(xcol,     "xcol")    # secondary sort col 
  gap = clgetd("gap")
  startrow = clgeti("startrow")
  tolerance = clgetd("tolerance")
  maxreports = clgeti("maxreports")

  # open table
  tb = tbtopn(Memc[tbname],READ_ONLY, 0)

  if (tb == NULL)
     call error(100,"cannot open table")

  # open columns
  call tbcfnd(tb, Memc[ycol], ycolcp,  1)  #
  call tbcfnd(tb, Memc[xcol], xcolcp,  1)  #

  if (ycolcp == NULL)
     call error(100,"cannot open primary column")

  if (xcolcp == NULL)
     call error(100,"cannot open secondary column")

  onecol = false
  if (strcmp(Memc[ycol], Memc[xcol]) == 0)
     onecol=true

  if (IS_INDEFD(gap)) gap = MAX_DOUBLE
  if (IS_INDEFI(maxreports)) maxreports = MAX_INT


  num_rows = tbpsta(tb, TBL_NROWS )

  call eprintf("reading table rows...\n")

  old_x = -MAX_DOUBLE
  old_y = -MAX_DOUBLE
  is_sorted = true
  tolcount = 0
  nreports = 0
  maxdiff = 0.0
  for (i=startrow; i<= num_rows; i=i+1) {

    pct = int(10.*real(i)/real(num_rows))*10
    if (pct != oldpct) {
        oldpct = pct
        call eprintf ("  %3d%% read\n")
        call pargi (pct)
    }

    call tbrgtd(tb, ycolcp, y, read_err, 1, i) 
    call tbrgtd(tb, xcolcp, x, read_err, 1, i) 

    # gap check 
    if (y-old_y > gap)
    {
       pct = int( (real(i)/real(num_rows))*100)
       call eprintf ("  %3d%% read, ")
       call pargi (pct)

       call eprintf("row=%d, ")
       call pargi(i)

       call eprintf("gap=%.0f\n")
       call pargd(y-old_y)


    }

    if (y < old_y)
    {
       diff = old_y-y
       if (diff <= tolerance)
       {
          tolcount = tolcount + 1
       }    
       else 
       {
	  call printf("row= %d  old_y=%16.5f y=%16.5f diff=%16.5f\n")
	  call pargi(i)
	  call pargd(old_y)
	  call pargd(y)
	  call pargd(diff)
	  is_sorted = false    

	  if (diff > maxdiff)
		maxdiff = diff

          nreports = nreports + 1
          if (nreports >= maxreports) break

       }
    }
    else if (y > old_y)
    {
       old_x = -1
    }


    if (x < old_x && !onecol)
    {
       call printf("row= %d old_x=%22.10f\n")
       call pargi(i-1)
       call pargd(old_x)

       call printf("row= %d     x=%22.10f\n")
       call pargi(i)
       call pargd(x)
       is_sorted = false    

       nreports = nreports + 1
       if (nreports >= maxreports) break

    }
    else
    {
       old_x = x
    }
    old_y = y


  }

  if (nreports >= maxreports)
  {
     call printf("Max number of reports exceeded, scan terminated.\n");
  }

  if (tolcount > 0)
  {
     call printf("%d rows were out of order, but within tolerance.\n")
     call pargi(tolcount)
  }

  if (is_sorted)
     call printf("Table is sorted by y,x.\n")
  else
  {
     call printf("Table is NOT sorted by y,x.\n")
     call printf("%d rows were out of order.\n")
     call pargi(nreports)
     call printf("Max shift back in y: %16.5f\n")
     call pargd(maxdiff)
  }

  # Clean up
  call tbtclo(tb)

  call sfree(sp)
end




