# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_euvcombine.x
# 	Author	:  ericco
# 	Date	:  03/24/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_euvcombine.x -r1.13 03/24/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	08/06/93   akeith	fixed mwcs problem
# 	07/16/93   akeith	took out debug statements
# 	04/30/93   gareth	added indef default
# 	04/30/93   gareth	does not divide by wpc when counts = yes
# 	04/30/93   gareth	added refdata
# 	01/04/93   gareth	added check that effective areas exist
# 	10/01/92   gareth	added mwcs. now accepts new and old images. writes new ones
# 	04/30/92   gareth	Changed to keep track of the edges of the bins as well as the bin centers
# 	03/24/92   ericco	date and time created 92/03/24 15:08:27 by ericco
# H-
# 
#

include	<imhdr.h>

#	Combine EUV 1d spectra using effective areas
#	and count rates.
#	This procedure currently computes the order separation
#	by using plfs.  This could be re-written to use
#	arrays instead.  This would speed up computations.
procedure	t_euvcombine ()

pointer		spectra
pointer		clpopnu()

pointer		sp
pointer		spec
pointer		det
pointer		et
pointer		cnts
pointer		imgl1r()

int		cnt
int		index
pointer		im
pointer		immap()
pointer		out
pointer		output
pointer		table
int		clplen()
int		total
int		orders
pointer		wl, ea
int		ln
pointer		detector
int		clgfil()
pointer		edd
int		current_orders
int		i
int		maxorders

pointer		wli, eai
int		lni
int		length
pointer		cwl, cea
int		clen
pointer		twl, tea
int		tlength
real		nw0, w0
real		nwpc, wpc
real		nwn, wn
real		imgetr()
pointer		xs, ys
pointer		cxs, cys
pointer		nxs, nys
int		nlength

int		nlen
pointer		outs
pointer		impl1r()
int		len
real		lambda0, lambdan, lambda
pointer		wl1, ea1
int		ln1
pointer		txs, tys
int		tlen
int		clgeti()
real		clgetr()
real		minwl, maxwl
bool		clgetb()
bool		autosize
bool		countp
real		exposure
pointer		in
pointer		los, his
real		xl, xh
real		xlo, xhi
real		plf_rwn

pointer         sam_ptr              # pointer to resampling array
int             resample             # resampling factor
int             samlen               # size of the resampling array
real            wzero                # resampled w0
real            rwpc                 # resampled wpc
bool            memo                 # true if we are to resample to conserve 
                                     # memory

#MWCS definitions
pointer mw
pointer ct
real	cd, r, w

pointer mw_sctran()
real	mw_c1tranr()

#FUNCTIONS:
pointer mw_open()
pointer mw_openim()

int imaccf()



##########################################################################
begin

  call smark(sp)
  call salloc(spectra, SZ_LINE, TY_CHAR)
  call salloc(detector, SZ_LINE, TY_CHAR)
  call salloc(table, SZ_LINE, TY_CHAR)
  call salloc(output, SZ_LINE, TY_CHAR)

  spec = clpopnu("spectra")
  det = clpopnu("detectors")
  call clgstr("refdata", Memc[table], SZ_LINE)
  call clgstr("output", Memc[output], SZ_LINE)
  maxorders = clgeti("maxorders")

  if ( IS_INDEFI(maxorders) ) maxorders = 0
 
  countp = clgetb("counts")
  exposure = clgetr("exposure")

  if ( IS_INDEFR(exposure) ) exposure = 0

  autosize = clgetb("autosize")
  resample = clgeti("resample")
  if (! autosize) {
    minwl = clgetr("minwavelength")
    maxwl = clgetr("maxwavelength")
  }

  if ( IS_INDEFI( resample ) ) memo = false
  else memo = true

  call et_open(et, Memc[table])

  if (clplen(spec) != clplen(det)) {
    call error (1, "Wrong number of elements in the operand lists")
  }

  total = clplen(det)
  call salloc(los, total, TY_REAL)
  call salloc(his, total, TY_REAL)

  # add counts from all detectors
  index = 0
  xlo = 0

  while (clgfil(spec, Memc[spectra], SZ_LINE) != EOF) {
    index = index + 1
    im = immap(Memc[spectra], READ_ONLY, 0)
    cnts = imgl1r(im)

    if ( imaccf( im, "W0" ) == YES ) {
      nw0 = imgetr(im, "W0")
      nwpc = imgetr(im, "WPC")
    } else {
      # open the mwcs
      mw = mw_openim( im )
      ct = mw_sctran(mw, "logical", "world", 1b)
      nw0 = mw_c1tranr(ct, 1.0)
      nwpc = mw_c1tranr(ct, 2.0) - nw0
    }

    nwn = nw0 + nwpc * ( IM_LEN(im,1) - 1 )
    if (!countp) {
      call amulkr(Memr[cnts], 1/nwpc, Memr[cnts], IM_LEN(im,1))
    }
    call plf_from_array(nw0, nwpc, Memr[cnts], IM_LEN(im,1),
			twl, tea, tlength)

    if (! autosize) xhi = maxwl
    else xhi = nwn + nwpc / 2

    call plf_clip(Memr[twl], Memr[tea], tlength, xlo, xhi, xs, ys, length)
    call plf_limits(Memr[xs], Memr[ys], length, 0.0,
		    Memr[los+index-1], Memr[his+index-1])

    call mfree( twl, TY_REAL )
    call mfree( tea, TY_REAL )


    if (index == 1) {
      in = im
      out = immap(Memc[output], NEW_COPY, im)

      if ( imaccf( out, "W0"  ) == YES ) call imdelf( out, "W0" )
      if ( imaccf( out, "WPC" ) == YES ) call imdelf( out, "WPC" )

      IM_NDIM(out) = 1
      IM_PIXTYPE(out) = TY_REAL
      # reserve memory for resampling plfs

      
      w0 = nw0
      wpc = nwpc
      wn = nwn
      cxs = xs
      cys = ys
      clen = length
    } else {
      if (nw0 < w0)	w0 = nw0
      if (nwpc > wpc)	wpc = nwpc
      if (nwn > wn)	wn = nwn


      call plf_binops(Memr[cxs], Memr[cys], clen,
		      Memr[xs], Memr[ys], length, '+',
		      nxs, nys, nlength)

      call mfree(cxs, TY_REAL)
      call mfree(cys, TY_REAL)
      call mfree(xs, TY_REAL)
      call mfree(ys, TY_REAL)
      cxs = nxs
      cys = nys
      clen = nlength
      call imunmap(im)
    }
  } # end while


  call clpcls(spec)
  len = (wn - w0) / wpc
  if (w0 + len * wpc < wn) len = len + 1
  len = len + 1
  IM_LEN(out,1) = len

  if ( memo ) {
    samlen = resample*IM_LEN( out,1 ) + 1
    call malloc( sam_ptr, samlen, TY_REAL )
  }
  
  # collect effective area information
  orders = 0
  wl = 0
  ea = 0
  ln = 0
  cnt = 0
  

  while (clgfil(det, Memc[detector], SZ_LINE) != EOF) {
    cnt = cnt + 1
    
    xl = Memr[los+cnt-1]
    xh = Memr[his+cnt-1]

    call edd_new(edd, et, Memc[detector])

    call edd_ea_orders(edd, current_orders)

    if (current_orders < 1) next
    if (maxorders > 0 && current_orders > maxorders) current_orders = maxorders

    # extend wl, ea, ln arrays to cover all orders
    if (current_orders > orders) {
      call realloc(wl, current_orders, TY_LONG)
      call realloc(ea, current_orders, TY_LONG)
      call realloc(ln, current_orders, TY_LONG)
      do i = orders+1, current_orders {
	Meml[wl+i-1] = 0
	Meml[ea+i-1] = 0
	Meml[ln+i-1] = 0
      }
      orders = current_orders
    }

    # combine all detector grouped by orders
    do i = 1, current_orders {

      wli = Memi[wl+i-1]
      eai = Memi[ea+i-1]
      lni = Memi[ln+i-1]

      call edd_ea_order(edd, i, length, cwl, cea)
      if (length == 0) next

      if (! autosize) xhi = maxwl
      else xhi = Memr[cwl+length-1]

      xlo = xl
      xhi = min(xh,xhi)
      call plf_clip(Memr[cwl], Memr[cea], length, xlo, xhi, twl, tea, tlength)


      if (lni == 0) {

	lni = tlength
	wli = twl
	eai = tea

      } else {

	call plf_binops(Memr[twl], Memr[tea], tlength,
			Memr[wli], Memr[eai], lni, '+',
			cwl, cea, length)

	call mfree(wli, TY_REAL)
	call mfree(eai, TY_REAL)
	call mfree( twl, TY_REAL )
	call mfree( tea, TY_REAL )
	wli = cwl
	eai = cea
	lni = length
      }
      
      Memi[wl+i-1] = wli
      Memi[ea+i-1] = eai
      Memi[ln+i-1] = lni

    } # end do


    call edd_close(edd)

  } # end while (clgfil)


  call clpcls(det)
  call et_close(et)

  if (! autosize) {
    w0 = minwl
    wn = maxwl
  }


  if (index > 0 && wl != 0) {
    plf_rwn = wn + wpc / 2
    call plf_flatten(Memr[cxs], Memr[cys], clen, nxs, nys, nlen)
    call mfree(cxs, TY_REAL)
    call mfree(cys, TY_REAL)
    cxs = nxs
    cys = nys
    clen = nlen

    # cxs, cys, clen contain the plf of combined counts
    # wl, ea, ln contains arrays of orders

    # compute lambda0.  the longest wavelength such that
    # first order counts less than lambda0 are all zero.
    call plf_limits(Memr[cxs], Memr[cys], clen, 0.0, lambda0, lambdan)

    if ( lambda0 <=0.0 ) {
      call printf("starting wavelength is <= 0.0 \n ")
      call flush(STDOUT)
      return
    }

    wl1 = Memi[wl]
    ea1 = Memi[ea]
    ln1 = Memi[ln]

    # break spectral band into orders of lambda0
    # note that we dont have to do the first order band


    for (lambda = 2 * lambda0; lambda < plf_rwn; lambda = lambda + lambda0) {
      # remove lower orders from this band

      do i = 2, orders { 

	wli = Memi[wl+i-1]
	eai = Memi[ea+i-1]
	lni = Memi[ln+i-1]

	xlo = lambda / i
	if (lambda0 + lambda > plf_rwn) {
	  xhi = plf_rwn / i
	} else {
	  xhi = (lambda + lambda0) / i
	}
	if (xhi <= lambda0) next	# nothing down here


	call plf_clip(Memr[wl1], Memr[ea1], ln1, xlo, xhi, txs, tys, tlen)
	call plf_binops(Memr[cxs], Memr[cys], clen,
			Memr[txs], Memr[tys], tlen, '/',
			nxs, nys, nlen)


        call mfree(txs, TY_REAL)
	call mfree(tys, TY_REAL)
	# nxs, nys contain first order counts/cm^2 restricted to clip region



	call plf_new_order(Memr[nxs], Memr[nys], nlen, -(i-1), 1,
			   twl, tea, tlen)
	call mfree(nxs, TY_REAL)
	call mfree(nys, TY_REAL)


	call plf_binops(Memr[twl], Memr[tea], tlen,
			Memr[wli], Memr[eai], lni, '*',
			nxs, nys, nlen)

	call mfree(twl, TY_REAL)
	call mfree(tea, TY_REAL)
	# nxs, nys has the first order counts that show up in the ith order


	# subtract ith order counts from raw counts

	call plf_binops(Memr[cxs], Memr[cys], clen,
			Memr[nxs], Memr[nys], nlen, '-',
			txs, tys, tlen)

	call mfree(nxs, TY_REAL)
	call mfree(nys, TY_REAL)
	call mfree(cxs, TY_REAL)
	call mfree(cys, TY_REAL)
	cxs = txs
	cys = tys
	clen = tlen
	if ( memo )
	{

	  # resample plf to conserve memory
	  wzero = w0 - ((resample/2.0 - 0.5 )*( wpc/resample ))
	  rwpc = wpc / resample	
	  
	  
	  call plf_to_array( Memr[cxs], Memr[cys], clen, wzero, rwpc,
			     Memr[sam_ptr], samlen )
	  call mfree( cxs, TY_REAL )
	  call mfree( cys, TY_REAL )
	  call plf_from_array( wzero, rwpc, Memr[sam_ptr], samlen,
			       cxs, cys, clen )


	      
	} # end if
	
      } #end do
    } # end for


    outs = impl1r(out)
    if (! countp) {
      call plf_binops(Memr[cxs], Memr[cys], clen,
		      Memr[wl1], Memr[ea1], ln1, '/',
		      nxs, nys, nlen)
      call mfree(cxs, TY_REAL)
      call mfree(cys, TY_REAL)
      cxs = nxs
      cys = nys
      clen = nlen


      if (exposure > 0) {
	call amulkr(Memr[cys], 1/exposure, Memr[cys], clen)
      }

    }

    call plf_to_array(Memr[cxs], Memr[cys], clen, w0, wpc, Memr[outs], len)

    # write out results
    mw = mw_open( NULL, 1 )
    call mw_newsystem( mw, "linear", 1 )
    call mw_swattrs( mw, 1, "label", "wavelength" )
    call mw_swattrs( mw, 1, "units", "angstroms" )
    cd = wpc
    w = w0
    r = 1

    call mw_swtermr( mw, r, w, cd, 1 )
    call mw_saveim( mw, out )

    call imunmap(out)
    call imunmap(in)
  }

  # release resources
  if (orders > 0) {
    do i = 1, orders {
      call mfree(Memr[wl+i-1], TY_REAL)
      call mfree(Memr[ea+i-1], TY_REAL)
    }
    call mfree(wl, TY_LONG)
    call mfree(ea, TY_LONG)
    call mfree(ln, TY_LONG)
  }

  if ( memo )
  {
    call mfree( sam_ptr )
  }

  call sfree(sp)


end
