# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_distcrt.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_distcrt.x -r1.3 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/29/91   ericco	convert integer values to real values
# 	05/07/91   ericco	date and time created 91/05/07 08:49:38 by ericco
# H-
# 
#

#	Revised: 4/30/92

include	<edata.h>
include <imhdr.h>

#	Create a distortion correction image.
#	Not only does this create the 3D image of correction
#	information, it also generates a description of the
#	data suitable for a reference text database.

procedure	t_distcrt ()		# create correction data

pointer		sp
pointer		input, output, record, image, xplaneid, outrec
pointer		disttype

int		dt, dtmap()
pointer		dist
pointer		img
int		yplane, xplane
int		clgeti()
int		sz[2]
real		xmin, xmax, ymin, ymax
pointer		wcs
real		xref, yref
pointer		fi, immap()
bool		fconst, clgetb()
real		fval, clgetr()
pointer		flat
pointer		fplaneid, fvalueid, yplaneid
int		fplane
pointer		mw, ct
pointer		mw_sctran()

begin
  call smark(sp)

  call euclstr(input,"input")
  call euclstr(record,"record")
  call euclstr(output,"output")
  call euclstr(outrec,"outrec")
  call euclstr(image,"image")
  fconst = clgetb("constantp")
  if (fconst == TRUE) {
    fval = clgetr("constant")
  } else {
    call euclstr(flat,"flat")
  }

  xplane = clgeti("xplane")
  if (xplane == 1) {yplane = 2} else {yplane = 1; xplane = 2}
  fplane = 3
  if (fconst == FALSE) {
    fi = immap(Memc[flat], READ_ONLY, 0)
    sz[1] = IM_LEN(fi, 1)
    sz[2] = IM_LEN(fi, 2)
  } else {
    sz[1] = clgeti("xdim")
    sz[2] = clgeti("ydim")
  }
  xmin = clgetr("xmin")
  xmax = clgetr("xmax")
  ymin = clgetr("ymin")
  ymax = clgetr("ymax")
  xref = clgetr("xref")
  yref = clgetr("yref")
  call euclstr(wcs,"wcsname")
  call euclstr(xplaneid,"xplaneid")
  call euclstr(yplaneid,"yplaneid")
  call euclstr(fplaneid,"fplaneid")
  call euclstr(fvalueid,"fvalueid")
  call euclstr(disttype,"disttype")

  call dc_load_corr(Memc[input],Memc[record], dist)
  if (dist != 0) {
    call dc_create(Memc[image], sz, img)
    call dc_wcs(img, Memc[wcs], xmin, xmax, ymin, ymax, xref, yref, mw)
    call imaddi(img, Memc[xplaneid], xplane)
    call imaddi(img, Memc[yplaneid], yplane)
    if (fconst == TRUE) {
      call imaddr(img, Memc[fvalueid], fval)
    } else {
      call imaddi(img, Memc[fplaneid], fplane)
    }
    ct = mw_sctran(mw, "physical", Memc[wcs], 3B)
    call dc_build(img, dist, xplane, ct)
    if (fconst == FALSE) call dc_copy(img, fi, fplane)
    call mw_ctfree(ct)
    call mw_close(mw)
    call imunmap(img)
    call ed_fr_corr(dist)
    dt = dtmap(Memc[output], APPEND)
    call dtptime(dt)
    call dtput(dt, "begin\t%s\n")
    call pargstr(Memc[outrec])
    call dtput(dt, "\tType\t%s\n"); call pargstr(Memc[disttype])
    call dtput(dt, "\tName\t%s\n"); call pargstr(Memc[image])
    call dtput(dt, "\tXPlaneref\t%s\n"); call pargstr(Memc[xplaneid])
    call dtput(dt, "\tYPlaneref\t%s\n"); call pargstr(Memc[yplaneid])
    call dtput(dt, "\twcsref\t%s\n"); call pargstr(Memc[wcs])
    call dtput(dt, "\tfvalref\t%s\n"); call pargstr(Memc[fvalueid])
    call dtput(dt, "\tfplaneref\t%s\n"); call pargstr(Memc[fplaneid])
    call dtput(dt, "\tfconstant\t%d\n"); call pargb(fconst)
    call dtunmap(dt)
  } else {
    call printf("Couldn't read distortion map\n")
  }

  call sfree(sp)
end

#	Initialize a 3d image from a 2d image and insert the 2d
#	image into one plane aligned with the x-y plane.

procedure	dc_copy(im, fi, fplane)		# initial 3d data
pointer		im				#I/O	3d image
pointer		fi				#I	2d image
int		fplane				#I	plane to use

pointer		impl3r(), imgl2r()
int		lines, width
int		index

begin
  IM_LEN(im,3) = 3
  lines = IM_LEN(fi, 1)
  width = IM_LEN(fi, 2)
  do index = 1, lines {
    call amovr(Memr[imgl2r(fi,index)],Memr[impl3r(im,index,fplane)], width)
  }
end

#	Save a MWCS system which linearly transforms the x and y
#	ranges into the image size.  The reference point must
#	correspond to the world coordinate's minimum point.
#	This is only for (x,y) transformations.

procedure	dc_wcs(im, name, xmin, xmax,
                       ymin, ymax, xref, yref, mw)	# create MWCS
pointer		im				#I/O image
char		name[ARB]			#I MWCS system name
real		xmin, xmax			#I x range
real		ymin, ymax			#I y range
real		xref, yref			#I reference position
pointer		mw				#O coordinates

double		r[2], w[2], cd[2,2]
pointer		mw_open()

begin
  mw = mw_open(NULL, 2)
  call mw_newsystem(mw, name, 2)
  
  w[1] = xmin
  w[2] = ymin
  r[1] = xref
  r[2] = yref
  cd[1,1] = (xmax-xmin)/IM_LEN(im,1);	cd[2,1] = 0
  cd[1,2] = 0;	cd[2,2] = (ymax-ymin)/IM_LEN(im,2)
  call mw_swtermd (mw, r, w, cd, 2)
  call mw_saveim (mw, im)
end

#	Fill X and Y distortion information in the appropriate
#	planes of the 3d data image.
procedure	dc_build(im, dist, xplane, ct)	# create distortion image
pointer		im				#I image
pointer		dist				#I distortion map
int		xplane				#I x plane
pointer		ct				#I transform

int		yplane

begin
  if (xplane == 1) yplane = 2; else {yplane = 1; xplane = 2}
  call dc_build1(im, VEC_ELEMENT(dist, 1), VEC_ELEMENT(dist, 3), xplane, ct)
  call dc_build1(im, VEC_ELEMENT(dist, 2), VEC_ELEMENT(dist, 4), yplane, ct)
end

#	Fill in distortion correction image from a geotrans
#	distortion record.
procedure	dc_build1(im, s1, s2, plane, ct)# fill in distortion plane
pointer		im				#I/O 3d correction data
pointer		s1, s2				#I distortion data
int		plane				#I correction plane
pointer		ct				#I transform

int		xdim, ydim
pointer		buf, impl3r()
real		gseval(), t
int		x, y
real		in[2], out[2]

begin
  xdim = IM_LEN(im,1)
  ydim = IM_LEN(im,2)
  do y = 1, ydim {
    buf = impl3r(im, y, plane)
    do x = 1, xdim {
      in[1] = x
      in[2] = y
      call mw_ctranr(ct, in, out, 2)
      if (s1 != NULL) t = gseval(s1, out[1], out[2]); else t = 0
      if (s2 != NULL) t = t + gseval(s2, out[1], out[2])
      Memr[buf+x-1] = t
    }
  }
end

#	Create an empty 3d correction data image
procedure	dc_create(image, sz, ptr)	# create distortion image
char		image[ARB]			#I image name
int		sz[2]				#I dimensions
pointer		ptr				#O new image

pointer		immap()
begin
  ptr = immap(image, NEW_IMAGE, 0)
  IM_PIXTYPE(ptr) = TY_REAL
  IM_LEN(ptr,1) = sz[1]
  IM_LEN(ptr,2) = sz[2]
  IM_LEN(ptr,3) = 2
  IM_NDIM(ptr) = 3
end

#	Read a geotrans correction map from a text
#	database.
procedure	dc_load_corr(db, record, correction) # load correction
char		db[ARB]				#I text database
char		record[ARB]			#I record name
pointer		correction			#O correction data

pointer	dt, dtmap()

begin
  dt = dtmap(db, READ_ONLY)
  call ed_rd_corr(dt, record, correction)
  call dtunmap(dt)
end
