# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  dqcload.x
# 	Author	:  akeith
# 	Date	:  04/27/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) dqcload.x -r1.6 04/27/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	02/26/97   mabbott	added support for multiple "time" columns
# 	07/16/93   akeith	fixed any YES/NO TRUE/FALSE problems and removed all the DEBUG stuff
# 	10/30/92   akeith	Only put in debug print statement of function name.
# 	06/26/92   akeith	Now calls dq_mkdict() to create the column name dictionary instead of dq_addstrg().
# 	04/27/92   akeith	date and time created 92/04/27 16:28:22 by akeith
# H-
# 
#

#  Date last modified: 2/26/97

include "dq.h"
include <tbset.h>

#
# dq_cload() deals with the messy business of loading a column. This means
#	getting the column from the table, dq_copen(), adding a new
#	cls to the list, dq_clsadd() and then it stores the dq tls index,
#	stores the column name, gets the rest of the cls structure info
#	adds the column name to the dictionary string and stores the data
#	in a vto.
#
#  MJA notes - I added code to call dq_otbclm instead of dq_copen if
#   tls_ptr is passed in non-NULL.  This occurs when we know what table
#   the column we are after is in, and do not need to search all tables
#   for it.  Because the search is not needed in this case, this also 
#   means that it is okay to load two columns with the same name in 
#   different tables, as long as tls_ptr is non-NULL.  When tls_ptr is
#   NULL, we can never load a given column name a second time, because
#   dq_copen will always find it in the first table in which it occurs,
#   effectively hiding all subsequent occurences.
#
#   A less crucial but useful addition is code I added to disambiguate
#   columns loaded more than once for display or filter generation.  This
#   is done by prepending the table name to the column name with a "." 
#   separator for the second and all subsequent occurences.

int	procedure	dq_cload(dq_ptr, tls_ptr, col_name, str_len, stype)

pointer	dq_ptr					#O dq structure pointer
pointer	tls_ptr					#I tls structure pointer
char	col_name[str_len]			#I table name
int	str_len					#I string length
int	stype					#I sample type

int	tls_indx				# tls structure index
int	cls_indx				# cls structure index

#int	lngth, l_neg				# length and negation
#real	l_low, l_high 				# low, high limits

pointer col_ptr					# table column pointer
pointer cls_ptr					# column list structure pointer
pointer str_ptr					# string pointer

int i
bool exists

int	tbcigi()
int	dq_vtoinit()
pointer dq_clsadd()
int dq_fndclsindx()

begin

  if (tls_ptr != NULL) {
	# must get tls index from tls ptr
	do i = 1, DQ_TLS_LEN(dq_ptr) {
	  if (tls_ptr == DQ_TLS(dq_ptr, i)) break
	}
	if (i > DQ_TLS_LEN(dq_ptr)) {
	  # this is an invalid state - should be impossible
	  # because dq_tbfndtime is not called until after
	  # a table is loaded.
	  return (INDEFI)
	}
	tls_indx = i
	call dq_otbclm(dq_ptr, col_name, col_ptr, tls_indx)
  }
  else call dq_copen(dq_ptr, col_name, col_ptr, tls_indx)

	if(col_ptr == NULL) return(INDEFI)

	cls_indx = INDEFI
	cls_ptr = dq_clsadd(dq_ptr, cls_indx)

    # Store the dq tls index.

	CLS_TLSINDX(cls_ptr) = tls_indx

    # Store the column pointer.

	CLS_COLPTR(cls_ptr) = col_ptr

    # Store the column name.

  exists = (dq_fndclsindx(dq_ptr, col_name, str_len) != INDEFI)
  if (tls_ptr != NULL && exists) {
	# prefix with table name if column already is in list and we have
	# table name available
	call malloc(str_ptr, TLS_TNAME_LEN(tls_ptr)+str_len+1, TY_CHAR)
	CLS_CNAME_PTR(cls_ptr) = str_ptr
	CLS_CNAME_LEN(cls_ptr) = TLS_TNAME_LEN(tls_ptr)+str_len+1
	call strcpy(TLS_TNAME(tls_ptr), CLS_CNAME(cls_ptr), TLS_TNAME_LEN(tls_ptr)+str_len)
	call strcat(".", CLS_CNAME(cls_ptr), TLS_TNAME_LEN(tls_ptr)+str_len)
	call strcat(col_name, CLS_CNAME(cls_ptr), TLS_TNAME_LEN(tls_ptr)+str_len)

	call eprintf(" Column \"%s\" in table \"%s\" has been renamed to \"%s\".\n")
	call pargstr(col_name)
	call pargstr(TLS_TNAME(DQ_TLS(dq_ptr, CLS_TLSINDX(DQ_CLS(dq_ptr, cls_indx)))))
	call pargstr(CLS_CNAME(DQ_CLS(dq_ptr, cls_indx)))

  } else {
	call malloc(str_ptr, str_len+1, TY_CHAR)
	CLS_CNAME_PTR(cls_ptr) = str_ptr
	CLS_CNAME_LEN(cls_ptr) = str_len+1
	call strcpy(col_name, CLS_CNAME(cls_ptr), str_len)
  }

    # Get the rest of the cls structure info.

	CLS_VTOINDX(cls_ptr) = dq_vtoinit(cls_ptr)

	CLS_COLINDX(cls_ptr) = tbcigi(col_ptr, TBL_COL_NUMBER)
	CLS_DTYPE(cls_ptr) = tbcigi(col_ptr, TBL_COL_DATATYPE)
	CLS_STYPE(cls_ptr) = stype

	call dq_mkdict(dq_ptr)

	call dq_mon2vto(cls_ptr)

	return(cls_indx)

end


