# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  dqgtime.x
# 	Author	:  akeith
# 	Date	:  04/27/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) dqgtime.x -r1.9 04/27/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	07/16/93   akeith	fixed any YES/NO TRUE/FALSE problems and removed all the DEBUG stuff
# 	10/30/92   akeith	uses new function dq_freeclsgt() and releases allocated mem. for good time intervals stored in CLS structure.
# 	07/29/92   akeith	takes care zero intervals.
# 	07/27/92   akeith	commented out a debug print statement
# 	07/23/92   akeith	This was completely reworked to handle multi limits.
# 	06/26/92   akeith	Now clears any previously allocated memory to the good time pointers in the cls structures.
# 	04/28/92   akeith	Fixed end time problem, which was putting out the next time point.
# 	04/27/92   akeith	date and time created 92/04/27 16:29:07 by akeith
# H-
# 
#

#  Date last modified: 11/23/93

include "dq.h"

#
# dq_gtime() looks for start and end times which consitute the good time
#	intervals by looking directly at the contents of the ST table.
#

procedure	dq_gtime(cls_ptr)

pointer	cls_ptr					#I/O cls structure pointer

int	i
int	sgt_indx, egt_indx			# start and end good time index

int	first_row				# first row to grab from table
int	last_row				# last row to grab from table
int	nrows					# number of row in table

double	last_time				# old time

real	lo_lim, hi_lim				# low and high limits
int	inclus					# inclusive limits
int	nlim					# number of limits

bool	done

int	start_end
bool	in_limit
bool	last_inlimit

pointer sp					# stack pointer
pointer null_ptr				# null value indicator buffer
pointer z_buff					# column values buffer
pointer t_buff					# time column values buffer

pointer dq_ptr					# dq structure pointer
pointer tls_ptr					# tls structure pointer
pointer tb_ptr					# table pointer
pointer tb_cptr					# table column pointer
pointer tb_tptr					# table time colomn pointer

bool	dq_inlimit()

begin

	nlim = 1
	first_row = 0
	last_row = 0
	sgt_indx = 0
	egt_indx = 0

	done = FALSE
	last_inlimit = FALSE

	dq_ptr = CLS_DQ(cls_ptr)
	call mls_limits(DQ_MLS(dq_ptr), CLS_CNAME(cls_ptr),
			lo_lim, hi_lim, inclus, nlim)

	if(nlim == 0) return

	tls_ptr = CLS_TLSPTR(cls_ptr)
	tb_cptr = CLS_COLPTR(cls_ptr)
	tb_tptr = TLS_IDSPTR(tls_ptr)
	tb_ptr = TLS_TBPTR(tls_ptr)
	nrows = TLS_NROWS(tls_ptr)

	if(nrows == 0) return

	call smark(sp)
	call salloc(z_buff, DQ_APAGE, TY_DOUBLE)
	call salloc(t_buff, DQ_APAGE, TY_DOUBLE)
	call salloc(null_ptr, DQ_APAGE, TY_INT)

#   Clear any previously allocate memory
#    for the good time list.

	call dq_freeclsgt(cls_ptr)

#   Allocate new memory for the
#    good time list.

	call dq_gtclsmalloc(cls_ptr, DQ_MEM_LEN)


#   Go through all the data
#    in the column.

	repeat {

		first_row = last_row + 1
		last_row = first_row + DQ_APAGE - 1

		if(last_row >= nrows) {
			last_row = nrows
			done = TRUE
		}


#   Get the column and the time
#    from the table in pages.


		call tbcgtd(tb_ptr, tb_cptr, Memd[z_buff],
				Memi[null_ptr], first_row, last_row)

		call tbcgtd(tb_ptr, tb_tptr, Memd[t_buff],
				Memi[null_ptr], first_row, last_row)


#   Go through every element in the
#    page and find start and end times.


		do i = 1, last_row - first_row + 1 {

#   Save the time value.

			in_limit = dq_inlimit(  ARRYD(z_buff, i),
					        lo_lim, hi_lim,
						inclus, nlim)


			if(in_limit && !last_inlimit) {

				# save start time

				start_end = 0
				sgt_indx = sgt_indx + 1
				CLS_SGT(cls_ptr, sgt_indx) = ARRYD(t_buff, i)

			} else if(!in_limit && last_inlimit) {

				# save end time

				start_end = 1
				egt_indx = egt_indx + 1
				CLS_EGT(cls_ptr, egt_indx) = last_time

				if(CLS_SGT(cls_ptr, sgt_indx) ==
				   CLS_EGT(cls_ptr, egt_indx) ) {
					sgt_indx = sgt_indx - 1
					egt_indx = egt_indx - 1
				}
			}

			last_inlimit = in_limit
			last_time = ARRYD(t_buff, i)

			if(sgt_indx == CLS_SGT_LEN(cls_ptr)) {
				call dq_gtclsmalloc(cls_ptr, sgt_indx+DQ_MEM_LEN)
			}

		}

	} until(done)

	if(start_end == 0) {
		start_end = 1
		egt_indx = egt_indx + 1
		CLS_EGT(cls_ptr, egt_indx) = last_time

		if(CLS_SGT(cls_ptr, sgt_indx) ==
		   CLS_EGT(cls_ptr, egt_indx) ) {
			sgt_indx = sgt_indx - 1
			egt_indx = egt_indx - 1
		}
	}

	call dq_gtclsmalloc(cls_ptr, sgt_indx)

	call sfree(sp)

end


#
#
#

bool	procedure	dq_inlimit( z, lo, hi, inclus, nlim)

double	z					#I value to check
real	lo[nlim]				#I list of low limits
real	hi[nlim]				#I list of high limits
int	inclus[nlim]				#I list of inclusive limits
int	nlim					#I number of limits

int	i

begin

	if(IS_INDEFD(z)) return(FALSE)

	do i = 1, nlim {
		if(inclus[i] == 1) {
			if( z >= lo[i] && z <= hi[i] ) return(TRUE)
		} else {
			if( z <= lo[i] || z >= hi[i] ) return(TRUE)
		}
	}

	return(FALSE)

end

