# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  dqgtmon.x
# 	Author	:  akeith
# 	Date	:  04/27/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) dqgtmon.x -r1.9 04/27/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	02/26/97   mabbott	added support for multiple time columns
# 	07/16/93   akeith	fixed any YES/NO TRUE/FALSE problems and removed all the DEBUG stuff
# 	12/10/92   akeith	Now contains a free function for the night time intervals.
# 	10/30/92   akeith	created dq_freeclsgt() and uncommented the use of dq_clsdel()
# 	07/29/92   akeith	Now the generation of the day gaps are only done when none have been produced before.
# 	07/29/92   akeith	makes the call to compute the day gaps and merge them into the gt.
# 	06/26/92   akeith	nothing that isn't commented out.
# 	04/27/92   akeith	date and time created 92/04/27 16:29:09 by akeith
# H-
# 
#

#  Date last modified: 2/26/97

include "dq.h"

#
# dq_gtmon() start the process of generating good time intervals by first
#	finding the named monitor and loading it if need be, then generating
#	the good time intervals for the spacific monitor.
#

procedure	dq_gtmon(dq_ptr, mon_name)

pointer	dq_ptr					#I dq structure pointer
char	mon_name[ARB]				#I monitor name

int	str_len					# string length

int	cls_indx				# dq's cls index
pointer	cls_ptr					# cls structure pointer
pointer	tls_ptr					# tls structure pointer

double	time_tol				# time tol

bool	load_flag				# did we load a monitor?

int	strlen()
int	dq_fndclsindx()
int	dq_cload()

begin

	call strlwr(mon_name)

        load_flag = FALSE

	str_len = strlen(mon_name)
	cls_indx = dq_fndclsindx(dq_ptr, mon_name, str_len)

	if(IS_INDEFI(cls_indx)) {

		cls_indx = dq_cload(dq_ptr, NULL, mon_name, str_len, DQ_SAMPLE)

		if(IS_INDEFI(cls_indx)) {

			call fprintf(STDERR, "\n Warning: Unable to find ")
			call fprintf(STDERR, "monitor %s.\n")
			call pargstr(mon_name)
			return

		} else load_flag = TRUE

	}


	cls_ptr = DQ_CLS(dq_ptr, cls_indx)

	call dq_gtime(cls_ptr)

#    Has the day gaps been computed for this table, if not, do that now.

	tls_ptr = CLS_TLSPTR(cls_ptr)

	if(TLS_SNT_PTR(tls_ptr) == NULL  ||
	   int( DQ_GAP(dq_ptr) - TLS_GTGAP(tls_ptr) ) != 0) {

		TLS_GTGAP(tls_ptr) = DQ_GAP(dq_ptr)
		time_tol = DQ_GAP(dq_ptr)*TLS_DELT(tls_ptr)
		call dq_fnddaygap(tls_ptr, time_tol)
		call dq_mergedaygap(dq_ptr, tls_ptr)

	}

#    Now merge the good times.

	call dq_merge(dq_ptr, cls_ptr)

	call dq_freeclsgt(cls_ptr)

	if(load_flag) call dq_clsdel(dq_ptr, cls_indx)

end


#
# dq_freeclsgt() frees the memory of the Column List Structure for
#	the Good Time intervals.
#

procedure	dq_freeclsgt(cls_ptr)

pointer	cls_ptr				#I/O column list structure

begin

	if(cls_ptr == NULL) return

	if(CLS_SGT_PTR(cls_ptr) != NULL)
		call mfree(CLS_SGT_PTR(cls_ptr), TY_DOUBLE)

	CLS_SGT_PTR(cls_ptr) = NULL
	CLS_SGT_LEN(cls_ptr) = 0

	if(CLS_EGT_PTR(cls_ptr) != NULL)
		call mfree(CLS_EGT_PTR(cls_ptr), TY_DOUBLE)

	CLS_EGT_PTR(cls_ptr) = NULL
	CLS_EGT_LEN(cls_ptr) = 0
	
end

#
# dq_freetlsnt() frees the memory of the Table List Structure for
#	the Night Time intervals.
#

procedure	dq_freetlsnt(tls_ptr)

pointer	tls_ptr				#I/O column list structure

begin

	if(tls_ptr == NULL) return

	if(TLS_SNT_PTR(tls_ptr) != NULL)
		call mfree(TLS_SNT_PTR(tls_ptr), TY_DOUBLE)

	TLS_SNT_PTR(tls_ptr) = NULL
	TLS_SNT_LEN(tls_ptr) = 0

	if(TLS_ENT_PTR(tls_ptr) != NULL)
		call mfree(TLS_ENT_PTR(tls_ptr), TY_DOUBLE)

	TLS_ENT_PTR(tls_ptr) = NULL
	TLS_ENT_LEN(tls_ptr) = 0
	
end

