# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_euvextract.x
# 	Author	:  akeith
# 	Date	:  04/30/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_euvextract.x -r1.17 04/30/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	03/06/98   mabbott	bug fix
# 	03/06/98   mabbott	added dispaxis parameter
# 	03/06/98   mabbott	added default for missing DISPAXIS
# 	09/20/94   akeith	removed unused var
# 	09/14/94   akeith	now uses de_bands()
# 	09/08/94   akeith	upgrade
# 	07/14/94   akeith	added shift mechanism
# 	04/04/94   akeith	the new optimal extraction
# 	04/04/94   akeith	new optimal extraction
# 	04/04/94   akeith	cleared for new euvextract
# 	04/27/93   ericco	updated parameter names
# 	09/28/92   gareth	moved image operations to file ex_head_wcs
# 	09/23/92   gareth	rewrite of erics code. removed w0 and wn parameters
# 	05/14/92   ericco	reduced WCSDIM if it exists
# 	05/14/92   ericco	added documentation
# 	04/30/92   akeith	date and time created 92/04/30 13:22:13 by akeith
# H-
# 
#


#	Version: @(#)t_euvextract.x	1.17 3/6/98 UCB CEA EGO

include	<imhdr.h>
include	<fset.h>
include	<eoxset.h>

define	R		Memr[($1)]
define	C		Memc[($1)]

#
# t_euvextract()
#

procedure t_euvextract()

pointer	sp			#
pointer	str			#
pointer	input			#
pointer	output			#
pointer cntr_ap			#
pointer spec_ap			#
pointer bkg_ap			#
pointer	spec			#
pointer	svar			#
pointer	bkg			#
pointer	bvar			#

pointer in, immap()		#
pointer	out, eox_mkout()	#

int	ncol			#
int	bkg_type		#
int	bkg_nline		#
int	spc_nline		#
int	disp_axis		#
int	img_axis		#
int	spc_band		#
int	sig_band		#
int	fltr_sz			#
int	icenter			#
int	radius			#
int	center			#

real	thres			#
int	eox_center()		#

real	clgetr()		#
bool	clgetb()		#
int	clgeti()		#
int	clgwrd()		#
int	imgeti()		#
int imaccf() 

begin

	call smark(sp)
	call salloc(str, SZ_LINE, TY_CHAR)
	call salloc(input, SZ_LINE, TY_CHAR)
	call salloc(output, SZ_LINE, TY_CHAR)
	call salloc(cntr_ap, SZ_LINE, TY_CHAR)
	call salloc(spec_ap, SZ_LINE, TY_CHAR)
	call salloc(bkg_ap, SZ_LINE, TY_CHAR)

#  Read primary input parameters.

	if(clgetb("verbose"))
		call fseti(STDOUT, F_FLUSHNL, YES)

	bkg_type = clgwrd("background", C(str), SZ_FNAME, BKG_DICT)

#  Open input image.

	call clgstr("input", C(input), SZ_LINE)
	in = immap(C(input), READ_ONLY, 0)

#  Open output image.

	call clgstr("output", C(output), SZ_LINE)
	if(bkg_type != BNONE)
		out = eox_mkout(C(output), in, 3)
	else 	out = eox_mkout(C(output), in, 2)

#  Get the dispersion axis and the imaging axis.

    if (imaccf(in, "DISPAXIS") == YES) {
	   disp_axis = imgeti(in, "DISPAXIS")
	} else {
       disp_axis = clgeti("dispaxis")
	}

	if(disp_axis == 1) img_axis = 2
	else img_axis = 1

#  Get the data bands.

	call de_bands(in, spc_band, sig_band)
	if(spc_band == -1) {
		spc_band = 1
		call eprintf("\nWarning: No data BANDID card.")
		call eprintf("\n         Assuming a data band of 1.\n")
	}

	ncol = IM_LEN(in, disp_axis)
	call salloc(spec, ncol, TY_REAL)
	call salloc(svar, ncol, TY_REAL)
	call salloc(bkg, ncol, TY_REAL)
	call salloc(bvar, ncol, TY_REAL)

#	   For fitting, we produce separate average backgrounds for
#	    the apertures(ap) the user specified on either side of the aperture.
#	    We then fit along these average backgrounds. For each column, we fit
#	    across the spectrum to give us the background value at any line in
#	    the aperture. (maybe)

#  Get the spectral center.

	center = clgeti("center")
	if(IS_INDEFI(center)) center = IM_LEN(in, img_axis)/2
	if(clgetb("seekspec")) {
		thres = clgetr("thres")
		call clgstr("cntr_ap", C(cntr_ap), SZ_LINE)
		radius = clgeti("radius")
		icenter = eox_center(in, -1, C(cntr_ap), center,
					radius, thres, disp_axis, spc_band)
		if(!IS_INDEFI(icenter))
			center = icenter
	}

#  Get the backgound.

	call clgstr("bkg_ap", C(bkg_ap), SZ_LINE)

	if(bkg_type == BFIT) {
	} else if(bkg_type == BNONE) {
	} else {

		call eox_ladd(in, center, C(bkg_ap), spc_band, disp_axis,
						R(bkg), ncol, bkg_nline)
		call adivkr(R(bkg), real(bkg_nline), R(bkg), ncol)

		if(sig_band == -1) {
			call adivkr(R(bkg), real(bkg_nline), R(bvar), ncol)
		} else {
			call eox_lsqradd(in, center, C(bkg_ap),
				sig_band, disp_axis, R(bvar), ncol, bkg_nline)
			call adivkr(R(bvar), real(bkg_nline*bkg_nline), R(bvar), ncol)
		}
	}

	if(bkg_type == BSMOOTH) {
		fltr_sz = clgeti("bkg_fltr_sz")
		call eox_smooth_bkg(R(bkg), R(bvar), ncol, fltr_sz)
	}


#  Now do ordinary extraction.  Optimal extraction requires ordinary 
#    extraction for an intial guess.

	call clgstr("spec_ap", C(spec_ap), SZ_LINE)
	call eox_ladd(in, center, C(spec_ap), spc_band, disp_axis,
						R(spec), ncol, spc_nline)

	if(sig_band == -1) {
		call amovr(R(spec), R(svar), ncol)
	} else {
		call eox_lsqradd(in, center, C(spec_ap),
				sig_band, disp_axis, R(svar), ncol, spc_nline)
	}
	call awsur(R(svar), R(bvar), R(svar), ncol, 1.0, real(spc_nline**2))

	if(bkg_type != BNONE)
		call awsur(R(spec), R(bkg), R(spec), ncol, 1.0, -real(spc_nline))


#  Write out the output image.

	call eox_mkheaders(in, out, C(spec_ap), C(bkg_ap), center,
						ENORMAL, bkg_type, fltr_sz)

	call eox_out(out, bkg_type, ENORMAL, R(bkg), R(spec), R(svar), ncol, spc_nline)

	call imunmap(out)

#   Clean up.

	call imunmap(in)

	call sfree(sp)

end

