# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_gettm.x
# 	Author	:  ericco
# 	Date	:  10/29/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_gettm.x -r1.30 10/29/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/24/96   tnk	increased param space
# 	10/21/96   tnk	changed deltime to timedel
# 	04/24/96   gareth	change RATE_%03d to RATE%-3d  for FITS compatiblity
# 	09/13/95   tnk	set max. table user params
# 	08/24/94   tnk	fixed call to eu_free_strings (added nlines)
# 	08/02/94   gareth	changed reals to doubles to stop headers rounding up
# 	08/01/94   gareth	added code to place the header cards in the tables as the right data type
# 	07/26/94   gareth	removed last fix ; added check for zero lines
# 	07/25/94   gareth	fixed error where the program alwasy prompted for text database file
# 	04/26/94   ericco	removed debug statements
# 	04/26/94   ericco	fixed env str for loop condition
# 	04/26/94   ericco	gets environment from tdb file
# 	02/26/94   ericco	added environment passing support
# 	11/20/93   ericco	added support of short columsn
# 	11/20/93   ericco	fixed fp check on close command
# 	08/02/93   gareth	realloced headers and initalised GT_ macros
# 	07/16/93   akeith	changed bp to TRUE/FALSE
# 	04/16/93   ericco	additional error handling during reads
# 	03/29/93   ericco	added header keyword handling
# 	10/27/92   ericco	extracted hex reading routines
# 	06/29/92   ericco	generalize and corrected argument processing
# 	06/29/92   ericco	moved header parameter definition to below the table create call
# 	06/29/92   akeith	Now accepts from egoport the delta time and time column name and places 
# them into header parameters of the st tables.
# 	06/26/92   ericco	fixed rate memory errors
# 	06/24/92   akeith	just changed the spacing
# 	04/30/92   ericco	changed parameter names
# 	12/13/91   ericco	fixed INDEF values on QPOE output format
# 	10/29/91   ericco	date and time created 91/10/29 16:35:56 by ericco
# H-
# 
#

#	Revised: 10/24/96 @(#)t_gettm.x	1.30 
#
#	Get telemetry from a service

include	<mach.h>
include <lexnum.h>
include "gettm.h"
include <tbset.h>

define MAX_USR_PARS   120       # max. number of user parameters in table

define	ERR_NONE	0
define	ERR_STAR	1
define	ERR_OPEN	2
define	ERR_MODE	3
define	ERR_PROC	4
define	ERR_DESC	5
define	ERR_SUPP	6
define	ERR_DETS	7
define	ERR_CONN	8
define	ERR_ENV		9

#	Add a column to the current table

procedure	tm_addcolumn(ptr, name, type)	 # add column

pointer		ptr			#I table set
char		name[ARB]		#I column name
char		type[ARB]		#I column type

begin
  call tm_addcol(ptr, name, type, "", "0")
end


procedure	tm_addcol(ptr, name, type, units, str_rate)	 # add column

pointer		ptr			#I table set
char		name[ARB]		#I column name
char		type[ARB]		#I column type
char		units[ARB]		#I column units
char		str_rate[ARB]		#I string column rate pred.

char		typ
pointer		sp
#bool		streq()
int		strlen()
int		len

begin

	len = strlen(name)
	call smark(sp)

	GT_CURCOLS(ptr) = GT_CURCOLS(ptr) + 1

	if (GT_CURCOLS(ptr) == 1) {

		GT_BYTES(ptr) = 0
		GT_LENGTH(ptr) = 0
		GT_TYPIP(ptr) = 1
		GT_RATES(ptr) = 0

		if (GT_EVENTP(ptr)) {
			call sprintf(GT_TYPEND(ptr), ARB, "{")
			GT_TYPIP(ptr) = 2
		} else {
			call sprintf(GT_TYPEND(ptr), ARB, "")
		}

		call eu_malloc(GT_CURTYP(ptr), 1, TY_INT)
		call eu_malloc(GT_CURNAM(ptr), 1, TY_INT)
		call eu_malloc(GT_CURUNT(ptr), 1, TY_INT)
		call eu_malloc(GT_CURRAT(ptr), 1, TY_INT)
	} else {
		call eu_realloc(GT_CURTYP(ptr), GT_CURCOLS(ptr), TY_INT)
		call eu_realloc(GT_CURNAM(ptr), GT_CURCOLS(ptr), TY_INT)
		call eu_realloc(GT_CURUNT(ptr), GT_CURCOLS(ptr), TY_INT)
		call eu_realloc(GT_CURRAT(ptr), GT_CURCOLS(ptr), TY_INT)
	}

	call eu_malloc(GT_THISNAMP(ptr), len, TY_CHAR)
	call strcpy(name, GT_THISNAME(ptr), ARB)

	len = strlen(units)
	call eu_malloc(GT_THISUNTP(ptr), len, TY_CHAR)
	call strcpy(units, GT_THISUNIT(ptr), ARB)

        GT_THISRATE(ptr) = 0
        if (str_rate[1] != '0') {
          GT_THISRATE(ptr) = 1
	  GT_RATES(ptr) = GT_RATES(ptr) + 1
	}

	typ = type[1]

#  if (GT_EVENTP(ptr) && GT_CURCOLS(ptr) <= 2) {
#    typ = 's'
#  } else if (streq(type, "integer\n")) {
##    typ = 'i'
#    typ = 's'
#  } else {
#    typ = 'r'
#  }

	switch (typ) {
		case 's' :	GT_THISTYPE(ptr) = TY_SHORT
		case 'i' :	GT_THISTYPE(ptr) = GT_ITYPE(ptr)
		case 'r' :	GT_THISTYPE(ptr) = TY_REAL
		case 'd' :	GT_THISTYPE(ptr) = TY_DOUBLE
		case 'l' :	GT_THISTYPE(ptr) = TY_LONG
		default  :	# error condition
	}

end

procedure	tm_puthdrs(tp, keys, hdrs, nhdrs)
pointer		tp		#I/O table pointer
pointer		keys[nhdrs]	#I header keywords
pointer		hdrs[nhdrs]	#I formated hdrs
int		nhdrs		#I number of headers

int		i,ret,valuei,ip,numch
int		strlen(),lexnum()
double		valued

begin
  do i = 1, nhdrs {
    ip = 1
    numch = strlen(Memc[hdrs[i]])
    ret = lexnum(Memc[hdrs[i]],ip,numch)
    if ( numch == strlen(Memc[hdrs[i]]) ) {
     	ip = 1
	switch ( ret ) {
	  case LEX_NONNUM :  
	    call tbhadt(tp, Memc[keys[i]], Memc[hdrs[i]])
	  case LEX_DECIMAL :
	    call ctoi(Memc[hdrs[i]],ip,valuei)
	    call tbhadi(tp, Memc[keys[i]], valuei)
	  case LEX_REAL :
	    call ctod(Memc[hdrs[i]],ip,valued)
	    call tbhadd(tp, Memc[keys[i]], valued) 
         default :
	    call tbhadt(tp, Memc[keys[i]], Memc[hdrs[i]])
	}
    } else {
	 call tbhadt(tp, Memc[keys[i]], Memc[hdrs[i]])
    }
  }
end

procedure	tm_addhdr(ptr, key, argv, argc)
pointer		ptr		#I/O table set
char		key[ARB]	#I header keyword
pointer		argv[argc]	#I arguments
int		argc		#I number of arguments

pointer		sp
pointer		buf
int		len
int		strlen()
int		i

begin
  call smark(sp)
  len = 1
  do i = 1, argc {
    len = len + strlen(Memc[argv[i]]) + 1
  }
  call malloc(buf, len, TY_CHAR)
  Memc[buf] = EOS
  do i = 1, argc {
    call strcat(Memc[argv[i]], Memc[buf], ARB)
    if (i < argc)
      call strcat(" ", Memc[buf], ARB)
  }

  GT_HDRS(ptr) = GT_HDRS(ptr) + 1
  call eu_realloc(GT_CURHDR(ptr), GT_HDRS(ptr), TY_INT)
  call eu_realloc(GT_CURKEY(ptr), GT_HDRS(ptr), TY_INT)
  GT_HDR(ptr,GT_HDRS(ptr)) = buf

  len = strlen(key)
  call malloc(buf, len, TY_CHAR)
  call strcpy(key, Memc[buf], ARB)
  GT_KEY(ptr,GT_HDRS(ptr)) = buf

  call sfree(sp)
end

#	Set up buffers for the current table, assuming
#	that all columns have been defined

procedure	tmsetbuf(ptr)	# add buffers

pointer		ptr		#I/O table set

int		size
int		i
int		si
int		order[5]
data		order / TY_DOUBLE, TY_LONG, TY_INT, TY_REAL, TY_SHORT /
int		coltyp[5]
data		coltyp/ TY_DOUBLE, TY_INT, TY_INT, TY_REAL, TY_SHORT /
int		sizes[5]
data		sizes / SZ_DOUBLE, SZ_INT, SZ_INT, SZ_REAL, SZ_SHORT /
char		letter[4] # wierd isn't it
data		letter/       'd',     'i',    'i',     'r',      's' /
char		value[10]
string		format "%g%d%d%f%d"

pointer		p
int		o
int		ncol
pointer		tmbuf
int		strlen()
char		table[10]
char		typ[10]
pointer		tp, cp, cps
pointer		tbtopn()

int		axlen[2]
pointer		qp_open()
int		ctyp
char		keyword[20]

begin

	ncol = GT_CURCOLS(ptr)
	call sprintf(Memc[GT_PREFIX(ptr)+GT_PRELEN(ptr)], 10, "%d")
	call pargi(GT_NTABLES(ptr)-1)

	if (!GT_EVENTP(ptr)) {
		tp = tbtopn(Memc[GT_PREFIX(ptr)], NEW_FILE, 0)

		GT_TABTP(ptr, GT_NTABLES(ptr)) = tp
		call eu_malloc(cps, ncol, TY_INT)
		GT_TABCP(ptr, GT_NTABLES(ptr)) = cps

		# set max no. of user params
		call tbpset (tp, TBL_MAXPAR, MAX_USR_PARS)

	} else {
		GT_QP(ptr) = qp_open(Memc[GT_PREFIX(ptr)], NEW_FILE, 0)
	}

	size = 0
	do o = 1, 5 {
		do i = 1, ncol {

			ctyp = GT_A_TYPE(ptr,i)

			if (GT_EVENTP(ptr) && i <= 2) ctyp = TY_SHORT

			if (ctyp == order[o]) {

				if (GT_EVENTP(ptr)) {

					call sprintf(GT_TYPEND(ptr), ARB, "%c")
					call pargc(letter[o])

					call sprintf(value, 10, "%c%d")
					call pargc(letter[o])
					call pargi(size*SZB_CHAR)
					call qp_add_macro(GT_QP(ptr),
						GT_A_NAME(ptr,i), value)
				} else {

					call sprintf(table, 10, "Table %d")
					call pargi(GT_NTABLES(ptr))
					call sprintf(typ, 10, "%c")
					call pargc(letter[o])
#					call qp_addf(GT_QP(ptr), GT_A_NAME(ptr,i),
#							typ, 0, table, 0)
					call strcpy(format[2*o-1], typ, 2)
					typ[3] = EOS
					call tbcdef(tp, cp, GT_A_NAME(ptr,i),
						    GT_A_UNIT(ptr,i), typ, coltyp[o], 1, 1)
					Memi[cps+i-1] = cp
				}

				size = size + sizes[o]
	
				GT_TYPIP(ptr) = strlen(GT_TYPE(ptr,1))+1

				if (GT_EVENTP(ptr)) {

					switch (i) {
					case 1 : call sprintf(GT_TYPEND(ptr), ARB, ":x");
					case 2 : call sprintf(GT_TYPEND(ptr), ARB, ":y");
					}

				} else call sprintf(GT_TYPEND(ptr), ARB, GT_A_NAME(ptr,i))

				GT_TYPIP(ptr) = strlen(GT_TYPE(ptr,1))+1
				call sprintf(GT_TYPEND(ptr), ARB, ",");
				GT_TYPIP(ptr) = strlen(GT_TYPE(ptr,1))+1
			}
		}
	}

	GT_TYPIP(ptr) = GT_TYPIP(ptr) - 1

	if (GT_EVENTP(ptr)) call sprintf(GT_TYPEND(ptr), ARB, "}");
	else call sprintf(GT_TYPEND(ptr), ARB, "");

	size = (size + SZ_LONG - 1) / SZ_LONG
	call eu_malloc (p, ncol, TY_INT)
	GT_TABBUF(ptr,GT_NTABLES(ptr)) = p
	GT_TABSIZ(ptr,GT_NTABLES(ptr)) = size
	call eu_malloc (tmbuf, size, TY_LONG)
	GT_TMBUF(ptr,GT_NTABLES(ptr)) = tmbuf

	si = P2S(tmbuf)
	do o = 1, 5 {
		do i = 1, ncol {

			ctyp = GT_A_TYPE(ptr,i)
			if (GT_EVENTP(ptr) && i <= 2) ctyp = TY_SHORT

#				convert to appropriate index
#			      round up to nearest size boundary
#		   shouldn't do anything if correctly ordered sizes are used
#			   add one size unit, then convert to shorts

			if (ctyp == order[o]) {
	Memi[p+i-1] = ((si-1) + sizes[o] - 1) * SZ_SHORT / sizes[o] + 1
	si = (Memi[p+i-1] - 1 + 1) * sizes[o] / SZ_SHORT + 1
			}

		}
	}
	GT_CURCOLS(ptr) = 0

	if (GT_EVENTP(ptr)) {
		axlen[1] = GT_QP_X(ptr)
		axlen[2] = GT_QP_Y(ptr)
		call euqpst(GT_QP(ptr), axlen, GT_TYPE(ptr,1), GT_QPIO(ptr))
	} else {
		call tbtcre(tp) 

		call tm_puthdrs(tp, Memi[GT_CURKEY(ptr)],
                                Memi[GT_CURHDR(ptr)], GT_HDRS(ptr))
                o = 1
                do i = 1, ncol {
                  if (GT_A_RATE(ptr,i) != 0) {
                    call sprintf(keyword, 20, "RATE%-3d")
                    call pargi(o)
                    o = o + 1
                    call tbhadt(tp, keyword, GT_A_NAME(ptr,i))
                  }
                }

		if (GT_TABTIMECOLP(ptr) != NULL ) {
                    call tbhadt(tp, "time", GT_TABTIMECOL(ptr))
		}

		if (!IS_INDEFR(GT_TABDELT(ptr))) {
                    call tbhadr(tp, "timedel", GT_TABDELT(ptr))
		}

#		call eutabst(qp, GT_NTABLES(ptr), GT_TYPE(ptr,1))
	}

			# restore prefix

	call sprintf(Memc[GT_PREFIX(ptr)+GT_PRELEN(ptr)], 10, "")

end


procedure	eu_free_strings(ptrs, len)
pointer		ptrs		#I/O pointers to strings
int		len		#I number of pointers

int		i

begin

  do i = 1, len
    call eu_free(Memi[ptrs+i-1], TY_CHAR)

  call eu_free(ptrs, TY_INT)
  ptrs = 0
end

#	Finish off the current table definition

procedure	tm_finishcolumn(ptr) # finished adding columns

pointer	ptr		#I/O table set

#pointer		qp

begin
	GT_NTABLES(ptr) = GT_NTABLES(ptr) + 1

	if (GT_NTABLES(ptr) == 1) {
		call eu_malloc (GT_TABCOLS(ptr), 1, TY_INT)
		call eu_malloc (GT_TABTYPS(ptr), 1, TY_INT)
		call eu_malloc (GT_TABBUFS(ptr), 1, TY_INT)
		call eu_malloc (GT_TMBUFS(ptr), 1, TY_INT)
		call eu_malloc (GT_TMCNTS(ptr), 1, TY_INT)
		call eu_malloc (GT_TABNAMS(ptr), 1, TY_INT)
		call eu_malloc (GT_TABSIZS(ptr), 1, TY_INT)
		call eu_malloc (GT_TABTPS(ptr), 1, TY_INT)
		call eu_malloc (GT_TABCPS(ptr), 1, TY_INT)
		call eu_malloc (GT_TABRATS(ptr), 1, TY_INT)
		call eu_malloc (GT_TABUNTS(ptr), 1, TY_INT)
	} else {
		call eu_realloc (GT_TABCOLS(ptr), GT_NTABLES(ptr), TY_INT)
		call eu_realloc (GT_TABTYPS(ptr), GT_NTABLES(ptr), TY_INT)
		call eu_realloc (GT_TABBUFS(ptr), GT_NTABLES(ptr), TY_INT)
		call eu_realloc (GT_TMBUFS(ptr), GT_NTABLES(ptr), TY_INT)
		call eu_realloc (GT_TMCNTS(ptr), GT_NTABLES(ptr), TY_INT)
		call eu_realloc (GT_TABNAMS(ptr), GT_NTABLES(ptr), TY_INT)
		call eu_realloc (GT_TABSIZS(ptr), GT_NTABLES(ptr), TY_INT)
		call eu_realloc (GT_TABTPS(ptr), GT_NTABLES(ptr), TY_INT)
		call eu_realloc (GT_TABCPS(ptr), GT_NTABLES(ptr), TY_INT)
		call eu_realloc (GT_TABRATS(ptr), GT_NTABLES(ptr), TY_INT)
		call eu_realloc (GT_TABUNTS(ptr), GT_NTABLES(ptr), TY_INT)
	}

	Memi[GT_TMCNTS(ptr)+GT_NTABLES(ptr)-1] = 0
	GT_TABCOL(ptr,GT_NTABLES(ptr)) = GT_CURCOLS(ptr)
	GT_TABRAT(ptr,GT_NTABLES(ptr)) = GT_CURRAT(ptr)
	GT_TABUNT(ptr,GT_NTABLES(ptr)) = GT_CURUNT(ptr)
	GT_TABTYP(ptr,GT_NTABLES(ptr)) = GT_CURTYP(ptr)
	GT_TABNAM(ptr,GT_NTABLES(ptr)) = GT_CURNAM(ptr)
	GT_TABTP(ptr, GT_NTABLES(ptr)) = 0
	GT_TABCP(ptr, GT_NTABLES(ptr)) = 0

	if (GT_CURCOLS(ptr) != 0) call tmsetbuf(ptr)

	GT_CURCOLS(ptr) = 0
	GT_CURTYP(ptr) = 0
	GT_CURNAM(ptr) = 0
	GT_CURRAT(ptr) = 0
	GT_CURUNT(ptr) = 0
	if (GT_CURHDR(ptr) != 0) {
	  call eu_free_strings(GT_CURHDR(ptr),GT_HDRS(ptr))
	  call eu_free_strings(GT_CURKEY(ptr),GT_HDRS(ptr))
          GT_CURHDR(ptr) = 0
          GT_CURKEY(ptr) = 0
        }
        GT_HDRS(ptr) = 0

	call eu_free(GT_TABTIMECOLP(ptr), TY_CHAR)
	GT_TABTIMECOLP(ptr) = NULL
	GT_TABDELT(ptr) = INDEFR
end

#	Echo server messages

int	procedure tmecho(res, ptr) # echo line

pointer	res			#I input line
pointer	ptr			#I table set

begin

	call printf("%s")
	call pargstr(Memc[res])
	return 0

end

#	Ignore server messages

int	procedure tmnoop(res, ptr) # echo line

pointer	res			#I input line
pointer	ptr			#I table set

begin
	return 0
end

#	Write server messages to a text database

int	procedure tmwhat(res, ptr) # tdb line

pointer	res			#I input line
pointer	ptr			#I table set

begin

	call dtput(GT_DT(ptr), "%s")
	call pargstr(Memc[res])
	return 0

end

#	Add supplementary files to observation

int	procedure tmsupp(res, ptr) # add supplement

pointer	res			#I input line
pointer	ptr			#I table set

int	fd
int	open()
errchk	open()

begin

	fd = GT_SUPP(ptr)

	if (Memc[res] == 'S') {

		call gt_end_supp(ptr)

		iferr (fd = open(Memc[res+1], NEW_FILE, TEXT_FILE)) {
			fd = 0
		}

		GT_SUPP(ptr) = fd
	}

	if (fd == 0) return 0

	if (Memc[res] != 'X') return 0

	call fprintf(fd, "%s")
	call pargstr(Memc[res+1])

	return 0

end

#	Finish off last supplement file

procedure	gt_end_supp(ptr) # finish supplement

pointer		ptr		#I table set

begin

	if (GT_SUPP(ptr) != 0) call close(GT_SUPP(ptr))
	GT_SUPP(ptr) = 0

end

#
#	According to gprof, most of the time is spent in
#	read under eusrcv().  This could be optimized by
#	reading in all the binary data for a row at once.
#	Currently, only the current data item in the
#	current row is read in on a single read call.
#
#	Process table input stream.

int	procedure tmprocess(res, ptr, fp) # process table data

pointer	res			#I input line
pointer	ptr			#I table set
int	fp			#I input descriptor

int	ac
pointer	av
pointer	av2[2]

int	ct
int	ip
pointer	sp
double	dval
int	ival
bool	streq()
pointer	typs
int	i
pointer	bufs
pointer	ca

int	recn
pointer	tp, cps
bool	bp
int	eusrcv(), euserr()

begin
  call smark(sp)

  if (GT_CURTAB(ptr) == -1) {
    ac = 2
    call eupart(res, ac, av2)
    if (streq(Memc[av2[1]], "Table")) {
      ip = 1
      call ctoi(Memc[av2[2]], ip, GT_CURTAB(ptr))
    }
  } else {
    if (GT_MODE(ptr) == MODE_BINARY) bp = TRUE; else bp = FALSE

    ct = GT_CURTAB(ptr)+1
    ac = GT_TABCOL(ptr,ct)

    call salloc(av, ac, TY_INT)
    call eupart(res, ac, Memi[av])

    if (streq(Memc[Memi[av]], "TableEnd")) {
      GT_CURTAB(ptr) = -1
    } else {		# row of current table data
      if (Memc[Memi[av]] != 'B') {
	bp = FALSE
      } else {
	ac = GT_TABCOL(ptr,ct)
      }
      if (bp) call eusmod(fp) # switch server to binary mode

      tp = GT_TABTP(ptr,ct)
      cps = GT_TABCP(ptr,ct)
      typs = GT_TABTYP(ptr,ct)
      bufs = GT_TABBUF(ptr,ct)
      recn = Memi[GT_TMCNTS(ptr)+ct-1] + 1
      Memi[GT_TMCNTS(ptr)+ct-1] = recn
      do i = 1, ac {
	ip = 1
	ca = Memi[av]

	switch(Memi[typs]) {
	case TY_SHORT :
		if (bp) {
			if (eusrcv(fp, Mems[Memi[bufs]], SZ_SHORT) != 0) {
			  GT_ERRNO(ptr) = euserr()
			  Mems[Memi[bufs]] = GT_SNULL(ptr)
			  break;
			}

			if (Mems[Memi[bufs]] == GT_SNULL(ptr)) {
				ival = INDEFI
				Mems[Memi[bufs]] = INDEFS
			} else {
				ival = Mems[Memi[bufs]]
			}

		} else {
			if (Memc[ca] == '*') {
				ival = INDEFI
				Mems[Memi[bufs]] = INDEFS
			} else {
				call ctoi(Memc[ca], ip, ival)
				Mems[Memi[bufs]] = ival
			}
		}

	case TY_INT :
	if (bp) {
		if (eusrcv(fp, Memi[Memi[bufs]], SZ_INT) != 0) {
		  GT_ERRNO(ptr) = euserr()
		  Mems[Memi[bufs]] = GT_INULL(ptr)
		  break;
		}

		if (Memi[Memi[bufs]] == GT_INULL(ptr)) {
			ival = INDEFI
			Memi[Memi[bufs]] = INDEFI
		} else {
			ival = Memi[Memi[bufs]]
		}
	} else {
		if (Memc[ca] == '*') {
			ival = INDEFI
			Memi[Memi[bufs]] = INDEFI
		} else {
			call ctoi(Memc[ca], ip, ival)
			Memi[Memi[bufs]] = ival
		}
	}

	case TY_LONG :
	if (bp) {
		if (eusrcv(fp, Meml[Memi[bufs]], SZ_LONG) != 0) {
		  GT_ERRNO(ptr) = euserr()
		  Mems[Memi[bufs]] = GT_LNULL(ptr)
		  break;
		}

		if (Meml[Memi[bufs]] == GT_LNULL(ptr)) {
			ival = INDEFI
			Meml[Memi[bufs]] = INDEFL
		} else {
			ival = Meml[Memi[bufs]]
		}

	} else {

		if (Memc[ca] == '*') {
			ival = INDEFI
			Meml[Memi[bufs]] = INDEFL
		} else {
			call ctoi(Memc[ca], ip, ival)
			Meml[Memi[bufs]] = ival
		}
	}

	case TY_DOUBLE :
	if (bp) {
		if (eusrcv(fp, Memd[Memi[bufs]], SZ_DOUBLE) != 0) {
		  GT_ERRNO(ptr) = euserr()
		  Mems[Memi[bufs]] = GT_DNULL(ptr)
		  break;
		}

		if (Memd[Memi[bufs]] == GT_DNULL(ptr)) {
			dval = INDEFD
			Memd[Memi[bufs]] = INDEFD
		} else {
			dval = Memd[Memi[bufs]]
		}
	} else {
		if (Memc[ca] == '*') {
			dval = INDEFD
			Memd[Memi[bufs]] = INDEFD
		} else {
			call ctod(Memc[ca], ip, dval)
			Memd[Memi[bufs]] = dval
		}
	}

	case TY_REAL :
	if (bp) {
		if (eusrcv(fp, Memr[Memi[bufs]], SZ_REAL) != 0) {
		  GT_ERRNO(ptr) = euserr()
		  Mems[Memi[bufs]] = GT_RNULL(ptr)
		  break;
		}

		if (Memr[Memi[bufs]] == GT_RNULL(ptr)) {
			dval = INDEFD
			Memr[Memi[bufs]] = INDEFR
		} else {
			dval = Memd[Memi[bufs]]
		}
	} else {
		if (Memc[ca] == '*') {
			dval = INDEFD
			Memr[Memi[bufs]] = INDEFR
		} else {
			call ctod(Memc[ca], ip, dval)
			Memr[Memi[bufs]] = dval
		}
	}

	default :	# error condition
	}

	if (!GT_QPOE_TABP(ptr,ct)) {

		switch(Memi[typs]) {

		case TY_SHORT, TY_INT, TY_LONG :
#			call qp_write(GT_QP(ptr), GT_A_TNAME(ptr,ct,i),
#					ival, 1, recn, "i")
		  	if (GT_ERRNO(ptr) == 0)
			  call tbrpti(tp, Memi[cps+i-1], ival, 1, recn)
		case TY_DOUBLE, TY_REAL :
#			call qp_write(GT_QP(ptr), GT_A_TNAME(ptr,ct,i),
#					dval, 1, recn, "d")
		  	if (GT_ERRNO(ptr) == 0)
			  call tbrptd(tp, Memi[cps+i-1], dval, 1, recn)
		}
	}

	typs = typs + 1
	av = av + 1
	bufs = bufs + 1

			}

			if (GT_QPOE_TABP(ptr,ct)) {
				call qpio_putevents (GT_QPIO(ptr),
					P2S(GT_TMBUF(ptr,ct)), 1)
			}

			if (bp) call eusmod(fp) # switch server to ascii mode

		}

	}

	call sfree(sp)
	if (GT_ERRNO(ptr) != 0) return -1;
	return 0

end

#	Toggle input mode

procedure	tmmode(ptr) # toggle input mode

pointer		ptr		#I table set

begin

	if (GT_MODE(ptr) == MODE_ASCII) {
		GT_MODE(ptr) = MODE_BINARY
	} else {
		GT_MODE(ptr) = MODE_ASCII
	}

end



#	Process the open server command

int	procedure tmopen(res, ptr) # open a table set

pointer	res			#I input line
pointer	ptr			#I table set

int	ip
int	argc
pointer	argv[25]

int	len
int	strlen()
bool	streq()
int	vsz
short	sval
long	lval

begin

	argc = 25

	call eupart(res, argc, argv)

	if (GT_CURTAB(ptr) == -1) {

		if (streq(Memc[argv[1]], "Table")) {
			ip = 1
			call ctoi(Memc[argv[2]], ip, GT_CURTAB(ptr))
			GT_CURCOLS(ptr) = 0
			if(argc >= 3) {
				ip = 1
				call ctor(Memc[argv[3]], ip, GT_TABDELT(ptr))
			}
			if(argc >= 4) {
				len = strlen(Memc[argv[4]])
				call eu_malloc(GT_TABTIMECOLP(ptr), len+1, TY_CHAR)
				call strcpy(Memc[argv[4]], GT_TABTIMECOL(ptr), ARB)
			}
		} else if (streq(Memc[argv[1]], "LongNull")) {
			ip = 1
			call readhexl(Memc[argv[3]], GT_LNULL(ptr))
		} else if (streq(Memc[argv[1]], "IntegerNull")) {

			ip = 1
			call ctoi(Memc[argv[2]], ip, vsz)

			if (vsz * SZB_CHAR == SZ_LONG) {
				call readhexl(Memc[argv[3]], lval)
				GT_INULL(ptr) = lval
				GT_ITYPE(ptr) = TY_LONG
			} else {
				call readhexs(Memc[argv[3]], sval)
				GT_INULL(ptr) = sval
				GT_ITYPE(ptr) = TY_SHORT
			}

		} else if (streq(Memc[argv[1]], "ShortNull")) {
			ip = 1
			call readhexs(Memc[argv[3]], sval)
			GT_SNULL(ptr) = sval
		} else if (streq(Memc[argv[1]], "DoubleNull")) {
			ip = 1
			call readhexd(Memc[argv[3]], GT_DNULL(ptr))
		} else if (streq(Memc[argv[1]], "RealNull")) {
			ip = 1
			call readhexr(Memc[argv[3]], GT_RNULL(ptr))
		}

	} else if (streq(Memc[argv[1]], "EndTable")) {
	  call tm_finishcolumn(ptr)
          GT_CURTAB(ptr) = -1
	} else if (streq(Memc[argv[1]], "Header")) {
	  call tm_addhdr(ptr, Memc[argv[2]], argv[3], argc-2)
	} else if (streq(Memc[argv[1]], "Column")) {
	  if (argc > 4) {
            call tm_addcol(ptr, Memc[argv[2]], Memc[argv[3]],
			       Memc[argv[4]], Memc[argv[5]])
          } else {
	    call tm_addcolumn(ptr, Memc[argv[2]], Memc[argv[3]])
          }
	}

return 0

end

#	Initialize the server read

pointer	procedure tmstart(textdb,prefix, qpoep) # initialize reader

char	textdb[ARB]			#I output text database
char	prefix[ARB]			#I table name prefix
bool	qpoep				#I output format

pointer	ptr
int	len, strlen()
int	dtmap()

begin

	len = strlen(prefix)
	call eu_malloc(ptr, GT_SIZE, TY_INT)
	GT_MODE(ptr) = MODE_ASCII
	GT_QP(ptr) = 0
	GT_CURTAB(ptr) = -1
	GT_CURCOLS(ptr) = 0
	GT_BYTES(ptr) = 0
	GT_LENGTH(ptr) = 0
	GT_TYPIP(ptr) = 0
	GT_CURTYP(ptr) = 0
	GT_NTABLES(ptr) = 0
	GT_TABCOLS(ptr) = 0
	GT_TABTYPS(ptr) = 0
	GT_LNULL(ptr) = 0
	GT_QPIO(ptr) = 0
	GT_TABBUFS(ptr) = 0
	GT_TMBUFS(ptr) = 0
	GT_TMCNTS(ptr) = 0
	GT_CURNAM(ptr) = 0
	GT_TABNAMS(ptr) = 0
	GT_TABSIZS(ptr) = 0
	GT_TABTPS(ptr) = 0
	GT_TABCPS(ptr) = 0
	GT_QP_X(ptr) = 2048
	GT_QP_Y(ptr) = 2048
	GT_SUPP(ptr) = 0
	GT_TABTIMECOLP(ptr) = NULL
	GT_TABDELT(ptr) = INDEFR
	GT_HDRS(ptr) = 0		# number of headers
	GT_CURHDR(ptr) = 0		# the current header in the list
	GT_CURKEY(ptr) = 0		# the current key
	GT_ERRNO(ptr) = 0		# set when socket is failing


	if (qpoep) GT_QPOE(ptr) = 1; else GT_QPOE(ptr) = 0

	call eu_malloc(GT_PREFIX(ptr), len + 10, TY_CHAR)
	call strcpy(prefix, Memc[GT_PREFIX(ptr)], ARB)
	GT_PRELEN(ptr) = len

	call eu_malloc(GT_DNULLP(ptr), 1, TY_DOUBLE)
	call eu_malloc(GT_RNULLP(ptr), 1, TY_REAL)
	call eu_malloc(GT_TYPBUF(ptr), 4096, TY_CHAR)

	call fclobber(textdb)
	GT_DT(ptr) = dtmap(textdb, NEW_FILE)
	call dtptime(GT_DT(ptr))

	return ptr

end

#	Finish a server reader

procedure tmend(ptr, fp) # finish reading

pointer	ptr			#I table set
int	fp			#I input file descriptor

int	i, j
int	cols
int	status, egosvr_get()
int	tmnoop()
extern	tmnoop

begin

	if (ptr == 0) return

	call gt_table_list(ptr)
	call gt_end_supp(ptr)

	if (GT_NTABLES(ptr) != 0 && fp != -1) {
		status = egosvr_get(fp, "close\n", tmnoop, ptr)
	}

	do i = 1, GT_NTABLES(ptr) {

		cols = GT_TABCOL(ptr,i)

		if (cols != 0) {

			call eu_mfree(GT_TMBUF(ptr,i), TY_LONG)
			call eu_mfree(GT_TABTYP(ptr,i), TY_INT)
			call eu_mfree(GT_TABBUF(ptr,i), TY_LONG)
			call eu_mfree(GT_TABCP(ptr,i), TY_INT)

			if (GT_TABTP(ptr,i) != 0) call tbtclo(GT_TABTP(ptr,i))

			do j = 1, cols {
				call eu_mfree(GT_A_TNAMEP(ptr,i,j), TY_CHAR)
			}

		}

	call eu_mfree(GT_TABNAM(ptr,i), TY_INT)

	}

	if (GT_QPIO(ptr) != 0) call qpio_close(GT_QPIO(ptr))

	call eu_mfree(GT_TABCOLS(ptr), TY_INT)
	call eu_mfree(GT_TABTYPS(ptr), TY_INT)
	call eu_mfree(GT_TABNAMS(ptr), TY_INT)
	call eu_mfree(GT_TABSIZS(ptr), TY_INT)
	call eu_mfree(GT_TABBUFS(ptr), TY_INT)
	call eu_mfree(GT_TMBUFS(ptr), TY_INT)
	call eu_mfree(GT_TMCNTS(ptr), TY_INT)
	call eu_mfree(GT_DNULLP(ptr), TY_DOUBLE)
	call eu_mfree(GT_RNULLP(ptr), TY_REAL)

	if (GT_QP(ptr) != 0) call qp_close(GT_QP(ptr))

	call eu_mfree(GT_TYPBUF(ptr), TY_CHAR)
	call eu_mfree(GT_TABTPS(ptr), TY_INT)
	call eu_mfree(GT_TABCPS(ptr), TY_INT)
	call eu_mfree(GT_PREFIX(ptr), TY_CHAR)
	call dtunmap(GT_DT(ptr))
	call eu_mfree(ptr, TY_INT)

end

#	Process the output of a server command

int procedure	egosvr_get(fp, cmd, func, state) # read command results

int		fp				#I input file descriptor
char		cmd[ARB]			#I server command
int		func				#I processing function
pointer		state				#I state information

extern		func

pointer		sp, recv
int		ret

int		eussnd(), eusrcv(), euserr()
bool		streq()
int		flag
int		strncmp()

begin

  ret = -1
  flag = 0
  if (eussnd(fp, cmd) != 0) return ret
  
  call smark(sp)
  call salloc(recv, SZ_COMLINE, TY_CHAR)
  
  repeat {
    if (eusrcv(fp, Memc[recv], SZ_COMLINE) != 0) {
       GT_ERRNO(state) = euserr()
       break
    }
    if (streq(Memc[recv],"END\n")) {
      ret = 0;
      break;
    }

    if (strncmp(Memc[recv],"err", 3) == 0) {
      call eprintf("%s")
      call pargstr(Memc[recv])
      while (eusrcv(fp, Memc[recv], SZ_COMLINE) != 0) {
	if (streq(Memc[recv],"END\n")) break
      }
      break
    }
    if (Memc[recv] != '#') {
      if (func(recv, state, fp) != 0) break
      flag = 1
    }
  }

  call sfree(sp)
  if ((ret == 0) && (flag == 0)) ret = 1
  return ret
end

#	Generate a text database record the tables generated.

procedure	gt_table_list(ptr) # make table list

pointer		ptr			#I table set

int	i, j
int	cols

begin

	call dtptime(GT_DT(ptr)) 
	call dtput(GT_DT(ptr), "begin\tTable List\n")
	call dtput(GT_DT(ptr), "\tType\tTableList\n")
	call dtput(GT_DT(ptr), "\tLength\t%d\n")
	call pargi(GT_NTABLES(ptr))

	do i = 1, GT_NTABLES(ptr) {
		call dtput(GT_DT(ptr), "\t\t%s%d\n")
		call pargstr(Memc[GT_PREFIX(ptr)])
		call pargi(i-1)
	}

	cols = 0
	do i = 1, GT_NTABLES(ptr) {
		cols = cols + GT_TABCOL(ptr,i)
	}

	call dtptime(GT_DT(ptr)) 
	call dtput(GT_DT(ptr), "begin\tColumn List\n")
	call dtput(GT_DT(ptr), "\tType\tColumnList\n")
	call dtput(GT_DT(ptr), "\tLength\t%d\n")
	call pargi(cols)

	do i = 1, GT_NTABLES(ptr) {

		cols = GT_TABCOL(ptr,i)
		do j = 1, cols {

			call dtput(GT_DT(ptr), "\t\t%d\t%s\n")
			call pargi(i)
			call pargstr(GT_A_TNAME(ptr,i,j))
		}
	}

end

#	Set output dimensions for QPOE format

procedure	gt_setdim(ptr,x,y)

pointer		ptr
int		x, y

begin

	if (GT_QPOE(ptr) != 0) {
		GT_QP_X(ptr) = x
		GT_QP_Y(ptr) = y
	}

end


#	Select telemetry from server

procedure	t_gettm ()	# get processing status

pointer		service
pointer		host

int		fp
int		eusopn()

int		tmopen()
extern		tmopen
int		tmecho()
extern		tmecho
int		tmwhat()
extern		tmwhat
int		tmsupp()
extern		tmsupp
int		tmprocess()
extern		tmprocess

pointer		state, tmstart()
int		status, egosvr_get()

pointer		obs, acode

pointer		sp
pointer		cmd
pointer		textdb
pointer		prefix
pointer		tdb
pointer		env
pointer		lines,cline
pointer		lbuf

bool		clgetb(), qpoep
int		x, y, clgeti()
bool		batchp, binaryp
int		err
int		nlines

string		efmt "error: %s\n"

begin
  call smark(sp)
  call salloc(cmd, SZ_COMLINE, TY_CHAR)
  call salloc(lbuf, SZ_COMLINE, TY_CHAR)

  call euclstr(textdb, "textdb")
  call euclstr(prefix, "prefix")
  call euclstr(host, "hostname")
  call euclstr(service, "service")
  call euclstr(obs, "observation")
  call euclstr(acode, "accesscode")
  call euclstr(tdb, "tmtextdb")
  call euclstr(env, "tmrecord")
  if (Memc[tdb] != 0 && Memc[env] != 0)
  {
    call fmtenv(Memc[tdb],Memc[env], lines)
    nlines=0
    while (Meml[lines+nlines] != 0)
       nlines = nlines + 1
  }
  else
  {
    lines = NULL
    nlines = 0
  }

  batchp = clgetb("asynchronous")
  binaryp = clgetb("binary")
  qpoep = ! clgetb("tablep")
  if (qpoep) {
    x = clgeti("xdim")
    y = clgeti("ydim")
  }

  call sprintf(Memc[cmd], SZ_COMLINE, "open %s %s\n")
  call pargstr(Memc[obs])
  call pargstr(Memc[acode])

  err = ERR_CONN
  fp = eusopn(Memc[service], Memc[host])
  if (fp >= 0) repeat {
    
    state = tmstart(Memc[textdb],Memc[prefix], qpoep)
    err = ERR_STAR
    if (state == 0) break
  
    call gt_setdim(state,x,y)
  
    err = ERR_ENV
    
# Pass the contents of the text database file to the server
    if ( lines != NULL) {    
      for (cline = lines; Meml[cline] != 0; cline = cline + 1) {
        status = egosvr_get(fp, Memc[Meml[cline]], tmecho, state)
        if (status < 0) break
      }
    }
    if (status < 0) break
  
    status = egosvr_get(fp, Memc[cmd], tmopen, state)
    err = ERR_OPEN
    if (status < 0) break
  
    status = egosvr_get(fp, "detector\n", tmecho, state)
    err = ERR_DETS
    if (status < 0) break
  
    err = ERR_MODE
    if (binaryp) {
      call tmmode(state)
      status = egosvr_get(fp, "binary 1\n", tmecho, state)
      if (status < 0) break
    }
  
    err = ERR_PROC
    if (batchp) {
      status = egosvr_get(fp, "batchprocess\n", tmprocess, state)
    } else {
      repeat {
        status = egosvr_get(fp, "process\n", tmprocess, state)
      } until (status != 0)
    }
    if (status < 0) break
    
    if (binaryp) {
      call tmmode(state)
    }

    err = ERR_DESC
    status = egosvr_get(fp, "describe\n", tmwhat, state)
    if (status < 0) break
  
    err = ERR_SUPP
    status = egosvr_get(fp, "supplement\n", tmsupp, state)

    err = ERR_NONE
    break
  }
  
  if (state != 0) {
    if (GT_ERRNO(state) != 0) {
      call eprintf("error: host system error number %d\n")
      call pargi(GT_ERRNO(state))
    }
  }
  call tmend(state, fp)
  if (lines != NULL) call eu_free_strings(lines,nlines)
  if (fp >= 0) call euscls(fp)

  switch (err) {
  case ERR_STAR: call eprintf(efmt); call pargstr("could not start processing");
  case ERR_OPEN: call eprintf(efmt); call pargstr("could not open observation");
  case ERR_MODE: call eprintf(efmt); call pargstr("could not set binary mode");
  case ERR_PROC: call eprintf(efmt); call pargstr("could not process data");
  case ERR_DESC: call eprintf(efmt); call pargstr("could not get description");
  case ERR_SUPP: call eprintf(efmt); call pargstr("could not get supplement");
  case ERR_DETS: call eprintf(efmt); call pargstr("could not set science");
  case ERR_CONN: call eprintf(efmt); call pargstr("could not locate service");
  case ERR_ENV: call eprintf(efmt); call pargstr("could not set environment");
  }
end
