# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_mkprimbsch.x
# 	Author	:  tnk
# 	Date	:  03/01/93
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_mkprimbsch.x -r1.19 03/01/93 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	02/20/97   tnk	disabled creation of error columns
# 	01/29/97   tnk	made output table row ordered
# 	10/24/96   tnk	incr MAX_USR_PARS
# 	10/24/96   tnk	changed column datatypes and units
# 	10/21/96   tnk	  now searches for TIMEDEL, then DELTIME
# 	04/24/96   gareth	changed the RATE card to have the format RATE%-3d instead of RATE_%03d
# 	10/04/95   tnk	added altdeadtime param
# 	04/13/95   tnk	reworked to handle new table0 formats; some extra error checking
# 	04/13/95   tnk	now returns last value if "off the end" of deadtime table
# 	11/30/94   tnk	fix to get_dir
# 	08/25/94   tnk	fixed histogram bin type (again)
# 	08/24/94   tnk	fixed hist datatype in incr_bin
# 	08/08/94   tnk	added history card (progname/rev)
# 	07/25/94   tnk	added code to handle table0 that has no wsz columns; fixed arg bug
# 	05/02/94   tnk	deleted epsilon compare routines
# 	07/12/93   tnk	changed tbrgtt call to tbrgttn
# 	04/30/93   tnk	rearranged parms
# 	03/16/93   tnk	disabled negative primbsch 'correction' code
# 	03/01/93   tnk	date and time created 93/03/01 15:58:32 by tnk
# H-
# 
#

# mkprimbsch
# Author:  T. Kilsdonk
## @(#)t_mkprimbsch.x	1.19 2/20/97 

# task to create a table of primbsch/deadtime correction factors

# Reads the ST tables "table[0,1,2]" in the observation directory.
#
# Builds a new table which contains all the quadrant counts
# from table2 and new monitors which tabulate how many counts in each
# quadrant actually made it into the telemetry, as well as
# the primbsch/deadtime correction.
#
# All of the  monitors in the new table apply to the two major frame time
# interval which ends at the corresponding TableTime. 
#
# You can use the unadvertised keyword "any" in the detector list
# to tell the program to look up the detectors in the data tables.
#
# known problems/bugs in the code:
#
#    Deadtime pos/neg errors have been disabled, because we don't
#    know yet how to do them properly.  The columns are left INDEF.
#
#    Assumes that there are 4 RATE??? headers for each of the existing
#    N detectors in table 2.  It creates new RATE??? headers in the output
#    file starting with a suffix of 4*N+1.
#
#
# FUTURE ENHANCEMENTS
#
#  Add an option to get the X & Y offsets & gains from the telemetry
# stream.
#
# Allow an override of the deadtime table for those cases where the 
# ADC counts are believed to be correct, i.e., instead of using the sum 
# of the quad counts (Det[d]Q[a]SF) and the deadtime table, use the actual
# ADC counts (DetXADcts).
#
#-------------------------------------------------------------------
#  permitted names for eu_detect_id, abbrevs are ok
define DET_NAMES "ScA,ScB,ScC,SWS,MWS,LWS,DS,1,2,3,4,5,6,7"

# output column names (most have detector and quad number embedded)
define TIME_COLNAME            "TableTime"        # time
define TM_COLNAME              "Det%1dQ%1dtm"     # telemetered counts
define QUAD_COLNAME            "Det%1dQ%1dSf"     # quad counts
define SUMQ_COLNAME            "Det%1dSumQ"       # sum of quad counts
define DEAD_COLNAME            "Det%1ddc"         # deadtime correction
define DEAD_NERR_COLNAME       "Det%1ddcnerr"     # deadtime neg err
define DEAD_PERR_COLNAME       "Det%1ddpnerr"     # deadtime pos err
define PRIMB_COLNAME           "Det%1dQ%1dpc"     # primbsch correction
define DEADPRIMB_COLNAME       "Det%1dQ%1ddpc"    # primb + dead correction

# Input column names

# from "detector" reference table
define  TIFNUM_COLNAME         "tif_number"
define  XOFF_COLNAME           "x_offset"
define  XGAIN_COLNAME          "x_gain"
define  YOFF_COLNAME           "y_offset"
define  YGAIN_COLNAME          "y_gain"
define  PRIMB_TYPE_COLNAME     "primbsch_type"
define  PRIMB_X_Y1_COLNAME     "primbsch_x_y1"
define  PRIMB_Y_Y2_COLNAME     "primbsch_y_y2"
define  PRIMB_Y3_COLNAME       "primbsch_y3"
define  REF_DEAD_COLNAME       "tifdeadtime"

# primbsch quadrant types
define  PRIMB_VERT_TYPE        2        # "quads" are really vertically stacked
define  PRIMB_SQR_TYPE         1        # "quads" really are quadrants


# from deadtime table
define  DEAD_CNTS_COLNAME      "counts"
define  DEAD_EFF_CNTS_COLNAME  "effcounts"

# from table0
define  PH_TIME_COLNAME        "PhotonT"
define  PH_X_COLNAME           "PhotonX"
define  PH_Y_COLNAME           "PhotonY"
define  PH_W_COLNAME           "PhotonW"
define  PH_S_COLNAME           "PhotonS"
define  PH_Z_COLNAME           "PhotonZ"
define  PH_D_COLNAME           "PhotonD"
define  PH_XORS_COLNAME        "PhotonXorS"
define  PH_YORZ_COLNAME        "PhotonYorZ"

# table0 formats
define  FORMATXY     1
define  FORMATXYWSZ  2
define  FORMATMERGE  3

# indices for the event table columns
define  MAX_TAB0_COLS  9
define  PH_TIME   1
define  PH_X      2
define  PH_Y      3
define  PH_W      4
define  PH_S      5
define  PH_Z      6
define  PH_D      7   # detector number
define  PH_XORS   8
define  PH_YORZ   9

# from table1
define  TAB1TIME_COLNAME       "Table1Time"

# from table2
define  TAB2_TIME_COLNAME      "Table2Time"
define  TAB2_QUAD_COLNAME      "Det%1dQ%1dSf"     # quad counts mask

define  PRIMARY_DELTIME_HDR    "TIMEDEL"
define  ALTERNATE_DELTIME_HDR  "DELTIME"    
	 
define PRIMB_INTERVAL 2.048D0  # time in secs for 1 primbsch interval
define HALF_PRIMB_INT 1.024D0  # time in secs for 1/2 primbsch interval
define N_QUADS        4        # number of "quadrants"
define MAX_DETECT     10       # max number of detectors allowed
define EVENT_EPSILON  0.0001D0 # Epsilon for FP comparisons.  If 2 values
                               # are within this, they are considered the
                               # same.   (Round-off compensation).
                               # Event resolution is >= 2 ms, so this value
                               # should be fine.
define MAX_USR_PARS   120       # max. number of user parameters in outfile
define MAX_REP_OFFSET 2.048D0  # max. report offset
define T2_T1_TIME_RATIO  2.0   # expected TABLE2/TABLE1 delta time ratio
define AUTO_DETLIST   "any"    # parm keyword: use detectors no.s in datafiles
#-------------------------------------------------------------------

include <error.h>
include <tbset.h>
include <mach.h>

# detector structure
include "mp.h"

# make a primbsch/deadtime correction table
procedure t_mkprimbsch()       #EXTERNAL

# LOCAL VARS

# parameter vars
pointer observe_dir      # observation directory
pointer event_table      # name of table of events (table0)
pointer majframe_tbl     # name of table of major frame times (table1)
pointer quadcnts_tbl     # name of table of quadrant counts (table2)
pointer output_file      # output file
pointer par_detlist      # detlist parameter
pointer ref_data         # reference directory/data table
double	report_offset    # quadrant count timestamp offset
bool    info_msgs        # give informational messages 
bool    diag_msgs        # give diagnostic messages 
bool    warn_msgs        # give warning messages flag
real    readper          # percent of data files to read
bool    do_neg_primb     # do negative primbsch corrections
pointer altdeadtbl	 # user specified deadtime table
#
int	nrows            # number of rows in table
int     i                # loop index
int     nbins            # number of histogram bins 
pointer	sp               # stack pointer
pointer times            # time stamps
pointer time_cp          # output table time column ptr
pointer dead_in          # deadtime input column
pointer dead_out         # deadtime output column
pointer out_tp           # output table pointer
pointer dptr[MAX_DETECT] # detector stucture pointers
pointer tp0              # pointer to table0 (events)
pointer tp1              # pointer to table1
pointer tp2              # table 2 pointer
pointer tab0_cptr[MAX_TAB0_COLS]  #  table0 column pointers
int     tab0_format      # format of table0 
pointer t1times          # column of times from table1
int     n_t1times        # number of rows in t1times table
int     deadtbl_nrows    # no. of rows in deadtime table
pointer ref_tp           # reference table pointer
pointer deadtblnam       # name/path of deadtime table
pointer deadtp           # deadtime table pointer
real    t1_deltime       # table1 deltatime
real    t2_deltime       # table2 deltatime
real    deltime_ratio    # deltatime ratio
pointer start_time       # bin start times for event histogram
pointer end_time         # bin end times for event histogram
pointer bin_valid        # marker: true if bin interval is valid
pointer str              # string buffer
pointer ref_detnums      # detector numbers from reference file
int     n_ref_detnums    # length of array
pointer par_detnums[MAX_DETECT]  # detector numbers from parameter file
int     n_par_detnums    # length of array
int     data_detnums[MAX_DETECT] # detector numbers from data file (table2)
int     n_data_detnums   # length of array
int     run_detnums[MAX_DETECT]  # detector numbers to actually use
int     n_run_detnums    # length of array
int     n_det            # no. of valid detectors 
int     rate_index       # index to use in header rate parameters
pointer ref_dir          # reference directory

# FUNCTIONS
real	 clgetr()
double   clgetd()
bool	 clgetb()
pointer  clpopnu()
pointer	 tbtopn()
int      tbpsta()
real     tbhgtr()
bool     verf_par_detnums()
bool     fp_equalr()
bool     ref_deadtbl_name()
bool     ref_offgains()
bool     ref_qbounds()
bool     streq()
bool     list_subset()
pointer  get_dir()
int	 strlen()

errchk tbrgtd

begin

  call smark (sp)

  # *** read parameter file ***
  call eu_clstr(output_file, "output")     # output directory/file
  par_detlist      = clpopnu("detectors")  # list of detectors
  info_msgs        = clgetb ("infomsgs")   # flag for info msgs
  warn_msgs        = clgetb ("warnmsgs")   # warnings
  diag_msgs        = clgetb ("diagmsgs")   # diagnostics
  do_neg_primb     = clgetb ("negprimb")   # test for neg primbsch
  readper          = clgetr ("readper")    # percent of data to read
  call eu_clstr(observe_dir, "obsdata")    # observation directory
  call eu_clstr(event_table, "events")     # table of det events
  call eu_clstr(majframe_tbl, "gaptimes")  # tbl of maj frame times
  call eu_clstr(quadcnts_tbl,"quadcts")    # table of quad counts
  call eu_clstr(ref_data,     "refdata")    # det ref table dir/table
  call eu_clstr(altdeadtbl,  "altdeadtime") # user selected deadtime table
  report_offset    = clgetd ("ctsoffset")  # report offset in seconds
 
  # --------- *** VERIFY PARAMETERS *** ------------------

  call salloc (str, SZ_LINE, TY_CHAR)

  # Get lists of detector numbers from
  #   1. the reference dataset  (possible detector numbers)
  #   2. the datafile (table2)  (detector numbers we have events for)
  #   3. the parameter file     (detectors requested by user)
  # and create a valid list of detectors to run with.
  # The reference detnums list is used to find the datafile detnums list.
  # The run detnums list must be a subset of the datafile detnums list.

  # *** open reference "detector" table ***
  ref_tp = tbtopn (Memc[ref_data], READ_ONLY, 0)

  # get list of detectors from reference dataset
  call colgeti(ref_tp,TIFNUM_COLNAME,ref_detnums,n_ref_detnums)
  if (n_ref_detnums == 0)
     call error(510,"Could not read detector numbers from ref file")

  if (diag_msgs)
      call pr_ivec(Memi[ref_detnums],n_ref_detnums,"Ref Det List")

  # open table 2
  call open_tab(Memc[observe_dir],Memc[quadcnts_tbl],tp2)

  # create detector list from table2
  call get_data_detnums(tp2,Memi[ref_detnums],n_ref_detnums,data_detnums,
      MAX_DETECT,n_data_detnums)
  if (diag_msgs)
     call pr_ivec(data_detnums,n_data_detnums,"Data Det List")
  if (n_data_detnums <= 0 || n_data_detnums > MAX_DETECT) {
     call eprintf("Could not find detectors in '%s' quadrant counts table")
     call pargstr(Memc[quadcnts_tbl])
     call eprintf(" that match reference table\n")
     call error(520," Exiting...")
  }

  # get first string from parameter det list
  call clgfil(par_detlist, Memc[str], SZ_LINE) 
  call clprew(par_detlist)

  # use detector list from data table?  (check for special token)
  if (streq(Memc[str],AUTO_DETLIST)) {
     # run_nums = table2nums
     call amovi(data_detnums,run_detnums,n_data_detnums)
     n_run_detnums = n_data_detnums
  }
  else {
     # get the list of detectors from parameter file
     if (!verf_par_detnums(par_detlist,par_detnums,MAX_DETECT,n_par_detnums)) {
         call error(530,"Invalid detector list in parameter file")
     }

     # copy par_detnums to run_detnums
     call amovi(par_detnums,run_detnums,n_par_detnums)
     n_run_detnums = n_par_detnums
     if (diag_msgs)
        call pr_ivec(run_detnums,n_run_detnums,"Run Det List")

     # make sure run_detnums list is a subset of the table2 list
     if (!list_subset(run_detnums,n_run_detnums,data_detnums,n_data_detnums)) {
        call error(540,"You requested detectors not found in the data tables")
     }

  }

   do i = 1,n_run_detnums {
      call mp_malloc (dptr[i])    # allocate detector structures
      MP_NUM(dptr[i]) = run_detnums[i]
   }
   n_det = n_run_detnums

   call salloc (deadtblnam, SZ_LINE, TY_CHAR)

    if ( ! ref_deadtbl_name(ref_tp,Memc[deadtblnam]))  # get deadtime tbl name
        call error(550,"Could not find name of deadtime table in ref table")

    if ( ! ref_offgains(ref_tp,dptr,n_det))     # get x&y offsets & gains
      call error(560,"Cannot find XY offsets & gains for all of the detectors")

    if ( ! ref_qbounds(ref_tp,dptr,n_det))      # get quadrant boundaries
      call error(570,"Cannot find primbsch bounds for all of the detectors")

  call tbtclo (ref_tp)                         

  # split apart ref_dir from refdata
  ref_dir = get_dir(Memc[ref_data])
  if (ref_dir == NULL) {
      call error(575,"Could not pull reference directory out of refdata")
  }

  #override deadtime?
  if ( strlen(Memc[altdeadtbl]) > 0)
  {
     # open user selected deadtime table
     deadtp =  tbtopn (Memc[altdeadtbl], READ_ONLY, 0)
  } 
  else
  {
     # open deadtime correction table in refdata
     call open_tab(Memc[ref_dir],Memc[deadtblnam],deadtp)
  }

  # get deadtime information 
  call colgetr(deadtp,DEAD_CNTS_COLNAME,dead_in,deadtbl_nrows)
  call colgetr(deadtp,DEAD_EFF_CNTS_COLNAME,dead_out,deadtbl_nrows)
  if (dead_in == NULL || dead_out == NULL)
     call error(580,"Could not read deadtime table columns")
  call tbtclo (deadtp)

  if (readper > 100.0 || readper <= 0.0) {
     call error(590,"Read percent must be > 0 and <= 100")
  }
  readper = readper / 100.0

  if (report_offset < 0.0 || report_offset > MAX_REP_OFFSET) {
     call eprintf("Report offset must be >= 0.0 and <= %8.3f\n")
     call pargd(MAX_REP_OFFSET)
     call error(600,"Exiting")
  }


  # open table 1
  call open_tab(Memc[observe_dir],Memc[majframe_tbl],tp1)

  # table1: look for primary, then alternate timdel/deltime header cards
  iferr (t1_deltime = tbhgtr(tp1, PRIMARY_DELTIME_HDR))
  {
     iferr (t1_deltime = tbhgtr(tp1, ALTERNATE_DELTIME_HDR))
     {
       call error(601,"Can't find deltime/timedel cards in gaptimes table")
     }
  }

  # table2: look for primary, then alternate timdel/deltime header cards
  iferr (t2_deltime = tbhgtr(tp2, PRIMARY_DELTIME_HDR))
  {
     iferr (t2_deltime = tbhgtr(tp2, ALTERNATE_DELTIME_HDR))
     {
       call error(601,"Can't find deltime/timedel cards in quadcts table")
     }
  }

  # output table is same length as table2, so get the no. of rows
  nbins = tbpsta (tp2, TBL_NROWS) 
  nbins = nbins * readper

  # verify table2 deltime / table1 deltime is the expected ratio
  deltime_ratio = t2_deltime / t1_deltime
  if (! fp_equalr(deltime_ratio,T2_T1_TIME_RATIO)) {
     call eprintf("Table %s and Table %s do not have compatible delta times\n")
     call pargstr(Memc[majframe_tbl])
     call pargstr(Memc[quadcnts_tbl])
     call error(610,"exiting...")    
  }

  #open table0 (events table) & verify it
  if (diag_msgs)
     call eprintf("Opening & verifying table0\n")
  call open_tab(Memc[observe_dir],Memc[event_table],tp0)
  call get_tab0_cptr(tp0,tab0_cptr,MAX_TAB0_COLS,tab0_format)

  # try to make output file as early as possible in case it fails
  if (diag_msgs)
     call eprintf ("Creating output table\n")
  out_tp = tbtopn (Memc[output_file], NEW_FILE, 0)           # open it
  call make_output_table(out_tp,time_cp,dptr,n_det,nbins)    # build it

  # ****** END VERIFICATION ***

  # get time column from table2
  if (diag_msgs)
     call eprintf("slurping up table2 times\n")

  call colgetd(tp2,TAB2_TIME_COLNAME,times,nrows)
  if (times == NULL) 
     call error(620,"Could not read time column from quadrant counts table")

  # *** get time column from table 1 ***
  if (diag_msgs)
     call eprintf("slurping up table1 times\n")

  call colgetd(tp1,TAB1TIME_COLNAME,t1times,n_t1times)
  if (t1times == NULL) 
     call error(630,"Could not read table1 times")

  call tbtclo(tp1)

  if (diag_msgs)
     call eprintf("Setting up\n")

  # *** Set up the detector histogram buffers  ***
  call malloc (start_time, nbins, TY_DOUBLE)
  call malloc (end_time,   nbins, TY_DOUBLE)
  call malloc (bin_valid,  nbins, TY_BOOL)

  # setup the start & end times for the bins, and
  # a validity indicator

  call setup_bins(Memd[times],nbins,Memd[start_time],Memd[end_time],
     Memb[bin_valid],Memd[t1times],n_t1times,report_offset,diag_msgs)

  call mfree (times, TY_DOUBLE)
  call mfree (t1times, TY_DOUBLE)

  # setup arrays for the bins
  do i = 1, n_det {
     call pb_det_setup (dptr[i], nbins)
  }

  # calculate the telemetered counts by histogramming all
  # of the table0 events

  call build_hist(tp0,tab0_cptr,tab0_format,Memd[start_time],Memd[end_time],
     nbins,dptr,n_det,readper,info_msgs)

  # output table stuff
  #
  # we copy all the quadrant counts over from table2 and add the 
  # telemetered quadrant counts which we have constructed and then
  # compute the correction

  # *** prepare output file ***
  if (diag_msgs)
     call eprintf ("Copying header parameters\n")
  call tbhcal (tp2, out_tp)                   # copy all header parameters
  call tbhptt (out_tp, "TIME", TIME_COLNAME)  # set TIME parameter 

  # add some history
  call tbhadt(out_tp, "HISTORY", "Generated by t_mkprimbsch.x, Rev. 1.19 2/20/97")

  # label Det#Q#tm monitors in the table header as rates
  if (diag_msgs)
     call eprintf ("Setting RATE header parameters\n")

  # assume for now that there are RATE??? headers, 4 for each
  # of the existing detectors in table 2.  
  rate_index = n_data_detnums * N_QUADS + 1
  call label_rates(out_tp,rate_index,dptr,n_det)

  #  write out bin end times as the time column
  if (diag_msgs)
     call eprintf ("Writing time column\n")
  call tbcptd (out_tp, time_cp, Memd[end_time], 1, nbins)

  # Write out histograms (ie, telemetered counts).  They
  # are saved as they are written out for use in the correction later.

  if (diag_msgs)
     call eprintf ("Writing telemetered counts\n")
  call wr_tm_counts(out_tp,dptr,n_det,nbins)

  # Get quadrant counts from Table2 and write out.  The copied
  # counts are saved in QCNT for use in computing the correction later

  if (diag_msgs)
     call eprintf ("Copying quadrant counts\n")
  call cp_quad_cnts(tp2, out_tp, dptr, n_det, readper)
  call tbtclo (tp2)

  # Make primbsch/deadtime correction

  if (diag_msgs) 
     call eprintf ("Computing correction\n")

  # do primbsch/deadtime corrections 
  call primb_dead(out_tp,Memr[dead_in],Memr[dead_out],deadtbl_nrows,dptr,
     n_det,Memb[bin_valid],nbins,warn_msgs,info_msgs,do_neg_primb)

  # *** cleanup ***
  if (diag_msgs)
     call eprintf ("Cleaning up\n")

  call tbtclo (out_tp)           # close output file

  call clpcls(par_detlist)       # free parameter det list

  do i = 1, n_det {
    call pb_det_free (dptr[i])   # Free histogram buffers
    call mp_free(dptr[i])
  }

  call mfree (ref_dir, TY_CHAR)
  call mfree (start_time, TY_DOUBLE)
  call mfree (end_time, TY_DOUBLE)
  call mfree (bin_valid, TY_BOOL)

  call mfree (dead_in, TY_REAL)
  call mfree (dead_out, TY_REAL)

  call mfree (ref_detnums, TY_INT)

  call sfree (sp)
end

#------------------------------------------------------

# verify list of detector names from parameter list
#
bool procedure verf_par_detnums(det_list,det_nums,max_det_nums,n_found)  #LOCAL
pointer det_list[ARB]      #I list of detector names
int     det_nums[ARB]      #O list of detector numbers
int     max_det_nums       #I length of allocated det_nums array
int     n_found            #I no. of valid det numbers found

# parse list of detector names & make sure they're all valid
# returns true if entire list is valid, else returns false

# LOCAL VARS
bool     status
int      det
pointer  sp
pointer  str
int      clplen()
int      i
int      n_det

begin

  call smark(sp)
  call salloc (str, SZ_LINE, TY_CHAR)
  status = true

  n_found = 0
  n_det = clplen(det_list)
  for (i=1; i<= max_det_nums && i <= n_det; i=i+1) {

     call clgfil(det_list, Memc[str], SZ_LINE)   # get next string

     call strupr(Memc[str])              # make uppercase

     call eu_detect_id(Memc[str],det)    # convert to number

     if (det == 0) {
        call eprintf("Invalid detector name: %s\n")
        call pargstr(Memc[str])
#        call eprintf("Must be one of: %s\n")
#        call pargstr(DET_NAMES)
        status = false
     }
     else {
        n_found = n_found + 1
        det_nums[n_found] = det
     }
  }

  call sfree(sp)
  return (status)

end

#------------------------------------------------------
# test if int a_list vector is a subset of b_list
bool procedure list_subset(a_list,n_a_list,b_list,n_b_list)  #LOCAL

int a_list[ARB]         #I subset list
int n_a_list            #I list length
int b_list[ARB]         #I master list
int n_b_list            #I list length

#LOCALS
int i
int j
bool all_found
bool found

begin

    # go thru every element of a_list and see if it's in b_list

    all_found = true

    for (i=1; i <= n_a_list; i=i+1) {

       found = false
       for (j=1; j<=n_b_list && !found; j=j+1) {
          if (a_list[i] == b_list[j]) {
            found = true
            break
          }
       }

       if (!found) {
         all_found = false
         call eprintf("Detector number %d not found in data tables\n")
         call pargi(a_list[i])
       }
    }

return(all_found)

end
#------------------------------------------------------

# get list of detector names from table2
# Look for DET[d]Q0Sf.  If it exists, we have a valid detector.

procedure get_data_detnums(tp,det_nums_avail,n_det_avail,det_nums,
   max_det_nums,n_found)  #LOCAL

pointer  tp                   #I pointer to open table
int      det_nums_avail[ARB]  #I list of possible detector numbers
int      n_det_avail          #I len of det_nums_avail
int      det_nums[ARB]        #O detector numbers found in file
int      max_det_nums         #I max len of det_nums array
int      n_found              #O no. of det numbers found in file

# Look in table:
# Using each possible detector number from det_nums_avail list,
# see if quadrants counts column exists.  If it does,
# add that detector number to the list.

# LOCAL VARS
int      det
pointer  sp
pointer  str
int      i
pointer  cp 
int      quad

#FUNCTIONS
#int      ctoi()


begin

  call smark(sp)
  call salloc (str, SZ_LINE, TY_CHAR)


  quad = 1         # always look for first quad (really quad 0)
  n_found = 0
  for (i=1; i<= n_det_avail && n_found < max_det_nums; i=i+1) {

     det = det_nums_avail[i]

     # look for DET[d]Q0Sf.  If it exists, we have a valid detector.
     call sprintf (Memc[str], SZ_LINE, TAB2_QUAD_COLNAME)
     call pargi (det)
     call pargi (quad-1)

     call tbcfnd (tp, Memc[str], cp, 1)

     if (cp != NULL) {
        n_found = n_found + 1
        det_nums[n_found] = det
     }

  }

  call sfree(sp)

end
#----------------------------------------------------------

# get deadtime table name
#
bool procedure ref_deadtbl_name(ref_tp,deadtblnam)  #LOCAL

pointer ref_tp             #I  opened reference detector table
char deadtblnam[ARB]       #O  name of deadtime table

# LOCAL VARS
pointer cp        # column pointer
bool junk         # junk value
bool found        # status return

begin

  found = false

  call tbcfnd (ref_tp, REF_DEAD_COLNAME, cp, 1)    
  if (cp != NULL) {
     call tbrgttn (ref_tp, cp, deadtblnam, junk, SZ_LINE, 1, 1)  # from 1st row
     found = true
  }

return(found)

end
#------------------------------------------------------
# read x & y offsets & gains from detector ref table
#
bool procedure ref_offgains(tp,dptr,n_det)   #LOCAL

pointer tp               #I table pointer
pointer dptr[ARB]        #I/O detector struct
int n_det                #I no. of detectors in dptr

# returns true on success

# LOCAL VARS
pointer detnum_cp        # detector no. column pointer
pointer xoff_cp          # x offset column pointer  
pointer xgain_cp	 # x gain column pointer  
pointer yoff_cp		 # y offset column pointer  
pointer ygain_cp	 # y gain column pointer  
int     nrows            # number of rows
int     row              # row index
int     tab_det          # detector number from the table
int     i                # index
bool    found            # flag

# FUNCTIONS
int     tbpsta()
begin

  n_rows = tbpsta (tp, TBL_NROWS)

  # get column pointers
  call tbcfnd (tp, TIFNUM_COLNAME, detnum_cp, 1)  # use this as det number
  call tbcfnd (tp, XOFF_COLNAME,   xoff_cp  , 1)
  call tbcfnd (tp, XGAIN_COLNAME,  xgain_cp,  1)
  call tbcfnd (tp, YOFF_COLNAME,   yoff_cp,   1)
  call tbcfnd (tp, YGAIN_COLNAME,  ygain_cp,  1)

  if (detnum_cp == NULL || 
      xoff_cp == NULL || xgain_cp == NULL ||
      yoff_cp == NULL || ygain_cp == NULL ) {
    call error(640,"Could not find ref columns for xy offsets & gains")
  }

  for (i=1; i<=n_det; i=i+1) {

     # go down each row in table, see if it matches the current detector 
     found = false

     for (row=1; row <=n_rows; row=row+1) {

        # get detector number of the row
        call tbegti(tp, detnum_cp, row, tab_det)

        if (tab_det == MP_NUM(d_ptr[i])) {
           found = true
           # get the offsets & gains 

           call tbegtd(tp, xoff_cp,  row,   MP_XOFF(dptr[i]))
           call tbegtd(tp, xgain_cp, row,  MP_XGAIN(dptr[i]))
           call tbegtd(tp, yoff_cp,  row,   MP_YOFF(dptr[i]))
           call tbegtd(tp, ygain_cp, row,  MP_YGAIN(dptr[i]))
           break
       }

     }

     if (! found) {
        call eprintf("Error:  Could not find xy offsets & gains for det %d\n")
        call pargi(MP_NUM(dptr[i]))
        break
     }
  }

return(found)

end
#------------------------------------------------------

# reads primbsch quadrant boundaries from detector ref table
#
bool procedure ref_qbounds(tp,dptr,n_det)      #LOCAL

pointer tp               #I   table pointer
pointer dptr[ARB]        #I/O detector struct
int n_det                #I   len of dtpr array

#returns true on success

# LOCAL VARS
pointer detnum_cp        # detector number column pointer
pointer x_y1_cp          # x/y1 boundary column pointer
pointer y_y2_cp          # y/y2 boundary column pointer
pointer y3_cp            # y3 boundary column pointer
pointer type_cp          # type of boundary column ptr
int     nrows            # number of rows
int     row              # current row no.
int     tab_det          # detector number from table
bool    found            # flag
int     i                # index

# FUNCTIONS
int     tbpsta()

begin

  n_rows = tbpsta (tp, TBL_NROWS)

  # get column pointers
  call tbcfnd (tp, TIFNUM_COLNAME,     detnum_cp, 1)  # use this as det number
  call tbcfnd (tp, PRIMB_TYPE_COLNAME, type_cp,   1)
  call tbcfnd (tp, PRIMB_X_Y1_COLNAME, x_y1_cp,   1)
  call tbcfnd (tp, PRIMB_Y_Y2_COLNAME, y_y2_cp,   1)
  call tbcfnd (tp, PRIMB_Y3_COLNAME,   y3_cp,     1)

  if (detnum_cp == NULL || 
      type_cp == NULL   || x_y1_cp == NULL ||
      y_y2_cp == NULL   || y3_cp == NULL ) {
    call error(650,"Could not find ref columns for primbsch quadrants")
  }

  for (i=1; i<=n_det; i=i+1) {

     # go down each row in table, see if it matches a detector in our
     # list
     found = false

     for (row=1; row <=n_rows; row=row+1) {
        # get detector number of each row
        call tbegti(tp, detnum_cp, row, tab_det)

        if (tab_det == MP_NUM(d_ptr[i])) {

           found = true
           # get the quadrant type & boundaries
           call tbegti(tp, type_cp,  row,   MP_QBOUND_TYPE(dptr[i]))
           call tbegti(tp, x_y1_cp,  row,   MP_QBOUND_X_Y1(dptr[i]))
           call tbegti(tp, y_y2_cp,  row,   MP_QBOUND_Y_Y2(dptr[i]))
           call tbegti(tp, y3_cp,    row,   MP_QBOUND_Y3(dptr[i]))
           break
        }
     }

     if (! found) {
        call eprintf("Error:  could not find primb quad bounds for det %d\n")
        call pargi(MP_NUM(dptr[i]))
        break
     }

  }

return(found)

end
#------------------------------------------------------

# open up an ST table in the given directory
#
procedure open_tab(dir,tab_name,tp)    #LOCAL

char dir[ARB]          #I directory name
char tab_name[ARB]     #I table name
pointer tp             #O pointer to opened table

# LOCAL VARS
pointer sp             #stack pointer
pointer str            #string buffer

# FUNCTIONS
pointer	 tbtopn()

begin

  call smark(sp)

  call salloc (str, SZ_LINE, TY_CHAR)

  # Open table 

  call strcpy(dir, Memc[str], SZ_LINE)
  call strcat("/", Memc[str], SZ_LINE)
  call strcat (tab_name, Memc[str], SZ_LINE)
  tp = tbtopn (Memc[str], READ_ONLY, 0)

  call sfree(sp)

end
#------------------------------------------------------

# create start & end times for histogram bins
#
procedure setup_bins(table2times,nbins,start_time,end_time,bin_valid,
t1times,n_t1times,report_offset,diag_msgs)    #LOCAL

double table2times[ARB]  #I  time stamps from table 2
int    nbins             #I  num elements in time arrays
double start_time[ARB]   #O  bin start times
double end_time[ARB]     #O  bin end times
double report_offset     #I  offset in seconds from quad counts report time
bool   bin_valid[ARB]    #O  flag marking which bins are valid
double t1times[ARB]      #I  table1 times (for telemetered intervals)
int    n_t1times         #I  length of t1times
bool   diag_msgs         #I  verbose messages


# LOCAL VARS
double  bin_width       # width of bin in seconds
double  diff_time       # difference in time
int     bin             # bin number

# FUNCTIONS
bool eq_epsilon()


  # The quadrant counts apply to the time interval of the two major frames
  # previous to the one in which the counts are telemetered.  


begin

  # make start & end times for each bucket.
  bin_width = PRIMB_INTERVAL    # buckets are 1 primbsch interval wide
  
  # bin end times are same as table2 times:
  # end_time = table2time
  call amovd(table2times,end_time,nbins)

  # Now take off amount of time from the end of the interval in which
  # the quad counts are telemetered, and the interval
  # to which the quadrant counts actually apply.  This value will
  # be either 1.024 or 2.048 seconds.  I.E, shift all the bins back.

  # end_time = end_time - report_offset
  call asubkd(end_time,report_offset,end_time,nbins)
  
  # Check for flushes, ie, where time diff between successive table2times
  # was 1 MF instead of two.  This means that the REPORT_OFFSET may have 
  # changed for a bin, and we need to add 1 Major Frame to compensate

  # offset the times & subtract them to get time difference

  for (bin = 2; bin <= nbins; bin=bin+1) {
     diff_time = end_time[bin] - end_time[bin-1]
     if ( eq_epsilon(diff_time,HALF_PRIMB_INT,EVENT_EPSILON)) {
        end_time[bin] = end_time[bin] + HALF_PRIMB_INT
        if (diag_msgs) {
           call eprintf("setup_bins:  flush in bin %d\n")
           call pargi(bin)
        }
     }
  }

  # make start time a bin width earlier than end time
  # start_time = end_time - BIN_WIDTH
  call asubkd(end_time,bin_width,start_time,nbins)


  # Take care of round off errors (very small bin gaps).
  # If end of bin A is closely adjacent to start of bin A + 1,
  # then make them match exactly.  Otherwise events will fall
  # between the bins.

  for (bin = 2; bin <= nbins; bin=bin+1) {
     if ( eq_epsilon(start_time[bin],end_time[bin-1],EVENT_EPSILON)) {
        start_time[bin] = end_time[bin-1]
     }
     else {  
         if (diag_msgs) {
            call eprintf("data gap before bin %d\n")
            call pargi(bin)
         }
     }

  }


  # now look in table 1 to see if either of the 2 Major Frames
  # making up each histogram interval is missing.  If so, the
  # telemetered counts
  # will be invalid (1/2 of the events will be missing).  So
  # mark the interval as invalid.

  call find_valid_bins(start_time,end_time,nbins,bin_valid,t1times,
     n_t1times,diag_msgs)
   
end
#------------------------------------------------------------

# find the histogram bins where either telemetered frame is missing
#
procedure find_valid_bins(start_time,end_time,nbins,bin_valid,
   t1times,n_t1times,diag_msgs)    #LOCAL

double start_time[ARB]    #I start time of each bin
double end_time[ARB]      #I end time of each bin
int    nbins              #I number of bins (length of above arrays)
bool   bin_valid[ARB]     #O flag marking which bins are valid
double t1times[ARB]       #I table1 times (for telemetered intervals)
int    n_t1times          #I length of t1times
bool   diag_msgs          #I give verbose information

  # look in table 1 times to see if either of the 2 Major Frames
  # making up each histogram bins is missing.  If so, the
  # telemetered counts
  # will be invalid (1/2 of the events will be missing).  So
  # mark the interval (bin) as invalid.

# LOCAL VARS
int bin                   # bin number
int t1                    # index into table1 times array
int i                     # loop index
int badcnt                # count of invalid bins

#FUNCTIONS
bool eq_epsilon()
bool lt_epsilon()

begin

   bin = 1
   t1 = 1
   badcnt = 0
   # mark all bins as invalid
   for (i=1; i<=nbins; i=i+1) {
      bin_valid[i]=false
   }

   # Go thru bin times, find the ones that have both major frame
   # entries in table1.
   # Note that we use epsilon comparisons to handle floating point
   # roundoff errors.

   while (bin <= nbins && t1 < n_t1times) {

      # **** if (time[t1] < end_time[bin]-HALF_PRIMB_INT) ****
      if ( lt_epsilon(t1times[t1],end_time[bin]-HALF_PRIMB_INT,
           EVENT_EPSILON)) { 
         # sync up times
         t1 = t1 + 1
      }
      # **** if ( time[t1] == end_time[bin]-HALF_PRIMB_INT &&  ****
      # ****      time[t1+1] == end_time[bin]) ****

      else if ( eq_epsilon(t1times[t1],end_time[bin]-HALF_PRIMB_INT,
         EVENT_EPSILON) && eq_epsilon(t1times[t1+1],end_time[bin],
         EVENT_EPSILON))  {
             bin_valid[bin] = true      
             t1 = t1 + 2
             bin = bin + 1
      }
      else {
         # skip to the next bin
         bin = bin + 1
      }
   }

   for (i=1; i<=nbins; i=i+1) {
      if (! bin_valid[i] && diag_msgs) {
      call eprintf("bin marked invalid: %d) start=%18.6f end=%18.6f\n")
      call pargi(i)
      call pargd(start_time[i])
      call pargd(end_time[i])
      badcnt = badcnt + 1
      }
   }

   if (badcnt == nbins) {
       call eprintf("Total misalignment between table2 times\n")
       call eprintf("and table1 times.  Cannot continue!\n")
       call error(660,"Aborting")
   }

end
#----------------------------------------------------------

# build a histogram of telemetered events
#
procedure build_hist(tp,cp,tab0_format,start_time,end_time,
   nbins,dptr,n_det,readper,info_msgs)  #LOCAL

pointer tp                 #I event table pointer
pointer cp[ARB]            #I event column pointers 
int     tab0_format        #I format of table0 (event table)
double  start_time[ARB]    #I histogram bin start times
double  end_time[ARB]      #I bin end times
int     nbins              #I number of histogram bins
pointer dptr[ARB]          #I/O detector structures 
int     n_det              #I number of elements in dptr
real    readper            #I percent of data to read (0.0 - 1.0)
bool    info_msgs          #I true = give informational msgs

# LOCAL VARS

int nrows                  # number of rows
int pct, oldpct            # percent, old percent read
int tossed                 # how many events we discard
double time_event          # time event occurred
int dptr_index             # index into dptr for event detector number
int quad                   # event "quadrant"
int row                    # row number
int bin                    # bin number
int count                  # how many we actually stuck in bins
int beyond                 # how many were after the last buffer
bool have_event            # flag: true if we have a good event
int  detoff_cnt            # how many were for turned off detectors

# FUNCTIONS
int  tbpsta()

bool get_next_event()
bool lt_epsilon()
bool ge_epsilon()

begin

  # get number of rows in event table

  nrows = tbpsta (tp, TBL_NROWS)
  nrows = nrows * readper

  # Find telemetered counts.
  # Go through all the table0 events & stuff them into histogram
  # buckets.  The buckets are all 2 Major Frames wide, 
  # to cover the same intervals as the 
  # quadrant counts monitors "det#Q#sf".
  # There may be time gaps between the buckets due to data gaps
  # and table1 flushes, so not all photons may go into a bucket.
  # This routine assumes that the events are time sorted.
  # Use epsilon comparisons so that we don't get into trouble
  # from floating point roundoff problems.
  
  if (info_msgs) {
     call eprintf ("Reading %3.0f percent of events table\n")
     call pargr(readper*100.0)
  }

  bin = 1
  row = 1
  tossed = 0
  count = 0
  detoff_cnt = 0

  # indicate percent read
  oldpct = 0

  have_event = get_next_event(tp,cp,tab0_format,time_event,dptr_index,
     quad,row,nrows,dptr,n_det,detoff_cnt)

  # go until we run out of events or histogram bins
  while ( (have_event) && (bin <= nbins)) {

      # *** if (time_event < start_time[bin])  ***
     if (lt_epsilon(time_event,start_time[bin],EVENT_EPSILON)) {

        # throw out current event & get the next one
        tossed = tossed + 1   

        have_event = get_next_event(tp,cp,tab0_format,time_event,dptr_index,
           quad,row,nrows,dptr,n_det,detoff_cnt)

     }
     # *** else if (time_event >= end_time[bin]) ***
     else if ( ge_epsilon(time_event,end_time[bin],EVENT_EPSILON)) {

        bin = bin + 1          # skip to next bin
     }
     else {  # if (time >= start && time < end)

        # valid event, increment histogram bin  
        call incr_bin(Mems[MP_HIST(dptr[dptr_index],quad)],bin)
        count = count + 1
        have_event = get_next_event(tp,cp,tab0_format,time_event,dptr_index,
           quad,row,nrows,dptr,n_det,detoff_cnt)
     }

     pct = int(10.*real(row)/real(nrows))*10
     if (pct != oldpct) {
        oldpct = pct
        if (info_msgs) { 
           call eprintf ("  %3d%% read\n")
	   call pargi (pct)

        }

     }

  }

  # stats:  account for the non-binned photons

  if ((nrows - count > 0) && info_msgs) {
     call eprintf("%10d events out of %d were not used\n")
     call pargi(nrows - count)
     call pargi(nrows)

     call eprintf("%10d came before the first quad counts interval")
     call pargi(tossed)
     call eprintf(" or in data gaps\n")

     call eprintf("%10d were beyond the last quad counts interval\n")
     beyond = nrows - (row - 1)
     if (have_event) beyond = beyond + 1  # were we holding one?
     call pargi(beyond)
     
     call eprintf("%10d were from detectors not selected\n")
     call pargi(detoff_cnt)

  }
  
  call tbtclo (tp)

end
#-------------------------------------------------------------

# get the next valid event from the events table
#
bool procedure get_next_event(tp,cp,tab0_format,time,dptr_index,quad,row,
   nrows,dptr,n_det,detoff_cnt)  #LOCAL

pointer tp      #I table pointer
pointer cp[ARB] #I column pointers
int tab0_format #I format of table0 (event table)
double time     #O event time 
int dptr_index  #O index into dptr for the returned event's detector number
int quad        #O quadrant number
int row         #I/O row in table to start fetching
int nrows       #I number of rows in table
pointer dptr[ARB]  #I detector structures 
int n_det       #I length of dptr array
int detoff_cnt  #I/O  count of events for turned off detectors

# LOCAL VARS
bool   bad_event           # loop control flag
int    y                   # detector y value
int    x                   # det x value
int    i                   # dptr index
int    event_det           # detector number of event

# FUNCTIONS
int    pb_quadrant()
int    det2dptr_index()

# read next event row from table tp
# skips rows with invalid detector numbers
# increments the row variable as needed to get a valid event

# returns true if valid event is found

begin 

    bad_event = true

    while (bad_event && row <= nrows) {

       # get event detector
       call tbegti (tp, cp[PH_D], row, event_det)

       # get the dptr index corresponding to this detector number
       i = det2dptr_index(event_det,dptr,n_det)

       if (i == INDEFI) {
          #  count photon for turned off detector
          detoff_cnt = detoff_cnt + 1
       }
       else {
   
          # get event time
          call tbegtd (tp, cp[PH_TIME], row, time)

          call get_xy(tp,cp,tab0_format,row,MP_YGAIN(dptr[i]), 
             MP_YOFF(dptr[i]), MP_XGAIN(dptr[i]), MP_XOFF(dptr[i]),
             x, y)

          quad = pb_quadrant (dptr[i], x, y) 
          quad = quad + 1
          dptr_index = i
          bad_event = false
      }
      row = row + 1
   }	

   return(!bad_event)

end

# ----------------------------------------------------------
# get xy value for a photon in table0, regardless of format
#
procedure  get_xy(tp,cp,tab0_format,row,ygain,yoff,xgain,xoff,x,y)
pointer tp          #I table pointer
pointer cp[ARB]     #I column pointers
int     tab0_format #I format of table0 (event table)
int     row         #I row in table to fetch
double  ygain       #I y gain
double  yoff        #I y offset
double  xgain       #I x gain
double  xoff        #I x offset for wsz to xy conversion
int     y           #O detector y value
int     x           #O det x value

# LOCAL VARS
double W,S,Z
double XorS,YorZ

begin

   x = INDEFI
   y = INDEFI

   # get x & y, converting as necessary
   if (tab0_format == FORMATXYWSZ)
   {
       call tbegti (tp, cp[PH_X], row, x)
       call tbegti (tp, cp[PH_Y], row, y)
       call tbegtd (tp, cp[PH_W], row, W)
       call tbegtd (tp, cp[PH_S], row, S)
       call tbegtd (tp, cp[PH_Z], row, Z)

       # if X & Y are not already filled in, then convert
       if (IS_INDEFI(y))
       {
          call pb_wsz_to_xy (ygain, yoff, xgain, xoff, W, S, Z, x, y)
       }
   }
   else if (tab0_format == FORMATXY)
   {
       call tbegti (tp, cp[PH_X], row, x)
       call tbegti (tp, cp[PH_Y], row, y)
   }
   else if (tab0_format == FORMATMERGE)
   {
       call tbegtd (tp, cp[PH_W], row, W)
       call tbegtd (tp, cp[PH_XORS], row, XorS)
       call tbegtd (tp, cp[PH_YORZ], row, YorZ)

       # if this is not a WSZ photon just copy it, else convert to X & Y 
       if (IS_INDEFD(W))
       {
          x = XorS
          y = YorZ
       }
       else
       {        
          call pb_wsz_to_xy (ygain, yoff, xgain, xoff, W, XorS,
             YorZ, x, y)
       }

   }
   else
   {
      call error(701,"Invalid tab0 format in get_xytd")
   }


end
#----------------------------------------------------------

#
# Increment histogram bin 
#
procedure incr_bin (hist, bin)    #LOCAL

short		hist[ARB]    #I/O histogram bin
int		bin          #I   bin number

begin
	hist[bin] = hist[bin] + 1
end
#-----------------------------------------------------

# convert from detector number to index in the dptr array
#
int procedure det2dptr_index(det,dptr,n_det)  #LOCAL

int det               #I det no. we're looking for
pointer dptr[ARB]     #I array of detector structures
int n_det             #I length of dtpr array

# LOCAL VARS
int i                 # loop index
int index             # returned dptr index

# returns the index of the dptr that contains det, 
# or INDEFI if not found

begin

   index = INDEFI

   for (i=1; i <= n_det; i=i+1)  {
      if (det == MP_NUM(dptr[i])) {
         index = i
         break
      }
   }
     
   return(index)

end
#-----------------------------------------------------

# create the output ST table
#
procedure make_output_table(out_tp,time_col,dptr,n_det,nrows) #LOCAL

pointer out_tp              #I/O pointer to opened output table
pointer time_col            #O pointer to time column 
pointer dptr[ARB]           #I detector structures
int     n_det               #I length of detector structures list
int     nrows               #I number of table rows to allocate

# LOCAL VARS
int  quad                   # quadrant number
pointer str                 # string buffer
pointer sp                  # stack pointer
int     i                   # loop index


begin

  # columns are:                           
  # -------------                          
  #    time                                
  #    telemetered counts       
  #    quad counts               
  #    sum quad counts 
  #    deadtime correction
  #    deadtime -err  
  #    deadtime +err  
  #    primbsch corr  
  #    primbsch + deadtime corr 

  call smark (sp)

  call salloc (str, SZ_LINE, TY_CHAR)

  # preallocate no. of rows and no. of user params so that
  # the output file doesn't have to be rewritten later.

  call tbpset (out_tp, TBL_ALLROWS, nrows)           # no. of rows
  call tbpset (out_tp, TBL_MAXPAR, MAX_USR_PARS)     # max no. of user params

  # TIME COLUMN
  call tbcdef (out_tp, time_col, TIME_COLNAME, "s", "%g", TY_DOUBLE, 1, 1)

  do i = 1, n_det {

    do quad = 1, N_QUADS {

      # TELEMETERED COUNTS

      call sprintf (Memc[str], SZ_LINE, TM_COLNAME)
      call pargi (MP_NUM(dptr[i]))
      call pargi (quad-1)
      call tbcdef (out_tp, MP_TM_COL(dptr[i],quad), Memc[str], "count", 
                   "%d", TY_SHORT, 1, 1)



      # QUADRANT COUNTS

      call sprintf (Memc[str], SZ_LINE, QUAD_COLNAME)
      call pargi (MP_NUM(dptr[i]))
      call pargi (quad-1)
      call tbcdef (out_tp, MP_QUAD_COL(dptr[i],quad), Memc[str], "count", 
                   "%d", TY_SHORT, 1, 1)

    }

    # SUM QUADRANT COUNTS
    call sprintf (Memc[str], SZ_LINE, SUMQ_COLNAME)
    call pargi (MP_NUM(dptr[i]))
    call tbcdef (out_tp, MP_SUMQ_COL(dptr[i]), Memc[str], "count", 
                 "%d", TY_INT, 1, 1)


    # DEADTIME CORRECTION
    call sprintf (Memc[str], SZ_LINE, DEAD_COLNAME)
    call pargi (MP_NUM(dptr[i]))
    call tbcdef (out_tp, MP_DEAD_COL(dptr[i]), Memc[str], "", 
                   "%g", TY_REAL, 1, 1)


    # DISABLED deadtime error columns
   if (false) {
    # DEADTIME NEG ERROR
    call sprintf (Memc[str], SZ_LINE, DEAD_NERR_COLNAME)
    call pargi (MP_NUM(dptr[i]))
    call tbcdef (out_tp, MP_NERR_DEAD_COL(dptr[i]), Memc[str], "error", 
                 "%g", TY_REAL, 1, 1)

    # POSITIVE DEADTIME ERROR
    call sprintf (Memc[str], SZ_LINE, DEAD_PERR_COLNAME)
    call pargi (MP_NUM(dptr[i]))
    call tbcdef (out_tp, MP_PERR_DEAD_COL(dptr[i]), Memc[str], "error", 
                 "%g", TY_REAL, 1, 1)
   }

  }

  do i = 1, n_det {

    do quad = 1, N_QUADS {

      # PRIMBSCH CORRECTION
      call sprintf (Memc[str], SZ_LINE, PRIMB_COLNAME)
      call pargi (MP_NUM(dptr[i]))
      call pargi (quad-1)
      call tbcdef (out_tp, MP_PRIM_COL(dptr[i],quad), Memc[str], "", 
                   "%g", TY_REAL, 1, 1)

      # PRIMBSCH + DEADTIME CORRECTION

      call sprintf (Memc[str], SZ_LINE, DEADPRIMB_COLNAME)
      call pargi (MP_NUM(dptr[i]))
      call pargi (quad-1)
      call tbcdef (out_tp, MP_COMB_CORR_COL(dptr[i],quad), Memc[str],
         "", "%g", TY_REAL, 1, 1)
    }
  }

  call tbtcre (out_tp)

  call sfree (sp)

end
#----------------------------------------------------------

# label Det#Q#tm monitors in the table header as rates
procedure label_rates(tp,rate_index_start,dptr,n_det)  #LOCAL

pointer tp               #I/O  table pointer
int     rate_index_start #I    rate index to start with 
pointer dptr[ARB]        #I    det pointer structures
int     n_det            #I    no. of structures

int i                # loop index
int rate_index       # suffix for RATE??? labels
pointer sp           # stack pointer
pointer colname_str  # column name string
pointer ratename_str # rate name label string
int quad             # quadrant number

begin 

  call smark (sp)

  call salloc (colname_str, SZ_LINE, TY_CHAR)
  call salloc (ratename_str, SZ_LINE, TY_CHAR)

  rate_index = rate_index_start

  do i = 1, n_det {

    do quad = 1, N_QUADS {

       # TELEMETERED COUNTS

       call sprintf (Memc[colname_str], SZ_LINE, TM_COLNAME)
       call pargi (MP_NUM(dptr[i]))
       call pargi (quad-1)

       call sprintf (Memc[ratename_str],SZ_LINE, "RATE%-3d")
       call pargi(rate_index)

       # label Det#Q#tm monitors in the table header as rates
       call tbhadt (tp, Memc[ratename_str], Memc[colname_str])

       rate_index = rate_index + 1

      }
   }

  call sfree (sp)

end

#----------------------------------------------------------

# compute the primbsch & deadtime corrections & write to output table
#
procedure primb_dead(out_tp,dead_in,dead_out,deadtbl_nrows,dptr,n_det,
   bin_valid,nbins,warn_msgs,info_msgs,do_neg_primb)   #LOCAL

pointer out_tp           #I/O output table pointer
real    dead_in[ARB]     #I deadtime table input
real    dead_out[ARB]    #I deadtime table output
int     deadtbl_nrows    #I no. of rows in deadtime table
pointer dptr[ARB]        #I/O ptr to detector information structs
int     n_det            #I length of dptr array
bool    bin_valid[ARB]   #I flag: true if bin is valid time interval
int     nbins            #I number of histogram bins
bool    warn_msgs        #I if true print warning messages
bool    info_msgs        #I if true give info messages
bool    do_neg_primb     #I if true then do neg primbsch corrections

# LOCAL VARS
pointer sumquad          # sum of quadrants array
pointer isumquad	 # sum of quad array as int
pointer deadtime         # deadtime array 
pointer deaderr          # deadtime error array
pointer sqrtsumq         # array for sqrt  of the sum of the quadrants
pointer pcorr            # primbsch correction array
int     i                # dptr index
int     quad             # quadrant number
int     neg_primb_err    # count of "negative primbsch" errors corrected
int     neg_primb_fnd    # no. of neg primbsch errors found
int     badcount         # count of bad values

begin

  # Compute the primbsch correction for each quadrant.  This is the quadrant
  # counts divided by telemetered counts and multiplied by the TIF deadtime 
  # correction which is read from a look-up table made by JVV and PATJ.
  #
  # Weird things in the data:  sometimes there is one more telemetered count
  # than quadrant counts.  This is not supposed to be possible.  Each such
  # "negative primbsching" event is compensated by a missing telemetered
  # count in the next reporting.  I "correct" these errors by canceling 
  # each "negative" event with the following "positive" one.  This is to
  # be considered merely a workaround.

  call calloc (sumquad, nbins, TY_REAL)
  call calloc (isumquad, nbins, TY_INT)
  call calloc (deadtime, nbins, TY_REAL)
  call calloc (deaderr, nbins, TY_REAL)
  call calloc (sqrtsumq, nbins, TY_REAL)
  call calloc (pcorr, nbins, TY_REAL)

  do i = 1, n_det {

    # find sum of quadrant counts
    call aclrr (Memr[sumquad], nbins)
    do quad = 1, N_QUADS {
      call iaddr (Memr[sumquad], Memr[MP_QCNT(dptr[i],quad)], 
                   Memr[sumquad], nbins)
    }
 
    # write out the sum of the quadrants as ints
    call ichtri (Memr[sumquad], Memi[isumquad], nbins)
    call tbcpti (out_tp, MP_SUMQ_COL(dptr[i]), Memi[isumquad], 1, nbins)

      call pb_make_dead (dead_in, dead_out, deadtbl_nrows, Memr[sumquad], 
                         Memr[deadtime], nbins, warn_msgs, badcount)

      if ((badcount > 0) && warn_msgs) {
        call eprintf("Warning: %d values not found in deadtime table")
        call pargi(badcount)
        call eprintf(" for detector %1d\n")
        call pargi(MP_NUM(dptr[i]))
      }

      call imulkr (Memr[deadtime], 2.0, Memr[deadtime], nbins)
      call idivr (Memr[deadtime], Memr[sumquad], Memr[deadtime], nbins)

      # write out the raw deadtimes
      call tbcptr (out_tp, MP_DEAD_COL(dptr[i]), Memr[deadtime], 1, nbins)

      # find neg & positive deadtime errors:
      #  Lookup 3 deadtimes:
      #      D  = deadtime(N)
      #      Da = deadtime(N+sqr(N))
      #      Db = deadtime(N-sqr(N))
      #      -err = D-Da   +err = Db-D

    #  ****  DEADTIME ERRORS DISABLED  ****
    if (false) { 
      # *** deadtime pos error ***
      call isqrr (Memr[sumquad], Memr[sqrtsumq], nbins)
      call iaddr (Memr[sumquad],Memr[sqrtsumq], Memr[sqrtsumq], nbins)
      call pb_make_dead (dead_in, dead_out, deadtbl_nrows, Memr[sqrtsumq], 
                         Memr[deaderr], nbins, warn_msgs, badcount)
      call imulkr (Memr[deaderr], 2.0, Memr[deaderr], nbins)
      call idivr (Memr[deaderr], Memr[sqrtsumq], Memr[deaderr], nbins)
      # diff from nominal deadtimes
      call isubr (Memr[deaderr], Memr[deadtime], Memr[deaderr], nbins)
 
      # write out the positive error deadtimes
      call tbcptr (out_tp, MP_PERR_DEAD_COL(dptr[i]), Memr[deaderr],
          1, nbins)


      # *** deadtime neg error ***
      call isqrr (Memr[sumquad], Memr[sqrtsumq], nbins)
      call isubr (Memr[sumquad],Memr[sqrtsumq], Memr[sqrtsumq], nbins)
      call pb_make_dead (dead_in, dead_out, deadtbl_nrows, Memr[sqrtsumq], 
                         Memr[deaderr], nbins, warn_msgs, badcount)
      call imulkr (Memr[deaderr], 2.0, Memr[deaderr], nbins)
      call idivr (Memr[deaderr], Memr[sqrtsumq], Memr[deaderr], nbins)
      # diff from nominal deadtimes
      call isubr (Memr[deadtime], Memr[deaderr], Memr[deaderr], nbins)
 
      # write out the negative error deadtimes
      call tbcptr (out_tp, MP_NERR_DEAD_COL(dptr[i]), Memr[deaderr],
          1, nbins)
    }

    do quad = 1, N_QUADS {

      if (do_neg_primb) {
         # test for "negative" primbsching
         call pb_corr_vals (Memr[MP_QCNT(dptr[i],quad)], 
                           Memr[MP_TCNT(dptr[i],quad)], nbins, 
                           neg_primb_fnd, neg_primb_err)

          # report counts even if info_msgs is off 
          if (neg_primb_fnd > 0)  {
            #if ((neg_primb_fnd > 0) && info_msgs) {
            call eprintf("Found %6d 'negative primbsch' errors")
            #*** DISBLED ****
            #  call eprintf("Corrected %6d of %6d 'negative primbsch' errors")
            #  call pargi(neg_primb_err)
            #*** DISBLED
            call pargi(neg_primb_fnd)
            call eprintf(" (det=%1d quad=%1d)\n")
            call pargi(MP_NUM(dptr[i]))
            call pargi(quad-1)
         }
      }

      # primbsch_correction = quad_cnts / telemetered_counts
      call idivr (Memr[MP_QCNT(dptr[i],quad)], 
                   Memr[MP_TCNT(dptr[i],quad)], Memr[pcorr], nbins)

      # mark those intervals where primbsch correction is not valid
      # (ie, where one or both of the 2 major frames is missing)
      call mark_indef(Memr[pcorr],bin_valid,nbins)      

      # write out just primbsch correction 
      call tbcptr (out_tp, MP_PRIM_COL(dptr[i],quad), Memr[pcorr], 1, nbins)

      # combined primbsch + deadtime correction
      call imulr (Memr[pcorr], Memr[deadtime], Memr[pcorr], nbins)
      call tbcptr (out_tp, MP_COMB_CORR_COL(dptr[i],quad), Memr[pcorr], 1, nbins)
    }
  }

  call mfree (sumquad,  TY_REAL)
  call mfree (isumquad, TY_INT)
  call mfree (deadtime, TY_REAL)
  call mfree (pcorr,    TY_REAL)
  call mfree (deaderr,  TY_REAL)
  call mfree (sqrtsumq,  TY_REAL)

end


#----------------------------------------------------------

# write telemetered counts to output ST table
#
procedure wr_tm_counts(out_tp,dptr,n_det,nbins)  #LOCAL

pointer out_tp           #I/O output table pointer
pointer dptr[ARB]        #I ptr to detector information structs
int     n_det            #I length of dptr array
int     nbins            #I number of histogram bins

# LOCAL VARS
int i                    # index
int quad                 # quad number

begin

  do i = 1, n_det {

    do quad = 1, N_QUADS {

      call tbcpts (out_tp, MP_TM_COL(dptr[i],quad), 
                   Mems[MP_HIST(dptr[i],quad)], 1, nbins)

      # convert short to real
      call ichtsr (Mems[MP_HIST(dptr[i],quad)],
          Memr[MP_TCNT(dptr[i],quad)],nbins)

    }
  }

end


#----------------------------------------------------------

# copy quadrant counts from table2 to the output table
#
procedure cp_quad_cnts(tp2, out_tp, dptr, n_det, readper)  #LOCAL

pointer tp2                 #I table2 pointer
pointer out_tp              #I/O output table pointer
pointer dptr[ARB]           #I/O detector structures 
int n_det                   #I length of dptr array
real readper                #I percent of data to read

  # Get quadrant counts from Table2 and write out.  The copied
  # counts are saved in QCNT for use in computing the correction later

# LOCAL VARS
int nrows                   # no. of rows
int i                       # loop index
int quad                    # quad no.
pointer sp                  # stack pointer
pointer str                 # string buffer

# FUNCTIONS
int tbpsta()

begin

  call smark (sp)

  call salloc (str, SZ_LINE, TY_CHAR)

  nrows = tbpsta (tp2, TBL_NROWS)
  nrows = nrows * readper

  do i = 1, n_det {

    do quad = 1, N_QUADS {

      call sprintf (Memc[str], SZ_LINE, TAB2_QUAD_COLNAME)
      call pargi (MP_NUM(dptr[i]))
      call pargi (quad-1)

      call pb_copy_col (tp2, nrows, out_tp, nrows, Memc[str], 
                        MP_QUAD_COL(dptr[i],quad), 
                        Memr[MP_QCNT(dptr[i],quad)])
    }
  }

  call sfree(sp)

end
#----------------------------------------------------------

#  allocate space for quadrant histograms and quadrant counts
#
procedure pb_det_setup (dp, nbins)    #LOCAL

pointer	   dp        #I/O pointer to detector structure
int	   nbins     #I no. of bins


# LOCAL VARS
int		quad

begin

  do quad = 1, N_QUADS {
    call calloc (MP_HIST(dp, quad), nbins, TY_SHORT)
    call calloc (MP_QCNT(dp, quad), nbins, TY_REAL)
    call calloc (MP_TCNT(dp, quad), nbins, TY_REAL)
  }
end
#----------------------------------------------------------

#  deallocate space for quadrant histograms
#
procedure pb_det_free (dp)   #LOCAL

pointer	   dp        #I/O pointer to detector structure

# LOCAL VARS
int	   quad      # quad no.

begin
  do quad = 1, N_QUADS {
    call mfree (MP_HIST(dp, quad), TY_SHORT)
    call mfree (MP_QCNT(dp, quad), TY_REAL)
    call mfree (MP_TCNT(dp, quad), TY_REAL)
  }

end
#---------------------------------------------------------
#
# copy a column:  read short, write short, change to real
#
procedure pb_copy_col (tpin, nin, tpout, nout, col_name, ocp, rbuf) #LOCAL

pointer tpin              #I input table ptr
int     nin               #I no. of input values
pointer tpout             #I/O output table pointer
int     nout              #I no. of output values
char    col_name[ARB]     #I name of column
pointer ocp               #I output column pointer
real    rbuf[ARB]         #I real buffer

# LOCAL VARS
pointer cp                # column pointer
pointer sp                # stack pointer
pointer vals              # array
pointer junk              # junk array

begin
  call smark (sp)

  call salloc (vals, nin, TY_SHORT)
  call salloc (junk, nin, TY_BOOL)

  call tbcfnd (tpin, col_name, cp, 1)
  if (cp == NULL) {
     call eprintf("Warning:  Could not find column %s\n")
     call pargstr(col_name)
  }
  else {
     call tbcgts (tpin, cp, Mems[vals], Memb[junk], 1, nin)
     call tbcpts (tpout, ocp, Mems[vals], 1, nout)
     call ichtsr (Mems[vals], rbuf, nout)

  }
  call sfree (sp)
end

#----------------------------------------------------------

#
#  Compute X&Y from W,S,Z
#
procedure pb_wsz_to_xy (ygain, yoff, xgain, xoff, w, s, z, x, y)  #LOCAL

double  ygain      #I  y gain
double  yoff       #I  y offset
double  xgain      #I  x gain
double  xoff       #I  x offset
int	y          #O  y position
int     x          #O  x position
double	w, s, z    #I  w,s,z values

# LOCAL VARS
double	ph         # pulse height

begin

   x = 0
   y = 0
   ph = w + s + 2*z
   if (ph > EPSILOND) {
      y = (ygain * s / ph + yoff )/ 2. + 0.5
      x = (xgain * w / ph + xoff )/ 2. + 0.5
   }
end

#----------------------------------------------------------

#
# Find quadrant of an event (labeled 0-3)
#
int procedure pb_quadrant (det_ptr, x, y)  #LOCAL

pointer  det_ptr   #I det struct
int      x         #I detector x-value
int      y         #I detector y-value

# LOCAL VARS
int      q         # quad num

#   TYPE VERT           TYPE SQUARE
#             3
#         Y3 -----        1  |  3
#             2              |
#         Y2 -----       ---------- Y
#             1              |
#         Y1 -----        0  |  2
#             0
#                            X

begin

  if (MP_QBOUND_TYPE(det_ptr) == PRIMB_VERT_TYPE) {

     # detector "quads" are really a vertical stack (Y1,Y2,Y3)
     if      ( y >= MP_QBOUND_Y3(det_ptr) )   q = 3
     else if ( y >= MP_QBOUND_Y_Y2(det_ptr))  q = 2
     else if ( y >= MP_QBOUND_X_Y1(det_ptr))  q = 1
     else q = 0
  }
  else {
     # detector is really quadrants (X,Y)
     if ( x >= MP_QBOUND_X_Y1(det_ptr)) {
          if (y >= MP_QBOUND_Y_Y2(det_ptr) ) q = 3
            else  q = 2
     }
     else  {
          if (y >= MP_QBOUND_Y_Y2(det_ptr) ) q = 1
            else q = 0
     }
  }
  return(q)

end
#---------------------------------------------------------

# Correct for miscounted values, i.e., "negative primbsching"
#
# Sometimes there is one more telemetered count
# than quadrant counts.  This is not supposed to be possible.  Each such
# "negative primbsching" event is compensated by a missing telemetered
# count in the next reporting.  I "correct" these errors by canceling 
# each "negative" event with the following "positive" one.  This is to
# be considered merely a workaround.
#
# NOTE:  the correction has been disabled, so that negative
#  primbsch corrections are merely reported.

procedure pb_corr_vals (quad_cnts, tm_cnts, n, errfnd, errcnt)  #LOCAL

real	quad_cnts[ARB]        #I quad counts
real    tm_cnts[ARB]          #I telemetered counts
int	n                     #I length of arrays
int     errfnd                #O number of errors found
int     errcnt                #O count of errors "corrected"

# LOCAL VARS      
int	dif           # difference
int     i             # loop index

begin

  errfnd = 0
  errcnt = 0

  # When the telemetered counts are greater than the quadrant
  # counts in the current bin, try to adjust the quadrant counts 
  # in this & the next bin.  If the adjustment would make the next
  # bin's tm counts exceed the quad counts, then don't adjust.

  # Note that this will not properly handle the case when an
  # invalid quadrant count has propagated across 2 bins.

  do i = 1, n-1 {

    if (IS_INDEFR(quad_cnts[i]) || IS_INDEFR(tm_cnts[i]) ||
        IS_INDEFR(quad_cnts[i+1]) || IS_INDEFR(tm_cnts[i+1]) ) {
       next
    }

    # are telemetered counts > quadrant counts?
    dif = nint(tm_cnts[i] - quad_cnts[i])
    if (dif > 0) {

       errfnd = errfnd + 1

       # will adjusting mess up the next bin?
       if (nint(quad_cnts[i+1] - dif) >= nint(tm_cnts[i+1]) ) {

          # ***** DISABLED ******
          #quad_cnts[i] = quad_cnts[i] + dif
   	  #quad_cnts[i+1] = quad_cnts[i+1] - dif

          errcnt = errcnt + 1

       }

     }
  }

  if (!IS_INDEFR(quad_cnts[n]) && !IS_INDEFR(tm_cnts[n])) {
    dif = nint(tm_cnts[n] - quad_cnts[n])
    if (dif > 0) {
          # *** DISABLED ***
	  # quad_cnts[n] = quad_cnts[n] + dif
          errfnd = errfnd + 1
          errcnt = errcnt + 1
    }
  }
end
#----------------------------------------------------------

#
# Compute the TIF deadtime correction from the sum of the quadrant counts
#  by simple look-up in the table created by JVV and PATJ
#
procedure pb_make_dead (deadtbl_in, deadtbl_out, dead_len, quad, dead, 
   nbins, warn_msgs, badcount)  #LOCAL

real	deadtbl_in[ARB]     #I  deadtime table input  values
real    deadtbl_out[ARB]    #I  deadtime table output values
int     dead_len            #I  length of deadtime table
real    quad[ARB]           #I  quad counts to look up in deadtime table
real    dead[ARB]           #O  looked-up deadtimes
int     nbins               #I  length of quad & dead vectors
bool    warn_msgs           #I  if false, don't print warnings
int     badcount            #O  count of values not in deadtime table

# LOCAL VARS
int	i                   # loop index

#FUNCTIONS
bool fp_equalr()

begin

  badcount = 0
  do i = 1, nbins {

    if (IS_INDEFR(quad[i])) {
	  dead[i] = INDEFR
    }
    else {

      if (fp_equalr(quad[i],0.0)) 
          dead[i] = 0.0 
      else {

         # search deadtime table, with linear interpolation
         # (return value of zero means it wasn't in the table)

	 # if we're off the high end of the table, return
	 # the last value 

         if (quad[i] > deadtbl_in[dead_len])
	    dead[i] = deadtbl_out[dead_len]
	 else
	 {
            call plf_value(deadtbl_in,deadtbl_out,dead_len,quad[i],dead[i])
            if (fp_equalr(dead[i],0.0)) {
               dead[i] = INDEFR
               badcount = badcount + 1
            }
         }
      }
    }
  }

end
#---------------------------------------

# mark invalid array elements as INDEF
procedure mark_indef(a,valid,len)  #LOCAL

real a[ARB]         #O marked array
bool valid[ARB]     #I good/bad list
int  len            #I length of both arrays

# LOCAL VARS
int i

begin

   # look at each element of 'valid' array
   # where false, mark 'a' as indef, else leave 'a' alone

   for (i=1;i<=len;i=i+1)  {
      if (! valid[i] ) a[i] = INDEFR
   }
end
#--------------------------------------

#
# change datatype of a vector with check for INDEF values
# (real to int) 

procedure ichtri (a, b, n)    #LOCAL

real    a[ARB]      #I int array
int     b[ARB]      #O real array
int     n           #I no. of elements

int             i

begin

  do i = 1, n {
        if (IS_INDEFR(a[i])) {
          b[i] = INDEFI
        }
        else {
          b[i] = a[i]
        }
  }
end

#--------------------------------------

#
# change datatype of a vector with check for INDEF values
# (short to real) 

procedure ichtsr (a, b, n)    #LOCAL

short   a[ARB]      #I short array
real    b[ARB]      #O real array
int     n           #I no. of elements

int             i

begin

  do i = 1, n {
        if (IS_INDEFS(a[i])) {
          b[i] = INDEFR
        }
        else {
          b[i] = a[i]
        }
  }
end

#----------------
# print out an integer vector, all on 1 line
procedure pr_ivec(iarray,len,msg)  #LOCAL 
int iarray[ARB]    #I vector
int len            #I length of vector
char msg[ARB]      #I message to prepend

int i

begin

   call eprintf("%s: ")
   call pargstr(msg)
   for (i=1; i<=len; i=i+1) {
      call eprintf(" %d")
      call pargi(iarray[i])
   }
   call eprintf("\n")


end

#-----------------------------------------
# get all the column pointers for table0
procedure get_tab0_cptr(tp,cptr,n_cptr,tab0_format)  #Local

pointer tp           #I table pointer
pointer cptr[ARB]    #O array of column pointers
int     n_cptr       #I len of cptr list
int     tab0_format  #O format of tab0

begin

  # Fetch pointers to all the event columns.
  # Only certain columns will be present, depending on the format.
  # Format:             Columns:  
  # XYWSZ,TELEMETRY:    time,x,y,w,s,z,d
  # XY format      :    time,x,y,d
  # MERGE format   :    time,w,XorS,YorZ,d
  
  call tbcfnd(tp, PH_TIME_COLNAME, cptr[PH_TIME], 1)
  call tbcfnd(tp, PH_X_COLNAME, cptr[PH_X], 1)
  call tbcfnd(tp, PH_Y_COLNAME, cptr[PH_Y], 1)
  call tbcfnd(tp, PH_W_COLNAME, cptr[PH_W], 1)
  call tbcfnd(tp, PH_S_COLNAME, cptr[PH_S], 1)
  call tbcfnd(tp, PH_Z_COLNAME, cptr[PH_Z], 1)
  call tbcfnd(tp, PH_D_COLNAME, cptr[PH_D], 1)
  call tbcfnd(tp, PH_XORS_COLNAME, cptr[PH_XORS], 1)
  call tbcfnd(tp, PH_YORZ_COLNAME, cptr[PH_YORZ], 1)

  # check for common columns
  if (cptr[PH_TIME] == NULL || cptr[PH_D] == NULL )
  {
      call error(670,"Invalid event table, could not find expected columns")
  }

  # verify particular formats
  if (cptr[PH_XORS] != NULL)
  {
     # must be MERGE format
     tab0_format = FORMATMERGE
     if (cptr[PH_W] == NULL || cptr[PH_YORZ] == NULL)
     {
	call error(671,"Missing columns for event table MERGE format")  
     }
  }
  else if (cptr[PH_W] == NULL)
  {
     # no PhotonW, must be XY mode
     tab0_format = FORMATXY
     if (cptr[PH_X] == NULL || cptr[PH_Y] == NULL)
     {
	call error(672,"Missing columns for event table XY format")  
     }
  }
  else  
  {
     # must be XYWSZ mode (this includes TELEMETRY mode, i.e., xy filled in)
     tab0_format = FORMATXYWSZ
     if (cptr[PH_X] == NULL || cptr[PH_Y] == NULL || 
         cptr[PH_W] == NULL || cptr[PH_S] == NULL || cptr[PH_Z] == NULL)
     {
	call error(673,"Missing columns for event table XYWSZ/TELEMETRY format")  
     }

  }

end
#------------------------------------------
pointer procedure get_dir (filepath)
char filepath[ARB]    #I path name including file

# given a full path filename, return just the path

# local vars
int len
int pos

pointer buff
char ch

# functions
int strldx()

begin

   buff = NULL

   # find last slash, return everything before 
   ch='/'
   pos = strldx (ch, filepath)
   len = pos - 1

   if (len > 0) {
      call malloc(buff, len+1, TY_CHAR)
      call strcpy (filepath, Memc[buff], len)
   }

   return(buff)

end
