# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  effexp.x
# 	Author	:  tnk
# 	Date	:  03/23/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) effexp.x -r1.3 03/23/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	09/07/94   tnk	rm duplicate df_print code
# 	03/25/94   tnk	converted effexp to use df
# 	03/23/94   tnk	date and time created 94/03/23 17:08:01 by tnk
# H-
# 
#

# @(#)effexp.x	1.3 9/7/94 UCB CEA EGO

include <dfset.h>

#  effexp - routine to compute effective exposure w/ corrections
#  Takes a set of valid time intervals, and a correction
#  column, and computes an effective exposure time	

real procedure effexp(xs, xe, xlen, cs, ce, corr, nc)

double	xs[xlen], xe[xlen]	#I valid time range list
int	xlen			#I range list length
double	cs[nc]			#I start time of correction values
double  ce[nc]			#I end time of correction values
real	corr[nc]		#I correction values
int	nc			#I number of correction values

real	out			#O effective exposure
int 	i
pointer df_gt
pointer df_corr
pointer df_product
real    c

# FUNCTIONS
double  df_integ()
pointer df_open()

begin

   # Exposure undefined due to open range or no time filter
   if (xlen < 1 || IS_INDEFD(xs[1]) || IS_INDEFD(xe[xlen]) ) {
	out = INDEFR
	return(out)
   }

   # turn the valid time intervals into a "boolean" PD df
   df_gt = df_open(DF_PD,TY_DOUBLE,TY_DOUBLE)
   call df_pedge(df_gt, 0.0d0, 0.0d0)

   i=1
   call dfpx0(df_gt,xs[i])
   call df_dputd(df_gt,xe[i],1.0d0,1)
   for (i=2; i <= xlen; i=i+1)
   {
      call df_dputd(df_gt,xs[i],0.0d0,1)   #interval start=0
      call df_dputd(df_gt,xe[i],1.0d0,1)   #interval end=1
   }

   # turn the correction table into a PD
   df_corr = df_open(DF_PD,TY_DOUBLE,TY_DOUBLE)
   call df_pedge(df_corr, 1.0d0, 1.0d0)

   for (i=1; i <= nc; i=i+1)
   {
      # if there's a gap (i.e., if the start of this interval is 
      # not the same as the end of the last one, put out a "zero"
      # for the start of interval
      if (i==1)
         call df_px0(df_corr,cs[i])         
      else if (ce[i-1] != cs[i])
         call df_dputd(df_corr,cs[i],0.0d0,1)   #interval start=0

      # put out the inverted correction factor at the end of interval
      c=corr[i]
      if ( (c == 0.0e0) || (c == INDEFR))
         c = 1.0e0
      c = 1.0e0 / c
      call df_dputd(df_corr,ce[i],double(c),1)   #interval end=correction
   }

   # multiply good times by the correction times (i.e., "AND" them)
   df_product = df_open(DF_PD,TY_DOUBLE,TY_DOUBLE)
   call df_pedge(df_product, 0.0d0, 0.0d0)

   call dfmul(df_gt, df_corr, df_product)

   # integrate to get the corrected exposure time
   out = real(df_integ(df_product))

   call df_close(df_gt)
   call df_close(df_corr)
   call df_close(df_product)

   return(out)

end

