# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  mkcorrim.x
# 	Author	:  tnk
# 	Date	:  03/23/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) mkcorrim.x -r1.14 03/23/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/20/97   bboyd	replaced illegal private IO_NEVENTS with qpio_stati call
# 	01/06/97   tnk	now checks for NULL correction factor vectors
# 	09/08/95   tnk	added check for correction values equalt to zero
# 	08/29/94   tnk	reworked qpoe event field offsets
# 	07/19/94   tnk	reworked error band wcs handling
# 	07/13/94   tnk	rect & block handling
# 	07/12/94   tnk	fixed WCS for rect & block factors
# 	07/07/94   tnk	added dx,dy check
# 	07/07/94   tnk	added error band handling
# 	05/02/94   tnk	adjusted epsilon compare routines
# 	03/30/94   tnk	switched event_epsilon arg
# 	03/28/94   tnk	added infomsgs flag
# 	03/25/94   tnk	converted effexp to use df
# 	03/23/94   tnk	date and time created 94/03/23 17:08:03 by tnk
# H-
# 
#

# @(#)mkcorrim.x	1.14 10/20/97 UCB CEA EGO

include <imhdr.h>
include	<qpioset.h>
include <imio.h>
include <mwset.h>


define   DATABANDID     1
define   ERRBANDID      2

# QPOE type conversion macros
define QPCONV_D Memd[coerce(($1)+($2), TY_SHORT, TY_DOUBLE)]
define QPCONV_S Mems[coerce(($1)+($2), TY_SHORT, TY_SHORT)]
define QPCONV_R Memr[coerce(($1)+($2), TY_SHORT, TY_REAL)]

# Creates corrected image from QPOE event records.
# I.e., merge QPOE file events and correction factors to
# produce a "corrected" output image.
#
#    For each QPOE event
#       C = correction_Factor( Time, corr_vect[n])
#       Image(x,y) += C
#    end
#
#    if errband is yes, then a second plane of the image will
#    be filled with sigma error factors:
#
#    For each QPOE event
#       C = correction_Factor( Time, corr_vect[n])
#       ErrorImage(x,y) += C^2
#    end
#    For all pixels
#       ErrorImage(x,y) = sqrt(ErrorImage(x,y))
#    end
#
#  Expected Event record:
#    TIME  double  
#    X     short   [1-2048]  image coordinates
#    Y     short   [1-2048]
#    DX    short   [0-2047]  detector coords
#    DY    short   [0-2047]
#
# If n_corr_vect ==1, there is a single correction vector in
# corr_fact[1]. 
# If n_corr_vect ==4, there are 4 correction vectors, one
# per quadrant, referenced by the dx,dy event values.
# (some of the 4 may be null pointers)
#
# Note:
#   When QPOE is blocked, it can return events that are outside
# the limits of the image ("partial pixels").  Hence some of the
# range checks below.  These events are discarded.
# 
# The forced buffer flushing code is pretty ugly.
#
#
procedure mk_corr_im(qp,qpio,mwout,im,corr_stime,corr_etime,corr_fact,
   n_corr_vect,lookup_len,event_epsilon,bound_type,bound_x_y1,bound_y_y2,
   bound_y3,infomsgs, errband, time_field, dx_field, dy_field)
pointer  qp              #I input QPOE file ptr
pointer  qpio            #I QP IO ptr
pointer  mwout           #I output wcs
pointer  im              #I/O output image ptr
double   corr_stime[ARB] #I correction time vector (interval start)
double   corr_etime[ARB] #I correction time vector (interval end)
pointer  corr_fact[ARB]  #I ptrs to correction factor vector(s) of type real
int      n_corr_vect     #I no. of corr_fact vectors (1 or 4)
int      lookup_len      #I length of stime/etime/fact vectors
double   event_epsilon   #I epsilon indicating timing accuracy
int 	 bound_type	 #I quad boundary type
int 	 bound_x_y1      #I quad boundaries
int 	 bound_y_y2      #I
int 	 bound_y3        #I
bool     infomsgs        #I give informational messages
bool     errband         #I fill in error band with 1 sigma errors
char     time_field[ARB] #I event struct time field
char	 dx_field[ARB]   #I event struct dx field
char     dy_field[ARB]   #I event struct dy field

# LOCAL VARS
int      n_events
int      n_indefs
int      n_bad_events
int      ev_count
int      pct
int      oldpct
int	 maskval
pointer	 ev		# event pointer
int	 nev
double   time
#real    lambda
#real    theta
int      x
int      y
short    dx
short    dy
real     x1,y1
real     x2,y2
real     c          	 # correction factor
int      out_y      	 # output y line
int      n_notfound
double   minval, maxval
real     minr, maxr
int      time_offset, x_offset, y_offset, dx_offset, dy_offset
char     fld_type
pointer	 out
pointer  databuff
pointer  errbuff
int	 im_size
int 	 quad            # quadrant
pointer  sp
int      linelen
int      xmax,ymax
bool     done
pointer  ct
pointer  param

# FUNCTIONS
long     clktime()
pointer	 coerce()	# pointer converter
pointer  impl3r()
int	 qpio_getevents()
real     corr_lookup()
int      xy2quad()
real	 negfactor()
pointer  mw_sctran()
int	 qpio_stati()
int 	 qplenf()

begin

  call smark(sp)

  param = qpio_stati(qpio, QPIO_PARAMP)
  n_events = qplenf(qp, Memc[param])

  if (infomsgs)
  {
     call eprintf("Number of prefiltered QPOE events: %d\n")
     call pargi(n_events)
  }
  im_size = IM_LEN(im,1)
  xmax = IM_LEN(im,1)
  ymax = IM_LEN(im,2)
  linelen = im_size

  # set up transformation MWCS for both x & y axes
  ct = mw_sctran(mwout, "physical", "logical", 3)

  # get qpoe field offsets
  x_offset = qpio_stati(qpio, QPIO_EVXOFF)
  y_offset = qpio_stati(qpio, QPIO_EVYOFF)
  call qpfld_off(qp, time_field, fld_type, time_offset)
  if (n_corr_vect > 1) {
     call qpfld_off(qp, dx_field,   fld_type, dx_offset)
     call qpfld_off(qp, dy_field,   fld_type, dy_offset)
  }
  
  call salloc(databuff, linelen+1, TY_REAL)  # plus 1 for partial junk pixel
  call amovkr(0.0e0, Memr[databuff], linelen)

  if (errband)
  {
     call salloc(errbuff, linelen+1, TY_REAL) # plus 1 for partial junk pix
     call amovkr(0.0e0, Memr[errbuff], im_size)
  }

  # do the data band and error band in separate passes
  # otherwise we take an enormous i/o hit
  
  if (infomsgs)
  {
     call eprintf("reading QPOE events...\n")
  }

  # Put each QPOE event into the line buffer by its X position.
  # Since the input is sorted by y, then x,
  # when y changes, we need a new line.
  maskval = 0
  y = -1
  x = -1
  out_y = -1
  n_indefs = 0
  n_notfound = 0
  ev_count = 0
  n_bad_events = 0
  minval = INDEFD
  maxval = -INDEFD
  out = NULL
  done=false
  oldpct = 0
  while (! done) {

    if (qpio_getevents(qpio, ev, maskval, 1, nev) == EOF)
    {
       done=true
    }
    else {

       ev_count = ev_count + 1
     
       if (infomsgs) 
       {
          # print percent read
          pct = int(10.*real(ev_count)/real(n_events))*10
          if (pct != oldpct) {
              oldpct = pct
              call eprintf ("  %3d%% read\n")
              call pargi (pct)
          }
       }

       # extract the field values from the QPOE event
       time = QPCONV_D(ev,time_offset)
       x1   = real(QPCONV_S(ev,x_offset))
       y1   = real(QPCONV_S(ev,y_offset))
       if (n_corr_vect > 1) {
          dx   = QPCONV_S(ev,dx_offset)
          dy   = QPCONV_S(ev,dy_offset)
       }
    }

    #convert coords (to handle rect regions & blocking)
    call mw_c2tranr(ct,x1,y1,x2,y2)
    x = nint(x2)
    y = nint(y2)

    if ( ( y < 1 || y > ymax+1 ) ||     # +1 for last partial pixel
         ( x < 1 || x > xmax+1 ) ) {

       n_bad_events = n_bad_events + 1
    }
    else {
       # flush buffer if we switch lines or we're done
       if (y > out_y || done) {  

         # (but not first time through)
         if (out_y >= 1 || done)
         {
            # make sure we're not processing an invalid partial pixel line
            if (out_y <= ymax)  
            {
               if (errband)  
               {
                   # take sqrt of error factors (to make stdev)
                   call asqrr(Memr[errbuff],Memr[errbuff], linelen, negfactor)
                   out = impl3r(im, out_y, ERRBANDID)
                   call amovr(Memr[errbuff], Memr[out], linelen)
                   call amovkr(0.0e0, Memr[errbuff], linelen)

               } else {
               
                  # data band
                  # update minmax
                  call alimr(Memr[databuff], linelen, minr, maxr)
                  if (minr < minval) minval = double(minr)
                  if (maxr > maxval) maxval = double(maxr)

                  out = impl3r(im, out_y, DATABANDID)
                  call amovr(Memr[databuff], Memr[out], linelen)
                  call amovkr(0.0e0, Memr[databuff], linelen)
               }
            }
         }
         out_y = y
       }

       quad = 1
       if (n_corr_vect > 1)
       {
         quad = xy2quad(bound_type,bound_x_y1,bound_y_y2,bound_y3,
            dx, dy)
         quad = quad + 1   # change to valid index
       }

       if (corr_fact[quad] == NULL)
          c = 1    # set correction factor to 1
       else
       {
          c = corr_lookup(time,corr_stime,corr_etime,
              Memr[corr_fact[quad]],lookup_len,n_indefs,n_notfound,
              event_epsilon)
       }

       if (errband)  # added in correction factor, squared
          Memr[errbuff+x-1] = Memr[errbuff+x-1] + (c * c)
       else
          Memr[databuff+x-1] = Memr[databuff+x-1] + c
    }
  }

  if (infomsgs) 
  {
     call eprintf("%d events out of %d processed\n")
     call pargi(ev_count)
     call pargi(n_events)

     if (n_bad_events > 0) {
        call eprintf("%d events had bad fields and were discarded\n")
        call pargi(n_bad_events)
     }

     if (n_indefs > 0) {
        call eprintf("%d zero or INDEF correction values were forced to 1\n")
        call pargi(n_indefs)
     }

     if (n_notfound > 0) {
        call eprintf("%d event times were not in the correction table ")
        call pargi(n_notfound)
        call eprintf("(factor forced to 1)\n")
     }
  }

  # update min/max & timestamps in header
  if (! errband)
  {
     IM_MIN(im) = minval
     IM_MAX(im) = maxval
     IM_MTIME(im) = clktime(long(0))
     IM_LIMTIME(im) = IM_MTIME(im) + 1
  }

  call mw_ctfree(ct)
  call sfree(sp)


end

#-----------------------------------------------
# return the correction factor for a given event
real procedure corr_lookup(ev_time,start_time,end_time,factor_table,table_len,
n_indefs,n_notfound,epsilon)
double 	ev_time	          #I time of event
double  start_time[ARB]   #I lookup start times
double  end_time[ARB]     #I lookup end times
real    factor_table[ARB] #I primbsch factors
int     table_len         #I
int     n_indefs          #I/O  number of indefs found
int     n_notfound        #I/O  number of ev times not found in table
double  epsilon           #I

real factor
int  index

begin

   factor = 1  # return a 1 if the search fails

   # look up the ev_time in the table & find the corresponding
   # correction factor.

   call bin_search(start_time,end_time,table_len,ev_time,epsilon,index)

   if (index > 0) {

      factor = factor_table[index]

      if ((factor == 0.0e0) || (factor == INDEFR))  {
          n_indefs = n_indefs + 1
          # lookup is indef, force to 1
          factor = 1
      }

   }
   else {

      n_notfound = n_notfound + 1
      # lookup failed, using correction of 1
   }   

   return factor
end

#---------------------------------------------
# binary search 
procedure bin_search(start_time,end_time,table_len,key,epsilon,index)

double start_time[ARB]   #I lookup table start time vector
double end_time[ARB]     #I end time vector
int    table_len         #I length of table
double key               #I value to look up
double epsilon           #I time epsilon value
int    index             #O index into table of found value


int left
int right
int x

bool lt_epsilon()
bool ge_epsilon()

begin

   # match is successful if (key >= start[x] && key < end[x] )
   
   index = -1

   left = 1
   right = table_len
   
   while (right >= left) {

      x = (left+right)/2

      #if (key < start_time[x]) 
      if (lt_epsilon(key,start_time[x],epsilon)) 
         right = x - 1
      else if ( ge_epsilon(key,end_time[x],epsilon))
      #else if (key >= end_time[x])
         left = x + 1
      else {
         index = x
         break
      }

   }

end
#---------------------------------------------
real procedure negfactor(x)
real x

begin

   call error(120,"Negative correction factor found, cannot compute error")

end

