# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_qpmkim.x
# 	Author	:  tnk
# 	Date	:  03/23/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_qpmkim.x -r1.20 03/23/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/13/97   mabbott	remvoed unneded qpoe$ includes
# 	01/06/97   tnk	added comments, slightly changed error messages
# 	01/06/97   tnk	bug fixes in new code
# 	01/03/97   tnk	preliminary rework to handle archive products
# 	10/21/96   tnk	now handles TIMEDEL or DELTIME headers
# 	10/03/96   tnk	reworked INSTRUME and DETNAM handling
# 	11/01/94   tnk	merged attribute & timefield params
# 	08/29/94   tnk	added time,dx,dy params
# 	07/19/94   tnk	reworked error band wcs handling
# 	07/15/94   tnk	missing arg in errband pass
# 	07/13/94   tnk	added fixmwcs parm
# 	07/12/94   tnk	fixed WCS for rect & block factors
# 	07/07/94   tnk	added error band handling
# 	05/02/94   tnk	adjusted epsilon compare routines
# 	03/31/94   tnk	misc. tweaks
# 	03/30/94   tnk	switched back to internal event_epsilon
# 	03/30/94   egos	reorganized options
# 	03/28/94   tnk	changed colget to getcol; added infomsgs; freed malloc'd vectors
# 	03/25/94   tnk	converted effexp to use df
# 	03/23/94   tnk	date and time created 94/03/23 17:08:06 by tnk
# H-
# 
#

#	@(#)t_qpmkim.x	1.20 10/13/97 UCB CEA EGO

include <imhdr.h>
include	<qpioset.h>
include <tbset.h>
include <imio.h>

#	Generate a "corrected" image from an EUVE QPOE file.
#
# Input:  QPOE file  (time,x,y,dx,dy)
#         correction (e.g., Mkprimbsch) ST table
#
# Output:  IRAF image 
#              imcopy
#              imcopy with adjusted exposure time
#              adjusted pixels
#      
#          IF in adjusted pixels mode, and user requests error band,
#          then output image has 2 planes, the 2nd of which has errors.
#    Limitations:
#    Adjust pixel mode:
#      Input file MUST be in sorted order by y, x
#
# 3 varieties of correction tables handled:
#
# 1)  Ordinary primbsch table (GO) products (mkprimbsch output).
#     Has data for all deadtime/primbsch combinations, with colnames
#     formed by the GO masks below
#
# 2)  Archive data products (massaged mkprimbsch output).
#     Deadtime/primbsch and deadtime corrections (no standalone primbsch).
#     Colnames formed from ARchive masks, below.
#     (Primbsch-only corrections are derived from the other two columns).
#
# 3) arbitrary correction column created by user


procedure	t_qpmkim()

define   N_REGIONS      4
# detector name in QPOE file
define   PRIMARY_DETNAME_HDR "DETNAM"
define   BACKUP_DETNAME_HDR  "INSTRUME"   
# delta time in correction table
define   PRIMARY_DELTIME_HDR "TIMEDEL"
define   BACKUP_DELTIME_HDR  "DELTIME"    

# correction column names (most have detector and quadrant number embedded)
# standard GO products
define TIME_COLNAME            "TableTime"        # time
define DEAD_COLMASK            "Det%1ddc"         # deadtime correction
define PRIMB_COLMASK           "Det%1dQ%1dpc"     # primbsch correction
define DEADPRIMB_COLMASK       "Det%1dQ%1ddpc"    # primb + dead correction

# archive products (embedded detname and quadrant number)
define AR_TIME_COLNAME         "Time"        # time
define AR_DEAD_COLMASK         "%sDC"        # deadtime correction
define AR_DEADPRIMB_COLMASK    "%sQ%1ddpc"   # primb + dead correction



# Default Epsilon for FP time comparisons.  If 2 values are within this,
# they are considered the same, i.e., round-off compensation.
# EUVE Event resolution is >= 2 ms, so this value should be fine.
define EVENT_EPSILON  0.0001D0 

# INPUT PARAMS
pointer	 input		 # QPIO filename param. ptr
pointer	 output          #
pointer  corr_tbl        # name of primbsch table
pointer  refdata         # reference directory
bool     applycorr       #
bool     adjexptime
bool     errband         # create error band
bool     do_dead         # do deadtime correct
bool     do_primb        # do primbsch correct
pointer  timecol         # name of time column
pointer  corrcol         # correction column name
double   time_epsilon    # epsilon value for time comparisons
double   deltime         # delta of table times
pointer  altdet		 # alternate detector name
bool     infomsgs        # informational messages
bool	 fixmwcs         # fix mwcs bug
pointer  time_field      # event struct time field
pointer	 dx_field        # event struct dx field
pointer  dy_field        # event struct dy field
int 	 user_src_quad	 # source quadrant number (INDEF or 0-3)

# LOCAL VARS
pointer	 qp, qpio	# QPOE pointers
pointer	 im             # Output image
pointer	 sp
pointer	 mw
pointer  mw2d
pointer  ref_tp     	 # reference table pointer
pointer  corr_tp   	 # primbsch table pointer
int 	 bound_type	 # boundary type
int 	 bound_x_y1      # boundaries
int 	 bound_y_y2      #
int 	 bound_y3        #
pointer  start_time      # primbsch time data array
pointer  end_time        # primbsch time data array
pointer  dead_vec        # deadtime factor
pointer  primb_vec[N_REGIONS]
pointer  deadprimb_vec[N_REGIONS]
pointer  corr_vec
pointer  corr_fact[N_REGIONS]
int      lookup_len      # length of primbsch lookup tab cols
int      det_num         # detector number
pointer  detector        #
pointer  text
int      stat
bool     adj_events
pointer  qpim
int      src_region
int      region
int      n_corr_vect
real     exptime
pointer  ex, xs, xe
int      xlen, nranges
pointer  field
bool     alt_corr    #alternate correction factor
bool     alt_time    #alternate correction factor
bool     mkerrband
bool     doerrband
int      indexlen        # length of QPOE index
real     shift[2]
int      blkfac
int      i
int	 n_rows
int      count
int	 quad

# FUNCTIONS
bool     clgetb()
pointer	 tbtopn()
int      tbpsta()
pointer  immap()
bool     rd_qbounds()
int      sscan()
real     imgetr()
real     effexp()
int      qpex_attrld()
int      stridx()
int      strlen()
double   tbhgtd()
double   clgetd()
int	 clgeti()
int      qpio_stati()
pointer  mw_openim()

begin

  call smark(sp)

  call eu_clstr(input,         "input")  # get qpoe filename 
  call eu_clstr(output,       "output")  # get output image filename
  call eu_clstr(corr_tbl,    "corrtab")  # correction table name
  call eu_clstr(refdata,     "refdata")  # det ref table
  applycorr = clgetb (     "applycorr")  # apply correction 
  adjexptime= clgetb (    "adjexptime")  #
  errband   = clgetb (       "errband")  # create error band
  do_dead   = clgetb (      "deadtime")  # correct for deadtime
  do_primb  = clgetb (      "primbsch")  # correct for primbsching
  call eu_clstr(time_field,"timefield")  # 
  call eu_clstr(timecol,     "timecol")  # alternate time column name
  call eu_clstr(corrcol,     "corrcol")  # alternate correction col name
  time_epsilon = clgetd (   "tepsilon")  # time epsilon value
  deltime   = clgetd (       "deltime")
  call eu_clstr(altdet,	    "detector")  # alternate detector
  user_src_quad  = clgeti  (     "quadrant")  # source quadrant number
  infomsgs  = clgetb (      "infomsgs")  # give informational messages
  fixmwcs   = clgetb (       "fixmwcs")  # use mwcs workaround code
  call eu_clstr(dx_field,    "dxfield")  # 
  call eu_clstr(dy_field,    "dyfield")  # 

  alt_corr = false
  if (strlen(Memc[corrcol]) > 0)
     alt_corr = true

  alt_time = false
  if (strlen(Memc[timecol]) > 0)
     alt_time = true

  adj_events = false
  if (applycorr && ! adjexptime)
     adj_events = true

  mkerrband = false
  if (adj_events && errband)
     mkerrband = true      # only make error band if asked & in pixel mode

  if (applycorr)
  {
     if (!do_dead && !do_primb && !alt_corr)
     {
        call eprintf("ERROR: You requested image correction (applycorr=yes)\n")
        call eprintf("but did not select any correction factor(s).\n")
        call error(122,"Must set deadtime and/or primbsch or corrcol parms")
     }
  }

  # open (filtered) QPOE file 
  call qpofilt(qp, qpio, Memc[input])

  # open QPOE as image also
  qpim = immap(Memc[input], READ_ONLY, 0) 

  # open output image 
  im = immap(Memc[output], NEW_FILE, 0)

  # setup image 
  if (mkerrband) {
     IM_NDIM(im) = 3
     IM_LEN(im,3) = 2
  }
  else {
     IM_NDIM(im) = IM_NDIM(qpim)
  }
  IM_LEN(im,2) = IM_LEN(qpim,2)	
  IM_LEN(im,1) = IM_LEN(qpim,1)
  if (adj_events)
     IM_PIXTYPE(im) = TY_REAL    
  else
     IM_PIXTYPE(im) = IM_PIXTYPE(qpim)

  # Set up the output WCS
  if (mkerrband)
  {
     # load 2d wcs info into a 3d wcs
     mw2d = mw_openim(qpim)  
     call mw2d3d(mw2d,mw)
     call mw_close(mw2d)
  }
  else {
     # just use 2d wcs
     mw = mw_openim(qpim)  
  }

  # This is a fix for a QPOE MWCS error in IRAF v2.10.2
  if (fixmwcs) 
  {
     blkfac = qpio_stati(qpio, QPIO_BLOCKFACTOR)
     shift[1] = 0.5 - 1/(2.0*blkfac)
     shift[2] = shift[1]
     call mw_shift(mw, shift, 3)
  }
  call mw_saveim (mw, im)


  # copy over header cards
  call hcpkill(qp, "Q", im, "I", "")

  # evaluate EXPTIME and put in output file
  exptime=INDEFR
  ifnoerr( exptime = imgetr(qpim, "EXPTIME")) {
     call imaddr(im, "EXPTIME", exptime)
  }

  # add history information
  call add_hist(im, Memc[input], applycorr, adjexptime, do_dead, do_primb,
     Memc[corrcol])

  if (!applycorr)
  {
     call im_cpy(qpim, im)
     call imastr(im, "BANDID1", "data - binned, uncorrected pixel values")
  }
  else 
  {
     # open correction table
     corr_tp = tbtopn(Memc[corr_tbl], READ_ONLY, 0)

     # get time column from correction table
     if (alt_time)
     {
        call getcold(corr_tp,Memc[timecol],end_time,lookup_len)
        if (lookup_len <= 0)
              call error(111,"Could not get correction table time column")
     }
     else  # look for default time columns
     {
        call getcold(corr_tp,TIME_COLNAME,end_time,lookup_len)
        if (lookup_len <= 0)
        {
           call getcold(corr_tp,AR_TIME_COLNAME,end_time,lookup_len)
           if (lookup_len <= 0)
              call error(111,"Could not get correction table time column")
        }
     }

     # if deltime parameter is undefined, get it from the correction table
     if (deltime == INDEFD)
     {
        # look for primary, then backup timdel/deltime header cards
        iferr (deltime = tbhgtd(corr_tp, PRIMARY_DELTIME_HDR))
        {
           iferr (deltime = tbhgtd(corr_tp, BACKUP_DELTIME_HDR))
           {
              call error(135,"Can't determine deltime/timedel from parameter or header cards in correction table")
           }
        }
     }

     # if time_epsilon is undefined, use the internal default
     if (time_epsilon == INDEFD)
        time_epsilon = EVENT_EPSILON

     # get either the "alternate" correction column or the usual
     # deadtime/primbsch columns
     corr_vec = NULL
     if (alt_corr)
     {
        call getcolr(corr_tp,Memc[corrcol],corr_vec,lookup_len)
        if (lookup_len <= 0)
           call error(111,"Problem getting correction column")
     }
     else  # get deadtime/primbsch columns
     {
        # get detector name.  look first in param, then in header cards
        call salloc(text, SZ_LINE, TY_CHAR)
        call salloc(detector, SZ_LINE, TY_CHAR)

	if (strlen(Memc[altdet]) > 0)
	   call strcpy(Memc[altdet], Memc[text], SZ_LINE)
        else 
        {
   	   iferr (call qp_gstr(qp, PRIMARY_DETNAME_HDR, Memc[text], SZ_LINE))
           {
              iferr (call qp_gstr(qp, BACKUP_DETNAME_HDR, Memc[text], SZ_LINE))
              {
		 call error(113,"No valid detector name given in parameter or header card")
              }
           }
	}

        # strip white space
        stat = sscan(Memc[text])
        call gargwrd(Memc[detector], SZ_LINE)

        # open reference detector table
        ref_tp = tbtopn (Memc[refdata], READ_ONLY, 0)

        # get detector number & source region from name
        call detinfo(ref_tp, Memc[detector],det_num,src_region)
        if (det_num < 0) {
           call eprintf("Invalid detector name: '%s'\n")
           call pargstr(Memc[detector])
           call error(100,"exiting...")
        }

	# determine which detector quadrant contains the source
	if (! IS_INDEFI(user_src_quad))
		src_region = user_src_quad

	if (IS_INDEFI(src_region))
        {
		call error(106,"Invalid detector source quadrant")
        }
        else
        {		
           src_region = src_region + 1  #convert region to index
           if (src_region > N_REGIONS)
              call error(105,"Invalid detector source quadrant")
        }

	# get deadtime, primbsch, & primb+dead columns
	# look for GO versions, then archive versions

	# get combined deadtime & primbsch correction
	for (i=1; i<= N_REGIONS; i=i+1)
           deadprimb_vec[i] = NULL

        call rd_quad_detnam_col(corr_tp,Memc[detector],src_region-1,
           AR_DEADPRIMB_COLMASK,deadprimb_vec[src_region])
	if (deadprimb_vec[src_region] == NULL)
        {
	   for (quad=1; quad <= N_REGIONS; quad=quad+1)
	   {
		 call read_quad_col(corr_tp,det_num,quad-1,DEADPRIMB_COLMASK,
                    deadprimb_vec[quad])
		 if (deadprimb_vec[quad] == NULL)
		    call error (222,"error fetching requested deadtime/primbsch correction data")
	   }
	}

	# get deadtime correction
	call rd_detnam_col(corr_tp,Memc[detector],AR_DEAD_COLMASK,
              dead_vec)
	if (dead_vec == NULL)
	{
	      call read_det_col(corr_tp,det_num,DEAD_COLMASK,dead_vec)
	      if (dead_vec == NULL)
		   call error (569,"error fetching requested deadtime correction data")
	}


	# get primbsch factor columns
        # if not there, compute vector from (deadtime & primbsch)
        # and deadtime factors

	for (i=1; i<= N_REGIONS; i=i+1)
           primb_vec[i] = NULL

	for (quad=1; quad <= N_REGIONS; quad=quad+1)
	{
              call read_quad_col(corr_tp,det_num,quad-1,PRIMB_COLMASK,
                 primb_vec[quad])
	      if (primb_vec[quad] == NULL)
              {
		  # compute primbsch correction from other factors
		  if (deadprimb_vec[quad] != NULL && dead_vec != NULL)
	          {
		        n_rows = tbpsta (corr_tp, TBL_NROWS)
			if (n_rows > 0)
                        {
			   call malloc(primb_vec[quad], n_rows, TY_REAL)
			   call idivr(Memr[deadprimb_vec[quad]],Memr[dead_vec],
                              Memr[primb_vec[quad]],n_rows)
		        }
		  }
	      } 
        }

	# make sure we have at least 1 primbsch correction column
	count = 0
	for (i=1; i<= N_REGIONS; i=i+1)
        {
           if (primb_vec[i] == NULL)
		count=count+1
        }
	if (count == N_REGIONS)
	   call error(568,"Could not fetch or reconstruct requested primbsch correction data")

        # get quadrant boundaries
        if ( ! rd_qbounds(ref_tp,det_num,bound_type,bound_x_y1,bound_y_y2,
           bound_y3))      
           call error(570,"Cannot find quadrant bounds")
     }


     # setup the start & end times for the lookup table
     call malloc (start_time, lookup_len, TY_DOUBLE)
     call setup_time(Memd[start_time],Memd[end_time],deltime,
        lookup_len,time_epsilon)

     for (region=1; region <= N_REGIONS; region=region+1)
        corr_fact[region] = NULL

     if (adjexptime) 
     {
        # verify time field (note that this only a test of attribute format)
        call salloc(field, SZ_LINE, TY_CHAR)
        call qp_expandtext(qp, Memc[time_field], Memc[field], SZ_LINE)
        if ( (stridx(Memc[field],"silrdSILRD") == 0) ||
          (stridx(Memc[field+1],"0123456789") == 0))
        {
           call error(101,"Invalid time field")
        }


        # assign the correction factor
        # note that only one detector region (the source region) is used
        if (corr_vec != NULL)
           corr_fact[1] = corr_vec
        else if (do_dead && !do_primb)
          corr_fact[1] = dead_vec
        else if (!do_dead && do_primb)
        {
          corr_fact[1] = primb_vec[src_region]
	  if (corr_fact[1] == NULL)
	     call error(588,"No primbsch correction factors available for the specified quadrant")
        }
        else # if (do_dead && do_primb)
        {
          corr_fact[1] = deadprimb_vec[src_region]
	  if (corr_fact[1] == NULL)
             call error(588,"No deadtime/primbsch correction factors available for the specified quadrant")

        }

        # fetch time intervals in QPOE time filter
        ex = qpio_stati(qpio, QPIO_EX)
        xs = NULL
        xe = NULL
        xlen = 0
        nranges = 0
        if (ex != NULL) {
	  nranges = qpex_attrld(ex, Memc[time_field], xs, xe, xlen)
        }

        # mv EXPTIME to RAWEXP, then update EXPTIME
        call imaddr(im, "RAWEXP", exptime)

        exptime = effexp(Memd[xs],Memd[xe],nranges,Memd[start_time],
          Memd[end_time],Memr[corr_fact[1]],lookup_len)
        call imaddr(im, "EXPTIME", exptime)

        # copy the data
        call im_cpy(qpim, im)

        call imastr(im, "BANDID1", "data - binned, uncorrected pixel values")

     }
     else  # make corrected image
     {
        # create a new image, where the events are adjusted
        # by the correction factors

        # verify that QPOE is in sorted order
        # This seems to return zero even for sorted files
        indexlen = qpio_stati(qpio,QPIO_INDEXLEN)

        # assign the primbsch/deadtime factors
        n_corr_vect = 1
        if (corr_vec != NULL)
           corr_fact[1] = corr_vec
        else if (do_dead && !do_primb)
           corr_fact[1] = dead_vec
        else if (!do_dead && do_primb)
        {
           n_corr_vect = N_REGIONS
           for (region=1; region <= N_REGIONS; region=region+1)
           {
              corr_fact[region] = primb_vec[region]
           }

        }
        else if (do_dead && do_primb)
        {
           n_corr_vect = N_REGIONS
           for (region=1; region <= N_REGIONS; region=region+1)
           {
              corr_fact[region] = deadprimb_vec[region]
           }

        }

        # do the data band and error band in separate passes
        # otherwise we take an enormous i/o hit

        doerrband = false
        call mk_corr_im(qp,qpio,mw,im,Memd[start_time],Memd[end_time],
             corr_fact,n_corr_vect,lookup_len,time_epsilon,bound_type,
             bound_x_y1,bound_y_y2,bound_y3, infomsgs, doerrband,
	     Memc[time_field], Memc[dx_field], Memc[dy_field])

        call imastr(im, "BANDID1", "data - binned, corrected pixel values")

        if (mkerrband) {

           if (infomsgs)
              call eprintf("making second pass for sigma band...\n")
      
           # rewind qpio pointer by closing & reopening
           call qpio_close(qpio)
           call qp_close(qp)

           # reopen (filtered) QPOE file 
           call qpofilt(qp, qpio, Memc[input])

           doerrband = true
           call mk_corr_im(qp,qpio,mw,im,Memd[start_time],Memd[end_time],
             corr_fact,n_corr_vect,lookup_len,time_epsilon,bound_type,
             bound_x_y1,bound_y_y2,bound_y3, infomsgs, doerrband,
	     Memc[time_field], Memc[dx_field], Memc[dy_field])

           call imastr(im, "BANDID2", 
                           "sigma - one stdev errors in pixel values")
        }

     }
  }

  # Clean up
  call imunmap(im)
  call imunmap(qpim)
  call qpio_close(qpio)
  call qp_close(qp)

  # free all the allocated vectors
  call mfree(start_time, TY_DOUBLE)
  call mfree(end_time, TY_DOUBLE)
  if (corr_vec != NULL)
     call mfree(corr_vec, TY_REAL)
  if (dead_vec != NULL)
     call mfree(dead_vec, TY_REAL)
  for (region=1; region <= N_REGIONS; region=region+1)
  {
     if (primb_vec[region] != NULL)
        call mfree(primb_vec[region], TY_REAL)
     if (deadprimb_vec[region] != NULL)
        call mfree(deadprimb_vec[region], TY_REAL)
  }

  call sfree(sp)
end
#----------------------------------------------------------
# read a table column, given detector number and colname mask
procedure read_det_col(tp,det_num,colname_format,vec_ptr)
pointer tp 	         #I table pointer
int det_num	         #I detector number
char colname_format[ARB] #I Det[d]xxx
pointer vec_ptr          #O ptr to data array, or NULL

pointer sp
pointer str
int len

begin

  # get a vector of data
  # colname_format has an sprintf mask that gives the column name

  call smark(sp)
  call salloc (str, SZ_LINE, TY_CHAR)

  # look for DET[d]Qxxx.
  call sprintf (Memc[str], SZ_LINE, colname_format)
  call pargi (det_num)

  call getcolr(tp,Memc[str],vec_ptr,len)


  call sfree(sp)

end

#----------------------------------------------------------
# read a table column, given detector name and colname mask
procedure rd_detnam_col(tp,det_name,colname_format,vec_ptr)
pointer tp 	         #I table pointer
char det_name[ARB]       #I detector name
char colname_format[ARB] #I column name format 
pointer vec_ptr          #O ptr to data array, or NULL

pointer sp
pointer str
int len

begin

  # get a vector of data
  # colname_format has an sprintf mask that gives the column name

  call smark(sp)
  call salloc (str, SZ_LINE, TY_CHAR)

  # look for column
  call sprintf (Memc[str], SZ_LINE, colname_format)
  call pargstr (det_name)

  call getcolr(tp,Memc[str],vec_ptr,len)

  call sfree(sp)

end

#----------------------------------------------------------
# read a column whose name is created with det_num and quad number
procedure read_quad_col(tp,det_num,quad,colname_format,vec_ptr)
pointer tp 	         #I table pointer
int det_num	         #I detector number
int quad		 #I quadrant number (zero based)
char colname_format[ARB] #I column name print format
pointer vec_ptr          #O ptr to data array

pointer sp
pointer str
int len

begin

  # get a vector of data for each quadrant
  # colname_format has an sprintf mask that gives the column name

  call smark(sp)
  call salloc (str, SZ_LINE, TY_CHAR)

  # create column name
  call sprintf (Memc[str], SZ_LINE, colname_format)
  call pargi (det_num)
  call pargi (quad)

  call getcolr(tp,Memc[str],vec_ptr,len)

  call sfree(sp)

end

#----------------------------------------------------------
# read a column whose name is created with det name and quad number
procedure rd_quad_detnam_col(tp,det_name,quad,colname_format,vec_ptr)
pointer tp 	         #I table pointer
char det_name[ARB]       #I detector name
int quad		 #I quadrant number (zero based)
char colname_format[ARB] #I format for column name
pointer vec_ptr          #O ptr to data array

pointer sp
pointer str
int len

begin

  # get a vector of data for each quadrant
  # colname_format has an sprintf mask that gives the column name

  call smark(sp)
  call salloc (str, SZ_LINE, TY_CHAR)

  # create column name
  call sprintf (Memc[str], SZ_LINE, colname_format)
  call pargstr (det_name)
  call pargi (quad)

  call getcolr(tp,Memc[str],vec_ptr,len)

  call sfree(sp)

end

#-----------------------------------------------------

# create start & end times for table lookup
#
procedure setup_time(start_time,end_time,bin_width,n_bins,epsilon)

double start_time[ARB]   #O  bin start times
double end_time[ARB]     #I  bin end times
double bin_width        #I width of primbsch intervals in seconds
int    n_bins            #I length of time arrays
double epsilon           #I epsilon for time comparisons

# LOCAL VARS
int     bin

# FUNCTIONS
bool eq_epsilon()


begin

  # make start & end times for each bucket.
  # bin end times are same as primbsch table times:

  # make start time a bin width earlier than end time
  # start_time = end_time - bin_width
  call asubkd(end_time,bin_width,start_time,nbins)


  # Take care of round off errors (very small bin gaps).
  # If end of bin A is closely adjacent to start of bin A + 1,
  # then make them match exactly.  Otherwise events will fall
  # between the bins.

  for (bin = 2; bin <= nbins; bin=bin+1) {
     if ( eq_epsilon(start_time[bin],end_time[bin-1],epsilon)) {
        start_time[bin] = end_time[bin-1]
     }
  }
   
end
#------------------------------------------
# add history information to the output image
procedure add_hist(im, infile, applycorr, adjexptime, do_dead, do_primb,
   corrcol)
pointer  im           #I/O image pointer
char     infile[ARB]  #I name of input file
bool     applycorr    #I
bool     adjexptime   #I
bool     do_dead      #I
bool     do_primb     #I
char     corrcol[ARB] #I name of (optional) override correction column

pointer sp
pointer text
pointer strlen

begin

  call smark(sp)
  call salloc(text, SZ_LINE, TY_CHAR)

  # add history information
  call sprintf(Memc[text],SZ_LINE,"Generated from '%s' by qpmkim.\n")
  call pargstr(infile)
  call strcat(Memc[text], IM_HISTORY(im), SZ_IMHIST)

  call strcpy("Correction applied: ",Memc[text],SZ_LINE)
  if (!applycorr)
  {
     call strcat("None\n",Memc[text],SZ_LINE)
     call strcat(Memc[text], IM_HISTORY(im), SZ_IMHIST)
  }
  else 
  {
     if (adjexptime)
        call strcat("adjusted exposure time\n",Memc[text],SZ_LINE)
     else
        call strcat("adjusted pixels\n",Memc[text],SZ_LINE)
     call strcat(Memc[text], IM_HISTORY(im), SZ_IMHIST)

     call strcpy("Correction factor(s): ",Memc[text],SZ_LINE)
     if (strlen(corrcol) > 0)
        call strcat(corrcol,Memc[text],SZ_LINE)
     else
     {
        if (do_dead) call strcat("deadtime ",Memc[text],SZ_LINE)
        if (do_primb) call strcat("primbsch",Memc[text],SZ_LINE)
     }
     call strcat("\n", Memc[text], SZ_LINE)
     call strcat(Memc[text], IM_HISTORY(im), SZ_IMHIST)

  }

  call sfree(sp)

end
