# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_specmain.x
# 	Author	:  tnk
# 	Date	:  04/28/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_specmain.x -r1.11 04/28/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	11/03/95   tnk	changed NR copyright to procedure call
# 	08/10/95   tnk	added Num. Rec. copyright
# 	11/28/94   tnk	switched to new image df i/o routine 
# 	09/14/94   tnk	df height/area fixes
# 	09/13/94   tnk	added output image data type
# 	09/12/94   tnk	added backrate to header cards
# 	09/12/94   tnk	fixed height/area problem with poisson noise
# 	09/01/94   tnk	added outsuffix
# 	07/05/94   tnk	added BANDID1 card
# 	05/02/94   tnk	fixed backrate computation
# 	04/28/94   tnk	date and time created 94/04/28 14:48:03 by tnk
# H-
# 
#

include <dfset.h>
include <tbset.h>
include <imhdr.h>
include <imio.h>

# @(#)t_specmain.x	1.11 11/3/95 UCB CEA EGO

# Instrument modeling software
# Input: model flux at telescope  (linelist or continuum)
#        reference (effective area)
# Output:
#        predicted counts
# Method:
#   Basically just multiplies flux by effective area to produce counts
#
# Makes extensive use of df (Discrete Function) library code
# df types:
#   PD = Pixel Data
#   PF = piecewise function  (sampled)
#
# Assumes model, effective area, and psf wavelengths
# all run from low to high.
#
# Makes use of Numerical Recipes library routines,
# Copyright (C) 1986 Numerical Recipes Software


define MAX_ORDERS 20    # max no. of effective area orders
define BANDID1    "spectrum - generated from model"

procedure t_specmain()

# PARAMETERS
pointer model           # name of model file
int     wavecol         # wavelength column 
int     fluxcol         # flux column 
pointer outroot         # root name of output file
pointer refdata         # name of reference data detector file
bool    linelist        # YES=linelist, NO=continuum
double  line_fwhm       # FWHM of lines in angstroms
bool    sampled         # input type:  YES=samples (PLF)  NO=histogram (PD)
bool    convolve        # boolean: apply convolution
bool    dopsf           # boolean:  yes, kernal=psf; no, kernal=collresp
double  n_sigma         # how many sigma out to go for psf gaussians
bool    addnoise        # boolean:  add poisson noise
int     seed            # seed for random number generator
# pset parameters
pointer detector        # name of detector
double  exptime         # exposure time
double  binsize         # output bin size
double  minwave         # first bin wavelength
double  maxwave         # last bin wavelength
int     nbins           # number of output bins
double  conv_step       # step to use for applying convolution (angstroms)
double  backrate        # background rate (counts/sec/angstrom)
int     order_list      # list of orders
pointer order_str       # list of orders as a string
pointer outsuffix       # optional suffix appended to output filename

# Other variables
pointer sp
pointer output
pointer out_file            # output file pointer
pointer model_df            # model Discrete Function
pointer model_in            # temporary DF used to read in the model
pointer ea_df[MAX_ORDERS]   # effective area DF`s
pointer psf_fwhm_df         # PSF df
pointer out_file_df
pointer out_bins
pointer out_bin_pf
pointer temp_df
pointer counts_pd
pointer counts_df
int     n_ea_orders
int     i
pointer et                  # EUVE table structure pointer
pointer dd                  # ref data (edd) structure pointer
double  x0
double  xmin, xmax, ymin, ymax
double  x_step
bool    out_table
pointer colptr
real    r[1]                # MWCS ref pixel
real    w[1]                # MWCS world coord at ref pixel
real    cd[1,1]             # coord xform
int     npts[1]
int     ndim       
double  half_bin 
double  ea_min, ea_max      # effective area domain extremes
pointer collresp_df
int     mod_dftype
char    ftype
int     out_dtype

# FUNCTIONS
pointer df_open()           # create DF in memory
double  clgetd()            #
int     clpopnu()           # unsorted list
bool    clgetb()            #
int     clgeti()
bool    verf_bins()
bool    get_file_df()
pointer tbl_df()
pointer tbtopn()
pointer eu_newimg()
int     df_len()
bool    is_incrdom()
int     strlen()
extern  tbhadt(), tbhadb(), tbhadd(), tbhadi()
extern  imastr(), imaddb(), imaddd(), imaddi()
int     hdr_fn_t, hdr_fn_b, hdr_fn_d, hdr_fn_i  # header function addr

begin

   # put in Numerical Recipes copyright string
   call nr_copyr()

   call smark(sp)

   # input parameters
   call eu_clstr(model,        "model")
   call eu_clstr(wavecol,    "wavecol")
   call eu_clstr(fluxcol,    "fluxcol")
   call eu_clstr(outroot,     "output")
   out_table =  clgetb(     "outtable")
   call eu_clstr(refdata,    "refdata")
   linelist  =  clgetb(     "linelist")
   line_fwhm =  clgetd(     "linefwhm")
   sampled   =  clgetb(      "sampled")
   convolve  =  clgetb(     "convolve")
   dopsf     =  clgetb(          "psf")
   n_sigma   =  clgetd(       "nsigma")
   addnoise  =  clgetb(     "addnoise")
   seed      =  clgeti(         "seed")
   # pset parameters
   call eu_clstr(detector,  "detector")
   exptime   =  clgetd(      "exptime")  
   binsize   =  clgetd(      "binsize")
   minwave   =  clgetd(      "minwave")
   maxwave   =  clgetd(      "maxwave")
   nbins     =  clgeti(        "nbins")
   conv_step =  clgetd(      "convstep")
   backrate  =  clgetd(     "backrate")
   order_list= clpopnu(       "orders")
   call eu_clstr(order_str,   "orders")  #easy way to get order string
   call eu_clstr(outsuffix,"outsuffix") 

   # verify output bin parameters
   if ( ! verf_bins(minwave,maxwave,binsize,nbins) )
      call error(101,"Invalid output bin specifications")

   # open reference table
   call et_open(et,Memc[refdata])
   if (et == NULL) {
      call eprintf("ERROR: Could not open reference table '%s'\n")
      call pargstr(Memc[refdata])
      call error(122,"Error opening refdata")
   }

   # pull out struct for desired detector
   call edd_new(dd, et, Memc[detector])
   if (dd == NULL) {
      call eprintf("ERROR: Could not find information for detector '%s'\n")
      call pargstr(Memc[detector])
      call error(124,"Invalid detector")
   }

   # determine model data type:  Linelist, samples (PF), or histogram (PD)
   # note that linelists get read in as PFs 
   mod_dftype = DF_PF
   if (!linelist && ! sampled)
      mod_dftype = DF_PD

   # get model (wavelength,flux) as doubles
   model_in  = df_open(mod_dftype,TY_DOUBLE,TY_DOUBLE)
   model_df  = df_open(mod_dftype,TY_DOUBLE,TY_DOUBLE)
   call df_pedge(model_df, 0.0d0, 0.0d0)

   if (! get_file_df(model_in,Memc[model],Memc[wavecol],Memc[fluxcol],ftype))
      call error(102,"Could not read model")
   call df_pedge(model_in, 0.0d0, 0.0d0)

   if (! is_incrdom(model_in))
      call error(104,"Input model wavelengths are out of order")

   # don't allow linelists in images
   if (linelist && (ftype == 'I'))
      call error(106,"Cannot have linelist stored in an image")

   # if input is a linelist, convert delta functions into triangles
   if (linelist) 
      call llpf2pf(model_in,model_df,line_fwhm) 
   else
      call df_mov(model_in,model_df)
   call df_close(model_in)

   # create output file name,  "outbase".<outsuffix> or outbase."detector"
   call salloc(output,SZ_LINE,TY_CHAR)
   call strcpy(Memc[outroot],Memc[output],SZ_LINE)
   call strcat(".",Memc[output],SZ_LINE)
   if (strlen(Memc[outsuffix]) > 0)
      call strcat(Memc[outsuffix],Memc[output],SZ_LINE)
   else
      call strcat(Memc[detector],Memc[output],SZ_LINE)

   if (out_table) {

      # open output table
      out_file = tbtopn(Memc[output],NEW_FILE,0)   
      call tbcdef(out_file, colptr, "wavelength", "", "",TY_DOUBLE,0,1)
      call tbcdef(out_file, colptr, "counts",     "", "",TY_DOUBLE,0,1)
      call tbtcre(out_file)
      out_file_df = tbl_df(DF_PD, out_file, "wavelength", "counts")

      # define header card output routines
      call zlocpr(tbhadt, hdr_fn_t)
      call zlocpr(tbhadb, hdr_fn_b)
      call zlocpr(tbhadd, hdr_fn_d)
      call zlocpr(tbhadi, hdr_fn_i)
   }
   else {

      # make an image, with appropriate WCS values
      r[1] = 1.0
      w[1] = real(minwave)
      cd[1,1] = real(binsize)
      npts[1] = nbins
      ndim = 1
      out_dtype = TY_REAL
      out_file = eu_newimg(Memc[output],r,w,cd,npts,ndim,out_dtype)

      # attach a DF 
      out_file_df = df_open(DF_PD, TY_DOUBLE, TY_DOUBLE)

      # define header card output routines
      call zlocpr(imastr, hdr_fn_t)
      call zlocpr(imaddb, hdr_fn_b)
      call zlocpr(imaddd, hdr_fn_d)
      call zlocpr(imaddi, hdr_fn_i)
   }

   # save parameters as output header cards          
   call zcall3(hdr_fn_t, out_file, "model",    Memc[model])    
   call zcall3(hdr_fn_t, out_file, "wavecol",  Memc[wavecol])  
   call zcall3(hdr_fn_t, out_file, "fluxcol",  Memc[fluxcol])  
   call zcall3(hdr_fn_t, out_file, "output",   Memc[output])   
   call zcall3(hdr_fn_b, out_file, "outtable", out_table)
   call zcall3(hdr_fn_t, out_file, "refdata",  Memc[refdata])  
   call zcall3(hdr_fn_b, out_file, "linelist", linelist)       
   call zcall3(hdr_fn_d, out_file, "linefwhm", line_fwhm)      
   call zcall3(hdr_fn_b, out_file, "sampled",  sampled)       
   call zcall3(hdr_fn_b, out_file, "convolve", convolve)          
   call zcall3(hdr_fn_b, out_file, "psf",      dopsf)          
   call zcall3(hdr_fn_d, out_file, "nsigma",   n_sigma)
   call zcall3(hdr_fn_b, out_file, "addnoise", addnoise)       
   call zcall3(hdr_fn_i, out_file, "seed",     seed)           
   call zcall3(hdr_fn_t, out_file, "detector", Memc[detector]) 
   call zcall3(hdr_fn_d, out_file, "exptime",  exptime)        
   call zcall3(hdr_fn_d, out_file, "binsize",  binsize)        
   call zcall3(hdr_fn_d, out_file, "minwave",  minwave)        
   call zcall3(hdr_fn_d, out_file, "maxwave",  maxwave)        
   call zcall3(hdr_fn_i, out_file, "nbins",    nbins)          
   call zcall3(hdr_fn_d, out_file, "convstep", conv_step)       
   call zcall3(hdr_fn_d, out_file, "backrate", backrate)       
   call zcall3(hdr_fn_t, out_file, "orders",   Memc[order_str])
   call zcall3(hdr_fn_t, out_file, "BANDID1",  BANDID1)


   # get effective areas as PF DF`s from EA table
   call get_ea_df(dd, ea_df, MAX_ORDERS, n_ea_orders)
   if (n_ea_orders <= 0)
      call error(121,"Could not find effective areas for detector")

   # get global ea min/max domain values
   call df_gminmax(ea_df, n_ea_orders, ea_min, ea_max)

   # make sure the EA files & model have some domain in common
   call df_minmax(model_df,xmin,xmax,ymin,ymax)
   if (xmax < ea_min || xmin > ea_max) {
      call eprintf("Error: The wavelength values of the effective area\n")
      call eprintf("files and the input model have no areas in common.\n")
      call error(115,"exiting...")
   }
   # for efficiency, trim the model by the ea extremes
   call df_trim(model_df,ea_min,ea_max,model_df)

   # get PSF fwhm list
   psf_fwhm_df = df_open(DF_PF,TY_DOUBLE,TY_DOUBLE)
   call get_psf_df(dd, psf_fwhm_df)

   # compute counts:
   # counts = effective_area(order) * flux * exposure time
   counts_df = df_open(mod_dftype,TY_DOUBLE,TY_DOUBLE)
   call df_pedge(counts_df,0.0d0,0.0d0)

   call comp_counts(order_list,model_df,ea_df,n_ea_orders,counts_df)
   call df_mulk(counts_df,exptime,counts_df)

   # trim the xdomain to match output bins
   call df_trim(counts_df,minwave-binsize,maxwave+binsize,counts_df)
 
   # create output bins
   out_bins = df_open(DF_PD,TY_DOUBLE,TY_DOUBLE)
   call df_pedge(out_bins,0.0d0,0.0d0)

   # convolve with psf 
   if (convolve)
   {
      temp_df = df_open(mod_dftype,TY_DOUBLE,TY_DOUBLE)
      call df_pedge(temp_df,0.0d0,0.0d0)

      # remap domain by adding in a finely spaced pf
      call df_minmax(counts_df,xmin,xmax,ymin,ymax)
      x_step = conv_step
      call gen_const_df(temp_df,xmin,xmax,x_step,0.0d0)
      call df_add(counts_df,temp_df,counts_df)
      call df_free(temp_df)

      counts_pd = df_open(DF_PD,TY_DOUBLE,TY_DOUBLE)
      call df_pedge(counts_pd,0.0d0,0.0d0)

      # convert pf to pd, if needed
      if (mod_dftype == DF_PF)
         call df_pfpd(counts_df,counts_pd)
      else
	 call df_mov(counts_df,counts_pd)

      if (do_psf) {          # convolve with PSF
         call psf_convolve(counts_pd,x_step,psf_fwhm_df,n_sigma)
      }
      else  #convolve with collimator response
      {
         # get collimator response
         collresp_df = df_open(DF_PF,TY_DOUBLE,TY_DOUBLE)
         call get_cr_df(Memc[detector],Memc[refdata],collresp_df)

         if (df_len(collresp_df) > 0)
            call df_convolve(counts_pd,x_step,collresp_df)
      }       
      call df_mov(counts_pd,out_bins)

   }
   else { # no convolve

      # make a DF with the same domain as the desired output bins,
      # only shifted by 1/2 bin
      out_bin_pf = df_open(mod_dftype,TY_DOUBLE,TY_DOUBLE)
      call df_pedge(out_bin_pf,0.0d0,0.0d0)
      half_bin = binsize/2.0d0                
      call gen_const_df(out_bin_pf,minwave+half_bin,maxwave+half_bin,
         binsize,0.0d0)

      # add output_bin_pf so that we split the PF into segments that
      # will fit within the desired bins
      call df_add(counts_df,out_bin_pf,counts_df)

      # convert pf to pd, if needed
      if (mod_dftype == DF_PF)
         call df_pfpd(counts_df,out_bins)
      else
         call df_mov(counts_df,out_bins)
   }

   # change the domain to match the desired output bins
   # (i.e., consolidate the pd segments)
   x0=minwave - binsize/2.0d0
   call df_rsdomain(out_bins,x0,binsize,nbins,out_bins)


   # add in the background rate
   # (note that we do not mulitply by binsize, because PD is stored
   # by height)
   backrate = backrate * exptime
   call df_addk(out_bins,backrate,out_bins)

   # convert from heights to areas
   call df_mbw(out_bins,out_bins)

   # add poisson noise?
   if (addnoise) {
      call add_poisson(out_bins,long(seed))
   }

   if (out_table) {
      # shift bins back by 1/2 bin size so that output values
      # reflect the center of the bin, and not the right edge

      call df_shift(out_bins, -binsize/2.0d0, out_bins)
   }

   # write it out 
   # (note that for images, bins automatically get shifted back 
   # by 1/2 bin size so that output values
   # reflect the center of the bin, and not the right edge)
   call df_mov(out_bins, out_file_df)


   # wrap it up
   call df_close(counts_df)
   call df_close(out_bins)
   call df_close(psf_fwhm_df)
   call df_close(collresp_df)
   call df_close(model_df)

   # free effective area df's
   for (i=1; i<= n_ea_orders; i=i+1)
         call df_close(ea_df[i])

   call edd_close(dd)
   call et_close(et)

   if (out_table) {
      call df_close(out_file_df)
      call tbtclo(out_file)
   }
   else {
      call pd_wrim(out_file, 1, 1, out_file_df)
      call df_close(out_file_df)
      call imunmap(out_file)
   }


   call sfree(sp)

end

