# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  verf_bins.x
# 	Author	:  tnk
# 	Date	:  04/28/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) verf_bins.x -r1.3 04/28/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	07/05/94   tnk	fixed calculations
# 	04/28/94   tnk	extra indef check
# 	04/28/94   tnk	date and time created 94/04/28 14:48:04 by tnk
# H-
# 
#


# @(#)verf_bins.x	1.3 7/5/94 UCB CEA EGO

# verify & compute bin specifications, i.e.,
# user should specify 3 out of 4 of minbin, maxbin, binsize, nbins.
# 4th one is calculated.  
# If all are specified, then nbins is ignored.

bool procedure verf_bins(minbin,maxbin,binsize,nbins)

double minbin   #I/O minimum bin value
double maxbin   #I/O maximum bin value
double binsize  #I/O width of bin
int    nbins    #I/O number of bins

int   n_undef
bool  status

begin

# see how many are undefined.

   status = true
   n_undef = 0
   if (minbin == INDEFD)  n_undef = n_undef + 1
   if (maxbin == INDEFD)  n_undef = n_undef + 1
   if (binsize == INDEFD) n_undef = n_undef + 1
   if (nbins == INDEFI)   n_undef = n_undef + 1

   if (n_undef > 1) {
      call eprintf("Must specify 3 out of 4 non-INDEF ")
      call eprintf("values for bin definitions\n")
      call eprintf("binsize, minwave, maxwave, nbins\n")
      status = false
   }
   else if (n_undef == 0) {
      call eprintf("One of the bin definitions must be ")
      call eprintf("set to INDEF.\n")
      call eprintf("(binsize, minwave, maxwave, nbins)\n")
      status = false
   }
   else {

      if ((minbin != INDEFD && minbin <=0.0d0) || 
          (maxbin != INDEFD && maxbin <=0.0d0) ||
          (binsize != INDEFD && binsize <= 0.0d0) ||
          (nbins != INDEFI && nbins <=0)) {
              call eprintf("All bin specifications must be > zero\n")
              call eprintf("(one of them can be INDEF)\n")
              status = false        
      } else {

         # Note the extra factor of 1 because we're using bin centers

         if (minbin == INDEFD)
            minbin = maxbin - (binsize * (nbins-1))

         if (maxbin == INDEFD)
            maxbin = minbin + (binsize * (nbins-1))

         if (binsize == INDEFD)
            binsize = (maxbin - minbin) / (nbins-1)

         if (nbins == INDEFI)
            nbins = (maxbin - minbin) / binsize + 1

         # check for maxbin > minbin 
         if (minbin >= maxbin) {
            call eprintf("Minbin must be < maxbin\n")
            status = false
         }
      }
   }

   return(status)

end
