# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_aspcorr.x
# 	Author	:  akeith
# 	Date	:  06/21/93
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_aspcorr.x -r1.17 06/21/93 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	01/08/97   tnk	tweaked comments
# 	01/08/97   tnk	renamed "position" to "ephemeris";  added INDEF time error reports
# 	01/08/97   tnk	major rework;  now handles INDEFs
# 	12/27/96   tnk	removed ra_i conversion to degrees (for RA_FIX card)
# 	12/26/96   tnk	reworked ra/dec handling
# 	10/24/96   tnk	incr MAX_USR_PARS
# 	10/24/96   tnk	fixed alignment problem causing 1st output row repeat
# 	10/22/96   tnk	moved EOF checks inside the table read loops
# 	09/13/95   tnk	set max table user params
# 	05/03/95   akeith	the function acwalk was being called with the wrong arg list
# 	08/24/94   akeith	cleaned lint out
# 	08/24/94   akeith	added header cards
# 	06/03/94   tnk	added hcpkill (header copy kill)
# 	04/11/94   akeith	fixed equal times error
# 	07/29/93   akeith	fixed one of the parameter names
# 	07/16/93   akeith	now utilizes the group library
# 	06/21/93   akeith	date and time created 93/06/21 16:51:35 by akeith
# H-
# 
#

#       Version: @(#)t_aspcorr.x	1.17 1/8/97 UCB CEA EGO

include	"atb.h"
include <tbset.h>
include <math.h>

define MAX_USR_PARS   120       # max. number of user parameters in table

#
# t_aspcorr()
#
# Perform aspect correction given a source ephemeris.
#
# Process entire aspect table, putting out
# a row for each input row.
#
# If any of the input aspect values (other than time)
# are INDEF, all the output aspect values for that
# row are forced to INDEF.
#
# If there's a valid interpolated target ephemeris position
# for the time of the aspect, then the aspect will
# be adjusted by it.
# If no valid ephemeris value is available (i.e., the
# bracketing earlier and later ephemeris position values are INDEF),
# the output aspect row will be INDEF.
#
# Assumes both aspect & ephemeris tables are time ordered.
# Rows with INDEF time values in either table are skipped.


procedure	t_aspcorr()

pointer	sp				# stack pointer
pointer	filename			# ephemeris filename
pointer	aspq				# aspect quaternion + time
pointer	nephemq				# new ephemeris quaternion + time
pointer	oephemq				# old ephemeris quaternion + time
pointer	qint				# interpolated quaternion
pointer	qfix				# fixed object quaternion
pointer	qtmp1				# temp quaternion + time
pointer	qtmp2				# temp quaternion + time

pointer pos				# table info structure
pointer asp				# table info structure
pointer out				# table info structure

double	ra, dec				#
bool    ephem_EOF				# eof indicator for ephemeris table

int	ac_gnephemeris()
int	ac_gnaspect()
bool	qisindef()
bool    fp_equald()

begin

   call smark(sp)
   call salloc(filename, SZ_LINE, TY_CHAR)
   call salloc(aspq, AC_NELMS*SZ_DOUBLE, TY_DOUBLE)
   call salloc(nephemq, AC_NELMS*SZ_DOUBLE, TY_DOUBLE)
   call salloc(oephemq, AC_NELMS*SZ_DOUBLE, TY_DOUBLE)
   call salloc(qfix, AC_NELMS*SZ_DOUBLE, TY_DOUBLE)
   call salloc(qint, AC_NELMS*SZ_DOUBLE, TY_DOUBLE)
   call salloc(qtmp1, AC_NELMS*SZ_DOUBLE, TY_DOUBLE)
   call salloc(qtmp2, AC_NELMS*SZ_DOUBLE, TY_DOUBLE)

   call ac_defpos(pos, Memc[filename], SZ_LINE)
   call ac_defasp(asp)
   call ac_deffix(Memd[qfix], ra, dec)
   call ac_defout(out, asp, Memc[filename], SZ_LINE, ra, dec)


   # load the old & new ephemeris quaternions
   if(ac_gnephemeris(pos, Memd[oephemq]) == EOF)
	   call ac_error(1, "Not enough rows in ephemeris table.", "")

   if(ac_gnephemeris(pos, Memd[nephemq]) == EOF)
	   call ac_error(1, "Not enough rows in ephemeris table.", "")


   # go through aspect table, put out a row for each input row
   ephem_EOF = false
   while (ac_gnaspect(asp, Memd[aspq]) != EOF)
   {
       # if the aspect is INDEF, or we've run out of ephemeris info,
       # put out an INDEF aspect row      
       if (qisindef(Memd[aspq]) || ephem_EOF)
       {
	   call setqindef(Memd[aspq])
	   call ac_output(out, Memd[aspq])
       }
       else
       {
	   # see if aspect time is before old ephemeris time
	   if (QTIME(aspq) < QTIME(oephemq))
	   {
	      call setqindef(Memd[aspq])
	      call ac_output(out, Memd[aspq])
	   }
	   else
	   {

	      # skip through ephemeris until we bracket aspect time 
	      while (QTIME(aspq) > QTIME(nephemq) &&
                  ! fp_equald(QTIME(aspq),QTIME(nephemq)))
	      {
		      # copy ephemeris to old ephemeris
		      call amovd(Memd[nephemq], Memd[oephemq], AC_NELMS)

		      if(ac_gnephemeris(pos, Memd[nephemq]) == EOF)
		      {
			 ephem_EOF = true
			 break
		      }
	      }


	      # if no valid ephemeris write INDEF aspect 
	      if (qisindef(Memd[nephemq]) || qisindef(Memd[oephemq]) || ephem_EOF)
	      {
		   call setqindef(Memd[aspq])
		   call ac_output(out, Memd[aspq])
	      }
	      else
	      {
		   # adjust aspect by interpolated ephemeris & write out 
		   QTIME(qint) = QTIME(aspq)
		   call ac_interp(nephemq, oephemq, QTIME(aspq), qtmp1)
		   call er_qmkrotate(Memd[qtmp2], Memd[qfix],  Memd[qtmp1])
		   call er_qmult(Memd[qtmp2], Memd[aspq], Memd[qint])
		   call ac_output(out, Memd[qint])

	      }

	   }


       }

   }

   call ac_close(pos, asp, out)
   call sfree(sp)

end


#
# qisindef() returns true if any of the w,x,y, or z components
# of a quaternion are indef

bool	procedure	qisindef(q)

double	q[AC_NELMS]			#I

begin

        if ( q[AC_ASPW] == INDEFD || q[AC_ASPX] == INDEFD ||
	     q[AC_ASPY] == INDEFD || q[AC_ASPZ] == INDEFD)
	{
	   return(true)
	}
	else
	{
	   return(false)
	}

end


#
# setqindef() sets the w,x,y, and z components  of a quaternion to indef

procedure	setqindef(q)

double	q[AC_NELMS]			#I

begin

        q[AC_ASPW] = INDEFD
        q[AC_ASPX] = INDEFD 
	q[AC_ASPY] = INDEFD
        q[AC_ASPZ] = INDEFD

end




#
# ac_interp() calculates the scalar and interpolates between the new and old
#	quaternion.
#

procedure	ac_interp(qnew, qold, time, qint)

pointer	qnew				#I
pointer	qold				#I
pointer	qint				#O
double	time				#I

double	scalar
bool    fp_equald()

begin

	if(fp_equald(QTIME(qnew), QTIME(qold)))
        {
	   call amovd(Memd[qold], Memd[qint], AC_NELMS)
	}
	else
	{
	   scalar = (time - QTIME(qold))/(QTIME(qnew) - QTIME(qold))
	   call er_qinterpolate(Memd[qold], Memd[qnew], scalar, Memd[qint])
	} 


end


#
# ac_gnephemeris() gets the next ephemeris value from the ephemeris table.
#	It is assumed that this value will be in terms of Ra and Dec, so
#	these values are converted to a quaternion.
#

int	procedure	ac_gnephemeris(pos, q)

pointer	pos				#I/O ephemeris structure
double	q[AC_NELMS]			#O quaternion

double	ra				# double value
double	dec				# double value
double	time				# double value
int     nrepeat

begin

   nrepeat = 0
   repeat {

	if(ATB_ROW(pos) > ATB_NROWS(pos))
		return(EOF)

	call tbegtd(ATB_TB(pos), ATB_RA(pos), ATB_ROW(pos), ra)
	call tbegtd(ATB_TB(pos), ATB_DEC(pos), ATB_ROW(pos), dec)
	call tbegtd(ATB_TB(pos), ATB_TIME(pos), ATB_ROW(pos), time)

	ATB_ROW(pos) = ATB_ROW(pos) + 1
	nrepeat = nrepeat + 1
   }
   until (time != INDEFD)

   if (nrepeat > 1)
   {
	call eprintf("Warning: INDEF time value(s) found in ephemeris table, ")
	call eprintf("row: %d\n");
	call pargi(ATB_ROW(pos)-2)
   }

   if ( ra != INDEFD && dec != INDEFD)
   {
	ra = ATB_RACONV(pos)*ra
	dec = ATB_DECCONV(pos)*dec
	call er_qfrompos(ra, dec, q)
	q[AC_TIME] = time
   }
   else
   {
	call setqindef(q)
	q[AC_TIME] = time
   }

   return(1)

end



#
# ac_gnaspect() gets the next aspect value from the aspect table. 
#

int	procedure	ac_gnaspect(asp, q)

pointer	asp				#I/O aspect structure
double	q[AC_NELMS]			#O

int     nrepeat

begin

   nrepeat = 0
   repeat {

	if(ATB_ROW(asp) > ATB_NROWS(asp))
		return(EOF)

	call tbegtd(ATB_TB(asp), ATB_ASPECTW(asp), ATB_ROW(asp), q[AC_ASPW])
	call tbegtd(ATB_TB(asp), ATB_ASPECTX(asp), ATB_ROW(asp), q[AC_ASPX])
	call tbegtd(ATB_TB(asp), ATB_ASPECTY(asp), ATB_ROW(asp), q[AC_ASPY])
	call tbegtd(ATB_TB(asp), ATB_ASPECTZ(asp), ATB_ROW(asp), q[AC_ASPZ])
	call tbegtd(ATB_TB(asp), ATB_TIME(asp), ATB_ROW(asp), q[AC_TIME])

	ATB_ROW(asp) = ATB_ROW(asp) + 1
	nrepeat = nrepeat + 1

   }
   until (q[AC_TIME] != INDEFD)

   if (nrepeat > 1)
   {
	call eprintf("Warning: INDEF time value(s) found in aspect table, ")
	call eprintf("row: %d\n");
	call pargi(ATB_ROW(asp)-2)
   }

   if ( q[AC_ASPW] == INDEFD || q[AC_ASPX] == INDEFD ||
	q[AC_ASPY] == INDEFD || q[AC_ASPZ] == INDEFD)
   {
   	call setqindef(q)
   }

   return(1)

end


#
# ac_output() writes out the quaternion to the output table.
#

procedure	ac_output(out, q)

pointer	out				#I/O output structure
double	q[AC_NELMS]			#I

begin

	call tbeptd(ATB_TB(out), ATB_ASPECTW(out), ATB_ROW(out), q[AC_ASPW])
	call tbeptd(ATB_TB(out), ATB_ASPECTX(out), ATB_ROW(out), q[AC_ASPX])
	call tbeptd(ATB_TB(out), ATB_ASPECTY(out), ATB_ROW(out), q[AC_ASPY])
	call tbeptd(ATB_TB(out), ATB_ASPECTZ(out), ATB_ROW(out), q[AC_ASPZ])
	call tbeptd(ATB_TB(out), ATB_TIME(out), ATB_ROW(out), q[AC_TIME])

	ATB_ROW(out) = ATB_ROW(out) + 1

end


#
# ac_defpos() gets the ephemeris table information and opens the table.
#

procedure	ac_defpos(pos, tb_name, len)

pointer	pos				#I/O ephemeris structure
char	tb_name[len]			#O   ephemeris table name
int	len				#I   length of table name

pointer sp
pointer units
pointer col_name

int	tbpsta()
pointer	tbtopn()

begin

	call atb_init(pos)

	call smark(sp)
	call salloc(units, SZ_LINE, TY_CHAR)

	call clgstr("ephemeris", tb_name, len)

	iferr(ATB_TB(pos) = tbtopn(tb_name, READ_ONLY, NULL))
		call ac_error(1, "No such table, ", tb_name)

	call eu_clstr(col_name, "racol")
	call tbcfnd(ATB_TB(pos), Memc[col_name], ATB_RA(pos), 1)
	if(ATB_RA(pos) == NULL)
		call ac_error(1, "No such column, ", Memc[col_name])

	call eu_clstr(col_name, "deccol")
	call tbcfnd(ATB_TB(pos), Memc[col_name], ATB_DEC(pos), 1)
	if(ATB_DEC(pos) == NULL)
		call ac_error(1, "No such column, ", Memc[col_name])

	call eu_clstr(col_name, "time")
	call tbcfnd(ATB_TB(pos), Memc[col_name], ATB_TIME(pos), 1)
	if(ATB_TIME(pos) == NULL)
		call ac_error(1, "No such column, ", Memc[col_name])

	ATB_NROWS(pos) = tbpsta(ATB_TB(pos), TBL_NROWS)

	call tbcigt(ATB_RA(pos), TBL_COL_UNITS,  Memc[units],  SZ_LINE)
	if(Memc[units] == 'h' || Memc[units] == 'H')
		ATB_RACONV(pos) = TWOPI / 24.0
	else if(Memc[units] == 'd' || Memc[units] == 'D')
		ATB_RACONV(pos) = TWOPI / 360.0
	else
	   call ac_error(1,
            "Invalid ephemeris table RA units, must be hours (H/h) or Degrees (D/d): ",
	    Memc[units])

	call tbcigt(ATB_DEC(pos), TBL_COL_UNITS,  Memc[units],  SZ_LINE)
	if(Memc[units] == 'd' || Memc[units] == 'D')
		ATB_DECCONV(pos) = TWOPI / 360.0
	else
	   call ac_error(1,"Invalid ephemeris table Dec units, must be Degrees (D/d): ",
            Memc[units])

	call sfree(sp)

end


#
# ac_defasp() gets the aspect table information and opens the table.
#

procedure	ac_defasp(asp)

pointer	asp				#I/O aspect structure

pointer sp
pointer grp
pointer tb_name
pointer col_name

int	tbpsta()
pointer	tbtopn()
pointer grp_tbggrp()

begin

	call atb_init(asp)

	call smark(sp)
	call salloc(col_name, 100, TY_CHAR)

	call eu_clstr(tb_name, "asptable")

	iferr(ATB_TB(asp) = tbtopn(Memc[tb_name], READ_ONLY, NULL))
		call ac_error(1, "No such table, ", Memc[tb_name])

	grp = grp_tbggrp(ATB_TB(asp), "aspectcols")

	call grp_list(grp, Memc[col_name])
	call tbcfnd(ATB_TB(asp), Memc[col_name], ATB_TIME(asp), 1)
	if(ATB_TIME(asp) == NULL)
		call ac_error(1, "No such column, ", Memc[col_name])

	call grp_list(grp, Memc[col_name])
	call tbcfnd(ATB_TB(asp), Memc[col_name], ATB_ASPECTW(asp), 1)
	if(ATB_ASPECTW(asp) == NULL)
		call ac_error(1, "No such column, ", Memc[col_name])

	call grp_list(grp, Memc[col_name])
	call tbcfnd(ATB_TB(asp), Memc[col_name], ATB_ASPECTX(asp), 1)
	if(ATB_ASPECTX(asp) == NULL)
		call ac_error(1, "No such column, ", Memc[col_name])

	call grp_list(grp, Memc[col_name])
	call tbcfnd(ATB_TB(asp), Memc[col_name], ATB_ASPECTY(asp), 1)
	if(ATB_ASPECTY(asp) == NULL)
		call ac_error(1, "No such column, ", Memc[col_name])

	call grp_list(grp, Memc[col_name])
	call tbcfnd(ATB_TB(asp), Memc[col_name], ATB_ASPECTZ(asp), 1)
	if(ATB_ASPECTZ(asp) == NULL)
		call ac_error(1, "No such column, ", Memc[col_name])

	ATB_NROWS(asp) = tbpsta(ATB_TB(asp), TBL_NROWS)

	call grp_close(grp)
	call sfree(sp)

end


#
# ac_defout() defines the output table by copying the column definitions
#	from the aspect table.
#

procedure	ac_defout(out, asp, ephem_file, len, ra, dec)

pointer out				#I/O output table
pointer asp				#I   aspect table
char	ephem_file[len]			#I   ephemeris filename
int	len				#I   length of string
double	ra				#I   fixed ra  position
double	dec				#I   fixed dec position

pointer	sp				# stack pointer
pointer t_col				# time column name
pointer tb_name				# output table name

pointer	tbtopn()
pointer tbccpy()

begin

	call atb_init(out)

	call smark(sp)
	call salloc(t_col, SZ_LINE, TY_CHAR)

	call eu_clstr(tb_name, "output")
	
	ATB_TB(out) = tbtopn(Memc[tb_name], NEW_FILE, NULL)

	ATB_TIME(out)    = tbccpy(ATB_TB(out), ATB_TIME(asp))
	ATB_ASPECTW(out) = tbccpy(ATB_TB(out), ATB_ASPECTW(asp))
	ATB_ASPECTX(out) = tbccpy(ATB_TB(out), ATB_ASPECTX(asp))
	ATB_ASPECTY(out) = tbccpy(ATB_TB(out), ATB_ASPECTY(asp))
	ATB_ASPECTZ(out) = tbccpy(ATB_TB(out), ATB_ASPECTZ(asp))

        # set max no. of user params
        call tbpset (ATB_TB(out), TBL_MAXPAR, MAX_USR_PARS)

	call tbtcre(ATB_TB(out))

	call tbcigt(ATB_TIME(out), TBL_COL_NAME,  Memc[t_col],  SZ_COLNAME)

        # copy header cards
        call hcpkill(ATB_TB(asp),"TABLE",ATB_TB(out),"TABLE","")

	call tbhadt(ATB_TB(out), "TIME", Memc[t_col])
	call tbhadt(ATB_TB(out), "APRT_POS", ephem_file)
	call tbhadd(ATB_TB(out), "RA_FIX", ra)
	call tbhadd(ATB_TB(out), "DEC_FIX", dec)

	call sfree(sp)

end


#
# ac_deffix() gets the fixed object position and turns it into the fixed
#	quaternion.
#

procedure	ac_deffix(q, ra_i, dec_i)

double	q[AC_NELMS]			#O quaternion
double	ra_i				#O input ra
double	dec_i				#O input dec

double	ra
double	dec
pointer sp
pointer ra_units

double	clgetd()

begin
	call smark(sp)

	ra_i = clgetd("ra")
	dec_i = clgetd("dec")
	call eu_clstr(ra_units, "raunits")

	if(Memc[ra_units] == 'h' || Memc[ra_units] == 'H')
	   ra = TWOPI * ra_i / 24.0
	else if(Memc[ra_units] == 'd' || Memc[ra_units] == 'D')
	   ra = TWOPI * ra_i / 360.0
	else
	   call ac_error(1,
            "Invalid ra units, must be hours (H/h) or Degrees (D/d): ",
	    Memc[ra_units])

	dec = TWOPI * dec_i / 360.0

	call er_qfrompos(ra, dec, q)

	call sfree(sp)
end


#
# atb_init() initializes the structure.
#

procedure	atb_init(atb)

pointer	atb				#I/O aspect table struct

begin

	call atb_malloc(atb)

	ATB_ROW(atb) = 1
	ATB_NROWS(atb) = 0
	ATB_TB(atb) = NULL
	ATB_TIME(atb) = NULL
	ATB_ASPECTW(atb) = NULL
	ATB_ASPECTX(atb) = NULL
	ATB_ASPECTY(atb) = NULL
	ATB_ASPECTZ(atb) = NULL
	ATB_RA(atb) = NULL
	ATB_DEC(atb) = NULL
	ATB_RACONV(atb) = 1.0
	ATB_DECCONV(atb) = 1.0

end



#
# ac_close() closes the three main aspect table structures.
#

procedure	ac_close(a, b, c)

pointer	a
pointer	b
pointer	c

begin

	call atb_close(a)
	call atb_close(b)
	call atb_close(c)

end


#
# atb_close() closes and releases the aspect table structure.
#

procedure	atb_close(atb)

pointer	atb				#I/O aspect table struct

begin

	call tbtclo(ATB_TB(atb))

	call atb_free(atb)

end




#
# ac_error() simply puts two strings together and outputs them as the error
#	message.
#

procedure	ac_error(errno, str1, str2)

int	errno				#I
char	str1[ARB]			#I
char	str2[ARB]			#I

pointer	sp
pointer err_msg

begin

	call smark(sp)
	call salloc(err_msg, SZ_LINE, TY_CHAR)
	call strcpy(str1, Memc[err_msg], SZ_LINE)
	call strcat(str2, Memc[err_msg], SZ_LINE)
	call error(errno, Memc[err_msg])
	call sfree(sp)

end

