# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  backmon.h
# 	Author	:  tnk
# 	Date	:  12/10/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) backmon.h -r1.4 12/10/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	08/09/95   tnk	changed from DSS to "Look" (made more generic)
# 	04/29/93   tnk	rearranged parameters
# 	12/17/92   tnk	added ram angle & velocity vectors
# 	12/10/92   tnk	date and time created 92/12/10 11:55:16 by tnk
# H-
# 
#


#  backmon.h
#  Background Monitors:
#  Contains info needed to create ST table columns:

#  @(#)backmon.h	1.4 8/9/95

#  Column  Quantity    units      description
#          --------    ------     -----------
#    1     TableTime   seconds    aspect time
#                   
#    2     LookLong     degrees    Look direction long (geographic coords)
#                   
#    3     LookLat      degrees    Look direction lat (geographic coords)
#                   
#    4     SatLong     degrees    Satellite position longitude (geo coords)
#                   
#    5     SatLat      degrees    Satellite position latitude (geo coords)
#                   
#    6     SatHt       km         Satellite height from earth's center
#                   
#    7     SunLong     degrees    Sun Position longitude (geo coords)
#                   
#    8     SunLat      degrees    Sun Pos. latitude (geo coords)
#                   
#    9     ASunZen     degrees    antisolar_zenith_angle
#                                 angle from antisun to spacecraft zenith
#                   
#    10    LookZen      degrees    Look_zenith_angle
#                                 angle from Look direction to s/c zenith
#                   
#    11    LookASun     degrees    Look_antisolar_ang
#                                 angle from Look dir. to antisun dir.
#                   
#    12    GMPos       degrees    pos. of Greenwich meridian at current time
#
#    13    MagLat      degrees    Magnetic Latitude
#
#    14    LookVelView  degrees    Look velocity-view angle
#
#    15    SatVelX     km/s       satellite velocity vector X
#
#    16    SatVelY     km/s       satellite velocity vector Y
#
#    17    SatVelZ     km/s       satellite velocity vector Z
#
#    18    MMJD        days       Modified Julian Day - 40000


# define output column indices

define      C_TIME         1 
define      C_LOOKLONG     2 
define      C_LOOKLAT      3 
define      C_SATLONG      4 
define      C_SATLAT       5 
define      C_SATHT        6 
define      C_SUNLONG      7 
define      C_SUNLAT       8 
define      C_ASUNZENITH   9 
define      C_LOOKZENITH  10
define      C_LOOKASUN    11
define      C_GMPOS       12
define      C_MAGLAT      13
define      C_LOOKVELVIEW 14
define      C_SATVELX     15
define      C_SATVELY     16
define      C_SATVELZ     17
define      C_MMJD        18


# define output column strings

string      S_TIME       "TableTime"
string      S_LOOKLONG   "LookLong"
string      S_LOOKLAT    "LookLat"
string      S_SATLONG    "SatLong"
string      S_SATLAT     "SatLat"
string      S_SATHT      "SatHt"
string      S_SUNLONG    "SunLong"
string      S_SUNLAT     "SunLat"
string      S_ASUNZENITH "ASunZen"
string      S_LOOKZENITH "LookZen"
string      S_LOOKASUN   "LookASun"
string      S_GMPOS      "GMPos"
string      S_MAGLAT     "MagLat"
string      S_LOOKVELVIEW "LookVelView"
string      S_MMJD       "MMJD"
string      S_SATVELX    "VelocityX"
string      S_SATVELY    "VelocityY"
string      S_SATVELZ    "VelocityZ"

define N_OUT_COLS 18            # number of output columns






