# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_backmon.x
# 	Author	:  tnk
# 	Date	:  12/10/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_backmon.x -r1.17 12/10/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	12/21/99   bboyd	 diagnostics added and then removed
# 	12/16/99   bboyd	fake for testing
# 	01/13/97   tnk	took out booleans controlling output columns
# 	12/19/96   tnk	reworked parameters
# 	10/24/96   tnk	incr MAX_USR_PARS
# 	10/21/96   tnk	took out extra bogus arg to tbhgtr calls
# 	10/21/96   tnk	now handles TIMEDEL, DELTIME, or no delta header cards; also DETNAM instead of INSTRUME
# 	09/13/95   tnk	set max table user params
# 	08/09/95   tnk	changed from DSS to "Look" (made more generic)
# 	02/03/95   tnk	added ds to detlist
# 	08/24/94   tnk	corrected badcol arg datatype
# 	08/08/94   tnk	replaced tbhckl with hcpkill
# 	03/09/94   tnk	fixed TIME card
# 	04/30/93   tnk	made detectors parm case insensitive
# 	04/29/93   tnk	rearranged parameters
# 	12/17/92   tnk	added ram angle & velocity vectors
# 	12/10/92   tnk	date and time created 92/12/10 11:55:14 by tnk
# H-
# 
#



# INCLUDE FILES
include <tbset.h>
include <math.h>
include <time.h>

# Create a table of monitors useful for background analysis.
# Reads aspect & satellite position (X,Y,Z) from Table1
# Creates a new table containing various background monitors
# (see backmon.h)

procedure t_backmon()


#AUTHOR : T. Kilsdonk
#DATE: 8 Dec 92
# @(#)t_backmon.x	1.17 12/21/99 CEA

#FUNCTIONS :
pointer tbtopn()                # table initialization
int     clgeti()                # get integer CL parameter
double  clgetd()                # get double CL parameter
int     edd_boresight()         # boresight info
real	tbhgtr()	        # get header table parameter value
long    clktime()               # time in seconds

define PROG_REV "@(#)t_backmon.x	1.17 12/21/99"

# Parameter file strings
define OUT_TABLE_PARM  "output"     # output table
define SKIP_PARM       "skip"       # input skip factor
define MAX_TIME_GAP_PARM "maxtimegap" # Max gap (secs) for sat vel & vel/view 
define REF_TABLE_PARM  "refdata"   # reference table
define DET_NAME_PARM   "detectors"   # detector to use for boresight 
define ORIENT_TABLE_PARM "orient"   # input table 
define TIME_PARM   "time"    # orientation time column
define ASP_W_PARM  "aspectw"    # aspect W,X,Y,Z columns
define ASP_X_PARM  "aspectx"    # 
define ASP_Y_PARM  "aspecty"    # 
define ASP_Z_PARM  "aspectz"    # 
define POS_X_PARM  "posx"       # position x, y, z 
define POS_Y_PARM  "posy"       #
define POS_Z_PARM  "posz"       #

string HIST_RECORD    "HISTORY"
define MAX_USR_PARS   120       # max. number of user parameters in table

# get all the info needed to define the output columns
include "backmon.h"

#LOCAL VARIABLES :       
pointer   stackp          # stack pointer 
pointer   in_name         # name of input ST table
pointer   out_name        # name of output ST table
pointer   ref_name        # name of reference ST table
pointer   in_tab          # input table ptr
pointer   out_tab         # output table ptr
pointer   ref_tab         # reference table pointer (for boresight)
pointer   det_name        # name of detector to use for boresight quat
pointer   time_colname    # aspect time colname
pointer   aspw_colname    # aspect w,x,y,z colnames
pointer   aspx_colname
pointer   aspy_colname
pointer   aspz_colname
pointer   posx_colname    # position x,y,z colnames
pointer   posy_colname
pointer   posz_colname
pointer   time_col           # aspect time column pointer
pointer   asp_col[4]          # aspect w,x,y,z column pointers
pointer   pos_col[3]          # position x,y,z column pointers
pointer   out_col[N_OUT_COLS] # output column pointers
int       skip                # input_file skip factor
pointer   edd                 # ptr to edd structure of ref info
double    bore_qfor[4]        # boresight forward quaternion
int       stat                # status return
real      deltime             # table delta time
pointer   hist_buff           # buffer to create history records
long      ltime               # time in seconds
pointer   time_buff           # ascii time buffer
double    max_time_gap        # max time allowed for vel-view & S/C vel vects


begin


  # mark the stack
  call smark(stackp)              

  # get parameters from param file
  call eu_clstr(out_name,   OUT_TABLE_PARM) # output table
  skip = clgeti(                 SKIP_PARM) # skip factor
  max_time_gap = clgetd( MAX_TIME_GAP_PARM) # max time gap
  call eu_clstr(in_name,    ORIENT_TABLE_PARM) #input aspect/pos table
  call eu_clstr(ref_name,   REF_TABLE_PARM) #reference table (detector info)
  call eu_clstr(det_name,    DET_NAME_PARM) # detector name
  call eu_clstr(time_colname,    TIME_PARM) # 
  call eu_clstr(aspw_colname,    ASP_W_PARM) # 
  call eu_clstr(aspx_colname,    ASP_X_PARM) # 
  call eu_clstr(aspy_colname,    ASP_Y_PARM) # 
  call eu_clstr(aspz_colname,    ASP_Z_PARM) # 
  call eu_clstr(posx_colname,    POS_X_PARM) # 
  call eu_clstr(posy_colname,    POS_Y_PARM) # 
  call eu_clstr(posz_colname,    POS_Z_PARM) # 


  # convert detector name to lower case
  call strlwr(Memc[det_name])

  # open the reference table
  call et_open(ref_tab, Memc[ref_name])
  if (ref_tab == NULL) {
     call eprintf("Cannot access reference table: %s\n")
     call pargstr(Memc[ref_name])
     return
  }

  # get the boresight quaternion for the specified detector
  call edd_new(edd, ref_tab, Memc[det_name])
  if (edd == NULL) {
     call eprintf("Cannot find detector '%s' in refdata\n")
     call pargstr(Memc[det_name])
     return
  }

  # open input table
  # append table name to obsdir
  in_tab = tbtopn(Memc[in_name], READ_ONLY, 0)

  # find the desired input column pointers
  call tbcfnd(in_tab, Memc[time_colname], time_col,    1)  # time col
  call tbcfnd(in_tab, Memc[aspw_colname], asp_col[1],  1)  # aspect W column
  call tbcfnd(in_tab, Memc[aspx_colname], asp_col[2],  1)  # aspect x column
  call tbcfnd(in_tab, Memc[aspy_colname], asp_col[3],  1)  # aspect y column
  call tbcfnd(in_tab, Memc[aspz_colname], asp_col[4],  1)  # aspect z column
  call tbcfnd(in_tab, Memc[posx_colname], pos_col[1],  1)  # position X
  call tbcfnd(in_tab, Memc[posy_colname], pos_col[2],  1)  # position y
  call tbcfnd(in_tab, Memc[posz_colname], pos_col[3],  1)  # position z

  # get the delta time from input table, create new delta time
  # look for TIMEDEL first, then DELTIME
  iferr (deltime = skip * tbhgtr(in_tab, "TIMEDEL"))
  {
     iferr (deltime = skip * tbhgtr(in_tab, "DELTIME"))
     {
        deltime = INDEFR
     }
  }

  # check them all
  if (time_col  == NULL) call bad_col(Memc[time_colname])
  if (asp_col[1] == NULL) call bad_col(Memc[aspw_colname])
  if (asp_col[2] == NULL) call bad_col(Memc[aspx_colname])
  if (asp_col[3] == NULL) call bad_col(Memc[aspy_colname])
  if (asp_col[4] == NULL) call bad_col(Memc[aspz_colname])
  if (pos_col[1] == NULL) call bad_col(Memc[posx_colname])
  if (pos_col[2] == NULL) call bad_col(Memc[posy_colname])
  if (pos_col[3] == NULL) call bad_col(Memc[posz_colname])


  stat= edd_boresight(edd, bore_qfor[1],bore_qfor[2],bore_qfor[3],bore_qfor[4])
  if (stat == NO) {
     call eprintf("Cannot read boresight information from reference tables\n")
     return
  }

  #open the output table
  out_tab = tbtopn(Memc[out_name], NEW_FILE, 0)


  #define the output columns
  #                   --->  Name        Unit       Format
  call tbcdef(out_tab,
        out_col[C_TIME],     S_TIME,     "Seconds", "%15.4f",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_LOOKLONG],  S_LOOKLONG,  "degrees", "%9.3f ",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_LOOKLAT],   S_LOOKLAT,   "degrees", "%9.3f ",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_SATLONG],  S_SATLONG,  "degrees", "%9.3f ",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_SATLAT],   S_SATLAT,   "degrees", "%9.3f ",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_SATHT],    S_SATHT,    "km",      "%9.3f ",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_SUNLONG],  S_SUNLONG,  "degrees", "%9.3f ",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_SUNLAT],   S_SUNLAT,   "degrees", "%9.3f ",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_ASUNZENITH],  S_ASUNZENITH,  "degrees", "%9.3f ",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_LOOKZENITH],   S_LOOKZENITH,   "degrees", "%9.3f ",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_LOOKASUN],  S_LOOKASUN,  "degrees", "%9.3f ",TY_DOUBLE,1,1)

  call tbcdef(out_tab,
        out_col[C_GMPOS],    S_GMPOS,   "degrees", "%9.3f ",TY_DOUBLE,1,1) 

  call tbcdef(out_tab,
        out_col[C_MAGLAT],   S_MAGLAT,   "degrees", "%9.3f ",TY_DOUBLE,1,1) 

  call tbcdef(out_tab,
        out_col[C_LOOKVELVIEW], S_LOOKVELVIEW, "degrees","%9.3f ",TY_DOUBLE,1,1) 

  call tbcdef(out_tab,
        out_col[C_SATVELX], S_SATVELX, "km/hr","%9.1f ",TY_DOUBLE,1,1) 

  call tbcdef(out_tab,
        out_col[C_SATVELY], S_SATVELY, "km/hr","%9.1f ",TY_DOUBLE,1,1) 

  call tbcdef(out_tab,
        out_col[C_SATVELZ], S_SATVELZ, "km/hr","%9.1f ",TY_DOUBLE,1,1) 

  call tbcdef(out_tab,
        out_col[C_MMJD],      S_MMJD,      "days", "%17.10f ",TY_DOUBLE,1,1) 

  # set max no. of user params
  call tbpset (out_tab, TBL_MAXPAR, MAX_USR_PARS)

  # create the table
  call tbtcre(out_tab)
  
  # copy header parameters from input file, with a kill list.
  call hcpkill(in_tab, "TABLE", out_tab, "TABLE", "")

  # write TIME & TIMEDEL parameters
  call tbhadt(out_tab, "TIME", S_TIME)  
  if (! IS_INDEFR(deltime)) call tbhadr(out_tab, "TIMEDEL", deltime)

  # write out DETNAM
  call tbhadt(out_tab, "DETNAM", Memc[det_name])  

  # add new header parameters (processing history)
     call salloc(hist_buff, SZ_LINE, TY_CHAR)

     # << prog revision number >>
     call sprintf(Memc[hist_buff],SZ_LINE,
        "backmon: Program revision number: %s")
     call pargstr(PROG_REV)
     call tbhadt(out_tab, HIST_RECORD, Memc[hist_buff])

     # << input table name >>
     call sprintf(Memc[hist_buff], SZ_LINE, "backmon: Input table: %s")
     call pargstr(Memc[in_name])
     call tbhadt(out_tab, HIST_RECORD, Memc[hist_buff])

     # << table creation date/time >>
     ltime = clktime(0)
     call salloc(time_buff, SZ_TIME, TY_CHAR)
     call cnvtime(ltime, Memc[time_buff], SZ_TIME)
     call sprintf(Memc[hist_buff],SZ_LINE,
         "backmon: Output table creation date: %s")
     call pargstr(Memc[time_buff])
     call tbhadt(out_tab, HIST_RECORD, Memc[hist_buff])

     # << detector used for boresight >>
     call sprintf(Memc[hist_buff],SZ_LINE,
        "backmon: Detector used for boresight: %s")
     call pargstr(Memc[det_name])
     call tbhadt(out_tab, HIST_RECORD, Memc[hist_buff])

  # Calculate the values for the output table
  call calc_cols(in_tab, skip, asp_col, time_col, pos_col, out_tab, 
     out_col, bore_qfor, max_time_gap)

  # Close the tables
  call tbtclo(in_tab)
  call tbtclo(out_tab)

  call edd_close(edd)
  call et_close(ref_tab)

  # free the stack
  call sfree(stackp)
end
#***************************************************
procedure bad_col(col_name)
char col_name[ARB]   #I name of column

# report bad column name & abort

begin
     call eprintf("Cannot access column: %s\n")
     call pargstr(col_name)
     call fatal(1, "Exiting...")
end     

