# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_compspec.x
# 	Author	:  akeith
# 	Date	:  04/30/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_compspec.x -r1.24 04/30/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	11/03/95   tnk	changed NR copyright to cal
# 	08/10/95   tnk	added Num. Recipes copyright
# 	09/14/94   tnk	adjusted df height/area code
# 	09/14/94   tnk	consolidated df input
# 	09/13/94   tnk	reworked df image i/o
# 	09/09/94   tnk	height to area conversions
# 	09/08/94   tnk	reworked error selection
# 	09/02/94   tnk	added one-to-many spectra comparisons
# 	09/02/94   tnk	removed subroutines
# 	09/02/94   tnk	major reworking for error handling
# 	08/29/94   tnk	changed list handling
# 	07/21/94   tnk	fixed areas, added title arg
# 	07/20/94   tnk	coded x1,x2,res options
# 	07/20/94   tnk	added minbinwid
# 	07/19/94   tnk	switch to using df
# 	07/14/94   tnk	reworded error messages
# 	01/11/93   tnk	added check for non-overlapping model & data
# 	11/13/92   akeith	commented out all code dealing with the KS test.
# 	10/02/92   gareth	added mwcs. now works with either
# 	05/14/92   ericco	added documentation
# 	05/13/92   gareth	corrected for the new plf libarary
# 	04/30/92   akeith	date and time created 92/04/30 13:08:19 by akeith
# H-
# 
#

include <dfset.h>
include <imhdr.h>

# @(#)t_compspec.x	1.24 11/3/95

#   Compare sets of spectra, stored in images, using chisqr test.
#   Error factors for chisqr calculation come from: image error 
#      bands of one-sigma values (if they exist), or
#      if no error band, then the poisson errors flag indicates error
#      source:
#         if yes, errors=spectral data values
#         if no, errors=0  (i.e., it's a model)
#   Output goes to STDOUT.
#
# Makes use of Numerical Recipes library routines,
# Copyright (C) 1986 Numerical Recipes Software


procedure	t_compspec ()

int		list1, list2
double		res
double		lowx, highx
int		constraints
bool		title
bool		poierr1
bool		poierr2
pointer		spec1
pointer		spec2
pointer		sp
real		chsq, prob
int		degfree
int		nsamp
pointer         df1, df2
pointer         dferr1, dferr2
pointer         x1,y1
pointer         x2,y2
pointer		err1,err2
int		len1, len2
int		nread
double		xmin1,xmax1
double		xmin2,xmax2
double		xmin,xmax
double		ymin,ymax
double		minwid1
double		minwid2
int		n_elems
int		datband1, datband2
int		errband1, errband2
bool		model1, model2
int		chistat
int		lenlist1, lenlist2

# Functions
int		imtopenp()
int		imtlen()
int		imtgetim()
double		clgetd()
bool		clgetb()
int		clgeti()
int		df_rgetr()
int		df_len()
double		minbinwid()
int  		chi_err()


#############################################################################
begin

  call smark(sp)

  # put in Numerical Recipes copyright string
  call nr_copyr()

  call salloc(spec1, SZ_LINE, TY_CHAR)
  call salloc(spec2, SZ_LINE, TY_CHAR)

  list1 = imtopenp("spectra1")     # list of spectra to compare
  list2 = imtopenp("spectra2")     # 2nd list of spectra
  poierr1  = clgetb("poiss1errs")  # poisson errors flag for list1
  poierr2  = clgetb("poiss2errs")  # poisson errors flag for list2
  res      = clgetd("resolution")  # wavelength resolution for comparisons
  lowx     = clgetd("x1")          # lower wavelength cutoff
  highx    = clgetd("x2")          # upper wavelength cutoff
  constraints = clgeti("constraints")    # no. of constraints
  title    = clgetb("title")       # produce title on output?

  lenlist1 = imtlen(list1)
  lenlist2 = imtlen(list2)
  if (lenlist1 > 1 && lenlist2 > 1)
  {
     if (lenlist1 != lenlist2)
     {
       call eprintf("If there is more than one item in both of the \n")
       call eprintf("'spectra1' and 'spectra2' lists, they must be the same length.\n")
       call error (101,"Invalid list lengths")
     }
  }

  if (title)
  {
    call printf("spectrum1 spectrum2 statistic probability degfree\n")
    call printf("==================================================\n")  
  }

  # compare spectral pairs from list1 & list2
  while (imtgetim(list1, Memc[spec1], SZ_LINE) != EOF &&	
	 imtgetim(list2, Memc[spec2], SZ_LINE) != EOF) {

    call get_de_bands(Memc[spec1],df1,dferr1,datband1,errband1)
    call get_de_bands(Memc[spec2],df2,dferr2,datband2,errband2)

    # find overlapping region
    call df_minmax(df1, xmin1, xmax1, ymin, ymax)
    call df_minmax(df2, xmin2, xmax2, ymin, ymax)

    xmin = xmin1
    if (xmin2 > xmin)  xmin = xmin2    
    if (lowx != INDEFD)
    {
       if (lowx > xmin)
          xmin = lowx     
    }
    xmax = xmax1
    if (xmax > xmax2)  xmax = xmax2
    if (highx != INDEFD)
    {
       if (highx < xmax)
          xmax = highx
    }

    if (xmin >= xmax)
    {
       call eprintf("%s and %s do not overlap\n")
       call pargstr(Memc[spec1])
       call pargstr(Memc[spec2])

    }
    else
    {

       if (res != INDEFD)
       { 
          # rebin to desired resolution
	  n_elems=(xmax-xmin)/res
          call df_rsdomain(df1,xmin,res,n_elems,df1)
          call df_rsdomain(df2,xmin,res,n_elems,df2)

          if (errband1 > 0)
             call df_rsdmnerr(dferr1,xmin,res,n_elems,dferr1)
          if (errband2 > 0)
             call df_rsdmnerr(dferr2,xmin,res,n_elems,dferr2)

       }
       else
       {
          # match domains to the coarser df
          minwid1 = minbinwid(df1)
          minwid2 = minbinwid(df2)

          if (minwid1 <= minwid2)
          {
             call df_dfdomain(df1,df2,df1)
             if (errband1 > 0)
                call df_dfdmnerr(dferr1,df2,dferr1)
          }
          else
          {
             call df_dfdomain(df2,df1,df2)
             if (errband2 > 0)
                call df_dfdmnerr(dferr2,df1,dferr2)
          }
       }

       # trim down to the overlapping region
       call df_trim(df1,xmin,xmax,df1)
       call df_trim(df2,xmin,xmax,df2)

       if (errband1 > 0)
          call df_trim(dferr1,xmin,xmax,dferr1)
       if (errband2 > 0)
          call df_trim(dferr2,xmin,xmax,dferr2)


       len1 = df_len(df1)
       len2 = df_len(df2)

       call malloc(x1, len1, TY_REAL)
       call malloc(y1, len1, TY_REAL)
       call malloc(x2, len2, TY_REAL)
       call malloc(y2, len2, TY_REAL)
       call malloc(err1, len1, TY_REAL)
       call malloc(err2, len2, TY_REAL)

       # convert heights to areas
       call df_mbw(df1,df1)
       call df_mbw(df2,df2)

       if (errband1 > 0)
          call df_mbw(dferr1,dferr1)
       if (errband2 > 0)
          call df_mbw(dferr2,dferr2)

       nread = df_rgetr(df1, Memr[x1], Memr[y1], len1)
       nread = df_rgetr(df2, Memr[x2], Memr[y2], len2)

       # select appropriate errors
       # first see if either spectrum is a model (no errband, no poisson errs)
       model1 = false
       if ((errband1 <= 0) && ! poierr1)
          model1 = true
       model2 = false
       if ((errband2 <= 0) && ! poierr2)
          model2 = true

       call aclrr(Memr[err1], len1)
       call aclrr(Memr[err2], len2)
       if (!model1 && !model2)
       {
          # both spectra are data only
          if (errband1 > 0)
             nread = df_rgetr(dferr1, Memr[x1], Memr[err1], len1)
          else if (poierr1)
             call amovr(Memr[y1], Memr[err1], len1)

          if (errband2 > 0)
             nread = df_rgetr(dferr2, Memr[x2], Memr[err2], len2)
          else if (poierr2)
             call amovr(Memr[y2], Memr[err2], len2)
       }
       else
       {
          # one or both spectra are models
          # if both, use first model's values as the error
          if (model1)
             call amovr(Memr[y1], Memr[err1], len1)             
          else
             call amovr(Memr[y2], Memr[err2], len2)
       }

       # compute chisqr
       chistat = chi_err(Memr[y1], Memr[y2], Memr[err1], Memr[err2], len1, 
          constraints, degfree, chsq, prob)
       nsamp = len1 + len2
       if (chistat != YES)
       {
          chsq=INDEFR
	  prob=INDEFR
       }

       # output is "spec1 spec2 statistic probability degfree"
       call printf("%s %s %f %f %d\n")
       call pargstr(Memc[spec1])
       call pargstr(Memc[spec2])
       call pargr(chsq)
       call pargr(prob)
       call pargi(degfree)

       call mfree(x1,TY_REAL)
       call mfree(y1,TY_REAL)
       call mfree(x2,TY_REAL)
       call mfree(y2,TY_REAL)
       call mfree(err1,TY_REAL)
       call mfree(err2,TY_REAL)
    }

    call df_close(df1)
    call df_close(df2)
    call df_close(dferr1)
    call df_close(dferr2)

    # if doing a one to many compare, rewind the list with one spectrum
    if (lenlist1 == 1 && lenlist2 > 1)
       call imtrew(list1)
    else if (lenlist2 == 1 && lenlist1 > 1)
       call imtrew(list2)

  }


  call clpcls(list1)
  call clpcls(list2)

  call sfree(sp)

end

#-----------------------------------------
# fetch data and error bands from an image
#
procedure get_de_bands(filename,df,dferr,datband,errband)
char     filename[ARB]  #I image name
pointer  df             #O data df
pointer  dferr          #O error df
int      datband        #O data band number
int      errband        #O error band number

int     line
pointer im
pointer immap()
pointer pd_rimg()

begin

    im = immap(filename,READ_ONLY,0)

    # look for data and error (sigma) bandid cards.
    call de_bands(im,datband,errband)

    line=1
    if (datband <= 0) datband=1
    df = pd_rimg(im, line, datband)
    if (df == NULL)
    {
         call eprintf("Could not read data band from image file %s\n")
         call pargstr(filename)
         call error(102,"input failure")
    }
    call df_pedge(df,0.0d0,0.0d0)

    # convert areas to heights
    call df_dbw(df,df)

    if (errband > 0)
    {
       dferr = pd_rimg(im, line, errband)
       if (dferr == NULL)
       {
         call eprintf("Could not read error band from image file %s\n")
         call pargstr(filename)
         call error(103,"input failure")
       }
       call df_pedge(dferr,0.0d0,0.0d0)

       # square sigma to get variance
       call df_powk(dferr,2,dferr)

       # convert from areas to heights
       call df_dbw(dferr,dferr)
    }

    call imunmap(im)

end
