# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  detmapos.x
# 	Author	:  akeith
# 	Date	:  04/04/95
# 	Synopsis:  
# 	Keywords:
# 		@(#) detmapos.x -r1.2 04/04/95 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	08/14/95   tnk	added keywords, aligned comments
# 	04/04/95   akeith	date and time created 95/04/04 16:03:46 by akeith
# H-
# 
#

# @(#)detmapos.x	1.2 8/14/95 UCB CEA
#
#  compute detector positions - part of detpos
#

include <tbset.h>
include <math.h>
include "detma.h"

procedure det_position(mas, num, asp, aq, atc, ra, dec)

pointer	mas[num]		#I array of detector info ptrs
int	num			#I number of detectors
pointer	asp			#I aspect table
pointer	aq[4]			#I aspect quaternion column ptrs
pointer	atc			#I aspect time column ptr
double	ra			#I source RA
double	dec			#I source DEC

int 	i			# loop counter
int 	num_rows		# number of rows on the aspect table
int	d			# detector id
int	orow			# output row
double	time			# aspect time
double	qasp[4]			# The j2000 to ACS quaternion (aspect)
double	qdet[4]			# The j2000 to detector quaternion
double	out[5]			# resulting table row
double	detlong			# detector longitude coordinate
double	detlat			# detector latitude coordinate
double	thetax			# off-boresight angle along x-axis
double	thetay			# off-boresight angle along y-axis
double	dx			# detector coordinate X
double	dy			# detector coordinate Y
bool	null[4], tnull		# table null values
pointer	ma			# aspect structure

int tbpsta()

begin
  # find num of rows in input table
  num_rows = tbpsta(asp, TBL_NROWS )

  orow = 1

  # loop through the aspect table
  do i = 1, num_rows {

	# read aspect time and quaternion
    call tbrgtd(asp, atc, time, tnull, 1, i)
    call tbrgtd(asp, aq, qasp, null, 4, i)

    if (tnull || null[1] || null[2] || null[3] || null[4]) ;
    else {
      do d = 1, num {
		ma = mas[d]

		# construct j2000 to detector quaternion by combining
		# aspect (j2000 to ACS) with ACS to detector quaternion
		call er_qmult(qasp, MA_QFOR(ma,1), qdet)

		# convert j2000 source position into detector coordinates which
		# are angles in radians in a spherical coordinate system that
		# has its origin at the boresight and x and y axes aligned with
		# the detector and positive z axis up the boresight.
		detlong = ra
		detlat = dec
		call er_qrotate(detlong, detlat, qdet)

		# convert spherical coordinates to off-axis angles and then to 
		# pixels on an ideal detector.  These are undistorted detector
		# coordinates.  The minus signs on the conversion to off-axis
		# angles are empirically determined to be neccessary.  We set the
		# coordinates to INDEF for angles larger than 20 degrees from the
		# boresight direction, which is way off the detector.

		# When a function becomes available to distort the detector 
		# coordinates, it will be called to produce DX and DY.  For now
		# we set those to INDEF.
		out[1] = time
		if (detlat >= 70.0*(PI/180.)) {
		  thetax = -cos(detlong)*(PI/2.0 - detlat) * (180./PI)
		  thetay = -sin(detlong)*(PI/2.0 - detlat) * (180./PI)
		  out[4] = thetax * MA_SCALE(ma) + MA_XBORE(ma)
		  out[5] = thetay * MA_SCALE(ma) + MA_YBORE(ma)
		  call det_distort(Memr[MA_DIST(ma)], MA_NX(ma), MA_NY(ma), MA_CT(ma), out[4], out[5], dx, dy)
		  out[2] = dx
		  out[3] = dy
		}
		else {
		  out[2] = INDEFD
		  out[3] = INDEFD
		  out[4] = INDEFD
		  out[5] = INDEFD
		}

		call tbrptd (MA_TP(ma), MA_COLUMN(ma,1), out, 5, orow)
      }
      orow = orow + 1
    }
  }
end



# Apply distortion correction to a pixel
procedure det_distort(dist, nx, ny, ct, x, y, dx, dy)

real	dist[nx,ny,2]	#I distortion image 
int	nx,ny		#I size of distortion image
pointer	ct		#I distortion image MWCS transformation pointer
double	x, y		#I undistorted coordinates
double	dx, dy		#O detector coordinates

int	ix, iy
real	y1, y2, y3, y4
double	lx, ly, t, u, delta

begin
  if (IS_INDEFD(x) || IS_INDEFD(y)) {
	dx = INDEFD
	dy = INDEFD
	return
  }

  # Convert from detector position to logical coords in distortion image
  call mw_c2trand(ct, x, y, lx, ly)

  ix = int(lx)
  iy = int(ly)

  # out of bounds of distortion map - apply no correction
  if (ix < 1 || iy < 1 || ix >= nx || iy >=ny) {
	dx = x
	dy = y
	return
  }

  t = lx - ix
  u = ly - iy

  # Bilinear interpolation

  y1 = dist[ix,iy,1]
  y2 = dist[ix+1,iy,1]
  y3 = dist[ix+1,iy+1,1]
  y4 = dist[ix,iy+1,1]

  # The distortion images contain the delta from distorted to undistorted
  # coordinates.
  delta = (1.0-t)*(1.0-u)*y1 + t*(1.0-u)*y2 + t*u*y3 + (1.0-t)*u*y4
  dx = x + delta

  y1 = dist[ix,iy,2]
  y2 = dist[ix+1,iy,2]
  y3 = dist[ix+1,iy+1,2]
  y4 = dist[ix,iy+1,2]

  # The distortion images contain the delta from distorted to undistorted
  # coordinates.
  delta = (1.0-t)*(1.0-u)*y1 + t*(1.0-u)*y2 + t*u*y3 + (1.0-t)*u*y4
  dy = y + delta
end
