# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_detpos.x
# 	Author	:  akeith
# 	Date	:  04/04/95
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_detpos.x -r1.10 04/04/95 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	12/27/96   tnk	changed pargs to pargstr in ra_units error msg
# 	12/26/96   tnk	added raunits param
# 	12/20/96   tnk	reworked parameters
# 	10/21/96   tnk	changed from DETECTOR to DETNAM
# 	10/21/96   tnk	checks for TIMEDEL, then DELTIME;  writes TIMEDEL
# 	09/21/95   mabbott	added check to skip any spectrometers in detector list
# 	08/14/95   tnk	changed TIME card to TableTime
# 	08/14/95   tnk	added et_close
# 	08/11/95   tnk	took out tabs, added sccs keywords
# 	04/04/95   akeith	date and time created 95/04/04 16:03:48 by akeith
# H-
# 
#

#
# Compute detector position of a point on the sky vs time
# @(#)t_detpos.x	1.10 12/27/96 UCB CEA

include <math.h>
include <tbset.h>
include <fset.h>
include <imhdr.h>
include "detma.h"

define DNUM_COLNAME "tif_number"
define XBORE_COLNAME "boresight_x_coord"
define YBORE_COLNAME "boresight_y_coord"
define SCALE_COLNAME "pixperdeg"
define DISTNAME_COLNAME "distort_map"

define PRIMARY_DELTIME_HDR    "TIMEDEL"
define ALTERNATE_DELTIME_HDR  "DELTIME"    

define INPUT_PARM  "input"      # aspect table name
define OUTPUT_PARM "output"     # output table name
define REFTAB_PARM "refdata"    # reference data table
define DETECT_PARM "detectors"
define RA_PARM     "ra"		# ra 
define DEC_PARM    "dec"
define RA_UNITS_PARM "raunits"  # units for ra
define TIME_PARM   "time"       # aspect time column
define ASP_W_PARM  "aspectw"    # aspect W,X,Y,Z columns
define ASP_X_PARM  "aspectx"    # 
define ASP_Y_PARM  "aspecty"    # 
define ASP_Z_PARM  "aspectz"    # 


procedure t_detpos()

pointer	asptab		# aspect table name
pointer	output		# output table prefix
pointer	reftab		# detector table name
pointer	out		# output table name
pointer	asp		# aspect table pointer
pointer	tp		# output table pointer

pointer	detectors	# detector list pointer
int	num		# number of list items
pointer	det		# detector name

pointer distname	# name of distortion map
pointer dim		# distortion map image ptr
pointer dmw		# distortion map MWCS ptr
pointer dct		# distortion map Ctrans ptr
pointer dist		# distortion map data
int	dnx, dny	# size of distortion map

pointer	aspw, aspx, aspy, aspz, aspt	# aspect column names
pointer	atc, aq[4]	# aspect column pointers
pointer ra_units	# units for RA

pointer	et		# reference structure
pointer ref_tp		# reference table 
pointer	edd		# detector information

pointer	sp		# stack marker

int	detnum		# detector number (list order)
int	dnum		# detector number (TIF number)
real	deltime		# table time interval
real	t0, t1		# time values for deltime
double	ra, dec		# source position
double	w, x, y, z	# quaternion
real	xbore, ybore, scale	# det info
pointer	ma, mas		# aspect structure pointers
pointer	str, buf

int	clplen(), clgfil()
real	tbhgtr()
double	clgetd()
bool	det_getinfo()
pointer	tbtopn(), clpopnu(), imgs3r(), immap(), mw_openim(), mw_sctran()

begin
  call smark(sp)

  call fseti(STDOUT, F_FLUSHNL, YES)

  # Process parameters
  call eu_clstr(asptab,  INPUT_PARM)
  call eu_clstr(output,  OUTPUT_PARM)
  call eu_clstr(reftab,  REFTAB_PARM)
  call eu_clstr(aspt,    TIME_PARM) # 
  call eu_clstr(aspw,    ASP_W_PARM) # 
  call eu_clstr(aspx,    ASP_X_PARM) # 
  call eu_clstr(aspy,    ASP_Y_PARM) # 
  call eu_clstr(aspz,    ASP_Z_PARM) # 
  call eu_clstr(ra_units, RA_UNITS_PARM)

  ra = clgetd(RA_PARM)
  if(Memc[ra_units] == 'h' || Memc[ra_units] == 'H')
     ra = 15.0 * ra * PI / 180.0
  else if(Memc[ra_units] == 'd' || Memc[ra_units] == 'D')
     ra = ra * PI / 180.0
  else
  {
     call eprintf("invalid ra_units: %s\n")
     call pargstr(Memc[ra_units])
     call error(1,"Ra units must be hours (H/h) or Degrees (D/d)")
  }

  dec = clgetd(DEC_PARM) * PI / 180.0

  # Open aspect table information
  asp = tbtopn(Memc[asptab], READ_ONLY, 0)

  call tbcfnd(asp, Memc[aspw], aq[1], 1)
    if(aq[1] == NULL) call eu_error(1, "Column not found, ", Memc[aspw])
  call tbcfnd(asp, Memc[aspx], aq[2], 1)
    if(aq[2] == NULL) call eu_error(1, "Column not found, ", Memc[aspx])
  call tbcfnd(asp, Memc[aspy], aq[3], 1)
    if(aq[3] == NULL) call eu_error(1, "Column not found, ", Memc[aspy])
  call tbcfnd(asp, Memc[aspz], aq[4], 1)
    if(aq[4] == NULL) call eu_error(1, "Column not found, ", Memc[aspz])
  call tbcfnd(asp, Memc[aspt], atc, 1)
    if(atc == NULL) call eu_error(1, "Column not found, ", Memc[aspt])


  iferr (deltime = tbhgtr(asp, PRIMARY_DELTIME_HDR))
  {
     iferr (deltime = tbhgtr(asp, ALTERNATE_DELTIME_HDR))
     {
        call tbegtr(asp, atc, 1, t0)
        call tbegtr(asp, atc, 2, t1)
        deltime = t1-t0
     }
  }

  # get reference data
  call et_open(et, Memc[reftab])
  if(et == NULL) {
	call sfree(sp)
	call eu_error(1, "Unable to find table, ", Memc[reftab])
  }
  call et_table (et, ref_tp)

  # process detector list
  detectors = clpopnu(DETECT_PARM)
  num = clplen(detectors)
  call salloc(det, SZ_LINE, TY_CHAR)
  call salloc(distname, SZ_LINE, TY_CHAR)
  call salloc(out, SZ_LINE, TY_CHAR)
  call salloc(str, SZ_LINE, TY_CHAR)
  call salloc(mas, num, TY_INT)

  detnum = 0
  while (clgfil(detectors, Memc[det], SZ_LINE) != EOF) {
	# make sure this is an imaging detector
	call eu_detect_id(Memc[det], dnum)
	if (dnum!=1 && dnum!=2 && dnum!=3 && dnum!=7) {
	  call eprintf("skipping %s - not an imaging instrument")
	  call pargstr(Memc[det])
	  next
	}

    detnum = detnum + 1

	# create a structure for each detector.
	call salloc(ma, MA_LEN, MA_TYP)
    Memi[mas+detnum-1] = ma

	# read in the detector to ACS quaternion from egodata
    call edd_new(edd, et, Memc[det])
    call edd_boresight(edd, w, x, y, z)
    MA_QFOR(ma,1) = w
    MA_QFOR(ma,2) = x
    MA_QFOR(ma,3) = y
    MA_QFOR(ma,4) = z
    call edd_close(edd)

	# get other detector info
	if ( ! det_getinfo(ref_tp, dnum, xbore, ybore, scale, Memc[distname], SZ_LINE)) {
	  call sfree(sp)
      call error(570,"Cannot find detector data for all of the detectors")
	}
	MA_XBORE(ma) = xbore
	MA_YBORE(ma) = ybore
	MA_SCALE(ma) = scale

	# open distortion image
	dim = immap(Memc[distname], READ_ONLY, 0)
	dmw = mw_openim(dim)
	dct = mw_sctran(dmw, "world", "logical", 3)
	dnx = IM_LEN(dim, 1)
	dny = IM_LEN(dim, 2)
	call malloc(dist, dnx*dny*2, TY_REAL)
	buf = imgs3r(dim, 1, dnx, 1, dny, 1, 2)
	call amovr(Memr[buf], Memr[dist], dnx*dny*2)
	call imunmap(dim)
	MA_DIST(ma) = dist
	MA_MW(ma) = dmw
	MA_CT(ma) = dct
	MA_NX(ma) = dnx
	MA_NY(ma) = dny

	# create an output table too.
    call sprintf(Memc[out], ARB, "%s%s")
    call pargstr(Memc[output])
    call pargstr(Memc[det])

    tp = tbtopn(Memc[out], NEW_FILE, 0)
    MA_TP(ma) = tp
    call tbcdef(tp, MA_COLUMN(ma,1), "TableTime", "Seconds", "%15.4f",
		TY_DOUBLE, 1, 1)
    call tbcdef(tp, MA_COLUMN(ma,2), "DX", "Pixels", "%g",
		TY_REAL, 1, 1)
    call tbcdef(tp, MA_COLUMN(ma,3), "DY", "Pixels", "%g",
		TY_REAL, 1, 1)
    call tbcdef(tp, MA_COLUMN(ma,4), "UDX", "Pixels", "%g",
		TY_REAL, 1, 1)
    call tbcdef(tp, MA_COLUMN(ma,5), "UDY", "Pixels", "%g",
		TY_REAL, 1, 1)

    call tbtcre(tp)

    # copy appropriate header cards
    call hcpkill(asp, "TABLE", tp, "TABLE", "")

    call tbhadt(tp, "TIME", "TableTime")
    call tbhadt(tp, "DETNAM", Memc[det])
    call tbhadd(tp, "DETASPW", w)
    call tbhadd(tp, "DETASPX", x)
    call tbhadd(tp, "DETASPY", y)
    call tbhadd(tp, "DETASPZ", z)

    call sprintf(Memc[str], SZ_LINE, "Created from %s for detector %s")
    call pargstr(Memc[asptab])
    call pargstr(Memc[det])
    call tbhadt(tp, "HISTORY", Memc[str])

    if (deltime != INDEFR) call tbhaddr(tp, PRIMARY_DELTIME_HDR, deltime)

  }

  # Calulate the values for the output table
  call det_position(Memi[mas], num, asp, aq, atc, ra, dec)

  # Close the tables
  do detnum = 1, num {
    ma = Memi[mas+detnum-1]
    call tbtclo(MA_TP(ma))

	# free distortion info
	call mw_ctfree(MA_CT(ma))
	call mw_close(MA_MW(ma))
	call mfree(MA_DIST(ma), TY_REAL)
  }
  call tbtclo(asp)

  call et_close(et)

  call sfree(sp)
end



bool procedure det_getinfo(tp, dnum, xbore, ybore, scale, name, maxch)

int		dnum, maxch
pointer	tp
real	xbore, ybore, scale
char	name[maxch]

int		nrows, row, tab_det, n
bool	found
pointer	sp, this, file, detnum_cp, xbore_cp, ybore_cp, scale_cp, distname_cp

int tbpsta(), fnldir()

begin
  call smark(sp)

  call salloc(this, SZ_LINE, TY_CHAR)
  call salloc(file, SZ_LINE, TY_CHAR)

  nrows = tbpsta (tp, TBL_NROWS)

  # get column pointers
  call tbcfnd (tp, DNUM_COLNAME, detnum_cp, 1)
  call tbcfnd (tp, XBORE_COLNAME, xbore_cp,   1)
  call tbcfnd (tp, YBORE_COLNAME, ybore_cp,   1)
  call tbcfnd (tp, SCALE_COLNAME, scale_cp,   1)
  call tbcfnd (tp, DISTNAME_COLNAME, distname_cp,   1)

  if (detnum_cp == NULL || xbore_cp == NULL ||
      ybore_cp == NULL   || scale_cp == NULL || distname_cp == NULL) {
	call sfree(sp)
    call error(650,"Could not find ref columns for boresight info")
  }

  # go down each row in table, see if it matches a detector
  found = false

  for (row=1; row <=nrows; row=row+1) {
	# get detector number of each row
	call tbegti(tp, detnum_cp, row, tab_det)

	if (tab_det == dnum) {
	  found = true

	  call tbegtr(tp, xbore_cp, row, xbore)
	  call tbegtr(tp, ybore_cp, row, ybore)
	  call tbegtr(tp, scale_cp, row, scale)
	  call tbegtt(tp, distname_cp, row, Memc[file], SZ_LINE)

	  break
	}
  }

  if (! found) {
	call eprintf("Error:  could not find boresight info for det %d\n")
	call pargi(dnum)
  } else {
	call tbtnam(tp, Memc[this], SZ_LINE)
	n = fnldir(Memc[this], name, maxch)
	call strcat(Memc[file], name, maxch)
  }

  call sfree(sp)

  return found
end
