# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_divea.x
# 	Author	:  akeith
# 	Date	:  04/21/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_divea.x -r1.7 04/21/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/21/96   tnk	searches for DETNAM, then INSTRUME, then uses detector param
# 	11/03/94   akeith	fixed so it works
# 	11/02/94   akeith	now takes indef for dx
# 	10/26/94   akeith	fixed exptime parameter
# 	10/10/94   akeith	added instrument parameter
# 	09/14/94   akeith	uses the spc library
# 	04/21/94   akeith	date and time created 94/04/21 17:52:25 by akeith
# H-
# 
#


#	Version: @(#)t_divea.x	1.7 10/21/96 UCB CEA EGO

include	"dva.h"
include	<dfset.h>
include	<imhdr.h>

define	SRC_STR		"spectrum - spectral flux"
define	ERR_STR		"sigma - spectral error flux"

#
# t_divea()
#

procedure	t_divea()

pointer	sp			#
pointer	ref			#
pointer	input			#
pointer	output			#

pointer	dva, dva_open()
pointer	in, out
pointer	immap()

begin

	call smark(sp)

	call eu_clstr(ref, "refdata")
	call eu_clstr(input, "input")
	call eu_clstr(output, "output")

	in = immap(Memc[input], READ_ONLY, NULL)
	out = immap(Memc[output], NEW_IMAGE, NULL)
	call hcpkill(in, "IMAGE", out, "IMAGE", NULL)

	dva = dva_open(in, Memc[ref], "exptime")

	call dva_mkout(dva, in, Memc[input], out, Memc[output])

	call spc_wspc(out, DVA_EA(dva), DVA_EXPT(dva), DVA_SPEC(dva))
	call spc_werr(out, DVA_EA(dva), DVA_EXPT(dva), DVA_ERR(dva))

	call imunmap(in)
	call imunmap(out)

	call dva_close(dva)

	call sfree(sp)

end


#
# dva_open()
#

pointer	procedure	dva_open(in, reference, exposure)

pointer	in			#I
char	reference[ARB]		#I
char	exposure[ARB]		#I

int	len			#
double	xo, x, y		#
pointer sp			#
pointer	det			#
pointer	dva			#

int	stat, df_dgetd()
double	imgetd()
double	clgetd()
pointer	spc_gspc()
pointer	spc_gerr()

begin

#    Initialize DVA_ structure to zero.

	call dva_malloc(dva)

	DVA_SPEC(dva) = NULL
	DVA_ERR(dva) = NULL
	DVA_EA(dva) = NULL
	DVA_DELTA(dva) = INDEFD
	DVA_EXPT(dva) = 0.0d0

#    Get the exposure time.
 
	iferr(DVA_EXPT(dva) = imgetd(in, exposure))
		DVA_EXPT(dva) = clgetd(exposure)

	if(IS_INDEFD(DVA_EXPT(dva)))
		DVA_EXPT(dva) = 1.0d0

#    Get the spectrum, if null leave.

	DVA_SPEC(dva) = spc_gspc(in)
	if(DVA_SPEC(dva) == NULL) {
		call dva_close(dva)
		call error(1, "Unable to find spectrum.")
	}

#    Get the spectral error.
 
        DVA_ERR(dva) = spc_gerr(DVA_SPEC(dva), in)
	if(DVA_ERR(dva) == NULL) {
		call dva_close(dva)
		call error(1, "Unable to find spectral error.")
	}

#    Get the detector name.

	call smark(sp)
	call salloc(det, SZ_LINE, TY_CHAR)
	iferr(call imgstr(in, "DETNAM", Memc[det], SZ_LINE)) {
  	   iferr(call imgstr(in, "INSTRUME", Memc[det], SZ_LINE)) {
		call eu_clstr(det, "detector")
	   }
	}

#    Get the effective areas as pf type df's.

	call spc_gea(reference, Memc[det], DVA_EA(dva), 1, len)
	call sfree(sp)
	if(len < 1) call error(1, "No effective areas.")
	call df_pedge(DVA_EA(dva), 1.0d0, 1.0d0)

#    Get the bin width.

	DVA_DELTA(dva) = clgetd("dx")
	if(!IS_INDEFD(DVA_DELTA(dva))) {
		if(DVA_DELTA(dva) <= 0.0d0)
			call error(1, "divea: Bad delta value")
	} else {
		call df_gx0(DVA_SPEC(dva), xo)
		stat = df_dgetd(DVA_SPEC(dva), x, y, 1)
		call df_rewind(DVA_SPEC(dva))
		DVA_DELTA(dva) = x - xo
	}

	return(dva)

end


#
# dva_mkout()
#

procedure	dva_mkout(dva, in, in_name, out, out_name)

pointer	dva			#I
pointer	in			#I
char	in_name[ARB]		#I
pointer	out			#I
char	out_name[ARB]		#I

int	npts			#
int	ndim			#
double	xl, xh			#
double	s_xl, s_xh		#
double	e_xl, e_xh		#

int	imaccf()		#
pointer buf, impl1r()		#

begin

#    Add header cards to the image.
 
	ndim = 3
	IM_PIXTYPE(out) = TY_REAL
	IM_NDIM(out) = ndim

        call strcpy(IM_HISTORY(in), IM_HISTORY(out), SZ_IMHIST)
        call strcat("Created by divea from ", IM_HISTORY(out), SZ_IMHIST)
        call strcat(in_name, IM_HISTORY(out), SZ_IMHIST)
        call strcat("\n", IM_HISTORY(out), SZ_IMHIST)
 
        call imastr(out, "BANDID1", SRC_STR)
        call imastr(out, "BANDID2", ERR_STR)
        if(imaccf(out, "BANDID3") == YES)
		call imdelf(out, "BANDID3")

#    Make the MWCS for out image.

	call df_gxrng(DVA_SPEC(dva), s_xl, s_xh)
	call df_gxrng(DVA_EA(dva), e_xl, e_xh)
	if(e_xl > s_xl) xl = e_xl
	else		xl = s_xl
	if(e_xh < s_xh) xh = e_xh
	else		xh = s_xh

	npts = int((xl - s_xl)/DVA_DELTA(dva))
	xl = DVA_DELTA(dva)*npts + s_xl

	npts = int((xh - xl)/DVA_DELTA(dva))
	if(npts <= 0)
		call error(1, "divea: Current parameters yield no pixels.")

	IM_LEN(out, 1) = npts
	IM_LEN(out, 2) = 1
	IM_LEN(out, 3) = 2
	call spc_mkmwcs(out, xl, DVA_DELTA(dva), ndim)

#    Flush IMIO

	buf = impl1r(out)
	call aclrr(Memr[buf], IM_LEN(out, 1))
	call imflush(out)

end


#
# dva_close()
#

procedure	dva_close(dva)

pointer	dva			#

begin

	if(dva == NULL) return

	if(DVA_SPEC(dva) != NULL)
		call df_close(DVA_SPEC(dva))

	if(DVA_ERR(dva) != NULL)
		call df_close(DVA_ERR(dva))

	if(DVA_EA(dva) != NULL)
		call df_close(DVA_EA(dva))

	call dva_free(dva)
	dva = NULL

end

