# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_dscen.x
# 	Author	:  kelley
# 	Date	:  04/13/95
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_dscen.x -r1.17 04/13/95 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	01/04/96   tnk	added missed 'verbose' flag
# 	11/22/95   tnk	added verbose flag
# 	11/03/95   tnk	changed NR copyright to procedure call
# 	09/22/95   tnk	changed half-width comments to sigma; make convolve kernal variable size
# 	08/21/95   tnk	renamed sigfix, enabled arith trap
# 	08/18/95   tnk	added threshsigma; made KERNSIZE a constant
# 	08/17/95   tnk	unwound convolve loop, pushed decisions to the top
# 	08/17/95   tnk	added xcentbox, ycentbox handling
# 	08/11/95   tnk	added arith trap, other defensive code
# 	08/11/95   tnk	wcs bypass, scale failure handling
# 	08/10/95   tnk	moved cl parameter fetching to the top, changed ints to bools
# 	08/10/95   tnk	add numerical recipes copyright
# 	05/04/95   akeith	handles 3d images, band id and uses eugwcs lib to do the wcs mapping
# 	05/04/95   akeith	removed bufptr, now uses image buf directly
# 	04/27/95   bboyd	changed output format to have more decimal places 
# 	04/14/95   kelley	none
# 	04/13/95   kelley	date and time created 95/04/13 15:08:23 by kelley
# H-
# 
#

# Task to read in an image and find the centroid of the brightest
# object near the center.
#
# @(#)t_dscen.x	1.17 1/4/96 UCB CEA EGO
# bboyd April 1995
#
# Makes use of Numerical Recipes library routines,
# Copyright (C) 1986 Numerical Recipes Software


include <error.h>
include <mach.h>
include <gset.h>
include <imhdr.h>
include <imio.h>
include <xwhen.h>
include <config.h>

procedure t_dscen ()

# This reads an image into a buffer and passes the image copy to
# the centroid procedure getcent.  It prints the final centroids
# in WCS coordinates.

int 	nc, nr			# image dimensions
int	data_band		# data band of 3d image
int	err_band		# error band of 3d image
pointer	im			# input image
pointer	buf			# pointer to image values
pointer	conptr			# convolved image if necessary
char    filename[80]		# input image file name
real	xcent, ycent		# centroid in logical image coordinates
pointer	mw, mw_openim()		# mwcs object pointer
real	xworld, yworld		# centroid wcs coordinates
int	boxsize			# size of initial estimate box
int     xcentbox	        # box center x position
int     ycentbox		# box center y position
int	cbox 			# size of centroid box
real 	sigma			# estimate of gaussian one-sigma width
real	threshsig		# threshold in sigma above the mean
bool	bkgfit			# Fit background
bool    sigfix			# Fixed sigma
real	lowrej			# lower rejection limit for profile one-sigma width
real 	highrej			# upper rejection limit for profile one-sigma width
bool    preconv			# use convolution if needed
bool	plotfit			# flag for displaying plot of gaussian fit
bool	verbose			# show intermediate results
int 	jumpbuf[LEN_JUMPBUF]    # error handler stuff
bool	use_convolve		#
common  /jmpcom/ jumpbuf        #
extern  error_hand                #
pointer arith_hand, junk_hand   #
pointer locpr()			#
int     status

#Functions
pointer	immap()
pointer imgs2r()
pointer imgs3r()
int     clgeti()
real    clgetr()
bool    clgetb()

begin

	# put in Numerical Recipes copyright string
	call nr_copyr()

	call clgstr ("image", filename, 80)
	sigma = clgetr ("psfsigma")	# user-estimated one-sigma width of profile
	threshsig = clgetr ("threshsigma")
        boxsize = clgeti("boxsize")
	xcentbox = clgeti("xcentbox")
	ycentbox = clgeti("ycentbox")
	cbox = clgeti("cbox")
	bkgfit = clgetb("bgfit")	# user parameter
	sigfix = clgetb("fixpsfsigma")
	lowrej = clgetr("lowrej")
	highrej = clgetr("highrej")
	preconv = clgetb("preconvolve")
	plotfit = clgetb("plotfit")
	verbose = clgetb("verbose")

	# post error handler for arithmetic errors
	call xwhen (X_ARITH, locpr(error_hand), arith_hand)	

	im = immap (filename, READ_ONLY, 0)	# Allocate memory for the image

	if(IM_NDIM(im) < 2) {
	   call printf("dscen: Number of dimensions must be at least 2.\n")
	   call imunmap(im)
	   return
	} else if(IM_NDIM(im) > 3) {
	   call printf("dscen: Number of dimensions can be no greater then 3.\n")
	   call imunmap(im)
	   return
	}

	nc = IM_LEN(im,1)       # get image dimensions
        nr = IM_LEN(im,2)
 	mw = mw_openim(im)

	# adjust box parameters
	call box_adj(boxsize, xcentbox, ycentbox, nr, nc)

	call malloc (conptr, nc*nr, TY_REAL) 

	if(IM_NDIM(im) == 3) {
		call de_bands(im, data_band, err_band)
		if(data_band < 0) {
			call printf("dscen: Using data band 1.\n")
			data_band = 1
		}
		buf = imgs3r(im, 1, nc, 1, nr, data_band, data_band)
	} else	buf = imgs2r(im, 1, nc, 1, nr)	# get ptr to image values

	# error handler return point
	call zsvjmp (jumpbuf, status)
	if (status == OK)
	{
  	   # pass local image copy to centroid routine
	   use_convolve=false
	   call getcent (Memr[buf], Memr[conptr], nc, nr, xcent, ycent,
	      boxsize, xcentbox, ycentbox, cbox, sigma, bkgfit, sigfix, 
              lowrej, highrej, plotfit, use_convolve, threshsig, verbose)

  	   # if we failed to find source, and user asked us to try
           # a convolution then reattempt	

	   if ((IS_INDEFR(xcent) || IS_INDEFR(ycent)) && preconv)
	   {
	      use_convolve=true
 	      call getcent (Memr[buf], Memr[conptr], nc, nr, xcent, ycent,
	         boxsize, xcentbox, ycentbox, cbox, sigma, bkgfit, sigfix, 
                 lowrej, highrej, plotfit, use_convolve, threshsig, verbose)
	   }


	} 
	else
	{
	   xcent=INDEFR
	   ycent=INDEFR
	}

	if (IS_INDEFR(xcent) || IS_INDEFR(ycent))
	{
	   call	printf("INDEF INDEF\n")
	}
	else
	{

	   if (verbose)
	   {
              call printf("final xcent=%7.2f ycent=%7.2f\n")
              call pargr(xcent)
              call pargr(ycent)
	   }

	   # Convert centroids to WCS world cooordinates
	   call lw_c2tranr(mw, "world", xcent, ycent, xworld, yworld)
           # call printf("RA and DEC:\n")
	   call printf("%9.5f  %9.5f\n")
	   call pargr(xworld)
	   call pargr(yworld)
        }	
   	call mw_close(mw)
	call imunmap (im)		# deallocate all memory
	call mfree (conptr, TY_REAL)

	#restore old error handler
	call xwhen (X_ARITH, arith_hand, junk_hand)

end



# This computes a centroid, assuming the target is the brightest 
# extended source within the coarse centroid box, centered on the
# middle of the image.  The centroid is computed roughly by a
# histogram maximum, and then by a gaussian fit.

procedure getcent (buffer, conbuf, ncols, nrows, xc, yc, boxsize, 
   xcentbox, ycentbox, cbox, sigma, bkgfit, sigfix, lowrej, 
   highrej, plotfit, use_convolve, threshsig, verbose)

real    buffer[ncols,nrows]	# image array
real 	conbuf[ncols,nrows]	# convolved image
int     ncols, nrows            #I image dimensions
real	xc, yc			#O final centroids
int	boxsize			# size of initial estimate box
int     xcentbox	        #I box center x position
int     ycentbox		#I box center y position
int	cbox 			# size of centroid box
real 	sigma			# estimate of gaussian one-sigma width
bool	bkgfit			#
bool    sigfix			#
real	lowrej		# lower rejection limit for profile one-sigma width
real 	highrej		# upper rejection limit for profile one-sigma width
bool	plotfit		# flag for displaying plot of gaussian fit
bool    use_convolve    #I pre-convolve image before fitting
real	threshsig	# threshold in sigma above the mean
bool	verbose		# show intermediate results

define  KERNSIZE 17
real	kern[KERNSIZE,KERNSIZE]		# gaussian kernel for convolution
pointer	tvec, tvec2			# temp vectors for convolution
pointer	xhist, yhist			# ptrs to x and y histogram vectors
real	maxx, maxy			# max values of histograms
real	dist
real    kerncent
int	i, j, k 			# loop indices
int	xoffset, yoffset, xsize, ysize	# subimage parameters
int	ix, iy				# indices of maxx, maxy
int	xcbox, xcrad, ycbox, ycrad	# sizes of x,y boxes (generally = cbox)
int 	tsiz				# temp vector dimension
bool    success
pointer	sp				# stack pointer

begin

    call smark(sp)

    call salloc (xhist, ncols, TY_REAL)
    call salloc (yhist, nrows, TY_REAL)

    xc = INDEF
    yc = INDEF

    if (cbox > boxsize) 
	cbox = boxsize

    xsize=boxsize
    xoffset=xcentbox - (xsize/2)

    ysize=boxsize
    yoffset=ycentbox - (ysize/2)

    call aclrr(Memr[xhist],ncols)         # initialize to zeros
    call aclrr(Memr[yhist],nrows)

    if (use_convolve) {
       # Make a KERNSIZExKERNSIZE gaussian kernel 
       kerncent=real(KERNSIZE+1)/2.0
       do i=1,KERNSIZE {
    	do j=1,KERNSIZE {
    	   dist = sqrt((i-kerncent)**2 + (j-kerncent)**2)
    	   kern[i,j]=exp(-0.5*(dist/sigma)**2)
    	}
       }

       tsiz = xsize + KERNSIZE - 1
       call salloc(tvec,tsiz,TY_REAL)
       call salloc(tvec2,xsize,TY_REAL)

       # Convolve the image with the kernel
          call aclrr(Memr[tvec],tsiz)
       k = yoffset - KERNSIZE/2         # start at the (bottom - kernsize/2) line
       do j=yoffset, yoffset+ysize {
          call aclrr(Memr[tvec2],xsize)
          do i=1,KERNSIZE {
            if ((k+i-1 < yoffset) || (k+i-1 > yoffset+ysize))
               call aclrr(Memr[tvec],tsiz)   # zeroes if outside image
            else
               call amovr(buffer[xoffset,k+i-1],Memr[tvec+8],xsize)
            call acnvr(Memr[tvec],Memr[tvec2],xsize,kern[1,i],KERNSIZE)
          }
          call amovr(Memr[tvec2],conbuf[xoffset,j],xsize)
          k = k+1
       }

       # Make the x and y histograms (the hard way)
       do i=xoffset, xoffset+xsize {
    	do j=yoffset, yoffset+ysize {
    	   Memr[yhist+j-1] = Memr[yhist+j-1] + conbuf[i,j]
    	   Memr[xhist+i-1] = Memr[xhist+i-1] + conbuf[i,j]
    	}
       }
    }
    else {
       # Make the x and y histograms (the hard way)
       do i=xoffset, xoffset+xsize {
            do j=yoffset, yoffset+ysize {
               Memr[yhist+j-1] = Memr[yhist+j-1] + buffer[i,j]
               Memr[xhist+i-1] = Memr[xhist+i-1] + buffer[i,j]
            }
       }
    } 

    # compute where max of each histogram lies
    # (same as code above)
    call ahivri(Memr[xhist], ncols, maxx, ix)
    call ahivri(Memr[yhist], nrows, maxy, iy)

    if (verbose)
    {
       call printf("raw ix=%d iy=%d\n")
       call pargi(ix)
       call pargi(iy)
    }


    # Make new histograms around the centroid estimate.
    xcbox = cbox
    ycbox = cbox
    xcrad = cbox/2
    ycrad = cbox/2
    
    call aclrr(Memr[xhist],ncols)         # initialize to zeros
    call aclrr(Memr[yhist],nrows)

    # If the box falls outside the image, shrink the box.
    if (ix-xcrad < 1) {
       xcrad = ix-1
       xcbox = xcrad*2
    }
    if (ix+xcrad > ncols) {
       xcrad = ncols-ix
       xcbox = xcrad*2
    }
    if (iy-ycrad < 1) {
       ycrad = iy-1
       ycbox = ycrad*2
    }
    if (iy+ycrad > nrows) {
       ycrad = nrows-iy
       ycbox = ycrad*2
    }

    if (use_convolve) {
       do i=ix-xcrad, ix+xcrad+1 {
            do j=iy-ycrad, iy+ycrad+1 {
               Memr[yhist+j-1] = Memr[yhist+j-1] + conbuf[i,j]
               Memr[xhist+i-1] = Memr[xhist+i-1] + conbuf[i,j]
            }
       }
    }
    else {
       do i=ix-xcrad, ix+xcrad+1 {
            do j=iy-ycrad, iy+ycrad+1 {
               Memr[yhist+j-1] = Memr[yhist+j-1] + buffer[i,j]
               Memr[xhist+i-1] = Memr[xhist+i-1] + buffer[i,j]
            }
       }
    }

    maxx = Memr[xhist+ix-1]		# Adjust height guess to new histograms
    maxy = Memr[yhist+iy-1]


    # Call gaussian fit routine for xhist 
    call ggfit(Memr[xhist+ix-xcrad],ix-xcrad, xcrad, maxx, xcbox+1, 
    	    xc, sigma, bkgfit, sigfix, lowrej, highrej,
            plotfit, threshsig, success)
    if (!success)
       xc=INDEFR

    # Call gaussian fit routine for yhist 
    call ggfit(Memr[yhist+iy-ycrad],iy-ycrad, ycrad, maxy, ycbox+1, 
    	    yc, sigma, bkgfit, sigfix, lowrej, highrej,
            plotfit, threshsig, success)
    if (!success)
       yc=INDEFR


    call sfree(sp)
end

# Set up parameters for a gaussian fit and call dofit from the
# deblend routine in noao.onedspec.
procedure ggfit (xcts, xoff, xest, yest, npts, cent,
   sigma, bkgfit, sigfix, lowrej, highrej, plotfit, threshsig, success)
real    xcts[npts]      # I: data values 
int     xoff            # I: offset for sub-vector of interest
int     xest            # I: estimate of gaussian centroid
real    yest            # I: estimate of height of gaussian
int     npts            # I: length of the sub-vector of interest
real	cent		# O: centroid from fit (output)
real 	sigma		# I: estimate of gaussian one-sigma width
bool	bkgfit		# I: 
bool    sigfix		# I:
real	lowrej		# I: lower rejection limit for profile one-sigma width
real 	highrej		# I: upper rejection limit for profile one-sigma width
bool	plotfit		# I: flag for displaying plot of gaussian fit
real	threshsig	# threshold in sigma above the mean
bool    success         # O: true if center is found

real	ycon		# background estimate for gaussian fit
real    savg            # average poisson error in background
pointer xpts            # index of abscissa values
pointer	gp		# GIO file descriptor
int     posfit, intfit, sigfit, nsub    # fit control params
int     nlines, i, navg
pointer x, y, s, xg, yg, sg	# local fit variables (x, y, error)
pointer temp
pointer	yfit, xadj		# variables for optional plot
real    scale, sscale, junk, slope, chisq, dw	# local fit variables
int	npix
real	asumr()
pointer	gopen(), sp
bool    fp_equalr()
real    assqr()

begin
	call smark(sp)

        # Set up fit estimates and parameters.

        posfit = 2	# position fit (1=fixed, 2=single, 3=all)
        intfit = 2	# intensity fit (1=fixed, 2=single, 3=all)
	if (sigfix)	# If user chose fixed sigma, set it
	   sigfit = 1
	else
           sigfit = 2	# sigma fit (1=fixed, 2=single, 3=all)
        nsub = 1
	dw = 1.0
	success = true
	cent = INDEF

	# make an index (abscissa) vector
        call salloc (xpts, npts, TY_REAL)
        do i=1,npts
                Memr[xpts+i-1] = i * 1.0

        # Allocate memory for fit params
        call malloc (x, npts, TY_REAL)	
        call malloc (y, npts, TY_REAL)
        call malloc (s, npts, TY_REAL)	
        call malloc (temp, npts, TY_REAL)	


        call amovr (Memr[xpts],Memr[x],npts)

	# Scale the data.
	scale=0.0
        do i = 1, npts {
            Memr[y+i-1] = xcts[i]
            scale = max (scale, abs (Memr[y+i-1]))
	    Memr[s+i-1] = sqrt (xcts[i])	# poisson errors
	    if (fp_equalr(xcts[i],0.)) {
                Memr[y+i-1] = 0.0001	# replace zeroes with 1e-4
                Memr[s+i-1] = sqrt(Memr[y+i-1])
            }
        }
	sscale = asumr (Memr[s], npts) / npts

	# get out if can't scale data
	if (scale <= 0.0 || sscale <= 0.0)
	{
	   cent=INDEF
	   call mfree (x, TY_REAL)
	   call mfree (y, TY_REAL)
	   call mfree (s, TY_REAL)
	   call mfree (temp, TY_REAL)
	   success = false
           return
        }

        call adivkr (Memr[y], scale, Memr[y], npts)
        call adivkr (Memr[s], sscale, Memr[s], npts)
	

        # Allocate space for fit params
        nlines = 1
        call malloc (xg, nlines, TY_REAL)
        call malloc (yg, nlines, TY_REAL)
        call malloc (sg, nlines, TY_REAL)

	Memr[xg] = xest * 1.0
        Memr[yg] = yest/scale
        Memr[sg] = sigma      


        # Estimate (flat) bkgd level - avg of first navg points
	if (npts < 40) 
		navg = 3
	else if ((npts < 100) && (npts >= 40))
		navg = 5
	else
		navg = 9
	call aavgr(Memr[y], navg, ycon, junk)  
        slope = 0.0

        call amulkr(Memr[s], sscale/scale, Memr[temp], navg)
	savg = sqrt(assqr(Memr[temp], navg))/navg

	# If npix peak values are less than threshsig sigma above bkgd.,
	# we failed to find the source
	npix = 3
	if (Memr[y+xest-2]+Memr[y+xest-1]+Memr[y+xest] < npix*ycon+npix*threshsig*savg) 
	   success = false
	else
	{
 	   # we have a source, now find the center 
	   # Use gaussian fitting routine dofit from noao.onedspec.splot 
           call dofit (bkgfit, posfit, intfit, sigfit, Memr[x],
              Memr[y], Memr[s], npts, dw, nsub, ycon, slope,
              Memr[xg], Memr[yg], Memr[sg], 1, chisq)

	   cent = Memr[xg]+xoff

	   # Throw out centroid if peak width is outside rejection limits.
	   if ((Memr[sg] < lowrej) || (Memr[sg] > highrej)) 
		success = false

	   if (plotfit) {
	      # generate a vector of the fit
	      call salloc (yfit, npts, TY_REAL)
	      call salloc (xadj, npts, TY_REAL)	
	      do i=1, npts {
		 Memr[yfit+i-1]=ycon+Memr[yg]*exp(-0.5*((Memr[x+i-1]-Memr[xg])/Memr[sg])**2)
		 Memr[xadj+i-1] = Memr[x+i-1] + xoff
	      }
	      # print out fit parameters
	      call printf ("Centroid = %10.5f,  sigma = %9.5f, chisqr = %9.5f\n")
	      call pargr (Memr[xg]+xoff)
	      call pargr (Memr[sg])
	      call pargr (chisq)

	      # show a plot of the data and the fit
	      call printf ("Hit <return> to continue...\n")
	      gp = gopen ("stdgraph", NEW_FILE, STDGRAPH)
	      call gascale (gp, Memr[xadj], npts, 1)
	      call gascale (gp, Memr[y], npts, 2)
	      call glabax (gp, "Profile vs. fit", "pixel", "scaled counts")
              #	   call gtext (gp, Memr[xadj], Memr[yg]+ycon-0.05, Memr[sg],"sigma = %6.8f")
	      call gpline (gp, Memr[xadj], Memr[y], npts)
	      call gpline (gp, Memr[xadj], Memr[yfit], npts)
	      call gclose (gp)
	      call scan()	# delay for a minute
	   }
	}

	call mfree (x, TY_REAL)
	call mfree (y, TY_REAL)
	call mfree (s, TY_REAL)
	call mfree (temp, TY_REAL)
        call mfree (xg, TY_REAL)
        call mfree (yg, TY_REAL)
        call mfree (sg, TY_REAL)
	
	call sfree(sp)

end

# --- Error handler ---
procedure error_hand (code, next_hand)
int code 	#I Error code that triggered this handler
int next_hand	#O Handler called after this handler exits

int 	jumpbuf[LEN_JUMPBUF]    # error handler stuff
common  /jmpcom/ jumpbuf        #

begin

   # Resume execution at zsvjmp
   next_hand = X_IGNORE
   call zdojmp(jumpbuf, code)

end

# adjust box size and position to fit within image size
procedure box_adj(boxsize, xcentbox, ycentbox, nrows, ncols)
int  boxsize	#I/O size of initial estimate box
int  xcentbox	#I/O box center x position
int  ycentbox	#I/O box center y position
int  nrows	#I num rows in image
int  ncols	#I num cols in image

int  xoffset
int  yoffset
int  xsize
int  ysize

begin

    if ((boxsize > ncols) || (boxsize > nrows)) {
	   if (ncols < nrows)
	  	boxsize = ncols - 2
	   else 
	  	boxsize = nrows - 2
    } 

    # find valid box center position 
    if (IS_INDEFI(xcentbox))
	xcentbox = ncols/2
    if (IS_INDEFI(ycentbox))
	ycentbox = nrows/2

    xsize=boxsize
    xoffset=xcentbox - (xsize/2)
    if (xoffset < 0 || (xoffset+xsize > ncols))
    {
	call error(123,"Can't put box of that size at given xcentbox position\n")
    }

    ysize=boxsize
    yoffset=ycentbox - (ysize/2)
    if (yoffset < 0 || (yoffset+ysize > nrows))
    {
	call error(123,"Can't put box of that size at given ycentbox position\n")
    }    

end
