# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  eqtoec.x
# 	Author	:  akeith
# 	Date	:  05/15/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) eqtoec.x -r1.3 05/15/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	04/12/94   tnk	took out diagnostics
# 	04/12/94   tnk	replaced cosd, sind, atand calls with radian equivalents
# 	05/15/92   akeith	date and time created 92/05/15 10:55:54 by akeith
# H-
# 
#

# EQTOEC -- Convert equatorial coordinates to ecliptic coordinates. -- revised 4/12/94

include <mach.h>
include <math.h>

procedure eqtoec (ra, dec, year, ec_long, ec_lat)

double ra		# Right ascension (hours)
double dec		# Declination (degrees)
double year		# Epoch of input coordinates
double ec_long	# Ecliptic longitude (degrees)
double ec_lat	# Ecliptic latitude (degrees)

double rar, x, y, z, r, phi, obec

begin
  obec = 23.439291D0 - 0.013004D0 * (year - 2000.D0)/100.D0

  rar = ra * 15.D0
  x = cos(DEGTORAD(dec)) * cos(DEGTORAD(rar))
  y = cos(DEGTORAD(obec)) * cos(DEGTORAD(dec)) * sin(DEGTORAD(rar)) + 
      sin(DEGTORAD(obec)) * sin(DEGTORAD(dec))
  z = cos(DEGTORAD(obec)) * sin(DEGTORAD(dec)) - sin(DEGTORAD(obec)) * 
      cos(DEGTORAD(dec)) * sin(DEGTORAD(rar))

  r = sqrt(x*x + y*y)
  if (r < EPSILOND) {
    if (abs(z) < EPSILOND) {
      ec_lat = 0.D0
	}
    else if (z > 0.D0) {
	  ec_lat = 90.D0
	}
    else {
	  ec_lat = -90.D0
	}
  }
  else {
	ec_lat = RADTODEG(atan(z/r))
  }

  if (abs(x) < EPSILOND && abs(y) < EPSILOND) {
	ec_long = 0.D0
  }
  phi = 2.D0 * RADTODEG(atan(y/(abs(x) + r)))
  if (x < 0.D0) {
	ec_long = 180.D0 - phi
  }
  else if (y < 0.D0) {
	ec_long = 360.D0 + phi
  }
  else {
	ec_long = phi
  }
end
