# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_ecliptic.x
# 	Author	:  akeith
# 	Date	:  05/15/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_ecliptic.x -r1.1 05/15/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	05/15/92   akeith	date and time created 92/05/15 10:55:56 by akeith
# H-
# 
#

# ECLIPTIC -- convert between equatorial and ecliptic coordinates. -- revised 5/15/92

include	<fset.h>

# ECLIPTIC -- convert between equatorial and ecliptic coordinates.

procedure t_ecliptic ()

int		filelist, prt_coords, in_coords
double	year
char	fname[SZ_FNAME]

int		clpopni(), clgfil(), btoi(), clgwrd()
double	clgetd()
bool	streq(), clgetb()

begin
  # Input can come from the standard input, a file, or a list of files.
  # The following procedure makes both cases look like a list of files.

  filelist = clpopni ("input")

  # Get output option
	
  in_coords = clgwrd ("in_coords", fname, SZ_FNAME,  "|equatorial|ecliptic|")
  prt_coords = btoi (clgetb ("print_coords"))
  year = clgetd("epoch")

  # Process each coordinate list.  If reading from the standard input,
  # set up the standard output to flush after every output line, so that
  # converted coords come back immediately when working interactively.

  while (clgfil (filelist, fname, SZ_FNAME) != EOF) {
	if (streq (fname, "STDIN")) {
	  call fseti (STDOUT, F_FLUSHNL, YES)
	}
	else {
	  call fseti (STDOUT, F_FLUSHNL, NO)
	}
	switch (in_coords) {
	case 1:
	  call in_equatorial (STDOUT, fname, prt_coords, year)
	case 2:
	  call in_ecliptic (STDOUT, fname, prt_coords, year)
	}
  }

  call clpcls (filelist)
end


# IN_EQUATORIAL -- convert a list of equatorial coordinates read from the
# named file to ecliptic coordinates, writing the results on the output file.  

procedure in_equatorial (out, listfile, prt_coords, defyear)

int		out					# output stream
char	listfile[SZ_FNAME]	# input file
int		prt_coords			# print coordinates in output file?
double	defyear				# default epoch of input coordinates

int		in
double	ra, dec, ec_long, ec_lat, year

int	fscan(), nscan(), open()
errchk	open, fscan, printf

begin
  in = open (listfile, READ_ONLY, TEXT_FILE)

  # Read successive RA,DEC coordinate pairs from the standard input,
  # converting and printing the result on the standard output.

  while (fscan (in) != EOF) {
	call gargd (ra)
	call gargd (dec)
	call gargd (year)

	# If there is something wrong with the input coords, print warning
	# and skip the conversion.  

	switch (nscan()) {
	case 2:
	  year = defyear
	case 3:
	default:
	  call eprintf ("Bad entry in coordinate list\n")
	  next
	}

	# Call routine to perform conversion and write ec_long, ec_lat to
	# the standard output.

	call eqtoec (ra, dec, year, ec_long, ec_lat)

	if (prt_coords == YES ) {
	  call fprintf (out, "%13.2h %12.1h")
	  call pargd (ra)
	  call pargd (dec)
	}
		
	call fprintf (out, "%13.4f %9.4f %7.2f" )
	call pargd (ec_long)
	call pargd (ec_lat)
	call pargd (year)

	call fprintf (out, "\n")
  }

  call close (in)
end


# IN_ECLIPTIC -- convert a list of ecliptic coordinates read from the
# named file to equatorial coordinates, writing the results on the output
# file.  

procedure in_ecliptic (out, listfile, prt_coords, defyear)

int		out					# output stream
char	listfile[SZ_FNAME]	# input file
int		prt_coords			# print coordinates in output file?
double	defyear				# default epoch of input coordinates

int		in
double 	ra, dec, ec_long, ec_lat, year

int		fscan(), nscan(), open()
errchk	open, fscan, printf

begin
  in = open (listfile, READ_ONLY, TEXT_FILE)

  # Read successive ecliptic longitude,latitude coordinate pairs from 
  # the standard input, converting and printing the result on the 
  # standard output.

  while (fscan (in) != EOF) {
	call gargd (ec_long)
	call gargd (ec_lat)
	call gargd (year)

	# If there is something wrong with the input coords, print warning
	# and skip the conversion.  

	switch (nscan()) {
	case 2:
	  year = defyear
	case 3:
	default:
	  call eprintf ("Bad entry in coordinate list\n")
	  next
	}

	# Call routine to perform conversion and write RA, Dec to
	# the standard output.

	call ectoeq (ec_long, ec_lat, year, ra, dec)

	if (prt_coords == YES ) {
	  call fprintf (out, "%13.4f %9.4f" )
	  call pargd (ec_long)
	  call pargd (ec_lat)
	}
		
	call fprintf (out, "%13.2h %12.1h %7.2f")
	call pargd (ra)
	call pargd (dec)
	call pargd (year)

	call fprintf (out, "\n")
  }

  call close (in)
end
