# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  qbound.x
# 	Author	:  tnk
# 	Date	:  08/15/95
# 	Synopsis:  
# 	Keywords:
# 		@(#) qbound.x -r1.1 08/15/95 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	08/15/95   tnk	date and time created 95/08/15 14:43:49 by tnk
# H-
# 
#

# @(#)qbound.x	1.1 8/15/95
# reads primbsch quadrant boundaries from detector ref table
#

include <tbset.h>
include "detmp.h"

bool procedure ref_qbounds(tp,dptr,n_det)

pointer tp               #I   table pointer
pointer dptr[ARB]        #I/O detector struct
int n_det                #I   len of dtpr array

pointer detnum_cp        # detector number column pointer
pointer x_y1_cp          # x/y1 boundary column pointer
pointer y_y2_cp          # y/y2 boundary column pointer
pointer y3_cp            # y3 boundary column pointer
pointer type_cp          # type of boundary column ptr
int     nrows            # number of rows
int     row              # current row no.
int     tab_det          # detector number from table
bool    found            # flag
int     i                # index

int     tbpsta()

begin

  n_rows = tbpsta (tp, TBL_NROWS)

  # get column pointers
  call tbcfnd (tp, TIFNUM_COLNAME,     detnum_cp, 1)
  call tbcfnd (tp, PRIMB_TYPE_COLNAME, type_cp,   1)
  call tbcfnd (tp, PRIMB_X_Y1_COLNAME, x_y1_cp,   1)
  call tbcfnd (tp, PRIMB_Y_Y2_COLNAME, y_y2_cp,   1)
  call tbcfnd (tp, PRIMB_Y3_COLNAME,   y3_cp,     1)

  if (detnum_cp == NULL || 
      type_cp == NULL   || x_y1_cp == NULL ||
      y_y2_cp == NULL   || y3_cp == NULL ) {
    call error(650,"Could not find ref columns for primbsch quadrants")
  }

  for (i=1; i<=n_det; i=i+1) {

     # go down each row in table, see if it matches a detector in our
     # list
     found = false

     for (row=1; row <=n_rows; row=row+1) {
        # get detector number of each row
        call tbegti(tp, detnum_cp, row, tab_det)

        if (tab_det == MP_NUM(d_ptr[i])) {

           found = true
           # get the quadrant type & boundaries
           call tbegti(tp, type_cp,  row,   MP_QBOUND_TYPE(dptr[i]))
           call tbegti(tp, x_y1_cp,  row,   MP_QBOUND_X_Y1(dptr[i]))
           call tbegti(tp, y_y2_cp,  row,   MP_QBOUND_Y_Y2(dptr[i]))
           call tbegti(tp, y3_cp,    row,   MP_QBOUND_Y3(dptr[i]))
           break
        }
     }

     if (! found) {
        call eprintf("Error:  could not find primb quad bounds for det %d\n")
        call pargi(MP_NUM(dptr[i]))
        break
     }

  }

return(found)

end



#
# Find quadrant of an event (labeled 0-3)
#
int procedure pb_quadrant (det_ptr, x, y)

pointer  det_ptr   #I det struct
real       x       #I detector x-value
real       y       #I detector y-value

# LOCAL VARS
int      q         # quad num

#   TYPE VERT           TYPE SQUARE
#             3
#         Y3 -----        1  |  3
#             2              |
#         Y2 -----       ---------- Y
#             1              |
#         Y1 -----        0  |  2
#             0
#                            X

begin

  if (MP_QBOUND_TYPE(det_ptr) == PRIMB_VERT_TYPE) {

     # detector "quads" are really a vertical stack (Y1,Y2,Y3)
     if      ( y >= MP_QBOUND_Y3(det_ptr) )   q = 3
     else if ( y >= MP_QBOUND_Y_Y2(det_ptr))  q = 2
     else if ( y >= MP_QBOUND_X_Y1(det_ptr))  q = 1
     else q = 0
  }
  else {
     # detector is really quadrants (X,Y)
     if ( x >= MP_QBOUND_X_Y1(det_ptr)) {
          if (y >= MP_QBOUND_Y_Y2(det_ptr) ) q = 3
            else  q = 2
     }
     else  {
          if (y >= MP_QBOUND_Y_Y2(det_ptr) ) q = 1
            else q = 0
     }
  }
  return(q)

end
