# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_effexp.x
# 	Author	:  tnk
# 	Date	:  08/15/95
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_effexp.x -r1.9 08/15/95 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	03/04/97   tnk	added error check: no detpos table w/ spectrometer
# 	02/26/97   tnk	reworked primbsch table loading to handle archive format
# 	02/24/97   tnk	added search for DETNAM header
# 	08/30/96   tnk	xs,xe error check (for missing qpoe filter)
# 	11/03/95   tnk	changed NR copyright to procedure call
# 	10/10/95   mabbott	added warning about non-empty mask parameter with QPOE input
# 	08/16/95   tnk	trims white space; more descriptive table open failure msgs
# 	08/15/95   tnk	renamed rlintegrate to avoid collision
# 	08/15/95   tnk	date and time created 95/08/15 14:43:50 by tnk
# H-
# 
#

#  @(#)t_effexp.x	1.9 3/4/97
#  effexp - task to compute effective exposure 
#
#
#    other sources:  rlintegrate.x
#                    hunt.f
#                    polint.f
#                    qbound.x      
#
# Makes use of Numerical Recipes library routines,
# Copyright (C) 1986 Numerical Recipes Software
#

include	<qpioset.h>
include <tbset.h>
include <fset.h>
include <imhdr.h>
include <ctype.h>
include "detmp.h"

define DISTORTION_COL "undistort_map"

procedure t_effexp()

int	np			# number of primbsch table rows
int	nd			# number of detpos table rows
int	nlt			# number of light curve bins
int	xlen, nranges		# number of attribute filter ranges
int	detnum			# detector TIF number
int	quad2use		# quadrant to use instead of detpos position
int	vnx, vny		# size of vignetting image
int	dnx, dny		# size of distortion image
int	i, n
real	binsize			# length of light curve bins
real	rate			# corrected count rate
real	rerr			# error in corrected count rate
real	net, neterr
double	sublen			# subinterval length
double	tmin, tmax		# time to integrate over
double	exposure, badexp, rawexp
bool	dead, prim, usevig, undistort
bool	null
bool	verbose
pointer	qp			# QPOE descriptor
pointer	qpio			# QPIO descriptor - event list for this QPOE
pointer	pm			# PLIO descriptor - mask for this event list
pointer	pim			# PMIO descriptor - mask opened as an image
pointer	ex			# QPEX descriptor - attribute filter expression
pointer	xs, xe			# attribute range start and end arrays
pointer	ptime, pcorr		# primbsch times and corrections
pointer	dtime, dx, dy		# detector position times and data
pointer	vim			# IMIO pointer to vignetting image
pointer	vig			# vignetting data buffer
pointer	vmw			# MWCS descriptor for vignetting image
pointer	vct			# MWCS compiled transformation descriptor for vmw
pointer	dim			# IMIO pointer to distortion image
pointer	dist			# distortion data buffer
pointer	dmw			# MWCS descriptor for distortion image
pointer	dct			# MWCS compiled transformation descriptor for dmw
pointer	tp			# ST table descriptor
pointer	nulls			# for array of bools when reading table columns
pointer	ref_tp			# detector.tab table pointer
pointer	detptr			# detector primbsch boundary structure pointer
pointer	det			# detector name from CL parameter
pointer	inst			# detector name from QPOE header parameter
pointer	refpath			# path to directory containing detector.tab
pointer	cp			# table column pointer
pointer	op[7]			# output column pointers
pointer	buf 			# IMIO buffer
pointer	sp, str, str2

int		tbpsta(), qpex_attrld(), fnldir(), qpgstr(), qp_access(), clgeti()
double	clgetd()
bool	clgetb(), streq(), ref_qbounds(), getvigname(), getdistname()
pointer	tbtopn(), qpio_stati(), immap(), mw_openim(), mw_sctran(), imgs2r(),
        im_pmmapo(), im_pmmap(), imgs3r()

begin
  call smark(sp)

  # put in Numerical Recipes copyright string
  call nr_copyr()

  call salloc(str, SZ_LINE, TY_CHAR)
  call salloc(str2, SZ_LINE, TY_CHAR)
  call salloc(det, SZ_LINE, TY_CHAR)
  call salloc(refpath, SZ_LINE, TY_CHAR)
  call salloc(inst, SZ_LINE, TY_CHAR)

  verbose = clgetb("verbose")
  if (verbose) call fseti(STDOUT, F_FLUSHNL, YES)

#####################################################################
# okay - here we go.  Start opening and reading input and calibration
# data files.  What a mess!
#####################################################################

  # open QPOE or table and get range list of times and optionally a mask
  call clgstr("intervals", Memc[str], SZ_LINE)
  call stripwhite(str)
  call qp_parse(Memc[str], Memc[str2], SZ_LINE, Memc[det], SZ_LINE)
  qp = NULL
  if (qp_access(Memc[str2], READ_ONLY) == YES) {
	# Yes, we have a QPOE.  Open up its QPEX structure.
	call qp_open_filtered(qp, qpio, Memc[str])
    ex = qpio_stati(qpio, QPIO_EX)

    call strcpy("time", Memc[str], SZ_LINE)
    xs = NULL
	xe = NULL
	xlen = 0
	nranges = 0
	if (ex != NULL) {
	  # get time filter from QPEX
	  nranges = qpex_attrld(ex, Memc[str], xs, xe, xlen)

	  iferr (n = qpgstr(qp, "DETNAM", Memc[inst], SZ_LINE)) {
	     iferr (n = qpgstr(qp, "INSTRUME", Memc[inst], SZ_LINE)) {
           	  call strcpy("", Memc[inst], SZ_LINE)
	     }
	  }

 	  call strlwr(Memc[inst])
	  call stripwhite(inst)

	  # get any region mask which may be active on the QPOE
	  pm = qpio_stati(qpio, QPIO_PL)
	  if (pm != NULL) {
		pim = im_pmmapo(pm, 0)
	  }
	  else pim = NULL

	  # warn of possible user error
	  call clgstr("mask", Memc[str], SZ_LINE)
	  call stripwhite(str)
	  if (!streq(Memc[str],"")) {
		call eprintf("Warning: mask parameter ignored.  When specifying QPOE\ninput, desired masks must be given in a QPOE expression.")
	  }	
	}
  }
  else {
	# not a QPOE, had better be a table.  open it up and read 
	# time intervals
	iferr (	tp = tbtopn(Memc[str], READ_ONLY, 0)){
	    call error(122,"Can't open intervals table")
	}
	nranges = tbpsta(tp, TBL_NROWS)
	call malloc(nulls, nranges, TY_BOOL)
	call malloc(xs, nranges, TY_DOUBLE)
	call malloc(xe, nranges, TY_DOUBLE)

	call clgstr("stime", Memc[str], SZ_LINE)
	call tbcfnd(tp, Memc[str], cp, 1)
	if (cp == NULL) call error(123,"start time column not found")
	call tbcgtd(tp, cp, Memd[xs], Memb[nulls], 1, nranges)

	call clgstr("etime", Memc[str], SZ_LINE)
	call tbcfnd(tp, Memc[str], cp, 1)
	if (cp == NULL) call error(123,"end column not found")
	call tbcgtd(tp, cp, Memd[xe], Memb[nulls], 1, nranges)

	call mfree(nulls, TY_BOOL)
	call tbtclo(tp)

	# open specified mask file, if any
	call clgstr("mask", Memc[str], SZ_LINE)
	call stripwhite(str)
	if (streq(Memc[str],"")) {
	  pim = NULL
	}
	else {
	  pim = im_pmmap(Memc[str], READ_ONLY, 0)
	}

    ex = 1
  }

  # Read detector reference data and correction values
  if (ex != NULL) {
    # open detector.tab
    call clgstr("refdata", Memc[str], SZ_LINE)
    iferr (ref_tp = tbtopn(Memc[str], READ_ONLY, 0)){
       call error(122,"Can't open refdata table")
    }

	n = fnldir(Memc[str], Memc[refpath], SZ_LINE)

	# get detector name and number.  use header value if needed
	call clgstr("detector", Memc[det], SZ_LINE)
        call stripwhite(det)
	if (streq(Memc[det],"")) {
	  if (streq(Memc[inst],"")) {
	    call error(765,"no detector name given in QPOE or parameter")
	  }
	  else {	
		call strcpy(Memc[inst], Memc[det], SZ_LINE)
	  }	
	}
	else if (qp != NULL) {
	  if (!streq(Memc[det], Memc[inst])) {
		call printf("Warning: parameter detector [%.3s] does not match QPOE header [%.3s]\n")
		call pargstr(Memc[det])
		call pargstr(Memc[inst])
	  }
	}
    call eu_detect_id(Memc[det], detnum)

	# fetch primbsch info
	call clgstr("primbsch", Memc[str], SZ_LINE)
	call stripwhite(str)
	dead = clgetb("usedead")
	prim = clgetb("useprim")
	np = 0
	if (!streq(Memc[str], "") && (dead || prim)) {

	  # Set up quadrant boundaries structure
	  call salloc(detptr, MP_LEN, MP_TYP)
	  MP_NUM(detptr) = detnum
	  if ( ! ref_qbounds(ref_tp, detptr, 1))
	     call error(456,"Unable to read quadrant boundaries")

	  # Read primbsch corrections
	  tp = tbtopn(Memc[str], READ_ONLY, 0)
	  np = tbpsta(tp, TBL_NROWS)
	  call malloc(nulls, np, TY_BOOL)
	  call malloc(ptime, np, TY_DOUBLE)
	  call malloc(pcorr, np*4, TY_REAL)

	  call clgstr("ptimecol", Memc[str], SZ_LINE)
	  call tbcfnd(tp, Memc[str], cp, 1)
	  if (cp == NULL) call error(123,"primbsch time column not found")
	  call tbcgtd(tp, cp, Memd[ptime], Memb[nulls], 1, np)


	  # get primbsch correction values
	  call get_primb_corr(tp, pcorr, nulls, np, prim, dead, Memc[det],
             detnum)

	  call mfree(nulls, TY_BOOL)
	  call tbtclo(tp)
	}

	# fetch fixed quadrant, if any
	quad2use = clgeti("quadrant")
	if (IS_INDEFI(quad2use)) {
	  quad2use = -1
	} else {
	  if (quad2use < 0 || quad2use > 3) call error(122, "Quadrant out of bounds")
	}

	# fetch detector position arrays
	call clgstr("detpos", Memc[str], SZ_LINE)
	call stripwhite(str)
	nd = 0
	usevig = false
	if (!streq(Memc[str], "")) {

          # verify imaging detector only (no spectrometers)
	  if (detnum >= 4 && detnum <= 6) {
	     call error(986, "Cannot use detpos table with a spectrometer detector")
	  }       

	  usevig = clgetb("usevignetting")

	  tp = tbtopn(Memc[str], READ_ONLY, 0)
	  nd = tbpsta(tp, TBL_NROWS)
	  if (nd < 4) call error(987,"detpos table must contain at least 4 rows")
	  call malloc(nulls, nd, TY_BOOL)
	  call malloc(dtime, nd, TY_DOUBLE)
	  call malloc(dx, nd, TY_REAL)
	  call malloc(dy, nd, TY_REAL)

	  call clgstr("dtimecol", Memc[str], SZ_LINE)
	  call tbcfnd(tp, Memc[str], cp, 1)
	  if (cp == NULL) call error(123,"detpos time column not found")
	  call tbcgtd(tp, cp, Memd[dtime], Memb[nulls], 1, nd)

	  call clgstr("dxcol", Memc[str], SZ_LINE)
	  call tbcfnd(tp, Memc[str], cp, 1)
	  if (cp == NULL) call error(123,"dx column not found")
	  call tbcgtr(tp, cp, Memr[dx], Memb[nulls], 1, nd)

	  call clgstr("dycol", Memc[str], SZ_LINE)
	  call tbcfnd(tp, Memc[str], cp, 1)
	  if (cp == NULL) call error(123,"dy column not found")
	  call tbcgtr(tp, cp, Memr[dy], Memb[nulls], 1, nd)

	  call mfree(nulls, TY_BOOL)
	  call tbtclo(tp)
	} else {
	  # uh oh, no detpos table and no primbsch quadrant - no way to determine
	  # correction values
	  if (quad2use < 0 && (dead || prim)) {
		call error(133, "Cannot compute primbsching and deadtime without either a quadrant or detpos table")
	  }
	}
	
	# open vignetting image
	if (usevig) {
	  undistort = clgetb("undistort")

	  call clgstr("filtertype", Memc[str2], SZ_LINE)
	  if ( ! getvigname(Memc[str2], ref_tp, detnum, Memc[str], SZ_LINE))
	    call error(543,"unable to get vignetting image name")
	  if (streq(Memc[str],"INDEF"))
	    call error(321,"invalid filtertype for this detector")

	  call sprintf(Memc[str2], SZ_LINE, "%s%s")
	  call pargstr(Memc[refpath])
	  call pargstr(Memc[str])

	  vim = immap(Memc[str2], READ_ONLY, 0)
	  vmw = mw_openim(vim)
	  vct = mw_sctran(vmw, "world", "logical", 3)
	  vnx = IM_LEN(vim, 1)
	  vny = IM_LEN(vim, 2)
	  call malloc(vig, vnx*vny, TY_REAL)
	  buf = imgs2r(vim, 1, vnx, 1, vny)
	  call amovr(Memr[buf], Memr[vig], vnx*vny)
	  call imunmap(vim)

	  # open distortion image
	  if (undistort) {
		if ( ! getdistname(ref_tp, detnum, Memc[str], SZ_LINE))
		  call error(543,"unable to get distortion image name")
		if (streq(Memc[str],"INDEF"))
		  call error(321,"not an imaging detector")

		call sprintf(Memc[str2], SZ_LINE, "%s%s")
		call pargstr(Memc[refpath])
		call pargstr(Memc[str])

		dim = immap(Memc[str2], READ_ONLY, 0)
		dmw = mw_openim(dim)
		dct = mw_sctran(dmw, "world", "logical", 3)
		dnx = IM_LEN(dim, 1)
		dny = IM_LEN(dim, 2)
		call malloc(dist, dnx*dny*2, TY_REAL)
		buf = imgs3r(dim, 1, dnx, 1, dny, 1, 2)
		call amovr(Memr[buf], Memr[dist], dnx*dny*2)
		call imunmap(dim)
	  }
    }	
  }

  # exposure integration subinterval
  sublen = clgetd("sublen")
  
  ######################################################################
  #  we are finally ready to compute some effective exposures
  #  there are two possibilities - a light curve or a single computation
  ######################################################################

  if (xs == NULL || xe == NULL)
     call error(123,"Couldn't get time intervals")

  call clgstr("ltcurv", Memc[str], SZ_LINE)
  call stripwhite(str)
  if (streq(Memc[str], "")) {
    # no light curve - do one integration

    # start time of integration
    tmin = clgetd("t0")
	if (IS_INDEFD(tmin)) tmin = Memd[xs]

	# end time of integration
	tmax = clgetd("t1")
	if (IS_INDEFD(tmax)) tmax = Memd[xe+nranges-1]

	# integrate
    if (ex == NULL) {
	  exposure = INDEFD
	  badexp = INDEFD
	  rawexp = INDEFD
	}
    else {
	  call rl_integ(tmin, tmax, Memd[xs], Memd[xe], nranges, 
					  Memd[ptime], Memr[pcorr], np, quad2use,
					  Memd[dtime], Memr[dx], Memr[dy], nd, 
					  Memr[vig], vnx, vny, vct, Memr[dist], dnx, dny, dct, 
					  detptr, pim, usevig, undistort, sublen, 
					  exposure, rawexp, badexp)
	}
  
    # output
    call clputd("exposure", exposure)
    call clputd("rawexposure", rawexp)
    if (verbose) {
	  call printf("\nExposure: %g seconds\n")
	  call pargd(exposure)
	  call printf("Raw exposure: %g seconds\n")
	  call pargd(rawexp)
	  call printf("\nThere were %g seconds of raw exposure with no\n")
	  call pargd(badexp)
	  call printf("  available correction values\n")
    }
  }
  else {
    # light curve file
    iferr (tp = tbtopn(Memc[str], READ_WRITE, 0)){
	call error(122,"Can't open ltcurv (light curve) table")
    }
    nlt = tbpsta(tp, TBL_NROWS)
    call tbcfnd(tp, "time", cp, 1)
    if (cp == NULL) call error(123,"ltc time column not found")

    call tbcfnd(tp, "rawexp", op[1], 1)
    if (op[1] == NULL) {
      # new light curve table
      call tbcdef(tp, op[1], "rawexp", "seconds", "%g", TY_REAL, 1, 1)
      call tbcdef(tp, op[2], "effexp", "seconds", "%g", TY_REAL, 1, 1)
      call tbcdef(tp, op[3], "invalid", "seconds", "%g", TY_REAL, 1, 1)

      call tbcdef(tp, op[6], "corr_rate", "cps", "%g", TY_REAL, 1, 1)
      call tbcdef(tp, op[7], "corr_err", "cps", "%g", TY_REAL, 1, 1)
    } else {
	  # overwrite a previously corrected table
      call tbcfnd(tp, "effexp", op[2], 1)
      call tbcfnd(tp, "invalid", op[3], 1)
      call tbcfnd(tp, "corr_rate", op[6], 1)
      call tbcfnd(tp, "corr_err", op[7], 1)
    }

    call tbcfnd(tp, "net", op[4], 1)
    call tbcfnd(tp, "neterr", op[5], 1)

    call tbrgtd(tp, cp, tmin, null, 1, 1)
    call tbrgtd(tp, cp, tmax, null, 1, 2)
    binsize = tmax - tmin
    do i = 1, nlt {
	  call tbrgtd(tp, cp, tmin, null, 1, i)
	  tmax = tmin + binsize

	  call rl_integ(tmin, tmax, Memd[xs], Memd[xe], nranges, 
					  Memd[ptime], Memr[pcorr], np, quad2use,
					  Memd[dtime], Memr[dx], Memr[dy], nd, 
					  Memr[vig], vnx, vny, vct, Memr[dist], dnx, dny, dct, 
					  detptr, pim, usevig, undistort, sublen, 
					  exposure, rawexp, badexp)

	  call tbrptd(tp, op[1], rawexp, 1, i)
	  call tbrptd(tp, op[2], exposure, 1, i)
	  call tbrptd(tp, op[3], badexp, 1, i)

	  call tbegtr(tp, op[4], i, net)
	  call tbegtr(tp, op[5], i, neterr)

	  if (exposure != 0.0) {
		rate = net / exposure
		rerr = neterr / exposure
	  }
	  else {
		rate = 0.0
		rerr = 0.0
	  }


	  call tbrptr(tp, op[6], rate, 1, i)
	  call tbrptr(tp, op[7], rerr, 1, i)
	}

	call tbtclo(tp)
  }


  # clean up

  if (ex != NULL) {
	call tbtclo(ref_tp)
	call mfree (xs, TY_DOUBLE)
	call mfree (xe, TY_DOUBLE)
	if (np != 0) {
	  call mfree (ptime, TY_DOUBLE)
	  call mfree (pcorr, TY_REAL)
	}
	if (nd != 0) {
	  call mfree (dtime, TY_DOUBLE)
	  call mfree (dx, TY_REAL)
	  call mfree (dy, TY_REAL)
	}
	if (pim != NULL) call imunmap(pim)
	if (usevig) {
	  call mfree (vig, TY_REAL)
	  call mw_ctfree(vct)
	  call mw_close(vmw)
	  if (undistort) {
		call mfree (dist, TY_REAL)
		call mw_ctfree(dct)
		call mw_close(dmw)
	  }
	}
  }
  if (qp != NULL) {
	call qpio_close(qpio)
	call qp_close(qp)
  }
  call sfree(sp)
end


#
# Open a QPOE event list
#
procedure	qp_open_filtered (qp, qpio, file)

char	file[ARB]
pointer	qp, qpio
  
int	ip

pointer	qp_open(), qpio_open()

begin

	# find char position of EOS or bracket
	# char position held by ip

	for (ip=1; file[ip] != EOS; ip = ip + 1) {
		if (file[ip] == '[') {
			break
		}
	}

	if (file[ip] == EOS) {
		qp = qp_open(file, READ_ONLY, 0)
		qpio = qpio_open(qp, "", READ_ONLY)
	} else {
		file[ip] = EOS
		qp = qp_open(file, READ_ONLY, 0)
		file[ip] = '['
		qpio = qpio_open(qp, file[ip], READ_ONLY)
	}

end



bool procedure getvigname(filt, tp, det, str, sz)

char	filt[ARB], str[sz]
int		det, sz
pointer	tp

int		nrow, tab_det, i
bool	found
pointer	sp, col, cp

int		tbpsta()

begin
  call smark(sp)

  call salloc(col, SZ_LINE, TY_CHAR)

  found = false

  nrow = tbpsta (tp, TBL_NROWS)
  call tbcfnd (tp, TIFNUM_COLNAME, cp, 1)
  if (cp == NULL) call error(567,"unable to find TIF number column")
  for (i=1; i<=nrow; i=i+1 ) {
	call tbegti(tp, cp, i, tab_det)

	if (tab_det == det) {
	  found = true
	  break
	}
  }	

  call sprintf(Memc[col], SZ_LINE, "%s_vignette")
  call pargstr(filt)
  call tbcfnd(tp, Memc[col], cp, 1)
  if (cp == NULL) call error(789,"unable to find vignetting column")
  call tbegtt(tp, cp, i, str, sz)

  call sfree(sp)

  return found
end



bool procedure getdistname(tp, det, str, sz)

char	str[sz]
int		det, sz
pointer	tp

int		nrow, tab_det, i
bool	found
pointer	sp, cp

int		tbpsta()

begin
  call smark(sp)

  found = false

  nrow = tbpsta (tp, TBL_NROWS)
  call tbcfnd (tp, TIFNUM_COLNAME, cp, 1)
  if (cp == NULL) call error(567,"unable to find TIF number column")
  for (i=1; i<=nrow; i=i+1 ) {
	call tbegti(tp, cp, i, tab_det)

	if (tab_det == det) {
	  found = true
	  break
	}
  }	

  call tbcfnd(tp, DISTORTION_COL, cp, 1)
  if (cp == NULL) call error(789,"unable to find distortion column")
  call tbegtt(tp, cp, i, str, sz)

  call sfree(sp)

  return found
end




procedure stripwhite (s)

pointer s

pointer p, i

begin
  if (Memc[s] == EOS) return
  p = s
  i = s
  while (IS_WHITE(Memc[i])) i = i + 1
  while (Memc[i] != EOS) {
    Memc[p] = Memc[i]
    i = i + 1
    p = p + 1
  }
  p = p - 1
  while (IS_WHITE(Memc[p])) p = p - 1
  Memc[p+1] = EOS
end


# Load primbsch and/or deadtime info from primbsch table into pcorr.
# Primbsch data is reconstructed from other columns if necessary and possible.
# Missing info is indicated by INDEF values.
#
procedure get_primb_corr(tp, pcorr, nulls, np, prim, dead, det, detnum)
pointer tp	  #I primbsch table
pointer pcorr	  #O filled in corrections for 4 columnsp
pointer nulls     #O temp nulls buffer to use for I/O
int     np	  #I len of nulls buffer, pcorr buffer (times 4)
bool    prim      #I use primbsch correction?
bool    dead	  #I use deadtime correction?
char	det[ARB]  #I detector name
int	detnum    #I detector number

pointer str	  # local string
pointer sp        # stack pointer
pointer dead_cp   # deadtime column pointer 
pointer primb_cp  # primbsch column pointer
pointer dpc_cp    # deadtime primbsch column pointer
pointer dead_vec  # deadtime info vector
pointer dpc_vec   # deadtime primbsch info vector
int     i
bool	data_found  # false if no data found at all (all set to INDEF)

begin

   call smark(sp)   
   call salloc(str, SZ_LINE, TY_CHAR)

   # try to open deadtime col by either name format
   call sprintf(Memc[str], SZ_LINE, "det%ddc")
   call pargi(detnum)
   call tbcfnd(tp, Memc[str], dead_cp, 1)

   if (dead_cp == NULL) {
      call sprintf(Memc[str], SZ_LINE, "%sdc")
      call pargstr(det)
      call tbcfnd(tp, Memc[str], dead_cp, 1)
   }

   data_found=false
   do i = 1, 4 {

      # try to open primbsch col
      call sprintf(Memc[str], SZ_LINE, "det%dq%dpc")
      call pargi(detnum)
      call pargi(i-1)
      call tbcfnd(tp, Memc[str], primb_cp, 1)

      if (primb_cp == NULL) {
	 call sprintf(Memc[str], SZ_LINE, "%sq%dpc")
	 call pargstr(det)
	 call pargi(i-1)
	 call tbcfnd(tp, Memc[str], primb_cp, 1)
      }


      # try to open deadtime primbsch col
      call sprintf(Memc[str], SZ_LINE, "det%dq%ddpc")
      call pargi(detnum)
      call pargi(i-1)
      call tbcfnd(tp, Memc[str], dpc_cp, 1)

      if (dpc_cp == NULL) {
	 call sprintf(Memc[str], SZ_LINE, "%sq%ddpc")
	 call pargstr(det)
	 call pargi(i-1)
	 call tbcfnd(tp, Memc[str], dpc_cp, 1)
      }


      if (!prim) {

         # deadtime only
         if (dead_cp == NULL)
            call error(123,"deadtime column not found")
         else {
	    call tbcgtr(tp, dead_cp, Memr[pcorr+np*(i-1)], Memb[nulls], 1, np)
	    data_found=true
         }
      }
      else {
         if (!dead) {

            # primbsch only
            if (primb_cp == NULL) {

               # see if we can reconstruct it from other columns
               if (dead_cp == NULL || dpc_cp == NULL) {

                  # fill with INDEFS
                  call amovkr(INDEFR, Memr[pcorr+np*(i-1)], np)

               }
               else {

		  # reconstruct primbsch column from deadtime & dpc cols
		  call malloc(dead_vec, np, TY_REAL)
		  call malloc(dpc_vec, np, TY_REAL)

		  call tbcgtr(tp, dead_cp, Memr[dead_vec], Memb[nulls], 1, np)
		  call tbcgtr(tp, dpc_cp, Memr[dpc_vec], Memb[nulls], 1, np)
		  call idivr(Memr[dpc_vec], Memr[dead_vec],
			  Memr[pcorr+np*(i-1)], np)

		  call mfree (dead_vec, TY_REAL)
		  call mfree (dpc_vec, TY_REAL)

		  data_found=true
               }
            }
            else {
               # read primbsch column
	       call tbcgtr(tp, primb_cp, Memr[pcorr+np*(i-1)], Memb[nulls], 1, np)
	       data_found=true

            }
         }
         else {

            # combined primbsch/deadtime
            if (dpc_cp == NULL) {
               # fill with INDEFS
               call amovkr(INDEFR, Memr[pcorr+np*(i-1)], np)
            }
            else {
               call tbcgtr(tp, dpc_cp, Memr[pcorr+np*(i-1)], Memb[nulls], 1, np)
               data_found=true
            }

         }
      }

   }

   if (! data_found)
	call error(123,"No deadtime and/or primbsch correction data found")

   call sfree(sp)

end
