# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  fappendtext.cl
# 	Author	:  tnk
# 	Date	:  11/13/96
# 	Synopsis:  
# 	Keywords:
# 		@(#) fappendtext.cl -r1.10 11/13/96 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	08/07/97   mabbott	removed unneeded deletion of a HISTORY card
# 	02/20/97   tnk	adjusted prompts, fixed typo
# 	02/20/97   tnk	made local copies of param variables to prevent prompts
# 	02/05/97   tnk	prepended blank COMMENT card to comment block
# 	02/05/97   tnk	changed TELTYPE to INSTTYPE
# 	01/31/97   tnk	updated header card list
# 	11/15/96   tnk	delete extra history card from fcreate
# 	11/13/96   tnk	header card massages
# 	11/13/96   tnk	more error checking &comments
# 	11/13/96   tnk	date and time created 96/11/13 08:30:13 by tnk
# H-
# 
#


procedure fappendtext (fitsfile, textfile)

# @(#)fappendtext.cl	1.10 8/7/97
# Appends textfile as a binary table extension named
# by extname to the end of fitsfile.  Copies selected header cards from
# the primary to the new extension.  Also adds COMMENT records
# to the new extension.

string fitsfile = ""		{prompt="Fits file to modify"}          
string textfile = ""		{prompt="Text file to append"}
string extname = "anomaly"	{prompt="Name of extension to create"}

struct *hdrlist {prompt="(Internal variable -- ignore)"}

begin

string tmp
struct line
string local_fitsfile
string local_textfile
string local_extname

local_fitsfile = fitsfile
local_textfile = textfile
local_extname = extname


# Make sure needed package is loaded
if (! defpac ("ftools") || ! defpac ("futils") ) {
    print ("You must load ftools and futils to run this task.")    
    bye
}

if (! access(local_fitsfile)) {
   print ("Can't access fits file: " // local_fitsfile)
   bye
}

if (! access(local_textfile)) {
   print ("Can't access text file: " // local_textfile)
   bye
}


print ("appending " // local_textfile // " to fits file " // local_fitsfile)

tmp="/tmp/fitsapp."

# Delete any leftover temp files
delete( files = tmp // "*", go_ahead = yes, verify = no,
        default_acti = yes, allversions = yes, subfiles = yes,
        >& "dev$null" )

# set up column definition
print ("LINE 80A text 1 80" , > tmp // "cd")

# turn text into temporary fits file
fcreate (cdfile = tmp // "cd",
   datafile = local_textfile,
   outfile =  tmp // "suffix.fit",
   headfile = "",
   tbltype = "binary",
   nskip = 0,
   nrows = 0,
   history = yes,
   morehdr = 0,
   extname = local_extname,
   anull = " ",
   inull = 0,
   clobber = no)

# delete HISTORY cards put in by fcreate
   fparkey (value = "",
     fitsfile = tmp // "suffix.fit[1]",
     keyword = "-HISTORY",
     comm = "",
     add = no)

#
# mod by MJA 8-7-97  fcreate appears to only put in one history card
#
#   fparkey (value = "",
#     fitsfile = tmp // "suffix.fit[1]",
#     keyword = "-HISTORY",
#     comm = "",
#     add = no)

# add TDISP1 card 
   fparkey (value = "A80",
     fitsfile = tmp // "suffix.fit[1]",
     keyword = "TDISP1",
     comm = "%80s",
     add = yes)



# copy selected headers from original fits primary to the temp fits file

print ("ORIGIN", >> tmp // "hdrlist")
print ("CREATOR", >> tmp // "hdrlist")
print ("TELESCOP", >> tmp // "hdrlist")
print ("INSTTYPE", >> tmp // "hdrlist")
print ("OBJECT", >> tmp // "hdrlist")
print ("RA_OBJ" , >> tmp // "hdrlist")
print ("DEC_OBJ" , >> tmp // "hdrlist")
print ("RA_PNT" , >> tmp // "hdrlist")
print ("DEC_PNT", >> tmp // "hdrlist")
print ("RA_PROC", >> tmp // "hdrlist")
print ("DEC_PROC", >> tmp // "hdrlist")
print ("OBSERVER", >> tmp // "hdrlist")
print ("DATE-OBS", >> tmp // "hdrlist")
print ("TIME-OBS", >> tmp // "hdrlist")
print ("DATE-END", >> tmp // "hdrlist")
print ("TIME-END", >> tmp // "hdrlist")
print ("OBS_MODE", >> tmp // "hdrlist")
print ("DITHER", >> tmp // "hdrlist")
print ("DETMODE", >> tmp // "hdrlist")
print ("OFF-AXIS", >> tmp // "hdrlist")
print ("MOVING", >> tmp // "hdrlist") 
print ("DAYNIGHT", >> tmp // "hdrlist") 
print ("VALIDTIM", >> tmp // "hdrlist")
print ("RA_UNIT", >> tmp // "hdrlist") 
print ("DEC_UNIT", >> tmp // "hdrlist") 
print ("EQUINOX" , >> tmp // "hdrlist")
print ("RADECSYS", >> tmp // "hdrlist")
print ("TIMESYS" , >> tmp // "hdrlist")
print ("TIMEZERO", >> tmp // "hdrlist")
print ("TIMEUNIT", >> tmp // "hdrlist")
print ("CLOCKCOR", >> tmp // "hdrlist")
print ("TIMEREF", >> tmp // "hdrlist")
print ("TASSIGN", >> tmp // "hdrlist")
print ("TSTART", >> tmp // "hdrlist")
print ("TSTOP", >> tmp // "hdrlist")
print ("MJDREF", >> tmp // "hdrlist")
print ("EGOCSVER", >> tmp // "hdrlist")
print ("REFVERS", >> tmp // "hdrlist") 


hdrlist = tmp // "hdrlist"

while (fscan (hdrlist, line) != EOF) {

   fkeypar (fitsfile = local_fitsfile // "[0]",
      keyword = line)

   if (fkeypar.exist) {

      fparkey (value = fkeypar.value,
	 fitsfile = tmp // "suffix.fit[1]",
	 keyword = line,
	 comm = fkeypar.comm,
	 add = yes)
   }
}

# append INHERIT flag
   fparkey (value = "F",
     fitsfile = tmp // "suffix.fit[1]",
     keyword = "INHERIT",
     comm = "Do not inherit cards from primary header",
     add = yes)

# append comment block

# append blank COMMENT 
   fparkey (value = "",
     fitsfile = tmp // "suffix.fit[1]",
     keyword = "COMMENT",
     comm = "",
     add = yes)


hdrlist = tmp // "commlist"

print ("This table is composed of 80 character lines of text", >> tmp // "commlist")
print ("containing anomaly information for this observation,", >> tmp // "commlist")
print ("in the form of comments recorded by EUVE personnel", >> tmp // "commlist")
print ("during the data collection and data analysis.", >> tmp // "commlist")
print ("The comments are presented as is, and may not always", >> tmp // "commlist")
print ("be of use.  The comments are not to be considered a", >> tmp // "commlist")
print ("complete or exhaustive record of all irregularities", >> tmp // "commlist")
print ("which may be present in a data set.", >> tmp // "commlist")

while (fscan (hdrlist, line) != EOF) {

      fparkey (value = line,
	 fitsfile = tmp // "suffix.fit[1]",
	 keyword = "COMMENT",
	 comm = "",
	 add = yes)
}


# append temp file to original fits file
fappend  (infile = tmp // "suffix.fit[1]",
   outfile = local_fitsfile,
   pkeywds = no,
   history = no)


# Delete the temp files
delete( files = tmp // "*", go_ahead = yes, verify = no,
        default_acti = yes, allversions = yes, subfiles = yes,
        >& "dev$null" )

end



