# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_mapcorr.x
# 	Author	:  tnk
# 	Date	:  01/02/97
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_mapcorr.x -r1.8 01/02/97 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	09/08/97   bboyd	commented out unused private include files
# 	03/05/97   mabbott	removed relx flip, added axis labels
# 	02/27/97   tnk	qp ra/dec as reals or doubles; ephem ra as hours or degrees
# 	01/14/97   tnk	changed qpl_open_filtered to qpofilt
# 	01/13/97   tnk	pulled out add_qp_hist routine
# 	01/08/97   mabbott	omitted read of pacol if rotate=no, fixed bug in INDEF detection
# 	01/02/97   tnk	renamed add_qp_hist to make it unique
# 	01/02/97   tnk	date and time created 97/01/02 10:43:40 by tnk
# H-
# 
#

# @(#)t_mapcorr.x	1.8 9/8/97 CEA SSL UCB 
#
#   MAPCORR
#
#  Remap a QPOE relative to a source ephemeris.  This is an alternate
#  algorithm to using aspcorr.  Instead of modifying the aspect and
#  then running CEP, as with aspcorr, one can run CEP on the telemetered
#  aspect and then take the resulting smeared out events and correct
#  them relative to an ephemeris using mapcorr.  This only works
#  for the DS and scanners.
#
#  Mapcorr output has positive offset in RA and DEC along the positive 
#  x and y axes. 
#
#  Mapcorr can also adjust the relative rotation of the events, correcting
#  the events to place any desired point on the source at the top of the
#  coordinate system.  For example, specifying a "pacol" column containing
#  the North Pole position angle will remap events to keep the North Pole
#  of the source pointed along the +y axis.  If rotate is "no", then the
#  events are remapped to place celestial north (positive dec offset) 
#  on the +y axis.
#
#  Note that the output QPOE file is not sorted or indexed.  
#
#  The input ephemeris is assumed to contain no large gaps. Linear
#  interpolation is used between ephemeris points no matter how far apart
#  the points may be in time.
#
#  Some parts of the code are hardcoded to the data types in EUVE QPOE 
#  files.
#

include	<qpioset.h>
# unused private include files
# include <qpoe$qpoe.h>
# include <qpoe$qpio.h>
include <qpset.h>
include <tbset.h>
include <mach.h>
include <ctype.h>

define SZ_PARAM 1024
define BUFFSIZE 8192


define RA 1
define DEC 2
define PA 3

define RA_OFF 1
define DEC_OFF 2
define TIME_OFF 3
define X_OFF 4
define Y_OFF 5


define PI double(3.14159265358979323846264338)
define DEG2RAD double(PI/180.0)

procedure	t_mapcor()

# input params
pointer	infile		 # input filename
pointer outfile		 # output filename
pointer time_att	 # TIME attribute
pointer ra_att		 # RA attribute
pointer dec_att		 # DEC attribute
pointer x_att		 # X attribute
pointer y_att		 # Y attribute

#local vars
pointer	inqp, inqpio	# input QPOE pointers
pointer	outqp, outqpio	# output QPOE pointers
int		maskval
pointer	iev[BUFFSIZE]	# input event buffer
pointer	oev[BUFFSIZE]	# output event buffer
int		nev				# number of events in input buffer
int		noev			# number of events in output buffer
int		offsets[5]		# offsets of attributes
int		nx, ny			# size of QPOE image
int		bucketlen
pointer		hist
int	ra_dtype, dec_dtype


pointer	mwcs			# MWCS handle for output coord system
pointer	ct				# coord trans handle
real	dx, dy			# desired scale of output image
real	r[2]			# MWCS physical origin
real	w[2]			# MWCS world origin
real	cd[2,2]			# MWCS scale

pointer	etp				# ephemeris table pointer
int		nephrow			# number of rows in ephemeris
pointer	tcp				# ephemeris time column pointer
pointer	cp[3]			# ephemeris column pointers
bool	null[3]			# nulls table flags
pointer	etime			# ephemeris times
pointer	ephem			# ephemeris data

bool	sourcerot
bool	verbose
int		intotal
int		outtotal

pointer	istr, ostr, pat
int		len, patlen, patstart, maxpatlen

pointer	sp
pointer	str
int		i, j
bool		ra_hours

int		strlen(), qp_gstr(), strsearch()
int		qpio_getevents()
int		qpio_stati()
pointer	qp_open()
pointer	qpio_open()
int		qp_geti()
pointer	mw_open()
pointer	mw_sctran()
int		tbpsta()
pointer	tbtopn()
int		clgeti()
real	clgetr()
bool	clgetb()

begin

  call smark(sp)
  call salloc(str, SZ_LINE, TY_CHAR)

  call eu_clstr(infile, "input")	# get qpoe filename 
  call eu_clstr(outfile, "output")  # get qpoe filename 
  call eu_clstr(time_att, "time")	# get TIME name
  call eu_clstr(ra_att, "ra")    	# get RA name
  call eu_clstr(dec_att, "dec")  	# get DEC name
  call eu_clstr(x_att, "x")  		# get X name
  call eu_clstr(y_att, "y")  		# get Y name

  dx = clgetr("dx")
  dy = clgetr("dy")

  if (IS_INDEFR(dx) || IS_INDEFR(dy)) {
	call error("You cannot specify an INDEF value for dx or dy")
  }

  nx = clgeti("nx")
  ny = clgeti("ny")

  sourcerot = clgetb("rotate")
  verbose = clgetb("verbose")

  ##########
  #
  # Open QPOE files
  #
  ##########

  # open input QPOE file, filtered
  call qpofilt(inqp, inqpio, Memc[infile])

  # open output QPOE as new copy of input qpoe
  outqp = qp_open (Memc[outfile], NEW_COPY, inqp)

  # set image size of output QPOE
  if (IS_INDEFI(nx)) {
	nx = qp_geti(inqp, "axlen[1]")
  }
  if (IS_INDEFI(ny)) {
	ny = qp_geti(inqp, "axlen[2]")
  }
  call qp_puti(outqp, "axlen[1]", nx)
  call qp_puti(outqp, "axlen[2]", ny)

  # set bucket length (QPOE only, not QPIO)
  bucketlen =  qpio_stati(inqpio, QPIO_BUCKETLEN)
  bucketlen = bucketlen + 2   # adjust for QPOE/QPIO bucketlen offset
  call qp_seti(outqp, QPOE_BUCKETLEN, bucketlen)
  
  # open for qpio, note that the event list name is the default
  outqpio = qpio_open(outqp, "", NEW_FILE)

  # add history comment
  call salloc(hist, SZ_LINE, TY_CHAR)
  call strcpy("mapcorr: created from ", Memc[hist], SZ_LINE)
  call strcat(Memc[infile], Memc[hist], SZ_LINE)
  call add_qp_hist(outqp, Memc[hist])  


  # rename RA and DEC in output QPOE to RELX and RELY
  iferr {
	call qp_renamef(outqp, Memc[ra_att], "relx")
	call qp_renamef(outqp, Memc[dec_att], "rely")
  } then {
	# ok, just skip this step
  }

  maxpatlen = max(strlen(Memc[ra_att]), strlen(Memc[dec_att]))+2
  call salloc(pat, maxpatlen, TY_CHAR)
  call salloc(istr, SZ_PARAM, TY_CHAR)
  call salloc(ostr, SZ_PARAM, TY_CHAR)

  ifnoerr (len = qp_gstr(outqp, "XS-EVENT", Memc[istr], SZ_PARAM)) {
	call sprintf(Memc[pat], maxpatlen, ":%s,")
	call pargstr(Memc[ra_att])
	patlen = strlen(Memc[pat])
	i = strsearch(Memc[istr], Memc[pat])
	if (i != 0) {
	  patstart = i-patlen-1
	  call strcpy(Memc[istr], Memc[ostr], patstart)
	  call sprintf(Memc[ostr+patstart], SZ_PARAM-(patstart), ":relx,")
	  call strcpy(Memc[istr+i-1], Memc[ostr+(patstart+6)], 
				  SZ_PARAM-(patstart+6))
	}
	call strcpy(Memc[ostr], Memc[istr], SZ_PARAM)

	call sprintf(Memc[pat], maxpatlen, ":%s,")
	call pargstr(Memc[dec_att])
	patlen = strlen(Memc[pat])
	i = strsearch(Memc[istr], Memc[pat])
	if (i != 0) {
	  patstart = i-patlen-1
	  call strcpy(Memc[istr], Memc[ostr], patstart)
	  call sprintf(Memc[ostr+patstart], SZ_PARAM-patstart, ":rely,")
	  call strcpy(Memc[istr+i-1], Memc[ostr+(patstart+6)], 
				  SZ_PARAM-(patstart+6))
	}

	call qp_pstr(outqp, "XS-EVENT", Memc[ostr])
  }

  ##########
  #
  # create a wcs in the output QPOE
  #
  ##########

  mwcs = mw_open(NULL, 2)
  call mw_newsystem(mwcs, "source-centered", 2)
  r[1] = real(nx) / 2 + 0.5
  r[2] = real(ny) / 2 + 0.5
  w[1] = 0.0
  w[2] = 0.0
  cd[1,1] = dx
  cd[2,2] = dy
  cd[1,2] = 0.0
  cd[2,1] = 0.0
  call mw_swtermr(mwcs, r, w, cd, 2)
  call mw_swattrs(mwcs, 1, "label", "Relx")
  call mw_swattrs(mwcs, 2, "label", "Rely")
  call mw_swattrs(mwcs, 1, "units", "degrees")
  call mw_swattrs(mwcs, 2, "units", "degrees")
  call qp_savewcs(outqp, mwcs)
  ct = mw_sctran(mwcs, "world", "physical", 3)


  ##########
  #
  #  get attribute offsets
  #
  ##########

  call mqp_offset(inqp, Memc[ra_att], offsets[RA_OFF])
  call mqp_offset(inqp, Memc[dec_att], offsets[DEC_OFF])
  call mqp_offset(inqp, Memc[time_att], offsets[TIME_OFF])
  call mqp_offset(inqp, Memc[x_att], offsets[X_OFF])
  call mqp_offset(inqp, Memc[y_att], offsets[Y_OFF])


  ############
  #
  # open and read ephemeris file
  #
  ############

  call clgstr("ephemeris", Memc[str], SZ_LINE)
  etp = tbtopn(Memc[str], READ_ONLY, 0)
  nephrow = tbpsta(etp, TBL_NROWS)

  call clgstr("ephtime", Memc[str], SZ_LINE)
  call tbcfnd(etp, Memc[str], tcp, 1)
  if (tcp == 0) call error(123, "ephemeris time column not found")

  call clgstr("racol", Memc[str], SZ_LINE)
  call tbcfnd(etp, Memc[str], cp[RA], 1)
  if (cp[RA] == 0) call error(123, "ephemeris RA column not found")

  call clgstr("deccol", Memc[str], SZ_LINE)
  call tbcfnd(etp, Memc[str], cp[DEC], 1)
  if (cp[DEC] == 0) call error(123, "ephemeris Dec column not found")

  if (sourcerot) {
	call clgstr("pacol", Memc[str], SZ_LINE)
	call tbcfnd(etp, Memc[str], cp[PA], 1)
	if (cp[PA] == 0) call error(123, "ephemeris position angle column not found")
  } else {
	# ugly (but quick) hack - just read the DEC twice if there is no pacol
	# this is wasteful but harmless
	cp[PA] = cp[DEC]
  }

  # determine ephem ra units, convert hours to degrees if necessary
  call get_ra_units(cp[RA], ra_hours)

  call malloc(etime, nephrow, TY_DOUBLE)
  call malloc(ephem, 3*nephrow, TY_DOUBLE)

  # read ephemeris data
  j = 0
  do i = 1, nephrow {

	call tbrgtd(etp, tcp, Memd[etime+i-1], null, 1, i)
	call tbrgtd(etp, cp, Memd[ephem+j], null, 3, i)
	if (ra_hours) {
	    # convert ra from hours to degrees
	    Memd[ephem+j] = Memd[ephem+j] * 15.0
	}
	j = j + 3
	
  }
  call tbtclo(etp)



  # get ra/dec field data types
  call mqp_att_dtype(inqp, Memc[ra_att], ra_dtype)
  call mqp_att_dtype(inqp, Memc[dec_att], dec_dtype)


  ##############
  #
  # copy & process events
  #
  ##############

  maskval = 0
  intotal = 0
  outtotal = 0
  while (qpio_getevents(inqpio, iev, maskval, BUFFSIZE, nev) != EOF) {

	call remap_events(iev, nev, offsets, ra_dtype, dec_dtype, Memd[etime], Memd[ephem], 
					  nephrow, sourcerot, ct, nx, ny, oev, noev)

	call qpio_putevents(outqpio, oev, noev)

	intotal = intotal + nev
	outtotal = outtotal + noev

  }

  if (verbose) {
	call printf("Read %d events\nWrote %d events\n")
	call pargi(intotal)
	call pargi(outtotal)
  }


  #############
  #
  # Clean up
  #
  #############

  call mw_ctfree(ct)
  call mw_close(mwcs)

  call mfree(etime, TY_DOUBLE)
  call mfree(ephem, TY_DOUBLE)

  call qpio_close(inqpio)
  call qp_close(inqp)

  call qpio_close(outqpio)
  call qp_close(outqp)

  call sfree(sp)
end


#
# remap the events relative to a tabulated position
#
procedure remap_events(iev, niev, offsets, ra_dtype, dec_dtype, etime,
	 ephem, neph, rotate, mwcs, nx, ny, oev, noev)

int		niev			#I number of events in buffer
pointer	iev[niev]		#I input buffer of events
int		offsets[5]		#I offsets in event record
int	ra_dtype, dec_dtype
int		neph	  		#I number of ephmeris rows
double	etime[neph] 	#I ephemeris times
double	ephem[3,neph] 	#I ephemeris data
bool	rotate			#I rotate the event into source frame?
pointer	mwcs			#I MWCS ct handle used to map world to pixel coords
int		nx, ny			#I size of output QPOE
pointer	oev[niev]		#O output buffer of events
int		noev			#O output number of events

int		i
double	time
double	ra, dec
double	ara, adec
double	xsrc, ysrc
double	rx, ry
double	era, edec, epa
short	x, y


real	qp_gvalr()
double	qp_gvald()

begin


  # loop over all events in the input buffer

  noev = 0
  do i = 1, niev {

    time = qp_gvald(iev[i], offsets[TIME_OFF])

    # get the RA & DEC of the event
    # must be either real or double
    #
    #  note: I have hardcoded the data types of the EUVE QPOE 
    #        attributes.  It would be better to be generic.

    if (ra_dtype == TY_REAL)
       ra = double(qp_gvalr(iev[i], offsets[RA_OFF]))
    else if (ra_dtype == TY_DOUBLE)
       ra = qp_gvald(iev[i], offsets[RA_OFF])
    else
       call error(123,"RA datatype must be real or double")

    if (dec_dtype == TY_REAL)
       dec = double(qp_gvalr(iev[i], offsets[DEC_OFF]))
    else if (dec_dtype == TY_DOUBLE)
       dec = qp_gvald(iev[i], offsets[DEC_OFF])
    else
       call error(123,"DEC datatype must be real or double")


	# correct negative RAs in old EUVE QPOEs
	if (ra < 0.0) {
	  ra = ra + 360.0
	}

	# compute source-centered coordinates
	call get_ephem(neph, etime, ephem, time, era, edec, epa)

	if (IS_INDEFD(era) || IS_INDEFD(edec)) {
	  xsrc = INDEFD
	  ysrc = INDEFD
	} else {
	  ara = (ra - era) * cos((edec+dec)/2*DEG2RAD)
	  adec = dec - edec
	  if (rotate) {
		# rotate the event so source has PA = 0 at the top
		if (! IS_INDEFD(epa)) {
		  epa = -epa * DEG2RAD
		  xsrc = ara * cos(epa) + adec * sin(epa)
		  ysrc = -ara * sin(epa) + adec * cos(epa)
		} else {
		  xsrc = INDEFD
		  ysrc = INDEFD
		}
	  } else {
		xsrc = ara
		ysrc = adec
	  }	  
	}

	if (! IS_INDEFD(xsrc) && ! IS_INDEFD(ysrc)) {

	  # compute new image pixels
	  call mw_c2trand(mwcs, xsrc, ysrc, rx, ry)
	  x = int(rx + 0.5)
	  y = int(ry + 0.5)

	  if (x >= 1 && x <= nx &&
		  y >= 1 && y <= ny) {
		# another output event
		noev = noev + 1

		# copy event to output buffer
		oev[noev] = iev[i]

		# write the new coords to output event
		#
		#  note: I have hardcoded the data types of the EUVE QPOE 
		#        attributes.  It would be better to be generic.
		#
		if (ra_dtype == TY_REAL)
		   call qp_pvalr(oev[noev], offsets[RA_OFF], real(xsrc))
		else
		   call qp_pvald(oev[noev], offsets[RA_OFF], xsrc)

		if (dec_dtype == TY_REAL)
		   call qp_pvalr(oev[noev], offsets[DEC_OFF], real(ysrc))
		else
		   call qp_pvald(oev[noev], offsets[DEC_OFF], ysrc)

		call qp_pvals(oev[noev], offsets[X_OFF], x)
		call qp_pvals(oev[noev], offsets[Y_OFF], y)
	  }
	}
  }
end
	

#
# look up values in an ephemeris
#
#   Linear interpolation is used between tabulated points.
#   The ephemeris is assumed to have no substantial gaps,
#   but may contain INDEF values.
#
procedure get_ephem(neph, etime, ephem, time, ra, dec, pa)

int		neph			#I number of ephmeris rows
double	etime[neph]		#I ephemeris times
double	ephem[3,neph]	#I ephemeris data
double	time			#I desired time 
double	ra, dec, pa		#O ephemeris coords

int		i
double	frac

begin
  # find time in the ephemeris
  call hunt(etime, neph, time, i)

  # if we are off the table, this time is invalid
  if (i <= 0 || i >= neph) {
	ra = INDEFD
	dec = INDEFD
	pa = INDEFD
	return
  }
					 
  if (IS_INDEFD(ephem[RA, i]) || IS_INDEFD(ephem[RA, i+1]) ||
	  IS_INDEFD(ephem[DEC, i]) || IS_INDEFD(ephem[DEC, i+1]) ||
	  IS_INDEFD(ephem[PA, i]) || IS_INDEFD(ephem[PA, i+1])) {
	# There is an INDEF row before or after this time.  This means
	# the ephemeris information is missing for any of a
	# number of reasons, usually related to end effects or telemetry
	# gaps.  The time is invalid.
	ra = INDEFD
	dec = INDEFD
	pa = INDEFD
	return
  }

  # Interpolate

  frac = (time - etime[i]) / (etime[i+1] - etime[i])
  ra = (ephem[RA, i] + frac * (ephem[RA, i+1] - ephem[RA, i]))
  dec = (ephem[DEC, i] + frac * (ephem[DEC, i+1] - ephem[DEC, i]))
  pa = (ephem[PA, i] + frac * (ephem[PA, i+1] - ephem[PA, i]))
end


#
#  Get the offset of a particular attribute in a QPOE file
#
#     the offset is returned in units of CHARS
#
procedure mqp_offset(qp, att, offset)

pointer	qp			#I QPOE file handle
char	att[ARB]	#I attribute name
int		offset		#O offset of attribute in CHARS

pointer	field		# full attribute name
char	kind		# attribute data type
char	firstdig	# first digit of attribute offset
int		ip			# string index
long	lval		# offset in bytes
pointer	sp

int		stridx()

begin
  call smark(sp)

  call salloc(field, SZ_LINE, TY_CHAR)

  # find full attribute name
  call qp_expandtext(qp, att, Memc[field], SZ_LINE)

  # verify attribute name
  # note that this only a test of attribute format
  kind = Memc[field]
  firstdig = Memc[field+1]
  if ( (stridx(kind,"silrdSILRD") == 0) ||
       (stridx(firstdig,"0123456789") == 0))
  {
     call error(101,"Invalid attribute")
  }

  # convert name to offset
  ip = 2
  call gctol(Memc[field], ip, lval, 10)
  offset = lval / SZB_CHAR

  call sfree(sp)
end



#
#  Get the datatype of a particular attribute in a QPOE file
#
#
procedure mqp_att_dtype(qp, att, dtype)

pointer	qp			#I QPOE file handle
char	att[ARB]	#I attribute name
int	dtype		#O datatype of attribute

pointer	field		# full attribute name
char	kind		# attribute data type
char	firstdig	# first digit of attribute offset
pointer	sp

int	stridx()
char  	chrlwr()
begin

  call smark(sp)

  call salloc(field, SZ_LINE, TY_CHAR)

  # find full attribute name
  call qp_expandtext(qp, att, Memc[field], SZ_LINE)

  # verify attribute name
  # note that this only a test of attribute format
  kind = chrlwr(Memc[field])
  firstdig = Memc[field+1]
  if ( (stridx(kind,"silrd") == 0) ||
       (stridx(firstdig,"0123456789") == 0))
  {
     call error(101,"Invalid attribute")
  }

  # convert name to datatype
  switch (kind) {
     case 's' : dtype = TY_SHORT
     case 'd' : dtype = TY_DOUBLE
     case 'r' : dtype = TY_REAL
     case 'i' : dtype = TY_INT
     default  : call error(102,"Invalid attribute datatype")
  }

  call sfree(sp)
end


#
#	Read a real value from a QPOE event
#
real procedure qp_gvalr(ev, offset)

pointer	ev		#I event pointer
int		offset	#I field offset

real	rval

pointer	coerce()	# pointer converter

begin
  rval = Memr[coerce(ev+offset, TY_SHORT, TY_REAL)]
  return rval
end


#
#	Read a double value from a QPOE event
#
double procedure qp_gvald(ev, offset)

pointer	ev		#I event pointer
int		offset	#I field offset

double	dval

pointer	coerce()	# pointer converter

begin
  dval = Memd[coerce(ev+offset, TY_SHORT, TY_DOUBLE)]
  return dval
end


#
#	Write a real value into a QPOE event.
#
procedure qp_pvalr(ev, offset, value)

pointer	ev		#I event pointer
int		offset	#I field offset
real	value	#I value to write

pointer	coerce()	# pointer converter

begin
  Memr[coerce(ev+offset, TY_SHORT, TY_REAL)] = value
end

#
#	Write a double value into a QPOE event.
#
procedure qp_pvald(ev, offset, value)

pointer	ev		#I event pointer
int		offset	#I field offset
double	value	#I value to write

pointer	coerce()	# pointer converter

begin
  Memd[coerce(ev+offset, TY_SHORT, TY_DOUBLE)] = value
end


#
#	Write a short value into a QPOE event.
#
procedure qp_pvals(ev, offset, value)

pointer	ev		#I event pointer
int		offset	#I field offset
short	value	#I value to write

pointer	coerce()	# pointer converter

begin
  Mems[coerce(ev+offset, TY_SHORT, TY_SHORT)] = value
end


procedure get_ra_units(cp, ra_hours)

pointer cp
bool    ra_hours

pointer sp
pointer units

begin
        call smark(sp)
        call salloc(units, SZ_LINE, TY_CHAR)

        call tbcigt(cp, TBL_COL_UNITS,  Memc[units],  SZ_LINE)
        if(Memc[units] == 'h' || Memc[units] == 'H')
                ra_hours=true
        else if(Memc[units] == 'd' || Memc[units] == 'D')
                ra_hours=false
        else
           call error(111,"Invalid ephemeris table RA units, must be hours (H/h) or Degrees (D/d)")

	call sfree(sp)
end
