# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_masktimes.x
# 	Author	:  akeith
# 	Date	:  04/04/95
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_masktimes.x -r1.7 04/04/95 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/24/96   tnk	incr MAX_USR_PARS
# 	11/03/95   tnk	changed NR copyright to procedure call; took out verbose statement
# 	09/21/95   mabbott	changed logic to make no output for no valid intervals
# 	09/13/95   tnk	set max table user params
# 	08/14/95   tnk	parm 'zero' to 'zerotimes'; renamed input to mask; added hcpkill
# 	08/10/95   tnk	added Num. Recipes copyright
# 	04/04/95   akeith	date and time created 95/04/04 16:05:03 by akeith
# H-
# 
#

#
# Determine time intervals spent in a region mask
# @(#)t_masktimes.x	1.7 10/24/96 UCB CEA 
#
# Makes use of Numerical Recipes library routines,
# Copyright (C) 1986 Numerical Recipes Software


include <tbset.h>
include <fset.h>
include <imhdr.h>

# time between successive rows in detpos table
define VINTLEN double(1.024)

# accuracy of test for real number equality
define EPS double(1.0e-5)

define MAX_USR_PARS   120       # max. number of user parameters in table

procedure t_masktimes 

int		nd, nx, n, maskext
bool	verbose, zero
double	sublen
pointer	sp, str, pim, tp, cp, nulls, dtime, dx, dy, xs, xe, cs, ce, out, otp,
        str2
int		tbpsta(), fnextn(), imaccess(), clgeti()
double	clgetd()
bool	clgetb(), streq()
pointer	im_pmmap(), tbtopn()


begin
  call smark(sp)

  # put in Numerical Recipes copyright string
  call nr_copyr()

  call salloc(str, SZ_LINE, TY_CHAR)
  call salloc(str2, SZ_LINE, TY_CHAR)
  call salloc(out, SZ_LINE, TY_CHAR)

  verbose = clgetb("verbose")
  if (verbose) call fseti(STDOUT, F_FLUSHNL, YES)

  sublen = clgetd("sublen")
  zero = clgetb("zerotimes")
  maskext = clgeti("maskextension")

  # open pixel list
  call clgstr("mask", Memc[str], SZ_LINE)
  n = fnextn(Memc[str], Memc[str2], SZ_LINE)
  if ((imaccess(Memc[str], READ_ONLY) == NO) || !streq(Memc[str2], "pl")) {
	call strcat(".pl", Memc[str], SZ_LINE)
  }
  pim = im_pmmap(Memc[str], READ_ONLY, 0)

  # fetch detector position arrays
  call clgstr("detpos", Memc[str], SZ_LINE)
  tp = tbtopn(Memc[str], READ_ONLY, 0)
  nd = tbpsta(tp, TBL_NROWS)
  if (nd < 4) call error(987,"detpos table must contain at least 4 rows")
  call malloc(nulls, nd, TY_BOOL)
  call malloc(dtime, nd, TY_DOUBLE)
  call malloc(dx, nd, TY_REAL)
  call malloc(dy, nd, TY_REAL)

  call clgstr("dtimecol", Memc[str], SZ_LINE)
  call tbcfnd(tp, Memc[str], cp, 1)
  if (cp == NULL) call error(123,"detpos time column not found")
  call tbcgtd(tp, cp, Memd[dtime], Memb[nulls], 1, nd)

  call clgstr("dxcol", Memc[str], SZ_LINE)
  call tbcfnd(tp, Memc[str], cp, 1)
  if (cp == NULL) call error(123,"dx column not found")
  call tbcgtr(tp, cp, Memr[dx], Memb[nulls], 1, nd)

  call clgstr("dycol", Memc[str], SZ_LINE)
  call tbcfnd(tp, Memc[str], cp, 1)
  if (cp == NULL) call error(123,"dy column not found")
  call tbcgtr(tp, cp, Memr[dy], Memb[nulls], 1, nd)

  call mfree(nulls, TY_BOOL)

  call clgstr("output", Memc[out], SZ_LINE)

  # I believe this is not needed, so comment out for now.
  #call remove_indef(Memd[dtime], Memd[dx], Memd[dy], nd)

  # Find intervals
  call mask_intervals(Memd[dtime], Memr[dx], Memr[dy], nd, zero, 
					  sublen, pim, maskext, xs, xe, nx)

  # check if there are any intervals to write
  if (nx > 0) {
	# open output table
	otp = tbtopn(Memc[out], NEW_FILE, 0)

	# Write output table
	call tbcdef(otp, cs, "stime", "Seconds", "%g", TY_DOUBLE, 1, 1)
	call tbcdef(otp, ce, "etime", "Seconds", "%g", TY_DOUBLE, 1, 1)

	# set max no. of user params
	call tbpset (otp, TBL_MAXPAR, MAX_USR_PARS)

	call tbtcre(otp)

	#copy header cards from detpos table to output
	call hcpkill(tp, "TABLE", otp, "TABLE", "TIME DETASP?")

	call tbcptd(otp, cs, Memd[xs], 1, nx)
	call tbcptd(otp, ce, Memd[xe], 1, nx)

	call tbtclo(otp)
  }	else {
	if (verbose) {
	  call printf("no intervals match criteria, no output table produced\n")
	}
  }
  
  # Clean up
  call tbtclo(tp)
  call imunmap(pim)
  call mfree(dtime, TY_DOUBLE)
  call mfree(dx, TY_REAL)
  call mfree(dy, TY_REAL)
  call mfree(xs, TY_DOUBLE)
  call mfree(xe, TY_DOUBLE)
  call sfree(sp)
end



procedure mask_intervals(time, x, y, n, zero, sublen, pim, maskext, xs, xe, nx)

double	time[n], sublen
real	x[n], y[n]
bool	zero
pointer	xs, xe, pim
int		n, nx, maskext

int		i, len
bool	inzero, lastinzero, openinterval
double	t

bool	mask_zero()

begin
  len = 100
  call malloc(xs, len, TY_DOUBLE)
  call malloc(xe, len, TY_DOUBLE)

  if (zero) lastinzero = false
  else lastinzero = true

  i = 0
  openinterval = false
  do t = time[1], time[n], sublen {

	#call printf("T: %g\n"); call pargd(t)

	inzero = mask_zero(time, x, y, n, pim, zero, maskext, t)

	#call printf("zero: %b\n"); call pargb(inzero)

	if (!inzero && lastinzero) {
	  if (zero) {
		# end interval
		Memd[xe+i-1] = t
		openinterval = false
 	  }
	  else {
		# start interval
		i = i + 1
		if (i > len) {
		  len = len + 100
		  call realloc(xs, len, TY_DOUBLE)
		  call realloc(xe, len, TY_DOUBLE)
		}
		Memd[xs+i-1] = t
		openinterval = true
	  }
	}
	else if (inzero && !lastinzero) {
	  if (zero) {
		# start interval
		i = i + 1
		if (i > len) {
		  len = len + 100
		  call realloc(xs, len, TY_DOUBLE)
		  call realloc(xe, len, TY_DOUBLE)
		}
		Memd[xs+i-1] = t
		openinterval = true
	  }
	  else {
		# end interval
		Memd[xe+i-1] = t
		openinterval = false
	  }
	}
	lastinzero = inzero
  }
  if (openinterval) Memd[xe+i-1] = t-sublen

  nx = i
end



bool	procedure mask_zero(time, x, y, n, pim, zero, maskext, t)

int		n, maskext
double	t, time[n]
real	x[n], y[n]
bool	zero
pointer	pim

int		i, k, npts, ix, iy
real	frac, dx, dy, dxerr, dyerr

bool	eq_epsilon(), im_pmsne2()

begin
  # find t in the detpos table
  call hunt(time, n, t, i)

  #call printf("  i: %d\n"); call pargi(i)

  # off the table, force a break if in an interval
  if (i <= 0 || i >= n) return !zero

  # in a gap, force a break if in an interval
  if (! eq_epsilon(time[i+1]-time[i], VINTLEN, EPS)) return !zero

  # There is an INDEF detector position before or after this time.  This
  # means we are way off the detector.  Assume mask value is maskext.
  if (IS_INDEFR(x[i]) || IS_INDEFR(x[i+1]) || IS_INDEFR(y[i]) || 
	  IS_INDEFR(y[i+1])) {
	if (maskext == 0)
	  return true
	else
	  return false
  }

  # now we do interpolation over 4 pts, taking care to avoid gaps or INDEFs
  # in between the points
  if (i<2 || i>n-2 || ! eq_epsilon(time[i+2]-time[i-1], 3.0*VINTLEN, EPS) ||
	  IS_INDEFR(x[i-1]) || IS_INDEFR(x[i+2]) || IS_INDEFR(y[i-1]) ||
	  IS_INDEFR(y[i+2])) {

	# We have a gap or INDEF during the nominal interpolation interval
	if (i<2 || ! eq_epsilon(time[i]-time[i-1], VINTLEN, EPS) ||
		IS_INDEFR(x[i-1]) || IS_INDEFR(y[i-1])) {

	  # Point i-1 is unusable

	  if (i>n-2 || ! eq_epsilon(time[i+2]-time[i+1], VINTLEN, EPS) ||
		  IS_INDEFR(x[i+2]) || IS_INDEFR(y[i+2])) {

		# Point i+2 is bad also

		# Case 3: gaps or INDEFs affecting both pts i-1 and i+2 
		# do linear interpolation from i to i+1

		#call printf("  case 3\n")

		frac = (t-time[i])/(time[i+1]-time[i])
		dx = x[i] + frac * (x[i+1]-x[i])
		dxerr = 0.0
		dy = y[i] + frac * (y[i+1]-y[i])
		dyerr = 0.0
		k = 0
		npts = 0
	  }
	  else {

		# Case 1: a gap between pts i-1 & i or i-1 is INDEF

		#call printf("  case 1\n")

		k = i
		npts = 3
		call polint(time[k], x[k], npts, t, dx, dxerr)
		call polint(time[k], y[k], npts, t, dy, dyerr)
	  }
	}
	else {
	  # Must be a problem with point i+2

	  # Case 2: a gap between pts i+2 & i+1 or i+2 is INDEF

	  #call printf("  case 2\n")

	  k = i-1
	  npts = 3
	  call polint(time[k], x[k], npts, t, dx, dxerr)
	  call polint(time[k], y[k], npts, t, dy, dyerr)
	}
  }
  else {

	#call printf("  normal\n")

	k = i-1
	npts = 4
	call polint(time[k], x[k], npts, t, dx, dxerr)
	call polint(time[k], y[k], npts, t, dy, dyerr)
  }

  # At this point we have a valid detector position.

  #call printf("  pos: %.2f %.2f\n"); call pargd(dx); call pargd(dy)

  # Check mask.  Add 1 to convert from detector coords to IRAF pixels.
  ix = nint(dx)+1
  iy = nint(dy)+1

  #call printf("  mask: %d %d\n"); call pargi(ix); call pargi(iy)

  # We are off the mask.  Assume mask extension value.
  if (ix<1 || ix>IM_LEN(pim,1) || iy<1 || iy>IM_LEN(pim,2)) {
	if (maskext == 0)
	  return true
	else
	  return false
  }

  # Actually look up the mask value
  if (! im_pmsne2(pim, ix, ix, iy, iy)) return true

  return false
end
