# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  mapos.x
# 	Author	:  ericco
# 	Date	:  10/21/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) mapos.x -r1.5 10/21/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	12/26/96   tnk	allow hours or degrees on output ra values
# 	01/04/93   ericco	corrected output to correspond with planning software
# 	12/10/92   ericco	added optional distance output
# 	10/21/92   ericco	date and time created 92/10/21 20:14:43 by ericco
# H-
# 
#

# @(#)mapos.x	1.5 12/26/96

include <tbset.h>
include <math.h>
include "ma.h"

# Converts the aspect that is returned by the space craft in to sky coordinates
# This is achived by asuming that the boresight is the z axis of the detector
# The quaternion that represents this is then rotated by the inverse of the
# the quaternion found in the reference dataset. This rotation converts from 
# detector coordinates to ACS cordinates. The resulting quaternion is then
# rotated by the inverse of the aspect quaternion that is given in the 
# st table ( input table ). The result is then converted to sky coordinates
# and output into the output table together with the other fields in the table.

procedure ma_position(mas, num, asp, aq, atc, skip, hours)

pointer	mas[num]		#I array of output tables
int	num			#I number of output tables
pointer	asp			#I aspect table
pointer	aq[4]			#I aspect columns
pointer atc			#I aspect time column
int	skip			#I number of rows to skip
bool    hours			#I ra in hours (true) or degrees (false)

int i				# loop counter
int num_rows			# of rows on the input table

double qfor[4]			# The ACS to detector quaternion
double qasp[4]			# The quaternion repotred by the space craft
double qdet[4]			# The defn of boresight
double qrev[4]			# inverse quaternion
double out[5]			# result vector

double ra			# right acension
double dec			# declination
double time			# aspect time
int	d			# detector id
pointer	ma			# aspect structure

bool	null[4], tnull		# table null values
int	orow			# output row

int tbpsta()
pointer	dist

begin
  orow = 1
  # find num of rows in input table
  num_rows = tbpsta(asp, TBL_NROWS )

  # loop through the aspect table
  do i = 1, num_rows, skip {
    call tbrgtd(asp, aq, qasp, null, 4, i)
    call tbrgtd(asp, atc, time, tnull, 1, i)
    if (tnull || null[1] || null[2] || null[3] || null[4]) ;
    else {
      do d = 1, num {
	# copy the time to the output table
	out[1] = time

	ma = mas[d]
	dist = MA_COLUMN(ma,5)
	call er_qmult(qasp, MA_QFOR(ma,1), qfor)
	call er_qinvert(qfor, qrev)

	# convert detector z into j2000 coordinates
	qdet[1] = 0
	qdet[2] = 0
	qdet[3] = 0
	qdet[4] = 1
	call er_qrotateq (qdet, qrev)

	# convert quaternion into a position
	call er_qtopos (qdet, ra, dec)
	if (ra < 0) ra = ra + TWOPI
	out[2] = ra * 180 / PI
	# convert ra to hrs min and sec
	if (hours)
	   out[2] = out[2]/15.0
	out[3] = dec * 180 / PI

	# convert j2000 z into detector coordinates
	qdet[1] = 0
	qdet[2] = 0
	qdet[3] = 0
	qdet[4] = 1
	call er_qrotateq (qdet, qfor)

	# convert quaternion into a "position"
	call er_qtopos( qdet, ra, dec )
	if (ra <= -PI) ra = ra + TWOPI
	else if (ra > PI) ra = ra - TWOPI
	out[4] = ra * 180 / PI

	# this position is from the +x axis to the j2000 +z axis
	# the convention is to report roll angles from the j2000 +z
	# axis to the detector +x axis.
	out[4] = -out[4]

	if (dist != 0) {
	  # convert j2000 center position into detector coordinates
	  qdet[1] = MA_QCEN(ma,1)
	  qdet[2] = MA_QCEN(ma,2)
	  qdet[3] = MA_QCEN(ma,3)
	  qdet[4] = MA_QCEN(ma,4)
	  call er_qrotateq (qdet, qfor)

	  # convert quaternion into a "position"
	  call er_qtopos( qdet, ra, dec )

	  # measure distance to z axis (boresight)
	  out[5] = 90 - dec * 180 / PI

	  call tbrptd (MA_TP(ma), MA_COLUMN(ma,1), out, 5, orow)
	} else {
	  call tbrptd (MA_TP(ma), MA_COLUMN(ma,1), out, 4, orow)
	}
      }
      orow = orow + 1
    }
  }
end
