# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_mkaspect.x
# 	Author	:  ericco
# 	Date	:  10/21/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_mkaspect.x -r1.19 10/21/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	12/27/96   tnk	changed pargs to pargstr in ra_units error code
# 	12/26/96   tnk	ra units now selectable
# 	12/26/96   tnk	made colnames for aspect time,w,x,y,z into parameters
# 	10/24/96   tnk	incr MAX_USR_PARS
# 	10/21/96   tnk	search for TIMEDEL, then DELTIME; put out TIMEDEL
# 	09/26/95   tnk	added et_close(et) call
# 	09/13/95   tnk	set max table user params
# 	08/08/94   tnk	renamed time column to TableTime
# 	06/03/94   tnk	changed tbhcal to hcpkill
# 	05/03/93   akeith	fixed seg viol problem w/ et == NULL
# 	04/30/93   gareth	added refdata
# 	04/29/93   gareth	hard coded aspect cols
# 	04/29/93   akeith	Now doesn't require DELTIME header in table.
# 	01/04/93   ericco	moved header copy before new header values
# 	12/10/92   ericco	added optional distance output
# 	10/21/92   ericco	date and time created 92/10/21 20:14:32 by ericco
# H-
# 
#

# Revised: @(#)t_mkaspect.x	1.19 12/27/96

include <math.h>
include <tbset.h>
include "ma.h"

define MAX_USR_PARS   120       # max. number of user parameters in table

define PRIMARY_DELTIME_HDR    "TIMEDEL"
define ALTERNATE_DELTIME_HDR  "DELTIME"    


# Creates aspect information from telescope pointing information

procedure t_mkaspect() #E

pointer		sp		# stack marker
pointer		table		# table
pointer		detectors	# detector list
pointer		det		# detector
pointer		output		# output table name
pointer		out		# output table
pointer		asptab		# aspect table name
pointer		aspw		# aspect column names
pointer		aspx
pointer		aspy
pointer		aspz
pointer		aspt
pointer		asp		# aspect table
pointer		atc		# aspect time column
real		deltime		# expected table time interval
pointer		aq[4]		# aspect quaternion
pointer		et		# euv table
int		num		# number of detectors
pointer		ma, mas		# mkaspect states
pointer		edd		# detector information
double		w, x, y, z	# quaternion
pointer		tp		# st table pointer
int		detnum		# detector number
int		skip		# skip rate
real		delta_t		# expected table time interval
real		t0, t1		# time values for delta_t

int		clgeti()
pointer		tbtopn()
int		clplen()
int		clgfil()
pointer		clpopnu()
real		tbhgtr()
bool		distp       
bool		clgetb()
double		ra, dec
pointer 	ra_units	# units for RA
double		clgetd()
pointer		buf
bool		hours

begin
  call smark(sp)
  call salloc(det, SZ_LINE, TY_CHAR)
  call salloc(out, SZ_LINE, TY_CHAR)
  call salloc(buf, SZ_LINE, TY_CHAR)

  # collect user input
  call eu_clstr(asptab, "input")
  call eu_clstr(output, "output")
  call eu_clstr(table, "refdata")
  # get the column names
  call eu_clstr(aspt, "time")
  call eu_clstr(aspw, "aspectw")
  call eu_clstr(aspx, "aspectx")
  call eu_clstr(aspy, "aspecty")
  call eu_clstr(aspz, "aspectz")
  call eu_clstr(ra_units, "raunits")

  if(Memc[ra_units] == 'h' || Memc[ra_units] == 'H')
     hours = true
  else if(Memc[ra_units] == 'd' || Memc[ra_units] == 'D')
     hours = false
  else
  {
     call eprintf("invalid ra_units: %s\n")
     call pargstr(Memc[ra_units])
     call error(1,"Ra units must be hours (H/h) or Degrees (D/d)")
  }

  skip = clgeti("skip")
  distp = clgetb("distance")
  if (distp) {

     ra = clgetd("ra")

     if (hours)
	ra = 15.0 * ra * PI / 180.0
     else
	ra = ra * PI / 180.0

     dec = clgetd("dec") * PI / 180
  }

  # open aspect table information
  asp = tbtopn(Memc[asptab], READ_ONLY, 0)

  call tbcfnd(asp, Memc[aspw], aq[1], 1)
    if(aq[1] == NULL) call eu_error(1, "Column not found, ", Memc[aspw])
  call tbcfnd(asp, Memc[aspx], aq[2], 1)
    if(aq[1] == NULL) call eu_error(1, "Column not found, ", Memc[aspx])
  call tbcfnd(asp, Memc[aspy], aq[3], 1)
    if(aq[1] == NULL) call eu_error(1, "Column not found, ", Memc[aspy])
  call tbcfnd(asp, Memc[aspz], aq[4], 1)
    if(aq[1] == NULL) call eu_error(1, "Column not found, ", Memc[aspz])
  call tbcfnd(asp, Memc[aspt], atc, 1)
    if(atc == NULL) call eu_error(1, "Column not found, ", Memc[aspt])


  iferr (delta_t = tbhgtr(asp, PRIMARY_DELTIME_HDR))
  {
     iferr (delta_t = tbhgtr(asp, ALTERNATE_DELTIME_HDR))
     {
        call tbegtr(asp, atc, 1, t0)
        call tbegtr(asp, atc, 2, t1)
        delta_t = t1-t0
     }
  }

  deltime = skip * delta_t

  # also get detector information
  call et_open(et, Memc[table])
  if(et == NULL) call eu_error(1, "Unable to find table, ", Memc[table])

  detectors = clpopnu("detectors")

  num = clplen(detectors)
  call salloc(mas, num, TY_INT)

  # create a structure for each detector.
  # create an output table too.
  detnum = 0
  while (clgfil(detectors, Memc[det], SZ_LINE) != EOF) {
    detnum = detnum + 1
    call ma_salloc(ma)
    Memi[mas+detnum-1] = ma

    call er_qfrompos (ra, dec, MA_QCEN(ma,1))

    call edd_new(edd, et, Memc[det])
    call edd_boresight(edd, w, x, y, z)
    MA_QFOR(ma,1) = w
    MA_QFOR(ma,2) = x
    MA_QFOR(ma,3) = y
    MA_QFOR(ma,4) = z
    call er_qinvert(MA_QFOR(ma,1), MA_QREV(ma,1))

    call sprintf(Memc[out], ARB, "%s%s")
    call pargstr(Memc[output])
    call pargstr(Memc[det])

    tp = tbtopn(Memc[out], NEW_FILE, 0)
    MA_TP(ma) = tp
    call tbcdef(tp, MA_COLUMN(ma,1), "TableTime", "Seconds", "%15.4f",
		TY_DOUBLE, 1, 1)

    if(hours)
    {
       call tbcdef(tp, MA_COLUMN(ma,2), "RA", "Hours", "%12.1h",
		TY_DOUBLE, 1, 1)
       call tbcdef(tp, MA_COLUMN(ma,3), "DEC", "Degrees", "%12.1h",
		TY_DOUBLE, 1, 1)
    }
    else
    {
       call tbcdef(tp, MA_COLUMN(ma,2), "RA", "Degrees", "%10.5f",
		TY_DOUBLE, 1, 1)
       call tbcdef(tp, MA_COLUMN(ma,3), "DEC", "Degrees", "%10.5f",
		TY_DOUBLE, 1, 1)
    }


    call tbcdef(tp, MA_COLUMN(ma,4), "ROLL", "Degrees", "%10.5f",
		TY_DOUBLE, 1, 1)

    if (distp) {
      call tbcdef(tp, MA_COLUMN(ma,5), "DISTANCE", "Degrees", "%10.5f",
		  TY_DOUBLE, 1, 1)
    } else {
      MA_COLUMN(ma,5) = 0
    }

    # set max no. of user params
    call tbpset (tp, TBL_MAXPAR, MAX_USR_PARS)

    call tbtcre(tp)

    # copy appropriate header cards
    call hcpkill(asp, "TABLE", tp, "TABLE", "")

    call tbhadt(tp, "TIME", "TableTime")
    call tbhadt(tp, "DETNAM", Memc[det])
    call tbhadd(tp, "DETASPW", w)
    call tbhadd(tp, "DETASPX", x)
    call tbhadd(tp, "DETASPY", y)
    call tbhadd(tp, "DETASPZ", z)

    call sprintf(Memc[buf], SZ_LINE, "Created from %s for detector %s")
    call pargstr(Memc[asptab])
    call pargstr(Memc[det])
    call tbhadt(tp, "HISTORY", Memc[buf])

    if (deltime != INDEFR) call tbhaddr(tp, PRIMARY_DELTIME_HDR, deltime)

    call edd_close(edd)
  }

  # Calulate the values for the output table
  call ma_position(Memi[mas], num, asp, aq, atc, skip, hours)

  # Close the tables
  do detnum = 1, num {
    ma = Memi[mas+detnum-1]
    call tbtclo(MA_TP(ma))
  }
  call et_close(et)
  call tbtclo(asp)

  call sfree(sp)
end
