# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_mkprofile.x
# 	Author	:  akeith
# 	Date	:  04/04/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_mkprofile.x -r1.9 04/04/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	07/12/94   akeith	added prf_interp()
# 	07/07/94   akeith	new and totally improved
# 	05/03/94   akeith	put in some error checking on the parameters
# 	04/26/94   akeith	added hcpykill
# 	04/04/94   akeith	date and time created 94/04/04 20:20:33 by akeith
# H-
# 
#


#	Version: @(#)t_mkprofile.x	1.9 7/12/94 UCB CEA EGO

#
# Construct spatial profiles from EUVE spectral images.
#

include	"prf.h"
include <dtset.h>
include <imhdr.h>
include <imio.h>

procedure	t_mkprofile()

int	ip			# string position
int	ncol, nline, tmp	# number of columns and lines
int	band			# spectral band
int	peak			# peak pix
int	fwhm			# full width half max
int	prof			# current profile index

pointer sp			# stack pointer
pointer	str			# general string
pointer	prf			# mkprofile structure
pointer	dto			# data table object
pointer	input			# input  file name
pointer	output			# output file name
pointer	cv			# curve fit structure

int	clgeti()		# cl get int
int	imgeti()		# get header param
int	imhdfnd()		# image header find
int	strcmp()		# string compare
int	stat, eu_grange()	# convert string to range

real	clgetr()		# cl get real
bool	clgetb()		# cl get bool
bool	prf_gpeak()		# find the profile peak

pointer	gopen()			# gio structure
pointer	gt_init()		# gtools structure
pointer	dt_open()		# data table open
pointer	prf_open()		# mkprofile structure open
pointer	prf_fit_bkg()		# mkprofile fit background
pointer in, out, immap()	# in, out image structure

begin

	call smark(sp)

	prf = prf_open()

#   Read some parameters.

	PRF_INTER(prf) = clgetb("interactive")
	PRF_TRIM(prf) = clgetb("trim")

	PRF_NFWHM(prf) = clgeti("nfwhm")
	if(PRF_NFWHM(prf) < 1) {
		call sfree(sp)
		call prf_close(prf)
		call error(1, "nfwhm parameter must be greater then zero.")
	}

	PRF_THRESH(prf) = clgetr("threshold")

#    Open input image.

	call eu_clstr(input, "input")
	iferr(in = immap(Memc[input], READ_ONLY, 0)) {
		call sfree(sp)
		call prf_close(prf)
		call error(1, "mkprofile: Unable to access input image.")
	}
	ncol = IM_LEN(in, 1)
	nline = IM_LEN(in, 2)

#    Open output image.

	call eu_clstr(output, "output")
	iferr(out = immap(Memc[output], NEW_IMAGE, NULL)) {
		call sfree(sp)
		call prf_close(prf)
		call error(1, "mkprofile: Unable to access output image.")
	}

#    Find the spectral band.

	band = imhdfnd(in, "BANDID*", "spectrum")
	if(band < 1) band = 1

	call prf_gband(in, band)

#    Get the dispersion axis.

	PRF_AXIS(prf) = imgeti(in, "DISPAXIS")
	if(PRF_AXIS(prf) == 2) {
		tmp = ncol
		ncol = nline
		nline = tmp
	}

#    Get the profile bounds.

	call prf_gbounds(prf, ncol)

#    Get the aperture bounds.

	call eu_clstr(str, "aperture")
	if(Memc[str] == EOS || strcmp(Memc[str], "*") == 0) {
		PRF_LO(prf) = 1
		PRF_HI(prf) = ncol
	} else {
		ip = 1
		stat = eu_grange(Memc[str], ip, PRF_LO(prf), PRF_HI(prf))
	}

#    Open the graphics.

	if(PRF_INTER(prf)) {
		call eu_clstr(str, "graphics")
		PRF_GP(prf) = gopen (Memc[str], NEW_FILE, STDGRAPH)
		PRF_GT(prf) = gt_init()
	}

	call ic_open(PRF_IC(prf))

	PRF_DT(prf) = dt_open()


#    Now make profiles.

	do prof = 1, PRF_NP(prf) {

		call prf_graw(prf, prof, in, band, dto)

		if( !prf_gpeak(prf, dto, peak, fwhm) ) {
			call prf_sindef(dto)
			next
		}

		cv = prf_fit_bkg(prf, dto, prof, peak, fwhm)

		call prf_clean(prf, dto, cv, peak, fwhm)

		call cvfree(cv)

	}


#    Open output image, copy header from input image.

	call hcpkill(in, "IMAGE", out, "IMAGE", "")

	IM_PIXTYPE(out) = TY_REAL
	IM_NDIM(out) = 2
	IM_LEN(out, 1) = PRF_NP(prf)
	IM_LEN(out, 2) = PRF_HI(prf) - PRF_LO(prf) + 1

	call prf_interp(prf)

	call prf_write(prf, out)

	call prf_mwcs(prf, in, out, band, Memc[input])

	call imunmap(in)
	call imunmap(out)

	call prf_close(prf)

	call sfree(sp)

end


#
# prf_write()
#

procedure	prf_write(prf, out)

pointer	prf			#I
pointer	out			#I

int	prof			#
int	len			#

pointer	sp			#
pointer	name			#

pointer	dto, dt_dto()		#
pointer	buf, imps2r()

begin

	call smark(sp)

	len = PRF_HI(prf) - PRF_LO(prf) + 1

	do prof = 1, PRF_NP(prf) {
		call prf_mkname(prof, name)
		dto = dt_dto(PRF_DT(prf), Memc[name])

		buf = imps2r(out, prof, prof, 1, len)
		call dto_gbufr(dto, Memr[buf], len)	
	}

	call sfree(sp)

end


#
# prf_sindef()
#

procedure	prf_sindef(dto)

pointer	dto			#I

int	i
int	len, dto_stati()
real	indef_val

begin

	indef_val = INDEFR
	len = dto_stati(dto, DTO_NELMS_PARM)
	call dto_seek(dto, 1)
	do i = 1, len
		call dto_pbufr(dto, indef_val, 1)
	
end

