include <iraf.h>

include	<ctype.h>
include	<lexnum.h>

include	<fset.h>
include	<math.h>

define	YYMAXDEPTH	150		# length of parser stack

#task	qcalc = t_qcalc

# Operand Structure (parser stack)
define	YYOPLEN		10		# size of operand structure
define	OPTYPE		Meml[P2L($1)+0]	# Operand type

define	OPVALI		Meml[P2L($1)+1]	# letter

define	OPVALW		Memd[P2D($1)+1]	# W component
define	OPVALX		Memd[P2D($1)+2]	# X component
define	OPVALY		Memd[P2D($1)+3]	# Y component
define	OPVALZ		Memd[P2D($1)+4]	# Z component

define	CONST		257
define	LETTER		258
define	YYEOF		259
define	UMINUS		260
define	yyclearin	yychar = -1
define	yyerrok		yyerrflag = 0
define	YYMOVE		call amovi (Memi[$1], Memi[$2], YYOPLEN)
define	YYERRCODE	256

# line 141 "t_qcalc.y"



# QCALC -- Main routine for the quaternion calculator.

procedure t_qcalc()

int	status
int	yyparse()
int	yylex()
extern	yylex()
bool	debug
bool	clgetb()
pointer	sp, in
int	strlen()
int	open()
int	stropen()
bool	streq()
bool	filep
int	fd
int	len

begin
  call smark(sp)
  call salloc(in, SZ_LINE, TY_CHAR)
  debug = FALSE
  call clgstr("input", Memc[in], SZ_LINE)
  filep = clgetb("file")
  if (filep) {
    if (streq(Memc[in], "STDIN")) {
      fd = open(Memc[in], READ_ONLY, TEXT_FILE)
    } else {
      fd = STDIN
    }
  } else {
    len = strlen(Memc[in])
    if (len < 1 || Memc[in+len-1] != ';') {
      Memc[in+len] = ';'
      len = len + 1
      Memc[in+len] = EOS
    }
    fd = stropen(Memc[in], strlen(Memc[in]), READ_ONLY)
  }

  repeat {
    status = yyparse (fd, debug, yylex)
    if (status == ERR)
      call eprintf ("syntax error")
    } until (status == EOF)

  if (filep) {
    if (fd != STDIN) call close(fd)
  } else {
    call strclose(fd)
  }

  call sfree(sp)
end

procedure binop (a, b, c, operation)

pointer	a, b, c			# c = a op b
int	operation		# i.e., '*', '/', etc.
double	qa[4], qb[4], qc[4]

begin
  qa[1] = OPVALW(a)
  qa[2] = OPVALX(a)
  qa[3] = OPVALY(a)
  qa[4] = OPVALZ(a)

  qb[1] = OPVALW(b)
  qb[2] = OPVALX(b)
  qb[3] = OPVALY(b)
  qb[4] = OPVALZ(b)

  switch (operation) {
    case '+':
      qc[1] = qa[1] + qb[1]
      qc[2] = qa[2] + qb[2]
      qc[3] = qa[3] + qb[3]
      qc[4] = qa[4] + qb[4]
    case '-':
      qc[1] = qa[1] - qb[1]
      qc[2] = qa[2] - qb[2]
      qc[3] = qa[3] - qb[3]
      qc[4] = qa[4] - qb[4]
    case '*':
      call er_qmult(qa, qb, qc)
    case '%':
      call er_qmkrotate(qc, qa, qb)
    case '@':
      qc[1] = qb[1]
      qc[2] = qb[2]
      qc[3] = qb[3]
      qc[4] = qb[4]
      call er_qrotateq(qc, qa)
    case '/':
      call er_qinvert(qb, qb)
      call er_qmult(qa, qb, qc)
    default:
      call error (1, "unknown binary operator")
  }

  OPVALW(c) = qc[1]
  OPVALX(c) = qc[2]
  OPVALY(c) = qc[3]
  OPVALZ(c) = qc[4]
end


procedure unop (a, b, operation)

pointer	a, b
int	operation
double	qa[4], qb[4], qt[4]
double	sqrt()

begin
  qa[1] = OPVALW(a)
  qa[2] = OPVALX(a)
  qa[3] = OPVALY(a)
  qa[4] = OPVALZ(a)

	switch (operation) {
	case '-':
	  call er_qinvert(qa, qb)
	case '|':
	  call er_qinvert(qa, qt)
	  call er_qmult(qa, qt, qb)
	  qb[1] = sqrt(qb[1])
	default:
	  call error (2, "unknown unary operator")
	}

  OPVALW(b) = qb[1]
  OPVALX(b) = qb[2]
  OPVALY(b) = qb[3]
  OPVALZ(b) = qb[4]
end


define	MAXREG		('z'-'a'+1)

procedure getreg (regchar, op)

int	regchar
pointer	op

bool	store
real	regbuf[MAXREG*YYOPLEN]
int	reg, offset

begin
	store = false
	goto 10

entry	putreg (regchar, op)
	store = true

	# Compute offset into storage.  Structures are stored in buffer
	# by a binary copy, knowing only the length of the structure.
10	if (IS_UPPER(regchar))
	    reg = regchar - 'A' + 1
	else
	    reg = regchar - 'a' + 1
	reg = max(1, min(MAXREG, reg))
	offset = (reg-1) * YYOPLEN + 1

	# Copy the operand structure either in or out.
	if (store)
	    call amovr (Memr[op], regbuf[offset], YYOPLEN)
	else
	    call amovr (regbuf[offset], Memr[op], YYOPLEN)
end


procedure	qc_dechour (a, b)
pointer		a, b

double	ra, dec
double	sin(), cos()
begin
  ra = OPVALW(a) * PI / 12
  dec = OPVALX(a) * PI / 180
  OPVALW(b) = 0
  OPVALX(b) = cos(ra)*cos(dec)
  OPVALY(b) = sin(ra)*cos(dec)
  OPVALZ(b) = sin(dec)
end

procedure	qc_decdeg (a, b)
pointer		a, b

double	ra, dec
double	sin(), cos()
begin
  ra = OPVALW(a) * PI / 180
  dec = OPVALX(a) * PI / 180
  OPVALW(b) = 0
  OPVALX(b) = cos(ra)*cos(dec)
  OPVALY(b) = sin(ra)*cos(dec)
  OPVALZ(b) = sin(dec)
end

procedure	qc_position (a, form)
pointer	a
int	form

double	ra, dec
double	atan2(), sqrt()

begin
    if (OPVALX(a) == 0 && OPVALY(a) == 0)
      ra = 0
    else
      ra = atan2(OPVALY(a), OPVALX(a))
    if (ra < 0) ra = ra + TWOPI

    dec = atan2(OPVALZ(a), sqrt(OPVALX(a)**2+OPVALY(a)**2))

    if (form == 0) {
      call printf("%g %g\n")
      call pargd(ra * 180 / PI)
      call pargd(dec * 180 / PI)
    } else {
      call printf("%h %h\n")
      call pargd(ra * 12 / PI)
      call pargd(dec * 180 / PI)
    }
end

# YYLEX -- Lexical input routine.  Return next token from the input
# stream.  Recognized tokens are CONST (numeric constants), LETTER,
# and the operator characters.

int procedure yylex (fd, yylval)

int	fd
pointer	yylval
char	ch, lbuf[SZ_LINE]
int	ip, junk
double	dval
int	getline(), gctod()
data	ip /0/
int	fstati()

begin
	# Fetch a nonempty input line, or advance to start of next token
	# if within a line.  Newline is a token.
	repeat {
	    if (ip <= 0 || lbuf[ip] == EOS) {
	        if (fd == STDIN && fstati(fd, F_REDIR) == NO	&&
		    fstati(STDOUT, F_REDIR) == NO) {
		  call printf("qcalc> ")
		  call flush(STDOUT)
		}
		if (getline (fd, lbuf) == EOF) {
		  if (fd == STDIN && fstati(fd, F_REDIR) == NO	&&
		      fstati(STDOUT, F_REDIR) == NO) {
		    call printf("\n")
		    call flush(STDOUT)
		  }
		  ip = 0
		  return (YYEOF)
		} else
		    ip = 1
	    }
	    while (IS_WHITE (lbuf[ip]))
		ip = ip + 1
	} until (lbuf[ip] != EOS)

	# Determine type of token.  If numeric constant, convert to binary
	# and return value in op structure (yylval).  If letter (register
	# variable) return value and advance input one char.  If any other
	# character, return char itself as the token, and advance input one
	# character.

	if ( !(IS_DIGIT(lbuf[ip]) || (lbuf[ip] == '.')) ) {
	    ch = lbuf[ip]
	    ip = ip + 1
	    if (IS_ALPHA (ch)) {
		OPTYPE(yylval) = LETTER
		OPVALI(yylval) = ch
		return (LETTER)
	    } else {
		OPTYPE(yylval) = ch
		return (OPTYPE(yylval))
	    }
	} else {
	    OPTYPE(yylval) = CONST
	    junk = gctod (lbuf, ip, dval)
	    OPVALW(yylval) = dval

	    junk = gctod (lbuf, ip, dval)
	    if (junk != 0)	OPVALX(yylval) = dval
	    else		OPVALX(yylval) = 0

	    junk = gctod (lbuf, ip, dval)
	    if (junk != 0)	OPVALY(yylval) = dval
	    else		OPVALY(yylval) = 0

	    junk = gctod (lbuf, ip, dval)
	    if (junk != 0)	OPVALZ(yylval) = dval
	    else		OPVALZ(yylval) = 0


	    return (CONST)
        }
end
define	YYNPROD		29
define	YYLAST		252
# Copyright(c) 1986 Association of Universities for Research in Astronomy Inc.

# Parser for yacc output, translated to the IRAF SPP language.  The contents
# of this file form the bulk of the source of the parser produced by Yacc.
# Yacc recognizes several macros in the yaccpar input source and replaces
# them as follows:
#	A	user suppled "global" definitions and declarations
# 	B	parser tables
# 	C	user supplied actions (reductions)
# The remainder of the yaccpar code is not changed.

define	yystack_	10		# statement labels for gotos
define	yynewstate_	20
define	yydefault_	30
define	yyerrlab_	40
define	yyabort_	50

define	YYFLAG		(-1000)		# defs used in user actions
define	YYERROR		goto yyerrlab_
define	YYACCEPT	return (OK)
define	YYABORT		return (ERR)


# YYPARSE -- Parse the input stream, returning OK if the source is
# syntactically acceptable (i.e., if compilation is successful),
# otherwise ERR.  The parameters YYMAXDEPTH and YYOPLEN must be
# supplied by the caller in the %{ ... %} section of the Yacc source.
# The token value stack is a dynamically allocated array of operand
# structures, with the length and makeup of the operand structure being
# application dependent.

int procedure yyparse (fd, yydebug, yylex)

int	fd			# stream to be parsed
bool	yydebug			# print debugging information?
int	yylex()			# user-supplied lexical input function
extern	yylex()

short	yys[YYMAXDEPTH]		# parser stack -- stacks tokens
pointer	yyv			# pointer to token value stack
pointer	yyval			# value returned by action
pointer	yylval			# value of token
int	yyps			# token stack pointer
pointer	yypv			# value stack pointer
int	yychar			# current input token number
int	yyerrflag		# error recovery flag
int	yynerrs			# number of errors

short	yyj, yym		# internal variables
pointer	yysp, yypvt
short	yystate, yyn
int	yyxi, i
errchk	salloc, yylex

short	yyexca[6]
data	(yyexca(i),i=  1,  6)	/  -1,   1,   0,  -1,  -2,   0/
short	yyact[252]
data	(yyact(i),i=  1,  8)	/   7,  46,  19,  10,   2,   6,  46,  10/
data	(yyact(i),i=  9, 16)	/  11,  46,  31,  25,  11,  46,  16,  25/
data	(yyact(i),i= 17, 24)	/  23,  21,  46,  22,  25,  26,  18,  14/
data	(yyact(i),i= 25, 32)	/  46,  23,  21,  14,  22,   3,  26,  10/
data	(yyact(i),i= 33, 40)	/   1,   0,   0,  25,  11,  27,  24,   0/
data	(yyact(i),i= 41, 48)	/  23,  28,  24,   0,  66,  26,   0,  24/
data	(yyact(i),i= 49, 56)	/   0,  29,  30,  14,  32,   0,   0,   0/
data	(yyact(i),i= 57, 64)	/  36,  37,   0,   0,   0,   8,  24,  38/
data	(yyact(i),i= 65, 72)	/  39,  40,  41,  42,  43,  44,  67,  17/
data	(yyact(i),i= 73, 80)	/  69,   0,   0,   0,   0,  51,   0,   0/
data	(yyact(i),i= 81, 88)	/   0,   0,   0,   0,   0,   0,  15,  12/
data	(yyact(i),i= 89, 96)	/   0,   0,  15,  12,  61,   0,  63,  64/
data	(yyact(i),i= 97,104)	/   5,   0,   0,   0,   0,   0,  65,   0/
data	(yyact(i),i=105,112)	/   0,   0,   0,   0,  33,  35,   0,  70/
data	(yyact(i),i=113,120)	/   0,   0,  15,  12,   0,   0,   0,   0/
data	(yyact(i),i=121,128)	/   0,  68,   0,  62,   0,  45,  47,  48/
data	(yyact(i),i=129,136)	/  49,  50,   0,   0,   0,  52,  53,  54/
data	(yyact(i),i=137,144)	/  55,  56,  57,  58,  59,  60,   0,   0/
data	(yyact(i),i=145,152)	/   0,   0,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=153,160)	/   0,   0,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=161,168)	/   0,   0,   0,   0,   0,   0,   0,  71/
data	(yyact(i),i=169,176)	/   0,   0,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=177,184)	/   0,   0,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=185,192)	/   0,   0,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=193,200)	/   0,   0,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=201,208)	/   0,   0,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=209,216)	/   0,   0,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=217,224)	/   0,   0,   0,   0,  13,  34,   0,   4/
data	(yyact(i),i=225,232)	/  13,   9,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=233,240)	/   0,   0,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=241,248)	/   0,   0,   0,   0,   0,   0,   0,   0/
data	(yyact(i),i=249,252)	/  13,  34,   0,  20/
short	yypact[72]
data	(yypact(i),i=  1,  8)	/-255, -33,-1000,  12,  -8, -26,-1000,-1000/
data	(yypact(i),i=  9, 16)	/-1000, -51,-1000, -37, -37,-1000,-1000,-1000/
data	(yypact(i),i= 17, 24)	/-1000,-1000,-1000,-1000,-1000,-1000,-1000,-1000/
data	(yypact(i),i= 25, 32)	/-1000,-1000,-1000,-1000,  -9,  -9,  -9, -37/
data	(yypact(i),i= 33, 40)	/  -9, -22,-1000, -26,  -9,  -9,  -9,  -9/
data	(yypact(i),i= 41, 48)	/  -9,  -9,  -9,  -9,  -9, -26,-1000, -26/
data	(yypact(i),i= 49, 56)	/ -26, -26, -26,  -1, -26, -26,  -2,  -2/
data	(yypact(i),i= 57, 64)	/ -22,-1000,-1000, -22, -26,   3,-1000,   8/
data	(yypact(i),i= 65, 72)	/  -4,  14,-1000,-1000,-1000,-1000,  -9, -17/
short	yypgo[6]
data	(yypgo(i),i=  1,  6)	/   0,  32,  29,  14,  96,  41/
short	yyr1[29]
data	(yyr1(i),i=  1,  8)	/   0,   1,   1,   1,   1,   1,   2,   2/
data	(yyr1(i),i=  9, 16)	/   2,   2,   2,   4,   4,   4,   4,   4/
data	(yyr1(i),i= 17, 24)	/   4,   4,   4,   4,   4,   4,   4,   4/
data	(yyr1(i),i= 25, 29)	/   4,   3,   3,   5,   5/
short	yyr2[29]
data	(yyr2(i),i=  1,  8)	/   0,   0,   3,   1,   3,   3,   1,   3/
data	(yyr2(i),i=  9, 16)	/   3,   3,   3,   5,   4,   4,   4,   4/
data	(yyr2(i),i= 17, 24)	/   4,   4,   2,   8,   4,   1,   1,   5/
data	(yyr2(i),i= 25, 29)	/   5,   1,   1,   0,   2/
short	yychk[72]
data	(yychk(i),i=  1,  8)	/-1000,  -1, 259,  -2, 256,  -4,  38,  33/
data	(yychk(i),i=  9, 16)	/  94, 258,  40,  45, 124, 257,  60, 123/
data	(yychk(i),i= 17, 24)	/  -3,  59,  10,  10, 259,  43,  45,  42/
data	(yychk(i),i= 25, 32)	/  64,  37,  47,  63,  -5,  -5,  -5,  61/
data	(yychk(i),i= 33, 40)	/  -5,  -4, 258,  -4,  -5,  -5,  -5,  -5/
data	(yychk(i),i= 41, 48)	/  -5,  -5,  -5,  -5,  -5,  -4,  10,  -4/
data	(yychk(i),i= 49, 56)	/  -4,  -4,  -4,  -5,  -4,  -4,  -4,  -4/
data	(yychk(i),i= 57, 64)	/  -4,  -4,  -4,  -4,  -4,  -5, 124,  -5/
data	(yychk(i),i= 65, 72)	/  -5,  -5,  41,  62, 125,  58,  -5,  -4/
short	yydef[72]
data	(yydef(i),i=  1,  8)	/   1,  -2,   3,   0,   0,   6,  27,  27/
data	(yydef(i),i=  9, 16)	/  27,  21,  27,   0,   0,  22,  27,  27/
data	(yydef(i),i= 17, 24)	/   2,  25,  26,   4,   5,  27,  27,  27/
data	(yydef(i),i= 25, 32)	/  27,  27,  27,  27,   0,   0,   0,   0/
data	(yydef(i),i= 33, 40)	/   0,  18,  21,  27,   0,   0,   0,   0/
data	(yydef(i),i= 41, 48)	/   0,   0,   0,   0,   0,   7,  28,   8/
data	(yydef(i),i= 49, 56)	/   9,  10,  27,   0,  27,  27,  12,  13/
data	(yydef(i),i= 57, 64)	/  14,  15,  16,  17,  27,   0,  20,   0/
data	(yydef(i),i= 65, 72)	/   0,   0,  11,  23,  24,  27,   0,  19/

begin
	call smark (yysp)
	call salloc (yyv, (YYMAXDEPTH+2) * YYOPLEN, TY_STRUCT)

	# Initialization.  The first element of the dynamically allocated
	# token value stack (yyv) is used for yyval, the second for yylval,
	# and the actual stack starts with the third element.

	yystate = 0
	yychar = -1
	yynerrs = 0
	yyerrflag = 0
	yyps = 0
	yyval = yyv
	yylval = yyv + YYOPLEN
	yypv = yylval

yystack_
	# SHIFT -- Put a state and value onto the stack.  The token and
	# value stacks are logically the same stack, implemented as two
	# separate arrays.

	if (yydebug) {
	    call printf ("state %d, char 0%o\n")
		call pargs (yystate)
		call pargi (yychar)
	}
	yyps = yyps + 1
	yypv = yypv + YYOPLEN
	if (yyps > YYMAXDEPTH) {
	    call sfree (yysp)
	    call eprintf ("yacc stack overflow\n")
	    return (ERR)
	}
	yys[yyps] = yystate
	YYMOVE (yyval, yypv)

yynewstate_
	# Process the new state.
	yyn = yypact[yystate+1]

	if (yyn <= YYFLAG)
	    goto yydefault_			# simple state

	# The variable "yychar" is the lookahead token.
	if (yychar < 0) {
	    yychar = yylex (fd, yylval)
	    if (yychar < 0)
		yychar = 0
	}
	yyn = yyn + yychar
	if (yyn < 0 || yyn >= YYLAST)
	    goto yydefault_

	yyn = yyact[yyn+1]
	if (yychk[yyn+1] == yychar) {		# valid shift
	    yychar = -1
	    YYMOVE (yylval, yyval)
	    yystate = yyn
	    if (yyerrflag > 0)
		yyerrflag = yyerrflag - 1
	    goto yystack_
	}

yydefault_
	# Default state action.

	yyn = yydef[yystate+1]
	if (yyn == -2) {
	    if (yychar < 0) {
		yychar = yylex (fd, yylval)
		if (yychar < 0)
		    yychar = 0
	    }

	    # Look through exception table.
	    yyxi = 1
	    while ((yyexca[yyxi] != (-1)) || (yyexca[yyxi+1] != yystate))
		yyxi = yyxi + 2
	    for (yyxi=yyxi+2;  yyexca[yyxi] >= 0;  yyxi=yyxi+2) {
		if (yyexca[yyxi] == yychar)
		    break
	    }

	    yyn = yyexca[yyxi+1]
	    if (yyn < 0) {
		call sfree (yysp)
		return (OK)			# ACCEPT -- all done
	    }
	}


	# SYNTAX ERROR -- resume parsing if possible.

	if (yyn == 0) {
	    switch (yyerrflag) {
	    case 0, 1, 2:
		if (yyerrflag == 0) {		# brand new error
		    call eprintf ("syntax error\n")
yyerrlab_
		    yynerrs = yynerrs + 1
		    # fall through...
		}

	    # case 1:
	    # case 2: incompletely recovered error ... try again
		yyerrflag = 3

		# Find a state where "error" is a legal shift action.
		while (yyps >= 1) {
		    yyn = yypact[yys[yyps]+1] + YYERRCODE
		    if ((yyn >= 0) && (yyn < YYLAST) &&
			(yychk[yyact[yyn+1]+1] == YYERRCODE)) {
			    # Simulate a shift of "error".
			    yystate = yyact[yyn+1]
			    goto yystack_
		    }
		    yyn = yypact[yys[yyps]+1]

		    # The current yyps has no shift on "error", pop stack.
		    if (yydebug) {
			call printf ("error recovery pops state %d, ")
			    call pargs (yys[yyps])
			call printf ("uncovers %d\n")
			    call pargs (yys[yyps-1])
		    }
		    yyps = yyps - 1
		    yypv = yypv - YYOPLEN
		}

		# ABORT -- There is no state on the stack with an error shift.
yyabort_
		call sfree (yysp)
		return (ERR)


	    case 3: # No shift yet; clobber input char.

		if (yydebug) {
		    call printf ("error recovery discards char %d\n")
			call pargi (yychar)
		}

		if (yychar == 0)
		    goto yyabort_		# don't discard EOF, quit
		yychar = -1
		goto yynewstate_		# try again in the same state
	    }
	}


	# REDUCE -- Reduction by production yyn.

	if (yydebug) {
	    call printf ("reduce %d\n")
		call pargs (yyn)
	}
	yyps  = yyps - yyr2[yyn+1]
	yypvt = yypv
	yypv  = yypv - yyr2[yyn+1] * YYOPLEN
	YYMOVE (yypv + YYOPLEN, yyval)
	yym   = yyn

	# Consult goto table to find next state.
	yyn = yyr1[yyn+1]
	yyj = yypgo[yyn+1] + yys[yyps] + 1
	if (yyj >= YYLAST)
	    yystate = yyact[yypgo[yyn+1]+1]
	else {
	    yystate = yyact[yyj+1]
	    if (yychk[yystate+1] != -yyn)
		yystate = yyact[yypgo[yyn+1]+1]
	}

	# Perform action associated with the grammar rule, if any.
	switch (yym) {
	    
case 2:
# line 41 "t_qcalc.y"
{
			return (OK)
		    }
case 3:
# line 44 "t_qcalc.y"
{
			return (EOF)
		    }
case 4:
# line 47 "t_qcalc.y"
{
			yyerrok
		    }
case 5:
# line 50 "t_qcalc.y"
{
			return (EOF)
		    }
case 6:
# line 55 "t_qcalc.y"
{
			# Print the value of an expression.
    call printf("%g %g %g %g\n")
    call pargd(OPVALW(yypvt))
    call pargd(OPVALX(yypvt))
    call pargd(OPVALY(yypvt))
    call pargd(OPVALZ(yypvt))
    call flush(STDOUT)
    }
case 7:
# line 64 "t_qcalc.y"
{
			# Print the value of an expression as vector
    call printf("%g %g %g\n")
    call pargd(OPVALX(yypvt))
    call pargd(OPVALY(yypvt))
    call pargd(OPVALZ(yypvt))
    call flush(STDOUT)
    }
case 8:
# line 72 "t_qcalc.y"
{
			# Print the value of an expression as position
    call qc_position(yypvt, 0)
    call flush(STDOUT)
    }
case 9:
# line 77 "t_qcalc.y"
{
			# Print the value of an expression as position
    call qc_position(yypvt, 1)
    call flush(STDOUT)
    }
case 10:
# line 82 "t_qcalc.y"
{
			# Set the value of a register (from a-z).
			call putreg (OPVALI(yypvt-2*YYOPLEN), yypvt)
		    }
case 11:
# line 88 "t_qcalc.y"
{
			YYMOVE (yypvt-2*YYOPLEN, yyval)
		    }
case 12:
# line 91 "t_qcalc.y"
{
			call binop (yypvt-3*YYOPLEN, yypvt, yyval, '+')
		    }
case 13:
# line 94 "t_qcalc.y"
{
			call binop (yypvt-3*YYOPLEN, yypvt, yyval, '-')
		    }
case 14:
# line 97 "t_qcalc.y"
{
			call binop (yypvt-3*YYOPLEN, yypvt, yyval, '*')
		    }
case 15:
# line 100 "t_qcalc.y"
{
			call binop (yypvt-3*YYOPLEN, yypvt, yyval, '@')
		    }
case 16:
# line 103 "t_qcalc.y"
{
			call binop (yypvt-3*YYOPLEN, yypvt, yyval, '%')
		    }
case 17:
# line 106 "t_qcalc.y"
{
			call binop (yypvt-3*YYOPLEN, yypvt, yyval, '/')
		    }
case 18:
# line 109 "t_qcalc.y"
{
			call unop (yypvt, yyval, '-')
		    }
case 19:
# line 112 "t_qcalc.y"
{
			if (OPVALW(yypvt-7*YYOPLEN) > 0)	YYMOVE (yypvt-4*YYOPLEN, yyval)
			else			YYMOVE (yypvt, yyval)
                    }
case 20:
# line 116 "t_qcalc.y"
{
                        call unop(yypvt-YYOPLEN, yyval, '|')
                    }
case 21:
# line 119 "t_qcalc.y"
{
			call getreg (OPVALI(yypvt), yyval)
		    }
case 23:
# line 123 "t_qcalc.y"
{
                        OPVALW(yyval) = 0
			call qc_decdeg(yypvt-2*YYOPLEN,yyval)
                    }
case 24:
# line 127 "t_qcalc.y"
{
                        OPVALW(yyval) = 0
			call qc_dechour(yypvt-2*YYOPLEN,yyval)
                    }	}

	goto yystack_				# stack new state and value
end
