# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_qp2evttab.x
# 	Author	:  tnk
# 	Date	:  07/10/96
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_qp2evttab.x -r1.8 07/10/96 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	05/14/98   bboyd	changed two "free" calls to "mfree" 
# 	10/13/97   mabbott	removed hlib$ from mach.h include
# 	09/25/97   bboyd	Replaced private QPOE constants, added call to qp_acquirefields,
# both of which are now within an include file or library routines in the euv
# 	08/11/97   mabbott	correct misdeclarations of access() and streq() as wrong type
# 	12/26/96   tnk	fixed length of vector read in qp_read call
# 	10/24/96   tnk	changed print formats
# 	07/10/96   tnk	moved header card copy before data copy 
# H-
# 
#

# qp2evttab @(#)t_qp2evttab.x	1.8 5/14/98 00/02/01 UCB CEA EGO
# Author : Gareth Bevan


# Include block

# include <qpoe$qpoe.h>
# include <qpoe$qpio.h>
include <qpeuv.h>
include <tbset.h>
include <mach.h>

define SZ_STRING	8


# define error codes
define	NO_ERROR 0
define	NO_QPOE 1
define	TABLE_EXISTS 2
define	QPOE_DD_ERROR 3
define	NO_MACRO 4
define	COL_FAILED 5
define	TOO_MANY_FIELDS 6
define	MAX_ERROR 7

# The maximum nimber of event record fields
define	MAX_EVT_REC_FIELDS	64



################
# T_QP2EVTTAB - copies the event list of a qpoe file
# into an ST table. The table compatible header cards are also 
# copied to the table header.

procedure t_qp2evttab()



# Local variable declarations

pointer	sp			# Stack marker
pointer	tabname			# table name
pointer	qpname			# qpoe name
pointer	filtname		# The name of the qpoe filter

pointer	qpptr			# Pointer the the open qpoe file
pointer	tabptr			# Pointer the the open table file

bool	verbose			# verbose flag

# Function declarations
int	stat, mkevtlst()

begin

  # Allocate the memory for the table name and qpoe file name
  call smark(sp)
  call salloc( tabname, SZ_FNAME, TY_CHAR)
  call salloc( qpname, SZ_FNAME, TY_CHAR)
  

  # Allocate memory for the qpoe filter expression
  call salloc( filtname, SZ_PATHNAME, TY_CHAR)
  
  # Read the parameters from the pf file
  call read_pf( tabname, qpname, verbose )

  # Validate that the table and QPOE file named exist/can be create
  call tryfiles( qpname, tabname, filtname, qpptr, tabptr )

  # If they can copy the event list and header cards from the QPOE to the table
  stat = mkevtlst( qpptr, filtname, tabptr, verbose )

  call closefile( qpptr, tabptr)
  call sfree(sp)


end # t_qp2evttab()





################
# CLOSEFILE - close any open files
#
procedure	closefile (qpptr, tabptr) #LOCAL

pointer	qpptr		#I - the qpoe pointer
pointer	tabptr		#I - the table pointer

# Local variable declarations

# Function declarations

begin
  # Check that the pointer is valid, if so close file 
  if ( qpptr != 0 ) {
    call qpclose( qpptr )
  }

  if ( tabptr != 0 ) {
    call tbtclo( tabptr )
  }
  


end # closefile()
       



################
# READ_PF - reads the the named parameters and returns
# them to the calling routine

procedure read_pf( tabname, qpname, verbose ) #LOCAL

pointer	tabname	#O	ST table name
pointer	qpname	#O	QPOE file name

bool	verbose	#O	verbose flag


# Define called functions
bool	clgetb()

begin
  #Read the table name ,qpoe name and verbose flag from the parameter
  #file.
  call clgstr( "qpoe", Memc[qpname], SZ_FNAME)
  call clgstr( "table", Memc[tabname], SZ_FNAME)

  verbose = clgetb( "verbose" )

end #  read_pf() 


################
# TRYFILES - test the user input files to see if they exist / can 
# be created. If the table file exists it is deleted

procedure tryfiles( qpname, tabname, filter, qpptr, tabptr ) #LOCAL

pointer	qpname		#I - the qpoe name + expression
pointer	tabname		#I/O - the table name
pointer filter		#O - the qpooe filter expression
pointer	qpptr		#O - qpoe pointer
pointer	tabptr		#O - table pointer

# Local variables

pointer	sp		# Stack pointer
pointer	root		# the root name of the qpoe file

# functions
int	access()
pointer	qp_open()
pointer	tbtopn()
int	tbtacc()

begin

  call smark(sp)
  call salloc( root, SZ_FNAME, TY_CHAR)

  # Parse the qpoe file name into file name and filter name
  call qp_parse( Memc[qpname], Memc[root], SZ_FNAME, Memc[filter], SZ_PATHNAME)

  # Test if the qpoe file exists
  if ( access(Memc[root],READ_ONLY,0 ) == YES ) {
    
    # If so open the qpoe file
    qpptr = qp_open( Memc[root], READ_ONLY, 0 )
    
    # Test if the table exists
    # Append the .tab file extention to the file name
    call tbtext( Memc[tabname], Memc[tabname], SZ_FNAME )

    if ( tbtacc( Memc[tabname] ) == YES || access( Memc[tabname], 0, 0 ) == YES )  {
      # If the table exists delete it
	call delete( Memc[tabname] )
    } 
    # Open a new table
    tabptr = tbtopn( Memc[tabname], NEW_FILE, NULL )
  } else {
    call error( NO_QPOE, "The QPOE file does not exists" )
  }

  call sfree(sp)
  
end # tryfiles()


       
################
# MKEVTLST - fill in the table with the qpoe event list and header cards
#

int procedure	mkevtlst ( qpptr, filter, tabptr, verbose ) #LOCAL

pointer	qpptr		#I - the qpoe pointer
pointer	filter		#I - the qpoe filter expression
pointer tabptr		#I - the table pointer
pointer	verbose		#I - the verbose flag

# Local variable declarations

pointer	ddptr		# pointer to the qpoe domain descriptor
pointer qpio		# pointer to the event list of the qpoe file

# Function declarations
pointer	qpio_open()


begin
  qpio = qpio_open( qpptr, Memc[filter], READ_ONLY )

  # Allocate space for the field descriptor struct
  call malloc ( ddptr, LEN_EUVDDDES, TY_STRUCT)

  # Load up the struct
  call qp_acquirefields(qpio, ddptr)

  # Get the column defs from the qpoe file and add them to the table
  call getqpcol( qpptr, ddptr, tabptr )

  # Copy the header from the qpoe file to the table
  # (Do this before the data to minimize file copying).
  call hcpkill( qpptr, "QPOE", tabptr, "TABLE", NULL )

  # Copy the event list from the qpoe file to the tables
  call cpevts( qpptr, qpio, ddptr, tabptr, verbose )
  call qpio_close( qpio )

  call mfree (ddptr, TY_STRUCT)

end # mkevtlst()




# GETQPCOL - Get the column definitions of the event list in the qpoe file
# and create a column in the table for each one.
#
procedure	getqpcol ( qpptr, ddptr, tabptr ) #LOCAL

pointer	qpptr		#I - the qpoe file pointer
pointer	ddptr		#I - the qpoe domain descriptor
pointer	tabptr		#I - the table pointer	

# Local variable declarations
pointer	sp		# Stack pointer
pointer	hdrname		# The name of the qpoe header parameter
pointer	qpdatatype	# The type of the data element in the qpoe file
pointer	comment		# Any comment that may have been given with a header card
pointer	value		# The string value of the macro
pointer	hdrlst		# A list of the qpoe header parameters
pointer mval		# The generated macro value


int	numelm		# The number of events read from the qpoe file
int	offset		# The offset of the event record in the event structure
int	numrec		# the number of records in the event stucture
int	recnum		# Loop counter for each event record
int	datatype	# The datatype of the event record
int	flags		# qpoe flags for header card
int	veclen		# The size of the qpoe header expression
int	maxelm		# Allocated length of parameter

bool	macfnd		# True if a macro has been found


# Function declarations
pointer	qp_ofnlu()
int	qp_queryf()
int	qp_gnfn()
bool streq()
int	qp_read()

int	evtoffset()

begin
  offset = 0

  # Allocate space for the qpoe macro names
  call smark(sp)
  call salloc( hdrname, SZ_COLNAME, TY_CHAR )
  call salloc( qpdatatype, SZ_LINE, TY_CHAR )
  call salloc( comment, SZ_LINE, TY_CHAR )
  call salloc( value, SZ_LINE, TY_CHAR )
  
  # Determine the number of events records
  numrec = EUVDD_NFIELDS( ddptr )
  
  # Loop over each record
  for ( recnum=1; recnum <= numrec ; recnum = recnum + 1 ) {
    # Get the datatype
    datatype = EUVDD_FTYPE( ddptr, recnum )
    
    # Calulate the offset in the event structure
    offset = evtoffset( ddptr, recnum, datatype )
    
    # Generate the macro name
    call macvalue( datatype, offset, mval )

    # Loop through all the macros looking for the value that 
    # corresponds to mval
    hdrlst = qp_ofnlu( qpptr, "*" )

    macfnd = FALSE
    while ( qp_gnfn( hdrlst, Memc[hdrname], SZ_FNAME ) != EOF ) {
      veclen = qp_queryf( qpptr, Memc[hdrname], Memc[qpdatatype], maxelm,
			 Memc[comment], flags )
      if ( streq( Memc[qpdatatype], "macro" )  ) {
	# Read the macro value
	call aclrc( Memc[value], SZ_STRING )
	numelm = qp_read( qpptr, Memc[hdrname], Memc[value], SZ_STRING, 1, Memc[qpdatatype] )

	# Compare it to the generated string
	if ( streq(Memc[value], Memc[mval] ) ) {
	  # Make a column of that name and type in the table
	  call mktabcol( tabptr, hdrname, datatype )
	  macfnd = TRUE
	}
      }
    }
    call qp_cfnl( hdrlst )

    # Display a warning message if the macro cannot be found
    if ( macfnd == FALSE ) {
	call eprintf("Macro not found with value %s no column created\n")
	call pargstr(Memc[value])
    }	
  }  # End loop

  # Create table
  call tbtcre( tabptr )

  call sfree(sp)
  call mfree(mval, TY_CHAR)

end # getqpcol()



################
# EVTOFFSET - Caluulate the offset of the record in the 
# event data structure     
#
int	procedure	evtoffset (ddptr, recnum, datatype ) #LOCAL

pointer	ddptr		#I - The event domain descriptor structure
int	recnum		#I - The number of the record in the structure
int	datatype	#I - The QPOE datatype of the record
int	offset		#I/O The byte aligned offset of the record in the strucure

# Local variable declarations

int	size		# The size of the datatype in bytes

# Function declarations

begin

  switch ( datatype ) {
  case TY_BOOL :
    size = SZ_BOOL
    
  case TY_SHORT :
    size = SZ_SHORT
    
  case TY_INT :
    size = SZ_INT
    
  case TY_LONG :
    size = SZ_LONG
    
  case TY_REAL :
    size = SZ_REAL
    
  case TY_DOUBLE :
    size = SZ_DOUBLE

  default:
    call fatal( QPOE_DD_ERROR, "Could't calulate event record size" )
    
  }
  # Multiply the qpoe offset by the number of bytes in a char to get
  # the real offset in bytes
  offset = size * EUVDD_FOFFSET( ddptr, recnum ) * SZB_CHAR
  return ( offset )

end # evtoffset()


################
# CPEVTS - Copy the events from the QPOE file to the to the table
#
procedure	cpevts (qpptr, qpioptr, ddptr, tabptr, verbose ) #LOCAL

pointer	qpptr		#I - the qpoe file pointer
pointer	qpioptr		#I - qpoe event pointer
pointer	ddptr		#I - the domain descriptor pointer
pointer	tabptr		#I - the table file pointer
bool	verbose		#I - the verbose flag

# Local variable declarations
int	numrecs		# The number of records in an event
int	numevts		# The number of events in the qpoe event list
int	maxelm		# allocated length of parameter
int	flags		# Header cards flags
int	recnum		# The current event record number
int	evtmsk		# Event mask
int	numevt		# The number of events returned
int	type		# The datatype of the event record
int	offset		# The qpoe offset of the record
int	rownum		# The current row number


real	perdone		# The percentage completed
real	decile		# The decile part of the percentage completed
real	lastdecile	# The previous value ( used so that progress is given 
			# only in 10% increments

pointer	sp		# Stack pointer
pointer	datatype	# Qpoe datatype
pointer comment		# Headre card comment string
pointer	ev		# Pointer to the current event
pointer	colptr[MAX_EVT_REC_FIELDS]	# Array of ST table column pointers

# Function declarations
int	qp_queryf()
int	qpio_getevents()
pointer	tbcnum()

begin
  call smark(sp)
  evtmsk = 0
  rownum = 0
  
  call salloc( datatype, SZ_STRING, TY_CHAR )
  call salloc( comment, SZ_LINE, TY_CHAR )
  
  
  # Get the number of events
  numevts = qp_queryf( qpptr, "events", Memc[datatype], maxelm, Memc[comment], flags )
  numrecs = EUVDD_NFIELDS( ddptr )
  if ( numrecs > MAX_EVT_REC_FIELDS ) {
    call fatal( TOO_MANY_FIELDS, "Error: There are too many fields in the event record" )
  }

  # Fill the column pointer array
  for ( recnum=1; recnum<=numrecs; recnum = recnum + 1 ) {
    colptr[recnum] = tbcnum( tabptr, recnum )
  }
  
  # Loop over each event
  while ( qpio_getevents( qpioptr, ev, evtmsk, 1, numevt ) != EOF ) {

    rownum = rownum + 1
    
    # For each record in the event
    for ( recnum=1; recnum<=numrecs; recnum = recnum + 1 ) {
      # Get the event offset and type
      type = EUVDD_FTYPE( ddptr, recnum  )
      offset = EUVDD_FOFFSET( ddptr, recnum )

      # Determine the type of the event
      # Add the record to the appropriate column in the table
      call addvaltotab ( tabptr, ev, type, offset, colptr[recnum], rownum )
    }

    # If the verbose flag is set calulate the percentage completed 
    if ( verbose ) {
      # Calculate percentage done to 10% accuracy
      perdone = rownum * 100.0 / numevts 
      decile = perdone / 10.0
      if ( decile  > lastdecile + 1.0 ) {
	call printf(" %d %% of events copied.\n")
	call pargr( perdone )
	lastdecile = decile
      }
    }
  } # End loop

  call sfree(sp)
  
end # cpevts()

################
# ADDVALTOTAB - add the value at the offset position in the event structure
# to the table
#
procedure	addvaltotab ( tabptr, ev, type, offset, colptr, rownum ) #LOCAL

pointer	tabptr		#I - the table pointer
pointer	ev		#I - event structure
int	type		#I - the datatype of the event record
int	offset		#I - the offset of the record in the event structure
pointer	colptr		#I - the column pointer for the data
int	rownum		#I - the rownumber to in which to place the data


# Local variable declarations
bool	boolval		
short	srtval
int	intval
real	realval
double	dblval


# Function declarations
int	coerce()

begin
      switch (type) {
      case TY_BOOL:
	boolval = Memb[coerce(ev,TY_SHORT,type) + offset ]
	call tbeptb( tabptr, colptr, rownum, boolval)
      case TY_SHORT:
	srtval = Mems[coerce(ev,TY_SHORT,type) + offset ]
	call tbepts( tabptr, colptr, rownum, srtval)
      case TY_INT:
	intval = Memi[coerce(ev,TY_SHORT,type) + offset ]
	call tbepti( tabptr, colptr, rownum, intval)
      case TY_REAL:
	realval = Memr[coerce(ev,TY_SHORT,type) + offset ]
	call tbeptr( tabptr, colptr, rownum, realval)
      case TY_DOUBLE:
	dblval = Memd[coerce(ev,TY_SHORT,type) + offset ]
	call tbeptd( tabptr, colptr, rownum, dblval)
	
      default:
	# If event cannot be added set error flag
        call eprintf("Skipping unknown event type : %d\n")
	call pargd( type )
      }

end # addvaltotab()



##################
# MACVALUE - generates the EUVE QPOE macro name from the datatype and 
# offset in the qpoe file domain descriptor macval
#
procedure	macvalue (datatype, offset, mname ) #LOCAL

int	datatype	#I - the qpoe datatype of the event record
int	offset		#I - the offset of the start of the record from the 
                        #    start of the events structure
pointer	mname		#O - the macro name


# Local variable declarations
pointer	offsetstr	# The string value of the record event offset

string	errstr "Unable to determine datatype of macro: xxxx" # The error string 
					                     # value if needed

# Function declarations

begin
  # Allocate the space for the string
  call calloc( mname, SZ_STRING, TY_CHAR )
  call calloc( offsetstr, SZ_STRING , TY_CHAR )
  
  # Generate the letter prefix for the datatype ( lower case )
  switch ( datatype ) {
  case TY_BOOL:
    call strcpy( "b", Memc[mname], SZ_STRING )
  case TY_SHORT:
    call strcpy( "s", Memc[mname], SZ_STRING )
  case TY_INT:
    call strcpy( "i", Memc[mname], SZ_STRING )
  case TY_LONG:
    call strcpy( "l", Memc[mname], SZ_STRING )
  case TY_REAL:
    call strcpy( "r", Memc[mname], SZ_STRING )
  case TY_DOUBLE:
    call strcpy( "d", Memc[mname], SZ_STRING )
  default:
    call sprintf(errstr, SZ_LINE, "Unable to determine datatype of macro: %d")
    call pargi( datatype )
    call error( NO_MACRO, errstr )
  }


  # Append the offset value
  call sprintf(Memc[offsetstr], SZ_STRING , "%d")
  call pargi( offset )					      
  
  call strcat( Memc[offsetstr], Memc[mname], SZ_STRING )
  call mfree(offsetstr, TY_CHAR)

end # macvalue()


##################
# MKTABCOL - Make the table column in the ST table of the given
# type
#

procedure	mktabcol ( tabptr, colname, datatype ) #LOCAL

pointer	tabptr		#I - the table pointer
pointer	colname		#I - the name of the column to create
int	datatype	#I - the datatype of the column to create

# Local variable declarations

pointer	colptr		# returned column pointer
pointer sp		# Stack pointer
pointer	units		# The column units
pointer	prtfmt		# The column print format

string	errstr "Column datatype xxxx unknown"

# Function declarations

begin
  # Allocate temparay variables
  call smark(sp)
  call salloc( units, SZ_COLUNITS, TY_CHAR)
  call salloc( prtfmt, SZ_COLFMT, TY_CHAR)

 # Depending on column datatype
    # Generate print format
    # Generate units
    # Create column

  switch ( datatype ) {
  case TY_BOOL:
    call strcpy("%b",Memc[prtfmt],SZ_COLFMT)
    call strcpy("",Memc[units],SZ_COLUNITS)
    call tbcdef( tabptr, colptr, Memc[colname], Memc[units], Memc[prtfmt],
		TY_BOOL, 1, 1 )
  case TY_SHORT:
    call strcpy("%d",Memc[prtfmt],SZ_COLFMT)
    call strcpy("",Memc[units],SZ_COLUNITS)
    call tbcdef( tabptr, colptr, Memc[colname], Memc[units], Memc[prtfmt],
		TY_SHORT, 1, 1 )

  case TY_INT:
    call strcpy("%d",Memc[prtfmt],SZ_COLFMT)
    call strcpy("",Memc[units],SZ_COLUNITS)
    call tbcdef( tabptr, colptr, Memc[colname], Memc[units], Memc[prtfmt],
		TY_INT, 1, 1 )

  case TY_LONG:
    call strcpy("%d",Memc[prtfmt],SZ_COLFMT)
    call strcpy("",Memc[units],SZ_COLUNITS)
    call tbcdef( tabptr, colptr, Memc[colname], Memc[units], Memc[prtfmt],
		TY_LONG, 1, 1 )

  case TY_REAL:
    call strcpy("%g",Memc[prtfmt],SZ_COLFMT)
    call strcpy("",Memc[units],SZ_COLUNITS)
    call tbcdef( tabptr, colptr, Memc[colname], Memc[units], Memc[prtfmt],
		TY_REAL, 1, 1 )

  case TY_DOUBLE:
    call strcpy("%g",Memc[prtfmt],SZ_COLFMT)
    call strcpy("",Memc[units],SZ_COLUNITS)
    call tbcdef( tabptr, colptr, Memc[colname], Memc[units], Memc[prtfmt],
		TY_DOUBLE, 1, 1 )
  default:
    call sprintf( errstr, SZ_LINE, "Column datatype %d unknown")
    call pargi( datatype )
    call error( COL_FAILED, errstr )

  }

  call sfree(sp)

end # mktabcol()



