# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_qpbin.x
# 	Author	:  ericco
# 	Date	:  08/14/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_qpbin.x -r1.14 08/14/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	01/13/97   tnk	changed from qpl_open_filtered to qpofilt
# 	01/13/97   tnk	fixed indef datatype checks in qpb_value
# 	02/06/95   akeith	Added integer case to qpb_value()
# 	11/10/94   tnk	changed system type to linear
# 	10/07/94   tnk	verify filter is specified for integrate mode
# 	09/15/94   tnk	changed 'range' to 'integrate'
# 	06/15/94   ericco	changed max output image size to 32M pixels
# 	03/10/94   tnk	added long to attribute check
# 	03/10/94   tnk	added attribute QA test, changed INDEFR cards, added hcpkill call
# 	12/01/92   ericco	changed real calculations to doubles. also fixed wcs edges.
# 	08/18/92   ericco	added range facility
# 	08/14/92   ericco	date and time created 92/08/14 19:49:08 by ericco
# H-
# 
#

#	@(#)t_qpbin.x	1.14 1/13/97

include <imhdr.h>
include	<qpioset.h>

define	SZ_MAX_IMAGE	(32 * 1024 * 1024)

#	Generate an image from a projection of a QPOE event field.

procedure	t_qpbin()

pointer		qpname				# QPIO filename param. ptr

double		clgetd()			# get cl parameter value

pointer		qp, qpio	# QPOE pointers
pointer		image
pointer		im, immap()	# Output image
int		maskval
pointer		ev		# event pointer
pointer		binfield
pointer		field
pointer		sp
int		offset
long		lval
char		kind
int		ip

double		bin_start, bin_end, bin_delta, bin_length
double		dval, qpb_value()
int		len

pointer		out, impl2l(), impl2r()
int		nev
int		qpio_getevents()
int		index
pointer		mw, mw_open()
pointer		qpio_stati()
bool		normp, clgetb()
pointer		units
int		nranges
int		qpex_attrld()
pointer		xs, xe
int		xlen
pointer		ex
int		mw_axes
double		cd[2,2], r[2], w[2]
int		clgeti()
char		firstdig
int 		stridx()
pointer         text

begin
  call smark(sp)
  call salloc(field, SZ_LINE, TY_CHAR)
  call salloc(text, SZ_LINE, TY_CHAR)

  call eu_clstr(qpname, "input")		# get filename parameter
  call qpofilt(qp, qpio, Memc[qpname])
  ex = qpio_stati(qpio, QPIO_EX)

  call eu_clstr(binfield, "attribute")		# get field parameter
  call qp_expandtext(qp, Memc[binfield], Memc[field], SZ_LINE)
  kind = Memc[field]
  firstdig = Memc[field+1]

  # verify attribute
  # note that this only a test of attribute format
  if ( (stridx(kind,"silrdSILRD") == 0) ||
       (stridx(firstdig,"0123456789") == 0))
  {
     call error(101,"Invalid attribute")
  }

  ip = 2
  call gctol(Memc[field], ip, lval, 10)
  offset = lval / 2

  call eu_clstr(image, "output")		# get output parameter

  # Get the limits on the output image
  bin_start = clgetd("x1")
  if (bin_start == INDEFD) {
    ev = qpio_stati (qpio, QPIO_MINEVP)
    dval = qpb_value(ev, offset, kind)
    if (dval == INDEFD) bin_start = INDEFD
    else		bin_start = dval
  }

  bin_end = clgetd("x2")
  if (bin_end == INDEFD) {
    ev = qpio_stati (qpio, QPIO_MAXEVP)
    dval = qpb_value(ev, offset, kind)
    if (dval == INDEFD) bin_end = INDEFD
    else		bin_end = dval
  }

  if (bin_start >= bin_end || bin_start == INDEFD || bin_end == INDEFD) {
    call error(1, "illegal range")
  }

  # Get remaining values
  call eu_clstr(units, "units")
  normp = clgetb("integrate")

  bin_delta = clgetd("dx")
  if (bin_delta == INDEFD) {
    len = clgeti("nx")
    bin_delta = (bin_end - bin_start) / len
    bin_length = bin_delta * len
  } else {
    # Adjust the image end point to fit the right edge of a pixel.
    bin_length = bin_end - bin_start
    len = bin_length / bin_delta
    if (len > SZ_MAX_IMAGE) len = SZ_MAX_IMAGE
  }
  bin_end = bin_start + len * bin_delta

  # open output image as necessary
  im = immap(Memc[image], NEW_IMAGE, 0)
  if (normp) {
    IM_NDIM(im) = 2
    IM_PIXTYPE(im) = TY_REAL
    IM_LEN(im,2) = 2
  } else {
    IM_NDIM(im) = 1
    IM_PIXTYPE(im) = TY_LONG
  }
  IM_LEN(im,1) = len


  # initialize bin counts
  out = impl2l(im, 1)
  call amovkl(0.0, Meml[out], len)

  # Put each event into its appropriate bin, ignoring the
  # events with undefined values.
  maskval = 0
  while (qpio_getevents(qpio, ev, maskval, 1, nev) != EOF) {
    dval = qpb_value(ev, offset, kind)
    if (dval == INDEFD) next
    index = (dval - bin_start) / bin_delta
    if (index < 0 || index >= len) next
    Meml[out + index] = Meml[out + index] + 1
  }

  # Handle the normalization information.
  # Extract the attribute list of the given field, if there is one.
  # Then integrate the run length encode ranges for each output pixel.
  if (normp) {
    out = impl2r(im, 2)
    xs = NULL
    xe = NULL
    xlen = 0
    nranges = 0
    if (ex != NULL) {
      nranges = qpex_attrld(ex, Memc[binfield], xs, xe, xlen)
    }

    if (nranges == 0)
       call error(102,"Valid filter must be given for integrate mode")
    else
    {
       call rl_integrate(Memd[xs], Memd[xe], nranges,
	   	      bin_start, bin_delta, Memr[out], len)

       call mfree (xs, TY_DOUBLE)
       call mfree (xe, TY_DOUBLE)
    }
 
  }

  # Set up the coordinate system
  mw = 0
  mw_axes = 1
  if (normp)	mw_axes = 2
  mw = mw_open(mw, mw_axes)

  call mw_newsystem(mw, "linear", mw_axes)

  call mw_swattrs(mw, 1, "wtype", "linear")
  call mw_swattrs(mw, 1, "label", Memc[binfield])
  call mw_swattrs(mw, 1, "units", Memc[units])

  # Setup the WCS.  Note that the left edge of an IRAF image
  # starts at 0.5, so that the middle of the first pixel is at
  # 1.0.
  cd[1,1] = bin_delta;  cd[2,1] = 0
  cd[1,2] = 0;          cd[2,2] = 1
  r[1]    = 0.5;        r[2]    = 1.0
  w[1]    = bin_start;  w[2]    = 1.0

  call mw_swtermd(mw, r, w, cd, mw_axes)
  call mw_saveim(mw, im)

  # copy header cards from qpoe file to output image
  call hcpkill(qp, "Q", im, "I", "")

  # add history information
  call sprintf(Memc[text],SZ_LINE,"Generated from '%s' by qpbin.\n")
  call pargstr(Memc[qpname])
  call strcat(Memc[text], IM_HISTORY(im), SZ_IMHIST)


  # Clean up
  call mw_close(mw)
  call imunmap(im)
  call qpio_close(qpio)
  call qp_close(qp)

  call sfree(sp)
end


#	Integrate a run length encoded segment.

procedure	rl_integrate(xs, xe, xlen, start, delta, out, len) #E
double		xs[ARB], xe[ARB]	#I range list
int		xlen			#I range list length
double		start, delta		#I output info
real		out[ARB]		#O output array
int		len			#I output length

int		i		# pixel index
int		j		# range index

double		xstart, xend	# save initial limit values
double		ps, pe		# current pixel start and end limits
double		rs, re		# current range start and end limits

begin
  # Compute simply case of zero length output array
  if (len < 1) return
  call amovkr(0.0, out, len)

  # Simply case of zero length input ranges.
  if (xlen < 1) {
    call amovkr(real(delta), out, len)
    return
  }

  # Save first and last range limits
  xstart = xs[1]
  xend = xe[xlen]

  # Truncate open ended ranges to array limits
  if (IS_INDEFD(xs[1])) {
    xs[1] = start
  }
  if (IS_INDEFD(xe[xlen])) {
    xe[xlen] = start + len * delta
  }

  # For each pair of intersecting pixel and range limits,
  # sum the intersection.
  i = 1
  j = 1

  while (i <= len && j <= xlen) {

    # This could be optimized to only be updated when
    # they actually change.
    ps = start + (i - 1) * delta
    pe = start + i * delta
    rs = xs[j]
    re = xe[j]

    # Handle simply non-intersecting cases first.
    if (rs >= pe) {
      i = i + 1
    } else if (re <= ps) {
      j = j + 1
    } else if (rs > ps) {		# ps < rs < pe
      if (re < pe) {			# ps < rs < re < pe
	out[i] = out[i] + (re - rs)
	j = j + 1
      } else {				#      pe < re
	out[i] = out[i] + (pe - rs)
	i = i + 1
      }
    } else {				# rs < ps
      if (re > pe) {			#      pe < re
	out[i] = out[i] + delta
	i = i + 1
      } else {				# ps < re < pe
	out[i] = out[i] + (re - ps)
	j = j + 1
      }
    }
  }

  # restore possibly open end ranges
  xs[1] = xstart
  xe[xlen] = xend
end


#	Convert any event field into a double value.
#	This could be optimized so that coerce() is not
#	called.

double procedure	qpb_value(ev, offset, kind) #I
pointer		ev		#I event pointer
int		offset		#I field offset
char		kind		#I field data type

double		dval
short		sval
real		rval
int		ival
long		lval

pointer		coerce()	# pointer converter

begin
  switch(kind) {
  case 's' : sval = Mems[coerce(ev+offset, TY_SHORT, TY_SHORT)]
    if (sval == INDEFS)	dval = INDEFD
    else		dval = sval
  case 'i' : ival = Memi[coerce(ev+offset, TY_SHORT, TY_INT)]
    if (ival == INDEFI)	dval = INDEFD
    else		dval = ival
  case 'l' : lval = Meml[coerce(ev+offset, TY_SHORT, TY_LONG)]
    if (lval == INDEFL)	dval = INDEFD
    else		dval = lval
  case 'r' : rval = Memr[coerce(ev+offset, TY_SHORT, TY_REAL)]
    if (rval == INDEFR)	dval = INDEFD
    else		dval = rval
  case 'd' : dval = Memd[coerce(ev+offset, TY_SHORT, TY_DOUBLE)]
  default  : dval = INDEFD
  }
  return dval
end
