# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_qpmedit.x
# 	Author	:  gareth
# 	Date	:  10/20/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_qpmedit.x -r1.6 10/20/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	09/12/95   tnk	added QPF_INHERIT; sccs keywords
# 	11/17/93   ericco	fixed pointer/bool error
# 	12/18/92   ericco	fixed long string output
# 	10/20/92   ericco	added qp_add_macro call
# 	10/20/92   gareth	date and time created 92/10/20 12:23:12 by gareth
# H-
# 
#

#
#  QPMEDIT -  edit QPOE file macros
#  @(#)t_qpmedit.x	1.6 9/12/95

include	<ctype.h>
include	<error.h>
include	<qpset.h>

define  SZ_MNAME        32              # max size macro name
define  SZ_MVBUF        8192            # max size macro value


define	OP_EDIT		1	
define	OP_ADD		2
define	OP_DELETE	3

define NOTMACRO	2

define ALLOC_UNIT 1024  # unit of string allocation for file indirection

procedure t_qpmedit()

pointer	macros			# template listing macros to be processed
pointer	valexpr			# the value expression (if op=edit|add)

pointer	sp, macro, sections, s_macros, s_valexpr, qp, ip, file, buf
pointer	mlist, qp_ofnlu()
int		operation, verify, show, update, qplist, nmacros

pointer	qp_open()
bool	clgetb(),  streq()
int		btoi(), imtopenp(), imtgetim(), qp_gnfn()

begin
	call smark (sp)
	call salloc (buf,       SZ_FNAME, TY_CHAR)
	call salloc (file,     SZ_FNAME, TY_CHAR)
	call salloc (macro,     SZ_FNAME, TY_CHAR)
	call salloc (s_macros,  SZ_LINE,  TY_CHAR)
	call salloc (s_valexpr, SZ_LINE,  TY_CHAR)
	call salloc (sections, SZ_FNAME, TY_CHAR)

	# Get the primary operands.
	qplist = imtopenp ("files")

	# Determine type of operation to be performed.  The default operation
	# is edit.

	operation = OP_EDIT
	if (clgetb ("add"))
	    operation = OP_ADD
	else if (clgetb ("delete"))
	    operation = OP_DELETE

	# Get list of macros to be edited, added, or deleted.
	call clgstr ("macros", Memc[s_macros], SZ_LINE)
	for (ip=s_macros;  IS_WHITE (Memc[ip]);  ip=ip+1)
    ;
	macros = ip

	# The value expression parameter is not used for the delete operation.
	if (operation != OP_DELETE) {
	    call clgstr ("value", Memc[s_valexpr], SZ_LINE)
	    for (ip=s_valexpr;  IS_WHITE (Memc[ip]);  ip=ip+1)
		;
	    valexpr = ip
	    while (Memc[ip] != EOS)
		ip = ip + 1
	    while (ip > valexpr && IS_WHITE (Memc[ip-1]))
		ip = ip - 1
	    Memc[ip] = EOS
	} else {
	    Memc[s_valexpr] = EOS
	    valexpr = s_valexpr
	}

	# Get switches.  If the expression value is ".", meaning print value
	# rather than edit, then we do not use the switches.

	if (operation == OP_EDIT && streq (Memc[valexpr], ".")) {
	    update = NO
	    verify = NO
	    show   = NO
	} else {
	    update = btoi (clgetb ("update"))
	    verify = btoi (clgetb ("verify"))
	    show   = btoi (clgetb ("show"))
	}

    if (streq(Memc[valexpr],"@stdin")) verify = NO

	# Main processing loop.  A QPOE file is processed in each pass through
	# the loop.

	while (imtgetim (qplist, Memc[file], SZ_FNAME) != EOF) {

	    # Open the file.
	    iferr {
		if (update == YES)
		    qp = qp_open (Memc[file], READ_WRITE, 0)
		else
		    qp = qp_open (Memc[file], READ_ONLY,  0)
	    } then {
		call erract (EA_WARN)
		next
	    }

	    if (operation == OP_ADD) {
		# Add a field to the image header.  This cannot be done within
		# the QP_GNFN loop because template expansion on the existing
		# fields of the image header would discard the new field name
		# since it does not yet exist.

		nmacros = 1
		call qm_addmacro (qp, Memc[file], Memc[macros], Memc[valexpr],
		    verify, show, update)
	    
	    } else {
		# Open list of macros to be processed.
		mlist = qp_ofnlu (qp, Memc[macros])

		nmacros = 0
		while (qp_gnfn (mlist, Memc[macro], SZ_FNAME) != EOF) {
		    switch (operation) {
		    case OP_EDIT:
			call qm_editmacro (qp, Memc[file], Memc[macro],
			    Memc[valexpr], verify, show, update)
		    case OP_DELETE:
			call qm_deletemacro (qp, Memc[file], Memc[macro],
			    Memc[valexpr], verify, show, update)
		    }
		    nmacros = nmacros + 1
		}

		if (nmacros == 0) {
		  call eprintf ("%s,%s : no macros match template\n")
		  call pargstr (Memc[file])
		  call pargstr (Memc[macros])
		}

		call qp_cfnl (mlist)
	    }

		call qp_close (qp)
	}

	call imtclose (qplist)
	call sfree (sp)
end


# QM_EDITMACRO -- Edit the value of the named macro of the indicated QPOE file.

procedure qm_editmacro (qp, file, macro, valexpr, verify, show, update)

pointer	qp				# image descriptor of QPOE file to be edited
char	file[ARB]		# name of file to be edited
char	macro[ARB]		# name of macro  to be edited
char	valexpr[ARB]	# value expression
int		verify			# verify new value interactively
int		show			# print record of edit
int		update			# enable updating of the file

int		goahead, nl, len, found, length
pointer	sp, ip, oldval, newval, defval

int		getline(), qm_macroexists(), strldxs(), strlen()
bool	streq()

errchk	getline, qm_gval, qm_indirect, qm_updatemacro, qm_macroexists

begin
  call smark (sp)

  # Verify that the named macro exists before going any further.

  length = strlen( macro )
  if ( length > SZ_MNAME )
  {
    call eprintf("ERROR: macro name too long \n")
    return
  }
  
  found = qm_macroexists (qp, macro)
  if (found == NO) {
	call eprintf ("parameter %s,%s not found\n")
	call pargstr (file)
	call pargstr (macro)
	call sfree (sp)
	return
  }
  if (found == NOTMACRO) {
	call sfree (sp)
	return
  }		

  # Get the old value.
  call qm_gval (qp, macro, oldval)

  # Evaluate the expression.  If file indirection, read the file.
  # Otherwise, assume that is is already a string literal.

  if (valexpr[1] == '@') {
	call qm_indirect (valexpr, newval)
  } else {
	call malloc (newval, SZ_LINE, TY_CHAR)
	call strcpy (valexpr, Memc[newval], SZ_LINE)
  }

  if (streq (Memc[newval], ".")) {
	# Merely print the value of the field.

	call printf ("%s,%s = ")
	call pargstr (file)
	call pargstr (macro)
	call prlstr (oldval)
	call printf("\n")

  } else if (verify == YES) {
	# Query for new value and edit the field.  If the response is a
	# blank line, use the default new value.  If the response is "$"
	# or EOF, do not change the value of the parameter.

	len = strlen(Memc[newval])
	call salloc (defval, len+1, TY_CHAR)
	call strcpy (Memc[newval], Memc[defval], len)
	call eprintf ("%s,%s (")
	call pargstr (file)
	call pargstr (macro)
	call eprlstr (oldval)
	call eprintf(" -> ")
	call eprlstr(defval)
	call eprintf("):")
	call flush (STDERR)

	if (getline (STDIN, Memc[newval]) != EOF) {
	  # Do not skip leading whitespace; may be significant in a
	  # string literal.

	  ip = newval

	  # Do strip trailing newline since it is an artifact of getline.
	  nl = strldxs ("\n", Memc[ip]) 
	  if (nl > 0)
	  Memc[ip+nl-1] = EOS

	  # Decode user response.
	  if (Memc[ip] == '\\') {
		ip = ip + 1
		goahead = YES
	  } else if (streq(Memc[ip],"n") || streq(Memc[ip],"no")) {
		goahead = NO
	  } else if (streq(Memc[ip],"y") || streq(Memc[ip],"yes") ||
				 Memc[ip] == EOS) {
		call strcpy (Memc[defval], Memc[newval], len)
		goahead = YES
	  } else {
		if (ip > newval)
		call strcpy (Memc[ip], Memc[newval], SZ_LINE)
		goahead = YES
	  }

	  # Edit field if so indicated.
	  if (goahead == YES && update == YES) {
		call qm_updatemacro (qp, file, macro, oldval, newval, 
							 show)
	  }

	  call flush (STDOUT)
	}

  } else if (update == YES) {
	call qm_updatemacro (qp, file, macro, oldval, newval,
						 show)
  }

  call mfree (oldval, TY_CHAR)
  call mfree (newval, TY_CHAR)
  call sfree (sp)
end


# QM_ADDMACRO -- Add a new macro to the indicated QPOE file.  If the field
# already exists, dp nothing.  The value expression is evaluated and the
# resulting value used as the initial value in adding the field to the
# image.

procedure qm_addmacro (qp, file, macro, valexpr, verify, show, update)

pointer	qp				# image descriptor of image to be edited
char	file[ARB]		# name of image to be edited
char	macro[ARB]		# name of field to be edited
char	valexpr[ARB]	# value expression
int		verify			# verify new value interactively
int		show			# print record of edit
int		update			# enable updating of the image

int		len, found
pointer	sp, newval, ip, ans

int		qm_macroexists(), strlen(), getline()

errchk	qm_indirect, qm_macroexists, qp_addf, qp_write

begin
  call smark (sp)

  call salloc (ans, SZ_LINE, TY_CHAR)

  # If the named field already exists, do nothing

  len = strlen( macro )
  if ( len > SZ_MNAME )
  {
    call eprintf("ERROR: macro name too long \n")
    return
  }
  
  found = qm_macroexists (qp, macro)
  if (found == YES || found == NOTMACRO) {
	call eprintf ("header parameter %s,%s already exists\n")
	call pargstr (file)
	call pargstr (macro)
	call sfree (sp)
	return
  }

  # Evaluate the expression.  If file indirection, read the file.
  # Otherwise, assume that is is already a string literal.

  if (valexpr[1] == '@') {
	call qm_indirect (valexpr, newval)
  } else {
	call malloc (newval, SZ_LINE, TY_CHAR)
	call strcpy (valexpr, Memc[newval], SZ_LINE)
  }

  # Add the field to the image.  

  len = strlen(Memc[newval])
  if ( len > SZ_MVBUF )
  {
    call eprintf("ERROR: macro exceeds max length %d \n" )
    call pargi( SZ_MVBUF )
    call sfree( sp )
    return
  }


  if (verify == YES) {
	# Add pending verification.

	call eprintf ("add %s,%s = ")
	call pargstr (file)
	call pargstr (macro)
	call eprlstr (newval)
	call eprintf(" ? (yes): ")
	call flush (STDERR)
	
	if (getline (STDIN, Memc[ans]) != EOF) {
	  # Strip leading whitespace and trailing newline.
	  for (ip=ans;  IS_WHITE(Memc[ip]);  ip=ip+1)
	  ;
	  if (Memc[ip] == '\n' || Memc[ip] == 'y') {
		iferr {
		  if (update == YES) {
#			call qp_addf (qp, macro, "macro", 0, 
#						  "qpmedit macro", QPF_INHERIT)
#			call qp_write (qp, macro, Memc[newval], len+1, 1, "macro")
			call qp_add_macro(qp, macro, Memc[newval])
		  }	
		} then {
		  call erract (EA_WARN)
		}
        else if (show == YES) {
		  call printf ("added %s,%s = ")
		  call pargstr (file)
		  call pargstr (macro)
		  call prlstr (newval)
		  call printf("\n")
		}
	  }
	}
	
  } else {
	# Add without verification.
	
	iferr {
	  if (update == YES) {
		call qp_addf (qp, macro, "macro", len, "macro edited by qpmedit", QPF_INHERIT)
		call qp_write (qp, macro, Memc[newval], len, 1, "macro")
	  }	
	} then {
	  call erract (EA_WARN)
	}
    else if (show == YES) {
	  call printf ("added %s,%s = ")
	  call pargstr (file)
	  call pargstr (macro)
	  call prlstr (newval)
	  call printf("\n")
	}
  }

  call mfree (newval, TY_CHAR)
  call sfree (sp)
end


# QM_DELETEMACRO -- Delete a macro from the indicated QPOE file.  If the
# field does not exist, print a warning message.

procedure qm_deletemacro (qp, file, macro, valexpr, verify, show, update)

pointer	qp				# image descriptor of image to be edited
char	file[ARB]		# name of image to be edited
char	macro[ARB]		# name of field to be edited
char	valexpr[ARB]	# not used
int		verify			# verify deletion interactively
int		show			# print record of edit
int		update			# enable updating of the image

int		found
pointer	sp, ip, ans

int		getline(), qm_macroexists()

errchk qm_macroexists, qp_deletef

begin
  call smark (sp)
  call salloc (ans, SZ_LINE, TY_CHAR)

  found = qm_macroexists (qp, macro)
  if (found == NO) {
	call eprintf ("nonexistent macro %s,%s\n")
	call pargstr (file)
	call pargstr (macro)
	call sfree (sp)
	return
  }
  if (found == NOTMACRO) {
	call sfree (sp)
	return
  }		
	    
  if (verify == YES) {
	# Delete pending verification.

	call eprintf ("delete %s,%s ? (yes): ")
	call pargstr (file)
	call pargstr (macro)
	call flush (STDERR)
	
	if (getline (STDIN, Memc[ans]) != EOF) {
	  # Strip leading whitespace and trailing newline.
	  for (ip=ans;  IS_WHITE(Memc[ip]);  ip=ip+1)
	  ;
	  if (Memc[ip] == '\n' || Memc[ip] == 'y') {
		iferr {
		  if (update == YES) {
			call qp_deletef (qp, macro)
		  }
		} then {
		  call erract (EA_WARN)
		}
	    else if (show == YES) {
		  call printf ("%s,%s deleted\n")
		  call pargstr (file)
		  call pargstr (macro)
		}	
	  }
	}
	
  } else {
	# Delete without verification.
	
	iferr {
	  if (update == YES) {
		call qp_deletef (qp, macro)
	  }
	} then {
	  call erract (EA_WARN)
	}
	else if (show == YES) {
	  call printf ("%s,%s deleted\n")
	  call pargstr (file)
	  call pargstr (macro)
	}	
  }

  call sfree( sp )
end


# QM_UPDATEMACRO -- Update the value of a QPOE file macro

procedure qm_updatemacro (qp, file, macro, oldval, newval, show)

pointer	qp				# QPOE file descriptor
char	file[ARB]		# file name
char	macro[ARB]		# field name
pointer	oldval   		# old value, encoded as a string
pointer	newval    		# new value, encoded as a string
int		show			# print record of update

int		len

int		strlen()

errchk qp_deletef, qp_addf, qp_write

begin
  len = strlen(Memc[newval])
  if ( len > SZ_MVBUF )
  {
    call eprintf("ERROR: macro exceeds %d \n")
    call pargi( SZ_MVBUF )
    return
  }

  iferr {
	call qp_deletef (qp, macro)
	call qp_addf (qp, macro, "macro", len, "macro edited by qpmedit", 0)
	call qp_write (qp, macro, Memc[newval], len, 1, "macro")
  } then {
	  call erract (EA_WARN)
  }	
  else if (show == YES) {
	call printf ("%s,%s: ")
	call pargstr (file)
	call pargstr (macro)
	call prlstr (oldval)
	call printf(" -> ")
	call prlstr (newval)
        call printf("\n")
      }
end


# QM_GVAL -- Get the value of an QPOE file macro and return it as a string.

procedure qm_gval (qp, macro, strptr)

pointer	qp			# QPOE file descriptor
char	macro[ARB]	# macro whose value is to be returned
pointer	strptr		# ptr to output string

int	len, maxlen, flags, nread
pointer dtype, comment, sp

int		qp_queryf(), qp_read()

errchk qp_queryf, qp_read

begin
  call smark(sp)

  call salloc (dtype, SZ_LINE, TY_CHAR)
  call salloc (comment, SZ_LINE, TY_CHAR)

  len = qp_queryf (qp, macro, Memc[dtype], maxlen, Memc[comment], flags)
  call malloc (strptr, len+1, TY_CHAR)
  nread = qp_read (qp, macro, Memc[strptr], len, 1, "macro")
  Memc[strptr+len] = EOS

  call sfree (sp)
end


# QM_INDIRECT -- Read a value string from a file

procedure qm_indirect (file, strptr)

char	file[ARB]	# file name
pointer	strptr		# ptr to output string

int		fd, len, pos
char	c

int		open(), getc()
bool	streq()

errchk open, close

begin 
  if (streq(file[2], "stdin")) fd = STDIN
  else fd = open (file[2], READ_ONLY, TEXT_FILE)

  len = ALLOC_UNIT
  pos = 0
  call malloc (strptr, len+1, TY_CHAR)

  while (getc(fd, c) != EOF) {
	if (c == '\n') break
	pos = pos + 1
	if (pos > len) {
	  len = len + ALLOC_UNIT
	  call realloc (strptr, len+1, TY_CHAR)
	}
	Memc[strptr+pos-1] = c
  }
  Memc[strptr+pos] = EOS

  if (fd != STDIN) call close (fd)
end


# QM_MACROEXISTS -- Check if a macro exists in a QPOE file header

int procedure qm_macroexists (qp, macro)

pointer	qp
char	macro[ARB]

int		len, maxlen, flags
pointer	sp, dtype, comment

int		qpaccessf(), qp_queryf()
bool	streq()

errchk qp_queryf, qp_accessf

begin
  if (qpaccessf (qp, macro) == YES) {
	call smark(sp)

	call salloc (dtype, SZ_LINE, TY_CHAR)
	call salloc (comment, SZ_LINE, TY_CHAR)

	len = qp_queryf (qp, macro, Memc[dtype], maxlen, Memc[comment], flags)

	if (streq(Memc[dtype], "macro")) {
	  call sfree (sp)
	  return YES
	}

	call sfree (sp)

	return NOTMACRO
  }

  return NO
end


# print a long string to a file descriptor

procedure fprlstr(fd, str)
int	fd			#I file descriptor
pointer	str			#I string to be printed

pointer	tmp

begin
  for (tmp = str; Memc[tmp] != EOS; tmp = tmp + 1) {
    call putc(fd, Memc[tmp])
  }
end

# Pass a string to a printf statement, enclosing the string
# in quotes if it contains any whitespace.

procedure prlstr(str)

pointer	str		# string to be printed

begin
  call fprlstr(STDOUT, str)
end


# Pass a string to a printf statement, enclosing the string
# in quotes if it contains any whitespace.

procedure eprlstr(str)

pointer	str		# string to be printed

begin
  call fprlstr(STDERR,str)
end



