# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  qpl_cmd.x
# 	Author	:  ericco
# 	Date	:  10/29/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) qpl_cmd.x -r1.10 10/29/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/14/97   mabbott	remove illegal QPOE private data and function dependencies
# 	07/16/93   akeith	corrected the TRUE/FALSE YES/NO problem
# 	03/24/93   akeith	changed version string
# 	05/14/92   akeith	documentation
# 	04/07/92   akeith	Changed the help page to coorsp. to the commands.
# 	04/03/92   akeith	All debug fprintf statements where removed.
# 	12/12/91   akeith	no
# 	10/30/91   akeith	documentation was added
# 	10/29/91   ericco	date and time created 91/10/29 15:09:52 by ericco
# H-
# 
#

#       Version: @(#)qpl_cmd.x	1.10 10/14/97 UCB CEA EGO


include <qpeuv.h>
include "qpl.h"

#
#	Command definintions
#

define	HELPTXT "[Hit any key to continue] "
define	ILLEGAL	"[?=help]"

define	HELP		'?'		# print helptxt
define	QUIT		'q'		# return to CL
define	FWD_SCREEN	'f'		# forward one full screen
define	BACK_SCREEN	'b'		# back one full screen
define	SCROLL_DOWN	'd'		# forward half a screen
define	SCROLL_UP	'u'		# back half a screen
define	PREV_LINE	'k'		# back one line
define	NEXT_LINE	'j'		# forward one line
define	LEFT		'h'		# left one column
define	RIGHT		'l'		# right one column
define	TO_BOF		'.'		# to beginning of file
define	TO_EOF		'G'		# to end of file
define	TO_EOF_ALT	'g'		# to end of file
define	REDRAW		'\014'		# <ctrl/l> redraw screen

define	NEXT_FILE	'N'		# goto next bucket
define	PREV_FILE	'P'		# goto previous bucket
define	TOP_FILE	'T'		# goto to top of bucket
define	NEXT_FILE_ALT	'\016'		# <ctrl/n>
define	PREV_FILE_ALT	'\020'		# <ctrl/p>


#	Reads commands from the command line and carries them out.
#
#	Structure values of qpl which may be reset:
#
#		QPL_EVI()	current event at top of page
#		QPL_TPEV()	current index within bucket at top of page
#		QPL_CURBK()	current bucket which means a change in
#					the evs (event struct.) array.
#

bool procedure qpl_cmd (qpl, clmn, fname, maxlines)

pointer         qpl                             #I/O The main structure
pointer		clmn				#I/O The column structure
char		fname[ARB]			#I file name
int		maxlines			#I part of screen for a page

pointer		io				# file ptr
pointer		dd				# data descriptor
int		tty				# tty descriptor

int		nev				# number of events in file
int		cmd				# command entered
bool		again				# true if we need to get cmd again
bool		ret				# return value

#int		tpev_tmp			# tmp values
int		evi_tmp

int		i
bool		done

int		pg_getcmd()			# gets command; returns cmd
int		qpl_move()			# move top_evi for scrolling
#int		qpl_gbckt()			# gets new bckt, sets evs array
int		qpl_topevi()			# sets evi, top_evi

begin

	io = QPL_IO(qpl)
	dd = QPL_DD(qpl)
	tty = QPL_TTY(qpl)

	ret = FALSE

	nev = QPL_NEVENTS(qpl)

	cmd = pg_getcmd (tty, fname, QPL_EVI(qpl), nev, QPL_EVI(qpl)+1, 1, 1)

	repeat {

		again = FALSE

		switch (cmd) {

		  case HELP		:

			call qpl_shwhelp(qpl, maxlines)
			cmd = pg_getcmd(tty,HELPTXT,evi_tmp,nev,evi_tmp+1,1,1)

		  case REDRAW		:

			QPL_REDRAW(qpl) = YES

		  case QUIT		:

			ret = TRUE

		  case FWD_SCREEN, ' '	:

			QPL_CURBK(qpl) = qpl_move(qpl, maxlines, maxlines, DOWN)

		  case BACK_SCREEN	:

			QPL_CURBK(qpl) = qpl_move(qpl, maxlines, maxlines, UP)

		  case SCROLL_DOWN	:

			QPL_CURBK(qpl) = qpl_move(qpl, maxlines/2, maxlines, DOWN)

		  case SCROLL_UP	:

			QPL_CURBK(qpl) = qpl_move(qpl, maxlines/2, maxlines, UP)

		  case PREV_LINE	:

			QPL_CURBK(qpl) = qpl_move(qpl, 1, maxlines, UP)

		  case NEXT_LINE	:

			QPL_CURBK(qpl) = qpl_move(qpl, 1, maxlines, DOWN)

		  case LEFT		:

			i = 1

			while(CLM_ON(CLM_PTR(clmn+i)) == NO) i = i + 1

			if(i != 1) {

				CLM_ON(CLM_PTR(clmn+i-1)) = YES

				done = FALSE
				while(i <= EUVDD_NFIELDS(dd) && !done){
					if(CLM_ON(CLM_PTR(clmn+i)) == NO){
						done = TRUE
					} else i = i + 1
				}

				CLM_ON(CLM_PTR(clmn+i-1)) = NO
				QPL_REDRAW(qpl) = YES

			}


		  case RIGHT		:

			i = EUVDD_NFIELDS(dd)

			while(CLM_ON(CLM_PTR(clmn+i)) == NO) i = i - 1

			if(i != EUVDD_NFIELDS(dd)) {

				CLM_ON(CLM_PTR(clmn+i+1)) = YES

				done = FALSE
				while(i >= 1 && !done){
					if(CLM_ON(CLM_PTR(clmn+i)) == NO){
						done = TRUE
					} else i = i - 1
				}

				CLM_ON(CLM_PTR(clmn+i+1)) = NO
				QPL_REDRAW(qpl) = YES

			}


		  case TO_BOF		:

			QPL_EVI(qpl) = 1
			QPL_CURBK(qpl) = qpl_topevi(qpl, DOWN)

		  case TO_EOF		:

			QPL_EVI(qpl) = nev
			QPL_CURBK(qpl) = qpl_topevi(qpl, UP)

		  case TO_EOF_ALT	:

			QPL_EVI(qpl) = nev
			QPL_CURBK(qpl) = qpl_topevi(qpl, UP)

#
#  MJA Oct 9 1997
#
#   remove all movement by bucket
#
		  case NEXT_FILE, NEXT_FILE_ALT	:

#			evi_tmp = QPL_EVI(qpl)
#			tpev_tmp = QPL_TPEV(qpl)

#			QPL_EVI(qpl) = IO_BKLASTEV(io) + 1

#			if(QPL_EVI(qpl) <= nev) {

#				QPL_CURBK(qpl) = qpl_topevi(qpl, DOWN)

#				if( QPL_CURBK(qpl) == EOF ){

#					QPL_EVI(qpl) = evi_tmp
#					QPL_TPEV(qpl) = tpev_tmp
#					QPL_CURBK(qpl) = qpl_gbckt(qpl)

#				} 

#			} else QPL_EVI(qpl) = evi_tmp

		  case PREV_FILE, PREV_FILE_ALT	:

#			evi_tmp = QPL_EVI(qpl)
#			tpev_tmp = QPL_TPEV(qpl)

#			QPL_EVI(qpl) = IO_BKFIRSTEV(io) - 1

#			if(QPL_EVI(qpl) >= 1) {

#				QPL_CURBK(qpl) = qpl_topevi(qpl, UP)

#				if(QPL_CURBK(qpl) == EOF){

#					QPL_EVI(qpl) = evi_tmp
#					QPL_TPEV(qpl) = tpev_tmp
#					QPL_CURBK(qpl) = qpl_gbckt(qpl)

#				} else {

#					QPL_EVI(qpl) = IO_BKFIRSTEV(io)
#					QPL_CURBK(qpl) = qpl_topevi(qpl, DOWN)

#				}

#			} else QPL_EVI(qpl) = evi_tmp

		  case TOP_FILE		:

#			QPL_EVI(qpl) = IO_BKFIRSTEV(io)
#			QPL_TPEV(qpl) = 0

#			QPL_CURBK(qpl) = qpl_topevi(qpl, DOWN)

		  default		:
			evi_tmp = QPL_EVI(qpl)
			cmd = pg_getcmd(tty,ILLEGAL,evi_tmp,nev,evi_tmp+1,1,1)
			again = TRUE
		}

	} until (again == FALSE)

	return ret

end


#
# qpl_shwhelp() displays the help page of commands.
#

procedure qpl_shwhelp(qpl, maxlines)

pointer         qpl                             #I/O The main structure
int		maxlines			#I part of screen for a page

int		i
int		tty				# tty descriptor

begin


	tty = QPL_TTY(qpl)

#
#  MJA Oct 9 1997
#
#  several changes to account for two less lines in use at top of page
#
#	call ttygoto(STDOUT, tty, 1, 6)
	call ttygoto(STDOUT, tty, 1, 4)

	do i = 1, maxlines {
		call ttyclearln(STDOUT, tty)
		call ttyputline(STDOUT, tty, "\n", QPL_MAPCC(qpl))
	}

#	call ttygoto(STDOUT, tty, 1, 6)
	call ttygoto(STDOUT, tty, 1, 4)
	call ttyclearln(STDOUT, tty)
	call ttyputline(STDOUT, tty, "\n", QPL_MAPCC(qpl))

	call ttyputline(STDOUT, tty,
		" ?		print helptxt\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" q		return to CL\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" f		forward one full screen\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" b		back one full screen\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" d		forward half a screen\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" u		back half a screen\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" k		back one line\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" j		forward one line\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" h		left one column\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" l		right one column\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" .		beginning of file\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" G, g		end of file\n", QPL_MAPCC(qpl))
	call ttyputline(STDOUT, tty,
		" ^l		redraw screen\n", QPL_MAPCC(qpl))
#	call ttyputline(STDOUT, tty,
#		" N, ^n		next bucket\n", QPL_MAPCC(qpl))
#	call ttyputline(STDOUT, tty,
#		" P, ^p		previous bucket\n", QPL_MAPCC(qpl))
#	call ttyputline(STDOUT, tty,
#		" T		top of bucket\n", QPL_MAPCC(qpl))

#	do i = 18, maxlines {
	do i = 15, maxlines {
		call ttyclearln(STDOUT, tty)
		call ttyputline(STDOUT, tty, "\n", QPL_MAPCC(qpl))
	}

	QPL_REDRAW(qpl) = YES

end

