# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  qpl_list.x
# 	Author	:  ericco
# 	Date	:  10/29/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) qpl_list.x -r1.10 10/29/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/14/97   mabbott	remove illegal QPOE private data and function dependencies
# 	07/16/93   akeith	corrected the TRUE/FALSE YES/NO problem
# 	03/24/93   akeith	changed version string
# 	06/17/92   ericco	added qpl_gclmn routine
# 	04/07/92   akeith	Changed the help page to coorsp. to the commands.
# 	04/03/92   akeith	Took out debug statements and fixed maxcols.
# 	12/12/91   akeith	no
# 	10/30/91   akeith	documentation was added
# 	10/29/91   ericco	date and time created 91/10/29 15:09:54 by ericco
# H-
# 
#

#       Version: @(#)qpl_list.x	1.10 10/14/97 UCB CEA EGO


include <qpeuv.h>
include <qpioset.h>
include "qpl.h"

#	This is the main procedure.
#	After doing various initializations including allocating memory
#		for the evs array, format string and the lbuf (line buffer)
#		string, its main task is to repeatedly call qpl_cmd() to 
#		execute the commands entered by the user.
#	In addition, it is responsible for printing to the tty the event
#		list.
#
#	Structure values of qpl which initialized:
#
#		QPL_DD()	pointer to domain descriptor
#		QPL_EVS()	pointer to the evs array
#		QPL_EVSLN()	evs array length = bucket length
#		QPL_EVI()	current event at top of page
#		QPL_TPEV()	current index within bucket at top of page
#		QPL_CURBK()	current bucket number
#		QPL_FRMT()	pointer to format string
#

int  procedure  qpl_list(qpl, clmn, fname)

pointer		qpl				#I/O The main structure
pointer		clmn				#I/O The column structure
char		fname[ARB]			#I ptr to filename

	# Elements of the structure

pointer		dd				# data descriptor
int		tty				# tty descriptor

pointer		sp				# stack ptr
pointer		evs				# ptr to event strct array
pointer		format				# ptr to format string
pointer		lbuf				# ptr to line buf string

int		i
int		tpev_tmp
int		evi_tmp
int		out_lines			# num. lines outputed to the screen

char		number[20]

int		ncols, nlines			# raw values pert. to screen
int		maxlines, maxcols		# part of screen for a page
int		old_bckt			# for print bckt info to screen
int		retval				# return status

pointer	eventlist

int		qpl_topevi()			# sets evi, top_evi init.
bool		qpl_cmd()			# gets and executes commands
pointer		qpl_gclmn()			# get column definition
pointer		qpio_stati()
int			qp_lenf()
begin


	retval = 1
	old_bckt = -1

	dd = QPL_DD(qpl)
	tty = QPL_TTY(qpl)

#  Get the number of columns and lines for the tty.
#  This doesn`t work very well, because iraf is unable to obtain
#	window size from the unix env. varibles.

	call xttysize (ncols, nlines)
#	maxlines = nlines - 1 - 5
#  MJA 9 Oct 97 - removed two output lines with bucket min/max
	maxlines = nlines - 1 - 3
	maxcols = ncols

# check that the initial event index is within the range
#	of existing events.

  eventlist = qpio_stati(QPL_IO(qpl), QPIO_PARAMP)
  QPL_NEVENTS(qpl) = qp_lenf(QPL_QP(qpl), Memc[eventlist])

	if(QPL_NEVENTS(qpl) == 0){
		call printf("\n No events in file. \n")
		retval = 0
		return retval
	}

	if ( QPL_EVI(qpl) > QPL_NEVENTS(qpl) ) {
		QPL_EVI(qpl) = QPL_NEVENTS(qpl)
	} else if ( QPL_EVI(qpl) < 1 ) QPL_EVI(qpl) = 1

# set up the stack and buffers

        call smark(sp)
        call salloc(evs, QPL_EVSLN(qpl), TY_INT)
        call salloc(format, 10 + EUVDD_NFIELDS(dd) * 10, TY_CHAR)
        call salloc(lbuf, 15 + EUVDD_NFIELDS(dd) * 15, TY_CHAR)

# init. the qpl structure

	QPL_EVS(qpl) = evs
	QPL_FRMT(qpl) = format

# initialize column structure

	clmn = qpl_gclmn(dd)

# initialize evi and top_evi to the first event which makes it through
#	the filter at or after the initial evi set by the user.

	QPL_CURBK(qpl) = qpl_topevi(qpl, DOWN)

	if(QPL_CURBK(qpl) == EOF){
		call printf("\n Sorry, event given and all those below it, \n")
		call printf("      have been filtered out. \n\n")
		call sfree(sp)
		retval = 0
		return retval
	}


# Create and store the format and the columns real names while also 
#     initializing the columns macro names.

        call qpl_gfrmt(qpl, clmn)

# Look in the header for those macro names representing the columns
#	and store those.

	call qpl_ghdr(qpl,clmn)

# clear screen

	call ttyclear (STDOUT, tty)

# set the format and print the header to the screen

        call qpl_phdr(qpl, clmn, maxcols)

# print the global info to the screen

        call qpl_pglob(qpl, clmn)

#
# This is where all the action is. The main work done here
#	is printing to tty and getting commands from qpl_cmd()
#	The action of those commands are carried out in qpl_cmd()
#	while the results are printed to the tty here.
#

        repeat {

		if(QPL_REDRAW(qpl) == YES){
			call qpl_phdr(qpl, clmn, maxcols)
			call qpl_pglob(qpl, clmn)
			#  MJA 9 Oct 97 - removed two output lines with bucket min/max
            #call qpl_wbckt(qpl, clmn)
			QPL_REDRAW(qpl) = NO
		}

        if(QPL_CURBK(qpl) != old_bckt){	     # updates the bckt info
		  #  MJA 9 Oct 97 - removed two output lines with bucket min/max
		  #call qpl_wbckt(qpl, clmn)    #   and writes to tty
		  old_bckt = QPL_CURBK(qpl)
		}

		call ttygoto(STDOUT, tty, 1, 4)
		#  MJA 9 Oct 97 - removed two output lines with bucket min/max
#		call ttygoto(STDOUT, tty, 1, 6)

		tpev_tmp = QPL_TPEV(qpl)	
		evi_tmp = QPL_EVI(qpl)
		out_lines = 1

#	writes events to tty for a full page.

		while(tpev_tmp < QPL_EVSLN(qpl) && out_lines <= maxlines){

			if(Memi[evs+tpev_tmp] != 0){

				call sprintf(number, ARB, "%d")
				call pargi(evi_tmp)

				call sprintf(Memc[lbuf], ARB, Memc[format])
				if(CLM_ON(CLM_PTR(clmn)) == YES)
					call pargstr(number)
				call qpl_format(qpl, clmn, Memi[evs+tpev_tmp])

				call ttyclearln(STDOUT, tty)
				call ttyputline (STDOUT, tty, 
						Memc[lbuf], QPL_MAPCC(qpl))

				out_lines = out_lines + 1

			} 

			tpev_tmp = tpev_tmp + 1
			evi_tmp = evi_tmp + 1

		}

#	if written events didn't make it to the bottom of the page, clear lines.

                do i = out_lines, maxlines {
                        call ttyclearln(STDOUT, tty)
                        call ttyputline (STDOUT, tty, "\n", QPL_MAPCC(qpl))
                }

        } until ( qpl_cmd (qpl, clmn, fname, maxlines) )

	call sfree(sp)

	return retval

end

