# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  readtabs.x
# 	Author	:  tnk
# 	Date	:  08/24/93
# 	Synopsis:  
# 	Keywords:
# 		@(#) readtabs.x -r1.12 08/24/93 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/02/97   bboyd	removed private qpoe headers, now using qp_parseeuv
# and qp_dtypeeuv in the euv library
# 	12/08/95   tnk	turned off inherit for event lists
# 	09/12/95   tnk	explicitly set QPOE inherit flag
# 	03/03/95   tnk	added trimwhite to table header reads (tables1.3.3 change)
# 	08/25/94   tnk	eliminated unused args
# 	08/09/94   tnk	restored +2 for bucketlen, only set qpoe_bucketlen
# 	08/09/94   tnk	new WCS header card handling
# 	08/08/94   tnk	doing work-arounds for portability problems
# 	06/02/94   tnk	forced bucketlen to 2048;  added routines to read WCS cards
# 	01/03/94   tnk	in string column, convert null to blank (FITS problem)
# 	11/23/93   tnk	now handles ST table shorts
# 	08/24/93   tnk	date and time created 93/08/24 15:04:39 by tnk
# H-
# 
#


include <qpeuv.h>
include <qpioset.h>
include <qpset.h>
include <iraf.h>
include <ctype.h>
include <tbset.h>

# @(#)readtabs.x	1.12 10/2/97 UCB CEA EGO

define QPSYMBOL_HDR "QPSYMBOL"
define QPDTYPE_HDR  "QPDTYPE"
define QPCMNT_HDR   "QPCMNT"
define QPMAXEL_HDR  "QPMAXEL"
define QPFLDLST_HDR "QPFLDLST"
define QPFLCMNT_HDR "QPFLCMNT"
define QPBUCKLN_HDR "QPBUCKLN"

define DEF_BUCKETLEN 2048      # default bucket length

#***********************************
# process an ST table (add its contents to the QPOE file)
#
procedure process_tab(qp, tabname, do_events, verbose, warnmsgs,
   mkindex,obucketlen,binary_wcs)
pointer qp        #I/O opened QPOE file
char tabname[ARB] #I name (including path) of ST table to add to qp
bool do_events    #I flag: do the event lists or do the non-event stuff
bool verbose      #I print informational messages
bool warnmsgs     #I print warning messages
bool mkindex      #I make a qpoe index
int  obucketlen   #I override bucket length
bool binary_wcs   #I use binary wcs info

# LOCAL VARIABLES
pointer sp        # stack pointer
pointer datatype  # parameter data type
pointer parmname  # parameter name
int     maxelem   # allocated length of parameter
pointer comment   # comment describing parameter
pointer fieldlist # field list descriptor
pointer fldcmnt   # field list descriptor comment
int     flags     # parameter flags
int     spp_dtype # spp datatype (number)
pointer tp        # table pointer
int     n_rows
int     fieldlen

# FUNCTIONS
int     tbpsta()
pointer tbtopn()
int     tbhgti()
int     qp_dtypeeuv()
bool    qp_accessf()
int     strlen()
int     tbtacc()
bool    fetch_header()


begin

   # either do all the event lists 
   # or everything BUT the event lists

   call smark(sp)

   call salloc(datatype,  SZ_LINE, TY_CHAR)
   call salloc(parmname,  SZ_LINE, TY_CHAR)
   call salloc(comment,   SZ_LINE, TY_CHAR)
   call salloc(fieldlist, SZ_LINE, TY_CHAR)
   call salloc(fldcmnt,   SZ_LINE, TY_CHAR)

   # open the table, get the appropriate information
   if (tbtacc(tabname) == NO) {
      if (warnmsgs) {
         call eprintf("Warning: Can't access ST table '%s'\n")
         call pargstr(tabname)
      }
      return
   }

   tp = tbtopn(tabname, READ_ONLY, 0)

   # maxelem is number of rows in table
   n_rows = tbpsta (tp, TBL_NROWS)

   if (n_rows == 0) {
      if (warnmsgs) {
         call eprintf("Warning: table %s is empty\n")      
         call pargstr(tabname)
      }
      call tbtclo(tp)
      return
   }
   

   # look for the QP symbol name, datatype, comment, maxelem cards

   # must have QP symbol name
   if ( ! fetch_header(tp, tabname, QPSYMBOL_HDR, Memc[parmname], 
      SZ_LINE, warnmsgs)) {
      call tbtclo(tp)
      return
   }

   # must have QP datatype
   if ( ! fetch_header(tp, tabname, QPDTYPE_HDR, Memc[datatype], 
      SZ_LINE, warnmsgs) ) {
      call tbtclo(tp)
      return
   }

   # comment card is optional
   iferr {
      call tbhgtt(tp, QPCMNT_HDR,   Memc[comment], SZ_LINE)
   } then {
      call strcpy("", Memc[comment], SZ_LINE)
   }

   # max elements card is optional
   iferr {
      maxelem = tbhgti(tp, QPMAXEL_HDR)
   } then {
      maxelem = 0
   }


   # lookup the SPP datatype for this parameter
   spp_dtype = qp_dtypeeuv(qp, Memc[datatype])

   # copy the data from the table to the qp file

   if (spp_dtype == TY_USER || spp_dtype == NULL) {


      if (do_events) {

         # don't set inherit flag for event lists
	 # (actually, all user defined types, unfortunately)
         flags=0

         if (verbose) {
            call eprintf("file:%s\n")
            call pargstr(tabname)
         }

         # get the field descriptor string and its comment (if any)
         if ( ! fetch_header(tp, tabname, QPFLDLST_HDR, Memc[fieldlist], 
            SZ_LINE, warnmsgs)) {
            call tbtclo(tp)
            return
         }

         # comment card is optional
         iferr {
            call tbhgtt(tp, QPFLCMNT_HDR, Memc[fldcmnt], SZ_LINE)
         } then {
            call strcpy("", Memc[fldcmnt], SZ_LINE)
         }

         # does this entry already exist?
         if (qp_accessf(qp, Memc[datatype])) {
            if (warnmsgs) {
               call eprintf("Warning: user field '%s' already exists.\n")
               call pargstr(Memc[datatype])
               call eprintf("Using previous definition.\n")
            }
         }
         else {
            iferr {
               fieldlen=strlen(Memc[fieldlist])
               call qp_addf(qp, Memc[datatype], Memc[fieldlist], fieldlen, 
                  Memc[fldcmnt], flags)
            } then {
               if (warnmsgs) { 
                   call eprintf("Warning: couldn't added user type definition\n")
                   call eprintf("'%s=%s'.\n")
                   call pargstr(Memc[datatype])
                   call pargstr(Memc[fieldlist])
               }
            }
         }

         # now add the actual user data structure
         call qp_addf(qp, Memc[parmname], Memc[datatype], maxelem, 
            Memc[comment], flags)

         call copy2dtab(tp, qp, Memc[parmname], Memc[datatype],
            n_rows, Memc[fieldlist], verbose, tabname, warnmsgs,
            mkindex,obucketlen,binary_wcs)
      }   

   }
   else {

      if (! do_events ) {

	 flags=QPF_INHERIT

         if (verbose) {
            call eprintf("file:%s\n")
            call pargstr(tabname)
         }

         # process a regular vector (not a user type)

         call qp_addf(qp, Memc[parmname], Memc[datatype], maxelem, 
            Memc[comment], flags)

         call copy1dtab(tp, qp, Memc[parmname], Memc[datatype], spp_dtype, 
            n_rows, tabname, warnmsgs)
      }
   }

   call sfree(sp)

   call tbtclo(tp)

end
#*********************************************
# get a string header from an ST table.  Print error message (optional).
bool procedure fetch_header(tp, tabname, card, value, valuelen, warnmsgs)
pointer tp             #I table pointer
char    tabname[ARB]   #I table name
char    card[ARB]      #I header card to look up
char    value[ARB]     #O returned value
int     valuelen        #I len of value buffer
bool    warnmsgs       #I boolean: print warning message

bool status

begin

   status=true

   iferr {
      call tbhgtt(tp, card, value, valuelen)
   } then {
      status=false
      if (warnmsgs) {
         call eprintf("Warning: %s header card not found in table %s\n")
         call pargstr(card)
         call pargstr(tabname)
      } 

   }
   call trimwhite(value)

   return (status)

end
#**********************************************
# copy a single data column from an ST table to a QP file
#
procedure copy1dtab(tp, qp, parmname, qpdatatype, spp_dtype, n_rows, tabname,
   warnmsgs)
pointer tp              #I input table
pointer qp              #I/O opened QPOE file pointer
char    parmname[ARB]   #I name of QPOE parameter to write
char    qpdatatype[ARB] #I datatype to copy
int     spp_dtype       #I spp data type (number)
int     n_rows          #I no. of rows in the ST table
char    tabname[ARB]    #I name of table
bool    warnmsgs        #I print warning messages

# LOCAL VARIABLES
pointer tp_colptr   # table column pointer
int     st_dtype
pointer sp
int     firstrow
int     lastrow
pointer buff
pointer nullflag

# FUNCTIONS
pointer tbcnum()

begin

   call smark(sp)

   # convert datatypes
   call spp_stdtype(spp_dtype,st_dtype)

   call salloc(buff, n_rows, st_dtype)
   call salloc(nullflag, n_rows, TY_BOOL)

   # grab the entire column
   firstrow=1
   lastrow=n_rows	

   # open first (and presumably only) column
   tp_colptr = tbcnum(tp, 1)

   if (tp_colptr == NULL) {
      if (warnmsgs) {
         call eprintf("Warning:  Cannot access first column of table %s\n")
         call pargstr(tabname)
      }
   }
   else {
  
      # read the entire column & write it to the qpoe file

      switch(st_dtype) {

         case TY_BOOL:
            call tbcgtb(tp, tp_colptr, Memb[buff], Memb[nullflag], firstrow,
               lastrow)
            call qp_write(qp, parmname, Memb[buff], n_rows, firstrow, 
               qpdatatype)

         case TY_CHAR:
            call read_str_col(tp, tp_colptr, Memc[buff], n_rows, firstrow, 
               lastrow)
            call qp_write(qp, parmname, Memc[buff], n_rows, firstrow, 
               qpdatatype)

         case TY_INT:
            call tbcgti(tp, tp_colptr, Memi[buff], Memb[nullflag], firstrow,
               lastrow)
            call qp_write(qp, parmname, Memi[buff], n_rows, firstrow, 
               qpdatatype)

         case TY_REAL:
            call tbcgtr(tp, tp_colptr, Memr[buff], Memb[nullflag], firstrow,
               lastrow)
            call qp_write(qp, parmname, Memr[buff], n_rows, firstrow, 
               qpdatatype)

         case TY_DOUBLE:
            call tbcgtd(tp, tp_colptr, Memd[buff], Memb[nullflag], firstrow,
               lastrow)
            call qp_write(qp, parmname, Memd[buff], n_rows, firstrow, 
               qpdatatype)
 
         default:
            call eprintf("can't handle data type: %d %s\n")
            call pargi(spp_dtype)
            call pargstr(qpdatatype)
      }

   }

   call sfree(sp)

end
#**********************************************
# copy a 2d data table from an ST table to a QP file
#
procedure copy2dtab(tp, qp, parmname, qpdatatype, n_rows,
  fieldlist,verbose,tabname,warnmsgs,mkindex,obucketlen,binary_wcs)

pointer  tp              #I input table
pointer  qp              #I/O opened QPOE file pointer
char     parmname[ARB]   #I name of QPOE parameter to write
char     qpdatatype[ARB] #I datatype to copy
int      n_rows          #I no. of rows in the ST table
char     fieldlist[ARB]  #I user field descriptor
bool     verbose         #I informational output
char     tabname[ARB]    #I name of st table
bool     warnmsgs        #I print warning messages
bool 	 mkindex         #I make a qpoe index
int      obucketlen      #I override bucket length
bool     binary_wcs      #I use binary wcs info

#LOCAL VARS
pointer sp
pointer dd
pointer qpio
int n_fields
int i
int n_cols
pointer tp_colptrs    # column pointers

# FUNCTIONS
int     qp_parseeuv()
int     tbpsta()
pointer tbcnum()

begin

   call smark(sp)

   call salloc(dd, LEN_EUVDDDES, TY_INT)
  
   # parse structure description string
   n_fields = qp_parseeuv(qp, fieldlist, dd)

   n_cols = tbpsta(tp, TBL_NCOLS)

   if (n_cols != n_fields) {
      if (warnmsgs) {
         call eprintf("Warning: Number of fields in QP field descriptor %s\n")
         call pargstr(fieldlist)
         call eprintf("does not match number of columns in ST table %s\n")
         call pargstr(tabname)
      }
      return
   }

   call salloc(tp_colptrs, n_cols, TY_POINTER)

   # open ST columns
   for (i=1; i<= n_cols; i=i+1) {
     Memi[tp_colptrs+i-1] = tbcnum(tp,i)
   }


   # see if this is an event list, ie, if
   # the x and y index fields are nonzero
   # and process accordingly

   if (EUVDD_XFIELD(dd) != 0 && EUVDD_YFIELD(dd) != 0) {
      # an event list

      # copy data from ST table to QPOE file   
      call get_ev_2d_data(tp,Memi[tp_colptrs],n_cols,n_rows,qp,qpio,dd,
         parmname,verbose,warnmsgs,tabname,mkindex,obucketlen,binary_wcs)

   }
   else {
      # a user-defined (non-event) list
      call get_user_2d_data(tp,Memi[tp_colptrs],n_cols,n_rows,qp,dd,
         parmname,qpdatatype)
   }

   call sfree(sp)

end

#**********************************************
# read a string written as an ST column, 1 char per row
#
procedure read_str_col(tp, tp_colptr, buff, bufflen, firstrow, lastrow)
pointer tp     #I/O output table pointer
pointer tp_colptr  #I column pointer
char buff[ARB]     #O char buffer
int  bufflen       #I length of buff
int  firstrow      #I first row to store text in
int  lastrow       #I last row to store text in

char line[SZ_LINE]       #one line of text (char + EOS)
int  i
int row

begin

  # take first character from each row, stuff it into a string buffer

  i = 1
  for (row=firstrow; row<=lastrow; row=row+1) {
     call tbegtt(tp, tp_colptr, row, line, SZ_LINE)
     if (line[1] == '\0') line[1]=' '   # restore null to blank
     buff[i]=line[1]
     i=i+1
  }

  # null terminate buffer
  line[i] = EOS

end

#****************************************************
# convert spp datatype to ST compatible datatype
#
procedure spp_stdtype(spp_dtype,st_dtype)
int spp_dtype       #I QPOE related spp datatype
int st_dtype        #O SPP datatype value

begin

switch(spp_dtype) {
   case TY_BOOL:
                st_dtype = TY_BOOL
   case TY_CHAR, TY_MACRO:
                st_dtype = TY_CHAR
   case TY_SHORT:
		st_dtype = TY_SHORT
   case TY_INT, TY_LONG, TY_OPAQUE:
                st_dtype = TY_INT
   case TY_REAL:
                st_dtype = TY_REAL
   case TY_DOUBLE:
                st_dtype = TY_DOUBLE
   case TY_COMPLEX:
                st_dtype = -1          # undefined
   default:   
                st_dtype = -1          # user defined
   }
end

#**********************************************************
# read the event list data out of an ST table & put into QPOE file
# (using qpio)
#
procedure get_ev_2d_data(tp,tp_colptrs,n_cols,n_rows,qp,qpio,dd,parmname,
   verbose,warnmsgs,tabname,mkindex,obucketlen,binary_wcs)

pointer tp          #I table pointer
int tp_colptrs[ARB] #I column pointers
int n_cols          #I number of columns in ST table
int n_rows          #I number of rows in ST table
pointer qp          #I/O QPOE file pointer
pointer qpio        #I/O qp io pointer
pointer dd          #I data descriptor
char parmname[ARB]  #I name of QPOE parameter
bool verbose        #I informational messages
bool warnmsgs       #I warning messages
char tabname[ARB]   #I name of table
bool mkindex        #I make a qpoe index
int  obucketlen     #I override bucket length
bool binary_wcs     #I use binary wcs info

# LOCAL VARS
int     row
pointer sp
pointer ev
int     ev_len     # length in bytes of event-like structure
int     i
int     offset
int     dtype
int     pct
int     oldpct
int     bucketlen

# FUNCTIONS
pointer qpio_open()
int     qpio_stati()
int     tbhgti()

begin

   call smark(sp)

   # get ready for QP I/O
   qpio = qpio_open (qp, parmname, NEW_FILE)

   # get QPIO bucketlen from ST table header (optional)
   bucketlen = DEF_BUCKETLEN   #internal default
   iferr {
      bucketlen = tbhgti(tp, QPBUCKLN_HDR)
      bucketlen = bucketlen + 2   # restore QPOE_BUCKETLEN from QPIO value
   }
   then {
      if (warnmsgs) {
         call eprintf("Warning: %s header card not found in table %s\n")
         call pargstr(QPBUCKLN_HDR)
         call pargstr(tabname)
      }
   }

   # did user give an override bucket length?
   if (obucketlen != INDEFI)
   {
      bucketlen = obucketlen
      if (verbose)
      {
          call eprintf("forcing QPOE bucket length to %d\n")
          call pargi(bucketlen)
      }
   }

   # set bucket length (QPOE only, not QPIO)
   call qp_seti(qp, QPOE_BUCKETLEN, bucketlen)

   # allocate an event structure
   ev_len = qpio_stati(qpio, QPIO_EVENTLEN)
   call salloc (ev, ev_len / SZ_SHORT, TY_SHORT)

   # for each ST row, read data element from each column,
   # stuff these into a QP ev structure, 
   # and write that to the qpoe file

   oldpct = 0
   for (row=1; row<=n_rows; row=row+1) {
      for (i=1; i<=n_cols; i=i+1) {

         offset = EUVDD_FOFFSET(dd,i)
         dtype  = EUVDD_FTYPE(dd,i)

         # load QP structure field
         call st2qp_fld(ev,offset,dtype,tp,tp_colptrs[i],row)

         if (verbose) {
            pct = int(10.*real(row)/real(n_rows))*10
            if (pct != oldpct) {
               oldpct = pct
               call eprintf ("  %3d%% read\n")
               call pargi (pct)
            }
         }
   
      }

      call qpio_putevents(qpio, ev, 1)

   }

   if (mkindex)
      call qpio_mkindex(qpio, "")

   call qpio_close(qpio)

   # set WCS from header cards stored in table?
   if (! binary_wcs)
   {
      if (verbose)
      {
         call eprintf("Loading WCS from header cards\n")
      }
      call qptabwcs(qp,tp)
   }

   call sfree(sp)

end

#**********************************************************
# read the user defined data out of an ST table & put into QPOE file
# (using qpwrite)
#
procedure get_user_2d_data(tp,tp_colptrs,n_cols,n_rows,qp,dd,parmname,
   qpdatatype)
pointer tp           # table pointer
int tp_colptrs[ARB]  # column pointers
int n_cols           # number of columns in ST table
int n_rows           # number of rows in ST table
pointer qp           # QPOE file pointer 
pointer dd           # data descriptor
char parmname[ARB]   # name of QPOE parameter
char qpdatatype[ARB] # data type of QPOE parameter (string)

# LOCAL VARS
int     row
pointer sp
pointer elem
int     elem_len         # length in bytes of element-like structure
int     i
int     offset
int     dtype

# FUNCTIONS
int sizeof()

begin

   call smark(sp)

   # figure out length of structure & allocate it
   elem_len = 0
   for (i=1; i<= EUVDD_NFIELDS(dd); i=i+1)
      elem_len = elem_len + sizeof(EUVDD_FTYPE(dd,i))
   call salloc (elem, elem_len / SZ_SHORT, TY_SHORT)

   # for each ST row, read data element from each column,
   # stuff these into a QP elem structure, 
   # and write that to the qpoe file

   for (row=1; row<=n_rows; row=row+1) {
      for (i=1; i<=n_cols; i=i+1) {

       offset = EUVDD_FOFFSET(dd,i)
       dtype  = EUVDD_FTYPE(dd,i)

       # load QP structure field
       call st2qp_fld(elem,offset,dtype,tp,tp_colptrs[i],row)

      }

      call qp_write(qp, parmname, Mems[elem], 1, row, qpdatatype)

   }

   call sfree(sp)

end
#**************************************************
# copy a field from an ST table to a qpoe structure
#
procedure st2qp_fld(qp_struct,offset,dtype,tp,tp_colptr,tab_row)
pointer qp_struct   #I QPOE user structure
int     offset      #I field offset into structure
int     dtype       #I QPOE field datatype
pointer tp          #I/O table pointer
pointer tp_colptr   #I/O table column pointer
int     tab_row     #I table row to write into


# LOCAL VARS
bool    bool_val
int     int_val
real    real_val
double  dbl_val
short   short_val

# FUNCTIONS
pointer coerce()

begin

   switch(dtype) {

      case TY_BOOL:
         call tbegtb(tp,tp_colptr, tab_row, bool_val)
         Memb[coerce(qp_struct,TY_SHORT,dtype) + offset] = bool_val

      case TY_SHORT:
         call tbegts(tp,tp_colptr, tab_row, short_val)
         Mems[coerce(qp_struct,TY_SHORT,dtype) + offset] = short_val

      case TY_INT:
         call tbegti(tp,tp_colptr, tab_row, int_val)
         Memi[coerce(qp_struct,TY_SHORT,dtype) + offset] = int_val

      case TY_LONG:
         call tbegti(tp,tp_colptr, tab_row, int_val)
         Meml[coerce(qp_struct,TY_SHORT,dtype) + offset] = long(int_val)

      case TY_REAL:
         call tbegtr(tp,tp_colptr, tab_row, real_val)
         Memr[coerce(qp_struct,TY_SHORT,dtype) + offset] = real_val

      case TY_DOUBLE:
         call tbegtd(tp,tp_colptr, tab_row, dbl_val)
         Memd[coerce(qp_struct,TY_SHORT,dtype) + offset] = dbl_val

      default:
          call eprintf("can't handle datatype %d\n")
          call pargi(dtype)
          call error(100,"bye")

      }

end
#**********************************************************
# Load a wcs from qprst table header cards into a qpoe
include	<imhdr.h>

procedure qptabwcs(qp,tb)
pointer qp
pointer tb

pointer	tempfile
pointer	value
#pointer path
pointer	sp
pointer im
pointer mw
int     ndim
int     ip
#int     len
pointer immap()
pointer mw_openim()
int     strlen()
int	tbhgti()
#int     getpath()

begin
	call smark (sp)
        call salloc(tempfile, SZ_LINE, TY_CHAR)
#        call salloc(path, SZ_LINE, TY_CHAR)
        call salloc(value, SZ_LINE, TY_CHAR)

#        # create temp image file in qpoe's directory
#        len =  getpath (QP_DFNAME(qp), Memc[path], SZ_LINE)
#        call strcat ("qprst", Memc[path], SZ_LINE)        
#        call mktemp(Memc[path], Memc[tempfile], SZ_LINE)

	# create a temp image header in tmp$
	call mktemp("tmp$qprst", Memc[tempfile], SZ_LINE)

        # open temporary image
        im = immap(Memc[tempfile], NEW_FILE, NULL)        
        if (im == NULL)
           call error(103,"cannot open new image")

        #get number of dimensions from WCSDIM
        iferr {
           ndim = tbhgti (tb, "WCSDIM")
        } then
        {
           # wcs dim stored as "COMMENT WCSDIM= n" ?
           Memc[value] = EOS
           call cmntval(tb,"WCSDIM",Memc[value],SZ_LINE)
           if (strlen(Memc[value]) <= 0)
           {
              call eprintf("Could not find WCSDIM information, ")       
              call eprintf("assuming ndim=2.\n")
              ndim=2
           } else {
   	      ip=1
   	      call ctoi(Memc[value],ip,ndim)
           }
        }

        IM_NDIM(im) = ndim

        call imaddi(im, "WCSDIM", ndim)

        # copy header cards from table to image
        call cp_tb_wcs(tb,im)

        call imunmap(im)
        im = immap(Memc[tempfile], READ_ONLY, NULL)        
        if (im == NULL)
           call error(103,"cannot open image read_only")

        # create mwcs from image & copy to qpoe
        mw = mw_openim (im)        
        call qp_savewcs (qp, mw)
        call mw_close (mw)

	# Clean up.
        call imunmap(im)
        call imdelete(Memc[tempfile])

	call sfree (sp)
end

#---------------------------------------------
# copy wcs info from table to image
procedure cp_tb_wcs(tb,im)
pointer tb    #I table of WCS cards
pointer im    #I/O image to copy cards to

pointer sp
pointer 

begin

   call smark(sp)

   call dblhdr(tb,im,"CRVAL1")
   call dblhdr(tb,im,"CRVAL2")
   call dblhdr(tb,im,"CRPIX1")
   call dblhdr(tb,im,"CRPIX2")
   call dblhdr(tb,im,"CDELT1")
   call dblhdr(tb,im,"CDELT2")
   call dblhdr(tb,im,"CD1_1")
   call dblhdr(tb,im,"CD2_2")
   call dblhdr(tb,im,"CD1_2")
   call dblhdr(tb,im,"CD2_1")
   call dblhdr(tb,im,"LTM1_1")
   call dblhdr(tb,im,"LTM2_2")
   call dblhdr(tb,im,"LTM1_2")
   call dblhdr(tb,im,"LTM2_1")
   call dblhdr(tb,im,"CROTA1")
   call dblhdr(tb,im,"CROTA2")

   call texthdr(tb,im,"WAT0_001")
   call texthdr(tb,im,"WAT1_001")
   call texthdr(tb,im,"WAT2_001")

end

#---------------------------------------------
# put out an image double header parameter from an ST table card
procedure dblhdr(tb,im,card)
pointer tb      #I/O output table  
pointer im      #I   input image   
char card[ARB]  #I   header card   

double d_val    #I   header double value
double tbhgtd()

begin

    ifnoerr {
        d_val = tbhgtd (tb, card)
    } then {
        call imaddd(im, card, d_val)
    }
end

#---------------------------------------------
# put out an image text header parameter from an ST table 
# card (or possibly encoded in a COMMENT card)
procedure texthdr(tb,im,card)
pointer tb      #I/O output table  
pointer im      #I   input image   
char card[ARB]  #I   header card   

pointer sp
pointer value

begin

    call smark(sp)
    call salloc(value, SZ_LINE, TY_CHAR)

    iferr {
        # text stored as regular header
        call tbhgtt(tb, card, Memc[value], SZ_LINE)
    } then {
       # text stored in "COMMENT <card>= <text>"
       call cmntval(tb,card,Memc[value],SZ_LINE)
    }
    # add new keyword & value to image
    call imastr(im,card,Memc[value])

    call sfree(sp)

end
#---------------------------------------------
# extract card value stored in a comment card
procedure cmntval(tb,key,value,maxvalue)
pointer tb         #I table of WCS cards
char    key[ARB]   #I
char    value[ARB] #O
int     maxvalue   #I

int i
pointer sp
pointer parmname
int     dtype
pointer buff
pointer card
int     last
int     index

bool    streq()
bool    strne()
int     strlen()
int     strmatch()

# the cards we want are of form:
# table input:
#     COMMENT HHHHHHHH=<sp><stuff>

begin

   call smark(sp)
   call salloc(parmname, SZ_LINE, TY_CHAR)
   call salloc(buff, SZ_PARREC, TY_CHAR)
   call salloc(card, SZ_LINE, TY_CHAR)

   i = 0
   call strcpy("xxx",Memc[parmname],SZ_LINE)
   while ( strne(Memc[parmname],""))
   {
      i = i + 1
      # get next header card from table
      call tbhgnp(tb, i, Memc[parmname], dtype, Memc[buff])

      if (streq(Memc[parmname],"COMMENT"))
      {

         # look for keyword
         # the cards we want are of form:
         # HHHHHHHH=<sp><stuff>
       
         index = strmatch(Memc[buff],key)
         last = strlen(Memc[buff])
         if (index > 0)
         {
            # pull out new value (add 2 to skip over "= ")
            call substr(Memc[buff],index+2,last,value,maxvalue)

            break
         }            
      }

   }

   call sfree(sp)

end
#---------------------------------------------
# copy instr[index1:index2] to outstr
procedure substr(instr,index1,index2,outstr,maxstr)
char instr[ARB]
int  index1
int  index2
char outstr[ARB]
int  maxstr

int i
int j

begin

   j=1
   for (i=index1; i<=index2; i=i+1)
   {
      outstr[j]=instr[i]
      j=j+1
      if (j >= maxstr)
         break
   }
   outstr[j]=EOS
end
#------------------------------------------------
#int procedure getpath (vfn, outstr, maxch)
#
#char    vfn[ARB], outstr[maxch]
#int     maxch
#int     root_offset, extn_offset, nchars_path
#int     gstrcpy()
#
#begin
#        call zfnbrk (vfn, root_offset, extn_offset)
#        nchars_path = max(0, min(maxch, root_offset-1))
#
#        return (gstrcpy (vfn[1], outstr, nchars_path))
#end
