# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_qprst.x
# 	Author	:  tnk
# 	Date	:  08/24/93
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_qprst.x -r1.4 08/24/93 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/13/97   mabbott	removed unneeded qpoe$ includes
# 	08/09/94   tnk	renamed wcs parm
# 	08/08/94   tnk	doing work-arounds for portability problems
# 	08/24/93   tnk	date and time created 93/08/24 15:04:40 by tnk
# H-
# 
#

include <iraf.h>

# @(#)t_qprst.x	1.4 10/13/97 UCB CEA EGO

# convert ST tables into a QPOE file
#
procedure t_qprst()

pointer qp                  # QPOE file pointer
pointer qpfname             # QPOE file name 
pointer tabprefix           # ST table filename prefix
pointer tabdir              # ST table directory
pointer list                # list of filenames
pointer template            # filename template
pointer sp                  # stack pointer
pointer parmfile_name       # name of parameter file
pointer pre_parmfile_name   # name of parameter file with prefix
bool    do_events           # flag: do / don't do events
bool    verbose             # verbose diagnostic output
bool    warnmsgs            # print warning messages
bool    mkindex             # make a qpoe index
bool    binary_wcs          # use binary wcs info
int     obucketlen          # override bucket length
int     nfiles              # no. of ST table files

# FUNCTIONS
pointer  fntopnb()   # expand template & open buffered filename list
pointer  qp_open()
bool     clgetb()   
int	 clgeti()
int      fntlenb()   # get number of filenames in a buffered list

define TABEXTN ".tab"

begin

   call smark(sp)

   call salloc(qpfname,   SZ_LINE, TY_CHAR)
   call salloc(tabdir,    SZ_LINE, TY_CHAR)
   call salloc(tabprefix, SZ_LINE, TY_CHAR)
   call salloc(parmfile_name, SZ_LINE, TY_CHAR)

   # get parameters
   call clgstr ("qpfilename",   Memc[qpfname], SZ_LINE)       #output file
   call clgstr ("tabdir",    Memc[tabdir], SZ_LINE)      #tables directory
   call clgstr ("tabprefix", Memc[tabprefix], SZ_LINE)   #tables prefix
   call clgstr ("parmtable", Memc[parmfile_name], SZ_LINE)   #parm table name
   verbose    = clgetb("verbose")
   warnmsgs   = clgetb("warnmsgs")
   mkindex    = clgetb("mkindex")
   obucketlen  = clgeti("bucketlen")
   binary_wcs = clgetb("wcsbin")

   # open the output qpoe file
   qp = qp_open (Memc[qpfname], NEW_FILE, 0)

   # get the QP parameters that were saved in the parmfile_name ST table
   call loadparms(qp, Memc[tabdir], Memc[tabprefix], Memc[parmfile_name],
      warnmsgs)

   # make a file template with the table prefix, and table extension
   call salloc (template,        SZ_LINE,        TY_CHAR)
   call strcpy (Memc[tabdir],    Memc[template], SZ_LINE)
   call strcat ("/",             Memc[template], SZ_LINE)
   call strcat (Memc[tabprefix], Memc[template], SZ_LINE)
   call strcat ("*",             Memc[template], SZ_LINE)
   call strcat (TABEXTN,         Memc[template], SZ_LINE)

   # step through files in the tables directory, and for each
   # file make a new entry in the QPOE file header (and symbol table)

   # add tabprefix to root filename of "parmfile_name" parameter
   call salloc(pre_parmfile_name, SZ_LINE, TY_CHAR)
   call strcpy(Memc[tabprefix], Memc[pre_parmfile_name], SZ_LINE)
   call strcat(Memc[parmfile_name], Memc[pre_parmfile_name], SZ_LINE)

   list = fntopnb(Memc[template],YES)        # open the list of files

   nfiles = fntlenb(list)
   if (nfiles <= 0) {
      call error(101,"No ST table files found")
   }

   # make 2 passes thru the data.  first do the non-event list
   # tables, then the event lists after everything is set up

   do_events = false
   call process_files(list, qp, Memc[pre_parmfile_name], do_events, verbose,
      warnmsgs,mkindex,obucketlen,binary_wcs)

   # rewind file list
   call fntrewb(list)

   do_events = true
   call process_files(list, qp, Memc[pre_parmfile_name], do_events, verbose,
      warnmsgs,mkindex,obucketlen,binary_wcs)

   call fntclsb(list)   # close the list
   call qp_close (qp)

   call sfree(sp)
end

#*************************************************
# process ST table files, loading them into the QPOE file
#
procedure process_files(list, qp, parmfile_name, do_events, verbose,
   warnmsgs,mkindex,obucketlen,binary_wcs)
pointer list          #I file list (ST tables)
pointer qp            #I/O qp file
char    parmfile_name #I name of special parameter file
bool    do_events     #I flag indicating whether to process event lists
bool    verbose       #I print informational messages
bool    warnmsgs      #I print warning messages
bool    mkindex       #I make a qpoe index
bool    binary_wcs    #I use binary wcs info
int     obucketlen    #I override bucket length

# LOCAL VARIABLES
pointer sp
int     fnum
int     nfiles
pointer fname
pointer root
int     status
int     len

# FUNCTIONS
int      fnroot()
int      fntlenb()   # get number of filenames in a buffered list
pointer  fntgfnb()   # get next filename from buffered list (sequential)
bool     streq()

begin

   call smark(sp)

   nfiles = fntlenb(list)

   call salloc(fname,   SZ_LINE, TY_CHAR)
   call salloc(root,   SZ_LINE, TY_CHAR)

   for (fnum=1; fnum<= nfiles; fnum=fnum+1){

      status = fntgfnb(list, Memc[fname], SZ_LINE)  # get next file

      len = fnroot(Memc[fname],Memc[root], SZ_LINE)

      # skip the "parmfile_name" parameters file already processed
      if (! streq(Memc[root],parmfile_name) ) {

         # add the table info to the qp file
         call process_tab(qp,Memc[fname],do_events,verbose, warnmsgs,
            mkindex,obucketlen,binary_wcs)
      }
   }

   call sfree(sp)

end
