# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  maketabs.x
# 	Author	:  tnk
# 	Date	:  08/24/93
# 	Synopsis:  
# 	Keywords:
# 		@(#) maketabs.x -r1.8 08/24/93 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/02/97   bboyd	removed private qpoe headers, now using qp_parseeuv in euv library
# 	10/24/96   tnk	incr MAX_USR_PARS
# 	09/13/95   tnk	set max table user params
# 	08/25/94   tnk	eliminated unused args
# 	08/09/94   tnk	adjusted comment
# 	11/23/93   tnk	now handles ST table shorts
# 	08/27/93   tnk	added WCS header cards to events tables
# 	08/24/93   tnk	date and time created 93/08/24 15:04:03 by tnk
# H-
# 
#

include <tbset.h>
include <ctype.h>
include <iraf.h>
include <qpset.h>
include <imhdr.h>
include	<qpioset.h>
include <qpeuv.h>
include <tbset.h>

# @(#)maketabs.x	1.8 10/2/97 UCB CEA EGO

# QPOE-related ST table header cards

define QPSYMBOL_HDR "QPSYMBOL"
define QPDTYPE_HDR  "QPDTYPE"
define QPCMNT_HDR   "QPCMNT"
define QPMAXEL_HDR  "QPMAXEL"
define QPFLDLST_HDR "QPFLDLST"
define QPFLCMNT_HDR "QPFLCMNT"
define QPBUCKLN_HDR "QPBUCKLN"

define MAX_USR_PARS   120       # max. number of user parameters in table

# create a 1d (1 column) ST table from a QPOE file entry
#
procedure mk_1d_table(destdir,tabprefix,parmname,qpdtype,nrows,maxelem,
   comment,qp,spp_dtype)  #LOCAL
char    destdir[ARB]    #I destination directory
char    tabprefix[ARB]  #I table name prefix
char    parmname[ARB]   #I name of parameter
char    qpdtype[ARB]    #I type of data column
int     nrows           #I number of rows
int     maxelem         #I max number of qp elements
char    comment[ARB]    #I QP comment
pointer qp              #I QP file pointer
int     spp_dtype       #I spp data type

# LOCAL VARS
pointer out_tp          # output table pointer
pointer tp_colptr       # column pointer

begin


   call open_new_tab(destdir,tabprefix,parmname,out_tp)

   call prep_1d_table(out_tp,tp_colptr,parmname,qpdtype,nrows,maxelem,
      comment,spp_dtype)

   call fill_1d_table(out_tp,tp_colptr,parmname,qpdtype,qp,nrows,spp_dtype)

   call tbtclo (out_tp)

end

#****************************************************
# create a 2d (multi-column) ST table from a QPOE file entry
#
procedure mk_2d_table(destdir,tabprefix,parmname,qpdtype,nrows,maxelem,
  comment,qp,descstring,desccmnt,spp_dtype,filter,verbose,im)
char    destdir[ARB]    #I destination directory
char    tabprefix[ARB]  #I table name prefix
char    parmname[ARB]   #I name of table
char    qpdtype[ARB]    #I type of data column
int     nrows           #I number of rows
int     maxelem         #I max number of qp elements
char    comment[ARB]    #I QP comment
pointer qp              #I QP file pointer
char    descstring[ARB] #I structure descriptor string
char    desccmnt[ARB]   #I structure descriptor comment
int     spp_dtype       #I spp data type
char    filter[ARB]     #I qpoe filter (user expression)
bool    verbose         #I verbose flag for debug messages
pointer im              #I image pointer to QPOE file

# LOCAL VARS
pointer dd              # pointer to domain descriptor
pointer sp
int     nfields
int     i
pointer out_tp          # output table pointer
pointer tp_colptrs
int     n_cols
pointer st_dtypes       # ST compatible data types


#FUNCTIONS
int qp_parseeuv()   #Parse the field list (declarations string) for user type

begin

   call smark(sp)

   call open_new_tab(destdir,tabprefix,parmname,out_tp)

   # parse structure description string
   call salloc(dd, LEN_EUVDDDES, TY_STRUCT)
   nfields = qp_parseeuv(qp, descstring, dd)
   n_cols=EUVDD_NFIELDS(dd)

   # convert QP datatypes to ST compatible data types
   call salloc(tp_colptrs, n_cols, TY_POINTER)
   call salloc(st_dtypes, n_cols, TY_INT)
   for (i=1; i<= n_cols; i=i+1)
      call spp2stdtype(EUVDD_FTYPE(dd,i),Memi[st_dtypes+i-1])

   call prep_2d_table(out_tp,Memi[tp_colptrs],n_cols,Memi[st_dtypes],parmname,
      qpdtype,maxelem,comment,spp_dtype)

   # put out field list description string & comment
   call tbhadt (out_tp, QPFLDLST_HDR, descstring)
   call tbhadt (out_tp, QPFLCMNT_HDR, desccmnt)

   # see if this is an event list, ie, if
   # the x and y index fields are nonzero
   # and process accordingly

   if (EUVDD_XFIELD(dd) != 0 && EUVDD_YFIELD(dd) != 0) {
      # an event list

      # put out WCS related header cards from QPOE (as image) to ST table
      call cp_wcs_hdrs(out_tp,im)

      call fill_ev_2d_table(out_tp,Memi[tp_colptrs],n_cols,qp,dd,
         Memi[st_dtypes],filter,verbose)
   }
   else {
      # a user-defined (non-event) list

      call fill_user_2d_table(out_tp,Memi[tp_colptrs],n_rows,n_cols,qp,dd,
         Memi[st_dtypes],parmname,qpdtype,verbose)
   }

   call tbtclo (out_tp)           # close output file

   call sfree(sp)

end


#****************************************************
# open up an ST table in the given directory
#
procedure open_new_tab(dir,tabprefix,parmname,tp)    #LOCAL

char    dir[ARB]        #I directory name
char    tabprefix[ARB]  #I filename prefix
char    parmname[ARB]   #I parameter name
pointer tp              #O pointer to opened table

# LOCAL VARS
pointer sp              #stack pointer
pointer str             #string buffer
pointer newparm

# FUNCTIONS
pointer  tbtopn()
int      strlen()
int      i

begin

  call smark(sp)

  call salloc (str, SZ_LINE, TY_CHAR)
  call salloc (newparm, SZ_LINE, TY_CHAR)

  #convert hyphens to underscores to take care of iraf pathname bug
  call strcpy (parmname, Memc[newparm], SZ_LINE)
  for (i=0; i < strlen(Memc[newparm]); i=i+1)
     if ( Memc[newparm+i] == '-' )
           Memc[newparm+i] = '_' 

  # Open table 
  # concatentate directory,filename prefix, and filename

  call strcpy(dir, Memc[str], SZ_LINE)
  call strcat("/", Memc[str], SZ_LINE)
  call strcat (tabprefix, Memc[str], SZ_LINE)
  call strcat (Memc[newparm], Memc[str], SZ_LINE)
  tp = tbtopn (Memc[str], NEW_FILE, 0)

  call sfree(sp)

end

#*****************************************
# prepare a 1d ST table
procedure prep_1d_table(out_tp,out_col,parmname,qpdtype,nrows,maxelem,
   comment,spp_dtype) #LOCAL
pointer out_tp              #I/O pointer to opened output table
pointer out_col             #I output column
char    parmname[ARB]       #I name of QP parameter (symbol)
char    qpdtype[ARB]        #I type of data column (QP string)
int     nrows               #I number of table rows to allocate
int     maxelem             #I max number of QP elements
char    comment[ARB]        #I QP comment
int     spp_dtype           #I spp data type (TY_CHAR, etc)

# LOCAL VARS
pointer str                 # string buffer
pointer sp                  # stack pointer
int     st_dtype


begin

  call smark (sp)

  call salloc (str, SZ_LINE, TY_CHAR)

  # specify no. of rows (don't really have to do this,
  # since table is row ordered)

  call tbpset (out_tp, TBL_ALLROWS, nrows)           # no. of rows

  call spp2stdtype(spp_dtype, st_dtype)

  # define column
  if (st_dtype == TY_CHAR) {
     st_dtype = -1  # make column 1 char wide
  }

  call tbcdef (out_tp, out_col, parmname, "", "", st_dtype, 1, 1)

  # set max no. of user params
  call tbpset (out_tp, TBL_MAXPAR, MAX_USR_PARS)

  call tbtcre (out_tp)

  # put in QP related header cards
  call mk_qp_cards(out_tp,parmname,qpdtype,comment,maxelem,spp_dtype)

  call sfree (sp)

end
#*****************************************
# prepare a 2d ST table
#
procedure prep_2d_table(out_tp,out_cols,n_cols,col_dtypes,parmname,qpdtype,
   maxelem,comment,spp_dtype) #LOCAL
pointer out_tp              #I/O pointer to opened output table
pointer out_cols[ARB]       #I/O output column pointers
int     n_cols              #I number of columns
int     col_dtypes[ARB]     #I datatypes of columns
char    parmname[ARB]       #I name of QP parameter (symbol)
char    qpdtype[ARB]        #I type of data column (QP string)
int     maxelem             #I max number of QP elements
char    comment[ARB]        #I QP comment
int     spp_dtype           #I spp data type (TY_CHAR, etc)

# LOCAL VARS
pointer str                 # string buffer
pointer sp                  # stack pointer
int     i                   # loop index
pointer colname

begin

  call smark (sp)

  call salloc (str, SZ_LINE, TY_CHAR)

  # create columns one at a time
  call salloc (colname, SZ_COLNAME, TY_CHAR)
  for (i=1; i<=n_cols; i=i+1) {
     call sprintf(Memc[colname],SZ_COLNAME,"Column%d")
     call pargi(i)
     call tbcdef (out_tp, out_cols[i], Memc[colname], "", "", col_dtypes[i],
        0, 1)
  }

  # set max no. of user params
  call tbpset (out_tp, TBL_MAXPAR, MAX_USR_PARS)

  call tbtcre (out_tp)

  # put in QP related header cards
  call mk_qp_cards(out_tp,parmname,qpdtype,comment,maxelem,spp_dtype)

  call sfree (sp)

end
#****************************************************
# put qp-related header cards into an ST table
procedure mk_qp_cards(out_tp,parmname,qpdtype,comment,maxelem,spp_dtype)
pointer out_tp          # I/O ST table pointer
char    parmname[ARB]   # QPOE parameter name
char    qpdtype[ARB]    # QPOE datatype string
char    comment[ARB]    # QPOE comment
int     maxelem         # max no. of QPOE elements
int     spp_dtype       # spp datatype

begin

   # put in header cards

   # QP parameter name (symbol name)
   call tbhadt (out_tp, QPSYMBOL_HDR, parmname)

   # QP data type
   call tbhadt (out_tp, QPDTYPE_HDR, qpdtype)

   # QP comment
   call tbhadt (out_tp, QPCMNT_HDR, comment)

   # QP max number of elements
   call tbhadi (out_tp, QPMAXEL_HDR, maxelem)

   # comment 
   call tbhadt (out_tp, 
     "COMMENT", "This table is one component of a QPOE file.")

   switch (spp_dtype) {

      case TY_BOOL, TY_CHAR, TY_SHORT, TY_INT, TY_LONG,
           TY_REAL, TY_DOUBLE, TY_OPAQUE:

           call tbhadt (out_tp, "COMMENT", "QPOE header")

      case TY_MACRO:

           call tbhadt (out_tp, "COMMENT", "QPOE macro")

      case TY_USER:

           call tbhadt (out_tp, "COMMENT", "QPOE user defined datatype")

      default:
           call eprintf("unknown data type encountered\n")

  }

end
#****************************************************
# add 1 column of QP data to an ST table
# note that this procedure slurps up the entire column at once into memory
#
procedure fill_1d_table(out_tp,tp_colptr,parmname,qpdtype,qp,nrows,spp_dtype)
pointer out_tp         #I/O output table pointer 
pointer tp_colptr      #I/O output column pointer
char    parmname[ARB]  #I name of QP parameter
char    qpdtype[ARB]   #I QP data type string
pointer qp             #I qp file pointer
int     nrows          #I no. of rows to add to table
int     spp_dtype      #I spp data type of column 

# LOCAL VARIABLES
int      nread
pointer  sp
pointer  buff
int      firstrow
int      lastrow
int      st_dtype
int      nelem
int      qp_read()

begin

   call smark(sp)

   nread=nrows
 
   call spp2stdtype(spp_dtype, st_dtype)
   call salloc(buff, nread, st_dtype)

   # put out an entire column
   firstrow=1
   lastrow=nread

   switch(spp_dtype) {

   case TY_BOOL:
       nelem = qp_read(qp, parmname, Memb[buff], nread, firstrow, qpdtype)
       call tbcptb(out_tp, tp_colptr, Memb[buff], firstrow, lastrow)

   case TY_CHAR, TY_MACRO:
       nelem = qp_read(qp, parmname, Memc[buff], nread, firstrow, qpdtype)
       call write_str_col(out_tp, tp_colptr, Memc[buff], nread, firstrow, 
       lastrow)

   case TY_SHORT:
       nelem = qp_read(qp, parmname, Mems[buff], nread, firstrow, qpdtype)
       call tbcpts(out_tp, tp_colptr, Mems[buff], firstrow, lastrow)

   case TY_INT, TY_LONG, TY_OPAQUE:
       nelem = qp_read(qp, parmname, Memi[buff], nread, firstrow, qpdtype)
       call tbcpti(out_tp, tp_colptr, Memi[buff], firstrow, lastrow)

   case TY_REAL:
       nelem = qp_read(qp, parmname, Memr[buff], nread, firstrow, qpdtype)
       call tbcptr(out_tp, tp_colptr, Memr[buff], firstrow, lastrow)

   case TY_DOUBLE:
       nelem = qp_read(qp, parmname, Memd[buff], nread, firstrow, qpdtype)
       call tbcptd(out_tp, tp_colptr, Memd[buff], firstrow, lastrow)

   default:
      call eprintf("can't handle data type: %d %s\n")
      call pargi(spp_dtype)
      call pargstr(qpdtype)
   }
                       
   call sfree(sp)

end
#****************************************************
# convert QPOE/spp datatype to ST compatible datatype
procedure spp2stdtype(spp_dtype,st_dtype)
int spp_dtype       #I QPOE datatype string
int st_dtype        #O SPP datatype value

begin

switch(spp_dtype) {
   case TY_BOOL:
                st_dtype = TY_BOOL
   case TY_CHAR, TY_MACRO:
                st_dtype = TY_CHAR
   case TY_SHORT:
                st_dtype = TY_SHORT
   case TY_INT, TY_LONG, TY_OPAQUE:
                st_dtype = TY_INT
   case TY_REAL:
                st_dtype = TY_REAL
   case TY_DOUBLE:
                st_dtype = TY_DOUBLE
   case TY_COMPLEX:
                st_dtype = -1          # undefined
   default:   
                st_dtype = -1          # user defined
   }
end

#****************************************************
# write a string as an ST column, 1 char per row
#
procedure write_str_col(out_tp, tp_colptr, buff, bufflen, firstrow, lastrow)
pointer out_tp     #I/O output table pointer
pointer tp_colptr  #I/O column pointer
char    buff[ARB]  #I char buffer
int     bufflen    #I length of buff
int     firstrow   #I first row to store text in
int     lastrow    #I last row to store text in

char    line[2]    # one line of text (char + EOS)
int     i
int     row

begin

  line[2] = EOS

  # take each char, stuff it into a string buffer, write it out
  i = 1
  for (row=firstrow; row<=lastrow; row=row+1) {
     line[1]=buff[i]
     call tbeptt(out_tp, tp_colptr, row, line)
     i=i+1
  }

end
#****************************************************
# copy qpoe events to an ST table, using qpio routines
#
procedure fill_ev_2d_table(out_tp,tp_colptrs,n_cols,qp,dd,st_dtypes,
   filter,verbose)
pointer out_tp            #I/O pointer to opened output table
pointer tp_colptrs[ARB]   #I/O output column pointers
int     n_cols            #I number of columns
pointer qp                #I QP file pointer
pointer dd                #I pointer to domain descriptor
pointer st_dtypes[ARB]    #I ST compatible data types
char    filter[ARB]       #I QP filter (user expressions)
bool    verbose           #I verbose flag

# LOCAL VARS
int       ev_count
int       pct
int       oldpct
int	  maskval
pointer	  ev	      # event pointer
pointer	  sp
int	  nev
int       i
int       n_events
int       tab_row
int       offset
int       dtype
pointer   qpio        #I QP I/O struct pointer
int       bucketlen
pointer	  param

# FUNCTIONS
pointer qpio_open()
int	qpio_getevents()
int     qpio_stati()
int	qplenf()

begin

   call smark(sp)

   # pass in the filter to pick up user specified QPOE expressions
   qpio = qpio_open(qp, filter, READ_ONLY)

   # save BUCKETLEN as ST header card
   bucketlen =  qpio_stati(qpio, QPIO_BUCKETLEN)
   call tbhadi (out_tp, QPBUCKLN_HDR, bucketlen)

   param = qpio_stati(qpio, QPIO_PARAMP)
   n_events = qplenf(qp, Memc[param])

   if (verbose) {
      call eprintf("reading QPOE events...\n")
   }

   maskval = 0
   ev_count = 0
   oldpct = 0

   while (qpio_getevents(qpio, ev, maskval, 1, nev) != EOF) {

      ev_count = ev_count + 1

      if (verbose) {

         pct = int(10.*real(ev_count)/real(n_events))*10
         if (pct != oldpct) {
            oldpct = pct
            call eprintf ("  %3d%% read\n")
            call pargi (pct)
         }
      }

      # pull off each field from the ev struct & stuff it into
      #  the appropriate ST table column 

      tab_row = ev_count

      for (i=1; i<=n_cols; i=i+1) {

        offset = EUVDD_FOFFSET(dd,i)
        dtype  = EUVDD_FTYPE(dd,i)

        call qp2st_fld(ev,offset,dtype,out_tp,tp_colptrs[i],tab_row)
      }

   }

  call qpio_close(qpio)

  call sfree(sp)

end
#****************************************************
# copy user structure values to an ST table, using qp_read
#
procedure fill_user_2d_table(out_tp,tp_colptrs,n_rows,n_cols,qp,dd,st_dtypes,
   parmname,qpdtype,verbose)
pointer out_tp            #I/O pointer to opened output table
pointer tp_colptrs[ARB]   #I/O output column pointers
int     n_rows            #I number of rows
int     n_cols            #I number of columns
pointer qp                #I QP file pointer
pointer dd                #I pointer to domain descriptor
pointer st_dtypes[ARB]    #I ST compatible data types
char    parmname[ARB]     #I name of user parameter
char    qpdtype[ARB]      #I QPOE data type of user parameter
bool    verbose           #I verbose flag

# LOCAL VARS
int       elem_count
int       elem_len
int       pct
int       oldpct
int	  maskval
pointer   elem		# element pointer
pointer	  sp
int       i
int       n_elements
int       tab_row
int       offset
int       dtype

# FUNCTIONS
int       qp_read()
int       sizeof()

begin

   call smark(sp)

   n_elements = n_rows

   # figure out length of structure & allocate it
   elem_len = 0
   for (i=1; i<= EUVDD_NFIELDS(dd); i=i+1)
      elem_len = elem_len + sizeof(EUVDD_FTYPE(dd,i))
   call salloc (elem, elem_len / SZ_SHORT, TY_SHORT)

   if (verbose) {
      call eprintf("reading QPOE elements...\n")
   }

   maskval = 0
   elem_count = 1
   oldpct = 0

   # get each structure & put it into the ST table columns  
   while (qp_read(qp, parmname, Mems[elem], 1, elem_count, qpdtype) > 0)  {

      tab_row = elem_count

      if (verbose) {
         pct = int(10.*real(elem_count)/real(n_elements))*10
         if (pct != oldpct) {
            oldpct = pct
            call eprintf ("  %3d%% read\n")
            call pargi (pct)
         }
      }

      # pull off each field from the elem struct & stuff it into
      #  the appropriate ST table column 

      for (i=1; i<=n_cols; i=i+1) {

         offset = EUVDD_FOFFSET(dd,i)
         dtype  = EUVDD_FTYPE(dd,i)

         call qp2st_fld(elem,offset,dtype,out_tp,tp_colptrs[i],tab_row)

      }

      elem_count = elem_count + 1

   }

  call sfree(sp)

end
#**************************************************
# copy a field from a qpoe structure to an ST table column
#
procedure qp2st_fld(qp_struct,offset,dtype,tp,tp_colptr,tab_row)
pointer qp_struct   #I QPOE user structure
int     offset      #I field offset into structure
int     dtype       #I QPOE field datatype
pointer tp          #I/O table pointer
pointer tp_colptr   #I/O table column pointer
int     tab_row     #I table row to write into

# LOCAL VARS
bool   bool_val
int    int_val
real   real_val
double dbl_val
short  short_val

# FUNCTIONS
pointer coerce()

begin

   switch (dtype) {

      case TY_BOOL:
         bool_val = Memb[coerce(qp_struct,TY_SHORT,dtype) + offset]
         call tbeptb(tp,tp_colptr, tab_row, bool_val)

      case TY_SHORT:
         short_val = Mems[coerce(qp_struct,TY_SHORT,dtype) + offset]
         call tbepts(tp,tp_colptr, tab_row, short_val)

      case TY_INT:
         int_val = Memi[coerce(qp_struct,TY_SHORT,dtype) + offset]
         call tbepti(tp,tp_colptr, tab_row, int_val)

      case TY_LONG:
         int_val = int(Meml[coerce(qp_struct,TY_SHORT,dtype) + offset])
         call tbepti(tp,tp_colptr, tab_row, int_val)

      case TY_REAL:
         real_val =Memr[coerce(qp_struct,TY_SHORT,dtype) + offset]
         call tbeptr(tp,tp_colptr, tab_row, real_val)

      case TY_DOUBLE:
         dbl_val = Memd[coerce(qp_struct,TY_SHORT,dtype) + offset]
         call tbeptd(tp,tp_colptr, tab_row, dbl_val)

    default:
        call eprintf("can't handle datatype %d\n")
        call pargi(dtype)
        call error(100,"bye")
  }


end
