# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_qpwst.x
# 	Author	:  tnk
# 	Date	:  08/24/93
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_qpwst.x -r1.3 08/24/93 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/02/97   bboyd	removed private qpoe headers, now using qp_dtypeeuv in euv library
# 	08/27/93   tnk	added WCS header cards to events tables
# 	08/24/93   tnk	date and time created 93/08/24 15:04:05 by tnk
# H-
# 
#

include <iraf.h>
include <qpeuv.h>

# convert a QPOE file into a set of ST tables
#
# @(#)t_qpwst.x	1.3 10/2/97 UCB CEA EGO
#
# parameters are:
#   fname:      name of input qpoe file
#   destdir:    directory to place the ST tables
#   tabprefix:  filename prefix to prepend to all ST table names
#   parmtable:  name of special QPOE parameter table
#   verbose:    warm & fuzzy informational messages
#
#   For each element in the QPOE symbol table, create an ST table.
#   The name of each ST table is the symbol prepended with "tabprefix".

procedure t_qpwst()

pointer qp         # qpoe file pointer
pointer list       # list of symbols in qpoe table
int     nelem      # no. of elements in a qpoe entry
int     maxelem    # max elements allowed in a qpoe entry (determines lfile)
int     flags      # qpoe flags
pointer qpdatatype # qp datatype string (i, d, macro, events, etc.)
pointer comment    # qpoe symbol comment string
pointer parmtable  # name of special qpoe parameter file
pointer fname      # filename
pointer parmname   # parameter name
pointer inpath     # input path
pointer tabprefix  # prefix prepended to every output table name
pointer destdir    # directory path for output tables
pointer filter     # QPOE filter specification 
int     sppdtype   # spp data type (TY_INT, etc)
pointer descparm   # parameter name of user defined datatype descriptor
pointer descstring # user defined datatype decriptor string eg, {d,d,i,i}
int     descmax    # max no. of elements in descriptor string
int     descnelem  # no. of elements in descriptor string
pointer desccmnt   # comment acommpanying descriptor string
pointer sp         # stack pointer
bool    verbose    # verbose flag (diagnostics/ warm & fuzzies)
pointer im         # image pointer

# FUNCTIONS
pointer qp_open()     # open qpoe file
pointer qp_ofnlu()    # Open an unsorted field name list
int     qp_queryf()   # Get information describing the named parameter
int     qp_gnfn()     # Return the next element from the field name list
int     qp_dtypeeuv() # Translate symbolic datatype name into int type code.
pointer immap()       # open an image
bool    clgetb()      # get boolean value


begin

   call smark(sp)

   call salloc(inpath,    SZ_PATHNAME, TY_CHAR)
   call salloc(destdir,   SZ_PATHNAME, TY_CHAR)
   call salloc(tabprefix, SZ_FNAME,    TY_CHAR)
   call salloc(parmtable, SZ_FNAME,    TY_CHAR)

   # get parameters
   call clgstr ("qpfilename", Memc[inpath],    SZ_PATHNAME)
   call clgstr ("tabdir",     Memc[destdir],   SZ_PATHNAME)
   call clgstr ("tabprefix",  Memc[tabprefix], SZ_FNAME)
   call clgstr ("parmtable",  Memc[parmtable], SZ_FNAME)
   verbose = clgetb("verbose")

   # split full qppath into filename & filter specification
   call salloc(fname, SZ_PATHNAME, TY_CHAR)
   call salloc(filter, SZ_PATHNAME, TY_CHAR)
   call qpparse (Memc[inpath], Memc[fname], SZ_PATHNAME, Memc[filter], SZ_LINE)

   # open the QPOE file
   qp = qp_open (Memc[fname], READ_ONLY, 0)

   # also open it as an image (to get header info later)
   im = immap (Memc[fname], READ_ONLY, 0)


   # save off general, QPOE wide parameters into special table
   call saveparms(qp, Memc[destdir], Memc[tabprefix], Memc[parmtable])

   call salloc(parmname,   SZ_FNAME, TY_CHAR)
   call salloc(qpdatatype, SZ_FNAME, TY_CHAR)
   call salloc(comment,    SZ_LINE,  TY_CHAR)
   call salloc(descparm,   SZ_LINE,  TY_CHAR)
   call salloc(desccmnt,   SZ_LINE,  TY_CHAR)
   call salloc(descstring, SZ_LINE,  TY_CHAR)

   # open QPOE header (symbol table) list
   # step through each entry
   list = qp_ofnlu (qp, "*")

   while (qp_gnfn (list, Memc[parmname], SZ_FNAME) != EOF) {

      nelem = qp_queryf (qp, Memc[parmname], Memc[qpdatatype], maxelem,
         Memc[comment], flags)
 
      sppdtype = qp_dtypeeuv(qp, Memc[qpdatatype])

      # build an appropriate table, depending on the spp data type
      switch (sppdtype){

         case TY_BOOL, TY_CHAR, TY_SHORT, TY_INT, TY_LONG,
              TY_REAL, TY_DOUBLE, TY_MACRO, TY_OPAQUE:

              if (verbose) {
                 call printf("Creating 1d table:  name=%15s type=%10s nrows=%9d maxelem=%d\n")
                 call pargstr(Memc[parmname])
                 call pargstr(Memc[qpdatatype])
                 call pargi(nelem)
                 call pargi(maxelem)
              }

            call mk_1d_table(Memc[destdir], Memc[tabprefix], Memc[parmname],
               Memc[qpdatatype], nelem, maxelem, Memc[comment], qp, sppdtype)
           
         case TY_COMPLEX:  
             # can't handle complex types
             if (verbose) {
                call eprintf("Skipping complex datatype parameter '%s'\n")
                call pargstr(Memc[parmname])
             }

         case TY_USER:

             # user defined data type

             # both the event structure descriptor string
             # eg, event='{d,s:x,s:y,r,r,s,s}'
             # and the event structure list itself,
             # eg, events=event[100]
             # get marked as TY_USER

             # skip the structure descriptor string
             # (we pick it up again inside the code below, 
             # once we find the event structure list).

             if ( Memc[qpdatatype+0] == '{' ) { 
                 # call eprintf("skipping descriptor: %s\n")
                 # call pargstr(Memc[parmname])
             }
             else {

                # get the associated descriptor (ie., this data type
                # is really the same of another symbol)

                if (verbose) {
                   call printf("Creating 2d table:  name=%15s type=%10s nrows=%9d maxelem=%d\n")
                   call pargstr(Memc[parmname])
                   call pargstr(Memc[qpdatatype])
                   call pargi(nelem)
                   call pargi(maxelem)
                }

                call strcpy (Memc[qpdatatype], Memc[descparm], SZ_LINE)
                descnelem = qp_queryf (qp, Memc[descparm], Memc[descstring],
                   descmax, Memc[desccmnt], flags)

                if (verbose) {
                   call printf("descriptor %s: %s\n")
                   call pargstr(Memc[descparm])
                   call pargstr(Memc[descstring])
                }

                # put the info in a 2d ST table 
                call mk_2d_table(Memc[destdir], Memc[tabprefix], 
                    Memc[parmname], Memc[qpdatatype], nelem,
                    maxelem, Memc[comment], qp, Memc[descstring],
                    Memc[desccmnt], sppdtype, Memc[filter],verbose,im)
            }

         default:
              call eprintf("****  bad datatype!!!!! ***\n")
      }

   }      

   call qp_cfnl (list)

   call imunmap (im)

   call qp_close (qp)

   call sfree(sp)

end

