# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  spcopen.x
# 	Author	:  akeith
# 	Date	:  06/27/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) spcopen.x -r1.3 06/27/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	09/21/94   akeith	now uses eu_clfile()
# 	09/14/94   akeith	reworked to use the spc library
# 	06/27/94   akeith	date and time created 94/06/27 15:04:24 by akeith
# H-
# 
#


#	Version: @(#)spcopen.x	1.3 9/21/94 UCB CEA EGO

include	"so.h"
include	"spc.h"
include	<dfset.h>


#
# spc_open()
#

pointer	procedure	spc_open(detector, reference, exposure)

char	detector[ARB]		#I detector name
char	reference[ARB]		#I reference file
char	exposure[ARB]		#I exposure image header

int	npts			#

pointer	spc			#

double	xl, xh			#

real	imgetr()		#
pointer	immap()			#
pointer	spc_gspc()		#
pointer	spc_init()		#

begin

#    Initialize SPC_ structure to zero.

	spc = spc_init()


#    Get the spectrum, if null leave.

	call so_gimname(detector, SPC_NAME(spc))
	if(SPC_NAME(spc) == EOS) {
		return(spc)
	}

	iferr(SPC_IM(spc) = immap(SPC_NAME(spc), READ_ONLY, NULL)) {
		SPC_IM(spc) = NULL
		return(spc)
	}

	SPC_ORG(spc) = spc_gspc(SPC_IM(spc))
	if(SPC_ORG(spc) == NULL) return(spc)

#    Get the exposure time.

	iferr(SPC_EXPT(spc) = imgetr(SPC_IM(spc), exposure)) {
		call spc_close(spc)
		call eprintf("suborder: unable to find exposer header, %s\n")
		call pargstr(exposure)
		call error(1, "Bad header name.")
	}
	SPC_EXISTS(spc) = TRUE
	SPC_ON(spc) = TRUE


#    Get the spectral error.
#	ORG_ERR(spc) = spc_gerr(SPC_ORG(spc), SPC_IM(spc))

#    Get the effective areas as pf type df's.

	call spc_gea(reference, detector, SPC_EA(spc,1), SPC_EA_LEN, SPC_NORDS(spc))

#    Get the bin width.

	call so_gdelta(detector, SPC_DELTA(spc))

#    Get the initial spectral range.

	call so_clgrng(detector, xl, xh)
	call so_chkrng(SPC_ORG(spc), xl, xh)

	npts = int((xh - xl)/SPC_DELTA(spc))
	SPC_SL(spc) = xl
	SPC_SH(spc) = xl + npts*SPC_DELTA(spc)

#    Create the new spectrum.

	call so_newdf(SPC_NEW(spc))
	call df_rsdomain(ORG(spc), SPC_SL(spc), SPC_DELTA(spc), npts, NEW(spc))
	SPC_NEW_ON(spc) = TRUE


#    Create the new error and order err spectrum.
#	call so_newdf(NEW_ERR(spc))
#	call df_rsdomain(ORG_ERR(spc), SPC_SL(spc), SPC_DELTA(spc), npts, NEW_ERR(spc))
#	call so_newdf(ORD_ERR(spc))


#    Create the order and leave it empty.

	call so_newdf(SPC_ORD(spc))
	SPC_OL(spc) = SPC_SL(spc)
	SPC_OH(spc) = SPC_SH(spc)


	return(spc)

end


#
# spc_init()
#

pointer	procedure	spc_init()

int	i
pointer	spc			#

begin

	call spc_malloc(spc)

	SPC_IM(spc) = NULL
	call malloc(SPC_NAME_PTR(spc), SZ_LINE, TY_CHAR)

	SPC_EXISTS(spc) = FALSE
	SPC_ON(spc) = FALSE

	ORG(spc) = NULL
	NEW(spc) = NULL
	ORD(spc) = NULL

	ORG_ERR(spc) = NULL
	NEW_ERR(spc) = NULL
	ORD_ERR(spc) = NULL

	ORG_ON(spc) = FALSE
	NEW_ON(spc) = FALSE
	ORD_ON(spc) = FALSE

	SPC_NORDS(spc) = 0

	do i = 1, SPC_EA_LEN
		SPC_EA(spc, i) = NULL

	SPC_XL(spc) = 0.0
	SPC_XH(spc) = 0.0
	SPC_YL(spc) = 0.0
	SPC_YH(spc) = 0.0

	SPC_SL(spc) = 0.0d0
	SPC_SH(spc) = 0.0d0

	SPC_OL(spc) = 0.0d0
	SPC_OH(spc) = 0.0d0

	SPC_DELTA(spc) = 0.0d0
	SPC_EXPT(spc) = 0.0d0

	return spc
end


#
# so_clgrng() gets the ranges from the cl
#

procedure	so_clgrng(detector, xl, xh)

char	detector[ARB]			#I
double	xl				#O
double	xh				#O

pointer	sp, param_name			#
double	clgetd()			#

begin

	call smark(sp)
	call salloc(param_name, SZ_LINE, TY_CHAR)

	call strcpy(detector, Memc[param_name], SZ_LINE)
	call strcat("_low", Memc[param_name], SZ_LINE)
	xl = clgetd(Memc[param_name])

	call strcpy(detector, Memc[param_name], SZ_LINE)
	call strcat("_high", Memc[param_name], SZ_LINE)
	xh = clgetd(Memc[param_name])

	call sfree(sp)

end


#
# so_newdf()
#

procedure	so_newdf(df)

pointer	df			#
pointer	df_open()

begin
	df = df_open(DF_PD, TY_DOUBLE, TY_DOUBLE)
	call df_pedge(df, 0.0d0, 0.0d0)
end


#
# so_chkrng()
#

procedure	so_chkrng(df, xl, xh)

pointer	df			#I
double	xl			#O
double	xh			#O

double	sl, sh

begin

	call df_gxrng(df, sl, sh)

	if(IS_INDEFD(xl)) xl = sl
	if(IS_INDEFD(xh)) xh = sh

	if(xl < sl) xl = sl
	if(xh > sh) xh = sh

end


#
# so_gimname()
#

procedure	so_gimname(detector, im_name)

char	detector[ARB]		#I
char	im_name[ARB]		#O

pointer	sp, str, ret		#

begin

	call smark(sp)
	call salloc(str, SZ_LINE, TY_CHAR)
	call strcpy(detector, Memc[str], SZ_LINE)
	call strcat("_in", Memc[str], SZ_LINE)
	call eu_clfile(ret, Memc[str])
	call strcpy(Memc[ret], im_name, SZ_LINE)
	call sfree(sp)

end


#
# so_gdelta()
#

procedure	so_gdelta(det, delta)

char	det[ARB]			#I
double	delta				#I

pointer	sp, str
double	clgetd()

begin

	call smark(sp)
	call salloc(str, SZ_LINE, TY_CHAR)
	call strcpy(det, Memc[str], SZ_LINE)
	call strcat("_delta", Memc[str], SZ_LINE)

	delta = clgetd(Memc[str])
	call sfree(sp)

end

